/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999 Emmanuel CURIS
   Cration de ce fichier : 2 avril 1999

   Modle structural tudi
   Gestion des absorbeurs (ajout, suppression, dition,...)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"
#include "atomes.h"
#include "feff.h"    /* Pour les modles FEFF */
#include "modeles.h" /* Pour les divers types de modles */
#include "absorption.h" /* Pour les types de seuils */
#include "interface.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define MAX_NOM_ABSORBEUR_VISIBLE 19 /* A modifier aussi dans molecules.c */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Ajouter un absorbeur  une molcule ----------------------------------\
 |   Renvoie l'absorbeur cr s'il a pu tre cr, NULL en cas d'erreur           |
 |    Cet absorbeur ne contient aucun modle : il faut le crer aprs             |
 |                                                                                |
 | atome        : l'atome qui doit tre utilis comme absorbeur                   |
 | nom          : le nom de l'absorbeur (NULL => 'sans nom')                      |
 | contribution : au signal EXAFS total de la molcule tudie pour cet lment   |
 |                (exprime en % du signal total, mais stocke en fraction [0,1]) |
 \------------------------------------------------------------------------------- */

Absorbeur *creer_absorbeur(Atome *atome, const char *nom, double contribution)
{
  Absorbeur *absorbeur;
  Molecule *molecule;

  /* Contrles */
  if (NULL == atome) return NULL;

  molecule = quelle_molecule(atome, NULL);
  if (NULL == molecule) {
    (void) fprintf(stderr, "creer_absorbeur : atome en-dehors des molcules ! Cration annule\n");
    return NULL;
  }

  /* Si dj absorbeur : on renvoie NULL */
  if ( VRAI == atome->marques.absorbeur ) {
    (void) fprintf( stderr, "%s,%s l.%u :  atome dj marqu comme absorbeur ! Cration annule\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return NULL;
  }


  /* On ajoute cet absorbeur  la fin de la table des absorbeurs de la molcule */
  if (molecule->absorbeurs != NULL) {
    absorbeur = (Absorbeur *) realloc(molecule->absorbeurs,
				      sizeof(Absorbeur) * (molecule->nbr_absorbeurs + 1) );
    if (absorbeur == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/absorbeur.c", "creer_absorbeur", "nv_absorbeur");
      return NULL;
    }
    molecule->absorbeurs = absorbeur; /* Nouvelle adresse des absorbeurs */
    absorbeur += molecule->nbr_absorbeurs; /* Le nouvel absorbeur est le dernier cr */
  } else {
    absorbeur = (Absorbeur *) malloc(sizeof(Absorbeur));
    if (absorbeur == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/absorbeur.c", "creer_absorbeur", "nv_absorbeur");
      return NULL;
    }
    molecule->absorbeurs = absorbeur; /* Premier cr dans la molcule */
  }
  molecule->nbr_absorbeurs++; /* Un absorbeur de plus */
  
  /* On initialise cet absorbeur */
  if (nom == NULL) nom = "Absorbeur sans nom";
  strncpy(absorbeur->nom, nom, LG_MAX_NOM_ABSORBEUR);
  absorbeur->nom[LG_MAX_NOM_ABSORBEUR] = 0; /* De toute faon */

  /* Atome utilis comme absorbeur */
  absorbeur->absorbeur = atome;
  atome->marques.absorbeur = VRAI;

  /* Aucun modle associ  cet absorbeur */
  absorbeur->nbr_modeles = 0;
  absorbeur->modeles = NULL;
  absorbeur->modele_actif = NULL;

  /* Contribution au signal total de la molcule */
  contribution /= 100.0;
  absorbeur->contribution.valeur_initiale = contribution; /* Par dfaut, constante connue */
  absorbeur->contribution.min = contribution;
  absorbeur->contribution.max = contribution;
  absorbeur->contribution.delta = 0;
  absorbeur->contribution.type = VAR_CONSTANTE;
  absorbeur->contribution.contenu.valeur = contribution;

  if (mol_courante == molecule) {
    MOL_PrepareListe();
  } /* Rq. MOL_aller ne fonctionne pas car molecule = molecule->courante : revient immdiatement ! */

  return absorbeur; /* Tout s'est bien pass */
}

int MOL_RenommerAbsorbeur(Absorbeur *absorbeur, const char *nouveau_nom)
{
  if ( NULL == absorbeur ) return 0;
  if ( NULL == nouveau_nom ) return -1;

  strncpy( absorbeur->nom, nouveau_nom, LG_MAX_NOM_ABSORBEUR );
  return 0;
}

/* -------- Propose de nouveaux absorbeurs pour une molcule -------- */

Absorbeur *nouvel_absorbeur(Molecule *molecule)
{
  Absorbeur *absorbeur;
  Atome *atm_absorbeur;

  if (molecule == NULL) return NULL;

  atm_absorbeur = ABS_ProposerAbsorbeur(molecule, VRAI);
  if (atm_absorbeur == NULL) return NULL; /* Aucun absorbeur propos par l'utilisateur */

  if (atm_absorbeur->marques.absorbeur == VRAI) {
    form_alert(1, "[2][---- Op"e_AIGU"ration interdite ----"
	             "| Cet atome est d"e_AIGU"j"a_GRAVE" utilis"e_AIGU
	             "|comme absorbeur dans une"
	             "|mol"e_AIGU"cule.][ Annuler ]");
    return NULL;
  }

  absorbeur = creer_absorbeur(atm_absorbeur, ATM_Z_vers_nom(atm_absorbeur->num_atomique), 100.0);
  if (absorbeur == NULL) return NULL;

  I_MOL_EditerAbsorbeur( absorbeur );

  /* On affiche ce qui correspond  cet absorbeur */
  if ( mol_courante == molecule ) {
    ABS_aller(absorbeur, VRAI);
  }

  return absorbeur;
}

/* -~-~-~-~-~-~-~-~ Gestion des modles d'un absorbeur -~-~-~-~-~-~-~-~ */

/* -------- Ajouter un modle existant  un absorbeur -------- */

void ABS_AjouterModele(Absorbeur *absorbeur, struct modele_ajuste *modele)
{
  if (NULL == absorbeur) return;
  if (NULL == modele) return;

  if (modele->suivant != NULL) {
    (void) fprintf(stderr, "ABS_AjouterModele : impossible d'ajouter un modle dj inclus dans un chanage\n");
    return;
  }

  absorbeur->nbr_modeles++; /* Un modle de plus */
  if (NULL == absorbeur->modeles) { /* On cre le premier modle */
    absorbeur->modeles = modele;
    absorbeur->modele_actif = modele;
  } else { /* On l'ajoute  la fin */
    struct modele_ajuste *avant;

    avant = absorbeur->modeles;
    while (avant->suivant != NULL)
      avant = avant->suivant;

    avant->suivant = modele;
  }
}

/* -------- Crer un modle pour un absorbeur -------- */

struct modele_ajuste *ABS_CreerModele(Absorbeur *absorbeur, int type, struct graphe *experience)
{
  struct modele_ajuste *modele;
  char *nom;

  if ( NULL == absorbeur ) return NULL;

  nom = TXT_genere_chaine( "Mod"e_GRAVE"le pour %s", absorbeur->nom );
  if ( NULL == absorbeur->modeles ) { /* Aucun modle pour l'instant : on cre le premier */
    modele = MODELE_CreerModele( NULL, nom, type, experience, VRAI );
    absorbeur->modeles = modele;
    absorbeur->modele_actif = modele;
  } else {
    modele = absorbeur->modeles;
    while (modele->suivant != NULL)
      modele = modele->suivant;
    modele = MODELE_CreerModele( modele, nom, type, experience, VRAI );
  }
  free( nom ); nom = NULL;
  if ( NULL == modele ) return NULL;

  absorbeur->nbr_modeles++; /* Un modle de plus */

  if ( MODELE_EXAFS_CHEMINS == type ) {
    struct modele_feff *feff;
    char *nom_molecule;
    Molecule *molecule;

    feff = modele->infos_modele.feff;

    /* On donne un nom raisonnable */
    molecule = quelle_molecule( absorbeur->absorbeur, NULL );
    if ( NULL == molecule ) {
      nom_molecule = "<<Probleme>>";
    } else {
      nom_molecule = molecule->nom;
    }
    feff->parametres.titre[ 0 ] = TXT_genere_chaine( "%s [%s]",
						     nom_molecule,
						     ATM_Z_vers_symbole( absorbeur->absorbeur->num_atomique ) );

    /* Atome absorbeur */
    if ( absorbeur->absorbeur->num_atomique >= IODE )
      feff->seuil = SEUIL_LIII; /* Seuils plus frquemment tudis (au LURE en tout cas)... */
    feff->absorbeur = absorbeur->absorbeur;
    feff->potentiels[ 0 ] = absorbeur->absorbeur->num_atomique;
  }

  return modele;
}

/* -------- Enlever un modle de l'absorbeur et l'effacer -------- */

int ABS_DetruireModele(Absorbeur *absorbeur, struct modele_ajuste *modele)
{
  if (NULL == absorbeur) return 0;
  if (NULL == modele) return -1;

  /* S'agit-il du modle actif ? */
  if (modele == absorbeur->modele_actif) {
    absorbeur->modele_actif = modele->suivant;
  }

  /* On efface le modle */
  if ( MODELE_EffacerModele( &(absorbeur->modeles), modele, FAUX ) != 0 ) {
    (void) fprintf(stderr, "ABS_DetruireModele : erreur en voulant dtruire le modle...\n");
    return -1;
  }

  absorbeur->nbr_modeles--; /* Un modle de moins */
  if ( (NULL == absorbeur->modele_actif) && ( absorbeur->modeles != NULL ) ) {
    /* On a effac le dernier de la liste, mais il en reste avant
       ==> le nouvel actif devient celui juste avant */
    for (absorbeur->modele_actif = absorbeur->modeles ; absorbeur->modele_actif->suivant != NULL ;
	 absorbeur->modele_actif = absorbeur->modele_actif->suivant);
  }

  return 0; /* Tout s'est bien pass */
}

/* -------- Cre une liste des absorbeurs plausibles d'une molcule et fait choisir l'utilisateur ---\
 |   Renvoie l'atome finalement choisi comme absorbeur (NULL en cas d'erreur)                        |
 |                                                                                                   |
 | molecule : la molcule o chercher les absorbeurs potentiels                                      |
 | strict   : VRAI ==> un atome dj marqu comme absorbeur n'est pas inclus  la liste              |
 \-------------------------------------------------------------------------------------------------- */

Atome *ABS_ProposerAbsorbeur(Molecule *molecule, boolean strict)
{
  Atome **choisis, *absorbeur;
  int num_choisi;
  unsigned long nbr_choisis;

  if (NULL == molecule) return NULL;
  if (NULL == molecule->premier) return NULL;

  /* Dcompte des atomes potentiellement absorbeurs */
  nbr_choisis = 0;
  for ( absorbeur = molecule->premier ; absorbeur != NULL ; absorbeur = absorbeur->suivant ) {
    if ( (VRAI == strict) && (VRAI == absorbeur->marques.absorbeur) ) /* Dj marqu comme absorbeur... */
      continue;

    if ( absorbeur->num_atomique >= Z_ABSORBEUR_MINIMAL ) {
      nbr_choisis++; /* Un atome plausible de plus */
    }
  }

  if (nbr_choisis == 0) {
    (void) fprintf(stderr, "propose_absorbeur: aucun atome ne convient\n");
    return NULL;
  }

  if ( 1 == nbr_choisis ) {
    printf("propose_absorbeur : un seul choix d'absorbeur, il faudrait ne pas afficher la liste...\n");
  }

  /* On cre la table de slection */
  choisis = (Atome **) calloc(nbr_choisis, sizeof(Atome *));
  if (NULL == choisis) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/absorbeur.c", "propose_absorbeur", "choisis");
    return NULL;
  }

  /* On remplit cette table */
  nbr_choisis = 0;
  for ( absorbeur = molecule->premier ; absorbeur != NULL ; absorbeur = absorbeur->suivant ) {
    if ( (VRAI == strict) && (VRAI == absorbeur->marques.absorbeur) ) /* Dj marqu comme absorbeur... */
      continue;

    if ( absorbeur->num_atomique >= Z_ABSORBEUR_MINIMAL ) {
      choisis[nbr_choisis] = absorbeur;
      nbr_choisis++; /* Un atome plausible de plus */
    }
  }

  /* On ralise la slection proprement dite (affichage de la liste et choix de l'utilisateur) */

  ATM_choisir(choisis, nbr_choisis, FAUX, &num_choisi);
  if (num_choisi < 0) { /* Slection annule */
    absorbeur = NULL;
  } else {
    absorbeur = choisis[num_choisi];
  }

  free(choisis);
  return absorbeur;
}

/* -------- ter un absorbeur d'une molcule ------------ */

boolean ABS_oter(Molecule *molecule, unsigned int i, boolean effacer_atome)
{
  Absorbeur absorbeur;
  unsigned int k;

  if (molecule == NULL) return VRAI;
  if (i >= molecule->nbr_absorbeurs) {
#ifdef _DEVELOPPE_
    fprintf(stderr, "ABS_oter : essai d'effacer l'absorbeur %u de la molcule %s qui n'en a que %u\n",
	    i + 1, molecule->nom, molecule->nbr_absorbeurs);
#endif
    return FAUX;
  }

  absorbeur = molecule->absorbeurs[i];
#ifdef _DEVELOPPE_
  printf("  Absorbeur : on enlve l'absorbeur %d de %s\n", i, molecule->nom);
#endif

  /* On l'te de la liste des absorbeurs */
  for (k = i + 1 ; k < molecule->nbr_absorbeurs ; k++)
    molecule->absorbeurs[k - 1] = molecule->absorbeurs[k];
  molecule->nbr_absorbeurs--; /* Un de moins dans la molcule */

  /* On libre ce qui va avec */
  absorbeur.absorbeur->marques.absorbeur = FAUX;

  /* MODELES : effacer les modles */
  while ( absorbeur.nbr_modeles > 0 ) {
    ABS_DetruireModele(&absorbeur, absorbeur.modeles);
  }

  if (effacer_atome == VRAI) {/* On demande d'effacer l'atome en mme temps */
    ATM_efface(molecule, absorbeur.absorbeur, VRAI);
  }

  if (molecule == mol_courante) { /* On a modifi la molcule courante */
    MOL_aller(mol_courante); /* Le plus simple pour mettre tout  jour... */
  }

  return VRAI; /* Effacement a pu tre fait */
}
