/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1997 - 2002, Emmanuel CURIS
   
   Modles structuraux
   Ajuster les spectres associs aux modles des molcules
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "atomes.h"
#include "couches.h"
#include "feff.h"    /* Pour ajustement / modles FEFF */
#include "modeles.h" /* Pour ajustement / absorbeur */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Ajustement d'une molcule complte ~-~-~-~-~-~-~-~- */

/* ------ Ajustement sur tous les absorbeurs d'une molcule ----
     Renvoie VRAI si l'ajustement a pu avoir lieu
*/

int MOL_ajuste(Molecule *molecule)
{
  Absorbeur **absorbeurs, *abs_mol;
  struct modele_ajuste **modeles, *modele_combine;
  unsigned int i, j;
  unsigned int nbr_modeles = 0, nbr_absorbeurs = 0;

  if (molecule == NULL) return FAUX;
  if ( molecule->modele != NULL ) {
    /* Modle dj prt : on l'ajuste directement ! */
    return MODELE_AjustementComplet( molecule->modele, NULL, VRAI, VRAI, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );
  }

  if (molecule->nbr_absorbeurs == 0) {
    (void) fprintf(stderr, "MOL_ajuste : la molcule %s n'a pas d'absorbeur ! Aucun ajustement possible.\n", molecule->nom);
    /* Molcule sans absorbeur */
    return FAUX;
  }

  if (molecule->nbr_absorbeurs == 1) { /* Un seul absorbeur : on l'ajuste */
    return ABS_Ajuster( molecule->absorbeurs );
  }

  /* S'il y a plusieurs absorbeurs : on retient au plus un modle par absorbeur
     1. Construire la table des absorbeurs */
  absorbeurs = (Absorbeur **) calloc( molecule->nbr_absorbeurs, sizeof(Absorbeur *) );
  if (absorbeurs == NULL) {
    (void) fprintf(stderr, "MOL_ajuste : impossible de crer la table des absorbeurs");
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/ajuster.c", "MOL_ajuste", "absorbeurs");
    return FAUX;
  }

  /* On remplit cette table d'absorbeurs */
  for ( i = 0, abs_mol = molecule->absorbeurs, j = 0 ; i < molecule->nbr_absorbeurs ; i++, abs_mol++ ) {
    if ( NULL == abs_mol->modeles ) continue; /* Pas de modle ! */
    if ( NULL == abs_mol->modele_actif ) {
      (void) fprintf(stderr, "MOL_ajuste : modles, mais pas de modle actif, pour l'absorbeur %u [molcule : %s]\n",
		     i + 1, molecule->nom);
      continue;
    }
    if ( NULL == abs_mol->modele_actif->experience ) continue; /* Pas de spectre exprimental ! */

    /* Si l'on est ici, c'est que le modle est acceptable */
    absorbeurs[ j ] = abs_mol;
    j++;
    nbr_modeles++; /* Pour l'instant, un seul modle par absorbeur, mais  faire voluer :
		      ajustement simultan seuils K & L... */
  }
  nbr_absorbeurs = j;

  /* On traite les cas simples */
  if ( 0 == nbr_modeles ) {
    (void) fprintf(stderr, "MOL_ajuste : aucun modle utilisable parmi les %u absorbeurs [molcule : %s]\n",
		   molecule->nbr_absorbeurs, molecule->nom);
    free( absorbeurs );
    return -2;
  }
  if ( ( 1 == nbr_absorbeurs ) || ( 1 == nbr_modeles ) ) {
    Absorbeur *abs_ajuste;

    abs_ajuste = absorbeurs[ 0 ];
    free( absorbeurs );
    return ABS_Ajuster( abs_ajuste );
  }

  /* 2. Grouper les absorbeurs de mme Z, pour faire les CL ensuite
     Tri brutal, car on ne devrait jamais avoir plus d'une dizaine d'absorbeurs ! */
  for ( i = 0 ; i < nbr_absorbeurs ; i++ ) {
    for ( j = i ; j < nbr_absorbeurs ; j++ ) {
      unsigned int Z_i, Z_j;

      Z_i = absorbeurs[ i ]->absorbeur->num_atomique;
      Z_j = absorbeurs[ j ]->absorbeur->num_atomique;
      if ( Z_i > Z_j ) {
	Absorbeur *echange;

	echange = absorbeurs[ i ];
	absorbeurs[ i ] = absorbeurs[ j ];
	absorbeurs[ j ] = echange;
      } else if ( Z_i == Z_j ) { /* On trie ensuite par graphe exprimental concern */
	Graphe *exp_i, *exp_j;

	exp_i = absorbeurs[ i ]->modele_actif->experience;
	exp_j = absorbeurs[ j ]->modele_actif->experience;
	if ( exp_i > exp_j ) {
	  Absorbeur *echange;

	  echange = absorbeurs[ i ];
	  absorbeurs[ i ] = absorbeurs[ j ];
	  absorbeurs[ j ] = echange;
	}
      }
    }
  }

  /* 3. Construction de la table des modles */
  modeles = (struct modele_ajuste **) calloc( nbr_modeles, sizeof( struct modele_ajuste * ) );
  if ( NULL == modeles ) {
    (void) fprintf(stderr, "MOL_ajuste : impossible de crer la table des modles\n");
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/ajuster.c", "MOL_ajuste", "modeles");
    return -1;
  }

  /* On cre les modles en fonction... (i : indice de lecture ; j : indice d'criture) */
  i = 0 ; j = 0;
  do {
    if ( i == nbr_absorbeurs - 1 ) { /* Dernier : isol */
      modeles[ j ] = absorbeurs[ i ]->modele_actif;
      j++;
      break;
    }
    if ( i >= nbr_absorbeurs ) { /* On a fini */
      break;
    }

    /* A-t-on des absorbeurs de mme Z et portant sur le mme graphe exprimental ? */
    if ( absorbeurs[ i ]->absorbeur->num_atomique == absorbeurs[ i + 1 ]->absorbeur->num_atomique ) {
      unsigned int nbr_termes, i_final = i + 1;

      while ( ( absorbeurs[ i_final ]->absorbeur->num_atomique == absorbeurs[ i ]->absorbeur->num_atomique ) &&
	      ( absorbeurs[ i_final ]->modele_actif->experience == absorbeurs[ i ]->modele_actif->experience ) ) {
	i_final++;
	if ( i_final >= nbr_absorbeurs )
	  break;
      }
      /* i_final est donc sur le premier absorbeur diffrent */
      nbr_termes = i_final - i;

      if ( 1 == nbr_termes ) {
	modeles[ j ] = absorbeurs[ i ]->modele_actif;
      } else {
	struct modele_ajuste *modele_CL = NULL;

	modele_CL = MODELE_CreerModele( NULL, "CL / absorbeur", MODELE_CL_MODELES, absorbeurs[ i ]->modele_actif->experience, VRAI );
	if ( NULL == modele_CL ) {
	  (void) fprintf(stderr, "MOL_ajuste : impossible de prparer la C.L. des absorbeurs\n");
	  free( absorbeurs );
	  free( modeles );
	  return -10;
	}

	printf("MOL_ajuste :  FAIRE --- modle de CL pour les absorbeurs de mme Z, %u termes\n", nbr_termes);
	modeles[ j ] = modele_CL;
      }
      j++;
      i = i_final;
    } else { /* Absorbeur de ce Z unique : on garde le modle tel quel */
      modeles[ j ] = absorbeurs[ i ]->modele_actif;
      j++;
    }
  } while ( i < nbr_absorbeurs );
  nbr_modeles = j; /* Nombre rel de modles */

  /* La table des absorbeurs devient inutile */
  free( absorbeurs );

  if ( 1 == nbr_modeles ) { /* Un seul modle : on l'ajuste donc classiquement ! */
    struct modele_ajuste *modele_ajustable;

    modele_ajustable = modeles[ 0 ];
    free( modeles );

    return MODELE_AjustementComplet( modele_ajustable, NULL, VRAI, VRAI, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );
  }

  /* 4. On cre le modle multiple */
  modele_combine = MODELE_CreerModele( NULL, "Modle multiple", MODELE_MULTI_MODELES, NULL, VRAI );
  if ( NULL == modele_combine ) {
    (void) fprintf(stderr, "MOL_ajuste : impossible de prparer le modle multiple combin\n");
    free( modeles );
    return -10;
  }

  /* La table des modles devient inutile */
  free( modeles );

  /* 5. On ajuste ce modle */
  MODELE_AjustementComplet( modele_combine, NULL, VRAI, VRAI, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );

  /* 6. On mmorise le modle */
  molecule->modele = modele_combine;

  return 0; /* On a pu lancer l'ajustement */
}

/* ~~~~~~~~ Ajustement structural d'un absorbeur ~~~~~~~~ */

/* -------- Cas simple : un seul absorbeur -------------------\
 |  renvoie VRAI si on a pu ajuster, <0 si erreur             |
 |                                                            |
 | molecule = la molcule contenant l'absorbeur  ajuster     |
 | num_absorbeur = le numro de l'absorbeur de cette molcule |
 \----------------------------------------------------------- */

int ABS_Ajuster(Absorbeur *absorbeur)
{
  int retour;

  if ( NULL == absorbeur ) return -1;
  if ( NULL == absorbeur->modeles ) {
    return -2;
  }

  /* Pour l'instant, on se contente d'ajuster btement le modle actif.
     Voir plus tard pour un ajustement multiple (exemple : seuils K & L simultans */
  retour = MODELE_AjustementComplet( absorbeur->modele_actif, NULL, VRAI, VRAI, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );

  return retour;
}
