/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1997 - 2001, Emmanuel CURIS
   
   Modles structuraux
   Chercher quelque chose dans les modles structuraux
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h> /* Pour isspace */
#include <string.h>

#include "systeme.h"

#include "atomes.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- L'atome correspond-il aux critres proposs ? --------\
 |  Renvoie VRAI si l'atome convient, FAUX sinon                  |
 |                                                                |
 | atome      : l'atome  vrifier                                |
 | x,y,z      : coordonnes de rfrence                          |
 | Z          : Z de rfrence (<=0 : on s'en fiche)              |
 | coordonnes : quel type de coordonnes utiliser ?               |
 \--------------------------------------------------------------- */

boolean ATM_AtomeConvient(Atome *atome, double x, double y, double z, int Z, int coordonnees)
{
  double x_atm, y_atm, z_atm;

  if (NULL == atome) return FAUX;

  if ( Z > 0 ) { /* On s'intresse au numro atomique */
    if (atome->num_atomique != Z) return FAUX;
  }

  /* Slection des coordonnes de contrle */
  switch ( coordonnees )
  {
  case ATM_COORD_IGNORER: /* On ne s'intresse pas aux coordonnes atomiques */
    x_atm = x; y_atm = y; z_atm = z;
    break;

  case ATM_COORD_NORMALE:
    x_atm = atome->position.x; y_atm = atome->position.y; z_atm = atome->position.z;
    break;

  case ATM_COORD_AJUSTE:
    x_atm = atome->position_ajustee.x; y_atm = atome->position_ajustee.y; z_atm = atome->position_ajustee.z;
    break;

  default:
    (void) fprintf(stderr, "ATM_AtomeConvient : mode de coordonnes inconnu - %d\n", coordonnees);
    return FAUX;
  }

  /* Distance entre la position de l'atome et la position choisie */
  if ( fabs(x_atm - x) > 1e-4 ) return FAUX; /* Tolrance de 0,0001  parat acceptable... */
  if ( fabs(y_atm - y) > 1e-4 ) return FAUX;
  if ( fabs(z_atm - z) > 1e-4 ) return FAUX;

  /* Pas de dsaccord trouv... */
  return VRAI;
}

/* -------- Chercher un atome prcis dans la liste des atomes --------\
 |  Renvoie le premier atome trouv qui satisfasse aux critres       |
 |     NULL en cas d'erreur ou si aucun atome n'est trouv            |
 |                                                                    |
 | x, y, z  : coordonnes o doit se trouver l'atome                  |
 | Z        : numro atomique de l'atome (<=0 : peu importe)          |
 | molecule : dans quelle molcule trouver l'atome (NULL : toutes)    |
 \------------------------------------------------------------------- */

Atome *ATM_chercher(Vecteur3D *position, int Z, Molecule *molecule)
{
  Atome *atome = NULL;

  if ( NULL == molecule ) { /* On cherche dans toutes les molcules */
    for (molecule = premiere_molecule ; molecule != NULL ; molecule = molecule->suivante) {
      atome = ATM_chercher(position, Z, molecule);
      if ( atome != NULL ) return atome; /* Premier trouv convient... */
    }

    return NULL; /* Aucun atome ne convient, dans aucune molcule */
  }

  /* On se limite  la molcule indique */
  if (NULL == molecule->premier) /* Aucun atome dans la molcule indique */
    return NULL;

  for (atome = molecule->premier ; atome != NULL ; atome = atome->suivant) {
    if ( VRAI == ATM_AtomeConvient(atome, position->x, position->y, position->z, Z, ATM_COORD_NORMALE) )
      return atome; /* Le premier trouv convient */
  }

#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "ATM_chercher : impossible de trouver l'atome Z = %d\n", Z);
#endif
  return NULL; /* Atome non trouv dans la molcule indique */
}

/* -------- Chercher un atome caractris par un index --------\
 |   Renvoie l'atome trouv, NULL si aucun ou erreur           |
 |                                                             |
 | molecule : la molcule dans laquelle chercher               |
 | i        : l'index que l'on doit trouver                    |
 \------------------------------------------------------------ */

Atome *ATM_index(Molecule *molecule, unsigned long i)
{
  Atome *atome;

  if ( NULL == molecule ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Pas de molcule indique (atome %lu cherch)\n",
		   __FILE__, __FUNCTION__, __LINE__, i );
#endif
    return NULL;
  }

#ifdef _DEVELOPPE_
  printf("ATM_index : recherche de l'atome d'index %lu", i);
#endif

  for ( atome = molecule->premier ; atome ; atome = atome->suivant )
    if ( i == atome->index ) {
#ifdef _DEVELOPPE_
      printf(" -- %d en (%f, %f, %f)\n", atome->num_atomique, atome->position.x, atome->position.y, atome->position.z);
#endif
      return atome;
    }

#ifdef _DEVELOPPE_
  printf(" -- NON TROUV !\n");
#endif

  return NULL; /* Aucun atome trouv */
}

/* -~-~-~-~-~-~-~-~ Reprages dans les structures -~-~-~-~-~-~-~-~ */

/* -------- Quelle molcule contient l'atome ? -------- */

Molecule *ATM_QuelleMolecule(Atome *atome)
{
  Molecule *molecule;

  if (NULL == atome) return NULL;
  if (NULL == premiere_molecule) {
    (void) fprintf(stderr, "ATM_QuelleMolecule : Aucune molcule en mmoire !\n");
    return NULL;
  }

  /* Remarque : si marqu absorbeur, plus rapide de regarder les absorbeurs
      FAIRE */

  for ( molecule = premiere_molecule ; molecule != NULL ; molecule = molecule->suivante ) {
    if ( NULL == molecule->premier )
      continue;

    if ( VRAI == ATM_EstDansMolecule(atome, molecule) )
      return molecule;
  }

  (void) fprintf(stderr, "ATM_QuelleMolecule : atome hors de toute molcule !\n");
  return NULL;
}

/* -------- La molcule contient-elle l'atome ? -------- */

boolean ATM_EstDansMolecule(Atome *atome, Molecule *molecule)
{
  Atome *atm;

  if (NULL == atome) return FAUX;
  if (NULL == molecule) return FAUX;
  if (NULL == molecule->premier) return FAUX;

  for ( atm = molecule->premier ; atm != NULL ; atm = atm->suivant) {
    if (atm == atome)
      return VRAI;
  }

  return FAUX;
}

/* -------- Quelle molcule contient le fragment ? -------- */

Molecule *FRG_QuelleMolecule(Fragment *fragment)
{
  Molecule *molecule;

  if (NULL == fragment) return NULL;

  if (NULL == premiere_molecule) {
    (void) fprintf(stderr, "FRG_QuelleMolecule : aucune molcule en mmoire !\n");
    return NULL;
  }

  for ( molecule = premiere_molecule ; molecule != NULL ; molecule = molecule->suivante ) {
    if ( NULL == molecule->fragments ) /* Pas de fragment dans la molcule */
      continue;

    if ( VRAI == FRG_EstDansMolecule(fragment, molecule) )
      return molecule;
  }

  (void) fprintf(stderr, "FRG_QuelleMolecule : aucune molcule ne contient le fragment indiqu !\n");
  return NULL;
}

/* -------- Le fragment appartient-il  la molcule indique ? -------- */

boolean FRG_EstDansMolecule(Fragment *fragment, Molecule *molecule)
{
  Fragment *frg;

  if (NULL == fragment) return FAUX;
  if (NULL == molecule) return FAUX;
  if (NULL == molecule->fragments) return FAUX;

  for ( frg = molecule->fragments ; frg != NULL ; frg = frg->suivant ) {
    if (frg == fragment) return VRAI;
  }

  /* Non trouv dans les fragments de la molcule */
  return FAUX;
}

/* -------- Quel absorbeur est associ  l'atome ? -------- */

Absorbeur *ATM_QuelAbsorbeur(Molecule *molecule, Atome *atome)
{
  unsigned int i;

  if (NULL == molecule) return NULL;
  if (NULL == atome) return NULL;

  if ( FAUX == atome->marques.absorbeur ) {
    (void) fprintf(stderr, "ATM_QuelAbsorbeur : l'atome indiqu n'est pas marqu comme absorbeur !\n");
    return NULL;
  }

  for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
    if ( molecule->absorbeurs[i].absorbeur == atome )
      return molecule->absorbeurs + i;
  }

  /* Absorbeur non trouv */
  return NULL;
}

/* -------- L'atome appartient-il au fragment indiqu ? -------- */

boolean ATM_EstDansFragment(Atome *atome, Fragment *fragment)
{
  unsigned int i;

  if (NULL == atome) return FAUX;
  if (NULL == fragment) return FAUX;

  for (i = 0 ; i < fragment->nbr_atomes ; i++) {
    if ( atome == fragment->atomes[i] ) /* Trouv ! */
      return VRAI;
  }

  /* Si on est l, c'est qu'il n'appartient pas au fragment */
  return FAUX;
}

/* -------- Quel(s) fragment(s) contien(nen)t l'atome indiqu ? -------- */

Fragment *ATM_QuelsFragments(Atome *atome, Molecule *molecule, unsigned int *nbr_fragments, Fragment ***fragments)
{
  unsigned int nbr_frg = 0;
  Fragment *frg, *frg_principal, **table;

  if (NULL == atome) return NULL;
  if (NULL == molecule) return NULL;

  if (NULL == nbr_fragments) /* On ne s'intresse qu'au fragment principal, s'il existe */
    return atome->fragment;

  /* Passe 1 : dcompte des fragments qui contiennent l'atome */
  frg_principal = atome->fragment;
  for (frg = molecule->fragments ; frg != NULL ; frg = frg->suivant) {
    if ( VRAI == ATM_EstDansFragment(atome, frg) ) {
      nbr_frg++;
      if (NULL == frg_principal)
	frg_principal = frg; /* On cherche le nouveau fragment principal */
    }
  }

  /* Vrifications de cohrence (version de dveloppement surtout) */
  if (0 == nbr_frg) { /* Aucun fragment ne contient l'absorbeur */
#ifdef _DEVELOPPE_
    if (atome->fragment != NULL) {
      (void) fprintf(stderr, "ATM_QuelsFragments : aucun fragment ne contient l'atome, mais atome->fragment != NULL !\n");
      abort();
    }
    if ( VRAI == atome->marques.multi_frg ) {
      (void) fprintf(stderr, "ATM_QuelsFragments : aucun fragment ne contient l'atome, mais atome marqu comme multi_frg !\n");
      abort();
    }
#endif
    *nbr_fragments = 0;
    if (NULL != fragments) *fragments = NULL;
    return NULL;
  }

#ifdef _DEVELOPPE_
  /* Remarque : on peut avoir atome->fragment = NULL car en cours d'limination d'un fragment
       ==> on recherche justement le nouveau fragment principal !
     Mme chose pour la marque multi_frg quand on a 1 seul fragment */
  if (nbr_frg > 1) {
    if ( FAUX == atome->marques.multi_frg ) { /* Par contre, ceci ne devrait jamais arriver */
      (void) fprintf(stderr, "ATM_QuelsFragments : atome dans plusieurs fragments, mais non marqu comme multi_frg !\n");
      abort();
    }
  }
#endif

  if (NULL == fragments) { /* On ne veut pas la table */
    *nbr_fragments = nbr_frg;
    return frg_principal;
  }

  /* On rserve la table des fragments, si ncessaire */
  table = (Fragment **) calloc(nbr_frg, sizeof(Fragment *));
  if (NULL == table) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/chercher.c", "ATM_QuelsFragments", "table");
    (void) fprintf(stderr, "ATM_QuelsFragments : impossible de rserver la table des fragments !\n");

    *nbr_fragments = 0;
    *fragments = NULL;
    return frg_principal;
  }

  *fragments = table;
  *nbr_fragments = nbr_frg;

  /* Passe 2 : remplissage de la table */
  frg_principal = atome->fragment;
  for (frg = molecule->fragments, nbr_frg = 0 ; frg != NULL ; frg = frg->suivant) {
    if ( VRAI == ATM_EstDansFragment(atome, frg) ) {
      table[nbr_frg] = frg;
      nbr_frg++;
    }
  }

  return frg_principal;
}
