/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1997 - 2000, Emmanuel CURIS
   Création : 17 janvier 2000

   Création d'un atome par l'utilisateur, en tenant compte des symétries cristallines
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

#include "systeme.h"
#include "ascii_utils.h"
#include "txt_utils.h"

#include "exafs.h"
#include "atomes.h"
#include "cristallo.h"
#include "interface.h"
#include "science.h" /* Pour M_PI */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ------ Les textes ------ */

#include "cristallo.txt"

#define LGMAX_NOM_RESEAU 20
#define LGMAX_NOM_GROUPE 13

/* Groupe P1 pour travail "par défaut" */
/* static float P1_identite[ 12 ] = { */
/*   1.0, 0.0, 0.0, 0.0, */
/*   0.0, 1.0, 0.0, 0.0, */
/*   0.0, 0.0, 1.0, 0.0 */
/* }; */

/* static struct groupe_espace groupe_P1 = { */
/*   1, "P1", NULL, 1, 1, P1_identite, P1_identite + 1 /\* Puisque l'on veut 0.0, 0.0, 0.0... *\/ */
/* }; */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* static int  GER_creer(unsigned int objet, int x, int y); */
/* static void prepare_arbre_creer(void); */

/* static void ActiverFractionnaires(int sens); */

static int CRIS_AnalyseTranslation(FILE *fichier, float *vecteur);
static int CRIS_AnalyseOperation  (FILE *fichier, float *matrice);
static int CRIS_AnalyseCoordonnee (char *coordonnee, float *ligne_matrice);
static int CRIS_OrdreGroupes      (const void *v1, const void *v2);

static int CRIS_GenererMaille(Molecule *molecule, unsigned long nbr_mailles, Fragment **mailles, Vecteur3D *position,
			      int num_atomique, double charge, double rayon);

inline static void CRIS_ViderSysteme(struct systeme_cristallin *systeme);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Applique une opération de symétrie, avec la matrice [3×4 ou 4×4, en ligne] indiquée -------- */

void CRIS_GenereSymetrique(Vecteur3D *original, float *matrice, Vecteur3D *image)
{
  double x_sym, y_sym, z_sym;

  /* Attention, la matrice contient les valeurs en short int et multipliées par 1000 */
  /* Rq : si nécessaire, il faudrait remplacer 333 par 1/3 et 667 par 2/3 */

  x_sym = original->x * matrice[ 0 ] + original->y * matrice[ 1 ] + original->z * matrice[  2 ] + matrice[  3 ];
  y_sym = original->x * matrice[ 4 ] + original->y * matrice[ 5 ] + original->z * matrice[  6 ] + matrice[  7 ];
  z_sym = original->x * matrice[ 8 ] + original->y * matrice[ 9 ] + original->z * matrice[ 10 ] + matrice[ 11 ];
  /* On suppose implicitement que l'on reste en coordonnée homogènes avec t = 1 */

  /* On vérifie que l'on reste dans la maille.
     Le seul problème vraiment envisageable est celui des valeurs négatives, mais mieux vaut tout prévoir ! */
  while ( x_sym < 0.0 ) x_sym += 1.0; while ( x_sym > 1.0 ) x_sym -= 1.0;
  while ( y_sym < 0.0 ) y_sym += 1.0; while ( y_sym > 1.0 ) y_sym -= 1.0;
  while ( z_sym < 0.0 ) z_sym += 1.0; while ( z_sym > 1.0 ) z_sym -= 1.0;

#ifdef _DEVELOPPE_
  printf("X symétrique : (%f, %f, %f)\n", x_sym, y_sym, z_sym);
#endif

  /* On mémorise l'image ainsi obtenue */
  image->x = x_sym; image->y = y_sym; image->z = z_sym;
}

/* -------- Normaliser les coordonnées fractionnaires : elles doivent êtres dans [0, 1[ -------- */

int CRIS_NormaliserCoordonneesFractionnaires(Vecteur3D *position)
{
  double p_entiere;

  if ( NULL == position ) return -1;

  /* Si coordonnée positive : il faut prendre la partie décimale
     Si coordonnée négative : il faut prendre 1 - partie décimale
     NB : modf = fonction non-ANSI, mais C99 : garde le signe... */

  position->x = modf( position->x, &p_entiere );
  if ( position->x < 0.0 ) { position->x += 1.0; }

  position->y = modf( position->y, &p_entiere );
  if ( position->y < 0.0 ) { position->y += 1.0; }

  position->z = modf( position->z, &p_entiere );
  if ( position->z < 0.0 ) { position->z += 1.0; }

  return 0;
}

/* -------- On vérifie que l'atome obtenu par symétrie n'est pas identique... ------ */

boolean CRIS_ControlerCreation(Vecteur3D *position, unsigned int nbr_crees, Point_3D *existants)
{
  double x, y, z;
  double d_x, d_y, d_z;

  /* Copie pour ne pas modifier l'original ! */
  x = position->x; y = position->y; z = position->z;

  /* Attention à bien toujours se ramener dans la maille, donc dans [0,1] :
     certaines opérations de symétrie peuvent créer des atomes équivalents
     mais dans la maille d'à côté (grrrr) */
  while ( x < 0.0 ) x += 1.0; while ( x >= 1.0 ) x -= 1.0;
  while ( y < 0.0 ) y += 1.0; while ( y >= 1.0 ) y -= 1.0;
  while ( z < 0.0 ) z += 1.0; while ( z >= 1.0 ) z -= 1.0;

  /* On compare aux images déjà obtenues par les autres opérations (y compris identité !) */
  if ( nbr_crees > 0 ) {
    unsigned int s;

    for ( s = 0 ; s < nbr_crees ; s++ ) {
      d_x = fabs( x - existants[ s ].x );
      d_y = fabs( y - existants[ s ].y );
      d_z = fabs( z - existants[ s ].z );

      if ( ( d_x < 0.01 ) && ( d_y < 0.01 ) && ( d_z < 0.01 ) ) {
#ifdef _DEVELOPPE_
	/*	printf("Atome déjà obtenu par une symétrie précédente\n"); */
#endif
	return VRAI; /* Atomes plus proches que 0.1 A -> supposés identiques */
      }
    }
  }

  /* C'est bien un nouvel atome */
  return FAUX;
}

/* -------- On vérifie qu'il n'est pas sur une face, et si oui on dit laquelle ---------
 |   Renvoie VRAI si l'atome est sur une face ET qu'un symétrique existe sur la face opposée
 |        ==> dans ce cas, il y a des risques d'apparition de doublons par translation
 |
 | x, y, z   : coordonnées de l'atome dans la maille que l'on veut vérifier
 | nbr_crees : le nombre d'atomes symétrique déjà créés
 | existants : la liste des atomes déjà existants
 | face      : si non-NULL, contient la face sur laquelle est l'atome
 \------------------------------------------------------------------------------------------ */

boolean CRIS_EstSurUneFace(Vecteur3D *position, unsigned int nbr_crees, Point_3D *existants, int *face)
{
  boolean probleme = FAUX;
  int face_atome = DANS_LA_MAILLE;
  Vecteur3D pos_image;

  /* On commence par vérifier si l'atome est sur une face ou non. */
  if ( fabs( position->x ) < 1.0e-4 ) face_atome |= SUR_FACE_BC; /* Atome en x = 0 */
  if ( fabs( position->y ) < 1.0e-4 ) face_atome |= SUR_FACE_AC; /* Atome en y = 0 */
  if ( fabs( position->z ) < 1.0e-4 ) face_atome |= SUR_FACE_AB; /* Atome en z = 0 */

  if ( fabs( position->x - 1.0 ) < 1.0e-4 ) face_atome |= SUR_FACE_BC_FOND; /* Atome en x = 1 */
  if ( fabs( position->y - 1.0 ) < 1.0e-4 ) face_atome |= SUR_FACE_AC_FOND; /* Atome en y = 1 */
  if ( fabs( position->z - 1.0 ) < 1.0e-4 ) face_atome |= SUR_FACE_AB_FOND; /* Atome en z = 1 */

  if ( face != NULL ) *face = face_atome;

  if ( DANS_LA_MAILLE == face_atome ) return FAUX; /* Aucun problème : l'atome n'est pas sur une face */

#ifdef _DEVELOPPE_
  /*  printf("************** Atome sur une face : (%f, %f, %f) *******************\n", x, y, z); */
#endif
  /* Maintenant, il faut vérifier que l'atome de la face d'en face n'a pas déjà été créé.
     Pour cela, on ruse et on utilise special()... */

  if ( face_atome & SUR_FACE_AB ) { /* Atome sur le fond. Problème si un atome en (x,y,1) a déjà été créé */
    pos_image = *position;
    pos_image.z = 1.0;

    probleme = CRIS_ControlerCreation( &pos_image, nbr_crees, existants );
    if ( VRAI == probleme ) return VRAI; /* Inutile de continuer */
  }
  if ( face_atome & SUR_FACE_AC ) { /* Atome devant. Problème si un atome en (x,1,z) a déjà été créé */
    pos_image = *position;
    pos_image.y = 1.0;

    /*    printf("Atome sur une face y = 0 = face horizontale, fond de la maille\n");
	  printf("On contrôle (%f %f %f) et (%f %f %f)\n", x, y, z, x, 1.0, z); */
    probleme = CRIS_ControlerCreation( &pos_image, nbr_crees, existants);
    if ( VRAI == probleme ) return VRAI; /* Inutile de continuer */
  }
  if ( face_atome & SUR_FACE_BC ) { /* Atome à gauche. Problème si un atome en (1,y,0) a déjà été créé */
    pos_image = *position;
    pos_image.x = 1.0;

    probleme = CRIS_ControlerCreation( &pos_image, nbr_crees, existants);
    if ( VRAI == probleme ) return VRAI; /* Inutile de continuer */
  }

  if ( face_atome & SUR_FACE_AB_FOND ) { /* Atome sur le couvercle. Problème si un atome en (x,y,0) a déjà été créé */
    pos_image = *position;
    pos_image.z = 0.0;

    probleme = CRIS_ControlerCreation( &pos_image, nbr_crees, existants);
    if ( VRAI == probleme ) return VRAI; /* Inutile de continuer */
  }
  if ( face_atome & SUR_FACE_AC_FOND ) { /* Atome derrière. Problème si un atome en (x,0,z) a déjà été créé */
    pos_image = *position;
    pos_image.y = 0.0;

    /*    printf("Atome sur une face y = 1 = face horizontale, couvercle de la maille\n");
	  printf("On contrôle (%f %f %f) et (%f %f %f)\n", x, y, z, x, 0.0, z); */
    probleme = CRIS_ControlerCreation( &pos_image, nbr_crees, existants);
    if ( VRAI == probleme ) return VRAI; /* Inutile de continuer */
  }
  if ( face_atome & SUR_FACE_BC_FOND ) { /* Atome à droite. Problème si un atome en (0,y,z) a déjà été créé */
    pos_image = *position;
    pos_image.x = 0.0;

    probleme = CRIS_ControlerCreation( &pos_image, nbr_crees, existants);
    if ( VRAI == probleme ) return VRAI; /* Inutile de continuer */
  }

  return FAUX; /* Si on est là, c'est qu'il n'y a pas de problème - normalement */
}

/* -------- Génération des copies par translation (= de l'ensemble des mailles) -------- */

static int CRIS_GenererMaille(Molecule *molecule, unsigned long nbr_mailles, Fragment **mailles, Vecteur3D *position,
			      int num_atomique, double charge, double rayon)
{
  Atome *atome = NULL;
  int i, j, k;			/* Indices de la maille */
  unsigned long idx_mailles;
  double c_a, c_b, c_g, s_g;
  struct maille *maille;

  maille = molecule->maille;
  c_a = cos( maille->alpha );
  c_b = cos( maille->beta );
  c_g = cos( maille->gamma ); s_g = sin( maille->gamma );

  for ( idx_mailles = 0 ; idx_mailles < nbr_mailles ; idx_mailles++ ) {
    /* On repère les indices de la maille désirée */
    i = mailles[ idx_mailles ]->i;
    j = mailles[ idx_mailles ]->j;
    k = mailles[ idx_mailles ]->k;

    /* On s'en sert pour obtenir les coordonnées absolues */
    double x_reel, y_reel, z_reel;
    Vecteur3D coord_absolues;

    x_reel = position->x + (double) i; /* Translation */
    y_reel = position->y + (double) j;
    z_reel = position->z + (double) k;

    coord_absolues.x = x_reel * maille->a + y_reel * maille->b * c_g + z_reel * maille->c * c_b;
    coord_absolues.y = y_reel * maille->b * s_g + z_reel * ( maille->c / s_g ) * ( c_a - c_b * c_g );
    coord_absolues.z = z_reel * maille->c * sqrt( fabs(1 - c_a*c_a - c_b*c_b - c_g*c_g + 2 * c_a * c_b * c_g) ) / s_g;

#ifdef _DEVELOPPE_
    printf("   conversion, maille (%d, %d, %d) : %f %f %f\n", i, j, k, coord_absolues.x, coord_absolues.y, coord_absolues.z);
#endif
    atome = ATM_creer(molecule, NULL, num_atomique, &coord_absolues, charge, rayon);

    /* On ajoute l'atome créé à la bonne maille, ie au bon fragment de la molécule */
    FRG_ajoute_atome( mailles[ idx_mailles ], 1, &atome );
  }

  return 0; /* Aucun problème rencontré */
}

/* ~~~~~~~~ Génération complète d'une maille avec ses symétriques éventuels
   La routine se charge de générer aussi les copies par translation de la maille dans chaque maille indiquée.
   Elle s'arrange pour éviter d'avoir en double des atomes :
       - elle vérifie que l'image par opération de symétrie n'est pas confondue avec un autre atome ;
       - elle vérifie que l'on ne crée pas d'image symétrique d'un atome sur une face de la maille
         lorsqu'elle correspond aussi à l'image par translation dans une autre maille demandée.
   Ainsi, dans une maille, on ne peut pas avoir d'atome avec une coordonnée fractionnaire >= 1 : il appartient à la maille d'à côté

 molecule : la molécule étudiée, qui doit contenir la maille ! [sinon, annulé]
 position : coordonnées fractionnaires de l'atome
 Z        : numéro atomique
 charge   : charge de l'atome
 rayon    : rayon de l'atome (en pm)
 nombre_mailles : nombre de mailles à créer
 mailles        : table des fragments associés à chaque maille [si NULL : annulé !]
 ~~~~~~~~ */

int CRIS_GenereCristal(struct molecule *molecule, Vecteur3D *position, int Z, double charge, double rayon,
		       unsigned long nombre_mailles, struct fragment **mailles)
{
  unsigned int nbr_images = 0;
  unsigned int idx_trans, idx_op;
  unsigned long nbr_operations;
  struct maille *maille;
  Vecteur3D pos_image;
  Point_3D *images = NULL;

  if ( ( NULL == molecule ) || ( NULL == molecule->maille ) || ( NULL == mailles ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Paramètre absent [molécule, maille de la molécule ou fragments] ! Abandon.\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  maille = molecule->maille;

  /* On normalise les coordonnées fractionnaires indiquées */
  CRIS_NormaliserCoordonneesFractionnaires( position );

  /* Combien d'opérations ? */
  nbr_operations = maille->nbr_operations * maille->nbr_translations;

  /* On prépare pour les symétriques */
  if ( nbr_operations > 0 ) { /* il y a des symétries */
    images = (Point_3D *) calloc( nbr_operations, sizeof(Point_3D) );
    if (images == NULL) {
      ERR_Memoire( "Table des opérations", nbr_operations * sizeof(Point_3D) );
      nbr_operations = 0;
    }
  }

#ifdef _DEVELOPPE_
  printf( "%s: Référence : Z = %d (%f, %f, %f)\n", __FUNCTION__, Z, position->x, position->y, position->z);
#endif

  /* Pour chacune des translations du groupe d'espace */
  for ( idx_trans = 0 ; idx_trans < maille->nbr_translations ; idx_trans++ ) {
#ifdef _DEVELOPPE_
    printf( "%s:   Translation (%g, %g, %g)\n", __FUNCTION__,
	    (maille->translations + 3 * idx_trans)[ 0 ], (maille->translations + 3 * idx_trans)[ 1 ], (maille->translations + 3 * idx_trans)[ 2 ]);
#endif
    /* On applique chacune des opérations de symétrie */
    for ( idx_op = 0 ; idx_op < maille->nbr_operations ; idx_op++ ) {
      float matrice[ 12 ]; /* Matrice de la transformation globale */

#ifdef _DEVELOPPE_
      printf("  Opération %u\n", idx_op + 1);
#endif
      /* Ligne des X */
      matrice[ 0 ] = (maille->operations + 16 * idx_op)[ 0 ];
      matrice[ 1 ] = (maille->operations + 16 * idx_op)[ 1 ];
      matrice[ 2 ] = (maille->operations + 16 * idx_op)[ 2 ];
      matrice[ 3 ] = (maille->operations + 16 * idx_op)[ 3 ] + (maille->translations + 3 * idx_trans)[ 0 ]; /* On n'oublie pas la translation */

      /* Ligne des Y */
      matrice[ 4 ] = (maille->operations + 16 * idx_op)[ 4 ];
      matrice[ 5 ] = (maille->operations + 16 * idx_op)[ 5 ];
      matrice[ 6 ] = (maille->operations + 16 * idx_op)[ 6 ];
      matrice[ 7 ] = (maille->operations + 16 * idx_op)[ 7 ] + (maille->translations + 3 * idx_trans)[ 1 ]; /* On n'oublie pas la translation */

      /* Ligne des Z */
      matrice[  8 ] = (maille->operations + 16 * idx_op)[  8 ];
      matrice[  9 ] = (maille->operations + 16 * idx_op)[  9 ];
      matrice[ 10 ] = (maille->operations + 16 * idx_op)[ 10 ];
      matrice[ 11 ] = (maille->operations + 16 * idx_op)[ 11 ] + (maille->translations + 3 * idx_trans)[ 3 ]; /* On n'oublie pas la translation */

      /* On génère l'image par la symétrie indiquée */
      CRIS_GenereSymetrique( position, matrice, &pos_image );

      /* On vérifie que l'atome n'a pas été déjà créé */
      if ( CRIS_ControlerCreation( &pos_image, nbr_images, images ) == FAUX ) { /* L'atome n'a pas déjà été créé */
	/* C'est bien une nouvelle copie par symétrie */
	boolean probleme = FAUX;
	int face_atome = DANS_LA_MAILLE;

	probleme = CRIS_EstSurUneFace( &pos_image, nbr_images, images, &face_atome);
	if ( VRAI == probleme ) {
	  /* L'atome a déjà été créé comme image par translation... */
	  AVERTISSEMENT( "Atome (%f, %f, %f) déjà créé : on l'ignore [ne devrait pas arriver !]", pos_image.x, pos_image.y, pos_image.z );
	} else { /* Aucun problème, on génère */
#ifdef _DEVELOPPE_
	  printf("  Image en (%f, %f, %f)\n", pos_image.x, pos_image.y, pos_image.z );
	  if ( face_atome != DANS_LA_MAILLE )
	    printf(" ** Atome sur une face : face %d **\n", face_atome);
#endif
	  CRIS_GenererMaille( molecule, nombre_mailles, mailles, &pos_image, Z, charge, rayon );
	}

	if ( images != NULL ) {
	  images[ nbr_images ] = pos_image;
	  nbr_images++; /* On a créé une image de plus */
	}
#ifdef _DEVELOPPE_
      } else {
	printf( "Atome rejeté en (%f, %f, %f) / (%f, %f, %f)\n",
		position->x, position->y, position->z, pos_image.x, pos_image.y, pos_image.z );
#endif
      }
    }
  }

  free( images ); images = NULL;
  return 0; /* Tout s'est bien passé */
}

/* ------ Génération des fragments associés à chaque maille ------ */

Fragment **FRG_mailles(Molecule *molecule,
		       int i_min, int i_max, int j_min, int j_max, int k_min, int k_max)
{
  Fragment **mailles = NULL;
  unsigned int nbr_mailles, i, j, k;

  nbr_mailles  = abs(i_max - i_min) + 1;
  nbr_mailles *= abs(j_max - j_min) + 1;
  nbr_mailles *= abs(k_max - k_min) + 1; /* Nombre de mailles à réserver */

  /* On réalise l'ensemble des fragments, un par maille si possible */
  mailles = (Fragment **) calloc(nbr_mailles, sizeof(Fragment *));
  if ( NULL == mailles ) {
    ERR_Memoire( "Table des mailles/fragments", sizeof( Fragment *) * nbr_mailles );
    return NULL;
  }

  for (i = 0 ; i < (unsigned) abs(i_max - i_min) + 1 ; i++)
    for (j = 0 ; j < (unsigned) abs(j_max - j_min) + 1 ; j++)
      for (k = 0 ; k < (unsigned) abs(k_max - k_min) + 1 ; k++) {
	char titre[ 50 ]; /* Texte : 19 caractères + 0 final si mailles de -9 à 99, donc devrait coller */

	sprintf( titre, txt_nom_fragment, i_min + i, j_min + j, k_min + k );
	mailles[ NUM_MAILLE(i, j, k) ] = FRG_creer( molecule, NULL, titre, 10, NULL, NULL );
	if ( NULL == mailles[ NUM_MAILLE(i, j, k) ] ) {
	  (void) fprintf( stderr, "%s, %s l.%u : impossible de créer la maille (%u, %u, %u)\n",
			  __FILE__, __FUNCTION__, __LINE__, i_min + i, j_min + j, k_min + k );
	  ERR_Memoire( "Fragment pour une maille", 0 );
	}
	mailles[ NUM_MAILLE(i, j, k) ]->drapeaux.maille = VRAI; /* C'est bien une maille */
	if ( (i_min + i == 0) && (j_min + j == 0) && (k_min + k == 0) )
	  mailles[ NUM_MAILLE(i, j, k) ]->drapeaux.origine = VRAI; /* C'est la maille de base */

	/* On mémorise les index de maille */
	mailles[ NUM_MAILLE(i, j, k) ]->i = i_min + i;
	mailles[ NUM_MAILLE(i, j, k) ]->j = j_min + j;
	mailles[ NUM_MAILLE(i, j, k) ]->k = k_min + k;
      }

/*   MOL_ajoute_fragment(molecule, nbr_mailles, mailles); */
  return mailles;
}

/* ~~~~~~~~ Interface utilisateur ~~~~~~~~ */

/* ~-~-~-~-~-~-~-~- Couplage avec les molécules ~-~-~-~-~-~-~-~- */

int MOL_AssocierCristallo(Molecule *molecule, struct maille *maille)
{
  if ( NULL == molecule ) return +1;
  if ( NULL == maille ) return 0;

  if ( molecule->maille != NULL ) {
    (void) fprintf(stderr, "%s, %s l.%u : Il y a déjà une maille !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -5;
  }

  molecule->maille = maille;
  return 0;
}

int MOL_EnleverCristallo(Molecule *molecule)
{
  if ( NULL == molecule ) return +1;
  
  /* On efface la maille et ses infos */
  CRIS_EffacerMaille( molecule->maille );
  molecule->maille = NULL;

  /* Terminé */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Lecture des tables d'opérations de symétrie dans un fichier ~-~-~-~-~-~-~-~- */

/* -------- Adapte les paramètres de maille indiqués, compte-tenu de la nature du système indiqué -------- */

int CRIS_AdapteParametres(struct systeme_cristallin *systeme, double *a, double *b, double *c, double *alpha, double *beta, double *gamma)
{
  int retour = 0; /* >0 pour erreurs mineures : non-respect des conventions */

  if ( NULL == systeme ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de système indiqué\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On place les angles dans 0-360°. Avant pour contrôles de cohérence, si on force de toute façon 90° ou 120° */
  while ( *alpha >= 360.0 ) *alpha -= 360.0; while ( *alpha < 0.0 ) *alpha += 360.0;
  while ( *beta  >= 360.0 ) *beta  -= 360.0; while ( *beta  < 0.0 ) *beta  += 360.0;
  while ( *gamma >= 360.0 ) *gamma -= 360.0; while ( *gamma < 0.0 ) *gamma += 360.0;

  switch ( systeme->type )
  {
  case 'C': /* Cubique : a=b=c, alpha = beta = gamma = 90° */
    *b = *a;
    *c = *a;

    *alpha = 90.0;
    *beta  = 90.0;
    *gamma = 90.0;
    break;

  case 'H': /* Hexagonal : a=b, alpha = beta = 90°, gamma = 120° */
  case 'r': /* Rhomboédrique, axes hexagonaux [P] */
    *b = *a;

    *alpha = 90.0;
    *beta  = 90.0;
    *gamma = 120.0;
    break;

  case 'M': /* Monoclinique (axe b) : alpha = gamma = 90° */
    *alpha = 90.0;
    *gamma = 90.0;
    if ( *beta < 90.0 ) {
      (void) fprintf(stderr, "%s,%sl.%u : Monoclinique axe b, angle bêta < 90° [%g°]\n",
		     __FILE__, __FUNCTION__, __LINE__, *beta);
      retour = +1;
    }
    break;

  case 'm': /* Monoclinique (axe c) : alpha = beta = 90° */
    *alpha = 90.0;
    *beta = 90.0;
    if ( *gamma < 90.0 ) {
      (void) fprintf(stderr, "%s,%sl.%u : Monoclinique axe c, angle gamma < 90° [%g°]\n",
		     __FILE__, __FUNCTION__, __LINE__, *gamma);
      retour = +1;
    }
    break;

  case 'O': /* Orthorhombique alpha = beta = gamma = 90° */
    *alpha = 90.0;
    *beta = 90.0;
    *gamma = 90.0;
    break;

  case 'Q': /* Quadratique : a = b, alpha = beta = gamma = 90° */
    *b = *a;

    *alpha = 90.0;
    *beta = 90.0;
    *gamma = 90.0;
    break;

  case 'R': /* Rhomboédrique (axes rhomboédriques) : a = b = c, alpha = beta = gamma */
    *b = *a;
    *c = *a;

    *beta = *alpha;
    *gamma= *beta;
    break;

  case 'T': /* Triclinique */
    break;
  }

  /* On contrôle que les valeurs indiquées ont un sens */
  if ( ( *a <= 0.0 ) || ( *b <= 0.0 ) || ( *c <= 0.0 ) ) {
    (void) fprintf(stderr, "%s,%sl.%u : Côtés des mailles incorrects : a = %g, b = %g, c = %g\n",
		   __FILE__, __FUNCTION__, __LINE__, *a, *b, *c);
    return -10;
  }
  if ( (   0.0 == *alpha ) || (   0.0 == *beta ) || (   0.0 == *gamma ) ||
       ( 180.0 == *alpha ) || ( 180.0 == *beta ) || ( 180.0 == *gamma ) ) {
    (void) fprintf(stderr, "%s,%sl.%u : Angles des mailles incorrects : alpha = %g°, beta = %g°, gamma = %g°\n",
		   __FILE__, __FUNCTION__, __LINE__, *alpha, *beta, *gamma);
    return -11;
  }

  /* Terminé : pas de problème détecté, ou juste un non-respect des conventions */
  return retour;
}

/* -------- Détection du classement en systèmes cristallins [fichier systemes.txt] -------- */

struct systeme_cristallin *CRIS_CreerListeSystemes(const char *dossier, unsigned long *nbr_systemes)
{
  int retour, c;
  struct systeme_cristallin *systemes = NULL;
  unsigned long i;
  FILE *fichier;

#ifdef _DEVELOPPE_
  if ( NULL == dossier ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de dossier indiqué\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }
  if ( NULL == nbr_systemes ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de variable indiquée pour stocker le nombre de systèmes [dossier %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, dossier);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }
#endif

  /* On se place dans le dossier indiqué pour travailler */
  retour = chdir( dossier );
  if ( -1 == retour ) {
    (void) fprintf(stderr, "%s,%sl.%u : ERREUR chdir(%s) --- %s\n",
		   __FILE__, __FUNCTION__, __LINE__, dossier, strerror( errno ) );
    return NULL;
  }

  /* On ouvre le fichier systemes.txt : contient la liste des systèmes disponibles */
  fichier = fopen( "systemes.txt", "rb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible d'ouvrir le fichier systemes.txt\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Une ligne par dossier : combien de lignes ? */
  *nbr_systemes = compte_lignes( fichier, '#' );
  if ( 0 == *nbr_systemes ) {
    (void) fprintf(stderr, "%s,%sl.%u : Aucun système cristallographique défini !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return NULL;
  }

  /* On réserve la liste des systèmes cristallins */
  systemes = ( struct systeme_cristallin * ) calloc( *nbr_systemes, sizeof( struct systeme_cristallin ) );
  if ( NULL == systemes ) {
    (void) fprintf(stderr, "%s,%sl.%u : Mémoire insuffisante pour la liste des systèmes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return NULL;
  }

  /* On lit système par système */
  i = 0;
  c = fgetc( fichier );
  ungetc( c, fichier );
  do {
    while ( '#' == c ) {
      c = ignore_ligne( fichier );
      if ( EOF == c ) break;
    }

    /* Ici : ligne pleine, on utilise donc les informations */
    /* a. Le dossier qui contient les groupes de ce système */
    systemes[ i ].dossier = lit_commande( fichier );
#ifdef _DEVELOPPE_
    printf("%s : ligne %lu, dossier = %s\n", __FUNCTION__, i, systemes[ i ].dossier);
#endif

    /* b. Le type de système (une lettre, colonne 2) */
    c = fgetc( fichier );
    while ( isspace( c ) ) {
      c = fgetc( fichier );
      if ( EOF == c ) {
	(void) fprintf(stderr, "%s,%sl.%u : Format de fichier incorrect (1 seule colonne !)\n",
		       __FILE__, __FUNCTION__, __LINE__);
	break;
      }
    }
    systemes[ i ].type = (short int) c;
    while ( !isspace(c ) ) { /* On ignore la suite de la colonne... */
      c = fgetc( fichier );
      if ( EOF == c ) {
	(void) fprintf(stderr, "%s,%sl.%u : Format de fichier incorrect (seulement 2 colonnes !)\n",
		       __FILE__, __FUNCTION__, __LINE__);
	break;
      }
    }

    /* c. Le sous-type de système (une lettre, colonne 3) */
    c = fgetc( fichier );
    while ( isspace( c ) ) {
      c = fgetc( fichier );
      if ( EOF == c ) {
	(void) fprintf(stderr, "%s,%sl.%u : Format de fichier incorrect (seulement 2 colonnes !)\n",
		       __FILE__, __FUNCTION__, __LINE__);
	break;
      }
    }
    systemes[ i ].simple = (short int) c;
    while ( !isspace(c ) ) { /* On ignore la suite de la colonne... */
      c = fgetc( fichier );
      if ( EOF == c ) {
	(void) fprintf(stderr, "%s,%sl.%u : Format de fichier incorrect (seulement 3 colonnes !)\n",
		       __FILE__, __FUNCTION__, __LINE__);
	break;
      }
    }

    /* d. Le nom du système : la fin de la ligne */
    systemes[ i ].nom = lit_ligne( fichier, 0 );
    if ( NULL == systemes[ i ].nom ) {
      (void) fprintf(stderr, "%s,%sl.%u : Format de fichier incorrect (seulement 3 colonnes, le nom manque !)\n",
		     __FILE__, __FUNCTION__, __LINE__);
      break;
    }

    /* e. Le nombre de groupes dans le système : 0 par défaut */
    systemes[ i ].nbr_groupes = 0;
    systemes[ i ].groupes = NULL;
    retour = CRIS_CreerListeGroupes( systemes + i );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%sl.%u : Impossible de créer la liste des groupes, système %s\n",
		     __FILE__, __FUNCTION__, __LINE__, systemes[ i ].nom);
    }

    /* On peut passer au système suivant */
    i++;
  } while ( ( c != EOF ) && ( i < *nbr_systemes ) );

  /* Lecture terminée */
  (void) fclose( fichier );

  return systemes;
}

/* -------- Effacement d'une liste de systèmes cristallins -------- */

int CRIS_EffacerListeSystemes(struct systeme_cristallin *systemes, unsigned long nbr_systemes)
{
  unsigned long i;

  if ( NULL == systemes ) return -1;
  if ( 0 == nbr_systemes ) {
    /* On veut le vider, mais sans libérer [ex. : système "en dur" dans chargement ASCII] */
    CRIS_ViderSysteme( systemes );
    return 0;
  }

  for ( i = 0 ; i < nbr_systemes ; i++ ) {
    CRIS_ViderSysteme( systemes + i );
  }

  /* On libère la table */
  free( systemes );

  /* Libération réussie */
  return 0;
}

/* -------- Effacement du contenu d'un système cristallin -------- */

inline static void CRIS_ViderSysteme(struct systeme_cristallin *systeme)
{
  /* On libère les groupes */
  CRIS_EffacerListeGroupes( systeme );

  /* On libère les textes */
  free( systeme->dossier ); systeme->dossier = NULL;
  free( systeme->nom ); systeme->nom = NULL; 
}

/* -------- Créer la liste des groupes associée à un système -------- */

int CRIS_CreerListeGroupes(struct systeme_cristallin *systeme)
{
  DIR *dossier;
  struct dirent *fichier;
  unsigned long i, nbr_fichiers = 0;

  if ( NULL == systeme ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de système indiqué\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  if ( NULL == systeme->dossier ) {
    AVERTISSEMENT( "Pas de dossier indiqué pour le système %s", systeme->nom );
    return -2;
  }

  /* On ouvre le dossier associé */
  dossier = opendir( systeme->dossier );
  if ( NULL == dossier ) {
    AVERTISSEMENT( "Le dossier indiqué pour le système %s ne peut être ouvert [%s]", systeme->nom, systeme->dossier );
    return +1; /* Erreur, mais mineure */
  }

  /* On compte le nombre de fichiers */
  fichier = readdir( dossier );
  while ( fichier != NULL ) {
    char *nom;
    int retour;
    struct stat infos_fichier;

    /* Chemin relatif du fichier */
    nom = TXT_genere_chaine( "%s/%s", systeme->dossier, fichier->d_name );
    if ( NULL == nom ) {
      (void) fprintf(stderr, "%s,%s l.%u : Problèmes de mémoire / recherche annulée\n",
		     __FILE__, __FUNCTION__, __LINE__);
      (void) closedir( dossier );
      return -3;
    }

    /* Infos sur le fichier */
    retour = stat( nom, &infos_fichier );
    if ( 0 == retour ) {
      if ( S_ISREG( infos_fichier.st_mode ) ) { /* Fichier classique */
	nbr_fichiers++;
      }
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR / fichier %s\n",
		     __FILE__, __FUNCTION__, __LINE__, nom );
    }

    free( nom ); nom = NULL;
    fichier = readdir( dossier );
  }

  /* Il y a un groupe par fichier */
  systeme->nbr_groupes = nbr_fichiers;
  if ( 0 == nbr_fichiers ) { /* Aucun groupe disponible */
    (void) fprintf(stderr, "%s,%sl.%u : Dossier indiqué vide [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, systeme->dossier);
    (void) closedir( dossier );
    return 0;
  }

  /* On réserve donc la liste des groupes */
  systeme->groupes = ( struct groupe_espace * ) malloc( nbr_fichiers * sizeof( struct groupe_espace ) );
  if ( NULL == systeme->groupes ) {
    (void) fprintf(stderr, "%s,%sl.%u : Mémoire insuffisante pour lire les groupes de %s (dossier %s)\n",
		   __FILE__, __FUNCTION__, __LINE__, systeme->nom, systeme->dossier);
    (void) closedir( dossier );
    systeme->nbr_groupes = 0;
    return -3;
  }

  /* Et on l'initialise */
  rewinddir( dossier );

  i = 0;
  fichier = readdir( dossier );
  while ( fichier != NULL ) {
    char *nom;
    int retour;
    struct stat infos_fichier;

    /* Chemin relatif du fichier */
    nom = TXT_genere_chaine( "%s/%s", systeme->dossier, fichier->d_name );
    if ( NULL == nom ) {
      (void) fprintf(stderr, "%s,%s l.%u : Problèmes de mémoire / recherche annulée\n",
		     __FILE__, __FUNCTION__, __LINE__);
      (void) closedir( dossier );
      return -3;
    }

    /* Infos sur le fichier */
    retour = stat( nom, &infos_fichier );
    if ( 0 == retour ) {
      if ( S_ISREG( infos_fichier.st_mode ) ) { /* Fichier classique */
	systeme->groupes[ i ].numero = 0;
	systeme->groupes[ i ].symbole = NULL;
	systeme->groupes[ i ].nom_fichier = strdup( fichier->d_name );
	systeme->groupes[ i ].nbr_operations = 0;
	systeme->groupes[ i ].operations = NULL;
	systeme->groupes[ i ].translations = NULL;
	i++;
      }
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR / fichier %s\n",
		     __FILE__, __FUNCTION__, __LINE__, nom );
    }

    free( nom ); nom = NULL;
    fichier = readdir( dossier );
  }

  /* Plus besoin du dossier */
  (void) closedir( dossier );

  /* On récupère le nom, le nombre d'opérations,... */
  for ( i = 0 ; i < systeme->nbr_groupes ; i++ ) {
    CRIS_LireInfosGroupes( systeme->groupes + i, NULL, systeme->dossier );
  }

  /* On trie les groupes par numéro IUCR croissant (pour faciliter le repérage...) */
  qsort( systeme->groupes, (size_t) systeme->nbr_groupes, sizeof( struct groupe_espace ), CRIS_OrdreGroupes );

  /* Initialisation terminée */
  return 0;
}

/* -------- Comparaison de deux groupes pour le tri -------- */

static int CRIS_OrdreGroupes(const void *v1, const void *v2)
{
  const struct groupe_espace *groupe_1, *groupe_2;

  groupe_1 = ( const struct groupe_espace * ) v1;
  groupe_2 = ( const struct groupe_espace * ) v2;

  if ( groupe_1->numero > groupe_2->numero ) return +1;
  if ( groupe_1->numero < groupe_2->numero ) return -1;
  return 0;
}

/* -------- Effacer la liste des groupes d'un système -------- */

int CRIS_EffacerListeGroupes(struct systeme_cristallin *systeme)
{
  unsigned long i;

  if ( NULL == systeme ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de système indiqué\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }
  if ( NULL == systeme->groupes ) {
    return 0;
  }

  /* On libère les informations associées à chaque groupe */
  for ( i = 0 ; i < systeme->nbr_groupes ; i++ ) {
    free( systeme->groupes[ i ].nom_fichier ); systeme->groupes[ i ].nom_fichier = NULL;
    free( systeme->groupes[ i ].symbole ); systeme->groupes[ i ].symbole = NULL;
    free( systeme->groupes[ i ].operations ); systeme->groupes[ i ].operations = NULL;
    free( systeme->groupes[ i ].translations ); systeme->groupes[ i ].translations = NULL;
  }

  /* On libère la table des groupes */
  free( systeme->groupes ); systeme->groupes = NULL;

  /* Libération terminée */
  return 0;
}

/* -------- Lire les informations associées à un groupe d'espace -------- */

int CRIS_LireInfosGroupes(struct groupe_espace *groupe, const char *racine, const char *dossier)
{
  char *nom_complet;
  FILE *fichier;
  int c;

  if ( NULL == groupe ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de groupe précisé\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( NULL == groupe->nom_fichier ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de fichier associé à ce groupe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* On construit le nom du fichier */
  if ( racine != NULL ) {
    nom_complet = TXT_genere_chaine( "%s/%s/%s", racine, dossier, groupe->nom_fichier );
  } else {
    nom_complet = TXT_genere_chaine( "%s/%s", dossier, groupe->nom_fichier );
  }
  if ( NULL == nom_complet ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible de générer le nom du fichier [dossier %s, racine %s, fichier %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, dossier, racine, groupe->nom_fichier);
    return -2;
  }

  /* On ouvre le fichier */
  fichier = fopen( nom_complet, "rb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible d'ouvrir le fichier \"%s\"\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_complet);
    free( nom_complet );
    return -5;
  }

  /* On ignore les commentaires initaux */
  c = fgetc( fichier );
  if ( EOF == c ) {
    (void) fprintf(stderr, "%s,%s l.%u : Fichier vide [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_complet );
    free( nom_complet );
    (void) fclose( fichier );
    return -48;
  }
  ungetc( c, fichier );
  while ( '#' == c ) {
    c = ignore_ligne( fichier );
    if ( EOF == c ) {
      (void) fprintf(stderr, "%s,%sl.%u : Fin de fichier atteinte prématurément [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, nom_complet);
      free( nom_complet );
      (void) fclose( fichier );
      return -8;
    }
  }

  /* On lit les informations */
  do {
    c = fgetc( fichier ); /* On relit le 1er caractère de la ligne (non-lu par ignore_ligne) */
    if ( 'N' == c ) {
      c = fgetc( fichier );
      if ( 'o' == c ) { /* No = Nombre : nombre d'opérations de symétrie (au total !) */
	while ( c != '=' ) {
	  c = fgetc( fichier );
	  if ( EOF == c ) {
	    goto sortie_EOF;
	  }
	}
	fscanf( fichier, "%u", &(groupe->nbr_operations) );
      } else if ( 'u' == c ) { /* Nu = Numéro, du groupe d'espace selon IUCR */
	while ( c != '=' ) {
	  c = fgetc( fichier );
	  if ( EOF == c ) {
	    goto sortie_EOF;
	  }
	}
	fscanf( fichier, "%d", &(groupe->numero) );
      }
      /* On ignore la fin de la ligne */
      c = ignore_ligne( fichier );
    } else if ( 'S' == c ) { /* S = Symbole, du groupe d'espace */
      while ( c != '=' ) {
	c = fgetc( fichier );
	if ( EOF == c ) {
	  goto sortie_EOF;
	}
      }
      groupe->symbole = lit_ligne( fichier, 0 );
      c = fgetc( fichier ); /* On lit le 1er caractère de la ligne suivante */
      ungetc( c, fichier ); /* sans le faire disparaître, pour être compatible avec ignore_ligne() */
    } else if ( 'T' == c ) { /* T = Translation, nombre de translations */
      while ( c != '=' ) {
	c = fgetc( fichier );
	if ( EOF == c ) {
	  goto sortie_EOF;
	}
      }
      fscanf( fichier, "%u", &(groupe->nbr_translations) );
      /* On ignore la fin de la ligne */
      c = ignore_ligne( fichier );
    }
  } while ( c != '*' );

  /* Ligne d'étoiles : séparateur En-tête / opérations / translations */
  c = ignore_ligne( fichier );

  /* Analyse terminée */
  free( nom_complet );
  (void) fclose( fichier );
  return 0;

  /* Erreur : fin de fichier atteinte en cours d'analyse */
sortie_EOF:
  (void) fprintf(stderr, "%s,%sl.%u : Fin de fichier atteinte prématurément [%s]\n",
		 __FILE__, __FUNCTION__, __LINE__, nom_complet);
  free( nom_complet );
  (void) fclose( fichier );
  return -1975;
}

/* -------- Lire les opérations (et translations) d'un groupe d'espace -------- */

int CRIS_LireOperationsGroupes(struct groupe_espace *groupe, const char *racine, const char *dossier)
{
  char *nom_complet;
  FILE *fichier;
  int c;
  unsigned long idx_ligne, nbr_lignes;
  unsigned int idx_operation, idx_translation;

  if ( NULL == groupe ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de groupe précisé\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( NULL == groupe->nom_fichier ) {
    (void) fprintf(stderr, "%s,%sl.%u : Pas de fichier associé à ce groupe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* On construit le nom du fichier */
  if ( racine != NULL ) {
    nom_complet = TXT_genere_chaine( "%s/%s/%s", racine, dossier, groupe->nom_fichier );
  } else {
    nom_complet = TXT_genere_chaine( "%s/%s", dossier, groupe->nom_fichier );
  }
  if ( NULL == nom_complet ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible de générer le nom du fichier [dossier %s, racine %s, fichier %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, dossier, racine, groupe->nom_fichier);
    return -2;
  }

  /* On ouvre le fichier */
  fichier = fopen( nom_complet, "rb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible d'ouvrir le fichier \"%s\"\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_complet);
    free( nom_complet );
    return -5;
  }

  /* On ignore les commentaires initiaux et l'en-tête, supposé déjà lus */
  c = fgetc( fichier );
  while ( c != '*' ) {
    c = ignore_ligne( fichier );
    if ( EOF == c ) {
      goto sortie_EOF;
    }
  }
  /* On ignore la ligne de séparation */
  c = ignore_ligne( fichier );
  if ( EOF == c ) {
    goto sortie_EOF;
  }

  /* Nombre de lignes */
  nbr_lignes = compte_lignes( fichier, '#' );
  if ( 0 == nbr_lignes ) {
    (void) fprintf(stderr, "%s,%sl.%u : Aucune opération dans le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_complet);
    free( nom_complet );
    (void) fclose( fichier );
    return -21;
  }
  if ( nbr_lignes != ( groupe->nbr_operations + groupe->nbr_translations ) ) {
    AVERTISSEMENT( "Incohérence dans %s : %u lignes attendues (%u opérations, %u translations), mais %lu trouvées",
		   nom_complet, groupe->nbr_operations + groupe->nbr_translations, groupe->nbr_operations, groupe->nbr_translations, nbr_lignes );
    free( nom_complet );
    (void) fclose( fichier );
    return -21;
  }

  /* On crée la table des opérations */
  groupe->operations = (float *) calloc( 16 * groupe->nbr_operations, sizeof( float ) );
  if ( NULL == groupe->operations ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible de construire la table des opérations\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( nom_complet );
    (void) fclose( fichier );
    return -48;
  }

  /* On crée la table des translations */
  groupe->translations = (float *) calloc( 16 * groupe->nbr_translations, sizeof( float ) );
  if ( NULL == groupe->translations ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible de construire la table des translations\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( groupe->operations ); groupe->operations = NULL;
    free( nom_complet );
    (void) fclose( fichier );
    return -48;
  }

  /* Pour chaque ligne, on lit et on analyse */
  for ( idx_ligne = 0, idx_operation = 0, idx_translation = 0 ; idx_ligne < nbr_lignes ; idx_ligne++ ) {
    int retour;

    c = fgetc( fichier );
    ungetc( c, fichier );

    if ( '(' == c ) {
      retour = CRIS_AnalyseTranslation( fichier, groupe->translations + 3 * idx_translation );
      idx_translation++;
    } else {
      retour = CRIS_AnalyseOperation( fichier, groupe->operations + 16 * idx_operation );
      idx_operation++;
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%sl.%u : ERREUR dans l'analyse de la ligne [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      free( nom_complet );
      (void) fclose( fichier );
      return -125;
    }
  }

  /* L'analyse s'est faite correctement */
  free( nom_complet );
  (void) fclose( fichier );
  return 0;

  /* Erreur : fin de fichier atteinte en cours d'analyse */
sortie_EOF:
  (void) fprintf(stderr, "%s,%sl.%u : Fin de fichier atteinte prématurément [%s]\n",
		 __FILE__, __FUNCTION__, __LINE__, nom_complet);
  free( nom_complet );
  (void) fclose( fichier );
  return -1974;
}

/* -------- Composante après une opération (= ligne d'une matrice 4 * 4 -------- */

static int CRIS_AnalyseCoordonnee(char *coordonnee, float *ligne_matrice)
{
  char const *c;
  double coefficient, numerateur = 0.0;

  /* On initialise les coefficients à zéro */
  ligne_matrice[ 0 ] = 0.0; /* x */
  ligne_matrice[ 1 ] = 0.0; /* y */
  ligne_matrice[ 2 ] = 0.0; /* z */
  ligne_matrice[ 3 ] = 0.0; /* translation */

  c = coordonnee;
  coefficient = 0.0;
  /* Version actuelle : juste des formules simples à coefficients entiers + fraction pour la translation, comme dans les tables UICR */
  while ( *c != 0 ) {
    if ( '-' == *c ) {
      if ( coefficient != 0.0 ) { /* C'est qu'il y a une translation */
	if ( numerateur != 0.0 ) { /* C'est une fraction qui la définit */
	  ligne_matrice[ 3 ] = numerateur / coefficient;
	} else {
	  ligne_matrice[ 3 ] = coefficient;
	}
      }
      coefficient = -1.0;
      numerateur = 0.0;
    } else if ( '+' == *c ) {
      if ( coefficient != 0.0 ) { /* C'est qu'il y a une translation */
	if ( numerateur != 0.0 ) { /* C'est une fraction qui la définit */
	  ligne_matrice[ 3 ] = numerateur / coefficient;
	} else {
	  ligne_matrice[ 3 ] = coefficient;
	}
      }
      coefficient = +1.0;
      numerateur = 0.0;
    } else if ( ( 'x' == *c ) || ( 'X' == *c ) ) {
      if ( 0.0 == coefficient ) /* Si x en début de formule */
	coefficient = 1.0;
      ligne_matrice[ 0 ] = coefficient;
      coefficient = 0.0; /* On recommence */
      numerateur = 0.0;
    } else if ( ( 'y' == *c ) || ( 'Y' == *c ) ) {
      if ( 0.0 == coefficient ) /* Si y en début de formule */
	coefficient = 1.0;
      ligne_matrice[ 1 ] = coefficient;
      coefficient = 0.0; /* On recommence */
      numerateur = 0.0;
    } else if ( ( 'z' == *c ) || ( 'Z' == *c ) ) {
      if ( 0.0 == coefficient ) /* Si z en début de formule */
	coefficient = 1.0;
      ligne_matrice[ 2 ] = coefficient;
      coefficient = 0.0; /* On recommence */
      numerateur = 0.0;
    } else if ( '/' == *c ) {
      numerateur = coefficient;
      coefficient = 1.0;
    } else if ( isdigit( *c ) || ( '.' == *c ) ) {
      char const *pos_suite;

      coefficient = TXT_CNV_vNombre_D( c, &pos_suite, NULL );
      c = pos_suite - 1;
    } else if ( ' ' == *c ) { /* Espaces ignorés */
    } else {
      (void) fprintf(stderr, "%s,%sl.%u : Caractère inconnu, %c [%s, reste %s]\n",
		     __FILE__, __FUNCTION__, __LINE__, *c, coordonnee, c);
      return -19;
    }

    c++;
  }
  if ( coefficient != 0.0 ) { /* Terme de translation à la fin */
    if ( numerateur != 0.0 ) {
      ligne_matrice[ 3 ] = numerateur / coefficient;
    } else {
      ligne_matrice[ 3 ] = coefficient;
    }
  }

#ifdef _DEVELOPPE_
  printf("%s : %s = ", __FUNCTION__, coordonnee);
  if ( ligne_matrice[ 0 ] != 0.0 )
    printf("%g * x + ", ligne_matrice[ 0 ]);
  if ( ligne_matrice[ 1 ] != 0.0 )
    printf("%g * y + ", ligne_matrice[ 1 ]);
  if ( ligne_matrice[ 2 ] != 0.0 )
    printf("%g * z + ", ligne_matrice[ 2 ]);
  if ( ligne_matrice[ 3 ] != 0.0 )
    printf("%g", ligne_matrice[ 3 ]);
  printf("\n");
#endif

  return 0;
}

/* -------- Ligne décrivant une opération : x, y, z -------- */

static int CRIS_AnalyseOperation(FILE *fichier, float *matrice)
{
  char *ligne;
  char **coords;
  unsigned int i;
  unsigned long nbr_coords;

  /* On lit la ligne entière */
  ligne = lit_ligne( fichier, 0 );
  if ( NULL == ligne ) {
    (void) fprintf(stderr, "%s,%sl.%u : Erreur en voulant lire la ligne de translation\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Séparation des coordonnées */
  coords = TXT_DecomposerChaine( ligne, ',', &nbr_coords );
  if ( NULL == coords ) {
    (void) fprintf(stderr, "%s,%sl.%u : La décomposition de la ligne \"%s\" a échoué\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne);
    free( ligne );
    return -5;
  }
  if ( nbr_coords < 3 ) {
    (void) fprintf(stderr, "%s,%sl.%u : La ligne \"%s\" ne comprend pas 3 coordonnées !\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne);
    free( coords[ 0 ] );
    free( coords );
    free( ligne );
    return -5;
  }

  /* Analyse coordonnée par coordonnée */
  for ( i = 0 ; i < 3 ; i++ ) {
    int retour;

    retour = CRIS_AnalyseCoordonnee( coords[ i ], matrice + 4 * i );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%sl.%u : ERREUR dans l'analyse de la coordonnée %s [ligne %s]\n",
		     __FILE__, __FUNCTION__, __LINE__, coords[ i ], ligne );
      free( coords[ 0 ] );
      free( coords );
      free( ligne );
      return -124;
    }
  }

  /* Dernière ligne : 0.0, 0.0, 0.0, 1.0 [pourrait être supprimée !] */
  matrice[ 12 ] = 0.0;
  matrice[ 13 ] = 0.0;
  matrice[ 14 ] = 0.0;
  matrice[ 15 ] = 1.0;

  /* Analyse terminée */
  free( coords[ 0 ] );
  free( coords );
  free( ligne );

  /* Sans erreur */
  return 0;
}

/* -------- Ligne décrivant une translation : (x, y, z) ou (x, y, z)+, forcément numérique mais 1/2 possible ! -------- */

static int CRIS_AnalyseTranslation(FILE *fichier, float *vecteur)
{
  char *ligne;
  char **coords, *valeur;
  unsigned int i;
  unsigned long nbr_coords;

  /* On lit la ligne entière */
  ligne = lit_ligne( fichier, 0 );
  if ( NULL == ligne ) {
    (void) fprintf(stderr, "%s,%sl.%u : Erreur en voulant lire la ligne de translation\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Séparation des coordonnées */
  coords = TXT_DecomposerChaine( ligne + 1, ',', &nbr_coords );
  if ( NULL == coords ) {
    (void) fprintf(stderr, "%s,%sl.%u : La décomposition de la ligne \"%s\" a échoué\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne);
    free( ligne );
    return -5;
  }
  if ( nbr_coords < 3 ) {
    (void) fprintf(stderr, "%s,%sl.%u : La ligne \"%s\" ne comprend pas 3 coordonnées !\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne);
    free( coords[ 0 ] );
    free( coords );
    free( ligne );
    return -5;
  }

  /* Analyse coordonnée par coordonnée */
  for ( i = 0 ; i < 3 ; i++ ) {
    char *pos_fraction;

    valeur = coords[ i ];

    pos_fraction = strchr( valeur, '/' );
    if ( NULL == pos_fraction ) { /* Valeur entière */
      vecteur[ i ] = (float) atof( valeur );
    } else {
      double num, den;

      *pos_fraction = 0;
      pos_fraction++;

      num = atof( valeur );
      den = atof( pos_fraction );
      if ( 0.0 == den ) {
	(void) fprintf(stderr, "%s,%sl.%u : Dénominateur absurde [ligne %s : %s / %s]\n",
		       __FILE__, __FUNCTION__, __LINE__, ligne, valeur, pos_fraction);
	vecteur[ i ] = 0.0;
      } else {
	vecteur[ i ] = (float) ( num / den );
      }
    }
  }

  /* Analyse terminée */
  free( coords[ 0 ] );
  free( coords );
  free( ligne );

#ifdef _DEVELOPPE_
  printf("%s : Translation de vecteur (%g, %g %g)\n", __FUNCTION__, vecteur[ 0 ], vecteur[ 1 ], vecteur[ 2 ]);
#endif

  /* Sans erreur */
  return 0;
}
