/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998, Emmanuel CURIS
   Cration : le 27 novembre 1998

   Module de cration d'un modle structural
   Gestion des fragments

   Un fragment est une runion d'atomes ayant un sens chimique
   ou une slection d'un ensemble d'atomes pour une raison quelconque.

   Les fragments sont runis entre eux pour former une molcule.

   Exemple de fragment :
   -> la molcule complte dans le cas de l'aspirine
   -> un ligand d'un complexe, p. ex. un imidazole dans ZnAc2Im2
   -> un acide amin d'une chane peptidique

   Les liaisons chimiques entre atomes d'un mme fragment sont dcrites
   par les liaisons dfinies dans le fragment.

   Les liaisons chimiques entre atomes de fragments diffrents sont dcrites
   par les liaisons dfinies dans la molcule qui contient le fragment.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "atomes.h"
#include "graphes.h" /* Pour GRF_efface() */

#include "atm_fragments.txt"

/* ~~~~~~~~ Variables locales  ce module ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales  ce module ~~~~~~~~ */

static int FRG_DelierAtomes(Fragment *fragment, Molecule *molecule);
static int defait_chainage(Molecule *molecule, Fragment *fragment);

static int Liaison_DevientInterfragments(Molecule *molecule, Fragment *fragment, Liaison *liaison);

static unsigned int construit_ligand(Molecule *molecule, Atome *central);

/* ~~~~~~~~ Ralisation pratique du module ~~~~~~~ */

/* -------- Cration d'un nouveau fragment ------------------------\
 |   Renvoie l'adresse du fragment cr                            |
 |           NULL en cas d'chec                                   |
 |                                                                 |
 | La fonction s'occupe du chanage en fin de chane               |
 | si premier != NULL, sinon le fragment est isol                 |
 |                                                                 |
 | molecule     : la molcule dans laquelle insrer le fragment    |
 | avant        : le fragment avant dans la liste considre       |
 | nom          : le nom  donner au fragment cr                 |
 | nbr_atomes   : le nombre d'atomes que peut contenir le fragment |
 | absorbeur    : l'atome absorbeur (NULL si aucun)                |
 | signal       : le signal associ (NULL si aucun)                |
 \---------------------------------------------------------------- */

Fragment *FRG_creer(Molecule *molecule, Fragment *avant, char *nom, unsigned int nbr_atomes,
		    Atome *absorbeur, struct graphe *signal)
{
  Fragment *nv_frag;

  if (nbr_atomes < 1) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "Aucun atome pour le fragment !\n");
#endif
    return NULL;
  }

  nv_frag = (Fragment *) malloc( sizeof(Fragment) );
  if (nv_frag == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "fragments.c", "FRG_creer", "nv_frag");
    return NULL;
  }

  /* Prparation du fragment */
  if (nom == NULL)
    strcpy(nv_frag->nom, "Fragment sans nom");
  else {
    strncpy(nv_frag->nom, nom, LG_MAX_NOM_FRAGMENT);
    (nv_frag->nom)[LG_MAX_NOM_FRAGMENT] = 0; /* Pour tre sr de finir la chane */
  }

  nv_frag->atomes = (Atome **) calloc(nbr_atomes, sizeof(Atome *));
  if (nv_frag->atomes == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "fragments.c", "FRG_creer", "atomes");
    free(nv_frag);
    return NULL;
  }
  nv_frag->nbr_atomes = 0; /* Encore aucun atome */
  nv_frag->max_atomes = nbr_atomes; /* Mais de la place pour ceux-ci */

  nv_frag->liaisons = NULL;
  nv_frag->nbr_liaisons = 0;            /* Encore aucune liaison */

  nv_frag->absorbeur = absorbeur;
  if ( absorbeur != NULL ) { /* Il y a un absorbeur */
    nv_frag->pos_barycentre = absorbeur->position; /* Position du barycentre = position de l'absorbeur */
    nv_frag->signal = signal; /* Le signal d'absorption associ */
  } else {
    nv_frag->pos_barycentre.x = 0.0; nv_frag->pos_barycentre.y = 0.0; nv_frag->pos_barycentre.z = 0.0;
    nv_frag->signal = NULL; /* Aucun signal associ, forcment */
  }

  nv_frag->translation = NULL; /* Pas de mouvement associ */
  nv_frag->rotation = NULL;

  nv_frag->drapeaux.maille = FAUX; /* Par dfaut, ce n'est pas une maille */
  nv_frag->drapeaux.origine = FAUX;
  nv_frag->drapeaux.ligand = FAUX; /* Ni un ligand */

  nv_frag->numero = 0;

  /* Le chanage */
  if (NULL == molecule) { /* Fragment isol */
    if (NULL == avant) {
      nv_frag->suivant = NULL; /* De toute faon, sera plac en dernier... */
    } else {
      nv_frag->suivant = avant->suivant;
      avant->suivant = nv_frag;
    }
  } else {
    if (NULL == molecule->fragments) { /* Premier de la liste */
      molecule->fragments = nv_frag;
      nv_frag->suivant = NULL;
    } else if (NULL == avant) { /* On l'insre en dbut de liste */
      nv_frag->suivant = molecule->fragments;
      molecule->fragments = nv_frag;
    } else {
      Fragment *frg_avant;

      for (frg_avant = molecule->fragments ; frg_avant != avant ; frg_avant = frg_avant->suivant) {
	if (NULL == frg_avant) {
	  (void) fprintf(stderr, "FRG_creer : le fragment indiqu [%s] n'est pas dans la molcule indique [%s]!\n",
			 frg_avant->nom, molecule->nom);
#ifdef _DEVELOPPE_
	  abort();
#endif
	  break;
	}
      }
      nv_frag->suivant = avant->suivant;
      avant->suivant = nv_frag;
    }

    molecule->nbr_fragments++;
  }

  return nv_frag; /* Cration termine et russie */
}

/* -------- Effacer un fragment ------------------------------------------\
 |   Renvoie <0 si erreur, FAUX si on n'a pas effac, VRAI si on a effac |
 |                                                                        |
 | molecule : la molcule contenant le fragment [NULL : fragment isol]   |
 | fragment : le fragment  effacer                                       |
 | forcer   : VRAI si on ne demande pas de confirmation                   |
 | eff_atm  : VRAI s'il faut aussi effacer les atomes qui le composent    |
 \----------------------------------------------------------------------- */

int FRG_effacer(Molecule *molecule, Fragment *fragment, boolean forcer, boolean effacer_atomes)
{
  if (fragment == NULL) return -1;
  if (forcer == FAUX) {
    int reponse;

    if ( VRAI == effacer_atomes )
      reponse = BI_ERR_alerte( NULL, &Q_EffacerFragment_Atomes, VRAI, fragment->nom );
    else
      reponse = BI_ERR_alerte( NULL, &Q_EffacerFragment, VRAI, fragment->nom );
    if ( 2 == reponse )
      return FAUX;
  }

#ifdef _DEVELOPPE_
  printf("    fragment : on efface  %s \n", fragment->nom);
#endif

  /* On efface les liaisons du fragment */
  while (fragment->liaisons != NULL)
    effacer_liaison(fragment->liaisons, VRAI);

  if ( NULL != molecule ) { /* Il appartient bien  une molcule */
    /* On dfait le chanage */
    defait_chainage(molecule, fragment );

    /* On s'occupe de mettre  jour les atomes */
    FRG_DelierAtomes(fragment, molecule);
  }

  /* Si ncessaire, on efface les atomes qui le constituent */
  if (effacer_atomes == VRAI) {
    unsigned int i;

    for (i = 0 ; i < fragment->nbr_atomes ; i++) {
#ifdef _DEVELOPPE_
      printf("%d... ", i);
#endif
      ATM_efface(molecule, fragment->atomes[i], VRAI);
      fragment->atomes[i] = NULL;
    }

    if (fragment->absorbeur) {
#ifdef _DEVELOPPE_
      printf("Absorbeur... ");
#endif
      ATM_efface(NULL, fragment->absorbeur, VRAI);
      GRF_efface(fragment->signal, FAUX, FAUX);
    }
  }

  /* On efface le fragment proprement dit */
  free(fragment->atomes);
  free(fragment->liaisons);
  free(fragment);
  return VRAI;
}


/* -------- te le fragment indiqu de la molcule indique --------
     Renvoie VRAI si la molcule contient encore des fragments,
             FAUX sinon
*/

static int defait_chainage(Molecule *molecule, Fragment *fragment)
{
  Fragment *avant;

  avant = molecule->fragments;
  if (avant == fragment) { /* On efface le premier fragment de la molcule */
#ifdef _DEVELOPPE_
    printf("   Effacement du premier fragment de %s\n", molecule->nom);
#endif
    molecule->fragments = fragment->suivant; /* On passe au fragment suivant */
  } else if (avant == NULL) {
#ifdef _DEVELOPPE_
    printf("Le fragment %s n'appartient pas  la molcule %s\n", fragment->nom, molecule->nom);
    abort();
#endif
    return VRAI;
  } else {
    while (avant->suivant != fragment) {
      avant = avant->suivant;
      if (avant == NULL) {
#ifdef _DEVELOPPE_
	fprintf(stderr, "defait_chainage : le fragment %s n'appartient pas  la molcule %s\n",
		fragment->nom, molecule->nom);
#endif
	return VRAI;
      }
    }
    avant->suivant = fragment->suivant; /* Liste simplement chane */
  }
  molecule->nbr_fragments--; /* Un fragment de moins */

#ifdef _DEVELOPPE_
  printf("   Effacement d'un fragment de %s, fragments restants\n", molecule->nom);
  for (avant = molecule->fragments ; avant != NULL ; avant = avant->suivant) {
    printf("    fragment %s\n", avant->nom);
  }
#endif

  return (molecule->nbr_fragments > 0);
}

static int FRG_DelierAtomes(Fragment *fragment, Molecule *molecule)
{
  Atome *atome;

  if (NULL == molecule->premier) return -1; /* Plus d'atomes dans la molcule */

  for ( atome = molecule->premier ; atome != NULL ; atome = atome->suivant ) {
    if ( fragment == atome->fragment )
      atome->fragment = NULL;

    if ( VRAI == atome->marques.multi_frg ) {
      fprintf(stderr, "FRG_DelierAtomes : ATOME MULTI-FRAGMENT,  CONTRLER\n");
    }
  }

  return 0;
}

/* -~-~-~-~-~-~-~-~ La gestion des liaisons dans le fragment -~-~-~-~-~-~-~-~ */

/* -------- Dplacer une liaison, pour qu'elle devienne entre-fragments --------\
 |   Renvoie 0 si tout s'est bien pass, != 0 sinon                             |
 |                                                                              |
 | molecule : la molcule concerne                                             |
 | fragment : le fragment dont on doit extraire la liaison                      |
 | liaison  : la liaison qu'il faut dplacer                                    |
 \----------------------------------------------------------------------------- */

static int Liaison_DevientInterfragments(Molecule *molecule, Fragment *fragment, Liaison *liaison)
{
  if ( NULL == molecule ) return -1;
  if ( NULL == fragment ) return -1;
  if ( NULL == liaison  ) return -1;

  /* On isole la liaison du fragment */
  if ( fragment->liaisons == liaison ) { /* C'est la premire */
    fragment->liaisons = liaison->suivante;
  } else {
    Liaison *avant;

    /* On cherche la liaison juste avant dans le chanage */
    for (avant = fragment->liaisons ; avant->suivante != liaison ; avant = avant->suivante) {
      if (NULL == avant) {
	(void) fprintf(stderr, "Liaison_DevientInterfragments : liaison pas dans le fragment indiqu !\n");
	return -2;
      }
    }
    avant->suivante = liaison->suivante;
  }
  fragment->nbr_liaisons--; /* Une liaison de moins, forcment */
  liaison->suivante = NULL; /* Celle-ci est isole */

  /* On insre la liaison dans la molcule, en fin de chane */
  if ( NULL == molecule->liaisons ) {
    molecule->liaisons = liaison;
  } else {
    Liaison *avant;

    /* On cherche la liaison juste avant dans le chanage */
    for (avant = molecule->liaisons ; avant->suivante != NULL ; avant = avant->suivante);
    avant->suivante = liaison;
  }
  molecule->nbr_liaisons++; /* Une liaison de plus, forcment */

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Ajouter un atome dans un fragment --------\
 |   Renvoie un code <0 en cas d'erreur               |
 |                                                    |
 | fragment   : le fragment o ajouter des atomes     |
 | nbr_atomes : le nombre d'atomes  ajouter          |
 | atome      : les atomes  ajouter (leurs adresses) |
 \--------------------------------------------------- */

int FRG_ajoute_atome(Fragment *fragment, unsigned int nbr_atomes, Atome **atome)
{
  unsigned int i;

  if ( (fragment == NULL) || (atome == NULL) || (nbr_atomes == 0) )
    return -1;

  if (fragment->max_atomes < fragment->nbr_atomes + nbr_atomes) /* On risque de dborder */
  {
    Atome **nv_atomes;

    nv_atomes = (Atome **) realloc(fragment->atomes, sizeof(Atome *) * (fragment->nbr_atomes + nbr_atomes));
    if (nv_atomes == NULL)
    {
      SYS_ERREUR(MANQUE_MEMOIRE, "fragments.c", "FRG_ajoute_atome", "nv_atomes");
      return -2;
    }

    fragment->max_atomes = fragment->nbr_atomes + nbr_atomes;
    fragment->atomes = nv_atomes;
  }

  for (i = 0 ; i < nbr_atomes ; i++)
  {
    if ( atome[i] != NULL )
    {
      fragment->atomes[fragment->nbr_atomes] = atome[i];
      (fragment->nbr_atomes)++; /* Un atome de plus */

      if (atome[i]->fragment == NULL) /* Aucun fragment associ  cet atome */
	atome[i]->fragment = fragment;
      else
	atome[i]->marques.multi_frg = VRAI; /* Il appartient  plusieurs fragments */
    }
#ifdef _DEVELOPPE_
    else
    {
      printf("L'atome %u de la liste est  NULL !\n", i);
    }
#endif
  }

  return 0; /* Tout s'est bien pass */
}

/* -------- Enlever un atome d'un fragment ---------------------\
 |  Renvoie un code <0 en cas d'erreur                          |
 |          VRAI si on a effac l'atome du fragment             |
 |          FAUX si l'atome n'a pas t trouv dans le fragment |
 |                                                              |
 | fragment = celui o il faut effacer l'atome                  |
 | atome    = l'atome  ter du fragment                        |
 | forcer   = VRAI si on ne demande aucune confirmation         |
 \------------------------------------------------------------- */

int FRG_efface_atome(Fragment *fragment, Atome *atome, boolean forcer)
{
  Molecule *molecule;
  unsigned int i, j, nv_nbr_atomes;

  if (fragment == NULL) return 0;
  if (atome == NULL) return 0;

  if ( FAUX == forcer ) {
    int reponse;

    reponse = BI_ERR_alerte( NULL, &Q_FRG_EffacerAtome, VRAI, fragment->nom,
			     ATM_Z_vers_nom( atome->num_atomique ), atome->position.x, atome->position.y, atome->position.z );
    if ( 2 == reponse ) {
      return FAUX;
    }
  }

  nv_nbr_atomes = fragment->nbr_atomes;
#ifdef _DEVELOPPE_
  printf("        fragment : [%s] on te le %s en (%f, %f, %f)\n", fragment->nom,
	 ATM_donnees[atome->num_atomique - 1].nom, atome->position.x, atome->position.y, atome->position.z);
#endif

  /* On vrifie que ce n'est pas l'absorbeur */
  if (atome == fragment->absorbeur) {
#ifdef _DEVELOPPE_
    printf("          absorbeur ; ");
#endif
    fragment->absorbeur= NULL;
    /* Effacer le graphe associ ? */
    fragment->signal = NULL;
  }

  /* On vrifie ensuite dans la table */
  for (i = 0, j = 0 ; i < fragment->nbr_atomes ; i++) {
    if (fragment->atomes[i] == atome) { /* Il faut donc l'ter */
#ifdef _DEVELOPPE_
      printf("%d, ", i);
#endif
      nv_nbr_atomes--; /* Nouveau nombre d'atomes dans le fragment */
    } else {
      fragment->atomes[j] = fragment->atomes[i];
      j++;
    }
  }
  if (nv_nbr_atomes == fragment->nbr_atomes) { /* Atome non trouv */
#ifdef _DEVELOPPE_
    printf("non trouv\n");
#endif
    return FAUX;
  }
#ifdef _DEVELOPPE_
  else
    printf("soit %d copie(s) {restent %d atomes}\n", fragment->nbr_atomes - nv_nbr_atomes, nv_nbr_atomes);
#endif

  for (j = nv_nbr_atomes ; j < fragment->nbr_atomes ; j++)
    fragment->atomes[j] = NULL; /* Ces atomes ne sont plus utiliss */

  fragment->nbr_atomes = nv_nbr_atomes;
  molecule = FRG_QuelleMolecule(fragment);
  if (atome->fragment == fragment) {
    atome->fragment = NULL; /* L'atome n'appartient plus au fragment, par dfinition... */

    if (VRAI == atome->marques.multi_frg) { /* Quel est le nouveau fragment ? */
      unsigned int nbr_fragments;

      atome->fragment = ATM_QuelsFragments(atome, molecule, &nbr_fragments, NULL);
      if (nbr_fragments < 2)
	atome->marques.multi_frg = FAUX;
    }
  } else { /* Il devait donc tre marqu comme multi_frg */
    unsigned int nbr_fragments;

    (void) ATM_QuelsFragments(atome, molecule, &nbr_fragments, NULL);
    if (nbr_fragments < 2) /* Ce n'est plus la peine */
      atome->marques.multi_frg = FAUX;
  }

  if (fragment->nbr_atomes == 0) { /* Il n'y a plus aucun atome dans le fragment */
    FRG_effacer(molecule, fragment, VRAI, FAUX); /* On l'efface donc */
    return VRAI; /* On aura aussi effac les liaisons... */
  }

  /* On dplace les liaisons si ncessaire */
  if ( ( fragment->liaisons != NULL ) && ( atome->nbr_liaisons > 0 ) ) {
    boolean ok;

    do {
      Liaison *liaison;
      ok = VRAI;

      /* On parcourt la liste des liaisons */
      liaison = fragment->liaisons;
      while ( liaison != NULL ) {
	if ( ATM_EstDansLiaison(atome, liaison) == VRAI ) { /* Liaison dans le fragment contient l'atome */
	  Liaison_DevientInterfragments(molecule, fragment, liaison);
	  ok = FAUX;
	  break;
	}
      }
    } while ( FAUX == ok );
  }

  return VRAI; /* On a effac l'atome et tout s'est bien pass */
}

/* -------- Recherche automatique des ligands dans une molcule
   Renvoie le nombre de fragments ainsi crs

   Part du principe que la molcule tudie est un complexe, donc sous-entendu
 de mtal de transition (voire d'un mtal tout court).

 molecule = la molcule qu'il faut traiter
 remplace : VRAI => on dtruit les anciens fragments
            FAUX => les fragments trouvs se rajoutent aux prcdents

Remarque : la marque mk_1 indique que l'atome est en cours de traitement
           la marque mk_2 indique que l'atome a t compltement trait

 -------- */

int FRG_chercher_ligands(Molecule *molecule)
{
  unsigned int i;
  Atome *atome_central = NULL;

  if (molecule == NULL) return 0;

  /* On commence par chercher un atome potentiellement au coeur du complexe
     C'est--dire un atome mtallique, de type mtal de transition */

  /* On cherche d'abord dans les absorbeurs */
  for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
    if ( VRAI == ZATM_forme_complexes(molecule->absorbeurs[i].absorbeur->num_atomique) ) {
      atome_central = molecule->absorbeurs[i].absorbeur;
      break;
    }
  }

  /* Si on n'a pas trouv : on cherche dans les fragments */
  if ( NULL == atome_central ) {
    Fragment *fragment;

    for (fragment = molecule->fragments ; fragment != NULL ; fragment = fragment->suivant) {
      for (i = 0 ; i < fragment->nbr_atomes ; i++) {
	if ( VRAI == ZATM_forme_complexes(fragment->atomes[i]->num_atomique) ) {
	  atome_central = fragment->atomes[i];
	  break;
	}
      }
    }
  }

  /* Si on n'a toujours pas trouv : pas d'atome central -> on annule ! */
  if ( NULL == atome_central ) {
    (void) fprintf(stderr, "FRG_cherche_ligands : pas d'atome formant potentiellement un complexe\n");
    return 0;
  }
  if ( 0 == atome_central->nbr_coval ) {
    (void) fprintf(stderr, "FRG_cherche_ligands : atome central li  aucun autre !\n");
    return 0;
  }

  /* On peut donc dmarrer. On met toutes les marques  zro. */
  MOL_tt_demarquer(molecule);

  printf("Construction des ligands en cours... Atome central : %d (%f, %f, %f)\n",
	 atome_central->num_atomique, atome_central->position.x, atome_central->position.y, atome_central->position.z);
  return construit_ligand(molecule, atome_central);
}

/* -------- Partie essentielle de la construction d'un ligand -------- */

static Atome *repere_ligands(Molecule *molecule, Atome *central)
{
  Fragment *fragment;
  Liaison *liaison;

  /* On cherche d'abord dans les liaisons entre fragments */
  for (liaison = molecule->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
    Atome *atm_ligand = NULL;

    if ( liaison->type != LIAISON_COVALENTE)
      continue; /* Seules les liaisons covalentes sont intressantes */

    if ( liaison->atome_1 == central )
      atm_ligand = liaison->atome_2;
    else if (liaison->atome_2 == central )
      atm_ligand = liaison->atome_1;

    if (NULL == atm_ligand) continue; /* Liaison ne concerne pas l'atome ! */
    if ( atm_ligand->marques.mk_1 == FAUX ) /* N'a pas dj t trait */
      return atm_ligand;
  }

  /* On cherche maintenant dans les liaisons intra-fragment */
  for (fragment = molecule->fragments ; fragment != NULL ; fragment = fragment->suivant) {
    for (liaison = fragment->liaisons ; liaison ; liaison = liaison->suivante) {
      Atome *atm_ligand = NULL;

      if ( liaison->type != LIAISON_COVALENTE)
	continue; /* Seules les liaisons covalentes sont intressantes */

      if ( liaison->atome_1 == central )
	atm_ligand = liaison->atome_2;
      else if (liaison->atome_2 == central )
	atm_ligand = liaison->atome_1;

      if (NULL == atm_ligand) continue; /* Liaison ne concerne pas l'atome ! */
      if ( atm_ligand->marques.mk_1 == FAUX ) /* N'a pas dj t trait */
	return atm_ligand;
    }
  }

  return NULL; /* Aucun atome li non trait repr */
}

/* -------- Ajoute les atomes lis  l'atome indiqu au fragment indiqu
   ide : appel rcursif pour reconstruire un ligand

   Si l'un des atomes lis appartient dj  un fragment de type ligand, les deux
fragments sont fusionns. La fonction renvoie donc le fragment final...
   On ne suit que les liaisons covalentes.
-------- */

static Fragment *refait_ligand(Molecule *molecule, Fragment *fragment, Atome *atome)
{
  unsigned int i;

  if (atome->nbr_coval < 1) /* Pas de liaison pour cet atome ! */
    return fragment;
  if (atome->marques.mk_2 == VRAI) /* Atome compltement trait */
    return fragment;

  printf(" atome trait : %d (%f, %f, %f)\n",
	 atome->num_atomique, atome->position.x, atome->position.y, atome->position.z);

  /* A AMELIORER : actuellement, fait plusieurs fois la boucle : sans intrt...
   * Il faudrait plutt faire un passage continu, pour chaque atome trouv vrifi
   * s'il appartient dj  un fragment, si non l'ajouter [si oui, fusionner]
   */

  for (i = 0 ; i < atome->nbr_coval ; i++) {
    Atome *lie;

    lie = repere_ligands(molecule, atome);
    if (NULL == lie)
      return fragment;

    if ( VRAI == ZATM_forme_complexes(lie->num_atomique) ) { /* Complexe polymtallique */
      /* A FAIRE : refaire une recherche exhaustive des ligands... */
    } else {
      FRG_ajoute_atome(fragment, 1, &lie);
      lie->marques.mk_1 = VRAI; /* On dmarre son traitement */
      fragment = refait_ligand(molecule, fragment, lie);
      lie->fragment = fragment;
      lie->marques.mk_2 = VRAI; /* Traitement de cet atome termin */
    }
  }

  return fragment;
}

/* Renvoie le nombre de fragments reprs */

static unsigned int construit_ligand(Molecule *molecule, Atome *central)
{
  unsigned int i;

  if (NULL == central) return 0;

  for (i = 0 ; i < central->nbr_coval ; i++) {
    Atome *lie;
    Fragment *ligand;
    char nom_ligand[100];

    lie = repere_ligands(molecule, central);
    if (NULL == lie) {
#ifdef _DEVELOPPE_
      printf("Plus de ligand trouv : %d fragments sur %d prvus - ligands polydentes ?\n",
	     i, central->nbr_liaisons);
#endif
      return i; /* Nombre de ligands reprs */
    }

    sprintf(nom_ligand, "Ligand %u", i + 1);
    ligand = FRG_creer(molecule, molecule->fragments, nom_ligand, 10, NULL, NULL);
    if (NULL == ligand) {
#ifdef _DEVELOPPE_
      fprintf(stderr, "Impossible de crer le fragment %s\n", nom_ligand);
#endif
      return i;
    }

    FRG_ajoute_atome(ligand, 1, &lie);
    lie->marques.mk_1 = VRAI; /* Cet atome est en cours de traitement */
    ligand->drapeaux.ligand = VRAI;

    /* Maintenant, on reconstitue le ligand en suivant les liaisons */
    ligand = refait_ligand(molecule, ligand, lie);
    lie->fragment = ligand;
    lie->marques.mk_2 = VRAI;
  }

  /* On renvoie le nombre de ligands trouv */
  return i;
}
