/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-2001, Emmanuel CURIS
   Ce fichier : cration le 6 juin 2001

   Cration d'un modle structural

   Gestion du schma de liaison d'une molcule
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"
#include "interface.h"
#include "tracer.h" /*Pour affichage des liaisons -  dplacer un jour... */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define INTRA_FRAGMENT  0 /* Types de liaisons possibles */
#define INTER_FRAGMENT  1
#define INTER_MOLECULES 2

#define TOLERANCE (1.1) /* De variation autour des sommes des rayons pour dfinir une liaison (1.1 = 10%) */

#define LGMAX_NOM_TYPE_LIAISON 20
#define LGMAX_NOM_ATOME        21 /* Dans le slecteur des atomes */
#define NOMBRE_LIAISONS_VISIBLES (CL_NOM_DER_ATOME - CL_NOM_ATOME_1 + 1)

static const char *txt_types_liaisons[] = {
  "Covalente",
  "Ionique",
  "Hydrog"e_GRAVE"ne",
  "Van der Waals",
  "M"e_AIGU"tallique",
  "R"e_AIGU"pulsive"
};

struct infos_liaisons {
  Atome *atome;
  Liaison **liaisons;
  unsigned int prem_affiche;
  unsigned int num_choisie;
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned int ATM_max_liaisons(int Z);

static Liaison *deja_lies(Liaison *depart, Atome *atm1, Atome *atm2);

static void affiche_liaisons  (struct infos_liaisons *infos);
static void desactive_tout    (void);
static void enregistre_liaison(Liaison *liaison);
static void prepare_arbre_liaisons(void);
static int GER_arbre_liaisons(unsigned int objet, int x, int y);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Cration relle de la liaison : structure contenant les infos, positionnement de cette structure
     Renvoie un pointeur sur la liaison cre - NULL en cas d'erreur

  molecule : si != NULL, on postule que les atomes sont dans cette molcule
  atome_1 : premier atome impliqu dans la liaison
  atome_2 : deuxime atome impliqu dans la liaison
  multiplicite = de la liaison (simple, double, triple,...)
  type = de liaison
*/

Liaison *creer_liaison(Molecule *molecule, Atome *atome_1, Atome *atome_2, int multiplicite, int type)
{
  Liaison *liaison = NULL;
  int sorte = 0; /* De liaison ainsi cre */

  if ( (atome_1 == NULL) || (atome_2 == NULL) )
    return NULL;

  if ( atome_1 == atome_2 ) {
    (void) fprintf( stderr, "creer_liaison : les deux atomes indiqus sont les mmes !\n" );
    return NULL;
  }

  /* Quel type de liaison doit-on crer ? */
  if (molecule == NULL) {
    Molecule *mol_1, *mol_2;

    mol_1 = quelle_molecule(atome_1, NULL);
    mol_2 = quelle_molecule(atome_2, NULL);

    if (mol_1 != mol_2) {
      sorte = INTER_MOLECULES;
    } else
      molecule = mol_1;
  }
  if ( atome_1->fragment != NULL ) {
    if ( atome_1->fragment == atome_2->fragment )
      sorte = INTRA_FRAGMENT;
    else if (molecule != NULL)
      sorte = INTER_FRAGMENT;
  } else {
    sorte = INTER_FRAGMENT; /* Puisqu'il n'y a pas de fragment... */
  }

  /* On vrifie que les atomes indiqus ne sont pas dj lis
     Suppose implicitement que les liaisons sont toujours bien places
     donc qu'un changement d'organisation des fragments est suivi d'une radaptation des liaisons 
  */
  if ( sorte == INTRA_FRAGMENT ) {
    liaison = deja_lies(atome_1->fragment->liaisons, atome_1, atome_2);
  } else if ( sorte == INTER_FRAGMENT ) {
    liaison = deja_lies(molecule->liaisons, atome_1, atome_2);
  } else {
    /* A FAIRE */
  }

  if ( liaison != NULL ) {
    (void) fprintf(stderr, "creer_liaison : les atomes indiqus sont dj lis !\n");
    /* Il faudrait, par exemple, augmenter la multiplicit... */
    return liaison;
  }

  /* On vrifie que l'on n'entre pas en dsaccord avec la valence de l'lment */
  if ( type == LIAISON_COVALENTE ) {
    if ( (atome_1->nbr_coval >= ATM_max_liaisons(atome_1->num_atomique)) ||
	 (atome_2->nbr_coval >= ATM_max_liaisons(atome_2->num_atomique)) ) {
      (void) fprintf(stderr, "creer_liaison : l'un des deux atomes indiqus  dj atteint le maximum de liaisons autorises\n");
      return NULL;
    }
  }

  liaison = (Liaison *) malloc(sizeof(Liaison));
  if (liaison == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/liaisons.c", "creer_liaison", "liaison");
    return NULL;
  }

  /* Initialisation des infos de la liaison */
  liaison->nbr_atomes = 2;
  liaison->atome_1 = atome_1;
  liaison->atome_2 = atome_2;
  liaison->autres_atomes = NULL;

  liaison->multiplicite = multiplicite;
  liaison->energie = 0.0;
  liaison->k_force = 0.0;
  liaison->type = (short int) type;

  liaison->longueur = ATM_distance(atome_1, atome_2);
  if (liaison->longueur > 4.0) {
    (void) fprintf(stderr, "creer_liaison : liaison inhabituellement longue : %f A\n", liaison->longueur);
  }
#ifdef _DEVELOPPE_
  printf("Cration d'une liaison entre %s (%f, %f, %f) et %s (%f, %f, %f) : %p\n",
	 ATM_donnees[atome_1->num_atomique - 1].symbole, atome_1->position.x, atome_1->position.y, atome_1->position.z,
	 ATM_donnees[atome_2->num_atomique - 1].symbole, atome_2->position.x, atome_2->position.y, atome_2->position.z, (void *) liaison);
#endif

  atome_1->nbr_liaisons++;
  atome_2->nbr_liaisons++;
  if (type == LIAISON_COVALENTE) { /* Pour contrle de valence maximale */
    atome_1->nbr_coval += multiplicite;
    atome_2->nbr_coval += multiplicite;
  }

  /* On place la liaison au bon endroit
     Au dbut de la liste simplement chane - reste  trouver la bonne liste... */
  if (sorte == INTER_MOLECULES) { /* Liaison intermolculaire */
    (void) fprintf(stderr, "creer_liaison : liaison intermolculaire - pas encore termin...\n");
  } else {
    if (sorte == INTRA_FRAGMENT) { /* Liaison dans le mme fragment */
#ifdef _DEVELOPPE_
      printf(" ... dans le mme fragment [suivante = %p]\n", (void *) atome_1->fragment->liaisons);
#endif
      liaison->suivante = atome_2->fragment->liaisons;
      atome_2->fragment->liaisons = liaison;
      atome_2->fragment->nbr_liaisons++;
    } else {
      liaison->suivante = molecule->liaisons;
      molecule->liaisons = liaison;
      molecule->nbr_liaisons++;
    }
  }

  return liaison;
}

/* -------- Effacer une liaison
     EN AUCUN CAS, ON N'EFFACE LES ATOMES ASSOCIS
 -------- */

void effacer_liaison(Liaison *liaison, boolean forcer)
{
  Liaison *depart = NULL;
  Molecule *molecule = NULL;
  int sorte;
  
  if (liaison == NULL) return;
  
  if (forcer == FAUX) {
  }

#ifdef _DEVELOPPE_
  printf("          on efface la liaison %p [%s-%s]", (void *) liaison,
	 ATM_donnees[liaison->atome_1->num_atomique - 1].symbole, ATM_donnees[liaison->atome_2->num_atomique - 1].symbole);
#endif

  /* On commence par dcompter la liaison des atomes concerns */
  liaison->atome_1->nbr_liaisons--;
  liaison->atome_2->nbr_liaisons--;
  if (liaison->type == LIAISON_COVALENTE) {
    liaison->atome_1->nbr_coval -= liaison->multiplicite;
    liaison->atome_2->nbr_coval -= liaison->multiplicite;
  }

  /* On extrait de la liste chane la liaison concerne */
  if ( (liaison->atome_1->fragment != NULL) &&
       (liaison->atome_1->fragment == liaison->atome_2->fragment) ) {
#ifdef _DEVELOPPE_
    printf(" - intra-fragment");
#endif
    sorte = INTRA_FRAGMENT;
    depart = liaison->atome_1->fragment->liaisons;
    liaison->atome_1->fragment->nbr_liaisons--; /* Une liaison de moins, quoi qu'il arrive */
  } else {
    Molecule *mol_1, *mol_2;
    
    mol_1 = quelle_molecule(liaison->atome_1, NULL);
    mol_2 = quelle_molecule(liaison->atome_2, NULL);
      
    if ( (mol_1 == NULL) || (mol_2 == NULL) || (mol_1 != mol_2) ) {
#ifdef _DEVELOPPE_
      printf(" - inter-molcules");
#endif
      sorte = INTER_MOLECULES;
      /* A FAIRE */
    } else {
#ifdef _DEVELOPPE_
      printf(" - intra-molcule");
#endif
      sorte = INTER_FRAGMENT;
      depart = mol_1->liaisons;
      molecule = mol_1;
      molecule->nbr_liaisons--; /* Une liaison de moins, quoi qu'il arrive */
    }
  }

  if (depart == liaison) { /* On supprime la premire */
#ifdef _DEVELOPPE_
    printf(" - premire de la liste\n");
#endif
    if (sorte == INTER_FRAGMENT) {
      molecule->liaisons = liaison->suivante;
    } else if (sorte == INTRA_FRAGMENT) {
      liaison->atome_1->fragment->liaisons = liaison->suivante;
    } else {
      /* A FAIRE */
    }
  } else {
    Liaison *avant = depart;
#ifdef _DEVELOPPE_
    printf("\n");
#endif
    while (avant->suivante != liaison)
      avant = avant->suivante;
    avant->suivante = liaison->suivante;
  }

  /* On libre la mmoire */
  free(liaison);
}

/* -------- Vrification : les atomes indiqus ne sont pas dj lis --------
      Renvoie la liaison entre les deux atomes, NULL si elle n'existe pas
 */

static Liaison *deja_lies(Liaison *depart, Atome *atm1, Atome *atm2)
{
  if ( (depart == NULL) || (atm1 == NULL) || (atm2 == NULL) )
    return NULL;

  for ( ; depart != NULL ; depart = depart->suivante) {
    if ( ( (depart->atome_1 == atm1) && (depart->atome_2 == atm2) ) ||
	 ( (depart->atome_1 == atm2) && (depart->atome_2 == atm1) ) )
      return depart;

    /* Il faudrait vrifier dans les atomes supplmentaires si liaisons multi-centres */
  }

  return NULL; /* Si on est l : aucune liaison n'existe entre ces atomes,
		  du moins dans la liste de liaisons indique */
}

/* -------- L'atome appartient-il  la liaison indique ? -------- */

boolean ATM_EstDansLiaison(Atome *atome, Liaison *liaison)
{
  if ( (NULL == liaison) || (NULL == atome) ) return FAUX;

  if (liaison->atome_1 == atome) return VRAI;
  if (liaison->atome_2 == atome) return VRAI;

  if (liaison->nbr_atomes > 2) { /* Liaison  + de 2 centres */
    unsigned int i;

    for (i = 0 ; i < liaison->nbr_atomes - 2U ; i++) {
      if ( liaison->autres_atomes[i] == atome ) return VRAI;
    }
  }

  return FAUX; /* L'atome n'est pas impliqu dans la liaison */
}

/* -------- Nombre maximal de liaisons de covalence pour un atome -------- */

static unsigned int ATM_max_liaisons(int Z)
{
  if ( (Z < HYDROGENE) || (Z > PLUTONIUM) )
    return 0; /* Pas de chimie connue pour ces lments... */

  switch (Z)
  {
  case HELIUM: /* Gaz rares du dbut : aucune liaison */
  case NEON:
  case ARGON:
    return 0;

  case HYDROGENE: /* Alcalins : une liaison covalente */
  case SODIUM:
  case POTASSIUM:
  case FLUOR: /* Fluor : pareil, car trop oxydant */
    return 1;

  case OXYGENE:
    return 3; /* Dans H3O+ par exemple */

  case CARBONE:
  case AZOTE: /* Dans les ions ammonium */
    return 4;

  case CHLORE: /* Halognes */
  case BROME:
  case IODE:
    return 7; /* Dans XO4- par exemple */

  case TERBIUM:
  case GADOLINIUM:
  case SAMARIUM:
  case EUROPIUM:
  case LANTHANE:
  case LUTETIUM:
    return 9;

  default: /* Semble le plus frquent : connu pour presque tout... */
    return 6;
  }

  return 0;
}

/* -------- Recherche automatique des liaisons dans une molcule
     Renvoie le nombre de liaisons trouves dans la molcule

     Ne cherche que les liaisons covalentes, pour le moment
-------- */

unsigned long MOL_chercher_liaisons(Molecule *molecule)
{
  unsigned long nbr_liaisons = 0;
  Atome *atome_1;

  if (molecule == NULL) return 0;

  /* Boucle sur l'atome 1 */
  for ( atome_1 = molecule->premier ; atome_1 != NULL ; atome_1 = atome_1->suivant ) {
    Atome *atome_2;
    double r_covalent;
    unsigned int max_liaisons;

    /* Contrle du rayon */
    r_covalent = (double) ATM_Z_vers_rayon( atome_1->num_atomique, ATM_RAYON_COVALENT );
    if (r_covalent <= 10.0) {/* Moins que 10 pm : absurde */
      (void) fprintf(stderr, "MOL_chercher_liaisons : atome %s, rayon covalent inconnu\n", ATM_Z_vers_symbole( atome_1->num_atomique ) );
      continue; /* Donnes inconnues pour cet atome */
    }

    /* Contrle du nombre de liaisons */
    max_liaisons = ATM_max_liaisons(atome_1->num_atomique);
    if (atome_1->nbr_coval >= max_liaisons) /* Toutes les liaisons autorises ont t trouves ! */
      continue;

    /* Boucle sur l'atome 2 */
    for (atome_2 = atome_1->suivant ; atome_2 != NULL ; atome_2 = atome_2->suivant ) {
      double r_cov2, dist;

      /* Contrle du rayon */
      r_cov2 = (double) ATM_Z_vers_rayon( atome_2->num_atomique, ATM_RAYON_COVALENT );
      if (r_cov2 <= 10.0) {/* Moins que 10 pm : absurde */
	/* Pas de message car sera affich dans la boucle principale aprs... */
	continue; /* Donnes inconnues pour cet atome */
      }

      /* Distance entre les atomes */
      dist = ATM_distance(atome_1, atome_2);

      /* La liaison est-elle chimiquement plausible ? */
      if (dist < TOLERANCE * (r_covalent + r_cov2) / 100.0) { /* Tolrance de 20% ; attention, rayons en pm et distance en angstrm */
	if (creer_liaison(molecule, atome_1, atome_2, 1, LIAISON_COVALENTE) != NULL)
	  nbr_liaisons++;
      }

      if (atome_1->nbr_coval >= max_liaisons) /* Toutes les liaisons autorises pour l'atome 1ont t trouves ! */
	break;
    }
  }

  /* Redessin ventuel du rsultat */
#ifdef AVEC_OPEN_GL_
  if ( (nbr_liaisons > 0) && (molecule == mol_courante) ) {
    if (liste_liaisons)
      {
	glNewList(liste_liaisons, GL_COMPILE);
	dessine_liaisons(mol_courante, mode_vue3d);
	glEndList(); /* On a fini le dessin */
      }
    FEN_redessin(fen_vue_3d, BI_RDS_GL); /* N'a pas d'intrt sinon ! */
  }
#endif

  /* Affichage du rsultat */
  HIST_Titre(1, "Recherche de liaisons", "LIE");
  HIST_Texte(2, 0, "Mol"e_AIGU"cule : %s\n", molecule->nom);
  HIST_Texte(3, 0, "Liaisons trouv"e_AIGU"es : %lu\n", nbr_liaisons);

  return nbr_liaisons;
}

/* ~~~~~~~~ Interface graphique : paramtres sur les liaisons ~~~~~~~~ */

void modifier_liaisons(Atome *atome, Molecule *molecule)
{
  struct infos_liaisons infos;

  if ( (atome == NULL) || (molecule == NULL) ) return;

  infos.atome = atome;
  if (atome->nbr_liaisons > 0) {
    Fragment *fragment;
    Liaison *liaison;
    unsigned int i = 0;

    infos.liaisons = (Liaison **) calloc(atome->nbr_liaisons, sizeof(Liaison *));
    if (infos.liaisons == NULL) {
      SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/liaisons.c", "modifier_liaisons", "infos.liaisons");
    }

    /* On construit la table des liaisons */
    /*   1) liaisons dans les fragments */
    for (fragment = molecule->fragments ; fragment != NULL ; fragment = fragment->suivant) {
      for (liaison = fragment->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
	if ( (liaison->atome_1 == atome) || (liaison->atome_2 == atome) ) {
	  if (i > atome->nbr_liaisons) {
	    fprintf(stderr, " ***** ERREUR GRAVE dans modifier_liaisons *****\n");
#ifdef _DEVELOPPE_
	    fprintf(stderr, "modifier_liaisons : plus de liaisons pour l'atome que n'indique atome->nbr_liaisons !\n");
	    fprintf(stderr, "     atome = %d en (%f, %f, %f)\n", atome->num_atomique, atome->position.x, atome->position.y, atome->position.z);
#endif
	    free(infos.liaisons);
	    return;
	  }
	  infos.liaisons[i] = liaison;
	  i++;
	}
      }
    }
    /*   2) liaisons dans la molcule */
    for (liaison = molecule->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
      if ( (liaison->atome_1 == atome) || (liaison->atome_2 == atome) ) {
	if (i > atome->nbr_liaisons) {
	  fprintf(stderr, " ***** ERREUR GRAVE dans modifier_liaisons *****\n");
#ifdef _DEVELOPPE_
	  fprintf(stderr, "modifier_liaisons : plus de liaisons pour l'atome que n'indique atome->nbr_liaisons !\n");
	  fprintf(stderr, "     atome = %d en (%f, %f, %f)\n", atome->num_atomique, atome->position.x, atome->position.y, atome->position.z);
#endif
	  free(infos.liaisons);
	  return;
	}
	infos.liaisons[i] = liaison;
	i++;
      }
    }

    OBJ_active(CREER_LIAISON, CL_TT_EFFACER, FORCE_OUI);
    OBJ_active(CREER_LIAISON, CL_OTER_LIAISON, FORCE_OUI);
    OBJ_active(CREER_LIAISON, CL_TRIER_LIAISON, atome->nbr_liaisons > 1);
    
    OBJ_active(CREER_LIAISON, CL_TYPE_LIAISON, FORCE_OUI);
  } else {
    infos.liaisons = NULL; /* Aucune liaison pour le moment, pour cet atome */
    desactive_tout();
  }
  infos.prem_affiche = 0;
  infos.num_choisie = 0;

  affiche_liaisons(&infos);

  if ( arbres[CREER_LIAISON].drapeaux.prepare == FAUX )
    prepare_arbre_liaisons();

  met_chaine( CREER_LIAISON, CL_TYPE_ATOME, 0,ATM_Z_vers_symbole( atome->num_atomique ) , TXT_COPIER);
  OBJ_MettreCoordonnees( CREER_LIAISON, CL_X_ATOME, CL_Y_ATOME, CL_Z_ATOME, &(atome->position), atome->err_position );

  ARB_MettreComplement( CREER_LIAISON, &infos );
  ARB_Bloquant( CREER_LIAISON, "Modifier les liaisons", fen_vue_3d, 0 );
  ARB_MettreComplement( CREER_LIAISON, NULL );

  if (infos.liaisons != NULL) /* Il reste au moins une liaison pour cet atome */
    enregistre_liaison(infos.liaisons[infos.num_choisie]);
  if (molecule == mol_courante) { /* Redessin du schma de liaison, car modifs potentielles */
#ifdef AVEC_OPEN_GL_
    if (liste_liaisons) {
      glNewList(liste_liaisons, GL_COMPILE);
      dessine_liaisons(mol_courante, mode_vue3d);
      glEndList(); /* On a fini le dessin */
    }
    FEN_redessin(fen_vue_3d, BI_RDS_GL);
#endif
  }
}

static void affiche_liaisons(struct infos_liaisons *infos)
{
  unsigned int i;

  ARB_ascenseur(CREER_LIAISON, CL_AT_AVANT, CL_FND_ASC_ATMS, CL_BOUTON_ASC, CL_ATM_APRES, ASC_VERTICAL,
		infos->atome->nbr_liaisons, NOMBRE_LIAISONS_VISIBLES, infos->prem_affiche, 0, -1, -1, METTRE);

  for (i = 0 ; i < NOMBRE_LIAISONS_VISIBLES ; i++) {
    if (i + infos->prem_affiche < infos->atome->nbr_liaisons) {
      Liaison *liaison = infos->liaisons[i + infos->prem_affiche];

      OBJ_active(CREER_LIAISON, CL_NOM_ATOME_1 + i, FORCE_OUI);
      if ( i + infos->prem_affiche == infos->num_choisie ) {
	met_chaine(CREER_LIAISON, CL_TYPE_LIAISON, LGMAX_NOM_TYPE_LIAISON, txt_types_liaisons[liaison->type], TXT_COPIER);
	OBJ_active(CREER_LIAISON, CL_MUTIPLICITE, liaison->type == LIAISON_COVALENTE); /* A de sens uniquement pour une liaison covalente */
	OBJ_active(CREER_LIAISON, CL_LG_EQUILIBRE, liaison->type != LIAISON_V_D_WAALS); /* Pas de sens avec le modle L-J pour VdW */
	OBJ_active(CREER_LIAISON, CL_VAL_CTE_FORCE, liaison->type != LIAISON_V_D_WAALS);
	OBJ_MettreNombre( CREER_LIAISON, CL_MUTIPLICITE, liaison->multiplicite, 0.0 );
	OBJ_MettreNombre( CREER_LIAISON, CL_LG_LIAISON, ATM_distance(liaison->atome_2, liaison->atome_1), 0.0 );
	OBJ_MettreNombre( CREER_LIAISON, CL_LG_EQUILIBRE, liaison->longueur, 0.0 );
	OBJ_MettreNombre( CREER_LIAISON, CL_VAL_CTE_FORCE, liaison->k_force, 0.0 );
      }
      if ( liaison->atome_1 == infos->atome) {
	met_chaine( CREER_LIAISON, CL_NOM_ATOME_1 + i, LGMAX_NOM_ATOME, 
		    ATM_Z_vers_nom( liaison->atome_2->num_atomique - 1 ), TXT_COPIER);
      } else {
	met_chaine( CREER_LIAISON, CL_NOM_ATOME_1 + i, LGMAX_NOM_ATOME,
		    ATM_Z_vers_nom( liaison->atome_1->num_atomique - 1 ), TXT_COPIER);
      }
      OBJ_choisit(CREER_LIAISON, CL_NOM_ATOME_1 + i, i + infos->prem_affiche == infos->num_choisie);
    } else {
      OBJ_choisit(CREER_LIAISON, CL_NOM_ATOME_1 + i, FORCE_NON);
      OBJ_active(CREER_LIAISON, CL_NOM_ATOME_1 + i, FORCE_NON);
      met_chaine(CREER_LIAISON, CL_NOM_ATOME_1 + i, LGMAX_NOM_ATOME, "", TXT_COPIER);
    }
  }
}

static void desactive_tout(void)
{
  (void) ARB_ascenseur(CREER_LIAISON, CL_AT_AVANT, CL_FND_ASC_ATMS, CL_BOUTON_ASC, CL_ATM_APRES, ASC_VERTICAL,
		       0, NOMBRE_LIAISONS_VISIBLES, 0, 0, -1, -1, METTRE);

  OBJ_active(CREER_LIAISON, CL_TT_EFFACER, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_OTER_LIAISON, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_TRIER_LIAISON, FORCE_NON);

  OBJ_active(CREER_LIAISON, CL_TYPE_LIAISON, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_MUTIPLICITE, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_LG_EQUILIBRE, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_VAL_CTE_FORCE, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_LJ_CTE_C12, FORCE_NON);
  OBJ_active(CREER_LIAISON, CL_LJ_VAL_C6, FORCE_NON);
}

static void enregistre_liaison(Liaison *liaison)
{
  if (liaison == NULL) return;

  liaison->multiplicite = OBJ_LireNombre( CREER_LIAISON, CL_MUTIPLICITE, NULL );
  /* Ces deux infos n'ont de sens que pour une liaison covalente, a priori */
  liaison->k_force  = OBJ_LireNombre( CREER_LIAISON, CL_VAL_CTE_FORCE, NULL );
  liaison->longueur = OBJ_LireNombre( CREER_LIAISON, CL_LG_EQUILIBRE, NULL ); /* Longueur  l'quilibre */
}

static int GER_arbre_liaisons(unsigned int objet, int x, int y)
{
  unsigned int i, nbr_liaisons;
  int nv_type;
  struct infos_liaisons *infos;
  Liaison *liaison = NULL;

  infos = (struct infos_liaisons *) ARB_LireComplement( CREER_LIAISON );
  if (infos->liaisons != NULL ) /* NULL si aucune liaison pour l'atome trait */
    liaison = infos->liaisons[infos->num_choisie];

  if ( (objet >= CL_NOM_ATOME_1) && (objet <= CL_NOM_DER_ATOME) ) {
    enregistre_liaison(liaison);
    infos->num_choisie = infos->prem_affiche + (objet - CL_NOM_ATOME_1);
    affiche_liaisons(infos);
  }

  OBJ_choisit(CREER_LIAISON, objet, FORCE_OUI);
  switch (objet)
  {
  case CL_AT_AVANT: /* Liste des atomes impliqus dans la liaison */
  case CL_FND_ASC_ATMS:
  case CL_BOUTON_ASC:
  case CL_ATM_APRES:
    infos->prem_affiche = ARB_ascenseur(CREER_LIAISON, CL_AT_AVANT, CL_FND_ASC_ATMS, CL_BOUTON_ASC, CL_ATM_APRES, ASC_VERTICAL,
					infos->atome->nbr_liaisons, NOMBRE_LIAISONS_VISIBLES, infos->prem_affiche,
					objet, x, y, ECHANGE);
    affiche_liaisons( infos );
    return VRAI;

  case CL_NV_LIAISON:
    {
      unsigned int *lst_atomes;
      int num_atome;

      lst_atomes = ATM_choisir(NULL, 0, FAUX, &num_atome);
      free(lst_atomes);
      if (num_atome >= 0) { /* On a bien choisi un atome */
	Atome *atm_2 = NULL;
	/* IL FAUDRAIT ADAPTER EN CONNAISSANT LA MOLCULE */
/*	for (atm_2 = ATM_premier, i = 0 ; i < num_atome ; atm_2 = atm_2->suivant, i++);*/

	liaison = creer_liaison(NULL, infos->atome, atm_2, 1, LIAISON_COVALENTE);
      }
    }
    return VRAI;

  case CL_OTER_LIAISON:
    liaison = infos->liaisons[infos->num_choisie]; /* La liaison qu'il faudra effacer */
    if (infos->atome->nbr_liaisons == 1) {/* On efface la dernire liaison */
      free(infos->liaisons);
      infos->liaisons = NULL;
      infos->num_choisie = 0;
      infos->prem_affiche = 0;
      desactive_tout();
    } else {
      for (i = infos->num_choisie ; i < infos->atome->nbr_liaisons ; i++) /* On remet  jour la table des liaisons */
	infos->liaisons[i] = infos->liaisons[i + 1];
      infos->liaisons[infos->atome->nbr_liaisons - 1] = NULL;
      if (infos->num_choisie + 1 == infos->atome->nbr_liaisons) /* On est sur la dernire */
	infos->num_choisie--;
    }
    effacer_liaison(liaison, FAUX); /* On efface _aprs_ car modifie atome->nbr_liaisons... */
    affiche_liaisons(infos);
    return VRAI;

  case CL_TT_EFFACER:
    nbr_liaisons = infos->atome->nbr_liaisons; /* Dans atome, sera modifi aprs, alors... */
    for (i = 0 ; i < nbr_liaisons ; i++)
      effacer_liaison(infos->liaisons[i], VRAI);
    free(infos->liaisons);
    infos->liaisons = NULL;
    infos->prem_affiche = 0;
    infos->num_choisie = 0;
    desactive_tout();
    affiche_liaisons(infos);
    return VRAI;

  case CL_TYPE_LIAISON:
    nv_type = SS_menu_texte(txt_types_liaisons, 6, "Type de liaison", CREER_LIAISON, CL_TYPE_LIAISON);
    if ( (nv_type >= 0) && (nv_type != liaison->type) ) {
      liaison->type = nv_type;
      met_chaine(CREER_LIAISON, CL_TYPE_LIAISON, LGMAX_NOM_TYPE_LIAISON, txt_types_liaisons[liaison->type], TXT_COPIER);
    }
    return VRAI;
  }
  OBJ_choisit(CREER_LIAISON, objet, FORCE_NON);

  return FAUX;
}

static void prepare_arbre_liaisons(void)
{
  unsigned int i;

  if ( FAUX == arbres[ CREER_LIAISON ].drapeaux.init ) {
    int retour;

    OPTION_A_FAIRE( "Documentation, gestion des liaisons" );
    retour = ARB_Initialiser( CREER_LIAISON, NULL, GER_arbre_liaisons, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  OBJ_agrandit(CREER_LIAISON, CL_FND_LST_ATMS, 0, NOMBRE_LIAISONS_VISIBLES - 1);
  for (i = 0 ; i < NOMBRE_LIAISONS_VISIBLES ; i++)
    OBJ_deplace(CREER_LIAISON, CL_NOM_ATOME_1 + i, 0, i, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_AT_AVANT, +1, 0, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_FND_ASC_ATMS, +1, +1, DEPLACE_PIXELS);
  OBJ_agrandit(CREER_LIAISON, CL_FND_ASC_ATMS, 0, NOMBRE_LIAISONS_VISIBLES - 2);
  OBJ_deplace(CREER_LIAISON, CL_ATM_APRES, 0, NOMBRE_LIAISONS_VISIBLES - 1, DEPLACE_PIXELS);

  OBJ_deplace(CREER_LIAISON, CL_TT_EFFACER, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace(CREER_LIAISON, CL_TXT_LG, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_LG_LIAISON, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_TXT_LG_EQ, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_LG_EQUILIBRE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_TXT_K, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_VAL_CTE_FORCE, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_TXT_LENNARD, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_TXT_LJ_C1, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_TXT_LJ_CTE2, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_LJ_VAL_C6, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_LIAISON, CL_LJ_CTE_C12, 0, +3, DEPLACE_PIXELS);

  arbres[CREER_LIAISON].drapeaux.prepare = VRAI;
}
