/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS

   Gestion d'une structure molculaire comme modle d'exafs
   Import des coordonnes cristallographiques rcupres de la Cambridge Cristallographic Databank C

  7, 12 janvier ; 5, 11 fvrier 1998
    * Modifi les noms des procdures de chargement en JNL_... (charge les fichiers *.JNL)
    * Ajout les fonctions de chargement des fichiers *.COR : COR_...
    * Chargement des fichiers *.COR fonctionne :-)
    * Travail sur le chargement : install le dbogage -> 2 des 3 fichiers passent
    * Rcupr le fichier sur la station - travail sur Atari maintenant
    * Chang les noms des fonctions : suffixe CCDB_
    * Ajout inclusion des fonctions GEM
    * Ajout procdure de choix du fichier : CCDB_charge()
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "systeme.h"

#include "atomes.h"
#include "cristallo.h"
#include "fichiers.h"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int CCDC_LitStructure(FILE *fichier, const char *nom_fichier, struct infos_lecture *infos);

static unsigned long CCDC_CompterStructures(FILE *fichier);

/* ~~~~~~~~~~~~~~~ Lecture des fichiers *.COR ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* -------------- Lecture proprement dite du fichier ---------------------------\
 |   Renvoie un code en cas d'erreur                                            |
 |                                                                              |
 | nom_fichier : le nom du fichier qu'il faut charger                           |
 \----------------------------------------------------------------------------- */

int MOL_Import_CCDC(const char *nom_fichier, struct infos_lecture *infos)
{
  FILE *fichier; /* le fichier que l'on va lire */
  int retour = -1;
  unsigned long nbr_structures;

  if (nom_fichier == NULL)
    return -1;

  fichier = fopen(nom_fichier, "rb");
  if (fichier == NULL) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    SYS_ERREUR(ERREUR_LECTURE, nom_fichier, NULL, NULL);
    return -1;
  }

  /* 1. On compte combien de structures sont prsentes dans le fichier */
  nbr_structures = CCDC_CompterStructures( fichier );
  if ( 0 == nbr_structures ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le dcompte des structures\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return -2;
  }

  if ( 1 == nbr_structures ) {
    retour = CCDC_LitStructure( fichier, nom_fichier, infos );
  } else {
    unsigned int i;

    OPTION_A_FAIRE( "Faire la liste des coordonnes prsente, proposer le choix" );
    for ( i = 0 ; i < nbr_structures ; i++ ) {
      retour |= CCDC_LitStructure( fichier, nom_fichier, infos );
    }
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le chargement de la ou des structures\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return -20;
  }

  /* Le chargement s'est fini sans encombre */
  (void) fclose( fichier );
  return 0;
}

/* -------- Compter le nombre de structures dans un fichier CCDC -------- */

static unsigned long CCDC_CompterStructures(FILE *fichier)
{
  unsigned long nombre = 0, num_ligne = 0;
  unsigned int nbr_colonnes;
  int c;

  do {
    num_ligne++; /* On passe  la ligne suivante */

    nbr_colonnes = compte_colonnes( fichier, 0 );
    if ( nbr_colonnes < 4 ) { /* Il faut au moins 4 colonnes pour une ligne de coordonnes : 3 (xyz) + 1 (Z) */
      if ( 0 == nbr_colonnes ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le dcompte des colonnes [ligne %lu]\n",
		       __FILE__, __FUNCTION__, __LINE__, num_ligne);
	break;
      }
      nombre++; /* Une structure de plus */
    }

    /* On ignore la ligne */
    c = ignore_ligne( fichier );
  } while ( c != EOF );

  /* On se replace au dbut du fichier */
  rewind( fichier );

  /* On a correctement dcompt */
  return nombre;
}

/* -------- Lire une (et une seule) structure dans le fichier -------- */

static int CCDC_LitStructure(FILE *fichier, const char *nom_fichier, struct infos_lecture *infos)
{
  char *nom, *reference;
  int c, retour;
  unsigned int nbr_colonnes;

  /* Rfrence lue */
  reference = lit_commande( fichier );
  if ( NULL == reference ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de lire la rfrence\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Nom de la molcule  crer */
  nom = TXT_genere_chaine( "%s (%F, %s)", infos->nom, nom_fichier, reference );

  /* On cre la molcule qui contiendra le rsultat */
  infos->molecule = MOL_creer( nom );
  free( nom ); nom = NULL;
  if ( NULL == infos->molecule ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la molcule [fichier %s, ref. %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier, reference);
    free( reference );
    return -5;
  }
  free( reference ); reference = NULL;

  /* On ignore la fin de la ligne */
  c = ignore_ligne( fichier );
  if ( EOF == c ) {
    (void) fprintf(stderr, "%s,%s l.%u : Rference sans coordonnes, en fin de fichier [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    return -10;
  }

  /* Nombre de colonnes de la 1re ligne */
  nbr_colonnes = compte_colonnes( fichier, 0 );
  while ( nbr_colonnes >= 4 ) {
    /* Ligne de coordonnes : on la lit */
    retour = CRIS_LireLigneAscii( fichier, infos );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la lecture de la ligne de coordonnes\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -20;
    }

    /* Nombre de colonnes de la ligne suivante */
    nbr_colonnes = compte_colonnes( fichier, 0 );
  }

  /* Si on est l : fin de fichier ou dbut de la rfrence suivante */

  /* On adapte donc la molcule */
  MOL_cherche_barycentre( infos->molecule, FAUX);
  MOL_initialise_vue( infos->molecule );

  return 0;
}
