/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2003, Emmanuel Curis

   Interface graphique : routines générales indépendantes du système

   Chargement d'un fichier RSC (de ses arbres, en fait)
*/

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef _UNIX_
# define _LIBRAIRIE_X_ /* Pour avoir accès aux fonctions internes */
#elif defined(_WINDOWS_)
# define _LIBRAIRIE_W_
#endif

#include "bib_interface.h"
#include "bin_utils.h"
#include "txt_utils.h"

/* ~~~~~ Variables locales ~~~~~ */

#define RSC_TAILLE_OBJET ( 7 * sizeof( int16_t ) + 3 * sizeof( uint16_t ) + sizeof( uint32_t ) )

/* ~~~~~ Fonctions locales ~~~~~ */

static int BI_RSC_LireTedinfo(struct rsc_infos *infos, unsigned int idx_tedinfo, TEDINFO *tedinfo);
static int BI_RSC_LireUnArbre(struct rsc_infos *infos, unsigned int num_arbre, unsigned int nbr_objets, long position);

/* ~~~~~ Réalisation pratique ~~ */

int BI_ChargerRSC(const char *nom_fichier, unsigned int idx_premier, unsigned int *nbr_arbres)
{
  int retour;
  struct rsc_infos infos;

  /* On ouvre le fichier, on lit ce qu'il doit contenir */
  retour = BI_RSC_LireEnTeteFichier( nom_fichier, &infos );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s, %s l. %u: ERREUR dans BI_RSC_LireFichier() - code %d\n",
		    __FILE__, __FUNCTION__, __LINE__, retour);
    return -1;
  }

  /* On lit les arbres, aux positions indiquées */
  retour = BI_RSC_LireArbres( &infos, idx_premier );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s, %s l. %u: ERREUR dans BI_RSC_LireFichier() - code %d\n",
		    __FILE__, __FUNCTION__, __LINE__, retour);
    (void) fclose( infos.fichier );
    return -1;
  }

  (void) fclose( infos.fichier );

  /* Tout s'est bien passe */
  if ( nbr_arbres != NULL ) *nbr_arbres = infos.nbr_arbres;
  return 0;
}

/* -------- Ouvre le fichier RSC, contrôle le format, prépare sa description -------- */

int BI_RSC_LireEnTeteFichier(const char *nom_fichier, struct rsc_infos *infos)
{
  FILE *fichier;
  unsigned int version;

#ifdef _DEVELOPPE_
  if ( NULL == infos ) {
    (void) fprintf( stderr, "%s, %s l. %u: Pas de structure donnée pour les infos\n",
		    __FILE__, __FUNCTION__, __LINE__);
    abort();
  }
#endif
  if ( NULL == nom_fichier ) {
    (void) fprintf( stderr, "%s, %s l. %u: Pas de fichier indiqué\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On essaye d'ouvrir le fichier */
  fichier = fopen( nom_fichier, "rb" );
  if ( NULL == fichier ) {
    (void) fprintf( stderr, "%s, %s l. %u: Impossible d'ouvrir le fichier %s\n",
		    __FILE__, __FUNCTION__, __LINE__, nom_fichier );
    return -2;
  }

  /* Numéro de version RSC */
  version = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB ); /* RSC codé au format motorola */
  if ( version > 3 ) {
    (void) fprintf( stderr, "%s, %s l. %u: Version RSC inconnue [%u]\n",
		    __FILE__, __FUNCTION__, __LINE__, version);
    (void) fclose( fichier );

    return -5;
  }

  if ( 3 == version ) { /* En-tête : avec codage sur 4 octets des valeurs */
    unsigned int ext_version __attribute__((__unused__));

    ext_version = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );

    infos->deb_objets       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_tedinfo      = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_iconblk      = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_bitblk       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_ch_libres    = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_chaines      = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_images       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_img_libres   = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->table_arbres     = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_objets       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_arbres       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_tedinfos     = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_iconblk      = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_bitblk       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_chaines      = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_images       = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
    infos->taille_ressource = BIN_Lire_32U( fichier, BIN_MODE_SOURCE_MSB );
  } else { /* En-tête : avec codage sur 2 octets des valeurs */
    infos->deb_objets       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_tedinfo      = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_iconblk      = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_bitblk       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_ch_libres    = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_chaines      = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_images       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->deb_img_libres   = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->table_arbres     = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_objets       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_arbres       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_tedinfos     = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_iconblk      = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_bitblk       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_chaines      = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->nbr_images       = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
    infos->taille_ressource = BIN_Lire_16U( fichier, BIN_MODE_SOURCE_MSB );
  }

#ifdef _DEVELOPPE_
  printf( "Début des objets  : %ld\n", infos->deb_objets   );
  printf( "Début des tedinfo : %ld\n", infos->deb_tedinfo  );
  printf( "Début des icônes  : %ld\n", infos->deb_iconblk  );
  printf( "Début des bitblk  : %ld\n", infos->deb_bitblk   );
  printf( "Début des chaînes libres : %ld\n", infos->deb_ch_libres );
  printf( "Début des chaînes : %ld\n", infos->deb_chaines  );
  printf( "Début des images  : %ld\n", infos->deb_images   );
  printf( "Nombre d'objets   : %ld\n", infos->nbr_objets   );
  printf( "Nombre d'arbres   : %ld\n", infos->nbr_arbres   );
  printf( "Nombre de TEDINFO : %ld\n", infos->nbr_tedinfos );
  printf( "Nombre d'ICONBLK  : %ld\n", infos->nbr_iconblk  );
  printf( "Nombre d'BITBLK   : %ld\n", infos->nbr_bitblk   );
  printf( "Nombre de chaînes : %ld\n", infos->nbr_chaines  );
  printf( "Nombre d'images   : %ld\n", infos->nbr_images   );
#endif

  /* La lecture des infos s'est bien déroulée */
  infos->fichier = fichier;
  return 0;
}

/* -------- Lit les arbres dans le fichier, mais AUCUN POST-TRAITEMENT -------- */

int BI_RSC_LireArbres(struct rsc_infos *infos, unsigned int idx_premier)
{
  int retour;
  unsigned long idx_arbre;
  long pos_debut, pos_suivant;

  /* On contrôle que les arbres demandés pourront être créés ! */
  if ( NULL == arbres ) {
    unsigned int nbr_arbres;

    nbr_arbres  = infos->nbr_arbres; /* Nombre d'arbres dans le RSC */
    nbr_arbres += BI_NombreArbresInternes; /* Arbres utilisés en interne par la bibliothèque d'interface */

    arbres = BI_ARB_CreerArbres( nbr_arbres );
    if ( NULL == arbres ) {
      (void) fprintf(stderr, "%s, %s l. %u: La table des arbres n'existe pas, et impossible de la créer !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
  }

  /* On se place au début de la table des arbres dans le fichier */
  if ( -1 == fseek( infos->fichier, infos->table_arbres, SEEK_SET ) ) {
    (void) fprintf(stderr, "%s, %s l. %u: ERREUR dans fseek, %s ! Impossible de trouver la table des arbres\n",
		   __FILE__, __FUNCTION__, __LINE__, strerror( errno ) );
    return -5;
  }

  /* Position du tout premier objet */
  pos_suivant = BIN_Lire_32( infos->fichier, BIN_MODE_SOURCE_MSB );
#ifdef _DEVELOPPE_
  if ( pos_suivant != infos->deb_objets ) {
    (void) fprintf(stderr, "%s, %s l. %u: Incohérence dans le fichier RSC. Selon la table, 1er objet en %lu ; selon l'en-tête, en %lu\n",
		   __FILE__, __FUNCTION__, __LINE__, pos_suivant, infos->deb_objets );
    return -666;
  }
#endif
  for ( idx_arbre = 0 ; idx_arbre < infos->nbr_arbres ; idx_arbre++ ) {
    unsigned long nbr_objets;

    pos_debut = pos_suivant;
    pos_suivant = BIN_Lire_32( infos->fichier, BIN_MODE_SOURCE_MSB );

    if ( idx_arbre != infos->nbr_arbres - 1 ) {
      nbr_objets = ( pos_suivant - pos_debut ) / RSC_TAILLE_OBJET;
      infos->nbr_objets -= nbr_objets;
    } else {
      nbr_objets = infos->nbr_objets;
    }
#ifdef _DEVELOPPE_
    printf("%s : arbre %lu / %lu commence en %ld, %lu objets\n", __FUNCTION__, idx_arbre + 1, infos->nbr_arbres, pos_debut, nbr_objets);
#endif

    /* On lit l'arbre */
    retour = BI_RSC_LireUnArbre( infos, idx_arbre + idx_premier, nbr_objets, pos_debut );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l. %u: ERREUR en voulant créer l'arbre %lu [%lu sur %lu] [code %d] !\n",
		     __FILE__, __FUNCTION__, __LINE__, idx_arbre, idx_arbre + 1, infos->nbr_arbres, retour);
      return -1;
    }
  }

  /* Ici, on a fini de lire les arbres, sans problème particulier */
  return 0;
}

/* -------- Lit un arbre en entier, le crée et lit les objets correspondants ! -------- */

static int BI_RSC_LireUnArbre(struct rsc_infos *infos, unsigned int num_arbre, unsigned int nbr_objets, long position)
{
  int retour;
  unsigned int idx_objet;
  long ex_position;
  OBJECT *objets;

  /* On prépare l'arbre, connaissant le nombre d'objets qu'il contient ; taille arbitraire, adaptée après */
  retour = BI_CreerArbre( num_arbre, nbr_objets, 10, 10 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l. %u: ERREUR en voulant créer l'arbre %u [code %d] !\n",
		   __FILE__, __FUNCTION__, __LINE__, num_arbre, retour);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -1;
  }

  /* On mémorise l'ancienne position dans le fichier */
  ex_position = ftell( infos->fichier );

  /* On se place au début de l'arbre dans le fichier */
  if ( -1 == fseek( infos->fichier, position, SEEK_SET ) ) {
    (void) fprintf(stderr, "%s, %s l. %u: Erreur de lecture [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, strerror( errno ) );
    return -42;
  }

  /* On lit tous les objets de l'arbre, un par un */
  for ( idx_objet = 0, objets = arbres[ num_arbre ].adresse ; idx_objet < nbr_objets ; idx_objet++, objets++ ) {
    unsigned long ob_spec;

    /* La hiérarchie */
    objets->ob_next = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
    objets->ob_head = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
    objets->ob_tail = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );

    /* Le type */
    objets->ob_type = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );

    /* Les options */
    objets->ob_flags = BIN_Lire_16U( infos->fichier, BIN_MODE_SOURCE_MSB );
    objets->ob_state = BIN_Lire_16U( infos->fichier, BIN_MODE_SOURCE_MSB );

    /* Le complément */
    objets->ob_spec.autre = BIN_Lire_32( infos->fichier, BIN_MODE_SOURCE_MSB );
    ob_spec = objets->ob_spec.autre;

    /* Position & dimensions */
    objets->ob_x = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
    objets->ob_y = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
    objets->ob_width  = BIN_Lire_16U( infos->fichier, BIN_MODE_SOURCE_MSB );
    objets->ob_height = BIN_Lire_16U( infos->fichier, BIN_MODE_SOURCE_MSB );

    /* Initialiser correctement l'objet */
    switch ( objets->ob_type )
    {
    case G_STRING: /* ob_spec.txt_malloc pointe sur la chaîne de caractère [créée avec malloc()] */
    case G_TITLE:  /* dans le fichier : contient la position du début de la chaîne dans le fichier */
    case G_BUTTON:
      objets->ob_spec.txt_malloc = BI_RSC_LireChaine( infos, ob_spec );
#ifndef _DEVELOPPE_
      if ( NULL == objets->ob_spec.txt_malloc ) {
	(void) fprintf( stderr, "%s, %s l. %u: ERREUR en voulant lire la chaîne de l'objet %u {%u sur %u} [code %d] !\n",
			__FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets, retour );
	return -2;
      }
#endif
      break;

    case G_TEXT:  /* ob_spec.tedinfo : pointeur sur une TEDINFO */
    case G_FTEXT: /* dans le fichier : numéro de la TEDINFO dans la liste des TEDINFOS du fichier */
    case G_BOXTEXT:
    case G_FBOXTEXT:
      /* On prépare une tedinfo */
      retour = OBJ_PreparerComplement( num_arbre, idx_objet );
      if ( retour != 0 ) {
	(void) fprintf( stderr, "%s, %s l. %u: ERREUR en voulant préparer la TEDINFO de l'objet %u {%u sur %u} [code %d] !\n",
			__FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets, retour );
	return -3;
      }

      /* On lit ses paramètres */
      retour = BI_RSC_LireTedinfo( infos, ob_spec, objets->ob_spec.tedinfo );
      if ( retour != 0 ) {
	(void) fprintf( stderr, "%s, %s l. %u: ERREUR en voulant lire la TEDINFO de l'objet %u {%u sur %u} [code %d] !\n",
			__FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets, retour );
	return -4;
      }
      break;

    case G_BOX:  /* ob_spec.gbox = diverses valeurs ; masque de bit obligé car ordre des bits / des mots change d'une machine à l'autre ! */
    case G_IBOX: /* dans le fichier : ces valeurs, avec masque de bits en format Motorola 68k [Silicon, PPC : bits dans l'autre sens] */
    case G_BOXCHAR:
      objets->ob_spec.gbox.couleur_fond  = (ob_spec & 0xF);
      objets->ob_spec.gbox.motif         = ( (ob_spec >>  4) & 0x7 );
      objets->ob_spec.gbox.opaque        = ( (ob_spec >>  7) & 0x1 );
      objets->ob_spec.gbox.couleur_texte = ( (ob_spec >>  8) & 0xF );
      objets->ob_spec.gbox.couleur_cadre = ( (ob_spec >> 12) & 0xF );
      objets->ob_spec.gbox.cadre         = ( (ob_spec >> 16) & 0xFF );
      objets->ob_spec.gbox.caractere     = ( (ob_spec >> 24) & 0xFF );
      break;

    case G_IMAGE: /* ob_spec.image = pointeur sur BITBLK ; dans le fichier : numéro de cette BITBLK dans la table */
      (void) fprintf(stderr, "%s, %s l. %u: À FAIRE --- BITBLK pour l'objet %u {%u sur %u}!\n",
		     __FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets);
      objets->ob_spec.image = NULL;
      break;

    case G_ICON: /* ob_spec.icone = pointeur sur ICONBLK ; dans le fichier : numéro de cet ICONBLK dans la table */
      (void) fprintf(stderr, "%s, %s l. %u: À FAIRE --- ICONBLK pour l'objet %u {%u sur %u}!\n",
		     __FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets);
      objets->ob_spec.icone = NULL;
      break;

    case G_CICON: /* ob_spec.icone_coul = pointeur sur CICONBLK ; dans le fichier : ?? */
      (void) fprintf(stderr, "%s, %s l. %u: À FAIRE --- CICONBLK pour l'objet %u {%u sur %u}!\n",
		     __FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets);
      objets->ob_spec.icone_coul = NULL;
      break;

    default: /* Autres objets : on prépare un complément, adapté à leur type */
      retour = OBJ_PreparerComplement( num_arbre, idx_objet );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l. %u: ERREUR en voulant préparer le complément de l'objet %u {%u sur %u} [code %d] !\n",
		       __FILE__, __FUNCTION__, __LINE__, idx_objet, idx_objet + 1, nbr_objets, retour);
	return -3;
      }
    }
  }

  /* On se remet à l'ancienne position dans le fichier */
  (void) fseek( infos->fichier, ex_position, SEEK_SET );

  /* Tous les objets ont été lus correctement : on a donc fini */
  return 0;
}

/* -------- Lire une TEDINFO -------- */

static int BI_RSC_LireTedinfo(struct rsc_infos *infos, unsigned int idx_tedinfo, TEDINFO *tedinfo)
{
  long ex_position, position;
  unsigned long pos_chaine;

  if ( NULL == tedinfo ) {
    (void) fprintf( stderr, "%s, %s l. %u: On veut lire la TEDINFO %u, mais la TEDINFO n'a pas été indiquée !\n",
		    __FILE__, __FUNCTION__, __LINE__, idx_tedinfo + 1 );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* À quelle position se trouve la TEDINO ?
     Rq : selon Atari Compendium, ob_spec contient l'index de la TEDINFO dans la table
     Mais vu les valeurs lues, serait directement la position dans le fichier ?
  */
  position = idx_tedinfo;

  /* On sauve l'ancienne position */
  ex_position = ftell( infos->fichier );

  /* On va à la TEDINFO indiquée */
  if ( -1 == fseek( infos->fichier, position, SEEK_SET ) ) {
    (void) fprintf( stderr, "%s, %s l. %u: Erreur de lecture [%s]\n",
		    __FILE__, __FUNCTION__, __LINE__, strerror( errno ) );
    return -42;
  }

  /* On peut lire effectivement */
  /* a. Le texte */
  pos_chaine = BIN_Lire_32U( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_ptext.editable = BI_RSC_LireChaine( infos, pos_chaine );
#ifndef _DEVELOPPE_
  if ( NULL == tedinfo->te_ptext.editable ) {
    (void) fprintf( stderr, "%s, %s l. %u: ERREUR dans la lecture de la chaîne éditable, TEDINFO %u !\n",
		    __FILE__, __FUNCTION__, __LINE__, idx_tedinfo + 1 );
    return -3;
  }
#endif

  /* b. Le masque */
  pos_chaine = BIN_Lire_32U( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_ptmplt = BI_RSC_LireChaine( infos, pos_chaine );
  /* On peut avoir NULL pour un objet non-éditable, a priori ( ? ) */
#ifndef _DEVELOPPE_
  if ( NULL == tedinfo->te_ptmplt ) {
    (void) fprintf( stderr, "%s, %s l. %u: ERREUR dans la lecture du masque, TEDINFO %u !\n",
		    __FILE__, __FUNCTION__, __LINE__, idx_tedinfo + 1 );
    return -4;
  }
#endif

  /* c. La chaîne de validation */
  pos_chaine = BIN_Lire_32U( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_pvalid = BI_RSC_LireChaine( infos, pos_chaine );
#ifndef _DEVELOPPE_
  if ( NULL == tedinfo->te_pvalid ) {
    (void) fprintf( stderr, "%s, %s l. %u: ERREUR dans la lecture de la chaîne de validation, TEDINFO %u !\n",
		    __FILE__, __FUNCTION__, __LINE__, idx_tedinfo + 1 );
    return -5;
  }
#endif

  /* Autres informations */
  tedinfo->te_font      = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_junk1     = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_just      = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_color     = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_junk2     = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_thickness = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_txtlen    = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );
  tedinfo->te_tmplen    = BIN_Lire_16( infos->fichier, BIN_MODE_SOURCE_MSB );

  /* Contrôles */
  /* a. te_tmplen : nombre de caractères dans le masque [0 final INCLUS]
        ATTENTION, avec l'éventuelle conversion en UTF-8,
	le nombre de _caractères_ peut être inférieur au nombre d'_octets_...
   */
  if ( tedinfo->te_tmplen != ( TXT_CompterCaracteres( tedinfo->te_ptmplt ) + 1 ) ) {
    printf( "%s : te_txtlen = %u strlen(texte=\"%s\") = %lu strlen(valide=\"%s\") = %lu\n", __FUNCTION__,
	    tedinfo->te_txtlen, tedinfo->te_ptext.constant, (unsigned long) strlen( tedinfo->te_ptext.constant ),
	    tedinfo->te_pvalid, (unsigned long) strlen( tedinfo->te_pvalid ) );
    (void) fprintf( stderr, "%s, %s l.%u : Incohérence, te_tmplen = %u, strlen(\"%s\") = %lu\n",
		    __FILE__, __FUNCTION__, __LINE__, tedinfo->te_tmplen, tedinfo->te_ptmplt,
		   (unsigned long) strlen( tedinfo->te_ptmplt ) );
    return -28;
  }

  /* b. te_txtlen : nombre de caractères dans le masque [0 final INCLUS]
        ATTENTION, avec l'éventuelle conversion en UTF-8,
	le nombre de _caractères_ peut être inférieur au nombre d'_octets_...
   */
  if ( tedinfo->te_txtlen != ( TXT_CompterCaracteres( tedinfo->te_ptext.constant ) + 1 ) ) {
    printf( "%s : te_tmplen = %u strlen(tmplt=\"%s\") = %lu strlen(valide=\"%s\") = %lu\n", __FUNCTION__,
	    tedinfo->te_tmplen, tedinfo->te_ptmplt, (unsigned long) strlen( tedinfo->te_ptmplt ),
	    tedinfo->te_pvalid, (unsigned long) strlen( tedinfo->te_pvalid ) );
    (void) fprintf( stderr, "%s, %s l.%u : Incohérence, te_txtlen = %u, strlen(\"%s\") = %lu\n",
		    __FILE__, __FUNCTION__, __LINE__, tedinfo->te_txtlen, tedinfo->te_ptext.constant,
		    (unsigned long) strlen( tedinfo->te_ptext.constant ) );
    return -29;
  }

  /* On se remet à l'ancienne position dans le fichier */
  (void) fseek( infos->fichier, ex_position, SEEK_SET );

  /* La lecture s'est effectuée sans difficulté */
  return 0;
}
