/* LASE - Logiciel d'analyse des spectres E.X.A.F.S.
   (c) 1996, 2001 - Emmanuel CURIS ; 
   Cration du fichier : 27 mai 1998

   Module d'extraction des oscillations EXAFS

   Premire tape de la soustraction de mu_0 : modlisation par un polynme
   Contient aussi la recherche automatique de E0
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"

#include "automatise.h"
#include "chimie_utils.h" /* Pour les symboles des lments */
#include "lissage.h"
#include "extrait.h"
#include "science.h"
#include "absorption.h" /* Pour la recherche de seuil */

#include "etape_1.txt"

/* ***************** Fonctions locales de ce module ************************************** */

/* *-*-*-*-*-*-*-*-* Calcul proprement dit *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* */

/* ----------------- Calcul du polynome servant  modliser au mieux mu_0 -----------------\
 |  ne renvoie rien
 |
 | brut  : le spectre brut dont on veut extraire l'EXAFS
 | infos : les informations ncessaires  la modlisation de mu_0
 \---------------------------------------------------------------------------------------- */

int XAS_MU01_CalculerMu0_I(const Graphe *brut, MU1_infos *infos)
{
  double erreur = 0.0;	   /* Erreur entre le polynme et la courbe */
  double bornes[ 2 ]; /* Intervalle d'approximation du polynme  ter */

  bornes[ 0 ] = infos->e_2;
  if ( infos->avec_e3 == VRAI )
    bornes[ 1 ] = infos->e_3;
  else
    bornes[ 1 ] = brut->x_max;

  if ( infos->methode == ELIM_MU0_RAPIDE ) {    /* Moindres carrs classiques */
    erreur = MC_Polynome_SurPartieGraphe( brut, infos->degre, 1, bornes, infos->coefs );
  } else if ( infos->methode == ELIM_MU0_LENT ) { /* Pondration par Dx demande */
    erreur = MC_Polynome_SurPartieGraphe_Juris( brut, infos->degre, 1, bornes, infos->coefs );
  }

  infos->residu = erreur;
  /* Tout s'est bien pass */
  return 0;
}

/* ----------------- Extraction proprement dite du spectre E.X.A.F.S. --------------------\
 |  Suppose - que le graphe brut soit tri par X croissants                               |
 |          - que les graphes exafs et mu0 soient rservs pour au moins autant de points |
 |            que le signal brut                                                          |
 \--------------------------------------------------------------------------------------- */

int XAS_MU01_ExtraireEXAFS(const Graphe *brut, MU1_infos *infos)
{
  register double *x_lu, *y_lu, *x_poly, *y_poly, *x_exafs, *y_exafs;
  unsigned long nbr_points, i, premier;
  boolean npt_change = FAUX;

  if (brut == NULL) return -1;

  /* Combien de points du spectre conviennent ? */
  nbr_points = GRF_intervalle(brut, infos->e_0, ( VRAI == infos->avec_e3 ) ? (infos->e_3) : (brut->x_max), &premier);
  if ( nbr_points < 3 ) {
    (void) BI_ERR_alerte(NULL, &ERR_AucunPoint, VRAI, infos->e_0, ( VRAI == infos->avec_e3 ) ? (infos->e_3) : (brut->x_max));
    return -2;
  }

  if (nbr_points != infos->exafs->nombre_point) {
    if ( FAUX == GRF_AdapteTaille( infos->exafs, nbr_points ) )
      return -3;
    if ( FAUX == GRF_AdapteTaille( infos->mu_0, nbr_points ) )
      return -3;
    npt_change = VRAI;
  }

  /* Calcul du polynme et du spectre extrait */
  x_poly  = infos->mu_0->x ; y_poly  = infos->mu_0->y;
  x_exafs = infos->exafs->x; y_exafs = infos->exafs->y;
  for ( i = 0, x_lu = brut->x + premier, y_lu = brut->y + premier ; i < nbr_points ; i++, x_lu++, y_lu++ ) {
    double x, y;
    short unsigned int j;

    x = *x_lu;
    *x_exafs = x;
    *x_poly  = x;
    y = infos->coefs[0];
    for (j = 1 ; j <= infos->degre ; j++) {
      y += x * infos->coefs[j];
      x *= *(x_lu);
    }
    *y_poly  = y;
    *y_exafs = *y_lu - y;
    x_poly++; y_poly++; x_exafs++; y_exafs++;
  }
  GRF_bornes(infos->exafs, DESORDONNE, DESORDONNE);
  GRF_bornes(infos->mu_0, DESORDONNE, DESORDONNE);

  /* On met les bonnes units en X, pour que la suite du traitement soit correcte */
  GRF_ChangerUnite( infos->exafs, AXE_X, GRF_QuelleUnite( brut, AXE_X ) );
  GRF_ChangerUnite( infos->mu_0, AXE_X, GRF_QuelleUnite( brut, AXE_X ) );

  /* On convertit en chelle des k (adapte les bornes en X en mme temps) */
  EXF_e_v_k(infos->exafs, infos->e_0, FAUX);
  EXF_e_v_k(infos->mu_0 , infos->e_0, FAUX);

  /* Multiplication par k^n */
  EXF_k_p_n(infos->exafs, infos->deg_affiche, FAUX); /* Multiplication par k^n du signal extrait */
  EXF_k_p_n(infos->mu_0,  infos->deg_affiche, FAUX); /* et du polynome soustrait                 */

  /* On informe le programme que le graphe a t modifi.
     La modif de 0 n'est pas signale, elle est implicite */
  GRF_modifier(infos->exafs, infos->exafs, npt_change);

  /* On calcule le 0 en E0 : donne une estimation de la hauteur de seuil;
     on la reporte dans l'arbre de normalisation de l'EXAFS */
  {
    double h_seuil, x;
    short unsigned int j;

    x = infos->e_0;
    h_seuil = infos->coefs[0];
    for (j = 1 ; j <= infos->degre ; j++) {
      h_seuil += x * infos->coefs[j];
      x *= infos->e_0;
    }
    OBJ_MettreNombre( NORM_EXAFS, NEXF_H_SEUIL, h_seuil, 0.0 );
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Traitement post-extraction du signal EXAFS -------- */

int XAS_MU01_PostExtraction(const struct graphe *brut, struct graphe *exafs_resultat, MU1_infos *infos)
{
  char *titre = NULL;

  if ( NULL == exafs_resultat ) return -1; /* Pas de graphe  amliorer ! */
  if ( NULL == brut ) return 0; /* Pas d'amlioration possible ! */

  /* On construit un titre adapt */
  if ( GR_MOINSMAT == brut->type ) {
    /* On te le fond continu  un spectre XAS */
    titre = TXT_genere_chaine( "%s - "MU"0,I", brut->nom );
  } else {
    /* Autre cas : titre plus gnrique */
    titre = TXT_genere_chaine( txt_titre_generique, brut->nom );
  }

  /* On installe ce titre */
  GRF_titre( exafs_resultat, titre );
  free( titre ); titre = NULL;

  /* On adapte si ncessaire le type */
  if ( GR_MOINSMAT == brut->type ) {
    /* On te l'absorption atomique  un spectre XAS */
    GRF_ChangerType( exafs_resultat, GR_EXAFS1 );
  } else {
    /* On garde le mme type que l'original ! */
  }

  /* On conserve les modes d'affichage & autres infos annexes */
  GRF_CopierAffichage( exafs_resultat, brut, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y
		       | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );
  /* Attention, il faut s'assurer que l'on a bien la bonne unit en X et pas l'unit par dfaut ! */
  GRF_ChangerUnite( exafs_resultat, AXE_X, GRF_QuelleUnite( brut, AXE_X ) );

  /* Conversion dans l'chelle des k */
  EXF_e_v_k( exafs_resultat, infos->e_0, VRAI );

  /* Pondration par k^n */
  EXF_k_p_n( exafs_resultat, infos->deg_affiche, VRAI );

  exafs_resultat->drapeaux.sauve = FAUX;   /* Graphe non sauv */

  /* Termin */
  return 0;
}

/* -------- Extraction effective pour un graphe avec propagation des erreurs,... -------- */

Graphe *MU1_ExtraitEXAFS(Graphe *brut, MU1_infos *infos, boolean recalcule)
{
  unsigned long nbr_pts, premier;
  int retour;

  if ( ( NULL == brut ) || ( NULL == infos ) ) return NULL;
  if ( brut->genre_donnees != GR_NORMAL ) {
    (void) fprintf(stderr, "MU1_ExtraitEXAFS : appel avec un graphe complexe ou 3D [%s] !\n", brut->nom);
    return NULL;
  }

  /* On cherche le nombre de points  conserver et leur position
     SUPPOSE que le graphe brut est tri par X croissants */
  {
    double x_min, x_max;

    x_min = infos->e_0; /* Dbut du spectre EXAFS */
    if ( VRAI == infos->avec_e3 )
      x_max = infos->e_3;
    else
      x_max = brut->x_max;

    nbr_pts = GRF_intervalle(brut, x_min, x_max, &premier);
    if ( nbr_pts < 3 ) { /* Trop peu de points */
      return NULL;
    }
  }

  if ( infos->exafs == NULL ) { /* On n'a pas encore le graphe, il faut le crer */
    infos->exafs = GRF_ExtraitGraphe( brut, premier, nbr_pts, VRAI );
    if ( NULL == infos->exafs ) {
      return NULL;
    }
#ifdef _DEVELOPPE_
    GRF_titre( infos->exafs, "EXAFS extrait [mu0,I], temporaire" );
#endif
  } else { /* On rextrait les infos pertinentes */
    memcpy( infos->exafs->x, brut->x + premier, sizeof(double) * nbr_pts);
    memcpy( infos->exafs->y, brut->y + premier, sizeof(double) * nbr_pts);
    GRF_bornes(infos->exafs, DESORDONNE, DESORDONNE);

    if ( brut->erreur_x != NULL ) {
      Graphe *erreur;

      erreur = GRF_creer_erreurs(infos->exafs, AXE_X);
      if (erreur != NULL) {
	memcpy(erreur->y, brut->erreur_x->y + premier, sizeof(double) * nbr_pts);
	memcpy(erreur->z.ddl, brut->erreur_x->z.ddl + premier, sizeof(int) * nbr_pts);

	GRF_bornes(erreur, DESORDONNE, DESORDONNE);
      }
    }
    if (brut->erreur_y != NULL) {
      Graphe *erreur;

      erreur = GRF_creer_erreurs(infos->exafs, AXE_Y);
      if (erreur != NULL) {
	memcpy(erreur->z.ddl, brut->erreur_y->z.ddl + premier, sizeof(int) * nbr_pts);
	memcpy(erreur->y, brut->erreur_y->y + premier, sizeof(double) * nbr_pts);

	GRF_bornes(erreur, DESORDONNE, DESORDONNE);
      }
    }
  }
  if ( NULL == infos->mu_0 ) {
    char *titre;

    titre = TXT_genere_chaine(MU"0,I pour %s", brut->nom);
    infos->mu_0 = GRF_nouveau( NULL, nbr_pts, GR_MU0, GR_NORMAL, titre );
    free(titre);
    if ( NULL == infos->mu_0 ) {
      GRF_efface(infos->exafs, VRAI, VRAI);
      return NULL;
    }
  }
  memcpy( infos->mu_0->x, brut->x + premier, nbr_pts * sizeof(double));
  GRF_bornes(infos->mu_0, DESORDONNE, NE_PAS_FAIRE);
  GRF_CopierAffichage( infos->mu_0, brut, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y );
  GRF_ChangerUnite( infos->mu_0, AXE_X, GRF_QuelleUnite( brut, AXE_X ) );

  /* En revanche, on refait toute l'extraction, pour faire en mme temps la propagation des erreurs */

  /* Polynme  ter [NULL : matrice de covariance des coefficients] */
  if ( VRAI == recalcule )
    XAS_MU01_CalculerMu0_I( brut, infos );
  GRF_GenerePolynome( infos->mu_0, FAUX, infos->degre, infos->coefs, NULL );

  /* Soustraction proprement dite -- s'occupe aussi des erreurs
     NULL :  remplacer par les termes de covariance... */
  retour = GRF_SoustraireGraphe( infos->exafs, infos->mu_0, VRAI, NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_SoustraireGraphe() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);

    GRF_efface(infos->mu_0, VRAI, VRAI);
    GRF_efface(infos->exafs, VRAI, VRAI);
    return NULL;
  }

  /* Log dans l'historique */
  HIST_Titre(1, "Premi"e_GRAVE"re "e_AIGU"tape de l'extraction de l'EXAFS", "MU0_I");
  HIST_Texte(1, 0, "Signal brut   : %s\n", brut->nom);
  HIST_Texte(1, 0, "EXAFS extrait : %s\n", infos->exafs->nom);
  HIST_Texte(1, 0, "Mod"e_GRAVE"le de "MU"0  : %s\n", infos->mu_0->nom);
  HIST_Texte(3, 0, E_AIGU"nergie de seuil : E0 = %.10f eV\n", infos->e_0);
  HIST_Texte(2, 0, "M"e_AIGU"thode : Soustraction d'un polyn"o_CIRC"me");
  HIST_Texte(3, 0, " de degr"e_AIGU" %u", (unsigned int) infos->degre);
  switch ( infos->methode ) {
  case ELIM_MU0_RAPIDE:
    HIST_Texte(4, 0, " (moindres carr"e_AIGU"s simples)");
    break;

  case ELIM_MU0_LENT:
    HIST_Texte(4, 0, " (moindres carr"e_AIGU"s pond"e_AIGU"r"e_AIGU"s par (x(i) - x(i-1)))");
    break;
  }
  HIST_Texte(2, 0, "\n"); /* Pour finir la ligne quel que soit le niveau de l'historique */
  HIST_Texte(3, 0, "Intervalle utilis"e_AIGU" : [%f, %f] (eV)\n", infos->e_2,
	     ( VRAI == infos->avec_e3 ) ? (infos->e_3) : (brut->x_max) );
  HIST_Polynome(4, 5, 2, infos->coefs, infos->degre, 'a');

  /* Conversion dans l'chelle des k */
  EXF_e_v_k(infos->mu_0 , infos->e_0, VRAI);

  /* Pondration par k^n */
  EXF_k_p_n(infos->mu_0, infos->deg_affiche, VRAI);

  /* On a fini */
  return infos->exafs;
}

/* -------- Automatisation de l'extraction -------- */

int XAS_MU01_Automatisation(struct automatiser *lot)
{
  Graphe *brut, *exafs_resultat = NULL/* , *mu0_resultat = NULL */;
  MU1_infos *infos = NULL;
  unsigned long i;

  if ( NULL == lot ) return -1;

  /* On rcupre les conditions d'extraction */
  infos = (MU1_infos *) lot->infos_necessaires;
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas d'infos d'extraction !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On traite chaque graphe de la srie */
  for ( i = 0 ; i < lot->nombre_objets ; i++ ) {
    brut = lot->objets[ i ].graphe;
    if ( NULL == brut ) {
      AVERTISSEMENT( "Graphe %lu [0..%lu] du lot NULL", i, lot->nombre_objets - 1 );
      continue;
    }

    /* On perd le lien avec les rsultats prcdents */
    infos->mu_0 = NULL;
    infos->exafs = NULL;
    exafs_resultat = NULL;

    /* En fonction de la mthode de calcul choisie... */
    switch ( lot->methode )
    {
    case AUTOMATISE_METHODE_ROBOT:
      /* Mme mthode, avec les mmes coefficients */
      /* Pas de rfrence */

      /* On calcule le 0 avec les coefficients indiqus */
      exafs_resultat = MU1_ExtraitEXAFS( brut, infos, FAUX );
      break;

    case AUTOMATISE_METHODE_ADAPTE:
      /* Mme mthode, avec les coefficients recalculs */
      /* Pas de rfrence */

      /* On calcule le 0 avec la meme mthode, mais des coefficients recalculs */
      exafs_resultat = MU1_ExtraitEXAFS( brut, infos, VRAI );
      break;

    case AUTOMATISE_METHODE_OPTIMISE:
      OPTION_A_FAIRE( "Optimiser le choix des coefficients pour coller  la rfrence" );
      /* La rfrence  utiliser */
/*       infos_mum->reference = lot->reference.graphe; */

      /* Le reste est comme la mthode "adapte" */
      break;
    }

    /* On fait la post-extraction */
    XAS_MU01_PostExtraction( brut, exafs_resultat, infos );

    /* On mmorise le rsultat dans la table des rsultats */
    lot->resultats[ i ].graphe = exafs_resultat;
    exafs_resultat = NULL;
  }

  /* Termin : tout s'est bien pass */
  return 0;
}

/* ------------ Recherche de E0 : maximum de la drive premire, en attendant mieux... -- */

double cherche_e0(Graphe *graphe, boolean affiche)
{
  register double *zone_x,*zone_y;
  double delta_x,derivee,derivee_max=0,*pos_x = NULL;
	
  if (graphe->type == GR_DERIVEE) { /* Le graphe est la drive => rapide */
    zone_y = graphe->y;
    while ( *(zone_y) < graphe->y_max ) /* On n'est pas au maximum */
      zone_y++;
    
    pos_x = graphe->x + (zone_y - graphe->y); /* X correspondant  ce Y */
  } else {
    unsigned long i;

    zone_x = graphe->x;
    zone_y = graphe->y;
    for (i = 0 ; i < graphe->nombre_point-1 ; i++) {
      delta_x = *(zone_x + 1) - *(zone_x);
      if (delta_x == 0) /* Drive infinie */
	return *(zone_x);
		
      derivee  = *(zone_y + 1) - *(zone_y);
      derivee /= delta_x;
      if (derivee > derivee_max) {
	derivee_max = derivee;
	pos_x = zone_x; /* Point juste avant la drive maximale... */
      }
      zone_x++;
      zone_y++;
    }
  }

  if (pos_x == NULL) {
    (void) BI_ERR_alerte(NULL, &err_seuil_pas_trouve, VRAI);
    return 0.0;
  }
	
  if (affiche == VRAI) { /* On affiche le rsultat */
    int Z = SYMBOLE_CHIMIQUE_INCONNU, seuil = 857;
    const char *type_seuil = "non trouv"e_AIGU;

    HIST_Titre(3, "Recherche d'un seuil", "SEUIL");
    HIST_Texte(3, 2, "Graphe : %s\n", graphe->nom);
    HIST_Texte(3, 2, "Seuil propos"e_AIGU" : E0 = %f eV", *pos_x);

    Z = TH_IdentifierSeuil(*pos_x, 10.0, &seuil, VRAI); /* 10 eV de marge : 5 eV insuffisant sur certains spectres Mn [5,05] */
    if ( Z >= HYDROGENE ) {
      type_seuil = TH_TexteSeuil( seuil );
      HIST_Texte(3, 0, " (Attribution : seuil %s, %s)\n", type_seuil, ATM_Z_vers_symbole( Z ) );
    } else {
      HIST_Texte(3, 0, " (Pas de seuil connu "a_GRAVE" moins de 10 eV\n");
      (void) fprintf(stderr, "cherche_e0 : Seuil inconnu, seuil = %d [E = %g, Z = %d]\n", seuil, *pos_x, Z);
    }

    (void) BI_ERR_alerte(NULL, &alerte_affiche_seuil, VRAI, (*pos_x) / 1000.0,
			 ATM_Z_vers_symbole(Z), type_seuil, TH_DonneesSeuil(Z, seuil, NULL, NULL) / 1000.0);
  }

  return *(pos_x);
}
