/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1996 - 1998, Emmanuel CURIS
   Création du fichier   :  2 octobre  1998
   Dernière modification : 24 novembre 1998

   Tout ce qui concerne l'extraction du signal EXAFS à partir du signal d'absorption brut

   Troisième étape de la soustraction de mu_0 (spline avec contrôle)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"

#include "extrait.h"
#include "filtrage.h" /* Pour TF_ouvre() dans traitement auto. */
#include "tracer.h"

#include "etape_3.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define CHAINE_COMPLETE  2	/* À remplacer par les I_XAS_CHAINE_* en séparant l'interface ! */
#define CHAINE_PARTIELLE 3

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int  CLV_mu03(Fenetre *fenetre, unsigned int touche, int etat);
static int  GER_mu03(unsigned int num_objet, int x, int y);
static int  MSG_mu03(Fenetre *fenetre, Message *message);
static void RDS_mu03(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);

static int  prepare_arbre_mu3(unsigned int);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~~~~~~~~~~~~~~~~~ Calculs effectifs de la correction à soustraire ~~~~~~~~~~~~~~~~~~~~~ */

#define avec_p_p /* Si défini, on utilise tableaux de type double **, sinon de type double *tableaux[10]
		    Sur Hades 40, compilateur = Turbo C 1.0 : méthode avec double ** un poil plus rapide
		    Sur alpha 433/au, compilateur = cc : indiscernable 
		 */

/* ----------------- Calcul de la courbe servant à modéliser au mieux mu_0 ----------------\
 |  ne renvoie rien                                                                        |
 |  Le pas du graphe est supposé constant (en-dehors du graphe, Dx = celui dans le graphe) |
 |  Le graphe est supposé nul en-dehors des points définis                                 |
 |  Le graphe doit être trié et sans doublon                                               |
 |                                                                                         |
 | brut  : le spectre brut dont on veut extraire l'EXAFS                                   |
 | infos : les informations nécessaires à la modélisation de mu_0                          |
 \---------------------------------------------------------------------------------------- */

void MU3_calc(Graphe *brut, MU3_infos *infos)
{
  register double *y_brut,*y_mu0;
  double d_x1, d_x2 = 0.0, y1, y2, p = 0.0, controle;
  unsigned long j, nbr_pts;
  boolean sortie = FAUX;

#ifdef _DEVELOPPE_
  unsigned long temps; /* Juste pour les opérations de chronométrage */
#endif

#ifdef avec_p_p
  double **tableaux;
  tableaux = (double **) calloc(10, sizeof( double *) );
  if (tableaux == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "etape_3.c", "", "tableaux");
    return;
  }
#else
  double *tableaux[10];
#endif

#ifdef _DEVELOPPE_
  temps = clock();
  printf("Valeur de lissage : %f\n", infos->parametre);
#endif

  nbr_pts = brut->nombre_point + 1;

  for (j = 0 ; j < 10 ; j++) {
    tableaux[j] = (double *) calloc( 1 + nbr_pts, sizeof(double) );
    if (tableaux[j] == NULL) {
      unsigned int i;

      (void) fprintf(stderr, "MU3_calc : mémoire insuffisante, tableau %lu sur 10\n", j + 1);
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "etape_3.c", "", "tableaux[i]");
      for (i = 0 ; i < j ; i++ )
	free(tableaux[i]);

#ifdef avec_p_p      
      free(tableaux);
#endif
      return;
    }
  }

  tableaux[3][0] = 0.0; tableaux[3][1] = 0.0;
  tableaux[3][nbr_pts - 1] = 0.0;
  tableaux[4][0] = 0.0; tableaux[4][1] = 0.0;
  tableaux[5][0] = 0.0; tableaux[5][1] = 0.0;
  tableaux[5][nbr_pts - 1] = 0.0;
  tableaux[6][0] = 0.0; tableaux[6][1] = 0.0;
  tableaux[7][0] = 0.0; tableaux[7][1] = 0.0;
  tableaux[8][0] = 0.0; tableaux[8][1] = 0.0; tableaux[8][2] = 0.0;
  tableaux[8][nbr_pts] = 0.0; tableaux[8][nbr_pts - 1] = 0.0;

  d_x1 = (brut->x)[1] - (brut->x)[0];
  if ( d_x1 == 0.0 ) {
    (void) EXF_ERREUR(ERR_DOUBLONS);
    for ( j = 0 ; j < 10 ; j++ ) /* On libère la mémoire */
      free(tableaux[ j ]);
#ifdef avec_p_p      
    free(tableaux);
#endif
    return;
  }
  y1 = ( (brut->y)[1] - (brut->y)[0] ) / d_x1;

  tableaux[3][1] = 1/d_x1; /* On aura besoin de cette valeur : vaut tab[5][0] en fait... */

  for ( j = 2 ; j < nbr_pts - 1 ; j++ ) { /* Brut[i] : 0 <= i < nbr_pts - 1 */
    d_x2 = d_x1; y2 = y1;
    d_x1 = (brut->x)[ j ] - (brut->x)[ j - 1 ];
    if ( d_x1 == 0.0 ) {
      (void) EXF_ERREUR(ERR_DOUBLONS);
      for ( j = 0 ; j < 10 ; j++ ) /* On libère la mémoire */
	free( tableaux[ j ] );
#ifdef avec_p_p      
      free(tableaux);
#endif
      return;
    }

    y1 = ( (brut->y)[j] - (brut->y)[j - 1] ) / d_x1;
    (infos->mu_0->y)[j - 1] = y1 - y2;
    tableaux[6][j] = 2 * (d_x1 + d_x2) / 3;
    tableaux[7][j] = d_x1 / 3;
    tableaux[3][j] = 1 / d_x1;
    tableaux[4][j] = - (tableaux[3][j - 1] + tableaux[3][j]);
  }

  /* On a rempli tab[   4, 5, 6, 7][           ] de 2 à nbr_pts - 2
                 tab[3            ][           ] de 1 à nbr_pts - 2
		 tab[3            ][0          ] = 0 toujours
		 tab[   4, 5, 6, 7][0, 1       ] = 0 toujours
		 tab[3, 4, 5, 6, 7][nbr_pts - 1] = 0
		 tab[3, 4, 5, 6, 7][nbr_pts    ] non définis     */

  /* Ici, on a besoin de tab[3][1 ... nbr_pts - 2]
	                 tab[4][2 ... nbr_pts - 1] ==> ok */

  for (j = 2 ; j <= nbr_pts - 2 ; j++) {
    tableaux[0][j] = tableaux[3][j] * tableaux[3][j]
	               + tableaux[4][j] * tableaux[4][j]
		       + tableaux[3][j - 1] * tableaux[3][j - 1];
    tableaux[1][j] = tableaux[3][j] * tableaux[4][j + 1]
                       + tableaux[4][j] * tableaux[3][j];
    tableaux[2][j] = tableaux[3][j] * tableaux[3][j + 1];
    tableaux[8][j] = 0.0; /* Supposé nul - rien à ce sujet dans le source de Juris */
  }
  /* On a rempli tab[0, 1, 2][                    ] de 2 à nbr_pts - 2
	         tab[0, 1, 2][0          , 1      ] non définis
	         tab[0, 1, 2][nbr_pts - 1, nbr_pts] non définis        */
	
  controle = - (infos->parametre);

  tableaux[3][1] = 0; /* Remise à cette valeur pour avoir le résultat correct */

  while (sortie == FAUX) {
    /* Ici, on a besoin de tab[3][0 ... nbr_pts - 4] ==> ok
	                   tab[0][2 ... nbr_pts - 2] ==> ok
	                   tab[1][2 ... nbr_pts - 2] ==> ok
			   tab[2][2 ... nbr_pts - 2] ==> ok
			   tab[6][2 ... nbr_pts - 2] ==> ok
			   tab[7][2 ... nbr_pts - 2] ==> ok
			   tab[8][0 ... nbr_pts - 4] ==> ok */

    for (j = 2 ; j <= nbr_pts - 2 ; j++) {
      tableaux[4][j - 1] = y1   * tableaux[3][j - 1];
      tableaux[5][j - 2] = d_x2 * tableaux[3][j - 2];
      tableaux[3][j] = 1 / ( p * tableaux[0][j]     +        tableaux[6][j]
			      - y1 * tableaux[4][j - 1] - d_x2 * tableaux[5][j - 2]);
      tableaux[8][j] = (infos->mu_0->y)[j - 1] - tableaux[4][j - 1] * tableaux[8][j - 1]
	                 - tableaux[5][j - 2] * tableaux[8][j - 2];
      y1 = p * tableaux[1][j] + tableaux[7][j] - d_x1 * tableaux[4][j - 1];
      d_x2 = d_x1;
      d_x1 = tableaux[2][j] * p;
    }
    /* On a rempli tab[4][ ] de 1 à nbr_pts - 3
                   tab[5][ ] de 0 à nbr_pts - 4
		   tab[3][ ] de 2 à nbr_pts - 2
		   tab[8][ ] de 2 à nbr_pts - 2 */

    tableaux[4][nbr_pts - 2] = y1   * tableaux[3][nbr_pts - 2]; /* Valeurs non définies dans le code de Juris   */
    tableaux[5][nbr_pts - 3] = d_x2 * tableaux[3][nbr_pts - 3]; /* Semble pas modifier le résultat, mais bon... */
    tableaux[5][nbr_pts - 2] = d_x1 * tableaux[3][nbr_pts - 2];

    /* On a besoin de tab[3][2 ... nbr_pts - 2] ==> ok
                      tab[4][2 ... nbr_pts - 2] ==> ?? pour nbr_pts-2 - modifié
		      tab[5][2 ... nbr_pts - 2] ==> ?? pour nbr_pts-2 et -3 - modifié
		      tab[8][2 ... nbr_pts    ] */
    for (j = nbr_pts - 2 ; j > 1 ; j--) {
      tableaux[8][j] = tableaux[3][j] * tableaux[8][j]
	                 - tableaux[4][j] * tableaux[8][j + 1]
			 - tableaux[5][j] * tableaux[8][j + 2];
    }
    /* On a rempli tab[8][ ] de 2 à nbr_pts - 2 */

    /* On a besoin de tab[8][1 ... nbr_pts - 1] ==> ?? pour 1 et nbr_pts - 1, pris nuls */
    tableaux[ 8 ][ 1 ] = 0.0;
    tableaux[ 8 ][ nbr_pts - 1 ] = 0.0;

    y2 = 0.0; d_x1 = 0.0;
    for (j = 1 ; j <= nbr_pts - 2 ; j++) {
      d_x2 = d_x1;
      d_x1 = ( tableaux[8][j + 1] - tableaux[8][j] )
	   / ( (brut->x)[j] - (brut->x)[j - 1] );
      tableaux[9][j] = (d_x1 - d_x2);
      y2 += tableaux[9][j] * (d_x1 - d_x2);
    }
    
    /* On a rempli tab[9][ ] de 1 à nbr_pts - 2 */
    tableaux[9][nbr_pts - 1] = -d_x1;

    d_x2 = -d_x1;
    y2  -= d_x2 * d_x1;
    d_x2 = controle;
    controle = y2 * p * p;
    if ( (controle >= infos->parametre) || (controle <= d_x2) ) {
      sortie = VRAI; /* On a fini la boucle... */
    } else {
      y1 = 0.0;

      /* On a besoin de tab[9][1 ... nbr_pts - 1]
	                tab[3][0 ... nbr_pts - 2]
			tab[4][1 ... nbr_pts - 3]
			tab[5][0 ... nbr_pts - 4] */
			                  
      d_x1 = ( tableaux[9][2] - tableaux[9][1] )
	   / ( (brut->x)[1] - (brut->x)[0] );
      for (j = 2 ; j <= nbr_pts - 2 ; j++) {
	d_x2  = d_x1;
	d_x1  = ( tableaux[9][j + 1] - tableaux[9][j] )
	      / ( (brut->x)[j] - (brut->x)[j - 1] );
	d_x2  = d_x1 - d_x2;
	d_x2 -= tableaux[4][j - 1] * tableaux[3][j - 1];
	d_x2 -= tableaux[5][j - 2] * tableaux[3][j - 2];
	y1   += d_x2 * d_x2 * tableaux[3][j];
	tableaux[3][j] = d_x2;
      }

      /* On a rempli tab[3][2 ... nbr_pts - 2] */
      d_x1 = y2 - p * y1;
      if (d_x1 <= 0 ) 
	sortie = VRAI; /* On sort de la boucle */
      else {
	double nv_p;

	nv_p = p + (infos->parametre - controle) / ( d_x1 * ( sqrt(infos->parametre / y2) + p ) );
	if ( nv_p == p ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Variation trop faible ou erreur quelque part\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  sortie = VRAI;
	}
	p = nv_p;
      }
    }
  }

  /* Fin du calcul */
  y_brut = brut->y;
  y_mu0 = infos->mu_0->y;	
  for (j = 0 ; j < brut->nombre_point ; j++, y_mu0++, y_brut++) {
    *y_mu0 = (*y_brut) - p * tableaux[9][j + 1];
  }

  for (j = 0 ; j < 10 ; j++) /* On libère la mémoire */
    free(tableaux[j]);

#ifdef avec_p_p      
  free(tableaux);
# ifdef _DEVELOPPE_
  printf("Temps "e_AIGU"coul"e_AIGU", avec double ** : %lu \n", clock() - temps);
# endif /* def _DEVELOPPE_ */
#else
# ifdef _DEVELOPPE_
  printf("Temps "e_AIGU"coul"e_AIGU", avec double *[] : %lu \n", clock() - temps);
# endif /* def _DEVELOPPE_ */
#endif /* def avec_p_p */

  GRF_bornes(infos->mu_0, ORDONNE, DESORDONNE);
}

/* ----------------- Extraction proprement dite du spectre E.X.A.F.S. -------------------- */

void MU3_extrait(Graphe *brut, MU3_infos *infos)
{
  register double *brut_y, *exafs_y, *mu0_y;
  unsigned long i;

  brut_y  = brut->y;
  exafs_y = infos->exafs->y;
  mu0_y   = infos->mu_0->y;
	
  for (i = 0 ; i < brut->nombre_point ; i++) {
    *exafs_y = *brut_y - *mu0_y;
    brut_y++; exafs_y++; mu0_y++;
  }

  /* Recherche des bornes */
  GRF_bornes(infos->exafs, ORDONNE, DESORDONNE); /* Le graphe est normalement trié */ 

  /* Le graphe a été modifié... */
  GRF_modifier(infos->exafs, infos->exafs, FAUX); /* Nbr de points n'a pas changé */
}

/* -------- Extraction complète avec propagation des erreurs et tout et tout -------- */

Graphe *MU3_ExtraitEXAFS(Graphe *brut, MU3_infos *infos, boolean recalcule __attribute__((__unused__)))
{
  char *titre;
  int retour;

  if ( (NULL == brut) || (NULL == infos) ) return NULL;
  if ( GR_NORMAL != brut->genre_donnees ) return NULL;

  if ( NULL == infos->exafs ) {
    infos->exafs = GRF_Copier( brut, NULL, FAUX );
    if ( NULL == infos->exafs ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de préparer l'EXAFS\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
    infos->exafs->type = GR_EXAFS3;
  } else {
    if ( infos->exafs->nombre_point != brut->nombre_point ) {
      if ( FAUX == GRF_AdapteTaille(infos->exafs, brut->nombre_point) ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible d'adapter la taille\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return NULL;
      }
      memcpy(infos->exafs->x, brut->x, sizeof(double) * infos->exafs->nombre_point);
      infos->exafs->x_min = brut->x_min;
      infos->exafs->x_max = brut->x_max;
    }
    memcpy(infos->exafs->y, brut->y, sizeof(double) * infos->exafs->nombre_point);
    infos->exafs->y_min = brut->y_min;
    infos->exafs->y_max = brut->y_max;    

    /* On copie les erreurs : inchangées par ce traitement */
    infos->exafs->erreur_x = GRF_Copier( brut->erreur_x, NULL, VRAI );
    infos->exafs->erreur_y = GRF_Copier( brut->erreur_y, NULL, VRAI );
  }

  if ( NULL == infos->mu_0 ) {
    /* On le crée, avec la même grille en X que le graphe original */
    infos->mu_0 = GRF_nv_TableX( brut, GR_MU0, GR_NORMAL, MU"0,III", FAUX, brut->x, brut->nombre_point, VRAI );
    if ( NULL == infos->mu_0 ) {
      GRF_efface(infos->exafs, VRAI, VRAI);
      return NULL;
    }
    recalcule = VRAI; /* Rq : en pratique, pas utilisé ! */
  } else {
    if ( infos->mu_0->nombre_point != brut->nombre_point ) {
      if ( FAUX == GRF_AdapteTaille(infos->mu_0, brut->nombre_point) ) {
	GRF_efface(infos->exafs, VRAI, VRAI);
	return NULL;
      }
      memcpy(infos->mu_0->x, brut->x, sizeof(double) * infos->mu_0->nombre_point);
      infos->mu_0->x_min = brut->x_min;
      infos->mu_0->x_max = brut->x_max;
      recalcule = VRAI;
    }
  }

  /* Adaptation titres, couleurs,... */
  titre = TXT_genere_chaine("%s - "MU"0,III", brut->nom);
  GRF_titre(infos->exafs, titre);
  free(titre);
  if ( NULL != infos->mu_0_ref )
    titre = TXT_genere_chaine(MU"0,III + %s", infos->mu_0_ref->nom); /* Contiendra la somme à la fin... */
  else
    titre = TXT_genere_chaine(MU"0,III "o_CIRC"t"e_AIGU" "a_GRAVE" %s", brut->nom);
  GRF_titre(infos->mu_0, titre);
  free(titre);

  if ( brut->affichage != NULL ) {
    GRF_CopierAffichage( infos->exafs, brut, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_COULEUR|GRF_CP_AFF_STYLE );
  }

  /* On refait toute l'extraction, pour propager les erreurs */

  MU3_calc( brut, infos );

  /* Soustraction proprement dite -- s'occupe aussi des erreurs
     NULL : à remplacer par les termes de covariance... */
  retour = GRF_SoustraireGraphe(infos->exafs, infos->mu_0, VRAI, NULL);
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_SoustraireGraphe() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);

    GRF_efface(infos->mu_0, VRAI, VRAI);
    GRF_efface(infos->exafs, VRAI, VRAI);
    return NULL;
  }

  /* Somme des µ0 pour avoir le modèle total */
  GRF_AjouterGraphe(infos->mu_0, infos->mu_0_ref, VRAI, NULL);

  /* Log dans l'historique */
  HIST_Titre(1, "Troisi"e_GRAVE"me "e_AIGU"tape de l'extraction de l'EXAFS", "MU0_III");
  HIST_Texte(1, 0, "Signal brut   : %s\n", brut->nom);
  HIST_Texte(1, 0, "EXAFS extrait : %s\n", infos->exafs->nom);
  HIST_Texte(1, 0, "Mod"e_GRAVE"le de "MU"0  : %s\n", infos->mu_0->nom);
  HIST_Texte(2, 0, "M"e_AIGU"thode : Soustraction d'une spline contrainte\n");
  HIST_Texte(3, 3, "Param"e_GRAVE"tre de lissage : %g\n", infos->parametre);

  /* On a fini */
  return infos->exafs;
}

/* ~~~~~~~~~~~~~~~~~ Gestion de l'interface ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ----------------- Ouverture de la fenêtre liée à cette opération -----------------------\
 |   Renvoie le graphe contenant le spectre EXAFS traité                                   |
 |                                                                                         |
 | brut   : le graphe qui va subir la soustraction                                         |
 | mu_0   : le polynôme soustrait utilisé à l'étape précédente (NULL : on en crée un)      |
 | chaine : VRAI si on est dans une procédure chaînée                                      |
 \---------------------------------------------------------------------------------------- */

Graphe *MU3_ouvre(Graphe *brut, Graphe *mu_0, boolean chaine)
{
  Fenetre *fenetre;
  MU3_infos *infos = NULL;
  char *titre = NULL;

  if (brut == NULL) return NULL; /* Vraiment rien à faire */

  if ( arbres[ELIM_MU0_TROIS].drapeaux.ouvert == VRAI ) { /* Arbre déjà ouvert */
    Graphe *vieux; /* Ancien spectre travaillé dans ce module */

    fenetre = ARB_QuelleFenetre( &(arbres[ELIM_MU0_TROIS]), FAUX );
    vieux = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
    infos = (MU3_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

    if ( vieux != NULL ) { /* Sinon, il suffit de faire comme si de rien n'était... */
      int reponse;

      reponse = EXF_ERREUR(DEJA_ACTIF); /* Module déjà ouvert */
      if ( ( 1 == reponse ) || ( vieux == brut ) ) { /* On active seulement */
	FEN_premierplan(fenetre); /* Fenêtre au premier plan */
	return (infos->exafs); /* Rien d'autre à faire */
      } else if (reponse == 2) { /* On rappelle, pour le nouveau graphe */
	vieux->drapeaux.protege = FAUX; /* On a fini de travailler avec */

	/* On délie les infos pour être tranquille */
	FEN_GRF_LierGraphe( fenetre, NULL );
	FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

	/* On efface les anciens graphes inutiles */
	GRF_efface( infos->exafs, VRAI, VRAI );
	GRF_efface( infos->mu_0 , VRAI, VRAI );

	FEN_premierplan(fenetre); /* Fenêtre au premier plan */
      } else { /* On annule l'appel */
	return NULL;
      }
    }
  } else { /* Arbre non ouvert : on l'ouvre */
    fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, txt_titre_fenetre, txt_infos_fenetre,
				  ELIM_MU0_TROIS, BI_FEN_ARBRE_HAUT, prepare_arbre_mu3,
				  RDS_mu03, CLV_mu03,
				  ( int  (*)(Fenetre *, int, int, int, int, int) ) NULL, 
				  MSG_mu03,
				  ( void (*)(Fenetre *, int, int, int, int) ) NULL,
				  ( int  (*)(Fenetre *) ) NULL);
    if (fenetre == NULL) return NULL; /* Rien à faire... */

    fenetre->drapeaux.util_1 = (chaine != 0);
    if (chaine == CHAINE_COMPLETE)
      FEN_dimensions(fenetre, -500, -500, -500, -500); /* Un quart d'écran, bas droit */
    else if (chaine == CHAINE_PARTIELLE)
      FEN_dimensions(fenetre, 0, -500, -500, -500); /* Un quart d'écran, bas gauche */

    FEN_Ouvrir( fenetre );
  }

  if (infos == NULL) { /* Aucune info n'existe déjà */
    infos = (MU3_infos *) malloc( (size_t) sizeof(MU3_infos) );
    if (infos == NULL) {
      FEN_Detruire( fenetre );
      SYS_ERREUR(MANQUE_MEMOIRE, "etape_3.c", "MU3_ouvre", "infos");
      return NULL;
    }
  }

  infos->exafs = GRF_nouveau(brut, brut->nombre_point, GR_EXAFS3, GR_NORMAL, nom_grf_exafs);
  if (infos->exafs == NULL) {
    free( infos ); infos = NULL;
    FEN_Detruire( fenetre );
    return NULL;
  }

  if ( mu_0 != NULL ) { /* On a envoyé un polynôme */
    if ( mu_0->nombre_point != brut->nombre_point ) {
      EXF_ERREUR(NBRPTS_INCORRECT);
      mu_0 = NULL;
    }
  }			

  infos->mu_0_ref = mu_0; /* Le graphe mu_0 de référence */
  /* Si aucun µ0, la normalisation par un µ0 calculé est impossible ! */
  OBJ_active(ELIM_MU0_TROIS, EM3_MU0_CALC, (mu_0 != NULL) );

  titre = TXT_genere_chaine( nom_grf_mu03, brut->nom );
  infos->mu_0 = GRF_nouveau( NULL, brut->nombre_point, GR_MU0, GR_NORMAL, titre );
  free( titre ); titre = NULL;
  if (infos->mu_0 == NULL) {
    /* Les infos ne sont pas encore liées à la fenêtre : pas de risque de boucle */
    GRF_efface( infos->exafs, VRAI, VRAI);
    free( infos ); infos = NULL;
    FEN_Detruire( fenetre );
    return NULL;
  }

  infos->parametre = OBJ_LireNombre( ELIM_MU0_TROIS, ELM3_VALEUR_S, NULL );

  /* On associe les infos à la fenêtre */
  FEN_GRF_LierGraphe( fenetre, brut );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, infos );
  brut->drapeaux.protege = VRAI; /* Calcul en cours sur ce spectre -> rien de faisable */

  /* On prépare les spectres pour gagner du temps */
  memcpy( infos->mu_0->x , brut->x, (size_t) ( brut->nombre_point * sizeof(double) ) );
  memcpy( infos->exafs->x, brut->x, (size_t) ( brut->nombre_point * sizeof(double) ) );

  infos->mu_0->x_min = brut->x_min; infos->mu_0->x_max = brut->x_max;
  infos->exafs->x_min = brut->x_min; infos->exafs->x_max = brut->x_max;

  /* On calcule la fonction à soustraire */
  infos->exafs->drapeaux.protege = VRAI; /* On protège le graphe de toute bêtise */
  infos->mu_0->drapeaux.protege  = VRAI; /* On protège le graphe de toute bêtise */

/*   graf_mouse(S_ABEILLE, NULL); */ /* Souris en abeille  */
  MU3_calc(brut, infos);       /* Calcul du polynôme */
  MU3_extrait(brut, infos);    /* Extraction         */
/*   graf_mouse(S_FLECHE, NULL);  */ /* Souris normale     */

  fenetre->drapeaux.redessin    = VRAI;
  fenetre->msg_masque.fermeture = VRAI; /* Messages traités par MSG_mu01 */

  return (infos->exafs);
}

/* -------- Gestion du clavier pour cette fenêtre --------------------------------\
 |  Renvoie VRAI si traité, FAUX sinon                                            |
 |                                                                                |
 | espace = quitter le module                                                     |
 | return = recalculer à l'aide de la valeur indiquée                             |
 | chiffres = servent à entrer directement la valeur du paramètre de lissage      |
 \------------------------------------------------------------------------------- */

static int CLV_mu03(Fenetre *fenetre, unsigned int touche, int etat)
{
  MU3_infos *infos;
  Graphe *brut;
  Message message;

  if ( etat & K_CTRL ) return BI_EVENEMENT_NON_TRAITE; /* Pas de raccourci avec Ctrl ! */

  brut  = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = (MU3_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  switch (touche)
  {
  case 13:
    infos->parametre = OBJ_LireNombre( ELIM_MU0_TROIS, ELM3_VALEUR_S, NULL );
    MU3_calc   (brut, infos); /* On recalcule    */
    MU3_extrait(brut, infos); /* Et on soustrait */
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case ' ':
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas où != 0, fermeture forcée... */
    return (fenetre->message)(fenetre, &message);
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Redessin de la fenêtre d'élimination de mu_0, troisième étape --------\
 |   Le clipping est déjà installé correctement lors de l'appel                   |
 |                                                                                |
 | fenetre    : structure associée à la fenêtre considérée                        |
 | x, y       : coordonnées de la zone de redessin                                |
 | larg, haut : dimensions de la zone de redessin                                 |
 \------------------------------------------------------------------------------- */

static void RDS_mu03(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Graphe *graphe;
  InfosRegion2D region;
  MU3_infos *infos;
  int retour;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  graphe = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (MU3_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  if ( ( NULL == infos ) || ( NULL == graphe ) ) { /* Pas encore toutes les infos nécessaires */
    TRC_Message( fenetre, txt_brut_nul );
    return;
  }

  if ( ( graphe->x_max == graphe->x_min ) || ( graphe->y_max == graphe->y_min ) ) {
    TRC_Message( fenetre, txt_brut_nul );
    return;
  }

  /* On prépare la région 2D : aide pour les dessins après en fait */
  retour = TRC_Z2D_Initialiser( &region, fenetre, graphe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la région 2D [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* On trace le graphe dans la région 2D et on récupère les échelles au passage */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );

  /* On trace la courbe à soustraire - en bleu */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->mu_0, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );

  /* On trace le graphe EXAFS extrait - en fin, car échelle différente */
  if ( infos->exafs->y_max != infos->exafs->y_min ) {
    int y_axe;
    double echelle_Y;

    /* On adapte la région 2D */
    retour = TRC_Z2D_Initialiser( &region, NULL, infos->exafs );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la région 2D [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return;
    }

    /* On trace le signal extrait dans la région 2D et on récupère les échelles au passage */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->exafs, GRIS_FONCE, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, &echelle_Y );

    /* On trace l'axe des X associé, en violet */
    y_axe = TRC_ConvertitY_PvE( 0.0, infos->exafs->y_min, fenetre->w_haut, echelle_Y );
    TRC_Couleur( VIOLET );
    TRC_LigneH( fenetre, 0, y_axe, fenetre->w_larg );
  } else {
    TRC_Message( fenetre, txt_exafs_nul );
  }
}

/* -------------- Messages reçus par la fenêtre d'élimination de mu_0, 3e étape ---------- */

static int MSG_mu03(Fenetre *fenetre, Message *message)
{
  int retour;
  Graphe *brut, *exafs, *mu_0, *mu_ref __attribute__((__unused__));
  MU3_infos *infos;

  brut  = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = (MU3_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  if ( message->type == MSG_GRAPHE_MODIFIE ) { /* Le brut a été modifié, par conséquent, le µ0 aussi
						  On ne reçoit de message que pour le brut */
    Graphe *nv_brut = (Graphe *) (message->divers.pointeur);
    Graphe *extrait = infos->exafs;

    if (nv_brut->type == GR_MU0) { /* On change de µ0 - devrait jamais arriver ! En effet,
				      dans l'étape d'avant, on travaille à coup de realloc sur (x, y)
				      donc sans toucher au graphe directement

				      Si ça arrive, normalement le brut a déjà été changé ->
				      c'est juste pour mettre à jour la structure d'infos */
#ifdef _DEVELOPPE_
      fprintf(stderr, "MSG_mu03 : µ0 de référence modifié, comportement anormal !\n");
#endif
      if (nv_brut->nombre_point != brut->nombre_point) {
	fprintf(stderr, "MSG_mu03 : µ0 modifié, en désaccord avec l'exafs extrait !\n");
	fprintf(stderr, "           comportement ultérieur du programme aléatoire...\n");
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      }
      infos->mu_0_ref = nv_brut;
      return BI_EVENEMENT_TRAITE;
    }

    if ( message->divers.drapeau == VRAI ) { /* Pas autant de point -> on joue sur realloc */
      if ( (GRF_AdapteTaille(extrait, nv_brut->nombre_point) == FAUX ) ||
	   (GRF_AdapteTaille(infos->mu_0, nv_brut->nombre_point) == FAUX ) ) {
	message->type = WM_CLOSED;
	message->fenetre.haut = -1; /* On force la fermeture de cette fenêtre */
	return MSG_mu03( fenetre, message );
      }
    }

    /* Note : les X peuvent avoir changé, si l'on a joué sur E0
       Or, la suite du calcul ne modifie _pas_ les X -> le faire ici */

    memcpy(extrait->x, nv_brut->x, sizeof(double) * nv_brut->nombre_point);
    memcpy(infos->mu_0->x, nv_brut->x, sizeof(double) * nv_brut->nombre_point);

    /* Ici, on a fait ce qu'il fallait pour s'adapter -> on refait les calculs */
    FEN_GRF_LierGraphe( fenetre, nv_brut );
    brut = nv_brut;
/*     graf_mouse(S_ABEILLE, NULL); */
    GRF_modifier(infos->exafs, infos->exafs, message->divers.drapeau); /* On informe _d'abord_         
									  pour informer du changement  
									  éventuel du nombre de points */
    MU3_calc   (brut, infos); /* On recalcule  */
    MU3_extrait(brut, infos); /* Et on extrait - se charger de faire suivre la modif. */
/*     graf_mouse(S_FLECHE, NULL); */
    FEN_redessin(fenetre, BI_RDS_UTIL); /* Puisqu'il y a eu modif */

    return BI_EVENEMENT_TRAITE;
  }

  if ( message->type == WM_CLOSED ) { /* On ferme la fenêtre */
    boolean erreur_norm = FAUX;

    if ( message->fenetre.haut == -1) { /* == -1 : on force la fermeture, type "annuler" */
      retour = 2;
    } else {
      retour = EXF_ERREUR(FIN_MODULE);
      if ( 3 == retour ) { /* On annule la fermeture */
	return BI_EVENEMENT_TRAITE;
      }
    }

    /* De toute façon, on ferme */
    exafs  = infos->exafs;
    mu_0   = infos->mu_0;
    mu_ref = infos->mu_0_ref;

    /* On délie donc les informations */
    FEN_GRF_LierGraphe( fenetre, NULL );
    FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

    /* On peut maintenant fermer la fenêtre sans risque */
    FEN_Detruire( fenetre );

    if ( infos->mu_0_ref != NULL ) { /* Existe bien... */
      infos->mu_0_ref->drapeaux.protege = FAUX;
    }
    if ( brut != NULL ) { /* NULL si fermeture forcée pour cause d'effacement de brut */
      brut->drapeaux.protege = FAUX;
    }

    if (retour == 2) { /* On ferme sans soustraire */
      GRF_efface( exafs, VRAI, VRAI );
      GRF_efface( mu_0,  VRAI, VRAI );
      free( infos ); infos = NULL;
      return BI_EVENEMENT_FEN_FERMEE;
    }

    exafs = MU3_ExtraitEXAFS(brut, infos, FAUX);
    if ( NULL == exafs ) {
      GRF_efface( mu_0,  VRAI, VRAI );
      free( infos ); infos = NULL;
      return BI_EVENEMENT_FEN_FERMEE;
    }

    /* ------ Que faire comme normalisation ? ------ */
    if ( OBJ_choisit(ELIM_MU0_TROIS, EM3_MU0_CALC, QUESTION) == VRAI) {
      if ( NULL == infos->mu_0_ref ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "%s,%s l.%u : on veut diviser par µ0 total qui n'existe pas !\n",
		       __FILE__, __FUNCTION__, __LINE__);
#endif
	erreur_norm = VRAI;
      } else { /* Mu_ref existe */
	int retour;
	/* infos->mu_0 contient alors le µ0 total, calculé dans MU3_ExtraitEXAFS */

	retour = EXF_NormaliseExp( infos->exafs, infos->mu_0, VRAI );
	if ( retour < 0 ) { /* Problème survenu */
	  /* On passe en normalisation théorique */
	  OPTION_A_FAIRE( "Message indiquant que cela a raté et que l'on passe en normalisation théorique" )
	  erreur_norm = VRAI;
	} else if ( retour > 0 ) {
	  OPTION_A_FAIRE( "Message indiquant que cela a raté et que l'on passe en normalisation théorique" )
	}
      /* Il faudrait diviser mu_ref par k^n, n affichage, pour bien faire
	 EXF_k_p_n(mu_ref, -n_affichage); */
      }
    }

    /* On ne met pas de else, parce que si la normalisation par µ0 a échoué, on passe en
       normalisation théorique... */
    if ( ( OBJ_choisit(ELIM_MU0_TROIS, EM3_THEORIQUE, QUESTION) == VRAI ) ||
	 ( VRAI == erreur_norm ) ) {
      /* On propose une normalisation théorique */
      Graphe *norm = I_XAS_NormaliserEXAFS( exafs );

      if (norm)
	exafs = norm;
      else
	EXF_ERREUR(NORM_IMPOSSIBLE);
    }

    exafs->drapeaux.protege = FAUX; /* On n'utilise plus ces graphes... */
    mu_0->drapeaux.protege  = FAUX;
    exafs->drapeaux.sauve   = FAUX; /* Graphe non sauvé */
    mu_0->drapeaux.sauve    = FAUX;

    GRF_aller(exafs); /* On va à ce graphe */

    OBJ_MettreNombre( EXAFS_GENERAL, EXF_VAL_S, infos->parametre, 0.0 );
    free( infos ); infos = NULL;

    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non traité [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Gestion de l'arbre qui est dans la fenêtre ----------------------------------------- */

static int GER_mu03(unsigned int num_objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  MU3_infos *infos;
  Graphe *brut;
  Fenetre *fenetre;

  fenetre = ARB_QuelleFenetre(arbres + ELIM_MU0_TROIS, FAUX); /* Fenêtre actuelle */
  brut  = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = (MU3_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  switch (num_objet)
  {
  case CHX_MU0_DEUX: /* Choix du mu0 de référence */
    {
      Graphe *nv_mu0 = GRF_choisir(GR_MU0, VRAI, VRAI, NULL);
      if (nv_mu0 == NULL) return BI_EVENEMENT_TRAITE;

      if (nv_mu0->nombre_point != brut->nombre_point) {
	(void) EXF_ERREUR(NBRPTS_INCORRECT);
	return BI_EVENEMENT_TRAITE;
      }
      if (nv_mu0->drapeaux.protege == VRAI) {
	(void) EXF_ERREUR(GRF_PROTEGE);
	return BI_EVENEMENT_TRAITE;
      }

      infos->mu_0_ref->drapeaux.protege = FAUX;
      infos->mu_0_ref = nv_mu0;
      nv_mu0->drapeaux.protege = VRAI;
      OBJ_active(ELIM_MU0_TROIS, EM3_MU0_CALC, FORCE_OUI); /* Elle est maintenant possible */
    }
    return BI_EVENEMENT_TRAITE;

  case MU3_VALIDER: /* On enregistre la valeur de s donnée par l'utilisateur */
    infos->parametre = OBJ_LireNombre( ELIM_MU0_TROIS, ELM3_VALEUR_S, NULL );
    MU3_calc   (brut, infos); /* On recalcule    */
    MU3_extrait(brut, infos); /* Et on soustrait */
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Préparation de l'arbre dans la fenêtre -------- */

static int prepare_arbre_mu3(unsigned int numero_arbre __attribute__((__unused__)))
{
  if ( FAUX == arbres[ ELIM_MU0_TROIS ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( ELIM_MU0_TROIS, "menu_exafs.html#etape_3", GER_mu03, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
  }
  OBJ_MettreNombre( ELIM_MU0_TROIS, ELM3_VALEUR_S, 100.0, 0.0 ); /* Valeur initiale pour le paramètre */

  OBJ_agrandit(ELIM_MU0_TROIS, ROOT, 0, +3);
  OBJ_cache(ELIM_MU0_TROIS, EM3_ATTENTION, FORCE_OUI); /* Message "Attention, µ0 négatif */
  OBJ_deplace(ELIM_MU0_TROIS, EM3_AUCUNE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(ELIM_MU0_TROIS, EM3_THEORIQUE, +1, +1, DEPLACE_PIXELS);
  OBJ_deplace(ELIM_MU0_TROIS, EM3_MU0_CALC, +2, +1, DEPLACE_PIXELS);

#ifdef AVEC_TRADUCTION_ /* Traduction */
  OBJ_TraduireTexte(ELIM_MU0_TROIS, CHX_MU0_DEUX, txt_choisir);
  OBJ_TraduireTexte(ELIM_MU0_TROIS, EM3_TXT_NORM, txt_normalisation);
  OBJ_TraduireTexte(ELIM_MU0_TROIS, MU3_VALIDER, txt_valider);
  OBJ_TraduireTexte(ELIM_MU0_TROIS, EM3_AUCUNE, txt_aucune);
  OBJ_TraduireTexte(ELIM_MU0_TROIS, EM3_THEORIQUE, txt_theorique);
  OBJ_TraduireTexte(ELIM_MU0_TROIS, EM3_MU0_CALC, txt_extrait);
  OBJ_TraduireTexte(ELIM_MU0_TROIS, MU03_TXT_PARAM, txt_parametre);
#endif

  arbres[ELIM_MU0_TROIS].drapeaux.prepare = VRAI;
  return 0;
}
