/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 1998, Emmanuel CURIS
   Cration de ce fichier : 22 septembre 1998

   Extraction du signal EXAFS proprement dit

   Normalisation du signal EXAFS
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"
#include "chimie_utils.h"

#include "absorption.h" /* Les seuils */
#include "extrait.h"
#include "graphes.h"
#include "interface.h"
#include "lissage.h"
#include "science.h"

#include "norm_exafs.txt"

/* ~~~~~~~~ Variables locales  ce fichier ~~~~~~~~ */

static const double k_v_E = H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM
                          / (2 * MASSE_ELECTRON * EV_V_JOULE);

/* ~~~~~~~~ Fonctions locales  ce fichier ~~~~~~~~ */

static int EXF_NormCoherent(Graphe *exafs, Graphe *mu0);
static int EXF_NormDivision(Graphe *exafs, Graphe *mu0);

static int EXF_NormThConstante(Graphe *exafs, double constante, double h_seuil);
static int EXF_NormThEbel     (Graphe *exafs, double e0, double h_seuil, double c[ 6 ]);
static int EXF_NormThHeitler  (Graphe *exafs, double e0, double h_seuil);
static int EXF_NormThLengeler (Graphe *exafs, double e0, double h_seuil);
static int EXF_NormThVictoreen(Graphe *exafs, double e0, double h_seuil, double C_av, double D_av, double C_ap, double D_ap);
static int EXF_NormThMcMaster (Graphe *exafs, int Z, double e0, double h_seuil, double A0_av, double A1_av, double A2_av, double A3_av,
			       double A0_ap, double A1_ap, double A2_ap, double A3_ap);

/* ~~~~~~~~ Ralisation pratique du module ~~~~~~~~ */

/* -~-~-~-~-~-~-~-~ Normalisation en divisant par un graphe contenant 0 -~-~-~-~-~-~-~-~ */

/* -------- Normalisation en divisant par 0, cas des X cohrents -------- */

static int EXF_NormCoherent(Graphe *exafs, Graphe *mu0)
{
  Graphe *erreurs;
  unsigned long i;
  double *y_exf, *y_mu0, *y_err = NULL;
  int retour = PAS_DE_PROBLEME;

  if ( ( NULL == exafs ) || ( NULL == mu0 ) ) return NORMEXAFS_NULL;
  if ( mu0->nombre_point != exafs->nombre_point ) /* La grille ne peut pas tre cohrente ! */
    return NORMEXAFS_X_INCOHERENTS;

  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  for ( i = 0, y_exf = exafs->y, y_mu0 = mu0->y ; i < exafs->nombre_point ; i++, y_exf++, y_mu0++ ) {
    if ( *y_mu0 == 0.0 ) {
      retour = NORMEXAFS_DIV_ZERO;
    } else if ( *y_mu0 <= 1e-4 ) { /* Valeur trs faible : risque d'erreurs d'arrondi */
      if ( retour != NORMEXAFS_DIV_ZERO ) retour = NORMEXAFS_ARRONDIS;
      *y_exf /= *y_mu0;
      if ( y_err != NULL ) {
	*y_err /= *y_mu0; /* On a vrifi que 0 > 0 en dbut de fonction */
	y_err++;
      }
    } else { /* Cas normal */
      *y_exf /= *y_mu0;
      if ( y_err != NULL ) {
	*y_err /= *y_mu0; /* On a vrifi que 0 > 0 en dbut de fonction */
	y_err++;
      }
    }
  }

  /* Traiter le cas des erreurs en Y sur 0 ? */
  if ( mu0->erreur_y != NULL ) {
    (void) fprintf( stderr, "%s,%s l.%u : ATTENTION, les erreurs en Y sur 0 ont t ngliges !\n",
		    __FILE__, __FUNCTION__, __LINE__);
    if ( PAS_DE_PROBLEME == retour ) {
      /* Mineur par rapport aux autres messages... */
      retour = NORMEXAFS_ERREURMU0_NEGLIGEES;
    }
  }

  /* Log dans l'historique */
  HIST_Texte(4, 0, "Les X sont suppos"e_AIGU"s identiques dans le graphe normalis"e_AIGU" et dans "MU"0\n");

  return retour;
}

/* -------- Normalisation en divisant par 0, cas gnral -------- */

static int EXF_NormDivision(Graphe *exafs, Graphe *mu0)
{
  Graphe *erreurs;
  double *b, *c, *d; /* Coefficients pour la spline */
  unsigned long i;
  double *x_exafs, *y_exafs, *y_err = NULL;
  int retour;

  if ( ( NULL == exafs ) || ( NULL == mu0 ) ) return NORMEXAFS_NULL;

  if ( ( mu0->x_min > exafs->x_max ) || ( mu0->x_max < exafs->x_min ) ) {
    /* ERREUR : Graphes disjoints */
    (void) fprintf( stderr, "%s,%s l.%u : graphes disjoints\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return NORMEXAFS_X_INCOHERENTS;
  }

  retour = SPL_tt_prep( mu0, &b, &c, &d );
  if ( retour  < 0 ) { /* Problme en prparant les splines */
    (void) fprintf( stderr, "%s,%s l.%u : impossible de faire la spline [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    (void) LIS_ERREUR( retour, mu0->nom );
    return NORMEXAFS_ERR_INTERPOLATION;
  }

  retour =  PAS_DE_PROBLEME;

  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  /* Normalisation proprement dite */
  for ( i = 0, x_exafs = exafs->x, y_exafs = exafs->y ; i < exafs->nombre_point ; i++, x_exafs++, y_exafs++ ) {
    double val_mu0;

    val_mu0 = SPL_tt_calc(mu0, *x_exafs, b, c, d);
    if ( val_mu0 == 0.0 )
      retour = NORMEXAFS_DIV_ZERO;
    else if ( val_mu0 <= 1e-4 ) { /* Valeur trs faible : risque d'erreurs d'arrondi, ou ngative : absurde */
      if ( retour != NORMEXAFS_DIV_ZERO ) retour = NORMEXAFS_ARRONDIS;
      *y_exafs /= val_mu0;
      if ( y_err != NULL ) {
	*y_err /= fabs(val_mu0); /* On a vrifi que 0 > 0 en dbut de fonction */
	y_err++;
      }
    } else { /* Cas normal */
      *y_exafs /= val_mu0;
      if ( y_err != NULL ) {
	*y_err /= fabs(val_mu0); /* On a vrifi que 0 > 0 en dbut de fonction */
	y_err++;
      }
    }
  }

  /* Traiter le cas des erreurs en Y sur 0 ? */
  /* Traiter le cas des erreurs en Y sur 0 ? */
  if ( mu0->erreur_y != NULL ) {
    (void) fprintf( stderr, "%s,%s l.%u : ATTENTION, les erreurs en Y sur 0 ont t ngliges !\n",
		    __FILE__, __FUNCTION__, __LINE__);
    if ( PAS_DE_PROBLEME == retour ) {
      /* Mineur par rapport aux autres messages... */
      retour = NORMEXAFS_ERREURMU0_NEGLIGEES;
    }
  }

  free( b ); free( c ); free( d ); /* Plus besoin des splines */
  b = NULL; c = NULL; d = NULL;

  /* Log dans l'historique */
  HIST_Texte(4, 0, "Interpolation de "MU"0 par spline\n");

  return retour;
}

/* -------- Normalisation en divisant par un graphe qui contient le 0 ---\
 |   Renvoie un code (0 : pas de problme ; autre en cas d'erreur)        |
 |     (normalisation EN PLACE)                                           |
 |                                                                        |
 | exafs = le signal  normaliser                                         |
 | mu0   = le graphe contenant le 0 utilis pour normaliser              |
 | coherent : VRAI si les deux graphes ont la mme grille en X            |
 \----------------------------------------------------------------------- */

int EXF_NormaliseExp(Graphe *exafs, Graphe *mu0, boolean coherent)
{
  int retour;

  if ( ( NULL == exafs ) || ( NULL == mu0 ) ) return NORMEXAFS_NULL;

  /* Pas de vrification de valeurs ngatives : rle de l'interface */
#ifdef ancien
  if ( mu0->y_min < 0.0 ) {
    /* ERREUR : normalisation par des valeurs ngatives, douteux... */
    if ( BI_ERR_alerte(NULL, &ERR_MU0_Negatif, VRAI) == 2) /* On annule le calcul */
      return NULL;
  }
#endif

  /* Log dans l'historique */
  HIST_Titre(1, "Normaliser les oscillations EXAFS", "NORM");
  HIST_Texte(2, 0, "Division par un graphe contenant un mod"e_GRAVE"le de "MU"0\n");
  HIST_Texte(1, 0, "Oscillations normalis"e_AIGU"es : %s\n", exafs->nom);
  HIST_Texte(1, 0, "Mod"e_GRAVE"le de "MU"0 : %s\n", mu0->nom);

  if ( VRAI == coherent ) {
    retour = EXF_NormCoherent( exafs, mu0 );
  } else {
    retour = EXF_NormDivision( exafs, mu0 );
  }
  if ( retour < 0 ) {
    /* Code grave : aucun calcul n'a t fait... */
    return retour;
  }

  /* Mises  jour internes */
  GRF_ChangerType( exafs, GR_EXAFS ); /* Le spectre est maintenant normalis */
  GRF_bornes( exafs, NE_PAS_FAIRE, DESORDONNE );
  GRF_bornes( exafs->erreur_y, NE_PAS_FAIRE, DESORDONNE );
  exafs->drapeaux.normalise = VRAI;
  exafs->drapeaux.sauve = FAUX;

  GRF_modifier( exafs, exafs, FAUX ); /* Nbr de points n'a pas chang, mais spectre si ! */

  return retour;
}

/* -~-~-~-~-~-~-~-~ Normalisation en divisant par un modle thorique de 0 -~-~-~-~-~-~-~-~ */

/* -------- Normalisation en divisant par une constante -------- */

static int EXF_NormThConstante(Graphe *exafs, double constante, double h_seuil)
{
  double facteur;

  if ( NULL == exafs ) return NORMEXAFS_NULL;

  facteur = constante * h_seuil;
  if ( 0.0 == facteur ) {
    return NORMEXAFS_FACTEUR_INCORRECT;
  }
  facteur = 1.0 / facteur;

  /* On divise */
  GRF_FacteurMultiplicatif( exafs, facteur, AXE_Y, NULL );

  return PAS_DE_PROBLEME;
}

/* -------- Modle de Ebel -------- */

static int EXF_NormThEbel(Graphe *exafs, double e0, double h_seuil, double c[ 6 ])
{
  int retour = PAS_DE_PROBLEME, type_unite;
  double *x, *y, *y_err = NULL;
  double facteur;
  unsigned long i;
  Graphe *erreurs;

  if ( NULL == exafs ) return NORMEXAFS_NULL;
  if ( 0.0 == h_seuil ) {
    (void) fprintf( stderr, "%s,%s l.%u : Hauteur de seuil nulle !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return NORMEXAFS_FACTEUR_INCORRECT;
  }
  facteur = TH_Ebel( e0, 0 /* Z est inutilis */, c );
  facteur /= h_seuil;

  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  x = exafs->x; y = exafs->y;

  /* Dtection du type d'unit en X */
  (void) TXT_NormaliserUnite( GRF_QuelleUnite( exafs, AXE_X ), FAUX, &type_unite, NULL );

  if ( UNITE_ENERGIE == type_unite ) {
    if ( exafs->x_min < e0 ) {
      /* Alerte : le graphe commence _avant_ e0 */
      (void) fprintf( stderr, "%s,%s l.%u : Des valeurs sont avant E0 !\n",
		      __FILE__, __FUNCTION__, __LINE__);
      retour = NORMEXAFS_K_NEGATIF;
    }

    for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
      double mu_0;

      mu_0  = TH_Ebel( *x, 0 /* Z est inutilis */, c );
      mu_0 /= facteur;

      if ( mu_0 != 0.0 ) {
	*y /= mu_0;
	if ( y_err != NULL ) {
	  *y_err /= fabs( mu_0 );
	  y_err++;
	}
      } else {
	retour = NORMEXAFS_DIV_ZERO;
	*y = 0.0;
	if ( y_err != NULL ) {
	  *y_err = 0.0; /* Puisque l'on impose 0, il n'y a plus d'erreur en ce point ! */
	  y_err++;
	}
      }
    }

    return retour;
  }

  /* Autre type d'unit en X */
  if ( type_unite != UNITE_NOMBRE_ONDE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Unit en X incorrecte [type %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_unite);
    return NORMEXAFS_UNITE_INCORRECTE;
  }

  /* Par dfaut, X contiennent des nombres d'onde */
  for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
    double nrj, mu_0;

    nrj = e0 + (*x * *x) * k_v_E; /* Valeur de l'nergie  partir de k */
    mu_0  = TH_Ebel( nrj, 0 /* Z est inutilis */, c );
    mu_0 /= facteur;
    
    if ( mu_0 != 0.0 ) {
      *y /= mu_0;
      if (y_err != NULL) {
	*y_err /= fabs(mu_0);
	y_err++;
      }
    } else {
      retour = NORMEXAFS_DIV_ZERO;
      *y = 0.0;
      if ( y_err != NULL ) {
	*y_err = 0.0; /* Puisque l'on impose 0, il n'y a plus d'erreur en ce point ! */
	y_err++;
      }
    }
  }

  return retour;
}

/* -------- Modle de Heitler -------- */

static int EXF_NormThHeitler(Graphe *exafs, double e0, double h_seuil)
{
  double *x, *y, *y_err = NULL;
  int retour = PAS_DE_PROBLEME, type_unite;
  unsigned long i;
  Graphe *erreurs;

  if ( NULL == exafs ) return NORMEXAFS_NULL;
  if ( 0.0 == e0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : nergie de seuil nulle\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return NORMEXAFS_FACTEUR_INCORRECT;
  }

  x = exafs->x; y = exafs->y;

  /* Adaptation des erreurs... */
  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  /* Dtection du type d'unit en X */
  (void) TXT_NormaliserUnite( GRF_QuelleUnite( exafs, AXE_X ), FAUX, &type_unite, NULL );

  if ( UNITE_ENERGIE == type_unite ) {
    if ( exafs->x_min < e0) {
      (void) fprintf(stderr, "EXF_NormThHeitler : on commence avant le seuil ! [min : %g, E0 : %g]", exafs->x_min, e0);
      return NORMEXAFS_K_NEGATIF;
    }

    for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
      double mu0, x_Heitler, racine, ln_mu0;

      x_Heitler = ( *x - e0 ) / e0;
      racine = sqrt( x_Heitler );

      ln_mu0 = 1.0 - atan( racine ) - log( 1.0 + x_Heitler );
      ln_mu0 *= 4.0;
      ln_mu0 -= log( 1.0 - exp( - 2.0 * M_PI / racine ) );

      mu0 = h_seuil * exp( ln_mu0 );
      if ( mu0 == 0.0 ) {
	retour = NORMEXAFS_DIV_ZERO;
	*y = 0.0; /* Normalisation impossible sur ce point... arbitrairement 0 */
	if (y_err != NULL) {
	  *y_err = 0.0;
	  y_err++;
	}
      } else {
	*y /= mu0;
	if (y_err != NULL) {
	  *y_err /= fabs(mu0);
	  y_err++;
	}
      }
    }

    return retour;
  }

  /* Autre type d'unit en X */
  if ( type_unite != UNITE_NOMBRE_ONDE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Unit en X incorrecte [type %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_unite);
    return NORMEXAFS_UNITE_INCORRECTE;
  }

  /* Par dfaut, contient des nombres d'onde */
  for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
    double mu0, x_Heitler, racine, ln_mu0;

    x_Heitler = *x * *x * k_v_E / e0;
    racine = sqrt( x_Heitler );
    if ( 0.0 == racine ) {
      ln_mu0 = 1.0; /* Extension par continuit */
    } else {
      ln_mu0 = /*1.0*/ - atan( racine ) - log( 1.0 + x_Heitler );
      ln_mu0 *= 4.0;
      ln_mu0 -= log( 1.0 - exp( - 2.0 * M_PI / racine ) );
    }

    mu0 = h_seuil * exp( ln_mu0 );
    if (mu0 == 0.0) {
      retour = NORMEXAFS_DIV_ZERO;
      *y = 0.0; /* Normalisation impossible sur ce point... arbitrairement 0 */
      if ( y_err != NULL ) {
	*y_err = 0.0;
	y_err++;
      }
    } else {
      *y /= mu0;
      if (y_err != NULL) {
	*y_err /= fabs(mu0);
	y_err++;
      }
    }
  }

  return retour;
}

/* -------- Modle de Lengeler-Eisenberger -------- */

static int EXF_NormThLengeler(Graphe *exafs, double e0, double h_seuil)
{
  int retour = PAS_DE_PROBLEME, type_unite;
  unsigned long i;
  double coef;
  double *x, *y, *y_err = NULL;
  Graphe *erreurs;

  if ( NULL == exafs ) return NORMEXAFS_NULL;

  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  x = exafs->x; y = exafs->y;

  /* Dtection du type d'unit en X */
  (void) TXT_NormaliserUnite( GRF_QuelleUnite( exafs, AXE_X ), FAUX, &type_unite, NULL );

  if ( UNITE_ENERGIE == type_unite ) { /* X contiennent des nergies */
    if ( exafs->x_min < e0 ) {
      retour = NORMEXAFS_K_NEGATIF;
    }

    coef = 8.0 / (3.0 * e0);

    for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
      double mu0;

      mu0  = 11.0/3.0 - coef * *x;
      mu0 *= h_seuil;
      if ( mu0 == 0.0 ) {
	retour = NORMEXAFS_DIV_ZERO;
	*y = 0.0; /* Normalisation impossible sur ce point... arbitrairement 0 */
	if ( y_err != NULL ) {
	  *y_err = 0.0 ;
	  y_err++;
	}
      } else {
	*y /= mu0;
	if ( y_err != NULL ) {
	  *y_err /= fabs( mu0 );
	  y_err++;
	}
      }
    }

    return retour;
  }

  /* Autre type d'unit en X */
  if ( type_unite != UNITE_NOMBRE_ONDE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Unit en X incorrecte [type %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_unite);
    return NORMEXAFS_UNITE_INCORRECTE;
  }

  /* Par dfaut, X contiennent des nombres d'onde */
  coef = 8.0 / (3.0 * e0) * k_v_E;
  
  for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
    double mu0;

    mu0  = 1.0 - coef * (*x * *x); /* Pourquoi 1.0 et pas 11/3 ? Vrifier ce que deviennent les coefficients dans la conversion E <-> k */
    mu0 *= h_seuil;
    if ( mu0 == 0.0 ) {
      retour = NORMEXAFS_DIV_ZERO;
      *y = 0.0; /* Normalisation impossible sur ce point... arbitrairement 0 */
      if ( y_err != NULL ) {
	*y_err = 0.0;
	y_err++;
      }
    } else {
      *y /= mu0;
      if ( y_err != NULL ) {
	*y_err /= fabs( mu0 );
	y_err++;
      }
    }
  }

  return retour;
}

/* -------- Modle de McMaster -------- */

static int EXF_NormThMcMaster(Graphe *exafs, int Z, double e0, double h_seuil, double A0_av, double A1_av, double A2_av, double A3_av,
			      double A0_ap, double A1_ap, double A2_ap, double A3_ap)
{
  int retour = PAS_DE_PROBLEME, type_unite;
  double *x, *y, *y_err = NULL;
  double facteur;
  unsigned long i;
  Graphe *erreurs;

  if ( NULL == exafs ) return NORMEXAFS_NULL;

  facteur  = TH_McMaster(e0, Z, A0_ap, A1_ap, A2_ap, A3_ap);
  facteur -= TH_McMaster(e0, Z, A0_av, A1_av, A2_av, A3_av);
  if ( ( 0.0 == facteur ) || ( 0.0 == h_seuil ) )
    return NORMEXAFS_FACTEUR_INCORRECT;
  facteur /= h_seuil;

  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  x = exafs->x; y = exafs->y;

  /* Dtection du type d'unit en X */
  (void) TXT_NormaliserUnite( GRF_QuelleUnite( exafs, AXE_X ), FAUX, &type_unite, NULL );

  if ( UNITE_ENERGIE == type_unite ) {
    if ( exafs->x_min < e0 ) {
      /* Alerte : le graphe commence _avant_ e0 */
      retour = NORMEXAFS_K_NEGATIF;
    }

    for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
      double mu_0;

      mu_0  = TH_McMaster(*x, Z, A0_ap, A1_ap, A2_ap, A3_ap);
      mu_0 -= TH_McMaster(*x, Z, A0_av, A1_av, A2_av, A3_av);
      mu_0 /= facteur;

      if (mu_0 != 0.0) {
	*y /= mu_0;
	if (y_err != NULL) {
	  *y_err /= fabs(mu_0);
	  y_err++;
	}
      } else {
	retour = NORMEXAFS_DIV_ZERO;
	*y = 0.0;
	if ( y_err != NULL ) {
	  *y_err = 0.0 ;
	  y_err++;
	}
      }
    }

    return retour;
  }

  /* Autre type d'unit en X */
  if ( type_unite != UNITE_NOMBRE_ONDE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Unit en X incorrecte [type %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_unite);
    return NORMEXAFS_UNITE_INCORRECTE;
  }

  /* Par dfaut, X contiennent des nombres d'onde */
  for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
    double nrj, mu_0;

    nrj = e0 + (*x * *x) * k_v_E; /* Valeur de l'nergie  partir de k */
    mu_0  = TH_McMaster(nrj, Z, A0_ap, A1_ap, A2_ap, A3_ap);
    mu_0 -= TH_McMaster(nrj, Z, A0_av, A1_av, A2_av, A3_av);
    mu_0 /= facteur;
    
    if ( mu_0 != 0.0 ) {
      *y /= mu_0;
      if (y_err != NULL) {
	*y_err /= fabs( mu_0 );
	y_err++;
      }
    } else {
      retour = NORMEXAFS_DIV_ZERO;
      *y = 0.0;
      if ( y_err != NULL ) {
	*y_err /= 0.0;
	y_err++;
      }
    }
  }

  return retour;
}

/* -------- Modle de Victoreen -------- */

static int EXF_NormThVictoreen(Graphe *exafs, double e0, double h_seuil, double C_av, double D_av, double C_ap, double D_ap)
{
  int retour = PAS_DE_PROBLEME, type_unite;
  double *x, *y, *y_err = NULL;
  double delta_c, delta_d, facteur;
  unsigned long i;
  Graphe *erreurs;

  if ( NULL == exafs ) return NORMEXAFS_NULL;

  /* Delta_x = x(aprs le seuil) - x(avant le seuil)
     Formule de Victoreen polynomiale => la diffrence correspond  la diffrence des coefficients */
  delta_c = C_ap - C_av;
  delta_d = D_ap - D_av;

  facteur = TH_Victoreen( e0, -1, delta_c, delta_d ); /* Z ne sert pas de toute faon... */
  if ( ( 0.0 == facteur ) || ( 0.0 == h_seuil ) )
    return NORMEXAFS_FACTEUR_INCORRECT;
  facteur /= h_seuil;

  erreurs = exafs->erreur_y;
  if ( NULL != erreurs )
    y_err = erreurs->y; /* Erreurs en Y sur l'exafs */

  x = exafs->x; y = exafs->y;

  /* Dtection du type d'unit en X */
  (void) TXT_NormaliserUnite( GRF_QuelleUnite( exafs, AXE_X ), FAUX, &type_unite, NULL );

  if ( UNITE_ENERGIE == type_unite ) {
    if ( exafs->x_min < e0 ) {
      /* Alerte : le graphe commence _avant_ e0 */
      retour = NORMEXAFS_K_NEGATIF;
    }

    for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
      double mu_0;

      mu_0  = TH_Victoreen( *x, -1, delta_c, delta_d ); /* Z inconnu, mais ne sert pas alors... */
      mu_0 /= facteur;

      if ( mu_0 != 0.0 ) {
	*y /= mu_0;
	if ( y_err != NULL ) {
	  *y_err /= fabs(mu_0);
	  y_err++;
	}
      } else {
	retour = NORMEXAFS_DIV_ZERO;
	*y = 0.0;
	if ( y_err != NULL ) {
	  *y_err = 0.0;
	  y_err++;
	}
      }
    }

    return retour;
  }

  /* Autre type d'unit en X */
  if ( type_unite != UNITE_NOMBRE_ONDE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Unit en X incorrecte [type %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_unite);
    return NORMEXAFS_UNITE_INCORRECTE;
  }

  /* Par dfaut, on suppose que X contient des nombres d'onde */

  for ( i = 0 ; i < exafs->nombre_point ; i++, x++, y++ ) {
    double nrj, mu_0;

    nrj = e0 + (*x * *x) * k_v_E; /* Valeur de l'nergie  partir de k */
    mu_0  = TH_Victoreen( nrj, -1, delta_c, delta_d ); /* Z inconnu, mais ne sert pas alors... */
    mu_0 /= facteur;

    if ( mu_0 != 0.0 ) {
      *y /= mu_0;
      if (y_err != NULL) {
	*y_err /= fabs(mu_0);
	y_err++;
      }
    } else {
      retour = NORMEXAFS_DIV_ZERO;
      *y = 0.0;
      if ( y_err != NULL ) {
	*y_err = 0.0;
	y_err++;
      }
    }
  }

  return retour;
}

/* -------- Normalisation / protocole d'appel gnral -----------------------------------------\
 |   Renvoie un code (0 : pas de problme ; autre en cas d'erreur)        |
 |     (normalisation EN PLACE)                                           |
 |                                                                                             |
 | exafs   : le graphe  normaliser                                                            |
 | methode : quel modle thorique  utiliser (cf. Include/absorption.h, MODELE_MU0_...)       |
 | e0      : nergie de seuil                                                                  |
 | seuil   : seuil tudi (K, L,... - cf Include/theorie.h)                                    |
 | Z       : lment tudi                                                                    |
 | h_seuil : hauteur de seuil exprimentale                                                    |
 | avant   : coefficients  utiliser avant seuil (le nombre dpend du modle, PAS DE CONTRLE) |
 | aprs   : coefficients  utiliser aprs seuil                                               |
 \-------------------------------------------------------------------------------------------- */

int EXF_NormaliseTh(Graphe *exafs, unsigned int methode, double e0, int seuil, int Z, double h_seuil,
		    double *avant, double *apres)
{
  int retour = NORMEXAFS_METHODE_INCONNUE;
  char *titre;

  if ( NULL == exafs ) return NORMEXAFS_NULL;

  if ( ( h_seuil <= 0.0 ) || ( e0 <= 0.0 ) ) {
    return NORMEXAFS_FACTEUR_INCORRECT;
  }

  /* Log dans l'historique */
  HIST_Titre(1, "Normaliser les oscillations EXAFS", "NORM");
  HIST_Texte(2, 0, "Division par un mod"e_GRAVE"le th"e_AIGU"orique de "MU"0");
  HIST_Texte(1, 0, "Oscillations normalis"e_AIGU"es : %s\n", exafs->nom);
  HIST_Texte(4, 0, "Seuil : %d, E0 = %f eV, hauteur = %g\n", seuil, e0, h_seuil);

  switch (methode)
  {
  case MODELE_MU0_CONSTANT: /* Facteur constant : dans apres[0], avant est ignor */
    if ( apres != NULL ) {
      retour = EXF_NormThConstante( exafs, apres[0], h_seuil );
      HIST_Texte(3, 0, "Mod"e_GRAVE"le : constante");
      HIST_Texte(5, 0, " (C = %g)", apres[0]);
      HIST_Texte(3, 0, "\n");
    }
    break;

  case MODELE_MU0_LENGELER: /* Modle de Lengeler-Eisenberger : E0 uniquement */
    retour = EXF_NormThLengeler( exafs, e0, h_seuil );
    HIST_Texte(3, 0, "Mod"e_GRAVE"le : Lengeler-Eisenberger\n");
    break;

  case MODELE_MU0_VICTOREEN:
    if ( ( avant != NULL ) && ( apres != NULL ) ) {
      retour = EXF_NormThVictoreen( exafs, e0, h_seuil, avant[0], avant[1], apres[0], apres[1] );
      HIST_Texte(3, 0, "Mod"e_GRAVE"le : Victoreen\n");
      HIST_Texte(5, 0, "  avant seuil, C/E"TROIS_EXPOSANT" - D/E^4, C = %f, D = %f\n", avant[0], avant[1]);
      HIST_Texte(5, 0, "  apr"e_GRAVE"s seuil, C/E"TROIS_EXPOSANT" - D/E^4, C = %f, D = %f\n", apres[0], apres[1]);
    }
    break;

  case MODELE_MU0_MCMASTER:
    if ( ( avant != NULL ) && ( apres != NULL ) ) {
      retour = EXF_NormThMcMaster( exafs, Z, e0, h_seuil, avant[0], avant[1], avant[2], avant[3], apres[0], apres[1], apres[2], apres[3] );
      HIST_Texte(3, 0, "Mod"e_GRAVE"le : McMaster\n");
      HIST_Texte(5, 0, "  avant seuil, A0 = %f, A1 = %f, A2 = %f, A3 = %f\n", avant[0], avant[1], avant[2], avant[3]);
      HIST_Texte(5, 0, "  apr"e_GRAVE"s seuil, A0 = %f, A1 = %f, A2 = %f, A3 = %f\n", apres[0], apres[1], apres[2], apres[3]);
    }
    break;

  case MODELE_MU0_EBEL:
    if ( ( avant != NULL ) && ( apres != NULL ) ) {
      double c[ 6 ];
      unsigned int i;

      for ( i = 0 ; i < 4 ; i++ ) {
	c[ i ] = avant[ i ];
      }
      c[ 4 ] = apres[ 0 ];
      c[ 5 ] = apres[ 1 ];

      retour = EXF_NormThEbel( exafs, e0, h_seuil, c );
      HIST_Texte(3, 0, "Mod"e_GRAVE"le : Ebel\n");
      HIST_Texte(5, 0, "  Coefficients : C0 = %g, C1 = %g, C2 = %g, C3 = %g, C4 = %g, C5 = %g\n", c[ 0 ], c[ 1 ], c[ 2 ], c[ 3 ], c[ 4 ], c[ 5 ]);
    }
    break;

  case MODELE_MU0_HEITLER:
    retour = EXF_NormThHeitler(exafs, e0, h_seuil);
    HIST_Texte(3, 0, "Mod"e_GRAVE"le : Heitler\n");
    break;
  }

  if ( retour < 0 ) {
    HIST_Titre(-1, "Impossible de normaliser", "NORM");
    return retour;
  }

  /* On met  jour le titre */
  titre = TXT_genere_chaine( txt_titre_resultat, exafs->nom );
  GRF_titre( exafs, titre );
  free( titre ); titre = NULL;

  /* On met  jour les bornes : seul Y a pu changer */
  GRF_bornes(exafs, NE_PAS_FAIRE, DESORDONNE);
  GRF_bornes(exafs->erreur_y, NE_PAS_FAIRE, DESORDONNE);

  /* On indique que le graphe obtenu est normalis */
  exafs->drapeaux.normalise = VRAI;
  GRF_ChangerType( exafs, GR_EXAFS );

  /* On prvient le coeur que le graphe a chang, seulement de contenu */
  exafs->drapeaux.sauve = FAUX;
  GRF_modifier( exafs, exafs, FAUX ); /* Nbr de points n'a pas chang, mais spectre si ! */

  /* On a fini... */
  return retour;
}
