/* LASE - Logiciel d'analyse des spectres EXAFS
   (C) Emmanuel CURIS, 1998 - Création du fichier le 14 juin 1998

   Sauvegarde au format postscript
   L'unité de base est le 1/100 de mm = taille d'un point donc 
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "math_utils.h" /* Pour Puissance */
#include "txt_utils.h"

#include "ps.h"
#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int PS_Graduer_Decimal(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe);
static int PS_GRD_Decimal    (FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, double tic_DH, double tic_GB, int p10);

static int PS_GraduerAxe_Decimal(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe,
				 double delta_texte, double taille_texte );

static int PS_Graduer_Log   (FILE *fichier, struct infos_graduation *infos, InfosAxe *axe);
static int PS_GRD_Log       (FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, double tic_DH, double tic_GB, int p10, int niveau, double val_ref);
static int PS_GraduerAxe_Log(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, double delta_texte, double taille_texte);
static int PS_SousGRD_Log   (FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, int p10, double val_ref, double place_legende, double x_texte, double y_texte);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Axes horizontaux ou verticaux ~-~-~-~-~-~-~-~-~ */

/* -------- Routine d'appel générique, qui appelle les vraies routines -------- */

int PS_AxeGradue(FILE *fichier, double x_debut, double y_debut, double x_fin, double y_fin, InfosAxe *axe,
		 double delta, double haut_texte)
{
  boolean vertical __attribute__((__unused__)) = FAUX;
  int retour = -1;
  struct infos_graduation infos = { 0.0, 0.0, 0.0, 0.0, 0.0, 0, 0 };

  /* Contrôles */
  if ( ( NULL == fichier ) || ( NULL == axe ) ) return -10;
  if ( GRADUATION_AUCUNE == axe->type_graduation ) return 0;
  if ( delta <= 0.0 ) {
    delta = 50.0; /* Valeur par défaut : 1/2 mm */
  }

  /* Dans quel sens graduer l'axe ? */
  if ( axe->proprietes & AXE_VERTICAL ) {
    vertical = VRAI;
    if ( y_debut == y_fin ) {
      (void) fprintf(stderr, "%s, %s l.%u : Axe réduit à un point verticalement\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
    (void) fprintf( fichier, "%% Graduation, axe vertical - de %g [%g] à %g [%g]\n", y_debut, axe->min, y_fin, axe->max );
  } else {
    if ( x_fin == x_debut ) {
      (void) fprintf(stderr, "%s, %s l.%u : Axe réduit à un point horizontalement\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
    (void) fprintf( fichier, "%% Graduation, axe horizontal - de %g [%g] à %g [%g]\n", x_debut, axe->min, x_fin, axe->max );
  }

  /* On prépare les graduations */
  retour = AXE_PreparerGraduations( x_debut, y_debut, x_fin, y_fin, delta, axe, &infos );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de préparer les graduations [code %d] !\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -10;
  }

  /* On met la couleur demandée pour les graduations, et le style en continu */
  (void) PS_Couleur( fichier, axe->couleur_graduations );
  (void) PS_StyleLigne( fichier, 0 );

  /* En fonction du type de graduation demandé */
  switch ( axe->type_graduation )
  {
  case GRADUATION_DECIMALE:
    retour = PS_Graduer_Decimal( fichier, &infos, axe );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant graduer [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
    if ( axe->proprietes & AXE_AVEC_GRADUATIONS ) {
      /* Texte des graduations : à 1 mm du bord, avec un texte de la hauteur indiquée (en points) */
      retour = PS_GraduerAxe_Decimal( fichier, &infos, axe, 100.0, haut_texte );
    }
    break;

  case GRADUATION_LOGARITHMIQUE:
    retour = PS_Graduer_Log( fichier, &infos, axe );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant graduer [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
    if ( axe->proprietes & AXE_AVEC_GRADUATIONS ) {
      /* Texte des graduations : à 1 mm du bord, avec un texte de la hauteur indiquée (en points) */
      retour = PS_GraduerAxe_Log( fichier, &infos, axe, 100.0, haut_texte );
    }
    break;

  case GRADUATION_PARABOLIQUE:
    OPTION_A_FAIRE("Graduation en unités paraboliques");
    break;

  case GRADUATION_GAUSSIENNE:
    OPTION_A_FAIRE("Graduation en unités gaussiennes");
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Méthode de graduation inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, axe->type_graduation);
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant graduer l'axe\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }
  return retour;
}

/* ~-~-~-~-~-~-~-~-~ Routines fondamentales ~-~-~-~-~-~-~-~-~ */

/* -------- Graduation décimale d'un axe -------- */

static int PS_Graduer_Decimal(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe)
{
  int niveau, retour;
  double tic_droite = 0.0, tic_gauche = 0.0;

  /* Définition de la taille de base des tics */
  tic_droite = axe->debord_droite; /* Débord par défaut [1/100 mm] */
  tic_gauche = axe->debord_gauche; /* Débord par défaut [1/100 mm] */
  if ( tic_droite < 0.0 ) tic_droite = 600.0; /* Tic de 6 mm de haut au plus */
  if ( tic_gauche < 0.0 ) tic_gauche = 0.0;

  /* Pour chaque puissance de 10, on gradue, en commençant par la plus élevée */
  for ( niveau = 0 ; niveau <= infos->niveau ; niveau++ ) {
    (void) fprintf( fichier, "%% -- en 10^%d\n", infos->p10_max - niveau );

    retour = PS_GRD_Decimal( fichier, infos, axe, tic_droite, tic_gauche, infos->p10_max - niveau );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans la graduation de niveau %d\n",
		     __FILE__, __FUNCTION__, __LINE__, niveau + 1);
      return +5;
    }

    /* Préparation pour le tour suivant */
    tic_droite -= 200.0; if ( tic_droite < 0.0 ) tic_droite = 0.0;
    tic_gauche -= 200.0; if ( tic_gauche < 0.0 ) tic_gauche = 0.0;
    if ( ( tic_droite <= 0.0 ) && ( tic_gauche <= 0.0 ) ) break; /* On a fini car les graduations sont invisibles */
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Graduation décimale pour une puissance de 10 imposée -------- */

static int PS_GRD_Decimal(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, double tic_DH, double tic_GB, int p10)
{
  double points[ 4 ] = { 0, 0, 0, 0 };  /* Init pour éviter un avertissement indû avec gcc 4.0 */
  double depart, puis_10;
  double delta_tics, valeur;
  int chiffre;

  /* Initialisation */
  puis_10 = Puissance( 10.0, p10 ); /* La puissance de 10 avec laquelle on travaille */
  delta_tics = puis_10 / infos->echelle;
  if ( delta_tics < 30.0 ) {
    /* Normalement, avec la préparation, on ne devrait jamais arriver ici ! */
    AVERTISSEMENT( "Intervalle entre deux tics trop faible [%g < 0,3 mm ]\n", delta_tics );
    return -5;
  }

  /* On récupère le chiffre des unités */
  if ( axe->min > 0.0 ) {
    double x_p;

    x_p = axe->min / puis_10;
    chiffre = (int) ( x_p - floor( x_p / 10.0 ) * 10.0);
    depart  = ( floor( x_p ) - ( double ) chiffre ) * puis_10;
  } else {
    double x_p;

    x_p = axe->min / puis_10;
    chiffre = (int) ( x_p - ceil( x_p / 10.0 ) * 10.0);
    depart  = ( ceil( x_p ) - ( double ) chiffre ) * puis_10;
  }

  /* On saute toutes les graduations avant le début de l'axe */
  valeur = depart + puis_10 * (double) chiffre;
  while ( valeur < axe->min ) {
    chiffre++;
    valeur = depart + puis_10 * (double) chiffre;
  }

  /* Préparation du tic : partie invariante */
  if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
    points[ 0 ] = infos->x_debut - tic_GB;
    points[ 2 ] = infos->x_debut + tic_DH;
  } else { /* Axe horizontal */
    points[ 1 ] = infos->y_debut - tic_GB;
    points[ 3 ] = infos->y_debut + tic_DH;
  }

  /* Tant que l'on est dans la partie visible, on fait les tics */
  while ( valeur  <= axe->max ) {
    /* Position des points */
    if ( axe->proprietes & AXE_VERTICAL ) {
      points[ 1 ] = infos->y_debut + ( valeur - axe->min ) / infos->echelle;
      points[ 3 ] = points[ 1 ];
    } else {
      points[ 0 ] = infos->x_debut + ( valeur - axe->min ) / infos->echelle;
      points[ 2 ] = points[ 0 ];
    }

    if ( ( chiffre % 5) == 0 ) { /* Multiple de 5 : trait un peu plus long */
      if ( axe->proprietes & AXE_VERTICAL ) {
	(void) fprintf( fichier,"%g %g moveto %g %g lineto ",
			points[ 0 ] - 100.0 * ( tic_GB > 0.0 ), points[ 1 ], points[ 2 ] + 100.0 * ( tic_DH > 0.0 ), points[ 3 ] );
      } else {
	(void) fprintf( fichier,"%g %g moveto %g %g lineto ",
			points[ 0 ], points[ 1 ] - 100.0 * ( tic_GB > 0.0 ), points[ 2 ], points[ 3 ] + 100.0 * ( tic_DH > 0.0 ) );
      }
    } else {
      (void) fprintf( fichier, "%g %g moveto %g %g lineto ",
		      points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
    }

    /* Passage au tic suivant */
    chiffre++;
    valeur = depart + puis_10 * (double) chiffre; /* Recalcul pour éviter de rester sur place si pas trop faible ! */
  }

  /* Terminé */
  (void) fprintf( fichier, "stroke\n" );

  return 0;
}

/* -------- Graduer en mettant le texte -------- */

static int PS_GraduerAxe_Decimal(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe,
				 double delta_texte, double taille_texte )
{
  double depart, valeur;
  unsigned int i = 0;
  double pas_graduation;
  char texte[ 100 ];

  /* Calcul du pas de la graduation */
  pas_graduation = axe->pas_graduation;
  if ( pas_graduation <= 0.0 ) {
    taille_texte *= 2540.0 / 72.0; /* Conversion pt -> 1/100 mm, 1 pt = 1/72 pouce environ, 1 pouce = 2,54 mm environ */
    if ( axe->proprietes & AXE_VERTICAL ) {
      pas_graduation = AXE_GraduationOptimale( axe->min, axe->max, infos->y_fin - infos->y_debut, taille_texte, VRAI );
    } else {
      pas_graduation = AXE_GraduationOptimale( axe->min, axe->max, infos->x_fin - infos->x_debut, taille_texte, FAUX );
    }
    if ( pas_graduation <= 0.0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de trouver le pas automatiquement ! [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, (int) pas_graduation);
      return -1;
    }
  }

  /* Où commence-t-on ? [on s'assure d'être sur un multiple du pas !] */
  depart = pas_graduation * floor( axe->min / pas_graduation );

  /* On se place au début des valeurs visibles */
  valeur = depart;
  while ( valeur < axe->min ) {
    i++;
    valeur = depart + pas_graduation * (double) i;
  }

  /* On affiche les valeurs dans la région choisie */
  while ( valeur <= axe->max ) {
    /* Correction des erreurs d'arrondi pour le 0 */
    if ( ( fabs( valeur ) < 1.0e-10 ) &&
	 ( ( fabs( axe->min ) > 1.0e-10 ) || ( fabs( axe->max ) > 1.0e-10 ) ) )
      valeur = 0.0;

    (void) TXT_CNV_D_vTexte( valeur, 0.0, 4, 3, 3, texte, NULL );
    if ( axe->proprietes & AXE_VERTICAL ) {
      if ( axe->proprietes & AXE_SECONDAIRE ) {
	PS_Texte( fichier, infos->x_debut + delta_texte, infos->y_debut + ( valeur - axe->min ) / infos->echelle, texte,
		  BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_CENTRER_V, 0.0, NOIR );
      } else {
	PS_Texte( fichier, infos->x_debut - delta_texte, infos->y_debut + ( valeur - axe->min ) / infos->echelle, texte,
		  BI_TEXTE_FER_A_DROITE, BI_TEXTE_CENTRER_V, 0.0, NOIR );
      }
    } else {
      if ( axe->proprietes & AXE_SECONDAIRE ) {
	PS_Texte( fichier, infos->x_debut + ( valeur - axe->min ) / infos->echelle, infos->y_debut + delta_texte, texte,
		  BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_BAS, 0.0, NOIR );
      } else {
	PS_Texte( fichier, infos->x_debut + ( valeur - axe->min ) / infos->echelle, infos->y_debut - delta_texte, texte,
		  BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_HAUT, 0.0, NOIR );
      }
    }

    /* On passe à la suivante */
    i++;
    valeur = depart + pas_graduation * (double) i;
  }

  /* Graduations terminées */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Graduation logarithmique ~-~-~-~-~-~-~-~-~ */

static int PS_Graduer_Log(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe)
{
  boolean tic_place = FAUX;
  int tic_droite = 0.0, tic_gauche = 0.0;
  int p10_min, p10_max, p10;
  double pos_depart, pos_fin, place_puissance, hauteur __attribute__((__unused__)), debut, fin;
  double points[ 4 ] = { 0.0, 0.0, 0.0, 0.0 }, position_tic;

  /* Définition de la taille de base des tics */
  tic_droite = axe->debord_droite; /* Débord par défaut [1/100 mm] */
  tic_gauche = axe->debord_gauche; /* Débord par défaut [1/100 mm] */
  if ( tic_droite < 0.0 ) tic_droite = 600.0; /* Tic de 6 mm de haut au plus */
  if ( tic_gauche < 0.0 ) tic_gauche = 0.0;

  /* Préparation du tic : partie invariante */
  if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
    points[ 0 ] = infos->x_debut - tic_gauche;
    points[ 2 ] = infos->x_debut + tic_droite;

    debut = infos->y_debut;
    fin   = infos->y_fin;
  } else { /* Axe horizontal */
    points[ 1 ] = infos->y_debut + tic_gauche;
    points[ 3 ] = infos->y_debut - tic_droite;

    debut = infos->x_debut;
    fin   = infos->x_fin;
  }
  hauteur = (double) ( fin - debut );

  /* Quelle place occupe une puissance complète sur l'axe [= un intervalle de 1 en log] */
  place_puissance = 1.0 / infos->echelle;
  if ( place_puissance < 1000.0 ) { /* Moins de 1 cm : seules certaines puissances seront placées */
    p10 = 1 + (int) rint( log10( infos->echelle ) );
    (void) PS_GRD_Decimal( fichier, infos, axe, tic_droite, tic_gauche, p10 );
    return 0;
  }

  /* Quelles sont les puissances de 10 extrêmes impliquées ? */
  p10_min = (int) floor( axe->min );
  p10_max = (int) ceil ( axe->max );

  /* On place les graduations principales */
  for ( p10 = p10_min ; p10 < p10_max ; p10++ ) {
    pos_depart  = p10 - axe->min; /* log10( 10^n ) = n : 1er tic */
    pos_depart /= infos->echelle;

    pos_fin = (p10 + 1) - axe->min; /* log10( 10^n+1 ) = n + 1 : dernier tic, 1er suivant */
    pos_fin /= infos->echelle;

    /* On place le tic de fin, s'il est dans la région */
    position_tic = debut + pos_fin;
    if ( ( position_tic >= debut ) && ( position_tic <= fin ) ) {
      if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
	points[ 1 ] = position_tic;
	points[ 3 ] = points[ 1 ];
      } else {
	points[ 0 ] = position_tic;
	points[ 2 ] = points[ 0 ];
      }

      (void) fprintf( fichier,"%g %g moveto %g %g lineto ",
		      points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
      tic_place = VRAI;		/* On a placé au moins un tic */
    }
  }
  if ( VRAI == tic_place ) {
    (void) fprintf( fichier, "stroke\n" );
  }

  /* Pour chacune de ces puissances, on gradue dans les niveaux inférieurs */
  for ( p10 = p10_min ; p10 < p10_max ; p10++ ) {
    /* On place le niveau en-dessous avec une taille inférieure [1 mm de moins] */
    PS_GRD_Log( fichier, infos, axe,
		( tic_droite > 0 ) ? ( tic_droite - 200.0 ) : ( 0 ), ( tic_gauche > 0 ) ? ( tic_gauche - 200.0 ) : ( 0 ),
		p10, 1, Puissance( 10.0, p10 ) );
  }

  /* Graduation effectuée */
  return 0;
}

static int PS_GRD_Log(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, double tic_DH, double tic_GB, int p10, int niveau, double val_ref)
{
  boolean tic_place = FAUX;
  double positions[ 10 ];
  double points[ 4 ] = { 0.0, 0.0, 0.0, 0.0 }, point_avant, position_tic;
  double debut, fin;
  unsigned int i;

  /* Les tics sont-ils assez grands ? [ie plus d'1 mm] */
  if ( ( tic_DH < 100.0 ) && ( tic_GB < 100.0 ) ) return +10;

  /* Commentaire pour s'y retrouver dans le fichier */
  (void) fprintf( fichier, "%% Graduations en semi-log, %g pour 10^%d\n", val_ref, p10 );

  /* Position des tics à la sortie */
  for ( i = 0 ; i < 10 ; i++ ) {
    positions[ i ] = val_ref +  Puissance( 10.0, p10 ) * (double) i; /* Valeur décimale */
    positions[ i ] = log10( positions[ i ] ); /* Correspondance en log */
  }

  /* Préparation du tic : partie invariante */
  if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
    points[ 0 ] = infos->x_debut - tic_GB;
    points[ 2 ] = infos->x_debut + tic_DH;

    debut = infos->y_debut;
    fin   = infos->y_fin;
  } else { /* Axe horizontal */
    points[ 1 ] = infos->y_debut + tic_GB;
    points[ 3 ] = infos->y_debut - tic_DH;

    debut = infos->x_debut;
    fin   = infos->x_fin;
  }
  point_avant = debut + ( positions[ 0 ] - axe->min ) / infos->echelle;

  /* On place les tics pour "2" à "9" [1 placé avant comme "x" d'un tour supérieur, "10" = "1" du suivant]  */
  for ( i = 1 ; i < 10 ; i++ ) {
    position_tic = debut + ( positions[ i ] - axe->min ) / infos->echelle;
    if ( ( position_tic - point_avant ) < 50.0 ) {
      /* Moins de 0,5 mm entre deux tics : inutile de les placer car serait illisible et les suivants sont de plus en plus serrés, donc on arrête */
      break;
    }
    point_avant = position_tic;

    if ( ( position_tic >= debut ) && ( position_tic <= fin ) ) {
      /* Le tic est bien visible ! */
      if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
	points[ 1 ] = position_tic;
	points[ 3 ] = points[ 1 ];
      } else {
	points[ 0 ] = position_tic;
	points[ 2 ] = points[ 0 ];
      }

      (void) fprintf( fichier,"%g %g moveto %g %g lineto ",
		      points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
      tic_place = VRAI;		/* On a placé au moins un tic */
    }
  }
  if ( VRAI == tic_place ) {
    /* On a bien placé au moins un tic [sinon, stroke sans pile : erreur PostScript */
    (void) fprintf( fichier, "stroke\n" );
  }

  /* On recommence pour placer, si nécessaire, les niveaux inférieurs de graduation */
  if ( ( axe->finesse != 0 ) && ( niveau > (int) axe->finesse ) ) {
    /* On ne s'intéresse pas à ces niveaux ! */
    return 0;
  }

  /* On essaye de placer les tics intermédiaires : puissance de 10 inférieure entre "1" & "2" : tiret plus court de 2 mm à chaque fois */
  PS_GRD_Log( fichier, infos, axe,
	      ( tic_DH > 0 ) ? ( tic_DH - 200.0 ) : ( 0 ), ( tic_GB > 0 ) ? ( tic_GB - 200.0 ) : ( 0 ),
	      p10 - 1, niveau + 1, val_ref );

  point_avant = debut + ( positions[ 0 ] - axe->min ) / infos->echelle;
  /* Idem, pour les autres puissances de 10 */
  for ( i = 1 ; i < 10 ; i++ ) {
    position_tic = debut + ( positions[ i ] - axe->min ) / infos->echelle;

    if ( ( position_tic - point_avant ) < 500.0 ) {
      /* Moins de 5 mm entre deux tics : inutile de placer des graduations en-dessous car serait illisible, donc on arrête */
      break;
    }
    point_avant = position_tic;

    if ( position_tic >= fin ) {	/* On est au-delà */
      /* Les autres le seront donc aussi : inutile de continuer ! */
      break;
    }

    if ( position_tic < debut ) { /* On est en-dessous */
      double position_tic_suivant;

      position_tic_suivant = debut + ( positions[ i + 1 ] - axe->min ) / infos->echelle;
      if ( position_tic_suivant < debut ) {
	continue;		/* On n'est pas visible pour cette série */
      }
    }

    /* On place les tics */
    PS_GRD_Log( fichier, infos, axe,
		( tic_DH > 0 ) ? ( tic_DH - 200.0 ) : ( 0 ), ( tic_GB > 0 ) ? ( tic_GB - 200.0 ) : ( 0 ),
		p10 - 1, niveau + 1, pow( 10.0, positions[ i ] ) );
  }

  return 0;
}

/* -------- Placer les textes / axe logarithmique -------- */

static int PS_GraduerAxe_Log(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, double delta_texte, double taille_texte)
{
  int p10_min, p10_max, p10;
  double pos_depart, pos_fin, place_puissance, hauteur __attribute__((__unused__));
  double place_legende;
  double x_texte = 0.0, y_texte = 0.0;
  char texte[ 100 ];

  /* Préparation du texte : partie invariante */
  if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
    x_texte = infos->x_debut - delta_texte;
    hauteur = (double) ( infos->y_fin - infos->y_debut );
    place_legende = taille_texte;
  } else { /* Axe horizontal */
    y_texte = infos->y_debut + delta_texte;
    hauteur = (double) ( infos->x_fin - infos->x_debut );
    place_legende = taille_texte * 8; /* À affiner plus tard */
  }
  place_legende *= 2540.0/72.0; /* Conversion pt -> 1/100 mm, 1 pt = 1/72 pouce environ, 1 pouce = 2,54 mm environ */

  /* Quelle place occupe une puissance complète sur l'axe [= un intervalle de 1 en log] */
  place_puissance = 1.0 / infos->echelle;
  if ( place_puissance < place_legende ) { /* Moins de 10 pixels : seules certaines puissances seront placées */
    place_legende = 3 + TXT_nbr_car_I( (int) axe->max ); /* 10^entier */
    p10 = (int) ceil( place_legende * infos->echelle );
    return 0;
  }

  /* Quelles sont les puissances de 10 extrêmes impliquées ? */
  p10_min = (int) floor( axe->min );
  p10_max = (int) ceil ( axe->max );

  /* Pour chacune de ces puissances, on place la légende */
  for ( p10 = p10_min ; p10 <= p10_max ; p10++ ) {
    pos_depart  = p10 - axe->min; /* log10( 10^n ) = n : 1er tic */
    pos_depart /= infos->echelle;

    pos_fin = (p10 + 1) - axe->min; /* log10( 10^n+1 ) = n + 1 : dernier tic, 1er suivant */
    pos_fin /= infos->echelle;

    (void) TXT_CNV_D_vTexte( Puissance( 10.0, p10 ), 0.0, 4, 3, 3, texte, NULL );

    /* On place la puissance du tic de début, s'il est dans la région */
    if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
      y_texte = infos->y_debut + pos_depart;
      if ( ( y_texte >= infos->y_debut ) && ( y_texte <= infos->y_fin ) ) {
	PS_Texte( fichier, x_texte, y_texte, texte, BI_TEXTE_FER_A_DROITE, BI_TEXTE_CENTRER_V, 0.0, NOIR );
      }
    } else {
      x_texte = infos->x_debut + pos_depart;
      if ( x_texte < infos->x_fin ) {
	PS_Texte( fichier, x_texte, y_texte, texte, BI_TEXTE_FER_A_DROITE, BI_TEXTE_CENTRER_V, 0.0, NOIR );
      }
    }

    /* On place le niveau en-dessous avec une taille inférieure */
    PS_SousGRD_Log( fichier, infos, axe, p10, Puissance( 10.0, p10 ), place_legende, x_texte, y_texte );
  }

  /* Graduation effectuée */
  return 0;
}

static int PS_SousGRD_Log(FILE *fichier, struct infos_graduation *infos, InfosAxe *axe, int p10, double val_ref, double place_legende, double x_texte, double y_texte)
{
  double positions[ 11 ], debut, fin, position_texte;
  double point_avant, point_fin;
  unsigned int i;
  char texte[ 100 ];

  /* Position des tics à l'écran */
  for ( i = 0 ; i <= 10 ; i++ ) {
    positions[ i ] = val_ref +  Puissance( 10.0, p10 ) * (double) i; /* Valeur décimale */
    positions[ i ] = log10( positions[ i ] ); /* Correspondance en log */
  }
  if ( ( ( positions[ 9 ] - positions[ 0 ] ) / infos->echelle ) <= place_legende ) return +5;

  /* Préparation du texte : partie invariante */
  if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
    point_avant = y_texte;

    debut = infos->y_debut;
    fin   = infos->y_fin;
  } else { /* Axe horizontal */
    point_avant = x_texte;

    debut = infos->x_debut;
    fin   = infos->x_fin;
  }
  point_avant = infos->y_debut + ( positions[  0 ] - axe->min ) / infos->echelle;
  point_fin   = infos->y_debut + ( positions[ 10 ] - axe->min ) / infos->echelle;
  
  /* On place les textes pour "2" à "9" [1 placé avant comme "x" d'un tour supérieur, "10" = "1" du suivant]  */
  for ( i = 1 ; i < 10 ; i++ ) {
    /* On calcule la position du texte */
    position_texte = debut + ( positions[ i ] - axe->min ) / infos->echelle;

    if ( ( position_texte >= debut ) && ( position_texte <= fin ) ) {
      /* Le texte est bien dans la région affichée */
      if ( ( ( position_texte - point_avant ) > place_legende ) && ( ( point_fin - position_texte ) > place_legende ) ) {
	/* Et on ne chevauche ni avec le texte précédent, ni avec le niveau supérieur ==> affichage possible */

	(void) TXT_CNV_D_vTexte( val_ref + Puissance( 10.0, p10 ) * (double) i, 0.0, 4, 3, 3, texte, NULL );

	if ( axe->proprietes & AXE_VERTICAL ) {
	  PS_Texte( fichier, x_texte, position_texte, texte, BI_TEXTE_FER_A_DROITE, BI_TEXTE_CENTRER_V, 0.0, NOIR );
	} else {
	  PS_Texte( fichier, position_texte, y_texte, texte, BI_TEXTE_CENTRER_H, BI_TEXTE_CENTRER_V, 0.0, NOIR );
	}

	point_avant = position_texte;
      }
    }
  }

  /* On place pour les niveaux en-dessous, si nécessaire */
  PS_SousGRD_Log( fichier, infos, axe, p10 - 1, val_ref, place_legende, x_texte, y_texte );

  point_avant = infos->y_debut + ( positions[  0 ] - axe->min ) / infos->echelle;
  for ( i = 1 ; i < 10 ; i++ ) {
    /* On calcule la position du texte */
    position_texte = debut + ( positions[ i ] - axe->min ) / infos->echelle;

    if ( position_texte > fin ) {
      /* On déborde par en haut, donc après aussi : inutile de continuer */
      break;
    }
    if ( position_texte < debut ) { /* On est en-dessous */
      double p_suivant;

      p_suivant = debut + ( positions[ i + 1 ] - axe->min ) / infos->echelle;
      if ( p_suivant <= debut ) {
	continue;
      }
    }

    PS_SousGRD_Log( fichier, infos, axe, p10 - 1, pow( 10.0, positions[ i ] ), place_legende, x_texte, y_texte );
  }

  return 0;
}
