/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2001 Emmanuel CURIS
   Cration : le 5 fvrier 1999

   Rcupration des donnes exprimentales pour obtenir  total

   Extraction proprement dite, qui redirige vers les bonnes routines
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "serie.h"

#include "import_xas.h"
#include "statistiques.h" /* Pour les moyennes / fusion */
#include "tri.h"

#include "ext_extraire.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

extern const char *textes_type_ligne[];

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned long EXTRAIT_CompterSpectres  (FILE *fichier, int commentaire, int ligne, unsigned int col_x);
static unsigned long EXTRAIT_ComptageXInverses(FILE *fichier, unsigned int col_x, int commentaire);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

int EXTRAIT_ChargerFichier(const char *nom, Infos_Extraction *infos, Serie_Graphes *serie, int type_ligne)
{
  FILE *fichier;
  int retour;
  unsigned long nbr_spectres, i_sp = 0;
  EXT_ListeGraphes *graphes = NULL;

  /* Vrifications prliminaires */
  if ( ( NULL == nom ) || ( NULL == infos ) || ( NULL == serie ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : paramtre incorrect (nom=%s, infos=%p, serie=%p)\n",
		   __FILE__, __FUNCTION__, __LINE__, nom, (void *) infos, (void *) serie);
    return -1;
  }

  /* On essaye d'ouvrir le fichier */
  fichier = fopen( nom, "rb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir le fichier [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
    (void) BI_ERR_alerte( NULL, &ERR_ImpossibleOuvrir, VRAI, nom );
    return -2;
  }

  /* Log dans l'historique */
  HIST_Titre(1, "Import de donn"e_AIGU"es exp"e_AIGU"rimentales", "E/S");
  HIST_Texte(1, 0, "Fichier   : %s\n", nom);
  HIST_Texte(1, 0, "Technique : absorption X");
  HIST_Texte(2, 0, " [");
  if ( infos->colonnes_trans != NULL )
    HIST_Texte(2, 0, "transmission");
  HIST_Texte(2, 0, " ; ");
  if ( infos->colonnes_fluo != NULL )
    HIST_Texte(2, 0, "fluorescence");
  HIST_Texte(2, 0, " ; ");
  if ( infos->colonnes_det != NULL )
    HIST_Texte(2, 0, "d"e_AIGU"tection d'"e_AIGU"lectrons");
  HIST_Texte(2, 0, "]");
  HIST_Texte(1, 0, "\n");
  HIST_OptionChoisie(4, "Ligne exp"e_AIGU"rimentale", NOMBRE_LIGNES_EXPERIMENTALES, type_ligne, textes_type_ligne);
  HIST_Texte(4, 0, "\n");

  /* S'il y a un en-tte, on le saute */
  retour = sauter_en_tete( fichier, infos->nombre_lignes_en_tete, infos->commentaire );
  if ( EOF == retour ) {
    (void) BI_ERR_alerte( NULL, &ERR_EnteteIncorrect, VRAI, nom, textes_type_ligne[ type_ligne ] );
    AVERTISSEMENT( "ERREUR dans sauter_en_tete [code %d ; fichier %s]", retour, nom );
    (void) fclose( fichier );
    return -3;
  }

  /* Compter le nombre de spectres, si ncessaire
     On utilise une variable speciale pour deux raisons :
      1) si 0, compte automatiquement -> plus 0 -> au fichier suivant, pas recompte automatiquement
      2) si erreur, diminue au bon nombre -> potentiellement faux pour le fichier suivant
 */
  if ( 0 == infos->nbr_successifs ) {
    AVERTISSEMENT( "On va compter le nombre de spectres dans %s", nom );
    /* Detection automatique du nombre de spectres */
    nbr_spectres = EXTRAIT_CompterSpectres( fichier, infos->commentaire, type_ligne, infos->col_x );
    if ( 0 == nbr_spectres ) {
      (void) BI_ERR_alerte( NULL, &ERR_AucunSpectre, VRAI, nom );
      AVERTISSEMENT( "ERREUR dans EXTRAIT_CompterSpectres [%s]", nom );
      (void) fclose( fichier );
      return -4;
    }
  } else {
    nbr_spectres = infos->nbr_successifs;
  }
  HIST_Texte(4, 0, "Format    : %d spectre(s) "a_GRAVE" la suite dans le fichier", infos->nbr_successifs);
  if ( ( DETECT_MULTI_MOYENNER == infos->MX_traitement ) ||
       ( DETECT_MULTI_MOY_LOG  == infos->MX_traitement ) )
    HIST_Texte(4, 0, " (moyenn"e_AIGU"s)");
  else if ( DETECT_MULTI_FONDRE == infos->MX_traitement )
    HIST_Texte(4, 0, " (fusionn"e_AIGU"s)");
  HIST_Texte(4, 0, "\n");

  /* On rserve pour _tous_ les spectres */
  graphes = ( EXT_ListeGraphes * ) calloc( sizeof( EXT_ListeGraphes ), nbr_spectres );
  if ( NULL == graphes ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mmoire insuffisante\n", __FILE__, __FUNCTION__, __LINE__);
    (void) BI_ERR_alerte( NULL, &ERR_MemoireInsuffisante1, VRAI, nom );
    (void) fclose( fichier );
    return -10;
  }

  /* Pour chacun des spectres... */
  while ( nbr_spectres > i_sp ) {
    int ignorer = FAUX;
    EXT_ListeGraphes *lst_graphes;

    /* Est-ce un spectre  conserver ? */
    lst_graphes = graphes + i_sp;
    if ( infos->filtre_MX != NULL ) {
      ignorer = !( TXT_est_dans_intervalle( 1 + i_sp, infos->filtre_MX, infos->nbr_I_MX ) );
      if ( VRAI == ignorer ) {
	lst_graphes = NULL;
      }
    }

    switch ( type_ligne )
    {
    case LIGNE_LURE_AQUVISU:
    case LIGNE_LURE_D21:
    case LIGNE_HASYLAB_DORIS:
    case LIGNE_ESRF_SNBL:
    case LIGNE_ESRF_ID20:
    case LIGNE_ESRF_ID21:
    case LIGNE_ESRF_ID22:
    case LIGNE_SOLEIL_SAMBA:
      retour = EXTRAIT_Generique( nom, fichier, infos, lst_graphes, type_ligne, nbr_spectres, serie, i_sp );
      break;

    case LIGNE_NSLS_MXANES:
      retour = NSLS_LireSpectre( nom, fichier, infos, lst_graphes, serie );
      break;

    case LIGNE_BESSY2_:
      (void) fprintf(stderr, "%s,%s l.%u : Bessy-2 non-ASCII, on ne devrait pas tre l !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      retour = BESSY2_LireSpectre( nom, infos, lst_graphes, serie, i_sp );
      break;

    default:
      (void) fprintf(stderr, "%s, %s l.%u : ligne inconnue [%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, type_ligne);
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      retour = -1;
      break;
    }

    /* Conventions : retour, valeur renvoye par la fonction de lecture d'un spectre
       0  = aucune erreur, chargement effectu correctement
       >0 = erreur mineure, le spectre n'a pas t charg entirement mais on peut continuer avec le fichier
            la valeur renvoye est le nombre de points qui ont pu tre lus
       <0 = erreur grave, on annule le chargement du spectre et du fichier
            si la valeur est < -100, on n'affiche pas de message,
            si la valeur est > -100, on affiche un message.
    */
    if ( retour != 0 ) {
      if ( retour < 0 ) { /* Chargement interrompu (erreur grave)... */
	if ( retour > -100 ) {
	  (void) BI_ERR_alerte( NULL, &ERR_AutreErreur, VRAI, 1 + i_sp, nbr_spectres, nom );
	}
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : erreur lors de l'extraction du fichier [%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour );
#endif
	EXTRAIT_DetruireGraphes( infos, lst_graphes );
	nbr_spectres = i_sp; /* Il n'y a donc que a comme spectres disponibles */
	break;
      } else { /* Chargement partiel : coupure faisceau... */
	if ( retour < 3 ) {
	  (void) BI_ERR_alerte( NULL, &ERR_TropPeuPoints, VRAI, nom );
	  EXTRAIT_DetruireGraphes( infos, lst_graphes );
	} else {
	  unsigned long i;
	  
	  /* Adaptation de la taille : troncature au nombre de points indiqu */
	  GRF_AdapteTaille( lst_graphes->i_0, retour );
	  if ( lst_graphes->i_1 != NULL ) {
	    for ( i = 0 ; i < lst_graphes->nbr_spectres ; i++ ) {
	      GRF_AdapteTaille( lst_graphes->i_1[ i ], retour );
	    }
	  }
	  if ( lst_graphes->spectres != NULL ) {
	    for ( i = 0 ; i < lst_graphes->nbr_spectres ; i++ ) {
	      GRF_AdapteTaille( lst_graphes->spectres[ i ], retour );
	    }
	  }
	}
      }
    }

    /* On ralise le post-traitement des graphes : bornes, ventuellement adaptation des X,... */
    retour = EXTRAIT_AdapterSpectres( type_ligne, infos, graphes + i_sp, nom );

    /* Et on passe au spectre suivant, s'il existe */
    i_sp++;
  }

  /* On a fini de lire le fichier, normalement ! */
  (void) fclose( fichier );

  /* Si plusieurs spectres : post-traitement commun, colonne par colonne */
  if ( ( nbr_spectres > 1 ) && ( infos->MX_traitement != DETECT_MULTI_SEPARER ) ) {
    Graphe **table_graphes;
    int methode_moyenne = DISP_X_RECALER;

    if ( DETECT_MULTI_MOYENNER == infos->MX_traitement ) {
      methode_moyenne = DISP_X_MOYENNER;
    } else if ( DETECT_MULTI_FONDRE == infos->MX_traitement ) {
      methode_moyenne = DISP_X_FUSION;
    }

    table_graphes = ( Graphe ** ) calloc( sizeof( Graphe *), nbr_spectres );
    if ( NULL == table_graphes ) {
      (void) BI_ERR_alerte( NULL, &ERR_MemoireInsuffisante2, VRAI, nom );
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante / table des graphes [moyenne/fusion]\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
    } else {
      Graphe *moyenne;
      unsigned long nbr_reel, idx_reel;
      unsigned int idx_colonnes, nbr_colonnes;

      /* Moyenne des I0, si ncessaire */
      for ( i_sp = 0, idx_reel = 0, nbr_reel = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
	if ( graphes[ i_sp ].i_0 != NULL ) {
	  table_graphes[ idx_reel ] = graphes[ i_sp ].i_0;
	  nbr_reel++;
	  idx_reel++;
	}
      }
      moyenne = MoyennerGraphes( table_graphes, nbr_reel, methode_moyenne, FAUX, FAUX );
      if ( moyenne != NULL ) {
	char *titre;

	/* On efface les originaux, puisque seule la moyenne/fusion nous intresse ! */
	for ( i_sp = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
	  GRF_efface( graphes[ i_sp ].i_0, VRAI, VRAI );
	  graphes[ i_sp ].i_0 = NULL;
	}

	/* On donne un joli titre */
	titre = TXT_genere_chaine( titre_I0_moyenne, nom, nbr_reel );
	GRF_titre( moyenne, titre );
	free( titre );
      }

      /* Combien de colonnes a-t-on au plus ? */
      nbr_colonnes = graphes[ 0 ].nbr_spectres;
      for ( i_sp = 1 ; i_sp < nbr_spectres ; i_sp++ ) {
	if ( graphes[ i_sp ].nbr_spectres > nbr_colonnes )
	  nbr_colonnes = graphes[ i_sp ].nbr_spectres;
      }

      /* Moyenne des graphes, si ncessaire, colonne par colonne */
      if ( graphes[ 0 ].spectres != NULL ) {
	for ( idx_colonnes = 0 ; idx_colonnes < nbr_colonnes ; idx_colonnes++ ) {
	  for ( i_sp = 0, idx_reel = 0, nbr_reel = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
	    if ( idx_colonnes < graphes[ i_sp ].nbr_spectres ) {
	      if ( graphes[ i_sp ].spectres[ idx_colonnes ] != NULL ) {
		table_graphes[ idx_reel ] = graphes[ i_sp ].spectres[ idx_colonnes ];
		nbr_reel++;
		idx_reel++;
	      }
	    }
	  }

	  moyenne = MoyennerGraphes( table_graphes, nbr_reel, methode_moyenne, FAUX, FAUX );
	  if ( moyenne != NULL ) {
	    char *titre;

	    /* On efface les originaux, puisque seule la moyenne/fusion nous intresse ! */
	    for ( i_sp = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
	      if ( idx_colonnes < graphes[ i_sp ].nbr_spectres ) {
		GRF_efface( graphes[ i_sp ].spectres[ idx_colonnes ], VRAI, VRAI );
		graphes[ i_sp ].spectres[ idx_colonnes ] = NULL;
	      }
	    }

	    /* On donne un joli titre */
	    titre = TXT_genere_chaine( titre_mu_moyenne, nom, idx_colonnes, nbr_reel );
	    GRF_titre( moyenne, titre );
	    free( titre );
	  }
	}
      }

      /* Moyenne des I1, si ncessaire, colonne par colonne */
      if ( graphes[ 0 ].i_1 != NULL ) {
	for ( idx_colonnes = 0 ; idx_colonnes < nbr_colonnes ; idx_colonnes++ ) {
	  for ( i_sp = 0, idx_reel = 0, nbr_reel = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
	    if ( idx_colonnes < graphes[ i_sp ].nbr_spectres ) {
	      if ( graphes[ i_sp ].i_1[ idx_colonnes ] != NULL ) {
		table_graphes[ idx_reel ] = graphes[ i_sp ].i_1[ idx_colonnes ];
		nbr_reel++;
		idx_reel++;
	      }
	    }
	  }

	  moyenne = MoyennerGraphes( table_graphes, nbr_reel, methode_moyenne, FAUX, FAUX );
	  if ( moyenne != NULL ) {
	    char *titre;

	    /* On efface les originaux, puisque seule la moyenne/fusion nous intresse ! */
	    for ( i_sp = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
	      if ( idx_colonnes < graphes[ i_sp ].nbr_spectres ) {
		GRF_efface( graphes[ i_sp ].i_1[ idx_colonnes ], VRAI, VRAI );
		graphes[ i_sp ].i_1[ idx_colonnes ] = NULL;
	      }
	    }

	    /* On donne un joli titre */
	    titre = TXT_genere_chaine( titre_I1_moyenne, nom, idx_colonnes, nbr_reel );
	    GRF_titre( moyenne, titre );
	    free( titre );
	  }
	}
      }
    }

    /* Plus besoin de cette table annexe */
    free( table_graphes );
  }

  /* Tout s'est bien pass : on efface donc les tables inutiles */
  for ( i_sp = 0 ; i_sp < nbr_spectres ; i_sp++ ) {
    free( graphes[ i_sp ].spectres );
    free( graphes[ i_sp ].i_1 );
  }
  free( graphes );
  return 0;
}

/* -~-~-~-~-~-~-~-~ Post-traitement des spectres -~-~-~-~-~-~-~-~ */

int EXTRAIT_AdapterSpectres(int type_ligne, Infos_Extraction *infos, EXT_ListeGraphes *graphes, const char *nom_fichier)
{
  unsigned int j;
  Graphe **g;

  if ( ( NULL == graphes ) || ( NULL == infos ) ) {
    (void) fprintf(stderr, "EXTRAIT_AdapterSpectres : paramtre incorrect (ligne=%d, infos=%p, graphes=%p) !\n",
		   type_ligne, (void *) infos, (void *) graphes);
    return -1;
  }

  /* 1. Si ncessaire, traitement des X
     Remarque : le traitement des doublons et tout ne peut pas tre fait ici, car joue aussi sur les Y
   */
  /* Conversion keV -> eV */
  if ( ( infos->post_traitement & DETECT_POST_X_KEV ) && ( graphes->x != NULL ) ) {
    unsigned long i, nbr_points = 0;

    /* Dtection du nombre de points */
    if ( graphes->i_0 != NULL )
      nbr_points = graphes->i_0->nombre_point;
    else if ( graphes->spectres != NULL ) {
      for ( j = 0 ; j < graphes->nbr_spectres ; j++ ) {
	if ( graphes->spectres[ j ] != NULL ) {
	  nbr_points = graphes->spectres[ j ]->nombre_point;
	  break;
	}
      }
    } else if ( graphes->i_1 != NULL ) {
      for ( j = 0 ; j < graphes->nbr_spectres ; j++ ) {
	if ( graphes->i_1[ j ] != NULL ) {
	  nbr_points = graphes->i_1[ j ]->nombre_point;
	  break;
	}
      }
    }

    /* Conversion */
    for ( i = 0 ; i < nbr_points ; i++ ) {
      graphes->x[ i ] *= 1000.0;
    }
  }

  /* 2. On met les X corrects partout (et les couleurs aussi) (et les units en X aussi) */
  if ( graphes->i_0 != NULL ) {
    if ( graphes->i_0->x != graphes->x ) {
      memcpy( graphes->i_0->x, graphes->x, sizeof( double ) * graphes->i_0->nombre_point );
    }
    GRF_ChangerUnite( graphes->i_0, AXE_X, "eV" );
  }
  if ( graphes->i_1 != NULL ) {
    unsigned int couleur = NOIR, style = BI_CONTINU, marque = 0;

    for ( j = 0, g = graphes->i_1 ; j < graphes->nbr_spectres ; j++, g++ ) {
      if ( *g != NULL ) {
	if ( (*g)->x != graphes->x ) {
	  memcpy( (*g)->x, graphes->x, sizeof( double ) * (*g)->nombre_point );
	}
	GRF_couleur( *g, couleur, style, marque );
	TRC_CouleurSuivante( &couleur, &style, &marque );
	GRF_ChangerUnite( *g, AXE_X, "eV" );
      }
    }
  }
  if ( graphes->spectres != NULL ) {
    unsigned int couleur = NOIR, style = BI_CONTINU, marque = 0;

    for ( j = 0, g = graphes->spectres ; j < graphes->nbr_spectres ; j++, g++ ) {
      if ( *g != NULL ) {
	if ( (*g)->x != graphes->x ) {
	  memcpy( (*g)->x, graphes->x, sizeof( double ) * (*g)->nombre_point );
	}
	GRF_couleur( *g, couleur, style, marque );
	TRC_CouleurSuivante( &couleur, &style, &marque );
	GRF_ChangerUnite( *g, AXE_X, "eV" );
      }
    }
  }

  /* 3. Tri ventuel [suppos peu important : routine peu efficace] */
  if ( infos->post_traitement & DETECT_POST_TRIER ) {
    GRF_Trier( graphes->i_0, GRF_TRI_RAPIDE );
    if ( graphes->i_1 != NULL ) {
      for ( j = 0, g = graphes->i_1 ; j < graphes->nbr_spectres ; j++, g++ ) {
	GRF_Trier( *g, GRF_TRI_RAPIDE );
      }
    }
    if ( graphes->spectres != NULL ) {
      for (  j = 0, g = graphes->spectres ; j < graphes->nbr_spectres ; j++, g++ ) {
	GRF_Trier( *g, GRF_TRI_RAPIDE );
      }
    }
  }

  /* 4. limination ventuelle des doublons */
  if ( infos->post_traitement & DETECT_POST_DOUBLONS ) {
    unsigned long nbr_doublons;
    int message = FAUX; /* VRAI si un message indiquant la prsence de doublons a dj t affich */

    nbr_doublons = GRF_TraiterDoublons( graphes->i_0, GRF_DOUBLONS_SIMPLE, 0.0, VRAI, VRAI );
    if ( nbr_doublons > 0 ) {
      message = VRAI;
      (void) BI_ERR_alerte( NULL, &MSG_SpectreAvecDoublons, VRAI, nbr_doublons, nom_fichier );
    }
    if ( graphes->i_1 != NULL ) {
      for ( j = 0, g = graphes->i_1 ; j < graphes->nbr_spectres ; j++, g++ ) {
	nbr_doublons = GRF_TraiterDoublons( *g, GRF_DOUBLONS_SIMPLE, 0.0, VRAI, VRAI );
	if ( ( FAUX == message ) && ( nbr_doublons > 0 ) ) {
	  message = VRAI;
	  (void) BI_ERR_alerte( NULL, &MSG_SpectreAvecDoublons, VRAI, nbr_doublons, nom_fichier );
	}
      }
    }
    if ( graphes->spectres != NULL ) {
      for (  j = 0, g = graphes->spectres ; j < graphes->nbr_spectres ; j++, g++ ) {
	nbr_doublons = GRF_TraiterDoublons( *g, GRF_DOUBLONS_SIMPLE, 0.0, VRAI, VRAI );
	if ( ( FAUX == message ) && ( nbr_doublons > 0 ) ) {
	  message = VRAI;
	  (void) BI_ERR_alerte( NULL, &MSG_SpectreAvecDoublons, VRAI, nbr_doublons, nom_fichier );
	}
      }
    }
  }

  /* 5. Calcul des bornes du spectre */
  GRF_bornes( graphes->i_0, DESORDONNE, DESORDONNE );
  if ( graphes->i_1 != NULL ) {
    for ( j = 0, g = graphes->i_1 ; j < graphes->nbr_spectres ; j++, g++ ) {
      GRF_bornes( *g, DESORDONNE, DESORDONNE );
    }
  }
  if ( graphes->spectres != NULL ) {
    for (  j = 0, g = graphes->spectres ; j < graphes->nbr_spectres ; j++, g++ ) {
      GRF_bornes( *g, DESORDONNE, DESORDONNE );
    }
  }

  /* 6. On normalise les spectres ainsi obtenus
     Normalisation brutale min = 0, max = 1 : pour la normalisation propre,
     mieux vaut utiliser l'interface graphique et reproduire sur la srie ! */
  if ( infos->post_traitement & DETECT_POST_NORMALISER ) {
    GRF_Normaliser( graphes->i_0, GRF_NORM_MIN_MAX, 0.0, 1.0 );
    if ( graphes->i_1 != NULL ) {
      for ( j = 0, g = graphes->i_1 ; j < graphes->nbr_spectres ; j++, g++ ) {
	GRF_Normaliser( *g, GRF_NORM_MIN_MAX, 0.0, 1.0 );
      }
    }
    if ( graphes->spectres != NULL ) {
      for (  j = 0, g = graphes->spectres ; j < graphes->nbr_spectres ; j++, g++ ) {
	GRF_Normaliser( *g, GRF_NORM_MIN_MAX, 0.0, 1.0 );
      }
    }
  }

  /* Termin, tout s'est bien pass */
  return 0;
}

/* -------- Compte combien de spectres il faut charger dans le fichier... -------- */

static unsigned long EXTRAIT_CompterSpectres(FILE *fichier, int commentaire, int ligne, unsigned int col_x)
{
  int c;
  unsigned long nbr_spectres = 0;
  long debut;

  /* On mmorise le dbut du fichier pour tre tranquille */
  debut = ftell( fichier );

  switch ( ligne )
  {
  case LIGNE_LURE_AQUVISU: /* Forcment un seul spectre, dans l'tat actuel */
    return 1;

  case LIGNE_LURE_D21: /* Rien ne signale : on utilise X(i + 1) < X(i) */
    nbr_spectres = EXTRAIT_ComptageXInverses( fichier, col_x, 0);
    break;

  case LIGNE_HASYLAB_DORIS: /* Les spectres commencent par %d : on compte le nombre de %d ! */
    c = fgetc( fichier );
    while ( c != EOF ) {
      if ( EOF == c ) break;
      if ( '%' == c ) {
	c = fgetc( fichier );
	if ( EOF == c ) break;
	if ( 'd' == c )
	  nbr_spectres++;
      }
      c = ignore_ligne( fichier );
    }
    break;

  case LIGNE_NSLS_MXANES: /* Format trs bizarre, mais probablement un seul spectre par fichier */
    return 1;

  case LIGNE_ESRF_SNBL: /* Compter les spectres vrais, mais aussi X(i+1) < X(i) ! */
    nbr_spectres = SNBL_CompterSpectres( fichier );
    break;

  case LIGNE_ESRF_ID21:
    nbr_spectres = ID21_CompterSpectres( fichier, "z3escan" );
    break;

  case LIGNE_ESRF_ID22:
    nbr_spectres = ID22_CompterSpectres( fichier, "energy" );
    break;

  case LIGNE_BESSY2_: /* Format HDF : on suppose un seul spectre, mais pas vident ! */
    return 1;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : type de ligne inconnu [%d][commentaire %c]\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne, commentaire);
  }

  /* On se remet au dbut */
  fseek( fichier, debut, SEEK_SET);
  return nbr_spectres;
}

/* -------- Compter le nombre de fois o X(i) > X(i + 1) --------
   Remarque : si interruption par des commentaires, on arrte la recherche...
 */

static unsigned long EXTRAIT_ComptageXInverses(FILE *fichier, unsigned int col_x, int commentaire)
{
  unsigned int nbr_inversions = 0;
  unsigned int j = 1;
  int c, n_lus;
  double X_ref, X;

  do {
    /* On commence un nouveau spectre */
    nbr_inversions++;

    /* On lit le X pour la premire ligne */
    j = 0;
    while ( j < col_x ) {
      (void) fscanf( fichier, "%*s" );
      j++;
    }
    n_lus = fscanf( fichier, "%lf", &X );
    if ( n_lus != 1 ) {
      (void) fprintf(stderr, "EXTRAIT_ComptageXInverses : impossible de lire X ref / bloc %u\n", nbr_inversions);
      return nbr_inversions - 1;
    }
    c = ignore_ligne( fichier );
    if ( EOF == c )
      return nbr_inversions;

    /* L, on lit les lignes ultrieures */
    do {
      X_ref = X;
      j = 0;
      while ( j < col_x ) {
	(void) fscanf( fichier, "%*s" );
	j++;
      }
      n_lus = fscanf( fichier, "%lf", &X );
      if ( n_lus != 1 ) {
	(void) fprintf(stderr, "EXTRAIT_ComptageXInverses : impossible de lire X / bloc %u\n", nbr_inversions);
	return nbr_inversions;
      }
      c = ignore_ligne( fichier );
      if ( EOF == c )
	return nbr_inversions;
    } while ( X > X_ref );

#ifdef _DEVELOPPE_
    printf("EXTRAIT_ComptageXInverses : inversion [X ref = %g, X = %g]\n", X_ref, X);
#endif
  } while ( c != commentaire );

  return nbr_inversions;
}
