/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1999, Emmanuel CURIS
   Ce fichier : cration le 12 janvier 1999

   Cration d'un modle thorique pour FEFF

   Gestion des atomes du modle structural
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"
#include "feff.h"
#include "interface.h"

#include "feff_atomes.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean FEFF_AjouterTousAtomes(Modele_FEFF *modele, Molecule *molecule);
static int     propose_potentiel     (Modele_FEFF *modele, int numero_atomique);

static void prepare_arbre_ajouter(void); /* Arbre pour ajouter plusieurs atomes au modle d'un coup */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -~-~-~-~-~-~-~-~ Ajouter des atomes au modle -~-~-~-~-~-~-~-~ */

/* -------- Ajouter un seul atome -------- */

FEFF_atome *FEFF_AjouterAtome(Modele_FEFF *modele, Atome *atome, int potentiel, boolean alerte)
{
  FEFF_atome *atome_feff;

  if (NULL == modele) return NULL;
  if (NULL == atome) return NULL;

  if ( ATM_EstDansModeleFEFF(modele, atome) == VRAI ) {
    if (VRAI == alerte) {
      (void) BI_ERR_alerte(NULL, &BI_ERR_DejaDansModele, VRAI);
    }
    return NULL;
  }

  atome_feff = (FEFF_atome *) malloc( sizeof(FEFF_atome) );
  if ( NULL == atome_feff ) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "", "", "atome_feff");
    return NULL;
  }

  if ( (potentiel <= 0) || (potentiel > 8) ) {
    potentiel = propose_potentiel(modele, atome->num_atomique);
    if (potentiel < 1) {
      free(atome_feff);
      return NULL;
    }
  }

  /* On initialise */
  atome_feff->atome     = atome;
  atome_feff->potentiel = potentiel;
  atome_feff->suivant   = NULL;

  if (NULL == modele->parametres.prem_atome) { /* Premier atome */
    modele->parametres.prem_atome = atome_feff;
    modele->parametres.nbr_atomes = 1;
  } else {
    double d_ref, i, d; 
    FEFF_atome *avant = modele->parametres.prem_atome;

    /* Distance  l'absorbeur [au carr] */
    i  = (atome->position.x - modele->absorbeur->position.x);
    d_ref  = i * i;
    i  = (atome->position.y - modele->absorbeur->position.y);
    d_ref += i * i;
    i  = (atome->position.z - modele->absorbeur->position.z);
    d_ref += i * i;

    /* Distance du plus proche actuel */
    i  = (avant->atome->position.x - modele->absorbeur->position.x);
    d  = i * i;
    i  = (avant->atome->position.y - modele->absorbeur->position.y);
    d += i * i;
    i  = (avant->atome->position.z - modele->absorbeur->position.z);
    d += i * i;

    if ( d > d_ref ) { /* On est plus prs que le premier */
      atome_feff->suivant = avant;
      modele->parametres.prem_atome = atome_feff;
    } else {
      FEFF_atome *apres = avant;

      /* On cherche l'atome immdiatement avant : tri par distance croissante  l'absorbeur */
      do {
	avant = apres;
	apres = avant->suivant;
	if (NULL == apres) break; /* On est aprs le dernier */

	i  = (apres->atome->position.x - modele->absorbeur->position.x);
	d  = i * i;
	i  = (apres->atome->position.y - modele->absorbeur->position.y);
	d += i * i;
	i  = (apres->atome->position.z - modele->absorbeur->position.z);
	d += i * i;

      } while ( d < d_ref );

      /* On insre dans le chanage */
      atome_feff->suivant = avant->suivant;
      avant->suivant = atome_feff;
    }

    modele->parametres.nbr_atomes++;
  }

  atome->nbr_modeles++; /* Il appartient  un modle de plus ! */

  return atome_feff;
}

/* -------- Ajouter tous les atomes d'une molcule -------- */

static boolean FEFF_AjouterTousAtomes(Modele_FEFF *modele, Molecule *molecule)
{
  Atome *atome;

  if (NULL == modele) return FAUX;
  if (NULL == molecule) return FAUX;

  for (atome = molecule->premier ; atome != NULL ; atome = atome->suivant) {
    if (atome == modele->absorbeur) continue;

    FEFF_AjouterAtome(modele, atome, -1, FAUX); /* Pas d'affichage si dj dans le modle */
  }

  return VRAI;
}

/* -------- Ajouter tous les voisins suivant certains critres -------- */

boolean FEFF_AjouterVoisins(Modele_FEFF *modele)
{
  Molecule *molecule;

  if (NULL == modele) return FAUX;
  if (NULL == modele->absorbeur) return FAUX;

  /* Contrle pralable */
  if (modele->parametres.prem_atome != NULL) { /* Dj des atomes -> que faire ? */
    int retour;

    retour = BI_ERR_alerte(NULL, &BI_Q_DejaAtomes, VRAI);
    if (retour == 3) /* On annule l'opration */
      return FAUX;

    if (retour == 1) { /* On efface les atomes dj existants */
      while (modele->parametres.prem_atome != NULL) { /* Tant qu'il y a un atome */
	FEFF_EffacerAtome(modele, modele->parametres.prem_atome, VRAI);
      }
    }
  }

  molecule = ATM_QuelleMolecule(modele->absorbeur);
  if (NULL == molecule) {
    (void) fprintf(stderr, "FEFF_AjouterVoisins : absorbeur du modle sans molcule !\n");
    return FAUX;
  }

  /* Interface : quel filtre utiliser ? */
  if (arbres[ AJ_TT_ATOMES ].drapeaux.prepare == FAUX)
    prepare_arbre_ajouter();

  if ( ARB_Bloquant( AJ_TT_ATOMES, titre_arbre_ajouter, principale, 0 ) == ATA_ANNULER)
    return 0;

  if ( OBJ_choisit(AJ_TT_ATOMES, ATA_TOUS, QUESTION) == VRAI ) {
    if ( FAUX == FEFF_AjouterTousAtomes(modele, molecule) )
      return FAUX;
  } else { /* Filtre sur la distance ou Z */
    Atome *atome;
    double distance, d_reel;
    int Z_ref;

    distance  = OBJ_LireNombre( AJ_TT_ATOMES, ATA_DIST_MAX, NULL );
    Z_ref     = OBJ_NumeroAtomique( AJ_TT_ATOMES, ATA_ELEMENT );

    for ( atome = molecule->premier ; atome != NULL ; atome = atome->suivant ) {
      boolean prendre = FAUX; /* VRAI si on doit ajouter l'atome au modle */

      /* Premier tri : critres lis  l'atome lui-mme */
      if ( VRAI == OBJ_choisit( AJ_TT_ATOMES, ATA_PROCHES, QUESTION ) ) { /* Seulement les atomes proches */
	d_reel = ATM_distance( atome, modele->absorbeur );
	if ( d_reel <= distance ) /* On est plus prs : on accepte l'atome */
	  prendre = VRAI;
      } else {
	prendre = VRAI;
      }

      if ( OBJ_choisit( AJ_TT_ATOMES, ATA_MEME_Z, QUESTION ) == VRAI ) { /* Ceux d'un Z donn */
	if (atome->num_atomique == Z_ref)
	  prendre &= VRAI;
      }

      if ( VRAI == prendre ) {
	FEFF_atome *atome_feff;

#ifdef _DEVELOPPE_
	d_reel = ATM_distance( atome, modele->absorbeur );
	printf("Atome slectionn : %s en (%f, %f, %f) a %g de l'absorbeur\n", ATM_donnees[atome->num_atomique - 1].nom,
	       atome->position.x, atome->position.y, atome->position.z, d_reel);
#endif
	atome_feff = FEFF_AjouterAtome(modele, atome, -1, FAUX); /* Pas d'affichage si dj dans le modle */
	if ( atome_feff != NULL ) { /* On a bien pu l'ajouter */
	  Fragment *fragment = atome->fragment;

	  if ( ( OBJ_choisit(AJ_TT_ATOMES, ATA_TOUT_FRAG, QUESTION) == VRAI ) && /* On veut tout le fragment */
	       ( fragment != NULL ) /* et il y en a bien un */ ) {
	    Atome *atome;
	    unsigned int i;

	    for (i = 0 ; i < fragment->nbr_atomes ; i++) {
	      atome = fragment->atomes[i];

	      FEFF_AjouterAtome(modele, atome, -1, FAUX);
	    }
	  }
	}
      }
    }
  }

  return VRAI;
}

/* -~-~-~-~-~-~-~-~ Un atome appartient-il  un modle FEFF ? -~-~-~-~-~-~-~-~ */

boolean ATM_EstDansModeleFEFF(Modele_FEFF *modele, Atome *atome)
{
  FEFF_atome *atome_feff;

  if (NULL == modele) return FAUX;
  if (NULL == atome) return FAUX;

  /* C'est l'absorbeur ! */
  if (modele->absorbeur == atome) return VRAI;

  for (atome_feff = modele->parametres.prem_atome ; atome_feff != NULL ; atome_feff = atome_feff->suivant) {
    if (atome_feff->atome == atome)
      return VRAI;
  }

  return FAUX;
}

/* -------- Proposition d'un potentiel FEFF pour un atome de Z donn --\
 |   Renvoie le numro du potentiel propos                            |
 |      (-1 en cas d'erreur ou si aucun potentiel n'est libre)         |
 |   La routine ne propose _jamais_ le potentiel 0 (absorbeur)         |
 |                                                                     |
 | num_atomique = le Z pour lequel chercher un potentiel               |
 \-------------------------------------------------------------------- */

static int propose_potentiel(Modele_FEFF *modele, int num_atomique)
{
  int potentiel, plus_proche, Z_plus_proche = -1; /* Potentiel propos */

  if ( (num_atomique < HYDROGENE) || (num_atomique > PLUTONIUM) ) /* Z absurde... */
    return -1;

  /* On commence par vrifier qu'il n'existe pas dj un potentiel
     associ  ce numro atomique */

  for (potentiel = 1 ; potentiel < 8 ; potentiel++) {
    if ( modele->potentiels[potentiel] == num_atomique)
      return potentiel;
  }

  /* Si on est l, c'est que l'on n'a pas trouv le potentiel -> il faut en crer un nouveau */

  /* Le premier convient-il ? */
  if ( 0 ==  modele->potentiels[ 1 ] ) {
    modele->potentiels[ 1 ] = num_atomique;
    return 1;
  }
  Z_plus_proche = modele->potentiels[ 1 ];
  plus_proche = 1;

  for (potentiel = 2 ; potentiel < 8 ; potentiel++) {
    if ( modele->potentiels[potentiel] == 0) { /* Pas de potentiel associ */
      modele->potentiels[potentiel] = num_atomique;
      return potentiel; /* On a pu crer un potentiel, tout va bien :-) */
    } else {
      if ( abs( num_atomique - modele->potentiels[potentiel]) < abs( num_atomique - Z_plus_proche ) ) {
	Z_plus_proche = modele->potentiels[ potentiel ];
	plus_proche = potentiel;
      }
    }
  }

  /* Si on est l, c'est qu'il a t impossible de crer un potentiel
     Les 7 potentiels FEFF sont occups, il faudra ruser -> c'est le boulot de l'utilisateur !
     On propose en attendant d'utiliser le potentiel le plus proche */

  if (2 == BI_ERR_alerte(NULL, &BI_Q_AucunPotentielLibre, VRAI,
			 ATM_Z_vers_nom( num_atomique ), plus_proche, ATM_Z_vers_nom( Z_plus_proche )))
    return -1;

  return plus_proche;
}

/* ~~~~~~~~ Interface graphique ~~~~~~~~ */

static void prepare_arbre_ajouter(void) /* Arbre pour ajouter plusieurs atomes au modle d'un coup */
{
  OBJ_deplace(AJ_TT_ATOMES, AA_TXT_DISTANCE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(AJ_TT_ATOMES, ATA_DIST_MAX, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace(AJ_TT_ATOMES, AA_TXT_ELEMENT, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(AJ_TT_ATOMES, ATA_ELEMENT, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(AJ_TT_ATOMES, EG_TXT_POTENTIEL, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(AJ_TT_ATOMES, ATA_POTENTIEL, 0, +1, DEPLACE_PIXELS);

  arbres[AJ_TT_ATOMES].drapeaux.prepare = VRAI;
}

/* -~-~-~-~-~-~-~-~ Enlever des atomes du modle -~-~-~-~-~-~-~-~ */

/* -------- Enlve un atome 3D du modle FEFF ------------------- */

void FEFF_EnleverAtome(Modele_FEFF *modele, Atome *atome, boolean alerte)
{
  FEFF_atome *atm_feff;

  if (NULL == modele) return;
  if (NULL == atome) return;

  /* C'est l'absorbeur ! */
  if (modele->absorbeur == atome) {
    (void) fprintf(stderr, "FEFF_EnleverAtome : on veut enlever l'absorbeur !\n");
    return;
  }

  for (atm_feff = modele->parametres.prem_atome ; atm_feff->atome != atome ; atm_feff = atm_feff->suivant) {
    if (NULL == atm_feff) {
      (void) fprintf(stderr, "FEFF_EnleverAtome : l'atome n'appartient pas aux atomes du modle FEFF\n");
      return;
    }
  }

  /* On efface l'atome FEFF correspondant */
  FEFF_EffacerAtome(modele, atm_feff, !alerte);
}

/* -------- Dtruire l'atome FEFF indiqu -------- */

void FEFF_EffacerAtome(Modele_FEFF *modele, FEFF_atome *atome, boolean forcer)
{
  if (NULL == modele) return;
  if (NULL == atome) return;

  /* 0. ventuellement, on demande confirmation */
  if ( FAUX == forcer ) {
    int reponse = 1;

    fprintf(stderr, "FEFF_EffacerAtome : il faudrait demander confirmation,  FAIRE\n");
    if ( 2 == reponse ) return; /* Effacement annul */
  }

  /* 1. On s'occupe du chanage */
  if ( atome == modele->parametres.prem_atome ) {
    modele->parametres.prem_atome = atome->suivant;
  } else {
    FEFF_atome *avant;

    for ( avant = modele->parametres.prem_atome ; avant->suivant != atome ; avant = avant->suivant) {
      if ( avant == NULL ) {
	(void) fprintf(stderr, "FEFF_EffacerAtome : l'atome FEFF n'est pas dans le modle indiqu !\n");
	return;
      }
    }
    avant->suivant = atome->suivant;
  }

  /* L'atome 3D associ appartient  un modle de moins */
  atome->atome->nbr_modeles--;

  /* On libre la mmoire */
  free(atome);
  modele->parametres.nbr_atomes--;
}
