/* LASE : Logiciel d'Analyse des Spectres Exprimentaux
   (c) Emmanuel CURIS 1998-2002

   Interface FEFF
   Modles d'ajustement par chemins, prvus pour FEFF

   Vrification de la cohrence du modle FEFF propos
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "atomes.h"
#include "feff.h"

#include "feff_verifier.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int FEFF_BoucheTrouPotentiel(Modele_FEFF *modele);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

boolean FEFF_VerifierModele(Modele_FEFF *modele)
{
  boolean pot_utilise[8] = {FAUX, FAUX, FAUX, FAUX, FAUX, FAUX, FAUX, FAUX };
  int  reponse;
  unsigned int i;
  FEFF_atome *atome;
  FEFF_SS *couche;

  if ( NULL == modele ) {
    modele = FEFF_ChoisirModele(FAUX); /* s'il y en a un, on le vrifie sans demander */
    if (NULL == modele)
      return FAUX;
  }

  /* 1. On vrifie qu'il existe un absorbeur, correctement caractris */
  if ( NULL == modele->absorbeur ) {
    (void) BI_ERR_alerte(NULL, &ERR_AucunAbsorbeur, VRAI);
    return FAUX;
  }

  if ( 0 == modele->potentiels[0] ) { /* On force celui de l'absorbeur */
    modele->potentiels[0] = modele->absorbeur->num_atomique;
  }

  if ( (modele->potentiels[0] < HYDROGENE) || (modele->potentiels[0] > PLUTONIUM) ) {
    (void) BI_ERR_alerte(NULL, &ERR_AbsorbeurAbsurde, VRAI, modele->potentiels[0]);
    (void) fprintf(stderr, "FEFF_VerifierModele : Z absurde pour l'absorbeur => demander le Z !\n");
    return FAUX;
  }

  if ( modele->potentiels[0] != modele->absorbeur->num_atomique ) {
    int reponse, Z;

    Z = modele->absorbeur->num_atomique;
    reponse = BI_ERR_alerte(NULL, &ERR_ConflitZAbsorbeur, VRAI, ATM_Z_vers_nom(Z), Z,
			    ATM_Z_vers_nom(modele->potentiels[0]), modele->potentiels[0]);
    if (1 == reponse) {
      modele->potentiels[0] = Z;
    }
  }

  /* 2. Contrle de la dfinition des potentiels */
  /*  a. Quels potentiels sont utiliss ? */
  for ( atome = modele->parametres.prem_atome ; atome != NULL ; atome = atome->suivant) {
    pot_utilise[ atome->potentiel - 1 ] = VRAI;
  }
  for ( couche = modele->parametres.prem_ss ; couche != NULL ; couche = couche->suivante) {
    pot_utilise[ couche->potentiel - 1 ] = VRAI;
  }

  /*  b. Ces potentiels sont-ils dfinis ? */
  for (i = 0 ; i < 8 ; i++) {
    if ( (VRAI == pot_utilise[i]) && ( 0 == modele->potentiels[i + 1] ) ) {
      (void) BI_ERR_alerte(NULL, &ERR_PotentielIndefini, VRAI, i + 1);
      (void) fprintf(stderr, "FEFF_VerifierModele : potentiel utilis, indfini. Proposer correction automatique / Z atome de ce potentiel\n");
      return FAUX;
    }

    if ( (FAUX == pot_utilise[i]) && ( 0 < modele->potentiels[i + 1] ) ) {
      int reponse;

      reponse = BI_ERR_alerte(NULL, &ERR_PotentielSansAtome, VRAI, i + 1, modele->potentiels[i + 1],
			      ATM_Z_vers_symbole(modele->potentiels[i + 1]));
      if (1 == reponse) {
	modele->potentiels[i + 1] = 0;
      } else {
	return FAUX;
      }
    }
    (void) fprintf(stderr, "FEFF_VerifierModele : il faudrait vrifier qu'un potentiel n'est pas utilis que dans une SS : il faudrait alors le faire sauter aussi...\n");
  }

  /* c. Les potentiels sont-ils contigus ? */
  while ( (reponse = FEFF_BoucheTrouPotentiel(modele)) == VRAI ); /* Tant qu'il y a des trous... */
  if ( -1 == reponse ) return FAUX; /* On a annul la vrification */

  /* 3. Cohrence potentiels / atomes */
  fprintf(stderr, "FEFF_VerifierModele :  FAIRE, contrle de cohrence potentiels/atomes\n");

  /* 4. Cohrence des distances entre atomes */
  fprintf(stderr, "FEFF_VerifierModele :  FAIRE, contrle de distance entre atomes (seuil FEFF : 1,75 rydberg)\n");

  /* Si on arrive l, aucune incohrence n'a t dtecte */
  (void) BI_ERR_alerte(NULL, &MSG_AucunProbleme , VRAI);
  return VRAI;
}

/* -------- Il y a des trous dans les potentiels ! -------- */

static int FEFF_BoucheTrouPotentiel(Modele_FEFF *modele)
{
  int i, prem_libre = 0;

  /* Quel est le premier potentiel inutilis ? */
  for (i = 1 ; i <= 8 ; i++) {
    if ( 0 == modele->potentiels[i] ) {
      prem_libre = i;
      break;
    }
  }
  if ( 0 == prem_libre ) { /* Tous les potentiels sont dfinis, pas de trou */
    return FAUX; /* Aucun trou  boucher */
  }

  /* Y a-t-il des potentiels dfinis aprs ? */
  for ( ; i <= 8 ; i++) {
    if ( 0 < modele->potentiels[i] ) {
      int reponse;

      reponse = BI_ERR_alerte(NULL, &ERR_TrouPotentiel, VRAI, i,
			      modele->potentiels[i], ATM_Z_vers_symbole(modele->potentiels[i]), prem_libre);
      if (3 == reponse) /* On arrte la vrification */
	return -1;
      if (2 == reponse) { /* On dplace le potentiel */
	FEFF_atome *atome;
	FEFF_SS *couche;

	/* Dfinition des potentiels */
	modele->potentiels[prem_libre] = modele->potentiels[i];

	/* Association aux atomes */
	for ( atome = modele->parametres.prem_atome ; atome != NULL ; atome = atome->suivant ) {
	  if ( atome->potentiel == i ) {
	    atome->potentiel = prem_libre;
	  }
	}

	/* Association aux couches */
	for ( couche = modele->parametres.prem_ss ; couche != NULL ; couche = couche->suivante ) {
	  if ( couche->potentiel == i) {
	    couche->potentiel = prem_libre;
	  }
	}

	/* Association aux commandes */
	modele->parametres.ION[prem_libre] = modele->parametres.ION[i];
	modele->parametres.FOLP[prem_libre] = modele->parametres.FOLP[i];

	(void) fprintf(stderr, "FEFF_BoucheTrouPotentiel : il faudrait s'occuper des OVERLAP\n");
      }
      if (1 == reponse) { /* On efface le potentiel */
	FEFF_atome *atome;
	FEFF_SS *couche;
	boolean trouve;

	/* Association aux atomes */
	do {
	  trouve = FAUX;
	  for ( atome = modele->parametres.prem_atome ; atome != NULL ; atome = atome->suivant ) {
	    if ( atome->potentiel == i ) {
	      FEFF_EffacerAtome(modele, atome, VRAI);
	      trouve = VRAI;
	      break;
	    }
	  }
	} while (VRAI == trouve);

	/* Association aux couches */
	do {
	  trouve = FAUX;
	  for ( couche = modele->parametres.prem_ss ; couche != NULL ; couche = couche->suivante ) {
	    if ( couche->potentiel == i) {
	      (void) FEFF_EffacerCouche(modele, couche, VRAI);
	      trouve = VRAI;
	      break;
	    }
	  }
	} while (VRAI == trouve);
      }

      modele->potentiels[i] = 0;
      return VRAI; /* On a trait un trou... */
    }
  }

  /* On n'a trouv aucun trou */
  return FAUX;
}
