/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Création de ce fichier : 23 juin 1998

   Module de filtrage des spectres (essentiellement par transformée de Fourier)

   Initialisation du calcul de la transformée de Fourier inverse
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"
#include "filtrage.h"
#include "lissage.h"
#include "science.h"
#include "utils_tables.h"

#include "prep_tf.txt"

/* -------- Extraction de la partie intéressante du signal, suivant la fenêtre --------
   Valeur renvoyée : - soit le graphe original si toutes ses valeurs sont utiles ;
                     - soit un sous-graphe ISOLÉ comprenant sa région utile
                     - NULL en cas d'erreur
 */

Graphe *TF_IsoleSignal(Graphe *original, struct infos_fenetre *filtre)
{
  unsigned long nbr_pts, premier;

  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'original !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if ( NULL == filtre ) return original; /* Pas de filtre... */

  if ( PAS_DE_FENETRE == filtre->type ) return original; /* Pas de filtre */
  if ( FEN_BESSEL     == filtre->type ) return original; /* Filtre jamais nul */
  if ( FEN_GAUSSIENNE == filtre->type ) return original; /* Filtre jamais nul */
  if ( FEN_NORTON_F3  == filtre->type ) return original; /* Filtre jamais nul */

  /* Si on est là, le filtre utilisé est nul en-dehors de (min, max)
     Suppose le graphe trié par X croissants */
/*  printf("Filtre conserve la région : %f - %f\n", filtre->x_min, filtre->x_max); */
  nbr_pts = GRF_intervalle(original, filtre->x_min, filtre->x_max, &premier);
  if (nbr_pts < 1) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun point dans l'intervalle de filtrage choisi\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  return GRF_ExtraitGraphe(original, premier, nbr_pts, VRAI);
}

/* ~~~~~~~~ Préparation dans le cas d'un calcul par algorithme rapide simple ~~~~~~~~~~
   Méthode : utilise l'intervalle demandé pour le spectre d'arrivé et le pas pour connaître le nombre de points
             calcule alors le R max correspondant
	     lisse et extrapole à zéro le spectre dont on veut la TF ensuite           */

/* -------- Préparation pour la TF rapide (directe ou inverse) -------------------------\
 |   Renvoie le graphe créé pour l'occasion                                             |
 |                                                                                      |
 | original     : le graphe que l'on veut transformer                                   |
 | r_min, r_max : l'intervalle intéressant dans l'espace du graphe après transformation |
 | pas          : le pas à utiliser dans l'espace après transformation                  |
 | sens         : préparation pour la TF directe ou inverse ?                           |
 \------------------------------------------------------------------------------------- */

Graphe *TF_prep_rapide(Graphe *original, double r_min, double r_max, double pas, int sens)
{
  unsigned long i, nbr_pts; /* Nombre de points de la TF */
  unsigned int bit_1 = 0;
  const char *titre;
  double coef __attribute__((__unused__));
  double pas_k;
  double *nv_x, *x;
  int type_graphe;
  int retour;
  Complexe *a, *b, *c; /* Pour la spline cubique */
  Complexe *nv_y, *y;
  Graphe *resultat;

  /* Contrôles */
  if ( NULL == original ) return NULL;
  if (pas <= 0.0) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas en R absurde [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, pas);
    EXF_ERREUR(PAS_FAUX);
    return NULL;
  }

  /* Calcul du pas en k nécessaire */
  pas_k = original->x_max - original->x_min; /* Le pas en k (grossier) */
  pas_k /= (double) (original->nombre_point - 1);
  if ( pas_k <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas en k absurde [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, pas_k);
    EXF_ERREUR(DIV_ZERO, NULL, NULL);
    return NULL;
  }

  nbr_pts = (unsigned long) ( M_PI / (pas * pas_k) );
  bit_1 = 0;
  while ( (1U << bit_1) < nbr_pts)
    bit_1++;
  nbr_pts = 1U << bit_1; /* On arrondi à la valeur supérieure */
  /* Rq : variante dans TFI_prep_rapide, paraît plus tordue que celle-ci... */

  pas_k = M_PI / ( pas * (double) nbr_pts ); /* Le pas à utiliser dans l'espace de l'original */

  /* On convertit l'original en graphe complexe */
  retour = GRF_ChangerGenre( original, GRF_FCT_COMPLEXE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : La conversion en complexes a échoué [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return NULL;
  }

  /* Réserver pour les nouvelles valeurs */
  /* a. Table des X */
  nv_x = TABLE_CreerTable_mNp( 0.0, nbr_pts, pas_k );
  if (nv_x == NULL) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table X de %lu points, pas de %g\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_pts, pas_k);
    SYS_ERREUR(MANQUE_MEMOIRE, "prep_tf.c", "TF_prep_rapide", "nv_x");
    return NULL;
  }

  /* b. Table des Y */
  nv_y = (Complexe *) calloc( nbr_pts, sizeof(Complexe) );
  if (nv_y == NULL) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table Y[Complexe] de %lu points\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_pts );
    free(nv_x); nv_x = NULL;
    SYS_ERREUR(MANQUE_MEMOIRE, "prep_tf.c", "TF_prep_rapide", "nv_y");
    return NULL;
  }

  /* Préparation */
  coef = pas_k * sqrt( 2.0 / M_PI ); /* Coefficient de normalisation */
  x = nv_x;
  y = nv_y;

  /* Calcul des tables de splines */
  retour = CSPL_tt_prep(original, &a, &b, &c);
  if ( retour  != 0) { /* Erreur dans le calcul... */
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de calculer les tables de splines\n",
		   __FILE__, __FUNCTION__, __LINE__);
    LIS_ERREUR(retour, original->nom);
    free(nv_y); nv_y = NULL;
    free(nv_x); nv_x = NULL;
    return NULL;
  }

  /* Calcul des valeurs interpolées */
  for (i = 0, x = nv_x, y = nv_y ; i < nbr_pts ; i++, x++, y++) {
    if ( ( *x < original->x_min ) || ( *x > original->x_max ) ) {
      y->re = 0.0;
      y->im = 0.0;
    } else {
      *y = CSPL_tt_calc(original, *x, a, b, c);
    }
  }

  free(a); free(b); free(c); /* Libération des tableaux de la spline, inutiles... */

  /* On remplace par les nouvelles valeurs */
  /* a. On efface les X existants */
  free( original->x ); original->x = NULL; /* On libère les anciens points */

  /* b. On efface les Y */
  free( original->y ); original->y = NULL;

  /* On installe les nouveaux X & Y */
  original->x = nv_x;
  original->y = (double *) nv_y;
  original->nombre_point = nbr_pts;

  /* Les bornes ont changé à cause du calcul, en particulier les X... */
  GRF_bornes(original, ORDONNE, DESORDONNE);

  /* On prépare maintenant le graphe contenant la TF ... */
  /* a. Dans quel sens fait-on la TF ? */
  if ( TF_SENS_INVERSE == sens ) {
    type_graphe = GR_TFI;
    titre = titre_grf_tfi_temp;
  } else {
    type_graphe = GR_TF;
    titre = titre_grf_tf_temp;
  }

  /* b. Création du graphe résultat */
  resultat = GRF_nv_GrilleXN( NULL, type_graphe, GRF_FCT_COMPLEXE, titre , FAUX, 0.0, pas * (double) nbr_pts, nbr_pts );
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer le graphe résultat\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  resultat->drapeaux.protege = VRAI; /* Graphe en cours de calcul : on le protège... */

  y = (Complexe *) resultat->y;
  for (i = 0 ; i < resultat->nombre_point ; i++, y++) {
    y->re = 0.0; y->im = 0.0; /* On met à zéro la TF - juste pour le dessin... */
  }

  /* Contrôle de l'intervalle après transformation */
#ifdef _DEVELOPPE_
  if ( r_min < 0.0 ) {
    (void) fprintf(stderr, "TF_prep_rapide : valeur de r_min incorrecte (ignorée) [%g]\n", r_min);
  }
#endif
  if ( ( r_max <= 0.0 ) || ( r_max > resultat->x_max ) ) {
    (void) BI_ERR_alerte(NULL, &ERR_Rmax_incorrect, VRAI, r_max, resultat->x_max);
  }

  return resultat;
}

/* ~~~~~~~~ Préparation dans le cas d'un calcul par la méthode brutale ~~~~~~~~~~~~~~~~ */

/* -------------------- Préparation du graphe qui contiendra la TF --------\
 |  Renvoie le graphe désiré (NULL si erreur)                              |
 |                                                                         |
 | r_min, r_max : intervalle à utiliser pour le calcul                     |
 | pas          : espacement des X à utiliser pour le calcul               |
 | sens         : calcul de la TF normale ou inverse ?                     |
 | erreurs      : VRAI = estimer les erreurs en Y                          |
 | covariance   : VRAI = estimer en plus la matrice de covariance en Y     |
 \------------------------------------------------------------------------ */

Graphe *TF_prep_normal(double r_min, double r_max, double pas, int sens, boolean erreurs, boolean covariance)
{
  unsigned long i;
  register Complexe *y;
  Graphe *resultat;
  int type_graphe;
  const char *titre;

  /* Dans quel sens fait-on la TF ? */
  if ( TF_SENS_INVERSE == sens ) {
    type_graphe = GR_TFI;
    titre = titre_grf_tfi_temp;
  } else {
    type_graphe = GR_TF;
    titre = titre_grf_tf_temp;
  }

  /* Création du graphe résultat */
  resultat = GRF_nv_GrilleXP( NULL, type_graphe, GRF_FCT_COMPLEXE, titre , FAUX, r_min, r_max, pas );
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer le graphe résultat\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  resultat->drapeaux.protege = VRAI; /* Graphe en cours de calcul : on le protège... */
  resultat->y_min = -2.0; /* À l'usage, suffit le plus souvent */
  resultat->y_max =  2.0;

  y = (Complexe *) resultat->y;
  for (i = 0 ; i < resultat->nombre_point ; i++, y++) {
    y->re = 0.0; y->im = 0.0; /* On met à zéro la TF */
  }

  if ( VRAI == erreurs ) { /* On veut les erreurs à la fin */
    GRF_creer_erreurs( resultat, AXE_Y );
    if ( VRAI == covariance ) {
      GRF_CreerCovariance( resultat );
    }
  }

  return resultat; /* Le graphe qui contiendra la TF */
}
