/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Gestion générique des graphes
   Gestion des barres d'erreur (en X, en Y) d'un graphe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "serie.h"
#include "lissage.h" /* pour le recalage d'un graphe */
#include "matrices.h"

#include "grf_erreurs.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_InitialiseErreurs(Graphe *erreurs, const Graphe *reference);
#ifdef maintenant_inutile
static int GRF_ExtraitErreurs_R (Graphe *erreurs, const Graphe *source_erreurs);
static int GRF_ExtraitErreurs_C (Graphe *erreurs, const Graphe *source_erreurs);
#endif

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Création d'un graphe contenant les erreurs ~-~-~-~-~-~-~-~- */

/* -------- Créer le graphe contenant les barres d'erreur associées à un graphe --------\
 |    Renvoie les erreurs créées (sur les Y si on crée pour X et Y en même temps)       |
 |                                                                                      |
 |  Les erreurs sont dans des graphes de type spécial, sans X [implicites]              |
 |  Ce sont des graphes isolés : ils ne peuvent pas etre partagés par d'autres !        |
 \------------------------------------------------------------------------------------- */

Graphe *GRF_creer_erreurs(Graphe *graphe, int axe)
{
  Graphe *err_x = NULL, *err_y = NULL;

  if ( NULL == graphe ) return NULL;
  if ( 0 == axe ) return NULL;	/* On ne crée aucune barre d'erreur... */

  /* S'il y a déjà des erreurs, elles sont perdues ! */

  if ( AXE_X & axe ) { /* Barres d'erreur pour les X */
    if ( graphe->erreur_x != NULL ) { /* Anciennes erreurs, graphe isolé => on les efface sinon impossible à libérer après ! */
      OPTION_A_FAIRE( "Comprendre pourquoi il y a déjà des erreurs en X !" );
      GRF_eff_isole( graphe->erreur_x );
      graphe->erreur_x = NULL;
    }

    /* On crée le graphe isolé qui contiendra les erreurs */
    err_x = GRF_nv_isole( graphe->nombre_point, GR_AUTRETYPE, GRF_ERREUR_R );
    if ( err_x != NULL ) { /* Création a réussi */
      (void) GRF_InitialiseErreurs( err_x, graphe );
      graphe->erreur_x = err_x;
      graphe->drapeaux.erreur_x = VRAI;
    }
  }

  if ( AXE_Y & axe ) { /* Barres d'erreur pour les Y */
    if ( graphe->erreur_y != NULL ) { /* Anciennes erreurs, graphe isolé => on les efface sinon impossible à libérer après ! */
      GRF_eff_isole( graphe->erreur_y );
      graphe->erreur_y = NULL;
    }

    /* On crée le graphe isolé qui contiendra les erreurs */
    if ( GRF_EstComplexe( graphe ) ) {
      err_y = GRF_nv_isole( graphe->nombre_point, graphe->type, GRF_ERREUR_C );
    } else {
      err_y = GRF_nv_isole( graphe->nombre_point, graphe->type, GRF_ERREUR_R );
    }

    if ( err_y != NULL ) { /* Création a réussi */
      (void) GRF_InitialiseErreurs( err_y, graphe );

      graphe->erreur_y = err_y;
      graphe->drapeaux.erreur_y = VRAI;
    }
  }

  if ( err_y != NULL )
    return err_y;
  else
    return err_x;
}

/* -------- Transformer un graphe pour qu'il devienne une barre d'erreurs --------
     Renvoie 0    si tout s'est bien passé
             != 0 en cas de problème

   graphe  : le graphe auquel on veut associer les erreurs
   erreurs : le graphe contenant les erreurs. Il doit avoir le meme nombre de points que le graphe modifié et etre isolé
   axe     : à quel axe doit-on associer les erreurs ?
 */

int GRF_InsererErreurs(Graphe *graphe, Graphe *erreurs, int axe)
{
  int retour;

  /* Controles */
  if ( ( NULL == graphe ) || ( NULL == erreurs ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Au moins un graphe manquant\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }
  if ( erreurs->nombre_point != graphe->nombre_point ) {
    (void) fprintf(stderr, "%s,%s l.%u : Les 2 graphes doivent avoir le meme nombre de points\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* Les paramètres d'affichage sont inutiles : on les efface */
  (void) GRF_LibererAffichage( erreurs );

  /* Les bornes en X sont forcément celles du graphe original */
  erreurs->x_min = graphe->x_min;
  erreurs->x_max = graphe->x_max;

  /* S'il s'agit des erreurs sur les X */
  if ( AXE_X == axe ) {
    if ( graphe->erreur_x != NULL ) {
      (void) fprintf(stderr, "%s,%s l.%u : Le graphe contient déjà des erreurs en X\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return +5;
    }

    retour = GRF_ChangerGenre( erreurs, GRF_ERREUR_R );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de convertir les erreurs en graphe du bon genre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -5;
    }

    graphe->erreur_x = erreurs;
    graphe->drapeaux.erreur_x = VRAI;

    /* Assignation terminée sans incident */
    return 0;
  }

  /* S'il s'agit des erreurs sur les Y */
  if ( AXE_Y == axe ) {
    if ( graphe->erreur_y != NULL ) {
      (void) fprintf(stderr, "%s,%s l.%u : Le graphe contient déjà des erreurs en X\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return +5;
    }

    if ( GRF_EstComplexe( graphe ) ) {
      retour = GRF_ChangerGenre( erreurs, GRF_ERREUR_C );
    } else {
      retour = GRF_ChangerGenre( erreurs, GRF_ERREUR_R );
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de convertir les erreurs en graphe du bon genre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -5;
    }

    graphe->erreur_y = erreurs;
    graphe->drapeaux.erreur_y = VRAI;

    /* Assignation terminée sans incident */
    return 0;
  }

  /* Si on est là : axe inconnu ! */
  (void) fprintf(stderr, "%s,%s l.%u : Axe à traiter inconnu [%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, axe);
  return -20;
}

/* -------- Créer une copie utilisateur des barres d'erreur --------
     Renvoie le graphe contenant la copie [en Y si X,Y demandés]
             NULL en cas d'erreur

     La copie est dans la meme série que l'original, juste après

   graphe : le graphe dont on veut extraire les erreurs [NON ISOLÉ]
   axe    : sur quel axe extraire les erreurs ?
 */

Graphe *GRF_SortirErreurs(Graphe *graphe, int axe)
{
  Graphe *erreurs = NULL;
  Serie_Graphes *serie;
  char *titre;

  /* Controles préliminaires */
  if ( NULL == graphe ) {
    (void) fprintf(stderr, "%s,%s l.%u : Appel avec graphe = NULL\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }
  if ( 0 == axe ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun axe indiqué\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* À quelle série appartient le graphe ? */
  serie = GRF_QuelleSerie(graphe);
  if ( NULL == serie ) {
    (void) fprintf(stderr, "%s,%s l.%u : Appel avec un graphe isolé !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  /* Conversion des erreurs en X */
  if ( axe & AXE_X ) {
    if ( NULL == graphe->erreur_x ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas d'erreur en X!\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Préparation du titre */
    titre = TXT_genere_chaine( titre_grf_erreurs_x, graphe->nom );

    /* Création du graphe */
    erreurs = Serie_nv_graphe( serie, graphe, graphe->nombre_point, GR_DEVIATION, GRF_FCT_REELLE, titre );
    free( titre ); titre = NULL;
    if ( NULL == erreurs ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer le graphe !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Contenu */
    /* a. les X */
    if ( graphe->x != NULL ) {
      memcpy( erreurs->x, graphe->x, graphe->nombre_point * sizeof( double ) );
    } else {
      unsigned long i;

      (void) fprintf(stderr, "%s,%s l.%u : Une série vraie ne devrait pas avoir d'erreur en X !\n",
		     __FILE__, __FUNCTION__, __LINE__);

      for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
	erreurs->x[ i ] = (double) i;
      }
    }
    /* b. les Y */
    memcpy( erreurs->y, graphe->erreur_x->y, graphe->nombre_point * sizeof( double ) );
    /* c. les DDL */
    memcpy( erreurs->z.ddl, graphe->erreur_x->z.ddl, graphe->nombre_point * sizeof( int ) );

    /* Les bornes */
    erreurs->x_min = graphe->x_min;
    erreurs->x_max = graphe->x_max;
    GRF_bornes( erreurs, NE_PAS_FAIRE, DESORDONNE );

    /* Légendes,... */
    if ( erreurs->affichage != NULL ) {
      int retour __attribute__((__unused__));

      /* Mêmes unités en X, même couleur & style */
      GRF_CopierAffichage( erreurs, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

      /* On construit la légende et l'unité pour l'axe des Y */
      titre = TXT_genere_chaine( "Erreurs sur %s", GRF_QuelleLegende( graphe, AXE_X ) );
      retour = GRF_ChangerLegende( erreurs, AXE_Y, titre, FAUX );
      GRF_ChangerUnite( erreurs, AXE_X, GRF_QuelleUnite( graphe, AXE_X ) );
    }

    /* Un mot dans l'historique */
    HIST_Texte(4, 0, "Barres d'erreur en X du graphe %s : %s\n", graphe->nom, erreurs->nom);
  }

  /* Conversion des erreurs en Y */
  if ( axe & AXE_Y ) {
    if ( NULL == graphe->erreur_y ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas d'erreur en Y!\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Préparation du titre */
    titre = TXT_genere_chaine( titre_grf_erreurs_y, graphe->nom );

    /* Création du graphe */
    if ( GRF_EstComplexe( graphe ) ) {
      erreurs = Serie_nv_graphe( serie, graphe, graphe->nombre_point, GR_DEVIATION, GRF_FCT_COMPLEXE, titre );
    } else if ( GRF_SERIE == graphe->genre_donnees ) {
      erreurs = Serie_nv_graphe( serie, graphe, graphe->nombre_point, GR_DEVIATION, GRF_SERIE, titre );
    } else {
      erreurs = Serie_nv_graphe( serie, graphe, graphe->nombre_point, GR_DEVIATION, GRF_FCT_REELLE, titre );
    }
    free( titre ); titre = NULL;
    if ( NULL == erreurs ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer le graphe !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Contenu */
    /* a. les X */
    if ( graphe->x != NULL ) {
      memcpy( erreurs->x, graphe->x, graphe->nombre_point * sizeof( double ) );
    }
    /* b. les Y */
    if ( GRF_EstComplexe( graphe ) ) {
      memcpy( erreurs->y, graphe->erreur_y->y, graphe->nombre_point * sizeof( Complexe ) );
    } else {
      memcpy( erreurs->y, graphe->erreur_y->y, graphe->nombre_point * sizeof( double ) );
    }
    /* c. les DDL */
    memcpy( erreurs->z.ddl, graphe->erreur_y->z.ddl, graphe->nombre_point * sizeof( int ) );

    /* Les bornes */
    erreurs->x_min = graphe->x_min;
    erreurs->x_max = graphe->x_max;
    GRF_bornes( erreurs, NE_PAS_FAIRE, DESORDONNE );

    /* Légendes,... */
    if ( erreurs->affichage != NULL ) {
      int retour __attribute__((__unused__));

      /* Memes unités en X et en Y, meme couleur & style */
      GRF_CopierAffichage( erreurs, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

      /* On construit la légende en Y */
      titre = TXT_genere_chaine( "Erreurs sur %s", GRF_QuelleLegende( graphe, AXE_Y ) );
      retour = GRF_ChangerLegende( erreurs, AXE_Y, titre, FAUX );

      if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
	erreurs->affichage->methode = GRF_C_RE_IM;
      }
    }

    /* Un mot dans l'historique */
    HIST_Texte( 4, 0, "Barres d'erreur en Y du graphe %s : %s\n", graphe->nom, erreurs->nom );
  }

  /* On renvoie ce que l'on a créé */
  return erreurs;
}

/* -------- Assigner des erreurs à un graphe -------- */

int GRF_AssignerErreurs(Graphe *graphe, Graphe *erreurs, boolean x_corrects, int axe)
{
  int retour;
  Serie_Graphes *sg_graphe __attribute__((__unused__)), *sg_erreurs;

  /* Vérifications préliminaires */
  if ( ( NULL == graphe ) || ( NULL == erreurs ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : pas de graphe, ou pas d'erreurs à associer\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( ( erreurs->x_max < graphe->x_max ) || ( graphe->x_min < erreurs->x_min ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : certains points du graphe n'existent pas dans les erreurs !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }
  if ( erreurs->y_min < 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : des erreurs doivent être nécessairement positives\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +2;
  }

  /* Contrôles spécifiques des axes */
  if ( AXE_X == axe ) {
    if ( ( FAUX == GRF_EstFonctionReelle( erreurs ) ) && ( erreurs->genre_donnees != GRF_ERREUR_R ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : les erreurs sur les X doivent être de type réel\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -2;
    }
  } else if ( AXE_Y == axe ) {
    if ( GRF_EstComplexe( graphe ) == VRAI ) {
      if ( FAUX == GRF_EstComplexe( erreurs ) ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreurs proposées non complexes !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -3;
      }
    } else {
      if ( ( FAUX == GRF_EstFonctionReelle( erreurs ) ) && ( erreurs->genre_donnees != GRF_ERREUR_R ) ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreurs proposées non réelles !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -4;
      }
    }
  } else {
    /* On ne devrait pas être là */
    (void) fprintf(stderr, "%s,%s l.%u : Type d'axe inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, axe);
    return -19;
  }

  /* À quelle série appartiennent les graphes ? */
  sg_graphe  = GRF_QuelleSerie( graphe );
  sg_erreurs = GRF_QuelleSerie( erreurs );

  /* Peut-on supposer que les X sont les mêmes ? */
  if ( VRAI == x_corrects ) {
    x_corrects = GRF_ComparerX( graphe, erreurs );
  }
  if ( VRAI == x_corrects ) { /* Mêmes X _exactement_ - Y compris le même nombre de points ! */
    /* On isole ce graphe */
    retour = GRF_IsolerGraphe( sg_erreurs, erreurs );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible d'isoler le graphe [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      x_corrects = FAUX;	/* On fait donc comme si les X étaient faux : copie du graphe */
    }
  }
  if ( FAUX == x_corrects ) {
    /* X différents ou identiques mais isolement impossible */
    int code_erreur;

    /* On recale donc les erreurs, en les copiant, sur les X corrects */
    erreurs = GRF_CalerSurGraphe( erreurs, INTERPOLE_SPLINE, INTERPOLE_BORDS_0_0, graphe, &code_erreur );
    if ( NULL == erreurs ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible d'interpoler le graphe [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, code_erreur);
      return -21;
    }
  }

  /* On insère comme erreurs le résultat */
  retour = GRF_InsererErreurs( graphe, erreurs, axe );

  /* Contrôles finaux */
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'insérer le graphe [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    if ( FAUX == x_corrects ) {
      /* On l'efface donc, puisqu'il est isolé & est une copie interpolée */
      GRF_eff_isole( erreurs );
    } else {
      /* On le réinsère dans la série d'origine */
      (void) GRF_InsererDansSerie( erreurs, sg_erreurs, NULL );
    }
    return -58;
  }

  /* Assignation effectuée */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Supprimer le graphe des barres d'erreur ~-~-~-~-~-~-~-~- */

int GRF_SupprimerErreurs(Graphe *graphe, int axe)
{
  if ( NULL == graphe ) return -1;

  if ( axe & AXE_X ) {
    /* On efface le graphe isolé des erreurs en X */
    GRF_eff_isole( graphe->erreur_x );
  }

  if ( axe & AXE_Y ) {
    /* On efface le graphe isolé des erreurs en Y */
    GRF_eff_isole( graphe->erreur_y );
  }

  /* Pas de problème particulier */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Travail sur le graphe des barres d'erreur ~-~-~-~-~-~-~-~- */

/* -------- Initialisation -------- */

static int GRF_InitialiseErreurs(Graphe *erreurs, const Graphe *reference)
{
  /* Vérifications préliminaires */
  if ( NULL == erreurs ) /* Rien à faire, mais étrange */
    return +1;
  if ( NULL == reference ) {
    (void) fprintf(stderr, "GRF_InitialiseErreurs : pas de graphe référence !\n");
    return -1;
  }

  /* Les X sont forcément les mêmes que ceux du graphe -> restent NULL ! */
  erreurs->x_min = reference->x_min;
  erreurs->x_max = reference->x_max;

  /* On met la même couleur et le même style, pour faciliter le repérage...
     L'axe des X est identique ; on ne sait rien sur les Y car peut venir de X ou de Y...
  */
  return 0;
}

/* -------- Extraire les valeurs des erreurs (réelles) d'un graphe de mêmes X -------- */
#ifdef maintenant_inutile
static int GRF_ExtraitErreurs_R(Graphe *erreurs, const Graphe *source_erreurs)
{
  unsigned long nbr_valeurs, premier;

  /* 1. Quelle sous-partie du graphe à assigner convient ? */
  nbr_valeurs = GRF_intervalle( source_erreurs, erreurs->x_min, erreurs->x_max, &premier);
  if ( nbr_valeurs != erreurs->nombre_point ) {
    (void) fprintf(stderr, "GRF_ExtraitErreurs_R : pas de correspondance simple entre les X du graphe et ceux des erreurs\n");
    return -3;
  }

  /* 2. On complète les Y et les ddl à partir de cette table */
  memcpy( erreurs->y, source_erreurs->y, sizeof(double) * nbr_valeurs );
  memcpy( erreurs->z.ddl, source_erreurs->z.ddl, sizeof(int) * nbr_valeurs );
  GRF_bornes( erreurs, NE_PAS_FAIRE, DESORDONNE );

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Extraire les valeurs des erreurs (complexes) d'un graphe de mêmes X -------- */

static int GRF_ExtraitErreurs_C(Graphe *erreurs, const Graphe *source_erreurs)
{
  unsigned long nbr_valeurs, premier;

  /* 1. Quelle sous-partie du graphe à assigner convient ? */
  nbr_valeurs = GRF_intervalle( source_erreurs, erreurs->x_min, erreurs->x_max, &premier);
  if ( nbr_valeurs != erreurs->nombre_point ) {
    (void) fprintf(stderr, "GRF_ExtraitErreurs_C : pas de correspondance simple entre les X du graphe et ceux des erreurs\n");
    return -3;
  }

  /* 2. On complète les Y et les ddl à partir de cette table */
  memcpy( erreurs->y, source_erreurs->y, sizeof( Complexe ) * nbr_valeurs );
  memcpy( erreurs->z.ddl, source_erreurs->z.ddl, sizeof(int) * nbr_valeurs );
  GRF_bornes( erreurs, NE_PAS_FAIRE, DESORDONNE );

  /* Tout s'est bien passé */
  return 0;
}
#endif
