/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Traitement simple du contenu d'un graphe :
    - extraction d'un sous-ensemble des points
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "graphes.h"
#include "matrices.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned long GRF_CompterSurAxeX(const Graphe *graphe, double x_min, double x_max, boolean dehors);
static unsigned long GRF_CompterSurAxeY(const Graphe *graphe, double y_min, double y_max, boolean dehors);

static unsigned long Extrait_CDedans(const Graphe *graphe, double x_min, double x_max, double *x, Complexe *y);
static unsigned long Extrait_CDehors(const Graphe *graphe, double x_min, double x_max, double *x, Complexe *y);
static unsigned long Extrait_RDedans(const Graphe *graphe, double x_min, double x_max, double *x, double *y);
static unsigned long Extrait_RDehors(const Graphe *graphe, double x_min, double x_max, double *x, double *y);
static unsigned long Extrait_SDedans(const Graphe *graphe, double x_min, double x_max, double *y);
static unsigned long Extrait_SDehors(const Graphe *graphe, double x_min, double x_max, double *y);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ---------------- Rechercher une valeur dans un graphe ---------------------------------
        Renvoie le nombre le plus proche de valeur trouv dans graphe
        mode = VRAI => chercher dans les x, FAUX => chercher dans les y                     */

double plus_proche(Graphe *graphe, double valeur, boolean mode)
{
  double erreur, retour;
  unsigned long i;

  if (mode == VRAI) { /* Quel est le x enregistr le plus proche de valeur ? */
    register double *x = graphe->x;

    if (valeur <= graphe->x_min) return graphe->x_min; /* On est au-del des bornes */
    if (valeur >= graphe->x_max) return graphe->x_max;

    erreur = fabs( valeur - *x );
    retour = *x; x++;
    for (i = 1 ; i < graphe->nombre_point ; i++, x++) {
      if ( fabs( valeur - *x ) < erreur ) {
	retour = *x;
	erreur = fabs(valeur - retour);
      }
      if (erreur == 0.0) /* La valeur est contenue dans le graphe */
	return retour;
    }
  } else { /* Quel est le y enregistr le plus proche de valeur ? */
    register double *y = graphe->y;

    if (valeur <= graphe->y_min) return graphe->y_min; /* On est au-del des bornes */
    if (valeur >= graphe->y_max) return graphe->y_max;

    erreur = fabs( valeur - *y );
    retour = *y; y++;
    for (i = 1 ; i < graphe->nombre_point ; i++, y++) {
      if ( fabs( valeur - *y ) < erreur ) {
	retour = *y;
	erreur = fabs(valeur - retour);
      }
    }
  }
  
  return retour;
}

/* -------- Nombre de points compris entre x_min et x_max ------------\
 |   Renvoie le nombre de points convenant et le numro du 1er        |
 |   Suppose le graphe tri par x croissants                          |
 |                                                                    |
 | graphe       = le graphe o chercher                               |
 | x_min, x_max = l'intervalle  conserver                            |
 | premier      = le numro du premier point valable                  |
 \------------------------------------------------------------------- */

unsigned long GRF_intervalle(const Graphe *graphe, double x_min, double x_max, unsigned long *debut)
{
  unsigned long i, nbr_pts = 0;
  double *x;

  ENTREE_FONCTION("GRF_intervalle", "graphe=%p x_min=%g x_max=%g debut=%p", graphe, x_min, x_max, debut);

  if ( (NULL == graphe) || (NULL == debut) ) {
    AVERTISSEMENT( "Appel avec un paramtre NULL [graphe=%p, debut=%p]", graphe, debut );
    SORTIE_FONCTION("GRF_intervalle", "%lu [Paramtres faux]", 0);
    return 0;
  }

  if ( x_max < x_min ) { /* On s'assure que min < max ! */
    double x = x_max; x_max = x_min; x_min = x;
    AVERTISSEMENT( "Permutation des bornes, x_min = %f, x_max = %g", x_min, x_max );
  }

  if ( (graphe->x_min > x_max) || (graphe->x_max < x_min) ) {
    SORTIE_FONCTION("GRF_intervalle", "%lu [hors-graphe]", 0);
    return 0;
  }
  if ( x_min < graphe->x_min ) x_min = graphe->x_min;
  if ( x_max > graphe->x_max ) x_max = graphe->x_max;

  /* On se place sur le premier point > x_min, il existe forcment sinon gr->x_min > x_max */
  for (i = 0, x = graphe->x ; *x < x_min ; i++, x++) {
#ifdef _DEVELOPPE_
    if ( i > graphe->nombre_point ) {
      (void) fprintf(stderr, "GRF_intervalle : ABSURDE, on dborde ! [x_min=%g, graphe->x_min=%g, graphe->x_max=%g, graphe->x[0]=%g]\n",
		     x_min, graphe->x_min, graphe->x_max, *(graphe->x));
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }
#endif
  }
  *debut = i; /* Index du premier point convenable */

  do {
    nbr_pts++;
    x++; i++;
    if (i >= graphe->nombre_point) break;
  } while (*x <= x_max);

  SORTIE_FONCTION("GRF_intervalle", "%lu", nbr_pts);
  return nbr_pts;
}

/* -------- Compter les points dans/hors d'un intervalle dans un graphe -------- */

static unsigned long GRF_CompterSurAxeX(const Graphe *graphe, double x_min, double x_max, boolean dehors)
{
  unsigned long i, nbr_pts = 0;
  double *x;

  if ( NULL == graphe->x ) { /* X implicites */
    unsigned long i_min, i_max;

    /* Index extrmes */
    if ( x_min <= 0.0 ) i_min = 0;
    else {
      i_min = (int) x_min; /* 3.0 -> 3 OK, mais 3.1 donne 3 il faut 4 */
      if ( x_min > (double) i_min ) i_min++;
    }

    if ( x_max >= (double) ( graphe->nombre_point - 1 ) ) i_max = graphe->nombre_point - 1;
    else {
      i_max = (int) x_max; /* 5.0 -> 5 OK, 5.2 donne 5 il faut 5 OK */
    }

    /* Ici, on i_min : indice du premier point >= x_min ; i_max : indice du dernier point <= x_max */
    /* Nombre de points entre ces deux valeurs
       = nombre de points entre x_min et x_max = i_max - i_min + 1 */
    nbr_pts = i_max - i_min + 1;

    if ( VRAI == dehors ) {
      /* On veut les points autre que ceux de [x_min, x_max] */
      if ( ( x_max < 0.0 ) || ( x_min > (double) (graphe->nombre_point - 1) ) ) return graphe->nombre_point;

      return graphe->nombre_point - nbr_pts;
    } else {
      if ( ( x_max < 0.0 ) || ( x_min > (double) (graphe->nombre_point - 1) ) ) return 0;

      return nbr_pts;
    }
  }

  if ( VRAI == dehors ) {
    /* Vrification grossire */
    if ( (x_min <= graphe->x_min) && (x_max >= graphe->x_max) ) return 0;

    /* Dcompte rel */
    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++) {
      if ( (*x < x_min) || (*x > x_max) )
	nbr_pts++;
    }
  } else {
    /* Vrifications grossires */
    if ( x_min > graphe->x_max ) return 0;
    if ( x_max < graphe->x_min ) return 0;

    /* Dcompte rel */
    for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++) {
      if ( (*x >= x_min) && (*x <= x_max) )
	nbr_pts++;
    }
  }

  return nbr_pts;
}

static unsigned long GRF_CompterSurAxeY(const Graphe *graphe, double y_min, double y_max, boolean dehors)
{
  unsigned long i, nbr_pts = 0;
  double *y;

  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    (void) fprintf(stderr, "GRF_CompterSurAxeY : dcompte impossible sur un graphe complexe ! [%s]\n", graphe->nom);
    return 0;
  }

  if ( VRAI == dehors ) {
    /* Vrification grossire */
    if ( (y_min <= graphe->y_min) && (y_max >= graphe->y_max) ) return 0;

    /* Dcompte rel */
    for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
      if ( (*y < y_min) || (*y > y_max) )
	nbr_pts++;
    }
  } else {
    /* Vrifications grossires */
    if ( y_min > graphe->y_max ) return 0;
    if ( y_max < graphe->y_min ) return 0;

    /* Dcompte rel */
    for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
      if ( (*y >= y_min) && (*y <= y_max) )
	nbr_pts++;
    }
  }

  return nbr_pts;
}

unsigned long GRF_CompterPoints(const Graphe *graphe, int axe, double min, double max, boolean dehors)
{
  if (NULL == graphe) return 0;

  if ( AXE_X == axe)
    return GRF_CompterSurAxeX(graphe, min, max, dehors);
  else
    return GRF_CompterSurAxeY(graphe, min, max, dehors);
}

/* -------- Extraire une table de valeurs d'un graphe --------------\
 |  Renvoie le nombre de valeurs dans les tables obtenues           |
 |                                                                  |
 | graphe : le graphe dont il faut extraire le sous-graphe          |
 | x_min, x_max : l'intervalle  considrer ([x_min, x_max])        |
 | dehors : VRAI ==> on garde les points en-dehors de l'intervalle  |
 | x, y   : les tables rserves, qui contiennent les valeurs       |
 \----------------------------------------------------------------- */

unsigned long GRF_extrait_table(const Graphe *graphe, double x_min, double x_max, boolean dehors, double **x, double **y)
{
  unsigned long nbr_pts;

  if ( NULL == graphe ) return 0;
  if ( (NULL == x) || (NULL == y) ) return 0;

  /* Combien de points ? */
  nbr_pts = GRF_CompterPoints(graphe, AXE_X, x_min, x_max, dehors);
  if (nbr_pts < 2) {
    (void) fprintf(stderr, "GRF_extrait_table : trop peu de points valables [%lu]\n", nbr_pts);
    return 0;
  }

  /* On rserve les tables */
  *x = (double *) malloc( sizeof(double) * nbr_pts );
  if ( NULL == *x ) {
    (void) fprintf(stderr, "GRF_extrait_table : mmoire insuffisante / table des X\n");
    return 0;
  }

  if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
    *y = (double *) malloc( sizeof(double) * nbr_pts );
    if ( NULL == *y ) {
      (void) fprintf(stderr, "GRF_extrait_table : mmoire insuffisante / table des Y rels\n");
      free( *x );
      *x = NULL;
      return 0;
    }

    if ( VRAI == dehors ) {
      nbr_pts = Extrait_RDehors(graphe, x_min, x_max, *x, *y);
    } else {
      nbr_pts = Extrait_RDedans(graphe, x_min, x_max, *x, *y);
    }
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    Complexe *table_y;

    table_y = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
    if ( NULL == table_y ) {
      (void) fprintf(stderr, "GRF_extrait_table : mmoire insuffisante / table des Y complexes\n");
      free( *x );
      *x = NULL;
      return 0;
    }

    if ( VRAI == dehors ) {
      nbr_pts = Extrait_CDehors(graphe, x_min, x_max, *x, table_y);
    } else {
      nbr_pts = Extrait_CDedans(graphe, x_min, x_max, *x, table_y);
    }

    *y = (double *) table_y; /* Assez sale, mais bon... */
  } else {
    (void) fprintf(stderr, "GRF_extrait_table :  FAIRE, graphe contenant des donnes de type %d\n", graphe->genre_donnees);
  }

  return nbr_pts;
}

/* ------ Extraction des tables. ATTENTION, les tables doivent avoir t rserves et prvues pour le bon nombre de points */
/* Valeurs complexes, dans [min, max] */
static unsigned long Extrait_CDedans(const Graphe *graphe, double x_min, double x_max, double *x_table, Complexe *y_table)
{
  unsigned long nbr_pts, i;
  register double *x;
  register Complexe *y;

  /* On recopie dans les tables les valeurs retenues */
  for (i = 0, nbr_pts = 0, x = graphe->x, y = (Complexe *) graphe->y ;
       i < graphe->nombre_point ; i++, x++, y++) {
    if ( (*x >= x_min) && (*x <= x_max) ) {
      (x_table)[ nbr_pts ] = *x;
      (y_table)[ nbr_pts ] = *y;
      nbr_pts++;
    }
  }

  return nbr_pts;
}

/* Valeurs complexes, hors de [min, max] */
static unsigned long Extrait_CDehors(const Graphe *graphe, double x_min, double x_max, double *x_table, Complexe *y_table)
{
  unsigned long nbr_pts, i;
  register double *x;
  register Complexe *y;

  /* On recopie dans les tables les valeurs retenues */
  for (i = 0, nbr_pts = 0, x = graphe->x, y = (Complexe *) graphe->y ;
       i < graphe->nombre_point ; i++, x++, y++) {
    if ( (*x < x_min) || (*x > x_max) ) {
      (x_table)[ nbr_pts ] = *x;
      (y_table)[ nbr_pts ] = *y;
      nbr_pts++;
    }
  }

  return nbr_pts;
}

/* Valeurs relles, dans [min, max] */
static unsigned long Extrait_RDedans(const Graphe *graphe, double x_min, double x_max, double *x_table, double *y_table)
{
  unsigned long nbr_pts = 0, i;
  register double *x, *y;

  /* On recopie dans les tables les valeurs retenues */
  for (i = 0, nbr_pts = 0, x = graphe->x, y = graphe->y;
       i < graphe->nombre_point ; i++, x++, y++) {
    if ( (*x >= x_min) && (*x <= x_max) ) {
      (x_table)[ nbr_pts ] = *x;
      (y_table)[ nbr_pts ] = *y;
      nbr_pts++;
    }
  }

  return nbr_pts;
}

/* Valeurs relles, hors de [min, max] */
static unsigned long Extrait_RDehors(const Graphe *graphe, double x_min, double x_max, double *x_table, double *y_table)
{
  unsigned long nbr_pts = 0, i;
  register double *x, *y;

  /* On recopie dans les tables les valeurs retenues */
  for (i = 0, nbr_pts = 0, x = graphe->x, y = graphe->y;
       i < graphe->nombre_point ; i++, x++, y++) {
    if ( (*x < x_min) || (*x > x_max) ) {
      (x_table)[ nbr_pts ] = *x;
      (y_table)[ nbr_pts ] = *y;
      nbr_pts++;
    }
  }

  return nbr_pts;
}

/* Valeurs d'une srie relle, dans [min, max] */
static unsigned long Extrait_SDedans(const Graphe *graphe, double x_min, double x_max, double *y_table)
{
  unsigned long nbr_pts = 0, i;
  register double *y;

  /* On recopie dans les tables les valeurs retenues */
  if ( NULL == graphe->x ) { /* X implicites */
    for (i = 0, nbr_pts = 0, y = graphe->y;
	 i < graphe->nombre_point ; i++, y++) {
      if ( ( i >= x_min) && ( i <= x_max) ) {
	y_table[ nbr_pts ] = *y;
	nbr_pts++;
      }
    }
  } else {
    double *x;

    for (i = 0, nbr_pts = 0, x = graphe->x, y = graphe->y;
	 i < graphe->nombre_point ; i++, x++, y++) {
      if ( ( *x >= x_min) && ( *x <= x_max) ) {
	y_table[ nbr_pts ] = *y;
	nbr_pts++;
      }
    }
  }

  return nbr_pts;
}

/* Valeurs relles, hors de [min, max] */
static unsigned long Extrait_SDehors(const Graphe *graphe, double x_min, double x_max, double *y_table)
{
  unsigned long nbr_pts = 0, i;
  register double *y;

  /* On recopie dans les tables les valeurs retenues */
  if ( NULL == graphe->x ) { /* X implicites */
    for (i = 0, nbr_pts = 0, y = graphe->y;
	 i < graphe->nombre_point ; i++, y++) {
      if ( ( ((double) i) < x_min) || ( ((double) i) > x_max) ) {
	y_table[ nbr_pts ] = *y;
	nbr_pts++;
      }
    }
  } else {
    double *x;

    for (i = 0, nbr_pts = 0, x = graphe->x, y = graphe->y;
	 i < graphe->nombre_point ; i++, x++, y++) {
      if ( ( *x < x_min) || ( *x > x_max) ) {
	(y_table)[ nbr_pts ] = *y;
	nbr_pts++;
      }
    }
  }

  return nbr_pts;
}

/* -------- Extraire un graphe d'un autre graphe -------------------\
 |  Renvoie le graphe extrait, NULL en cas d'erreur                 |
 |   les points extraits sont entre premier et premier + nombre - 1 |
 |                                                                  |
 | graphe : le graphe dont il faut extraire le sous-graphe          |
 | isole  : VRAI => on gnre un graphe isol                       |
 \----------------------------------------------------------------- */

Graphe *GRF_ExtraitGraphe(Graphe *graphe, unsigned long premier, unsigned long nombre, boolean isole)
{
  Graphe *resultat;

  if (NULL == graphe) return NULL;
  if ( (nombre > graphe->nombre_point) || (premier > graphe->nombre_point) ||
       (premier + nombre > graphe->nombre_point) ) {
    (void) fprintf(stderr, "GRF_ExtraitGraphe : intervalle indiqu absurde !\n");
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL; /* On dborde ! */
  }

  if ( VRAI == isole ) {
    resultat = GRF_nv_isole(nombre, graphe->type, graphe->genre_donnees);
  } else {
    resultat = GRF_nouveau(graphe, nombre, graphe->type, graphe->genre_donnees, "Extraction de graphe");
  }
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe rsultat\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On recopie les donnes */
  switch ( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
    memcpy( resultat->x, graphe->x + premier, sizeof(double) * nombre );
    memcpy( resultat->y, graphe->y + premier, sizeof(double) * nombre );
    break;

  case GRF_FCT_COMPLEXE:
    memcpy( resultat->x, graphe->x + premier, sizeof(double) * nombre );
    memcpy( resultat->y, ((Complexe *) graphe->y) + premier, sizeof(Complexe) * nombre );
    break;

  case GRF_SERIE:
    if ( NULL != graphe->x ) {
      int retour;

      /* On doit avoir de vrais X dans le rsultat pour garder la correspondance */
      retour = GRF_ChangerGenre( resultat, GRF_FCT_REELLE );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de garder la correspondance avec les X -> X implicites\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	memcpy( resultat->x, graphe->x + premier, sizeof(double) * nombre );
      }
    }
    memcpy( resultat->y, graphe->y + premier, sizeof(double) * nombre );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Genre de graphe inconnu ou non-trait [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->genre_donnees );
  }

  if ( GR_DEVIATION == graphe->type ) {
    memcpy( resultat->z.ddl, graphe->z.ddl + premier, sizeof(int) * nombre );
  }
  GRF_bornes(resultat, DESORDONNE, DESORDONNE);

  /* On conserve les modes d'affichage & autres infos annexes */
  GRF_CopierAffichage( resultat, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y |
		       GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* On s'occupe des barres d'erreur */
  if ( graphe->erreur_x != NULL ) {
    Graphe *erreurs;

    erreurs = GRF_creer_erreurs(resultat, AXE_X);
    if (erreurs != NULL) {
      memcpy(erreurs->y, graphe->erreur_x->y + premier, sizeof(double) * nombre);
      memcpy(erreurs->z.ddl, graphe->erreur_x->z.ddl + premier, sizeof(int) * nombre);

      GRF_bornes(erreurs, DESORDONNE, DESORDONNE);
    }
  }

  if ( graphe->erreur_y != NULL ) {
    Graphe *erreurs;

    erreurs = GRF_creer_erreurs(resultat, AXE_Y);
    if (erreurs != NULL) {
      if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
	memcpy(erreurs->y, graphe->erreur_y->y + premier, sizeof(double) * nombre);
	memcpy(erreurs->z.ddl, graphe->erreur_y->z.ddl + premier, sizeof(int) * nombre);
      } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
	memcpy(erreurs->y, ((Complexe *) graphe->erreur_y->y) + premier, sizeof(Complexe) * nombre);
      }

      GRF_bornes(erreurs, DESORDONNE, DESORDONNE);
    }
  }

  /* On s'occupe de la matrice de covariance */
  if ( graphe->covariance != NULL ) {
    /* On extrait le morceau de la matrice de covariance qui convient */
    if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
      resultat->covariance = MAT_ExtraitBloc( graphe->covariance, premier, premier, nombre, nombre );
    } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
      resultat->covariance = MAT_ExtraitBloc( graphe->covariance, 2 * premier, 2 * premier, 2 * nombre, 2 * nombre );
    }
  }

  return resultat;
}

/* -------- Extraire un graphe  partir d'un autre graphe et d'un intervalle (pas forcment tri) -------- */

Graphe *GRF_ExtraitGrapheI(Graphe *graphe, double x_min, double x_max, boolean dehors, boolean isole)
{
  Graphe *resultat;
  double *x;
  unsigned long nbr_pts;

  if ( NULL == graphe ) return NULL;

  /* On compte le nombre de points convenable - attention, graphe pas forcment tri */
  nbr_pts = GRF_CompterPoints(graphe, AXE_X, x_min, x_max, dehors);
  if ( nbr_pts < 2 ) return NULL;
#ifdef _DEVELOPPE_
/*   printf("GRF_ExtraitGrapheI : %lu points conservs dans \"%s\"\n", nbr_pts, graphe->nom); */
#endif

  /* On cre les graphes qui vont contenir le rsultat */
  if ( VRAI == isole ) {
    resultat = GRF_nv_isole(nbr_pts, graphe->type, graphe->genre_donnees);
  } else {
    resultat = GRF_nouveau(graphe, nbr_pts, graphe->type, graphe->genre_donnees, "Extraction de graphe");
  }
  if (NULL == resultat) return NULL;

  /* On fait l'extraction proprement dite */
  if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
    if ( VRAI == dehors ) {
      nbr_pts = Extrait_RDehors(graphe, x_min, x_max, resultat->x, resultat->y);
    } else {
      nbr_pts = Extrait_RDedans(graphe, x_min, x_max, resultat->x, resultat->y);
    }
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    if ( VRAI == dehors ) {
      nbr_pts = Extrait_CDehors(graphe, x_min, x_max, resultat->x, (Complexe *) (resultat->y) );
    } else {
      nbr_pts = Extrait_CDedans(graphe, x_min, x_max, resultat->x, (Complexe *) (resultat->y) );
    }
  } else if ( GRF_SERIE == graphe->genre_donnees ) {
    if ( VRAI == dehors ) {
      nbr_pts = Extrait_SDehors(graphe, x_min, x_max, resultat->y );
    } else {
      nbr_pts = Extrait_SDedans(graphe, x_min, x_max, resultat->y );
    }
  } else {
    (void) fprintf(stderr, "GRF_ExtraitGrapheI :  FAIRE, donnes de type %d\n", graphe->genre_donnees );
  }

  if (GR_DEVIATION == graphe->type) { /* On s'occupe des ddl */
    unsigned long i;
    int *ddl_source, *ddl_extrait;

    if (VRAI == dehors) {
      for (i = 0, ddl_source = graphe->z.ddl, ddl_extrait = resultat->z.ddl, x = graphe->x ;
	   i < graphe->nombre_point ;
	   i++, ddl_source++, x++) {
	if ( (*x < x_min) || (*x > x_max) ) { /* Point  garder */
	  *ddl_extrait = *ddl_source;
	  ddl_extrait++;
	}
      }
#ifdef _DEVELOPPE_
/*       printf(" --- %lu points d'erreur conservs, en %p / %p\n", i, (void*) graphe->z.ddl, (void *) resultat->z.ddl); */
#endif
    } else {
      for (i = 0, ddl_source = graphe->z.ddl, ddl_extrait = resultat->z.ddl, x = graphe->x ;
	   i < graphe->nombre_point ;
	   i++, ddl_source++, x++) {
	if ( (*x >= x_min) && (*x <= x_max) ) { /* Point  garder */
	  *ddl_extrait = *ddl_source;
	  ddl_extrait++;
	}
      }
    }
  }

  /* On s'occupe maintenant des barres d'erreur.
     Comme les X des barres d'erreur sont forcment ceux du graphe, il suffit de rappeler rcursivement... */
  if ( graphe->erreur_x != NULL ) {
    resultat->erreur_x = GRF_ExtraitGrapheI(graphe->erreur_x, x_min, x_max, dehors, isole);
  }
  if ( graphe->erreur_y != NULL ) {
    resultat->erreur_y = GRF_ExtraitGrapheI(graphe->erreur_y, x_min, x_max, dehors, isole);
  }

  if ( graphe->covariance != NULL ) {
    fprintf(stderr, "GRF_ExtraitGrapheI :  FAIRE, traitement des matrices de covariance !\n");
  }

  /* On conserve les modes d'affichage & autres infos annexes */
  resultat->pointeur.autre = GRF_CopierComplements(graphe);
  GRF_CopierAffichage( resultat, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y |
		       GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* On a fini ! */
  GRF_bornes(resultat, DESORDONNE, DESORDONNE);
  return resultat;
}
