/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS
   Cration : 28 janvier 2002

   Gestion gnrique des graphes

   Gnration de graphes  partir des formules
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "utils_tables.h"

#include "c_formel.h"
#include "gen_graphe.h"
#include "images.h"
#include "matrices.h"
#include "serie.h"

#include "grf_generer.txt"

/* ~~~~~~~~ Structures locales ~~~~~~~ */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_GRAPHE_X 37

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

Graphe *gen_grf_2D_cart (double x_min, double pas, unsigned long nbr_pts, Expression *fonction);
Graphe *gen_grf_2D_Pcart(double t_min, double pas, unsigned long nbr_pts, Expression *fct_X, Expression *fct_Y);

static double *VAR_GenererTable(struct info_variable *variable, unsigned long *nombre_valeurs);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -~-~-~-~-~-~-~-~ Gnration  partir de formules simples -~-~-~-~-~-~-~-~ */

/* -------- Cration d'un polynme ---------------------------------\
 | Renvoie 0 si tout s'est bien pass, != 0 sinon                   |
 |                                                                  |
 | graphe : le graphe  gnrer. Calcule les Y pour les X du graphe |
 | ajouter: VRAI ==> on ajoute au graphe existant                   |
 |          FAUX ==> on remplace (on commence par mettre  zro)    |
 | degre  : le degr du polynme                                    |
 | coefs  : les coefficients : y = coefs[0] + coefs[1] * x + ...    |
 | covariance : la matrice de covariance des coefficients           |
 |              (NULL : ils sont parfaitement connus)               |
 \----------------------------------------------------------------- */

int GRF_GenerePolynome(Graphe *graphe, boolean ajouter, unsigned int degre, double *coefs, Matrice *covariance)
{
  Graphe *erreurs;
  unsigned long i;
  double *x, *y, y_poly;

  if ( ( NULL == graphe ) || (NULL == coefs) ) return -1;

  if ( FAUX == ajouter ) {
    for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
      *y = 0.0;
    }
  }

  /* On calcule le polynome proprement dit */
  for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
    unsigned int j = degre;

    /* Mthode de Horner : P(x) = (((An)*x + An-1)*x +... )*x + A0 */
    y_poly = coefs[ degre ];
    while (j > 0) {
      y_poly *= *x;
      y_poly += coefs[ --j ];
    }

    *y += y_poly;
  }

  GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
  if ( covariance == NULL ) return 0; /* Pas d'erreur sur les Y */

  /* On s'occupe des erreurs sur les Y  cause des coefficients */
  erreurs = GRF_creer_erreurs(graphe, AXE_Y);
  if ( NULL == erreurs ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_creer_erreurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return FAUX; /* Il y a eu un problme */
  }
  if ( MATRICE_DIAGONALE == covariance->type ) { /* Coefficients indpendants */
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Estimation des erreurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  return VRAI;
}

/* -~-~-~-~-~-~-~-~ Gnration  partir de formules complexes -~-~-~-~-~-~-~-~ */

/* -------- Gnrer une table de valeurs de la variable  partir de ses infos
   Fonctionne dans tous les cas, SAUF SI LA VARIABLE EST CALCULE PAR UNE FONCTION
   Renvoie la table cre
           NULL en cas de problme

   variable       : la variable pour laquelle il faut gnrer une table de valeurs
   nombre_valeurs : contiendra le nombre de valeurs dans la table ainsi cre
 */

static double *VAR_GenererTable(struct info_variable *variable, unsigned long *nombre_valeurs)
{
  double *table;

  if ( NULL == variable ) return NULL;
  if ( NULL == nombre_valeurs ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de mmoriser la taille de la table !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* La taille de la table ne sera mmorise que si tout s'est bien pass */
  *nombre_valeurs = 0;

  switch ( variable->type )
  {
  case VARIABLE_MIN_MAX_PAS:
    table = TABLE_CreerTable_mMp( variable->min, variable->max, variable->pas, nombre_valeurs );
    break;

  case VARIABLE_MIN_PAS_NOMBRE:
    table = TABLE_CreerTable_mNp( variable->min, variable->nombre, variable->pas );
    *nombre_valeurs = variable->nombre;
    break;

  case VARIABLE_MAX_PAS_NOMBRE:
    table = TABLE_CreerTable_MNp( variable->max, variable->nombre, variable->pas );
    *nombre_valeurs = variable->nombre;
    break;

  case VARIABLE_MIN_MAX_NOMBRE:
    table = TABLE_CreerTable_mMN( variable->min, variable->max, variable->nombre );
    *nombre_valeurs = variable->nombre;
    break;

  case VARIABLE_X_GRAPHE:
    if ( NULL == variable->graphe ) {
      (void) fprintf(stderr, "%s, %s l.%u : les valeurs doivent venir d'un graphe, mais pas de graphe indiqu\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
    if ( NULL == variable->graphe->x ) {
      (void) fprintf(stderr, "%s, %s l.%u : les valeurs doivent venir des X d'un graphe, mais le graphe %s n'a pas de X !\n",
		     __FILE__, __FUNCTION__, __LINE__, variable->graphe->nom );
      return NULL;
    }

    /* Le nombre de points */
    *nombre_valeurs = variable->graphe->nombre_point;

    /* On pr-gnre le graphe avec les bons X */
    table = TABLE_CopierTable( variable->graphe->x, *nombre_valeurs );
    break;

  case VARIABLE_Y_GRAPHE:
    if ( NULL == variable->graphe ) {
      (void) fprintf(stderr, "%s, %s l.%u : les valeurs doivent venir d'un graphe, mais pas de graphe indiqu\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
    if ( NULL == variable->graphe->y ) {
      (void) fprintf(stderr, "%s, %s l.%u : les valeurs doivent venir des Y d'un graphe, mais le graphe %s n'a pas de X !\n",
		     __FILE__, __FUNCTION__, __LINE__, variable->graphe->nom );
      return NULL;
    }

    /* Le nombre de points */
    *nombre_valeurs = variable->graphe->nombre_point;

    /* On pr-gnre le graphe avec les bons X */
    table = TABLE_CopierTable( variable->graphe->y, *nombre_valeurs );
    break;

  case VARIABLE_VALEURS:
    table = TABLE_CreerTable_SV( variable->valeurs, nombre_valeurs, ';' );
    break;

  case VARIABLE_FORMULE:
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de calculer pour une variable dfinie par une fonction\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type de variable non-gr [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, variable->type);
    return NULL;
  }

  /* On a fini */
  return table;
}

struct graphe *GRF_GenererFonctionReelle(const char *nom, int type, struct info_variable *X, struct info_variable *Y, struct info_variable *I,
					 unsigned long nbr_parametres, struct parametre *parametres,
					 struct graphe *avant, struct serie *serie)
{
  int retour;
  unsigned long i, nombre_points, nombre_Y;
  double *table_X = NULL, *table_Y = NULL;
  Graphe *resultat = NULL;

  if ( ( NULL == X ) || ( NULL == Y ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Informations manquantes (X [%p] ou Y [%p])\n",
		   __FILE__, __FUNCTION__, __LINE__, X, Y);
    return NULL;
  }

  /* On prpare la table des X, puis on fait Y = f(X) */
  if ( X->type != VARIABLE_FORMULE ) {
    table_X = VAR_GenererTable( X, &nombre_points );
  } else {
    if ( NULL == I ) {
      (void) fprintf(stderr, "%s, %s l.%u : X=f(I), mais I non dfinie !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* I : numro de point, donc min = 0, max = N, pas = 1 forcment */
    nombre_points = I->nombre;

    /* On gnre une table  partir de la variable I */
    table_X = TABLE_Creer_Formule( NULL, nombre_points, X->valeurs, nbr_parametres, parametres );
  }

  /* On contrle que l'on a pu prparer la table des X ! */
  if ( NULL == table_X ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de prparer la table des X\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On en fait le graphe */
  resultat = GRF_nv_TableX( NULL, type, GRF_FCT_REELLE, nom, VRAI, table_X, nombre_points, FAUX );

  /* On contrle que l'on a pu prparer le graphe */
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de prparer le graphe %s - Opration annule\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
    return NULL;
  }

  /* On construit maintenant les Y */
  if ( Y->type != VARIABLE_FORMULE ) {
    printf( "%s : Y de type %d\n", __FUNCTION__, Y->type );
    table_Y = VAR_GenererTable( Y, &nombre_Y );

    if ( nombre_Y != nombre_points ) {
      int reponse;

      reponse = BI_ERR_alerte( NULL, &ERR_NYDifferent, VRAI, nombre_points, nombre_Y );
      if ( 2 == reponse ) {	/* On annule */
	GRF_eff_isole( resultat );
	free( table_Y ); table_Y = NULL;
	return NULL;
      }
    }
  } else {
    nombre_Y = nombre_points;
    table_Y = TABLE_Creer_Formule( table_X, nombre_points, Y->valeurs, nbr_parametres, parametres );
  }
  if ( NULL == table_Y ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur en voulant gnrer les Y\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( resultat );
    return NULL;
  }

  /* On remplit la table des Y */
  for ( i = 0 ; i < nombre_points ; i++ ) {
    if ( i < nombre_Y ) {
      resultat->y[ i ] = table_Y[ i ];
    } else {
      resultat->y[ i ] = 0.0;
    }
  }

  /* On n'a plus besoin de la table des Y */
  free( table_Y ); table_Y = NULL;

  /* On calcule les bornes */
  GRF_bornes( resultat, DESORDONNE, DESORDONNE );

  /* On insre le graphe dans la bonne srie */
  retour = GRF_InsererDansSerie( resultat, serie, avant );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'insrer le graphe dans la srie indique [erreur : %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    GRF_eff_isole( resultat ); resultat = NULL;
    return resultat;
  }

  /* On installe les lgendes et les units [aprs la srie, pour qu'il y ait les infos d'affichage !] */
  GRF_ChangerUnite  ( resultat, AXE_X, X->unite );
  GRF_ChangerLegende( resultat, AXE_X, X->nom, VRAI );

  GRF_ChangerUnite  ( resultat, AXE_Y, Y->unite );
  GRF_ChangerLegende( resultat, AXE_Y, Y->nom, VRAI );

  /* On a termin */
  return resultat;
}

/* -------- Gnration d'une srie de valeurs. Les X sont donc implicites.
    Renvoie NULL en cas de problme
            le graphe cr si tout a pu se faire correctement
   -------- */

struct graphe *GRF_GenererSerieValeurs(const char *nom, int type, struct info_variable *Y, struct info_variable *I,
				       unsigned long nbr_parametres, struct parametre *parametres,
				       struct graphe *avant, struct serie *serie)
{
  int retour;
  unsigned long nombre_points;
  double *table_Y = NULL;
  Graphe *resultat = NULL;

  if ( NULL == Y ) {
    (void) fprintf(stderr, "%s, %s l.%u : Informations pour Y manquantes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On prpare une table des valeurs en Y */
  if ( Y->type != VARIABLE_FORMULE ) {
    table_Y = VAR_GenererTable( Y, &nombre_points );
  } else {
    if ( NULL == I ) {
      (void) fprintf(stderr, "%s, %s l.%u : Y=f(I), mais I non dfinie !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* I : numro de point, donc min = 0, max = N, pas = 1 forcment */
    nombre_points = I->nombre;

    /* On gnre une table  partir de la variable I */
    table_Y = TABLE_Creer_Formule( NULL, nombre_points, Y->valeurs, nbr_parametres, parametres );
  }

  /* On contrle que l'on a pu prparer la table des Y ! */
  if ( NULL == table_Y ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de prparer la table des Y\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On en fait le graphe */
  resultat = GRF_nv_isole( nombre_points, type, GRF_SERIE );

  /* On contrle que l'on a pu prparer le graphe */
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de prparer le graphe %s - Opration annule\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
    return NULL;
  }

  /* On permute les Y du graphe et ceux de la table */
  free( resultat->y ); resultat->y = table_Y;

  /* On calcule les bornes */
  GRF_bornes( resultat, ORDONNE, DESORDONNE );

  /* On insre le graphe dans la bonne srie */
  retour = GRF_InsererDansSerie( resultat, serie, avant );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'insrer le graphe dans la srie indique [erreur : %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    GRF_eff_isole( resultat ); resultat = NULL;
    return resultat;
  }

  /* On installe les noms, lgendes et les units [aprs l'insertion en srie, pour que les infos d'affichage existent] */
  GRF_titre( resultat, nom );

  GRF_ChangerUnite  ( resultat, AXE_Y, Y->unite );
  GRF_ChangerLegende( resultat, AXE_Y, Y->nom, VRAI );

  /* On a termin */
  return resultat;
}

/* -------- Graphe 2D y = f(x) -------- */

Graphe *gen_grf_2D_cart(double x_min, double pas, unsigned long nbr_pts, Expression *fonction)
{
  unsigned long i;
  register double *x, *y;
  Graphe *graphe;

  if (fonction == NULL) return NULL;
  graphe = GRF_nouveau( NULL, nbr_pts, GR_AUTRETYPE, GR_NORMAL, "Sans nom / Y=F(X)" );
  if (graphe == NULL) return NULL;

  Var_T.valeur_initiale = 0.0;
  Var_Y.valeur_initiale = 0.0;

  for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
    double f_x = 0.0;

    *x = x_min + i * pas;
    Var_I.valeur_initiale = (double) i;
    Var_X.valeur_initiale = *x;
    Var_R.valeur_initiale = *x; /* Dans le cas polaire */

    if ( calcule_arbre(fonction, &f_x) == VRAI )
      *y = f_x;
  }

  GRF_bornes(graphe, ORDONNE, DESORDONNE);
  return graphe;
}

/* -------- Graphe 2D paramtrique (x = f(t), y = g(t)) -------- */

Graphe *gen_grf_2D_Pcart(double t_min, double pas, unsigned long nbr_pts, Expression *fct_X, Expression *fct_Y)
{
  unsigned long i;
  register double *x, *y;
  Graphe *graphe;

  if ( (fct_X == NULL) || (fct_Y == NULL) ) return NULL;
  graphe = GRF_nouveau( NULL, nbr_pts, GR_AUTRETYPE, GR_NORMAL, "Sans nom / X=f(T), Y=g(T)" );
  if (graphe == NULL) return NULL;

  Var_X.valeur_initiale = 0.0;
  Var_Y.valeur_initiale = 0.0;

  for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
    double f = 0.0, g = 0.0;

    Var_T.valeur_initiale = t_min + i * pas;
    Var_I.valeur_initiale = (double) i;
    Var_X.valeur_initiale = *x;

    if ( calcule_arbre(fct_X, &f) == VRAI )
      *x = f;
    if ( calcule_arbre(fct_Y, &g) == VRAI )
      *y = g;
  }

  GRF_bornes(graphe, DESORDONNE, DESORDONNE);
  return graphe;
}
