/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2003, Emmanuel CURIS

   Trier les donnes dans le graphe
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "math.h"

#include "graphes.h"
#include "tri.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_Trier_P     (Graphe *graphe);
static int GRF_Trier_Rapide(Graphe *graphe);

static int GRF_TrierComplexe(Graphe *graphe);
static int GRF_TrierReel    (Graphe *graphe);

static void tri_Rfusion(Graphe *graphe, double *x_tri, double *y_tri,
			unsigned long i1, unsigned long i2, unsigned long i3);
static void tri_Cfusion(Graphe *graphe, double *x_tri, Complexe *y_tri,
			unsigned long i1, unsigned long i2, unsigned long i3);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Tri en place d'un graphe par X croissants --------\
 |                                                            |
 | graphe  : le graphe que l'on veut trier                    |
 | methode : quelle mthode de tri utiliser ?                 |
 \----------------------------------------------------------- */

int GRF_Trier(Graphe *graphe, int methode)
{
  int retour;

  if ( graphe == NULL ) return 0; /* Pas de tri si pas de graphe... */

  /* On contrle la mthode utilise */
  if ( graphe->erreur_x != NULL ) {
    methode = GRF_TRI_UNIVERSEL;
  }
  if ( graphe->erreur_y != NULL ) {
    methode = GRF_TRI_UNIVERSEL;
  }

  /* Historique */
  HIST_Titre(4, "Tri", "Tri");
  HIST_Texte(4, 0, "Graphe tri"e_AIGU" : %s\n", graphe->nom);

  /* On fait le tri, en fonction de la mthode */
  switch( methode )
  {
  case GRF_TRI_UNIVERSEL:
    /* Tri en place par la mthode de permutations */
    retour = GRF_Trier_P( graphe );
    break;

  case GRF_TRI_RAPIDE:
    retour = GRF_Trier_Rapide( graphe );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Mthode de tri inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, methode);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    retour = -5;
    break;
  }

  if ( retour != 0 ) {
    if ( methode != GRF_TRI_UNIVERSEL ) {
      (void) fprintf(stderr, "%s,%s l.%u : Tri astucieux impossible [code %d]. On se rabat sur le tri universel...\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      retour = GRF_Trier_P( graphe );
    }
  }

  /*  Tri termin */
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de trier le graphe [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  } else {
    graphe->drapeaux.X_tries = VRAI;  /* Les X sont maintenant tris par ordre croissant ! */
  }
  return retour;
}

/* -------- Tri total d'un graphe, sur place, mthode de permutations ----------\
 |      Suppose que le graphe est presque tri pour tre efficace               |
 |     (c'est le cas des spectres exprimentaux, en gnral ! donc utilisable)  |
 |                                                                              |
 | graphe  : le graphe que l'on veut trier                                      |
 \----------------------------------------------------------------------------- */

static int GRF_Trier_P(Graphe *graphe)
{
  switch( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
    return GRF_TrierReel( graphe );

  case GRF_FCT_COMPLEXE:
    return GRF_TrierComplexe( graphe );

  case GRF_SERIE:
/*     return GRF_TrierSerie( graphe ); */

  case GRF_COURBE_PLANE:
  case GRF_COURBE_ESPACE:
  case GRF_SURFACE:
    (void) fprintf(stderr, "GRF_Trier : Tri impossible pour ce type de graphes [%d]\n", graphe->genre_donnees);
    return -1;

  default:
    (void) fprintf(stderr, "GRF_Trier : Type de graphe inconnu [%d]\n", graphe->genre_donnees);
  }

  return -2;
}

/* -------- Tri en place d'un graphe par X croissants : valeurs complexes -------- */

static int GRF_TrierComplexe(Graphe *graphe)
{
  double *x, *x_suivant;
  unsigned long i;
  int non_trie;

#ifdef _DEVELOPPE_
  unsigned long nbr_tours = 0;
#endif

  do {
    non_trie = FAUX; /* On recommence un tour... */
    x = graphe->x; x_suivant = x + 1;
    for (i = 1 ; i < graphe->nombre_point ; i++, x++, x_suivant++) {
      if (*x_suivant < *x) { /* Problme de tri */
	double ech_x = *x_suivant;
	Complexe y;

	*x_suivant = *x; *x = ech_x; /* On change les X */
	non_trie = VRAI; /* Le spectre n'tait pas tri */

	y = ((Complexe *) graphe->y)[i];
	((Complexe *) graphe->y)[i]     = ((Complexe *) graphe->y)[i - 1];
	((Complexe *) graphe->y)[i - 1] = y;

	/* On s'occupe des ventuelles incertitudes associes
	   Les X des graphes d'erreur, s'ils existent, sont juste figuratifs -> on ne s'en occupe pas ! */
	if ( graphe->erreur_x != NULL ) {
	  int ddl;
	  double sx;

	  sx = (graphe->erreur_x->y)[ i ];
	  (graphe->erreur_x->y)[ i ] = (graphe->erreur_x->y)[ i - 1 ];
	  (graphe->erreur_x->y)[ i - 1 ] = sx;

	  ddl = (graphe->erreur_x->z.ddl)[ i ];
	  (graphe->erreur_x->z.ddl)[ i ]= (graphe->erreur_x->z.ddl)[ i - 1 ];
	  (graphe->erreur_x->z.ddl)[ i - 1 ] = ddl;
	}

	if ( graphe->erreur_y != NULL ) {
	  int ddl;

	  y = ( (Complexe *) graphe->erreur_y->y)[ i ];
	  ( (Complexe *) graphe->erreur_y->y)[ i ] = ( (Complexe *) graphe->erreur_y->y)[ i - 1 ];
	  ( (Complexe *) graphe->erreur_y->y)[ i - 1 ] = y;

	  ddl = (graphe->erreur_x->z.ddl)[ i ];
	  (graphe->erreur_y->z.ddl)[ i ]= (graphe->erreur_y->z.ddl)[ i - 1 ];
	  (graphe->erreur_y->z.ddl)[ i - 1 ] = ddl;
	}
      }
    }
#ifdef _DEVELOPPE_
    nbr_tours++;
#endif
  } while (non_trie == VRAI);

#ifdef _DEVELOPPE_
  printf("GRF_TrierComplexe : %lu points, tris en %lu passages [%s]\n", graphe->nombre_point, nbr_tours, graphe->nom);
#endif

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Tri en place d'un graphe par X croissants : valeurs relles -------- */

static int GRF_TrierReel(Graphe *graphe)
{
  double *x, *x_suivant;
  unsigned long i;
  int non_trie;

#ifdef _DEVELOPPE_
  unsigned long nbr_tours = 0;
#endif

  do {
    non_trie = FAUX; /* On recommence un tour... */
    x = graphe->x; x_suivant = x + 1;
    for (i = 1 ; i < graphe->nombre_point ; i++, x++, x_suivant++) {
      if (*x_suivant < *x) { /* Problme de tri */
	double ech_x = *x_suivant, y;

	*x_suivant = *x; *x = ech_x; /* On change les X */
	non_trie = VRAI; /* Le spectre n'tait pas tri */

	y = (graphe->y)[ i ];
	(graphe->y)[ i ] = (graphe->y)[ i - 1 ];
	(graphe->y)[ i - 1 ] = y;

	/* On s'occupe des ventuelles incertitudes associes
	   Les X des graphes d'erreur, s'ils existent, sont juste figuratifs -> on ne s'en occupe pas ! */
	if ( graphe->erreur_x != NULL ) {
	  int ddl;

	  y = (graphe->erreur_x->y)[ i ];
	  (graphe->erreur_x->y)[ i ] = (graphe->erreur_x->y)[ i - 1 ];
	  (graphe->erreur_x->y)[ i - 1 ] = y;

	  ddl = (graphe->erreur_x->z.ddl)[ i ];
	  (graphe->erreur_x->z.ddl)[ i ]= (graphe->erreur_x->z.ddl)[ i - 1 ];
	  (graphe->erreur_x->z.ddl)[ i - 1 ] = ddl;
	}

	if ( graphe->erreur_y != NULL ) {
	  int ddl;

	  y = (graphe->erreur_y->y)[ i ];
	  (graphe->erreur_y->y)[ i ] = (graphe->erreur_y->y)[ i - 1 ];
	  (graphe->erreur_y->y)[ i - 1 ] = y;

	  ddl = (graphe->erreur_x->z.ddl)[ i ];
	  (graphe->erreur_y->z.ddl)[ i ]= (graphe->erreur_y->z.ddl)[ i - 1 ];
	  (graphe->erreur_y->z.ddl)[ i - 1 ] = ddl;
	}
      }
    }
#ifdef _DEVELOPPE_
    nbr_tours++;
#endif
  } while (non_trie == VRAI);

#ifdef _DEVELOPPE_
/*   printf("GRF_TrierReel : %lu points, tris en %lu passages [%s]\n", graphe->nombre_point, nbr_tours, graphe->nom); */
#endif

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~- Tri d'un graphes selon les X avec rarrangement en parallle des Y ~-~-~-~-~-~- */

/* -------- Tri total d'un graphe en utilisant le tri par fusion ----------\
 |      (mme algorithme que pour le tri d'un tableau, cf. ci-dessus       |
 |   Renvoie VRAI si le tri a pu tre effectu, FAUX sinon                 |
 \------------------------------------------------------------------------ */

static int GRF_Trier_Rapide(Graphe *graphe)
{
  double *x_tri; /* Position des points tris */

  if (graphe == NULL) return FAUX; /* Pas de graphe */

  x_tri = (double *) calloc( graphe->nombre_point, sizeof(double) );
  if (x_tri == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Stats/tri.c", "tri_graphe", "x_tri");
    return FAUX;
  }

  if ( GRF_FCT_REELLE == graphe->genre_donnees ) { /* Graphe  valeurs relles */
    double *y_tri, *temp;
    unsigned long j = 1, i;

    y_tri = (double *) calloc( graphe->nombre_point, sizeof(double) );
    if (y_tri == NULL) {
      SYS_ERREUR(MANQUE_MEMOIRE, "Stats/tri.c", "tri_graphe", "y_tri / r"e_AIGU"el");
      free(x_tri);
      return FAUX;
    }

    while (j < graphe->nombre_point) {
      for (i = 0 ; i < graphe->nombre_point ; i += 2 * j) {
	tri_Rfusion(graphe, x_tri, y_tri, i, i + j, i + 2 * j);
      }
      j <<= 1;

      temp = x_tri; x_tri = graphe->x; graphe->x = temp;
      temp = y_tri; y_tri = graphe->y; graphe->y = temp;
    }

    free(y_tri);
  } else { /* Graphe  valeurs complexes */
    Complexe *y_tri, *C_temp;
    double *d_temp;
    unsigned long j = 1, i;

    y_tri = (Complexe *) calloc( graphe->nombre_point, sizeof(Complexe) );
    if (y_tri == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "Stats/tri.c", "tri_graphe", "y_tri / complexe");
      free(x_tri);
      return FAUX;
    }

    while (j < graphe->nombre_point) {
      for (i = 0 ; i < graphe->nombre_point ; i += 2 * j) {
	tri_Cfusion(graphe, x_tri, y_tri, i, i + j, i + 2 * j);
      }
      j <<= 1;

      d_temp = x_tri; x_tri = graphe->x; graphe->x = d_temp;
      C_temp = y_tri; y_tri = (Complexe *) graphe->y; graphe->y = (double *) C_temp;
    }

    free(y_tri);
  }

  free(x_tri);
  return VRAI; /* Le tri a russi */
}

/* -------- Fusion de deux zones prtris, le rsultat est tri par ordre croissant ----------\
 |    Code adapt de l'exemple fourni par Bruno Blanchet                                      |
 |                                                                                            |
 | graphe     = le graphe  trier 			                                      |
 | nv_x, nv_y = adresse o crire les donnes tries                                          |
 | i1, i2     = bornes du premier morceau  fusionner (fusionne [i1, i2[ et [i2, i3[)         |
 | i2, i3     = bornes du second morceau  fusionner                                          |
 \------------------------------------------------------------------------------------------- */

static void tri_Rfusion(Graphe *graphe, double *x_tri, double *y_tri,
			unsigned long i1, unsigned long i2, unsigned long i3)
     /* Version pour graphes rels */
{
  unsigned long i,j,k;
  double x_1, x_2, y_1, y_2;

  /* On vrifie qu'on ne dpasse pas le tableau */
  if (i2 > graphe->nombre_point) i2 = graphe->nombre_point;
  if (i3 > graphe->nombre_point) i3 = graphe->nombre_point;
 
  i = i1; j = i2;
  if (i < i2) {
    x_1 = graphe->x[i];
    y_1 = graphe->y[i];
  } else {
    x_1 = HUGE_VAL;
    y_1 = HUGE_VAL;
  }
  if (j < i3) {
    x_2 = graphe->x[j];
    y_2 = graphe->y[j];    
  } else {
    x_2 = HUGE_VAL;
    y_2 = HUGE_VAL;
  }

  for (k = i1 ; k < i3 ; k++) {
    if (x_1 < x_2) {
      x_tri[k] = x_1;
      y_tri[k] = y_1;
      if (++i < i2) {
	x_1 = graphe->x[i];
	y_1 = graphe->y[i];
      } else {
	x_1 = HUGE_VAL;
	y_1 = HUGE_VAL;
      }
    } else {
      x_tri[k] = x_2;
      y_tri[k] = y_2;
      if (++j < i3) {
	x_2 = graphe->x[j];
	y_2 = graphe->y[j];    
      } else {
	x_2 = HUGE_VAL;
	y_2 = HUGE_VAL;
      }
    }
  }
}

static void tri_Cfusion(Graphe *graphe, double *x_tri, Complexe *y_tri,
			unsigned long i1, unsigned long i2, unsigned long i3)
     /* Version pour graphes complexes */
{
  unsigned long i,j,k;
  double x_1, x_2;
  Complexe y_1 = { 0.0, 0.0 }, y_2 = { 0.0, 0.0 };  /* Init. pour viter un avertissement avec gcc 4.0 */

  /* On vrifie qu'on ne dpasse pas le tableau */
  if (i2 > graphe->nombre_point) i2 = graphe->nombre_point;
  if (i3 > graphe->nombre_point) i3 = graphe->nombre_point;
 
  i = i1; j = i2;
  if (i < i2) {
    x_1 = graphe->x[i];
    y_1 = ( (Complexe *) graphe->y)[i];
  } else {
    x_1 = HUGE_VAL;
    /* Valeur de y_1 inutile car x_1 sera forcment le plus grand des deux */
  }
  if (j < i3) {
    x_2 = graphe->x[j];
    y_2 = ( (Complexe *) graphe->y)[j];
  } else {
    x_2 = HUGE_VAL;
    /* Valeur de y_2 inutile car x_2 sera forcment le plus grand des deux */
  }

  if ( ( x_1 == x_2 ) && ( x_1 == HUGE_VAL ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Initialisation trs bizarre [i1=%lu i2=%lu i3=%lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, i1, i2, i3);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }

  for (k = i1 ; k < i3 ; k++) {
    if (x_1 < x_2) {
      x_tri[k] = x_1;
      y_tri[k] = y_1;
      if (++i < i2) {
	x_1 = graphe->x[i];
	y_1 = ( (Complexe *) graphe->y)[i];
      } else {
	x_1 = HUGE_VAL;
      }
    } else {
      x_tri[k] = x_2;
      y_tri[k] = y_2;
      if (++j < i3) {
	x_2 = graphe->x[j];
	y_2 = ( (Complexe *) graphe->y)[j];
      } else {
	x_2 = HUGE_VAL;
      }
    }
  }
}
