/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Transformation d'une image pour tenir compte des effets d'paisseur et d'orientation de l'chantillon
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"

#include "exafs.h"
#include "images.h"
#include "science.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Interface pour les effets d'paisseur de l'chantillon et d'inclinaison
     Renvoie l'image obtenue, NULL en cas d'erreur ou d'annulation

     source : contient les donnes pour une tranche infiniment fine que l'on suppose reprsentative
     de toute l'paisseur
   -------- */

Image *effets_epaisseur(Image *source)
{
  Image *image;
  double epais_coupe, epais_totale;
  double angle_faisceau, inclinaison, rotation; /* Orientation de l'chantillon */
  int retour;

  if (source == NULL) return NULL;

  retour = ARB_Bloquant( IMG_EFFETS_3D, "Effets d'"e_AIGU"paisseur", principale, 0 );
  if (retour == IE3D_ANNULER)
    return NULL;

  epais_coupe  = OBJ_LireNombre(IMG_EFFETS_3D, IE3D_EPAIS_COUPE, NULL );
  epais_totale = OBJ_LireNombre(IMG_EFFETS_3D, IE3D_EPAIS_ECH, NULL );

  if (epais_totale < 2.0 * epais_coupe) {
    fprintf(stderr, "Au moins deux tranches ncessaires pour que le rsultat ait un sens..\n");
    return NULL;
    /* En fait, on pourrait aussi faire une simple projection mais bon... */
  }

  OPTION_A_FAIRE( "Convertir en LireAngle" );
  angle_faisceau = OBJ_LireNombre(IMG_EFFETS_3D, IE3D_ANGLE_FAIS, NULL );
  inclinaison = OBJ_LireNombre(IMG_EFFETS_3D, IE3D_ANGLE_NORM, NULL );
  rotation = OBJ_LireNombre(IMG_EFFETS_3D, IE3D_ANGLE_HOR, NULL );

  image = ECH_radial(source, epais_coupe, epais_totale, angle_faisceau, inclinaison, rotation);
  return image;
}

/* -------- Modle d'chantillon  radial par empilement 
   On empile l'image indiqu comme tranche unitaire suivant un axe.
   Idalement, cet axe est la normale au plan illumin par le faisceau.
   On peut indiquer deux corrections :
       - l'inclinaison par rapport  cette normale, \  /<- normale
                                                     \/_____
                                                      \

       - la rotation par rapport  l'horizontale pour se placer sur la droite

   On calcule en fin de compte la somme obtenue en traversant tout a
   paralllement au faisceau, point par point. On suppose ici le
   faisceau ponctuel.
*/

Image *ECH_radial(Image *tranche, double epais_tranche, double epais_totale, 
		  double angle_faisceau, double inclinaison, double rotation)
{
  unsigned long larg_image, haut_image, n_case;
  unsigned int x, y, nbr_tranches;
  double lgreel_source, htreel_source, lgreel_final, htreel_final;
  double delta_x, delta_y, d_x, d_y, d_z;
  Image *image;

  if (tranche == NULL) return NULL;

  /* Conversion des angles en radians */
  angle_faisceau *= M_PI / 180.0;
  inclinaison *= M_PI / 180.0;
  rotation *= M_PI / 180.0;

  /* Importance des dbordements */
  d_x = tan(angle_faisceau) + cos(rotation) * sin(inclinaison); /* Incrments unitaires */
  d_y = sin(rotation) * sin(inclinaison);
  delta_x = epais_totale * d_x;
  delta_y = epais_totale * d_y;

  /* Dimensions de l'image finale, compte-tenu des dbordements */
  lgreel_source = tranche->larg * tranche->l_reel;
  htreel_source = tranche->haut * tranche->h_reel;
  lgreel_final = lgreel_source + fabs(delta_x);
  htreel_final = htreel_source + fabs(delta_y);

  larg_image = (unsigned long) rint(lgreel_final / tranche->l_reel); /* Mme unit de cases */
  haut_image = (unsigned long) rint(htreel_final / tranche->h_reel);

  image = IMG_CreerImage( NULL, tranche, tranche->genre, tranche->type, larg_image, haut_image);
  if (image == NULL)
    return NULL;
  image->l_reel = tranche->l_reel;
  image->h_reel = tranche->h_reel;

  image->x_min = tranche->x_min - delta_x;
  image->y_min = tranche->y_min - delta_y;

  /* Dcomposition en tranches */
  nbr_tranches = (unsigned int) rint(epais_totale / epais_tranche);
  d_z = epais_totale / (double) nbr_tranches;

  for (x = 0 ; x < image->larg ; x++) {
    double x_reel;

    x_reel = x * image->l_reel + image->x_min;
    for (y = 0 ; y < image->haut ; y++) {
      double y_reel;
      unsigned int z;

      IMG_IJ_v_N( image, x, y, &n_case );
/*       image->valeurs[ IMG_IJ_v_N( image, x, y ) ] = 0.0; */
      y_reel = y * image->h_reel + image->y_min;
      for (z = 0 ; z < nbr_tranches ; z++) {
	double xo, yo;
	unsigned long valeur;

	xo = x_reel + d_x * z * d_z;
	yo = y_reel + d_y * z * d_z;
	valeur = IMG_lit_valeur(tranche, xo, yo, IMG_INTERPOLE_BILINEAIRE, NULL);
	image->valeurs.R[ n_case ] += valeur;
      }
    }
  }

  /* On renormalise par rapport  l'original */
  IMG_FacteurMultiplicatif( image, 1.0 / nbr_tranches, AXE_Z );

  /* On tient compte du dcentrage */
/*  image->x_centre -= delta_x / 2.0;
  image->y_centre -= delta_y / 2.0;*/

  IMG_bornes(image);
  return image;
}
