/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2003, Emmanuel CURIS
   Cration : 13 juin 2003

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Sauvegarde des images
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "bin_utils.h"

#include "fichiers.h"
#include "images.h"

#include "img_sauver.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define VERIFIE(n,t,r) if ( n_ecrit < (n) ) {\
                         (void) SYS_ERREUR(ERREUR_ECRITURE, "Fichier en cours", fichier, NULL); \
                         (void) fprintf(stderr, "VERIFIE, sauvegarde, %s : %s\n", __FUNCTION__, t); \
                         return (r); }

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int SV_ImageDeSerie(FILE *fichier, Serie_Images *serie, Image *image);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Sauvegarde d'une srie d'images : interface pour fichier isol... -------- */

int IMG_SERIE_SauverIsolee(Serie_Images *serie, const char *nom_fichier, unsigned int options)
{
  FILE *fichier;
  int retour;

  if ( ( NULL == serie ) || ( NULL == nom_fichier ) ) return -1;

  /* On ouvre le fichier */
  fichier = fopen( nom_fichier, "wb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir le fichier \"%s\" en criture\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    (void) SYS_ERREUR( ERREUR_ECRITURE, nom_fichier, NULL, NULL );
    return -2;
  }

  /* On prpare l'en-tte */
  if ( SV_EnTete( fichier, FICHIER_SERIE_IMAGE ) == FAUX ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_EnTete\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  retour = IMG_SERIE_Sauver( serie, fichier, options );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_Sauver [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -4;
  }

  /* Plus besoin du fichier */
  (void) fclose(fichier);

  /* On mentionne la sauvegarde dans l'historique, puisqu'elle a russi */
  HIST_Titre(1, "Sauvegarde", "E/S");
  HIST_Texte(1, 0, "Fichier : %s\n", nom_fichier);
  HIST_Texte(1, 0, "Srie d'images sauv"e_AIGU"e : %s\n", serie->nom);

  /* Sauvegarde russie */
  return 0;
}

/* -------- Sauvegarde d'une srie d'images, comme bloc dans un fichier
     Renvoie 0 si a pu sauver, !=0 en cas de problme

   Format utilis :
   << chaine : le nom de la srie >>
   nbr_images                   - unsigned long, nombre d'images
   <<image 1>>
   <<image 2>>
   ...
   <<image n>>
 */

int IMG_SERIE_Sauver(Serie_Images *serie, FILE *fichier, unsigned int options)
{
  int retour;
  int n_ecrit;

  if ( NULL == fichier ) return -1;
  if ( NULL == serie ) return -2;

  /* Le nom de la srie : une chane de caractres */
  retour = BIN_SV_chaine( fichier, serie->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant le nom\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  if ( 0 == options ) {
    Image *image;

    /* On sauve tous les graphes */
    n_ecrit = fwrite( &(serie->nbr_images), sizeof(unsigned long), 1, fichier );
    VERIFIE(1, "Nombre d'images [toute la srie]", -12);

    for ( image = serie->premiere_image ; image != NULL ; image = image->suivante ) {
      retour = SV_ImageDeSerie( fichier, serie, image );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Sauvegarde du image %s\n",
		       __FILE__, __FUNCTION__, __LINE__, image->nom);
	return -13;
      }
    }
  } else {
    OPTION_A_FAIRE( "Sauver les images avec un filtre indiqu !" );
    return +255;
  }

  /* Sauvegarde du bloc finie, tout s'est bien pass */
  return 0;
}

/* -------- Sauvegarde d'une image, quand elle fait partie de la sauvegarde d'une srie
   Isol car revient souvent dans sv_serie()
   -------- */

static int SV_ImageDeSerie(FILE *fichier, Serie_Images *serie __attribute__((__unused__)), Image *image)
{
  int retour = 0;

  /* Ssauvegarde directe */
  retour = IMG_Sauver( fichier, image );

  return retour;
}

/* -------- Sauvegarde d'une image : interface pour fichier isol... -------- */

int IMG_SauverIsolee(const char *nom_fichier, Image *image)
{
  FILE *fichier;
  int retour;

  if ( ( NULL == image ) || ( NULL == nom_fichier ) ) return -1;

  /* On ouvre le fichier */
  fichier = fopen( nom_fichier, "wb" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir le fichier \"%s\" en criture\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    (void) SYS_ERREUR( ERREUR_ECRITURE, nom_fichier, NULL, NULL );
    return -2;
  }

  /* On prpare l'en-tte */
  if ( SV_EnTete( fichier, FICHIER_IMAGE ) == FAUX ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans SV_EnTete\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  retour = IMG_Sauver( fichier, image );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_Sauver [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -4;
  }

  /* Plus besoin du fichier */
  (void) fclose(fichier);

  /* On mentionne la sauvegarde dans l'historique, puisqu'elle a russi */
  HIST_Titre(1, "Sauvegarde", "E/S");
  HIST_Texte(1, 0, "Fichier : %s\n", nom_fichier);
  HIST_Texte(1, 0, "Image sauv"e_AIGU"e : %s\n", image->nom);

  /* Sauvegarde russie */
  return 0;
}

/* -------- Sauvegarde d'une image [donnes brutes] --------

int              Genre des donnes [VERSION 3.2]
int              Type d'image
<< chaine >>     Nom de l'image
2 * unsigned int Nombre de valeurs (larg=X, haut=Y)
2 * double       Pas en X, en Y
2 * double       Position du coin infrieur gauche dans le repre absolu (x,y)
2 * double       Valeurs extrmes (min, max)
Suivant le genre
  << table D >>    Valeurs  IMG_GENRE_REELLE
  << table C >>    Valeurs  IMG_GENRE_COMPLEXE
  << n*graphes >>  Valeurs  IMG_GENRE_GRAPHE
*/

int IMG_Sauver(FILE *fichier, Image *image)
{
  int retour;

  if ( NULL == fichier ) return -1;
  if ( NULL == image ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'image\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Le genre des donnes dans l'image */
  fwrite( &(image->genre), sizeof( int ), 1, fichier );

  /* Le type d'image */
  fwrite( &(image->type), sizeof( int ), 1, fichier );

  /* Le nom */
  retour = BIN_SV_chaine( fichier, image->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans BIN_SV_Chaine(%s) [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, image->nom, retour);
    return -3;
  }

  /* Nombre de valeurs en X et Y */
  fwrite( &(image->larg), sizeof( unsigned int ), 1, fichier );
  fwrite( &(image->haut), sizeof( unsigned int ), 1, fichier );

  /* Tailles de la grille */
  fwrite( &(image->l_reel), sizeof( double ), 1, fichier );
  fwrite( &(image->h_reel), sizeof( double ), 1, fichier );

  /* Position du coin bas-gauche dans le repre absolu */
  fwrite( &(image->x_min), sizeof( double ), 1, fichier );
  fwrite( &(image->y_min), sizeof( double ), 1, fichier );

  /* Valeurs extrmes */
  fwrite( &(image->val_min), sizeof( double ), 1, fichier );
  fwrite( &(image->val_max), sizeof( double ), 1, fichier );

  /* Table des valeurs */
  if ( IMG_GENRE_REELLE == image->genre ) {
    fwrite( image->valeurs.R, sizeof( double ), image->larg * image->haut, fichier );
  } else if ( IMG_GENRE_COMPLEXE == image->genre ) {
    fwrite( image->valeurs.C, sizeof( Complexe ), image->larg * image->haut, fichier );
  } else if ( IMG_GENRE_GRAPHE == image->genre ) {
    unsigned long i, nbr_valeurs;

    nbr_valeurs = image->larg * image->haut;
    for ( i = 0 ; i < nbr_valeurs ; i++ ) {
      retour = SV_Graphe( fichier, image->valeurs.G[ i ] );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR en sauvant le graphe %lu\n",
		       __FILE__, __FUNCTION__, __LINE__, i);
	return -20;
      }
    }
  } else if ( IMG_GENRE_DIAGRAMME == image->genre ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE / diagramme\n",
		   __FILE__, __FUNCTION__, __LINE__);
  } else if ( IMG_GENRE_PHOTO == image->genre ) {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE / photo\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }


  /* Sauvegarde russie */
  return 0;
}
