/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 1998
   Dernière modification : 17 novembre 1998

   Définitions pour l'utilisation de modèles structuraux

   L'unité fondamentale est la molécule (équivalente au graphe).
   Chaque molécule comprend une liste chaînée d'atomes, qui lui sont uniques.
   Les atomes peuvent être groupés en fragments, propres à la molécule
*/

#ifndef ATOMES_H_
# define ATOMES_H_ /* Protection contre l'inclusion multiple */

#include "chimie_utils.h" /* Pour les numéros atomiques */
#include "c_formel.h"     /* Pour Variable, utilisé dans Absorbeur */
#include "utils_3d.h"     /* Pour Vecteur3D, coordonnées d'un point */

#define Z_ABSORBEUR_MINIMAL POTASSIUM /* Les éléments légers (trois premières périodes)
					 ne sont pas considérés comme pouvant être
					 des absorbeurs EXAFS plausibles */

/* Note : l'atome est la brique constitutive, dans le programme, des structures.
   Les structures sont groupées en une sur-unité : la molécule, qui est celle manipulée par l'utilisateur
   Toutes les données structurales sont données dans un repère absolu, orthonormé,
   défini par la première série de coordonnées entrée par une méthode quelconque.
   Toutes les autres sont donc supposées être dans le même repère.
   L'unité de base du repère est l'angstrom; pour des raisons de commodité,
   les rayons atomiques sont supposés exprimés en picomètres (à cause de ATM_infos...)

   Les atomes forment une liste chaînée par molécule et sont uniques pour une molécule ;
   les autres unités constitutives pointent sur la sous-partie de cette liste chaînée
   qui correspond aux atomes qui les forment.

   Exemple : si l'on charge un dipeptide NH2-Gly-Ala-COOH, on aura l'agencement suivant en mémoire
   (si l'on suppose que les H ne sont pas disponibles ; <--- signifie pointe sur pour un atome ;
    <~~~~ Signifie pointe sur pour une liaison), pour la molécule créée :

   Atomes           Fragments         Molécule
    ________________________________________
    |                                      |
    v                                      |
   N terminal    <---\                     |
    |    \           |                     |
    |     ~~~~~~~~~~~~~~~~~~\              |
    v    /           |      |              |
   C alpha (Gly) <---+-fragment Gly <------+ [Gly : 4 atomes, 3 liaisons]
    |    \           |      |              |
    |     ~~~~~~~~~~~~~~~~~~|              |
    v    /           |      |              |
   C du COO      <---|      |              |
    |    \           |      |              |
    |     ~~~~~~~~~~~~~~~~~~/              |
    v    /           |                     |
   O du C=O      <---/                     |
    |        \                             |
    |         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+-Dipeptide H N-Gly-Ala-COOH
    v        /                             |            2        
   N liaison     <---\                     |  [1 liaison, 2 fragments]
    |    \           |                     |
    |     ~~~~~~~~~~~~~~~~~~\              |
    v    /           |      |              |
~> C alpha (Ala) <---+-fragment Ala <------/ [Ala : 5 atomes, 5 liaisons]
|   |    \           |      |
|   |     ~~~~~~~~~~~~~~~~~~|
|   v    /           |      |
|  C bêta (Ala)  <---|      |
|   |                |      |
~~~~~~~~~~~~~~~~~~~~~~~~~~~~|
|   v                |      |
~> C du COO      <---| <~\  |
    |    \           |   |  |
    |     ~~~~~~~~~~~~~~~~~~|
    v    /           |   |  |
   O du C=O      <---|   |~~/
    |                |   |
    v                |   |
   O terminal    <---/ <~/

   Avantage supposé : permet de faire des fragments virtuels = sélection d'une région et tout et tout
                      devrait aussi permettre, en créant de nouvelles choses, de définir des entités
		      plus compatibles avec des solides cristallisés (type ZnS,...) en utilisant
		      des symétries de maille sans avoir à, nécessairement, refaire les atomes
		      - à voir quand la nécessité apparaîtra... 

   Inconvénient supposé : même repère, nécessairement, pour toutes les molécules et tous les fragments
           Remède : correction globale par fragment en définissant un changement de repère / fragment ?
	            pas encore fait car utilité pas encore rencontrée
*/

struct fragment; /* Définie ultérieurement */

#define HYBRIDE_S          0 /* Pour l'hydrogène... */
#define HYBRIDE_SP3        1 /* aliphatique */
#define HYBRIDE_SP2        2 /* type éthylène */
#define HYBRIDE_SP         3 /* type acétylène */
#define HYBRIDE_AROMATIQUE 4 /* type benzène */

typedef struct atome
{
  struct atome  *suivant;      /* Atome précédent dans le modèle             */

  /* Caractéristiques de l'atome étudié */
  unsigned char  num_atomique; /* Numéro atomique de l'élément               */
  unsigned char  nbr_liaisons; /* Nombre de liaisons "informatiques" pour cet atome, pour accélérer l'effacement */
  unsigned char  nbr_coval;    /* Nombre de liaisons covalentes (en équivalents simples...) - pour contrôle de la valence */
  unsigned char  type_atome;   /* Hybridation, en gros... */

  Vecteur3D      position;     /* Coordonnées atomiques absolues             */
  Vecteur3D_Erreurs *err_position; /* Incertitudes sur la position [cristallographique ?]
				      pointeur, car beaucoup de structures n'en ont pas -> économise de la place :
				      1 pointeur = 8 octets, 6 doubles = 24 octets... */
  float          charge;       /* Charge (partielle) de l'atome              */
  float          rayon;        /* Rayon de l'atome considéré (exprimé en pm) */

  /* Normalement, tout ce qui concerne les liaisons est défini à l'aide
     de structures de type Liaison */

  /* Voir pour éventuellement rajouter un fondement quantique, pour les calculs théoriques
     Soit au moins une base d'OA, un rayon de muffin-tin, un potentiel... */

  struct fragment *fragment; /* Pour faciliter la recherche du fragment dans lequel se trouve l'atome
				Le fragment indiqué ici est celui chimiquement logique, mais d'autres
				fragments peuvent aussi contenir l'atome (sélection, ... ) */

  /* Plus spécifique de la modélisation */
  Vecteur3D position_ajustee;  /* Position estimée après ajustement. [si flag avec_modele mis] */

  /* nécessaires pour le fonctionnement interne */
  unsigned int nbr_modeles; /* Nombre de modèles auxquels appartient l'atome */
  unsigned long index; /* 0 en général, utilisé p. ex. dans chargement PDB pour préparer
			  la réalisation des liaisons */
  struct {
    unsigned visible     :1; /* VRAI ssi l'atome est marqué visible */
    unsigned absorbeur   :1; /* VRAI ssi l'atome est utilisé comme absorbeur */
    unsigned multi_frg   :1; /* VRAI ssi l'atome appartient à plusieurs fragments */
    unsigned avec_modele :1; /* VRAI ssi l'atome a des coordonnées estimées après modèle (x_aj a un sens) */
    unsigned mk_3 :1;
    unsigned mk_2 :1;
    unsigned mk_1 :1;
  } marques; /* Diverses marques pour les recherches éventuelles */

  struct {
    unsigned protonable : 1; /* VRAI ssi l'atome est protonable : base de Lewis */
    unsigned ac_Lewis   : 1; /* VRAI ssi l'atome est un centre acide de Lewis */
    unsigned squelette  : 1; /* VRAI ssi l'atome appartient au squelette de la chaîne polymère */
  } chimie;

  /* Voir pour une caractérisation du désordre de cet atome */

} Atome;

/* Les différents types de liaison pouvant exister
   Une liaison est en fait un lien entre n atomes (n >= 2, très souvent n = 2)
   caractérisé par un certain nombre de propriétés

   Les liaisons sont agencées en liste simplement chaînée, démarrant dans le fragment, la molécule ou
   directement dans le programme pour les liaisons intermoléculaires

   Dans l'état actuel, elles sont là par souci de complétude,
   mais elles devraient pouvoir servir à une recherche des contraintes
   (devrait aussi améliorer l'affichage, mais c'est surtout visuel...)

   Prévu : trois types de liaisons, 1) entre molécules [donc externes à toute structure]
                                    2) entre fragments [dans une liste chaînée simple démarrant dans
				                        une structure Molecule]
				    3) entre atomes    [dans une liste chaînée simple démarrant dans
				                        une structure Fragment]
*/

#define LIAISON_COVALENTE  0
#define LIAISON_IONIQUE    1
#define LIAISON_HYDROGENE  2
#define LIAISON_V_D_WAALS  3
#define LIAISON_METALLIQUE 4
#define LIAISON_REPULSIVE  5 /* Ne devrait avoir d'intérêt que pour des contraintes */

typedef struct liaison
{
  struct liaison *suivante;   /* La liaison suivante dans le fragment */

  float multiplicite; /* Caractéristiques de la liaison */
  double longueur; /* Longueur _à l'équilibre_ (pour une covalente) */
  double energie;
  double k_force; /* Constante de force, modèle de l'oscillateur harmonique */
  short int type; /* Type de liaison : covalente, ionique, métallique, H, VdW,... */

  unsigned short int nbr_atomes; /* Nombre d'atomes impliqués dans la liaison
				    Cas le plus fréquent : deux ! sauf modèles de liaisons à n centres */

  Atome *atome_1;                /* Les atomes impliqués dans la liaison */
  Atome *atome_2;
  Atome **autres_atomes; /* Pour les atomes en +, s'ils existent */

  /* Voir pour enregistrer une description quantique de la liaison :
     Dans ce cas, on assimilerait une liaison à une OM et il faudrait donc la base des OA
     ainsi que les coefficients de chaque OA dans le modèle */

} Liaison;

struct molecule; /* Sera définie par la suite */

Liaison *creer_liaison  (struct molecule *molecule, Atome *atm_1, Atome *atm_2, int multiplicite, int type);
void     effacer_liaison(Liaison *liaison, boolean forcer);

#define LG_MAX_NOM_FRAGMENT 95 /* Longueur maximale du nom d'un fragment -> +1 = 96 : multiple de 8, ok */

/* Définition d'un fragment : association 1) d'une famille d'atomes
                                          2) d'un ensemble de liaisons entre ces atomes
   Voir plus haut et plus bas pour plus de détails + des exemples sur le rôle de cette structure

   Chaque molécule pointe sur le premier élément de la liste chaînée de ses fragments
   Chaque fragment contient la liste des ADRESSES des atomes qui le constituent
*/

struct graphe; /* Défini dans graphes.h */

typedef struct fragment
{
  struct fragment *suivant;          /* Fragment suivant en mémoire 
					Liste simplement chaînée devrait suffire */
  char nom[LG_MAX_NOM_FRAGMENT + 1]; /* Nom du fragment */

  Vecteur3D pos_barycentre;          /* Coordonnées _absolues_ du barycentre */

  struct { /* Quelques drapeaux pouvant servir, parfois */
    unsigned maille    : 1; /* VRAI : le fragment correspond à une maille    */
    unsigned origine   : 1; /* VRAI : cette maille est la maille (0, 0, 0)   */
    unsigned ligand    : 1; /* VRAI : le fragment correspond à un ligand     */
    unsigned polydente : 1; /* VRAI : le ligand est polydente (réellement)   */
    unsigned aamine    : 1; /* VRAI : il s'agit d'un acide aminé -> protéine */
    unsigned nucleotide: 1; /* VRAI : il s'agit d'un ADN/ARN ou équivalent   */
  } drapeaux;

  /* Constitution du fragment : atomes le formant
                                liaisons _entre ces atomes_ */

  unsigned int nbr_atomes; /* Nombre d'atomes actuellement définis                  */
  unsigned int max_atomes; /* Nombre d'atomes au maximum dans ce fragment           */
  Atome **atomes;          /* Table des adresses des atomes constituant le fragment */

  unsigned int nbr_liaisons; /* Nombre de liaisons définies */
  Liaison *liaisons;         /* Liste simplement chaînée */

  Atome *absorbeur;        /* Atome absorbeur, si ce fragment en contient un */
  struct graphe *signal;   /* Signal EXAFS associé à ce fragment, si nécessaire
			      défini uniquement si contient un atome absorbeur ! */

  /* Définition du mouvement d'un fragment :
       - position initiale = celle dans x_g, y_g, z_g;
       - translation contient les déplacements du barycentre;
            tout est repéré en positions absolues du barycentre
       - rotation contient les rotations du fragment autour des 3 axes absolus (x, y, z)
            tout est repéré par rapport à la position précédente (rotations RELATIVES);
       - les deux graphes doivent être de type 3D et avoir le MÊME nombre de points
       - le numéro du point dans le graphe sert de référence temporelle; la conversion éventuelle
         en véritable unité temporelle est à la charge de l'utilisateur */
  struct graphe *translation;
  struct graphe *rotation;

  /* Informations internes */
  unsigned int numero; /* Pour n° aa ou nucléotide - essentiellement PDB */
  unsigned int code;   /* Pour un codage quelconque... Exemple :
			  - type de structure 3D pour un aa (hélice, feuillet,...) */
  int i, j, k;	       /* Pour les mailles */
} Fragment;

#define SANS_CODE 0 /* Cas général */

#define HELICE_ALPHA  10 /* 10 -> 19 : les types d'hélices / acides aminés */

#define FEUILLET_BETA 20 /* 20 -> 29 : les types de feuillets / acides aminés */

#define COUDE_BETA    30 /* 30 -> 39 : les types de coudes / acides aminés */


/* Définition d'une molécule : ensemble 1) de fragments
                                        2) de liaisons entre ces fragments
   plus quelques données globales
   Voir exemple plus bas pour la distinction fragment / atome

   La molécule est l'entité logique de base dans LASE : c'est l'équivalent du graphe pour la structure
   --> pour voir plusieurs molécules, il faudra prévoir des modules de superposition...

   Sauf précision contraire, les opérations s'appliquent à la molécule courante
*/

struct modele_ajuste; /* Dans modeles.h */
struct maille;        /* Dans cristallo.h */

typedef struct molecule
{
  struct molecule *suivante; /* Molécule suivante en mémoire */

  unsigned long nbr_atomes;
  Atome *premier;            /* Liste chaînée des atomes de la molécule */

  /* Quelques informations sur la molécule */
  char  *nom; /* Nom de la molécule [à libérer avec free()]*/
  float masse_molaire;

  /* Informations sur la cristallo */
  struct maille *maille;

  /* Voir pour coder la formule chimique, si nécessaire */
  Vecteur3D pos_barycentre;          /* Coordonnées _absolues_ du barycentre */

  /* Définition d'une molécule = assemblage d'une série de fragments.

     Exemples : 1) complexe de coordination
                      -> absorbeur = atome central
		      -> un fragment par ligand
		      -> une liaison par liaison fragment / ligand

		2) molécule simple, type solvant, inhibiteur ou autre
		      -> absorbeur s'il en existe un (me semble douteux)
		      -> un seul fragment qui définit toute la molécule, y compris l'absorbeur
		      -> aucune liaison nécessaire

		3) métalloprotéine, protéine, AD/RN,...
             	      -> absorbeur = atome absorbeur, métal
		      -> un fragment par acide aminé
		      -> les liaisons correspondent en fait aux liaisons peptidiques

		      NB : voir comment gérer les hétéro-atomes : un fragment spécial sans
		      liaisons ? ou une liste d'atomes ? Ou une liste de fragments ?
  */

  unsigned int nbr_fragments; /* Nombre de fragments actuellement définis */
  Fragment *fragments; /* Premier élément d'une liste simplement chaînée */

  /* Pour les liaisons entre fragments, afin de former la molécule */
  unsigned int nbr_liaisons; /* Nombre de liaisons définies */
  Liaison *liaisons;         /* Liste simplement chaînée */

  unsigned int nbr_absorbeurs;  /* Nombre d'absorbeurs dans la molécule */
  struct absorbeur *absorbeurs; /* Description de ces absorbeurs (tableau dynamique, car
				   le nombre d'absorbeur/molécule devrait rester limité...) */
  unsigned int absorbeur_courant; /* Absorbeur actuellement actif (modèles, ajustements,...) */

  struct
  {
    unsigned sauve  : 1; /* Elle a été sauvée après création             */
    unsigned marque : 1; /* Elle est marquée pour une raison quelconque  */
    unsigned visible: 1; /* On demande son affichage dans la fenêtre 3D  */
    unsigned protege: 1; /* Elle est protégée car en cours d'utilisation */
  } drapeaux;       /* Une série de drapeaux d'usage varié et encore non défini */

  double parametre; /* Un paramètre dans le cas d'une famille de molécules
		       Exemple : évolution temporelle -> ce paramètre est le temps */

  /* Définitions de l'angle de vue et tout et tout */
  Vecteur3D pos_oeil;   /* Position de l'oeil */
  Vecteur3D point_vise; /* Point visé */
  float x_h, y_h, z_h;          /* Vecteur définissant le haut */

  float angle1, angle2;         /* Pour la projection (utile ?) */

  /* Modèle global de la molécule */
  struct modele_ajuste *modele;
} Molecule;

#define MOL_EstSauvee(molecule_testee) ( (VRAI==molecule_testee->drapeaux.sauve)?(VRAI):(FAUX) )

/* ~~~~~~~~ Entités davantage spécifiques de la modélisation EXAFS ~~~~~~~~ */

/* Caractérisation d'un site d'absorption des photons.
   Contient -> l'atome qui absorbe effectivement
            -> le seuil qui se veut décrit par ce site
	    -> la polarisation du faisceau utilisée (<=> l'orientation de l'échantillon)
   Cette structure apparaît essentiellement dans une molécule.
   On peut avoir plusieurs absorbeurs par molécule, a priori.
   Remarque : les #define sont dans Include/theorie.h

   Si on demande une modélisation couplée par FEFF, les paramètres généraux sont utilisés,
   la table de potentiel aussi et les atomes des chemins sont utilisés pour créer le modèle.
*/

#define LG_MAX_NOM_ABSORBEUR 95

typedef struct absorbeur
{
  char nom[LG_MAX_NOM_ABSORBEUR + 1]; /* Le nom - donne un multiple de 8 */

  Atome *absorbeur; /* Quel atome joue le rôle d'absorbeur ? */

  /* ---- Données pour les modèles ---- */

  unsigned int nbr_modeles;           /* Nombre de modèles pour cet absorbeur (un par seuil étudié au moins) */
  struct modele_ajuste *modeles;      /* Liste simplement chaînée des modèles */
  struct modele_ajuste *modele_actif; /* Le modèle courant, pour ajustements semi-automatiques complexes */

  Variable contribution; /* En % du signal expérimental - pour le cas où une molécule contient
			    plusieurs atomes absorbeurs non-équivalents.
			    Sous forme de variable car peut faire l'objet d'un ajustement si l'on
			    travaille avec un mélange de composés dont on veut connaître la composition.
			 */

  /* Note : rien pour le lpm (en particulier Gamma) : on suppose implicitement que choisir Gamma comme
     paramètre ajustable n'a plus de sens maintenant que feff calcule bien les lpm théoriques
     Il semble que l'on ne gagne rien à jouer sur la variation du lpm avec k,
     donc autoriser un graphe lpm n'a pas trop de sens.
  */

  /* Le déphasage de l'atome central est a priori inclus dans le déphasage associé à chaque chemin */

  /* Voir comment décrire effectivement la polarisation, quand elle est utile */
} Absorbeur;

/* ~~~~~~~~ Fonctions de gestion des molécules ~~~~~~~~ */

Molecule *MOL_creer  (const char *nom);
int       MOL_effacer(Molecule *molecule, boolean sauver, boolean forcer);
boolean   MOL_tteffacer(void);

int MOL_SauverIsolee(const char *nom_fichier, Molecule *molecule);

int MOL_RenommerAbsorbeur(Absorbeur *absorbeur, const char *nouveau_nom);
int MOL_RenommerMolecule (Molecule *molecule, const char *nouveau_nom);

int       MOL_AssocierCristallo(Molecule *molecule, struct maille *maille);
int       MOL_EnleverCristallo (Molecule *molecule);
Fragment *MOL_AjouterMaille    (Molecule *molecule, int i, int j, int k);

int MOL_ajoute_fragment(Molecule *molecule, unsigned int nbr_fragments, Fragment **fragment);
int MOL_efface_fragment(Molecule *molecule, Fragment *fragment);

int MOL_ajoute_absorbeur(Molecule *molecule, Atome *absorbeur, struct graphe *signal);
int MOL_efface_absorbeur(Molecule *molecule, Atome *absorbeur);

int           MOL_creer_liaison (Molecule *molecule, unsigned int nbr_atomes, Atome **atomes,
				 Liaison modele);
int           MOL_retire_liaison(Molecule *molecule, Liaison *liaison, boolean forcer);
unsigned long MOL_chercher_liaisons(Molecule *molecule);

int MOL_cherche_barycentre(Molecule *molecule, boolean centrer);

Molecule *MOL_isoler(Molecule *molecule, Atome *atome);

Fragment *MOL_centre      (Molecule *molecule);
Fragment *MOL_creer_maille(Molecule *molecule, int i, int j, int k);

void MOL_tt_demarquer(Molecule *molecule);

#define AST_VUE_FEFF_ET_VISIBLE -1 /* Pas d'objet associé -> valeur pour être tranquille ! */

#define LIAISON_AUCUNE     0 /* On ne voit pas les liaisons */
#define LIAISON_SQUELETTE  1 /* Liaisons = bâtonnets homogènes */
#define LIAISON_SQUEL_TYPE 2 /* Liaisons = bâtonnets homogènes mais avec influence de la nature */
#define LIAISON_BATONS     3 /* Liaisons = bâtons colorés suivant les atomes mis en jeu */
#define LIAISON_BATON_TYPE 4 /* Liaisons = comme ci-dessus, mais type et multiplicité pris en compte */

int MOL_initialise_vue    (Molecule *molecule);

/* -------- Localisation des divers éléments dans les structures -------- */

Molecule *ABS_QuelleMolecule(Absorbeur *absorbeur);
Molecule *ATM_QuelleMolecule(Atome *atome);
Molecule *FRG_QuelleMolecule(Fragment *fragment);

Fragment *ATM_QuelsFragments(Atome *atome, Molecule *molecule, unsigned int *nbr_fragments, Fragment ***fragments);

Absorbeur *ATM_QuelAbsorbeur(Molecule *molecule, Atome *atome);

boolean   ABS_EstDansMolecule(Absorbeur *absorbeur, Molecule *molecule);
boolean   ATM_EstDansFragment(Atome *atome, Fragment *fragment);
boolean   ATM_EstDansLiaison (Atome *atome, Liaison *liaison);
boolean   ATM_EstDansMolecule(Atome *atome, Molecule *molecule);
boolean   FRG_EstDansMolecule(Fragment *fragment, Molecule *molecule);

/* -------- Aller à tel ou tel élément du modèle -------- */

void ABS_aller(Absorbeur *absorbeur, boolean redessin);
void ATM_aller(Atome *atome);
int  MOL_aller(Molecule *molecule);

void ABS_AllerModele   (struct modele_ajuste *modele, boolean redessin);
void MOL_AllerAbsorbeur(Molecule *molecule, unsigned int num_absorbeur);

/* -------- Sélecteurs des divers niveaux du modèle ----- */

void ABS_PrepareListe(void);
void ATM_PrepareListe(void);
void MOL_PrepareListe(void);

/* -------- Un sous-groupe : la gestion des liaisons ---------- */

void modifier_liaisons(Atome *atome, Molecule *molecule);

/* -------- Un sous groupe : la gestion des absorbeurs -------- */

#define LG_MAX_TXT_ABSORBEUR 32 /* Nom de l'absorbeur dans la fenêtre d'affichage 3d */

Absorbeur *creer_absorbeur   (Atome *atome, const char *nom, double contribution);
boolean    modifier_absorbeur(Absorbeur *absorbeur);
Absorbeur *nouvel_absorbeur  (Molecule *molecule);

Atome     *ABS_ProposerAbsorbeur(Molecule *molecule, boolean strict);
boolean    ABS_oter(Molecule *molecule, unsigned int i, boolean effacer_atome);

struct modele_ajuste *ABS_CreerModele   (Absorbeur *absorbeur, int type, struct graphe *experience);
void                  ABS_AjouterModele (Absorbeur *absorbeur, struct modele_ajuste *modele);
int                   ABS_DetruireModele(Absorbeur *absorbeur, struct modele_ajuste *modele);

/* ~~~~~~~~ Fonctions de gestion des fragments ~~~~~~~~ */

Fragment *FRG_creer (Molecule *molecule, Fragment *premier, char *nom, unsigned int nbr_atomes, Atome *absorbeur, struct graphe *signal);
int FRG_effacer     (Molecule *molecule, Fragment *fragment, boolean forcer, boolean effacer_atomes);

int FRG_ajoute_atome(Fragment *fragment, unsigned int nbr_atomes, Atome **atome);
int FRG_efface_atome(Fragment *fragment, Atome *atome, boolean forcer);

int FRG_creer_liaison (Fragment *fragment, unsigned int nbr_atomes, Atome **atomes,
		       Liaison modele);
int FRG_retire_liaison(Fragment *fragment, Liaison *liaison, boolean forcer);

int FRG_chercher_ligands(Molecule *molecule);

/* ~~~~~~~~ Fonctions de gestion des atomes ~~~~~~~~~~~ */

void   ATM_init(void); /* Initialisation du module */

Atome *ATM_chercher(Vecteur3D *position, int Z, Molecule *molecule);
unsigned int *ATM_choisir (Atome **atomes, unsigned int nbr_atomes, boolean multiple, int *choisi);
Atome *ATM_copier  (Atome *original, Molecule *cible);
Atome *ATM_creer   (Molecule *molecule, Atome *avant, int numero_atomique, Vecteur3D *position, double charge, double rayon);
int    ATM_efface  (Molecule *molecule, Atome *atome, boolean forcer);
Atome *ATM_index   (Molecule *molecule, unsigned long i);
Atome *ATM_nouveau (Molecule *molecule, boolean complet);
int    ATM_ttefface(int numero_atomique); /* Efface tous les atomes du même type */
void   ATM_zero    (Molecule *molecule);

int    ATM_CreerErreurs(Atome *atome);
#define ATM_EffacerErreurs( atome ) { free(atome->err_position); atome->err_position = NULL; }

Molecule *quelle_molecule(Atome *atome, Absorbeur **absorbeur);

/* Diverses fonctions structurales */

#define ATM_COORD_IGNORER 0
#define ATM_COORD_NORMALE 1
#define ATM_COORD_AJUSTE  2 /* Coordonnées après reconstruction du modèle */

boolean ATM_AtomeConvient(Atome *atome, double x, double y, double z, int Z, int coordonnees);

double ATM_distance(Atome *A, Atome *B);
double ATM_angle3  (Atome *A, Atome *B, Atome *C);

boolean eloigne_absorbeur(Atome *absorbeur, Atome *atome);

/* Remplace un ensemble d'atomes par un « atome moyen » */

Atome *ATM_moyenne(int nbr_atomes, Atome **atomes, int effacer);

void   ATM_termine(void); /* On quitte le module */

/* ~~~~~~~~ Fonctions de dessin ~~~~~~~~ */

void I_MOL_Visualisation3D(void); /* Appel de la fenêtre de vue en 3d */

void VUE3D_ChoisirChemin(void);

/* ~~~~~~~~ Fonctions d'ajustement structural ~~~~~~~~ */

int ABS_Ajuster      (Absorbeur *absorbeur);
int MOL_ajuste       (Molecule *molecule);

#ifndef VARIABLES_ATOMES

extern Atome *ATM_courant;

extern Fragment *premier_fragment;     /* Uniquement pour les sélections extra-moléculaires */

extern Molecule *premiere_molecule, *mol_courante;
extern Fenetre *fen_vue_3d;            /* la fenêtre qui contient la vue 3d */
extern unsigned long nombre_molecules; /* Le nombre de molécules en mémoire */
extern int mode_vue3d; /* Comment tracer la molécule ? */

# ifdef AVEC_OPEN_GL_
extern GLuint liste_molecule; /* liste d'affichage des atomes de la molécule */
extern GLuint liste_chemin;   /* liste d'affichage du chemin actuel */
extern GLuint liste_liaisons; /* liste d'affichage du schéma de liaisons */
# endif /* AVEC_OPEN_GL_ */

#endif /* VARIABLES_ATOMES */

/* ************ Fonctions de chargement des atomes ****************************** */

void COR_charge(void); /* Lecture d'un fichier *.COR de la CCDC */

#if defined(_TURBO_C_) || defined(_PURE_C_)
# include "graphes.h" /* Obligatoire car Pure C n'aime pas les typedef struct... */
# include "couches.h" /*  sans vraie définition */
#endif

#ifdef _DEVELOPPE_
void MOL_Controle_Affiche_Structure(Molecule *molecule, int niveau);
#endif

#endif /* ATOMES_H_ */
