/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1997 - 2000, Emmanuel CURIS
   Cration : le 12 avril 2000

   Infos et fonctions ncessaires  la description d'une structure cristallographique
   Attention, atomes.h doit tre inclus avant
*/

#ifndef CRISTALLO_H_

#define CRISTALLO_H_ /* Protection contre l'inclusion multiple */

#include "utils_3d.h" /* Dfinition d'un point dans l'espace ! */

struct fragment;
struct molecule; /* Dans atomes.h */

/* -------- Dfinition des 14 rseaux de Bravais -------- */

struct groupe_espace {
  int numero; /* Selon l'IUCR */
  char *symbole; /* Idem + synonymes ventuels */
  char *nom_fichier; /* Le nom du fichier qui dcrit ce groupe */
  unsigned int nbr_operations; /* De symtrie dans le groupe, en plus de l'identit */
  unsigned int nbr_translations;
  float *operations; /* Les matrices dcrivant les oprations [44, en ligne] */
  float *translations; /* Les vecteurs de translations [x, y, z]*/
};

struct systeme_cristallin {
  char *nom; /* Nom du systme cristallin */
  short int type;  /* C = cubique, H = hexagonal, R = rhombodrique, Q = quadratique, O = orthorhombique, M = monoclinique, T = triclinique, 0 = autre, inconnu */
  short int simple;/* 0, P ou S = simple (primitif), I = centr, C = bases centres, F = faces centres */
  char *dossier;   /* Le nom du dossier qui contient tous les groupes pour ce systme */
  unsigned int nbr_groupes; /* Le nombre de groupes dans ce systme (= de fichiers dans le dossier !) */
  struct groupe_espace *groupes; /* Les diffrents groupes de ce systme */
};

/* ------ Dfinition d'une maille cristalline -------- */

typedef struct maille {
  double a, b, c;            /* Cts de la maille */
  double alpha, beta, gamma; /* Angles de la maille (en _radians_, pour acclrer les calculs) */

  double s_a, s_b, s_c;            /* Incertitude sur les cts */
  double s_alpha, s_beta, s_gamma; /* Incertitude sur les angles */

  int type; /* Comme dans le systme cristallin */
  unsigned int Z; /* Nombre de motifs par maille */

  unsigned int nbr_operations; /* De symtrie dans le groupe, en plus de l'identit */
  unsigned int nbr_translations;
  float *operations; /* Les matrices dcrivant les oprations [44, en ligne] */
  float *translations; /* Les vecteurs de translations [x, y, z]*/
} MailleCristalline;

/* ------ Lecture des tables de paramtres cristallins -------- */

struct systeme_cristallin *CRIS_CreerListeSystemes  (const char *dossier, unsigned long *nbr_systemes);
int                        CRIS_EffacerListeSystemes(struct systeme_cristallin *systemes, unsigned long nbr_systemes);

int CRIS_CreerListeGroupes    (struct systeme_cristallin *systeme);
int CRIS_EffacerListeGroupes  (struct systeme_cristallin *systeme);
int CRIS_LireInfosGroupes     (struct groupe_espace *groupe, const char *racine, const char *dossier);
int CRIS_LireOperationsGroupes(struct groupe_espace *groupe, const char *racine, const char *dossier);

int CRIS_AdapteParametres(struct systeme_cristallin *systeme, double *a, double *b, double *c, double *alpha, double *beta, double *gamma);

extern const char *titre_choix_reseau;
 
/* -------- Les fonctions permettant de grer tout a -------- */

/* Cration d'une dfinition de maille */
struct maille *CRIS_CreerMaille  (void);
struct maille *CRIS_CopierMaille (struct maille *maille);
int            CRIS_EffacerMaille(struct maille *maille);

/* Demander les trois indices dfinissant la maille */

int CRIS_DemanderIndicesMaille(BI_Boite_Demande *boite, int *i, int *j, int *k);

/* Conversion (i, j, k) -> numro de la maille dans le tableau mailles */
#define NUM_MAILLE(i, j, k) ( ( (i) * (j_max-j_min+1) + (j)) * (k_max-k_min+1) + (k) )

/* ------ Gnration des positions symtriques ------ */

typedef Vecteur3D Point_3D;

void CRIS_GenereSymetrique(Vecteur3D *original, float *matrice, Vecteur3D *image);
int  CRIS_GenereCristal   (struct molecule *molecule, Vecteur3D *position, int Z, double charge, double rayon,
			   unsigned long nombre_mailles, struct fragment **mailles);

int CRIS_NormaliserCoordonneesFractionnaires(Vecteur3D *position);

int CRIS_AbsoluVersFrac(struct maille *maille, Vecteur3D *absolu, Vecteur3D *fractionnaire);
int CRIS_FracVersAbsolu(struct maille *maille, Vecteur3D *fractionnaire, Vecteur3D *absolu);
int CRIS_VecteursMaille(struct maille *maille, Vecteur3D *a, Vecteur3D *b, Vecteur3D *c);

/* Les positions particulires sur la maille (faces). Fonctionne comme
   un masque de bits : le bit est mis si l'atome appartiet  la face
   considre. Donc un atome sur une arte aura deux bits mis, un atome  un coin aura trois bits mis.
*/

#define DANS_LA_MAILLE    0
#define SUR_FACE_AB       1 /* Fond du paralllpipde      : plan (x, y, 0) */
#define SUR_FACE_AC       2 /* Ct vertical face  nous    : plan (x, 0, z) */
#define SUR_FACE_BC       4 /* Ct vertical de ct        : plan (0, y, z) */
#define SUR_FACE_AB_FOND  8 /* Couvercle du paralllpipde : plan (x, y, 1) */
#define SUR_FACE_AC_FOND 16 /* Ct vertical face  nous    : plan (x, 1, z) */
#define SUR_FACE_BC_FOND 32 /* Ct vertical de ct        : plan (1, y, z) */

boolean CRIS_ControlerCreation(Vecteur3D *position, unsigned int nbr_crees, Point_3D *existants);
boolean CRIS_EstSurUneFace    (Vecteur3D *position, unsigned int nbr_crees, Point_3D *existants, int *face);

struct fragment **FRG_mailles(struct molecule *molecule,
			      int i_min, int i_max, int j_min, int j_max, int k_min, int k_max);

/* -------- Import / Export ASCII -------- */

#define TROISD_LIRE_ERREURS  1 /* Bit mis : on lit aussi les erreurs */
#define TROISD_COORD_FRAC    2 /* Bit mis : les coordonnes sont fractionnaires */
#define TROISD_SYMETRIES     4 /* Bit mis : on gnre les symtries   */
#define TROISD_MULTIMAILLES  8 /* Bit mis : on veut plusieurs mailles */
#define TROISD_FRG_MAILLE   16 /* Bit mis : un fragment = une maille  */

struct infos_lecture {
  unsigned int drapeaux;            /* Masque de bits : ce qu'il faut faire */

  unsigned long lg_en_tete; /* Nombre de lignes  ignorer au dbut [hors commentaires] */
  unsigned int nbr_colonnes; /* Pour les lignes "utiles"; 0 = automatique */
  int commentaire; /* Caractre introduisant les lignes de commentaire */
  int separateur;  /* Sparateur de colonnes */

  unsigned int col_x, col_y, col_z; /* Colonnes des coordonnes */
  unsigned int col_Z;     /* Colonne des symboles atomiques */
  unsigned int col_txt;   /* Colonne des tiquettes */
  unsigned int col_rayon; /* Colonne des rayons */
  unsigned int col_charge;/* Colonne des charges */

  unsigned int col_a, col_b, col_c, col_alpha, col_beta, col_gamma; /* Colonnes pour les paramtres de maille */
  unsigned long ligne_maille; /* Numro de la ligne contenant les paramtres de maille */
  unsigned long ligne_groupe; /* Numro de la ligne contenant le groupe d'espace */

  /*struct groupe_espace      *groupe;*/ /* Le groupe d'espace  utiliser, si ncessaire */
  struct systeme_cristallin systeme;/* Le systme cristallin  utiliser, si ncessaire [partiel !] */
  unsigned int groupe_choisi; /* Index du groupe dans le systme choisi */

  struct maille *maille;
  struct fragment **table_mailles;
  int i_min, i_max;
  int j_min, j_max;
  int k_min, k_max;

  const char *nom; /* Indiqu par l'utilisateur */
  struct molecule *molecule; /* La molcule qui contient le rsultat ! */
  struct fragment **mailles; /* La table des fragments associs  chaque maille */
};

/* Lit une ligne, renvoie le premier caractre de la ligne suivante (EOF si problme) non-lu */

int CRIS_LireLigneAscii(FILE *fichier, struct infos_lecture *infos);

#endif /* CRISTALLO_H_ */
