/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modles structurels d'chantillons
   Fichiers de dclarations
*/

#ifndef ECHANTILLON_H_
# define ECHANTILLON_H_ /* Protection contre l'inclusion multiple */

#include "chimie_utils.h" /* Pour la composition chimique lmentaire */
#include "utils_3d.h"

struct faisceau;
struct graphe; /* Dans graphes.h */
struct image;  /* Dans images.h */

/* -------- Briques lmentaires ----------- */

#ifdef ancien
struct spectre_local {
  unsigned long nbr_energies; /* Nombre de points */
  double e_max, delta_e;      /* On part de Emax et on descend par pas de DE */
  double *valeur;             /* -> Nombre de photons [intensit vraie] pour chacune de ces nergies
				 -> Valeur du coefficient d'absorption (total, diffusion,...)
			      */ 
};
#else
# define spectre_local graphe
#endif

#define ZMAX_TABLE_CONSTITUANTS PLUTONIUM /* Peu probable que l'on utilise plus lourd, mais... */

struct composant {
  struct composant *suivant; /* Liste simplement chaine */
  char *nom;                 /* Nom du composant [allou dynamiquement, longueur libre] */
  float atomes[ ZMAX_TABLE_CONSTITUANTS ]; /* A priori, inutile d'aller au-del, mais bon...
				Contient les fractions massiques de chaque lment*/
  float densite;             /* La densit du composant tudi, pour la conversion mu/rho ==> mu linique */
  float masse_molaire;       /* La masse molaire du constituant, pour la formule indique */
  struct graphe *mu[ ZMAX_TABLE_CONSTITUANTS ]; /* Si non-NULL : contient le seuil XANES / EXAFS
						   sinon on utilise un modle classique */
  struct graphe *mu_total;        /* Si non-NULL, contient un spectre propre */
};

struct composant *CreerComposant  (struct composant *avant);
struct composant *ChoisirComposant(Fenetre *fenetre_mere);

int EffacerComposant( struct composant *, boolean forcer );

int Composant_AnalyserFormule(struct composant *composant, const char *formule);
int Composant_CalculerFluo   (struct composant *composant, struct graphe *graphe, double f_massique, double energie, double mu_total);
int Composant_CalculerMu     (struct composant *composant, struct graphe *graphe, double f_massique);
int Composant_CalculerTau    (struct composant *conposant, struct graphe *graphe, double f_massique, boolean elastique, int methode );
int Composant_ChangerNom     (struct composant *composant, char *nom);
int Composant_NombreElements (struct composant *composant);

int Composant_EstimerDensite(struct composant *composant);

double Composant_CalculerMu_1p(struct composant *composant, double f_massique, double energie);

unsigned long     Composant_Compter (void);
unsigned long     Composant_C_Vers_N(struct composant *premier, struct composant *composant);
struct composant *Composant_N_Vers_C(struct composant *premier, unsigned long n);

struct composant *Composant_Charger(FILE *fichier, int version, const char *nom_fichier);
int Composant_Sauver(FILE *fichier, struct composant *composant);

struct matiere {
  struct matiere *suivante; /* Liste simplement chane */
  char *nom;

  unsigned int nbr_composants;
  double densite;
  float *fraction_massique;
  struct composant **composants;
  struct graphe *mu_total; /* Spectre d'absorption total [idal] pour ce composant */
  struct graphe *te_total; /* Spectre de diffusion lastique total [idal] pour ce composant */
  struct graphe *ti_total; /* Spectre de diffusion inlastique total [idal] pour ce composant */
  struct graphe *mu;       /* Spectre d'absorption d _au seul seuil tudi_ (pour la fluo...) */
  struct graphe *fluo;     /* Spectre d'mission idal du  ce composant */

  unsigned int couleur;	/*  utiliser pour les cartes de composition */
};

#ifndef VARIABLES_GENERALES_
extern struct composant *premier_composant;
extern struct matiere   *premiere_composition;
#endif

struct matiere *Matiere_CreerComposition  (struct matiere *avant, unsigned int nbr_composants);
int             Matiere_EffacerComposition(struct matiere *composition, boolean forcer);

struct composant *Matiere_AjouterComposant( struct matiere *composition, struct composant *composant );
int               Matiere_EffacerComposant( struct matiere *composition, unsigned int numero_composant );

struct graphe    *Matiere_CalculerMu      ( struct matiere *composition, double *table, unsigned long N, double E_min, double E_max, double pas );
int               Matiere_CalculerFluo    ( struct matiere *composition, double energie, boolean zero );
int               Matiere_CalculerTau    ( struct matiere *composition, struct graphe *graphe, boolean elastique, int methode, boolean ajouter );

struct matiere   *Matiere_ChoisirComposition ( unsigned int arbre, unsigned int objet, unsigned long lg_max, struct matiere *premiere );
int               Matiere_ModifierComposition( struct matiere *composition );

int Matiere_Nommer(struct matiere *composition, const char *nom);
int Matiere_ControlerComposition(struct matiere *composition);
int Matiere_EstimerDensite(struct matiere *composition);

inline unsigned long   COMPOSITION_Compter(void);
inline unsigned long   COMPOSITION_C_Vers_N(struct matiere *premiere, struct matiere *composition);
inline struct matiere *COMPOSITION_N_Vers_C(struct matiere *premiere, unsigned long n);

struct matiere *Composition_Charger(FILE *fichier, int version, const char *nom_fichier, struct composant **table_composants);
int Composition_Sauver(FILE *fichier, struct matiere *composition);

/* -------- Qu'est-ce qu'un chantillon ? --------

  Pour les besoins informatiques, c'est un pav rectangulaire, compatible avec le repre.
  Par convention, le repre est centr sur la face d'entre du faisceau,
  avec des axes compatibles avec les bords de l'chantillon.

  Dans le cas idal d'un faisceau arrivant orthogonalement  une face, on a les conventions :
  Largeur   = axe des X
  Hauteur   = axe des Z
  paisseur = axe des Y = direction de propagation du faisceau

  Trois cas pour le codage des donnes :

- chantillon homogne (si proprietes & ECHANTILLON_HOMOGENE)
  Alors valeurs->homogene.composition contient la composition & la densit
        valeurs->homogene.concentration est NULL
        valeurs->homogene.composition->mu_total contient le coefficient d'absorption LINIQUE

- chantillon de composition fixe mais de concentration variable (si type == ECHANTILLON_GRADIENT)
  Alors valeurs->homogene.composition contient la composition
        valeurs->homogene.concentration contient la correction  la densit en chaque point [1 : densit de la composition ; facteur correctif multiplicatif]
        valeurs->homogene.composition->mu_total contient le coefficient d'absorption MASSIQUE (mu/rho)

- chantillon de composition variable (si type == ECHANTILLON_MATERIEL)
  Alors valeurs->inhomogenes contient la composition & la densit en chaque point
        valeurs->inhomogenes->mu_total contient le coefficient d'absorption LINIQUE en ce point

  Les valeurs sont connues uniquement pour les noeuds de la grille;
  on suppose que chaque case est homogne par rapport  la valeur connue
  => pavage 3D (comme sur l'image affiche)
*/

#define ECHANTILLON_MATERIEL  0 /* Contient la composition, en chaque point */
#define ECHANTILLON_GRADIENT  1 /* Idem, mais composition unique et juste un nombre = pondration */
#define ECHANTILLON_SPECTRE  10 /* Contient le nombre de photons en chaque point */

#define ECHANTILLON_HOMOGENE  ( 1U << 0 ) /* Bit mis : l'chantillon est homogne */
#define ECHANTILLON_DIFFUSE   ( 1U << 1 ) /* Bit mis : on tient compte de la diffusion lastique */
#define ECHANTILLON_DIFF_INEL ( 1U << 2 ) /* Bit mis : on tient compte de la diffusion inlastique */
#define ECHANTILLON_AUGER     ( 1U << 3 ) /* Bit mis : on tient compte des lectrons Auger */

#define IndexTableEchantillon(x, y, z, nbr_x, nbr_y) ( ( ( ( z ) * ( nbr_y ) ) + ( y ) * ( nbr_x ) ) + ( x ) )

typedef struct echantillon {
  struct echantillon *suivant;
  char *nom;

  int type; /* Que contient comme info chaque case de l'chantillon ? */

  double larg, haut, epaisseur; /* Dimensions, en microns [larg : X, haut : Z, epaisseur : Y] */
  double pas_x, pas_y, pas_z;   /* Pas de la grille 3D utilise, en microns aussi */
  unsigned int nbr_x, nbr_y, nbr_z; /* Nombre de cases dans chaque direction. Redondant avec ce qui prcde,
				       juste pour faciliter les calculs (donc acclrer lgrement) */
  unsigned int proprietes;

  /* Codage de l'chantillon : la grille 3D dfinit des n~uds en (i, j, k) : coordonnes locales entires
       les valeurs sont dfinies sur cette grille, pour 0 <= i < nbr_x, 0 <= j < nbr_y, 0 <= k < nbr_z
       pour les autres valeurs : on se ramne au (i,j,k) _le plus proche_
       ==> les n~uds sont au centre des zones de mme valeur
       ==> revient  dire que la valeur connue est au centre de la case

     Donc : (i, j, k) = repre rel   => (x_min + i * pas_x, y_min + j * pas_y, z_min + k * pas_z )
            (x, y, z) = repre grille => ((x - x_min)/pas_x, (y - y_min)/pas_y, (z - z_min)/pas_z )
	          valeurs non-entires ; pour savoir dans quelle case _lire_ : partie entire
   */

  /* Stockage des donnes dcrivant l'chantillon */
  union {
    struct {
      struct matiere *composition;
      double *concentration;
    } homogene;
    struct matiere **inhomogene;
    struct spectre_local **spectre;
  } valeurs;

  /* Ncessaires pour le modle */
  /* Nombre de photons absorbs en chaque point --- pour le premier tour du modle !
     Correspond  l'absorption pour arriver du point prcdent au point courant.
     Ainsi, pour une coupe 2D :
     ----------
     |  |   |  |
   ^-|--+---*--|
   h |  |   |  | * : absorption aprs traverse de h
   v_|--+---*--|
     |  |   |  |  : absorption aprs traverse de h/2
     ----------
 */
  double *nombre; 

  struct spectre_local **avant;   /* Nombre de photons de chaque nergie en chaque point, au coup prcdent */
  struct spectre_local **photons; /* Nombre de photons de chaque nergie en chaque point, actuel */
  struct spectre_local **apres;   /* Nombre de photons de chaque nergie en chaque point, au coup suivant */
  /* On doit mmoriser 3 coups, car il faut pouvoir retrouver les directions de propagation des photons
     pour la diffusion inlastique ! */

  struct spectre_local **electrons; /* lectrons Auger gnrs en chaque point et leur nergie */
} Echantillon;

Echantillon *ECH_CreerEchantillon (void); /* Interface, en fait... */

Echantillon *ECH_CreerEchantillonP(double larg, double haut, double epaisseur, double pas_l, double pas_h, double pas_e,
				   int type, unsigned int proprietes);
Echantillon *ECH_CreerEchantillonN(double larg, double haut, double epaisseur,
				   unsigned int n_l, unsigned int n_h, unsigned int n_e,
				   int type, unsigned int proprietes);

int ECH_ChangerGrille (Echantillon *echantillon, double nv_largeur, double nv_hauteur, double nv_epaisseur,
		       double nvpas_l, double nvpas_h, double nvpas_e);
int ECH_Redimensionner(Echantillon *echantillon, double nv_largeur, double nv_hauteur, double nv_epaisseur);

int ECH_ChainerEchantillon  (Echantillon *avant, Echantillon *echantillon);
int ECH_DechainerEchantillon(Echantillon *echantillon);

int ECH_EffacerEchantillon  (Echantillon *echantillon, boolean forcer);

int ECHANTILLON_Nommer    (Echantillon *echantillon, const char *nom);
int ECHANTILLON_Parametrer(Echantillon *echantillon);

/* Version pour le slecteur universel */
int ECHANTILLON_Charger_I (unsigned long n_echantillon);
int ECHANTILLON_Copier_I  (unsigned long n_original);
int ECHANTILLON_Effacer_I (unsigned long n_original);
int ECHANTILLON_Modifier_I(unsigned long n_original);
int ECHANTILLON_Nouveau_I (unsigned long n_avant);
int ECHANTILLON_Sauver_I  (unsigned long n_echantillon);

unsigned long   ECHANTILLON_Compter(void);
unsigned long   ECHANTILLON_E_Vers_N(Echantillon *premier, Echantillon *echantillon);
Echantillon    *ECHANTILLON_N_Vers_E(Echantillon *premier, unsigned long n);
const char     *ECHANTILLON_TrouverNom(unsigned int arbre, unsigned long objet);

Echantillon *ECHANTILLON_Charger(FILE *fichier, int version, const char *nom_fichier, struct matiere **table_compositions);
int          ECHANTILLON_Sauver (FILE *fichier, Echantillon *echantillon);

/* Sous-menu pour changer la composition */
int ECH_SousMenuComposition(Echantillon *echantillon, struct image *coupe, unsigned int n_tranche);

int    ECH_ControlerOptions(Echantillon *echantillon);

inline unsigned long ECH_IJK_v_Case(Echantillon *echantillon, unsigned int i, unsigned int j, unsigned int k);
inline unsigned long ECH_XYZ_v_Case(Echantillon *echantillon, double x, double y, double z);
inline int           ECH_Case_v_IJK(Echantillon *echantillon, unsigned long n_case, unsigned int *i, unsigned int *j, unsigned int *k);
inline int           ECH_Case_v_XYZ(Echantillon *echantillon, unsigned long n_case, double *x, double *y, double *z);

inline int ECH_IJK_v_XYZ(Echantillon *echantillon, unsigned int i, unsigned int j, unsigned int k, double *x, double *y, double *z);
inline int ECH_XYZ_v_IJK(Echantillon *echantillon, double x, double y, double z, unsigned int *i, unsigned int *j, unsigned int *k);

double ECH_LireDensite(Echantillon *echantillon, Vecteur3D *point);

double ECH_CalculerMuTotal      (Echantillon *echantillon, Vecteur3D *point, double energie, double *mu_photoelectrique);
double ECH_TrouverCaseSuivante  (Echantillon *echantillon, Vecteur3D *direction, Vecteur3D *p_entree,
				 unsigned int *i_x, unsigned int *i_y, unsigned int *i_z, boolean *sortie);
int    ECH_TrouverSortieFaisceau(Echantillon *echantillon, Vecteur3D *point_entree, Vecteur3D *direction, Vecteur3D *point_sortie, boolean tester_face_entree);

struct image *ECH_TrancheVersImage_X(Echantillon *echantillon, unsigned int tranche_x, struct image *image);
struct image *ECH_TrancheVersImage_Y(Echantillon *echantillon, unsigned int tranche_y, struct image *image);
struct image *ECH_TrancheVersImage_Z(Echantillon *echantillon, unsigned int tranche_z, struct image *image);

int ECH_ImageVersTranche_X(Echantillon *echantillon, unsigned int tranche_x, struct image *image);
int ECH_ImageVersTranche_Y(Echantillon *echantillon, unsigned int tranche_y, struct image *image);
int ECH_ImageVersTranche_Z(Echantillon *echantillon, unsigned int tranche_z, struct image *image);

int FLUO_InitialiserEchantillon(Echantillon *echantillon, struct faisceau *faisceau, double energie);
double FLUO_PhotonsAbsorbes(Echantillon *echantillon, Vecteur3D *point_entree, Vecteur3D *point_sortie, double distance, double mu, double mu_total,
				   double energie, double *attenuation);

#endif /* ECHANTILLON_H_ */
