/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1996 - 2002
   Création : 5 décembre 1997

   Tout ce qui concerne le paramétrage de FEFF (shell en particulier)
   Fichier d'en-tête : toutes les déclarations nécessaires
*/

#ifndef INC_FEFF_

#define INC_FEFF_ /* Protection contre l'inclusion multiple */

#include "atomes.h"   /* Nécessaire dans les déclarations  */
#include "c_formel.h" /* id. pour Variable, dans le modèle */
#include "utils_3d.h" /* Pour les vecteurs */

/* ------------ Toutes les infos nécessaires à FEFF ---------------------------- */

typedef struct atome_feff {
  struct atome_feff *suivant; /* NULL : on est sur le dernier                                 */
  struct atome *atome;  /* L'atome utilisé : x, y, z, charge, symbole, nom                           */
  int potentiel; /* Le potentiel FEFF associé à cet atome (0 = absorbeur, 1...8 = réflecteur) */
} FEFF_atome;

typedef struct ss_feff
{
  struct ss_feff *suivante;
  unsigned int index; /* Couche que l'on considère         */
  int potentiel;      /* Potentiel des atomes de la couche */
  int degenerescence; /* Dégénérescence de la couche       */
  float distance;     /* Distance à l'atome central        */
} FEFF_SS; /* Commandes de réflexion simple */

typedef struct overlap_feff
{
  struct overlap_feff *suivante;
  int potentiel;           /* Potentiel de l'atome avec lequel on recouvre */
  unsigned int nbr_atomes; /* Nombre d'atomes de ce genre                  */
  float distance;          /* Distance de ces atomes                       */
} FEFF_OVERLAP;

/* -------- Versions de FEFF supportées -------- */

#define FEFF_VERSION_300  0	/* FEFF 3.00 */
#define FEFF_VERSION_400  1	/* FEFF 4.00 */
#define FEFF_VERSION_500  2	/* FEFF 4.00 */
#define FEFF_VERSION_600  3	/* FEFF 6.00 */
#define FEFF_VERSION_6L   4	/* FEFF 6 « lite » */
#define FEFF_VERSION_700  5	/* FEFF 7.00 */
#define FEFF_VERSION_702X 6	/* FEFF 7.02x [extension XANES] */
#define FEFF_VERSION_800  7	/* FEFF 8.00 */
#define FEFF_VERSION_810  8	/* FEFF 8.10 */
#define FEFF_VERSION_820  9  	/* FEFF 8.20 */
#define FEFF_VERSION_840 10	/* FEFF 8.40 */

#define FEFF_NOMBRE_VERSIONS 11
#define VERSION_FEFF_PAR_DEFAUT FEFF_VERSION_820 /* Version installée sur ma machine... */

/* ------------ Commandes possibles dans FEFF.INP ----------------------------- */

#define ABSENT -1 /* Pour les commandes qui doivent ne pas être présentes */
#define NON_LANCE -1 /* Module de FEFF non lancé au démarrage */

/* Les modules FEFF utilisables */
#define POTPH  0		
#define PATHS  1
#define GENFMT 2
#define FF2CHI 3
#define XSPH   4		/*  FEFF 8.2 : module XSPh */
#define FMS    5		/*  FEFF 8.2 : module FMS  */

typedef struct
{
  unsigned int nbr_atomes;  /* On limite à 65536 atomes... devrait suffire */
  FEFF_atome  *prem_atome;  /* Liste des atomes des autres couches         */

  short int CONTROL[ 6 ]; /* NON_LANCE : non lancé, sinon lancé avec sortie indiquée */

  struct {
    float temperature; /* Température          */
    float t_debye;     /* Température de Debye */
  } DEBYE;             /* Nulles -> option non utilisée */

  float SIG2; /* Facteur de Debye-Waller */

  char *titre[10]; /* Les lignes de titre */

  struct {
    float decalage;
    float elargissement;
  } CORRECTIONS;
  
  struct {
    float critcw, critpw;
  } CRITERIA;

  struct {
    int circulaire; /* FAUX : linéaire, VRAI : circulaire, ABSENT : on s'en fiche */
    struct vect_3d vecteur; /* Vecteur d'ellipticité */
  } ELLIPTICITY;

  int NEMAX; /* Intervalle de calcul */

  int NLEG; /* Nombre maximal de réflexions */

  int XANES;  /* Commande XANES présente, si oui valeur ? */

  struct {
    float conservation, tas;
  } PCRITERIA;

  struct {
    int presente;   /* FAUX : linéaire, VRAI : circulaire, ABSENT : on s'en fiche */
    struct vect_3d vecteur; /* Vecteur de polarisation */
  } POLARIZATION;

  float RMAX; /* Longueur maximale des chemins (0 : absent...) */

  float RMULTIPLIER; /* Facteur multiplicatif des distances */

  unsigned int nbr_ss; /* Nombre de commandes SS */
  FEFF_SS *prem_ss;    /* Paramètres de la première commande SS */

  float FOLP[9];     /* Recouvrements pour chacun des potentiels     */
  float ION[9];      /* Charge portée par les atomes au potentiel... */

  struct {
    int type; /* 0 : H-L, 1 : D-H, 2 : fondamental, 3 : D-H/H-L */
    float decalage;
    float p_imag;
  } EXCHANGE;

  FEFF_OVERLAP *prem_couche[8]; /* Pour instruction OVERLAP */

  /* -------- Options ajoutées par Bruce Raven pour xanes ---------- */

  boolean Utiliser_Xanes; /* VRAI si on veut utiliser les options de XANES */

  struct {
    int nombre;
    float position[5]; /* Définition des intervalles pour la grille */
    float delta_x[5];  /* Ceci ira en fait dans EMESH */
  } EGRID;
  float VINTFIX; /* Démarrage du calcul, relatif au 0 des potentiel « muffin-tin » */

} FEFF_INP; /* Tout ce qu'il faut pour un fichier FEFF.INP */

/* ---- Pour le chargement des résultats FEFF [paths.dat] --- */

#define PATHSDAT_FORCER_ABSORBEUR_INDIQUE 1 /* Mis : on force à l'absorbeur indiqué ; si incompatible : échec */
#define PATHSDAT_LIMITER_LONGUEUR         2 /* Mis : on se limite à une certaine longueur de chemins */
#define PATHSDAT_LIMITER_NOMBRE_ATOMES    4 /* Mis : on se limite aux chemins passant par moins de X atomes  */
#define PATHSDAT_LIMITER_NOMBRE_CHEMINS   8 /* Mis : on ne charge que les X premiers chemins */
#define PATHSDAT_N_CONSTANT              16 /* Mis : on génère des couches avec N fixe */
#define PATHSDAT_R_CONSTANT              32 /* Mis : on génère des couches avec R fixe */
#define PATHSDAT_INCLURE_LPM             64 /* Mis : on inclue le libre parcours moyen à l'amplitude */

#define PATHSDAT_ERREUR_FICHIER_INACCESSIBLE -1
#define PATHSDAT_ERREUR_FICHIER_ILLISIBLE    -2
#define PATHSDAT_ERREUR_Z_INCONNU            -3
#define PATHSDAT_ERREUR_Z_INCOMPATIBLE       -4
#define PATHSDAT_ERREUR_PAS_DE_MOLECULE      -5
#define PATHSDAT_ERREUR_PAS_DE_MODELE        -6

typedef struct infos_pathsdat {
  unsigned int options;		/* Options de chargement : cf #define ci-dessus */
  char *chemin_travail;		/* Le dossier où trouver les feffXXXX.dat */
  double longueur_maximale;	/* Longueur maximale de chemin */
  unsigned int n_maximal;	/* Nombre maximal de réflexions */
  unsigned int i_maximal;	/* Index maximal du chemin à charger */
  Vecteur3D translation;	/* Pour recaser l'absorbeur au bon endroit */
  struct absorbeur *absorbeur;	/* L'atome absorbeur trouvé */
  struct molecule *molecule;	/* Si une nouvelle molécule doit être créée */
  unsigned int Z_absorbeur_feff; /* Z de l'absorbeur dans paths.dat */
  int code_erreur;		/* 0 : Pas d'erreur */
} FEFF_PATHS_DAT;

/* ---- Modèle obtenu par FEFF ---- */

struct couche; /* Dans couches.h */

typedef struct modele_feff {
  int seuil; /* Seuil d'absorption */
  /* Attention, FEFF : 0 = aucune lacune, 1 = seuil K, 2 = L1, 3 = L2, 4 = L3... */

  struct atome *absorbeur;	/* L'atome central, forcément unique, de potentiel nul */
  int potentiels[ 9 ];		/* Table des potentiels - 0 à 8, 0 = absorbeur */

  FEFF_INP parametres; /* Paramètres utilisés pour utiliser FEFF */

  Variable S0_2;    /* S0_2 associé à cet absorbeur */
  Variable Delta_E; /* Décalage du E0 associé à cet absorbeur */
  Variable Gamma;   /* Libre parcours moyen, pour toutes les couches */

  struct {
    unsigned DE0_general :1; /* VRAI : même valeur de DE0 pour toutes les couches */
    unsigned GAM_general :1; /* VRAI : même valeur de GAMMA pour toutes les couches */
    unsigned av_AFOLP    :1; /* VRAI : le modèle FEFF contient la commande AFOLP */
    unsigned av_NOGEOM   :1; /* VRAI : le modèle FEFF contient la commande NOGEOM */
    unsigned av_XANES    :1; /* VRAI : le modèle FEFF contient la commande XANES */
  } drapeaux;

  unsigned int nbr_chemins; /* Nombre de chemins générés par FEFF */
  unsigned int nbr_couches; /* Nombre de _couches_ générées pour le modèle */
  struct couche   *couches; /* Liste simplement chaînée des couches du modèle */

  unsigned short int ponderation; /* n, pour pondération par k^n */

  unsigned short int type_lpm; /* Modèle de libre parcours moyen à utiliser ? */
  struct graphe *lpm;          /* Graphe du libre parcours moyen */
  double *b_lpm, *c_lpm, *d_lpm; /* Coefficients des splines associés */
  double lpm_val_eta;

  struct graphe *phase_central; /* Déphasage de l'atome absorbeur (rarement utilisé) */
} Modele_FEFF;

Modele_FEFF *FEFF_CreerModele  (void);
Modele_FEFF *FEFF_ChoisirModele(boolean proposer);
int          FEFF_EffacerModele(Modele_FEFF *modele);

void FEFF_termine(void); /* Libérer la mémoire occupée par le modèle FEFF */

/* --------------- Fonctions d'entrée-sortie ------------------------------------ */

void         I_FEFF_Sauver(Modele_FEFF *modele, unsigned int version); /* Sauver FEFF.INP  */
void           FEFF_Sauver(Modele_FEFF *modele, unsigned int version, const char *nom_fichier); /* Sauver FEFF.INP  */
char        *I_FEFF_SauverTemporaire(Modele_FEFF *modele);
Modele_FEFF *FEFF_charge(void); /* Charger FEFF.INP */

int lancer_feff(void);  /* Demande du lieu de travail */
int attend_feff(const char *nom); /* Lancer feff en utilisant nom comme feff.inp */

struct modele_ajuste *FEFF_LirePathsDat(FEFF_PATHS_DAT *infos, struct absorbeur *absorbeur, const char *nom_fichier);

/* --------------- Fonctions de paramétrage proprement dites -------------------- */

void FEFF_ParametresGeneraux(Modele_FEFF *modele); /* Paramètres généraux */
void FEFF_DefinirPotentiels (Modele_FEFF *modele); /* Potentiels & associés */
void FEFF_DefinirStructure  (Modele_FEFF *modele); /* Structure 3D */
void FEFF_ParametresXANES   (Modele_FEFF *modele); /* Commandes FEFF pour XANES */

boolean FEFF_VerifierModele (Modele_FEFF *modele);

/* --------------- Fonctions de gestion des formulaires (du SHELL donc) --------- */

/* --------------- Structures de recouvrement (OVERLAP) - - - - - - - - - - - - - */

FEFF_OVERLAP *FEFF_CreerOverlap  (Modele_FEFF *modele, int potentiel, int pot_recouvre, int nbr_atomes, double distance);
boolean       FEFF_EffacerOverlap(Modele_FEFF *modele, FEFF_OVERLAP *overlap, boolean forcer);

/* --------------- Structures de réflexion simple (SS) - - - - - - - - - - - - -  */

FEFF_SS *FEFF_AjouterCouche(Modele_FEFF *modele, unsigned int numero, int potentiel, int degenerescence, double distance);
boolean  FEFF_EffacerCouche(Modele_FEFF *modele, FEFF_SS *couche, boolean forcer);

/* --------------- Atomes FEFF (ATOMS) - - - - - - - - - - - - - - - - - - - - -  */

boolean     FEFF_AjouterVoisins(Modele_FEFF *modele);
FEFF_atome *FEFF_AjouterAtome  (Modele_FEFF *modele, Atome *atome, int potentiel, boolean alerte);
FEFF_atome *FEFF_CreerAtome    (Modele_FEFF *modele);
void        FEFF_EffacerAtome  (Modele_FEFF *modele, FEFF_atome *atome, boolean forcer);
void        FEFF_EnleverAtome  (Modele_FEFF *modele, Atome *atome, boolean alerte);

boolean     ATM_EstDansModeleFEFF(Modele_FEFF *modele, Atome *atome);
FEFF_atome *ATM_QuelAtomeFEFF    (Modele_FEFF *modele, Atome *atome);

#endif /* INC_FEFF_ */
