/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 2008 Emmanuel CURIS
   Création : le 5 février 1999

   Récupération des données expérimentales pour obtenir µ total

   Interface générale : conditions d'enregistrement, traitement à faire,...
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "exafs.h"
#include "import_xas.h"
#include "interface.h"
#include "serie.h"

#include "int_import_XAS.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LIGNE_PAR_DEFAUT LIGNE_SOLEIL_SAMBA /* Ligne Exafs 2 par défaut */

static BI_InfosSelecteur infos_selecteur = {
  NULL,
  BI_SELECTEUR_MULTIPLE|BI_SELECTEUR_FILTRE_UTILISATEUR|BI_SELECTEUR_AJOUTER_PREDEFINI|BI_SELECTEUR_AVEC_JOKERS|BI_SELECTEUR_MEMORISER_CHEMIN,
  NULL,
  0, NULL,
  0, NULL,
  (void (*)(unsigned int, unsigned int, const char *)) NULL
};

static const char *filtres[] = {
  /* Filtres acquvisu/exafs 2 */
  "*.flu", "*.dat", "*.xan",
  "*.flu, *.xan", "*.flu, *.dat", "*.dat, *.xan",
  "*.dat, *.xan, *.flu"
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

inline static void I_ES_ImportXAS_ActiverFluorescence(int sens);
inline static void I_ES_ImportXAS_ActiverSuccessifs(int sens);
inline static void I_ES_ImportXAS_ActiverTransmission(int sens);

static void I_ES_ImportXAS_InstallerLigne(unsigned int arbre, unsigned int objet, unsigned long ligne);
static void I_ES_ImportXAS_Parametrer_CorrectionTempsMort(struct infos_corrige_tps_mort *infos);

static boolean I_ES_ImportXAS_LigneEstASCII(unsigned int ligne);

static unsigned int I_ES_ImportXAS_LireParametres(Infos_Extraction *infos);

static void I_ES_ImportXAS_PreparerArbre(void);
static int  I_ES_ImportXAS_GererArbre(unsigned int num_objet, int x, int y);

static int I_ES_ImportXAS_FichierASCII  (unsigned int type_ligne, Infos_Extraction *infos);
static int I_ES_ImportXAS_FichierBinaire(unsigned int type_ligne, Infos_Extraction *infos);

static void I_ES_ImportXAS_PreparerSelecteur(unsigned int type_ligne, BI_InfosSelecteur *infos_selecteur);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Paramétrage par l'utilisateur et répartition des tâches ------ */

int I_IMPORT_DonneesXAS(void)
{
  unsigned int ligne_choisie;
  int retour;
  Infos_Extraction infos;  /* Comment gérer les différentes colonnes du fichier à traiter ? */

  if ( FAUX == arbres[ TYPE_LIGNE ].drapeaux.prepare ) {
    I_ES_ImportXAS_PreparerArbre();
    I_ES_ImportXAS_InstallerLigne( TYPE_LIGNE, TL_CHOIX_LIGNE, LIGNE_PAR_DEFAUT );
  }

  retour = ARB_Bloquant( TYPE_LIGNE, txt_titre_arbre_extrait, principale, 0 );
  if ( EXTRAIT_ANNULER == retour ) {
    return +1; /* Annulation : erreur minime */
  }

  /* --- Paramètres d'extraction --- */
  ligne_choisie = I_ES_ImportXAS_LireParametres( &infos );
  if ( LIGNE_ERREUR == ligne_choisie ) {
    /* Erreur dans le paramétrage ! */
    return +2;
  }

  /* Préparation préalable au chargement, en fonction de la ligne */
  switch ( ligne_choisie )
  {
  case LIGNE_LURE_AQUVISU:
  case LIGNE_LURE_D21:
  case LIGNE_HASYLAB_DORIS:
  case LIGNE_NSLS_MXANES:
  case LIGNE_ESRF_SNBL:
  case LIGNE_SOLEIL_SAMBA:
    break;

  case LIGNE_ESRF_ID21:
  case LIGNE_ESRF_ID22:
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Voir pour demander la commande à repérer pour le µ-XANES ?\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    break;

  case LIGNE_BESSY2_:
    /* On force les intervalles à être corrects */
#ifndef AVEC_HDF_
    (void) BI_ERR_alerte( NULL, &ERR_SupportHDFAbsent, VRAI );

    free( infos.colonnes_trans ); infos.colonnes_trans = NULL;
    free( infos.colonnes_fluo ); infos.colonnes_fluo = NULL;
    free( infos.colonnes_det ); infos.colonnes_det = NULL;

    /* On n'a pas encore créé la série */
    return -2;
#endif
    break;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : type de ligne expérimentale inconnu [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, ligne_choisie );
    OPTION_A_FAIRE( "Gérer ce type de ligne expérimentale..." );
    break;
  }

  if ( I_ES_ImportXAS_LigneEstASCII( ligne_choisie ) == VRAI ) {
    /* Import d'un fichier ASCII : routines génériques */
    retour = I_ES_ImportXAS_FichierASCII( ligne_choisie, &infos );
  } else {
    /* Import d'un fichier binaire : forcément (très) dépendant de la ligne ! */
    retour = I_ES_ImportXAS_FichierBinaire( ligne_choisie, &infos );
  }
  if ( retour != 0 ) {
    /* Une erreur s'est produite pendant le chargement... */
  }

  /* On efface les infos d'extraction : plus nécessaires... */
  free( infos.colonnes_trans ); infos.colonnes_trans = NULL;
  free( infos.colonnes_fluo ); infos.colonnes_fluo = NULL;
  free( infos.colonnes_det ); infos.colonnes_det = NULL;

  /* Normalement, on a créé des graphes, voire des séries : on met à jour */
  I_PrincipaleGRF_RefaireAscenseurs( VRAI, VRAI );

  /* On a fini, sans problème */
  return 0;
}

/* -------- Lecture des paramètres indiqués par l'utilisateur --------
      Renvoie le numéro de ligne retenu
 */

static unsigned int I_ES_ImportXAS_LireParametres(Infos_Extraction *infos)
{
  unsigned int ligne;

  if ( NULL == infos ) {
    (void) fprintf( stderr, "%s, %s l.%u : Appel avec infos = NULL !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return LIGNE_ERREUR;
  }

  /* 1. Initialisation de la structure */

  /* Aucune propriété par défaut */
  infos->proprietes = 0;

  /* Aucune colonne par défaut */
  infos->colonnes_trans = NULL; infos->nbr_I_trans = 0;
  infos->colonnes_fluo  = NULL; infos->nbr_I_fluo  = 0;
  infos->colonnes_det   = NULL; infos->nbr_I_det   = 0;

  /* Aucune table de sélection de spectres successifs par défaut */
  infos->filtre_MX = NULL; infos->nbr_I_MX = 0;

  /* 2. On demande le numéro de ligne */
  ligne = OBJ_OptionMenu( TYPE_LIGNE, TL_CHOIX_LIGNE, 0, QUESTION );

  /* 3. Que garde-t-on comme données du fichier */
  infos->que_garder = 0;
  if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_GARDER_I0, QUESTION ) ) {
    infos->que_garder |= DETECT_GARDER_I0;
  }
  if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_TRANSMISSION, QUESTION ) ) { /* Détection en transmission */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_GARDER_ITRANS, QUESTION ) ) {
      infos->que_garder |= DETECT_GARDER_TRANS;
    }
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_GARDER_SIGNAL, QUESTION ) ) {
      infos->que_garder |= DETECT_GARDER_SIGNAL;
    }
  }
  if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_FLUORESCENCE, QUESTION ) ) { /* Détection en fluorescence */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_GARDER_IFLUO, QUESTION ) ) {
      infos->que_garder |= DETECT_GARDER_FLUO;
    }
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_GARDER_SIGNAL, QUESTION ) ) {
      infos->que_garder |= DETECT_GARDER_SIGNAL;
    }
  }
  if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_ELECTRONS, QUESTION ) ) { /* Détection d'électrons */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_EL_GARDER, QUESTION ) ) {
      infos->que_garder |= DETECT_GARDER_EL;
    }
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_GARDER_SIGNAL, QUESTION ) ) {
      infos->que_garder |= DETECT_GARDER_SIGNAL;
    }
  }
  if ( 0 == infos->que_garder ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : aucune information conservée !\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    (void) BI_ERR_alerte( NULL, &ERR_AucuneInformation, VRAI );
    return LIGNE_ERREUR; /* Erreur embêtante ! */
  }

  /* 4. Si la ligne s'y prête : description du codage ASCII*/
  if ( I_ES_ImportXAS_LigneEstASCII( ligne ) == VRAI ) {
    /* Combien de lignes d'en-tête (commentaires exclus) ? */
    infos->nombre_lignes_en_tete = OBJ_LireEntier_U( TYPE_LIGNE, TL_ENTETE );

    /* Quel caractère introduit les lignes de commentaires ? */
    infos->commentaire = *lit_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, NULL, TXT_COPIER );

    /* Quel caractère introduit les lignes avec des indications complémentaires ? */
    infos->indication = '?';

    /* Combien de colonnes au total ? */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, QUESTION ) ) {
      infos->nbr_colonnes = 0; /* Détection automatique du nombre de colonnes */
    } else {
      infos->nbr_colonnes = OBJ_LireEntier_U( TYPE_LIGNE, TL_COLONNES );
    }

    /* Quelle est la colonne des X ? [attention, numérotées à partir de 1, 0 = absente] */
    infos->col_x = OBJ_LireEntier_U( TYPE_LIGNE, TL_COLONNE_X );

    /* Quelle est la colonne des I0 ? [attention, numérotée à partir de 0 car indispensable !]*/
    infos->col_i0 = OBJ_LireEntier_U( TYPE_LIGNE, TL_COL_I0 ) - 1;

    /* Quelles sont les colonnes des I/transmission ? */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_TRANSMISSION, QUESTION ) ) {
      infos->colonnes_trans = TXT_genere_intervalles( lit_chaine( TYPE_LIGNE, TL_TRANS_COLS, 0, NULL, TXT_REMPLACE ),
						      &(infos->nbr_I_trans) );
    }

    /* Quelles sont les colonnes des I/fluorescence ? */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_FLUORESCENCE, QUESTION ) ) {
      infos->colonnes_fluo = TXT_genere_intervalles( lit_chaine( TYPE_LIGNE, TL_FLUO_COLS, 0, NULL, TXT_REMPLACE ),
						     &(infos->nbr_I_fluo) );
    }

    /* Quelles sont les colonnes des I/détection d'électrons ? */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_ELECTRONS, QUESTION ) ) {
      infos->colonnes_det = TXT_genere_intervalles( lit_chaine( TYPE_LIGNE, TL_EL_COLONNES, 0, NULL, TXT_REMPLACE ),
						    &(infos->nbr_I_det) );
    }
  } else {
    /* On l'indique dans les propriétés */
    infos->proprietes |= DETECT_FICHIER_BINAIRE;
  }

  /* 5. Comment traiter plusieurs détecteurs ? */
  if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_MD_MOYENNER, QUESTION ) ) {
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, MELM_SIGMA_LOG, QUESTION ) )
      infos->MD_traitement = DETECT_MULTI_MOY_LOG;
    else
      infos->MD_traitement = DETECT_MULTI_MOYENNER;
  } else {
    infos->MD_traitement = DETECT_MULTI_SEPARER; /* Par défaut, un spectre par détecteur */
  }

  /* 6. Y a-t-il plusieurs spectres à la suite ? */
  if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_SUCCESSIFS, QUESTION ) ) {
    infos->proprietes |= DETECT_MULTI_SPECTRE;

    /* Comment les traiter ? */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_MX_MOYENNER, QUESTION ) ) {
      infos->MX_traitement = DETECT_MULTI_MOYENNER;
    } else if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_MX_FONDRE, QUESTION ) ) {
      infos->MX_traitement = DETECT_MULTI_FONDRE;
    } else {
      infos->MX_traitement = DETECT_MULTI_SEPARER; /* Par défaut, un spectre par détecteur */
    }

    /* Quels spectres garder ? */
    infos->filtre_MX = TXT_genere_intervalles( lit_chaine( TYPE_LIGNE, TL_FILTRE_MX, 0, NULL, TXT_REMPLACE ),
					       &(infos->nbr_I_MX) );

    /* Combien y en a-til ? */
    if ( OBJ_choisit( TYPE_LIGNE, TL_AUTO_SUCC, QUESTION ) == VRAI ) {
      infos->nbr_successifs = 0; /* Détection automatique */
    } else {
      infos->nbr_successifs = OBJ_LireEntier_S( TYPE_LIGNE, TL_NBR_SUCCESSIF);
    }
  } else {
    infos->nbr_successifs = 1;	/* Un seul bloc-spectre dans le fichier */
  }

  /* 7. Que faire après le chargement d'un fichier / spectre ? */
  infos->post_traitement = 0;
  if ( VRAI == OBJ_choisit(TYPE_LIGNE, EXTRAIT_TRIER, QUESTION) )
    infos->post_traitement |= DETECT_POST_TRIER;
  if ( VRAI == OBJ_choisit(TYPE_LIGNE, EXTRAIT_DOUBLONS, QUESTION) )
    infos->post_traitement |= DETECT_POST_DOUBLONS;
  if ( VRAI == OBJ_choisit(TYPE_LIGNE, TL_NORMALISER, QUESTION) ) {
    infos->post_traitement |= DETECT_POST_NORMALISER;

    /* À revoir : comment normaliser ? */
  }
  if ( VRAI == OBJ_choisit(TYPE_LIGNE, EXTRAIT_CORRIGE, QUESTION) ) {
    infos->post_traitement |= DETECT_POST_CORRECT;

    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_FLUORESCENCE, QUESTION ) ) {
      I_ES_ImportXAS_Parametrer_CorrectionTempsMort( &(infos->temps_mort ) );
    }
  }
  if ( VRAI == OBJ_choisit(TYPE_LIGNE, TL_NRJ_KEV, QUESTION) )
    infos->post_traitement |= DETECT_POST_X_KEV;

  /* 8. Informations complémentaires à chaque format/ligne */
  switch ( ligne )
  {
  }

  /* Terminé : on renvoie le numéro de ligne */
  return ligne;
}

/* -------- Gestion de l'arbre de paramétrage -------- */

static int I_ES_ImportXAS_GererArbre(unsigned int numero_objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  boolean etat;

  switch ( numero_objet )
  {
  case TL_COL_AUTO:
    etat = OBJ_choisit( TYPE_LIGNE, numero_objet, ECHANGE);
    OBJ_active( TYPE_LIGNE, TL_COLONNES, !etat );
    return BI_EVENEMENT_TRAITE;

  case TL_TRANSMISSION: /* Type de manipulation */
    etat = OBJ_choisit( TYPE_LIGNE, numero_objet, ECHANGE);
    I_ES_ImportXAS_ActiverTransmission( etat );
    return BI_EVENEMENT_TRAITE;

  case TL_FLUORESCENCE:
    etat = OBJ_choisit(TYPE_LIGNE, numero_objet, ECHANGE);
    I_ES_ImportXAS_ActiverFluorescence( etat );
    return BI_EVENEMENT_TRAITE;

/*   case TL_CHOIX_LIGNE: */ /* Type ligne EXAFS III du LURE */
/*     { */
/*       int nv_ligne; */

/*       OBJ_choisit(TYPE_LIGNE, TL_CHOIX_LIGNE, FORCE_OUI); */
/*       nv_ligne = SS_menu_texte( textes_type_ligne, NOMBRE_LIGNES_EXPERIMENTALES, titre_menu_lignes, TYPE_LIGNE, TL_CHOIX_LIGNE); */
/*       if ( ( nv_ligne >= 0 ) && ( nv_ligne != (int) type_ligne ) ) { */
/* 	type_ligne = (unsigned int) nv_ligne; */
/* 	I_ES_ImportXAS_InstallerLigne( type_ligne ); */
/*       } */
/*       OBJ_choisit(TYPE_LIGNE, TL_CHOIX_LIGNE, FORCE_NON); */
/*     } */
/*     return BI_EVENEMENT_TRAITE; */

  case TL_SUCCESSIFS:
    I_ES_ImportXAS_ActiverSuccessifs( ECHANGE );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

inline static void I_ES_ImportXAS_ActiverFluorescence(int sens)
{
  OBJ_active( TYPE_LIGNE, TL_NBR_FLUO, sens );
  OBJ_active( TYPE_LIGNE, TL_GARDER_IFLUO, sens );
  OBJ_active( TYPE_LIGNE, TL_FLUO_COLS, sens );    
}

inline static void I_ES_ImportXAS_ActiverSuccessifs(int sens)
{
  int etat, etat_auto;

  etat = OBJ_choisit( TYPE_LIGNE, TL_SUCCESSIFS, sens );
  OBJ_active( TYPE_LIGNE, TL_MX_MOYENNER, etat );
  OBJ_active( TYPE_LIGNE, TL_MX_FONDRE, etat );
  OBJ_active( TYPE_LIGNE, TL_MX_SEPARER, etat );
  OBJ_active( TYPE_LIGNE, TL_AUTO_SUCC, etat );
  etat_auto = OBJ_choisit( TYPE_LIGNE, TL_AUTO_SUCC, QUESTION );
  OBJ_active( TYPE_LIGNE, TL_NBR_SUCCESSIF, etat & !etat_auto);
  OBJ_active( TYPE_LIGNE, TL_FILTRE_MX, etat );
}

inline static void I_ES_ImportXAS_ActiverTransmission(int sens)
{
  OBJ_active( TYPE_LIGNE, TL_NBR_TRANS, sens );
  OBJ_active( TYPE_LIGNE, TL_GARDER_ITRANS, sens );
  OBJ_active( TYPE_LIGNE, TL_TRANS_COLS, sens );
}

static void I_ES_ImportXAS_InstallerLigne(unsigned int arbre __attribute__((__unused__)), unsigned int objet __attribute__((__unused__)), unsigned long ligne)
{
  unsigned int nbr_colonnes;

  if ( ligne >= NOMBRE_LIGNES_EXPERIMENTALES ) {
    (void) fprintf( stderr, "%s, %s l.%u : type de ligne inconnu [%lu]\n",
		    __FILE__, __FUNCTION__, __LINE__, ligne );
    OPTION_A_FAIRE( "Gérer ce type de ligne expérimentale..." );
    return;
  }

  /* On adapte les différents champs */
  switch ( ligne )
  {
  case LIGNE_LURE_AQUVISU:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "#", TXT_COPIER ); /* Commentaires : # */
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_NBR_TRANS, 1 ); /* Un seul détecteur en transmission, en général... */
    nbr_colonnes = 2; /* X, I0 */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_TRANSMISSION, QUESTION ) )
      nbr_colonnes += OBJ_LireEntier_U( TYPE_LIGNE, TL_NBR_TRANS );
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_FLUORESCENCE, QUESTION ) )
      nbr_colonnes += OBJ_LireEntier_U( TYPE_LIGNE, TL_NBR_FLUO );
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_ELECTRONS, QUESTION ) )
      nbr_colonnes += OBJ_LireEntier_U( TYPE_LIGNE, TL_EL_NBR_DET );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNES, nbr_colonnes );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en eV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_EV, FORCE_OUI );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 2); /* I0 en 2e colonne */
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_NON ); /* On ne peut pas avoir plusieurs spectres à la suite */ 
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_NON );
    met_chaine( TYPE_LIGNE, TL_TRANS_COLS, 0, "3", TXT_COPIER );
    break;

  case LIGNE_LURE_D21:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 2 ); /* 2 lignes d'en-tête à ignorer */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, ";", TXT_COPIER ); /* Pas de commentaire, mais ; pourrait servir... */
    nbr_colonnes = 2; /* X, I0 */
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_TRANSMISSION, QUESTION ) )
      nbr_colonnes += OBJ_LireEntier_U( TYPE_LIGNE, TL_NBR_TRANS );
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_FLUORESCENCE, QUESTION ) )
      nbr_colonnes += OBJ_LireEntier_U( TYPE_LIGNE, TL_NBR_FLUO );
    if ( VRAI == OBJ_choisit( TYPE_LIGNE, TL_ELECTRONS, QUESTION ) )
      nbr_colonnes += OBJ_LireEntier_U( TYPE_LIGNE, TL_EL_NBR_DET );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNES, nbr_colonnes );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en eV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_EV, FORCE_OUI );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 2); /* I0 en 2e colonne */
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_OUI ); /* On peut avoir plusieurs spectres à la suite... */ 
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_NON ); /* ... mais c'est rare */
    break;

  case LIGNE_HASYLAB_DORIS:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "!", TXT_COPIER ); /* Commentaires : ! */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */
    OBJ_active( TYPE_LIGNE, TL_COLONNE_X, VRAI ); /* La colonne des X peut changer ! */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en eV, en colonne 1 en general mais pas toujours */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_EV, FORCE_OUI );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 4); /* I0 en 4e colonne [fixe ?] */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_TRANS_COLS, 6 ); /* I1 en colonne 8 [fixe ?] */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_FLUO_COLS, 8 ); /* If en colonne 8 [fixe ?] */
    I_ES_ImportXAS_ActiverFluorescence( FORCE_OUI ); /* En général, on travaille en fluorescence */
    OBJ_choisit(TYPE_LIGNE, TL_FLUORESCENCE, FORCE_OUI);
    break;

  case LIGNE_NSLS_MXANES:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 1 ); /* Pas d'en-tête à ignorer */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "$", TXT_COPIER ); /* pas de commentaire a priori... */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 0 ); /* X : énergies en eV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_AUTRE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_KEV, FORCE_OUI );
    break;

  case LIGNE_ESRF_SNBL:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer : ce seront des commentaires */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "#", TXT_COPIER ); /* Commentaires : # */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en eV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_EV, FORCE_OUI );
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_OUI ); /* On peut avoir plusieurs spectres à la suite... */ 
    OBJ_choisit( TYPE_LIGNE, TL_AUTO_SUCC , FORCE_OUI ); /* et on laisse LASE les compter */
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_OUI ); /* et c'est le plus fréquent */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 3); /* I0 en 3e colonne */
    break;

  case LIGNE_ESRF_BM05:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer : ce seront des commentaires */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "#", TXT_COPIER ); /* Commentaires : # */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */

    OPTION_A_FAIRE( "Definir le format, ligne BM05, EXAFS" );
    break;

  case LIGNE_ESRF_ID21:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer : ce seront des commentaires */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "#", TXT_COPIER ); /* Commentaires : # */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en keV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_KEV, FORCE_OUI );
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_OUI ); /* On peut avoir plusieurs spectres à la suite... */ 
    OBJ_choisit( TYPE_LIGNE, TL_AUTO_SUCC , FORCE_OUI ); /* et on laisse LASE les compter */
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_OUI ); /* et c'est le plus fréquent */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 6); /* I0 en 6e colonne */
    break;

  case LIGNE_ESRF_ID22:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer : ce seront des commentaires */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "#", TXT_COPIER ); /* Commentaires : # */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en keV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_KEV, FORCE_OUI );
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_OUI ); /* On peut avoir plusieurs spectres à la suite... */ 
    OBJ_choisit( TYPE_LIGNE, TL_AUTO_SUCC , FORCE_OUI ); /* et on laisse LASE les compter */
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_OUI ); /* et c'est le plus fréquent */
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 7); /* I0 en 7e colonne [?] */
    break;

  case LIGNE_BESSY2_:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer : fichier binaire... */
    OBJ_choisit( TYPE_LIGNE, TL_COL_AUTO, FORCE_OUI ); /* Détection automatique du nombre de colonnes */
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_NON ); /* On ne peut pas avoir plusieurs spectres à la suite [actuellement] */ 
    OBJ_choisit( TYPE_LIGNE, TL_AUTO_SUCC , FORCE_NON );
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_NON );
    I_ES_ImportXAS_ActiverFluorescence( FORCE_OUI ); /* En général, on travaille en fluorescence */
    OBJ_choisit(TYPE_LIGNE, TL_FLUORESCENCE, FORCE_OUI);
    I_ES_ImportXAS_ActiverTransmission( FORCE_NON );
    OBJ_choisit(TYPE_LIGNE, TL_TRANSMISSION, FORCE_NON);
    met_chaine( TYPE_LIGNE, TL_FLUO_COLS, 0, "1-", TXT_COPIER ); /* Pour être sûr de créer ! */
    break;

  case LIGNE_SOLEIL_SAMBA:
    OBJ_MettreEntier_U( TYPE_LIGNE, TL_ENTETE, 0 ); /* Pas d'en-tête à ignorer */
    met_chaine( TYPE_LIGNE, TL_COMMENTAIRE, 0, "#", TXT_COPIER ); /* Commentaires : # */
    OPTION_A_FAIRE( "Vérifier le nombre de colonnes réel..." );
    nbr_colonnes = 50;
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNES, nbr_colonnes );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COLONNE_X, 1 ); /* X : énergies en eV, en colonne 1 */
    OBJ_choisit( TYPE_LIGNE, TL_X_ENERGIE, FORCE_OUI );
    OBJ_choisit( TYPE_LIGNE, TL_NRJ_EV, FORCE_OUI );
    OBJ_MettreEntier_U ( TYPE_LIGNE, TL_COL_I0, 5 ); /* I0 en colonne 5 */
    OBJ_active ( TYPE_LIGNE, TL_SUCCESSIFS, FORCE_NON ); /* On ne peut pas avoir plusieurs spectres à la suite */
    I_ES_ImportXAS_ActiverSuccessifs( FORCE_NON );
    met_chaine( TYPE_LIGNE, TL_TRANS_COLS, 0, "6", TXT_COPIER );
    met_chaine( TYPE_LIGNE, TL_FLUO_COLS, 0, "13-19", TXT_COPIER );
    break;

  default:
    (void) fprintf( stderr, "%s, %s l.%u : Ligne inconnue [%lu]\n",
		    __FILE__, __FUNCTION__, __LINE__, ligne );
    OPTION_A_FAIRE( "Gérer ce type de ligne expérimentale..." );
  }
}

/* -------- Préparation de l'arbre de paramétrage -------- */

static void I_ES_ImportXAS_PreparerArbre(void)
{
  if ( FAUX == arbres[ TYPE_LIGNE ].drapeaux.init ) {
    int retour;

    /* On prépare les sous-menus */
    OBJ_ChangerType( TYPE_LIGNE, TL_CHOIX_LIGNE, G_SSMENU );
    OBJ_MettreMenu ( TYPE_LIGNE, TL_CHOIX_LIGNE,
		     NOMBRE_LIGNES_EXPERIMENTALES, textes_type_ligne, titre_menu_lignes,
		     I_ES_ImportXAS_InstallerLigne );
    retour = OBJ_OptionMenu( TYPE_LIGNE, TL_CHOIX_LIGNE, LIGNE_PAR_DEFAUT, METTRE );

    retour = ARB_Initialiser( TYPE_LIGNE, "menu_absorption.html#extraire", I_ES_ImportXAS_GererArbre, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return;
    }
  }

  /* Ligne du haut */
  OBJ_deplace( TYPE_LIGNE, TL_TXT_ENTETE  , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_ENTETE      , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_LIGNES  , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_COMMENT , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_COMMENTAIRE , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_COLONNES, 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_COLONNES    , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_COL_AUTO    , 0, -1, DEPLACE_PIXELS );

  /* Analyse des X */
  OBJ_deplace( TYPE_LIGNE, TL_X_ENERGIE    , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_ENERGIE  , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_FOND_ENERGIES, 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_ANGLES       , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_ANGLE    , 0, -1, DEPLACE_PIXELS );

  OBJ_agrandit(TYPE_LIGNE, TL_FOND_MULTIX, 0, +3 );
  OBJ_deplace( TYPE_LIGNE, TL_MX_SEPARER, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_MX_FONDRE , 0, +2, DEPLACE_PIXELS );

  /* Détecteurs */
  OBJ_deplace( TYPE_LIGNE, TL_TXT_TRANS_COL, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TRANS_COLS   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_FLUO_COL , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_FLUO_COLS    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_EL_COLS  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( TYPE_LIGNE, TL_EL_COLONNES  , 0, +1, DEPLACE_PIXELS );

  /* Les sur-titres... */
  OBJ_deplace( TYPE_LIGNE, TL_TXT_X        , 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( TYPE_LIGNE, TL_TXT_DETECTEUR, 0, 500, DEPLACE_RELATIF );

  OBJ_deplace( TYPE_LIGNE, TL_FILTRE_MX, 0, -500, DEPLACE_RELATIF );

#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_LIGNE , txt_ligne_acquisition);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_LIGNE_AUTO, txt_automatique);

  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_ENTETE  , txt_en_tete);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_LIGNES  , txt_lignes);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_COMMENT , txt_commentaires);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_COLONNES, txt_colonnes);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_COL_AUTO    , txt_automatique);

  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_NLE_SERIE, txt_nouvelle_serie);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_FICHIER  , txt_une_serie_par_fichier);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_GARDER   , txt_garder_les_spectres);

  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_X        , txt_abscisses);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_COL_NRJ  , txt_colonne);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_ENERGIE  , txt_energies);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_ANGLE    , txt_angles);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_AUTRES   , txt_autre);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_SUCCESSIF, txt_spec_successifs);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_AUTO_SUCC    , txt_auto);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_MX_MOYENNER  , txt_moyenner);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_MX_SEPARER   , txt_separer);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_MX_FONDRE    , txt_fondre);

  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_DETECTEUR, txt_Detecteurs);
  OBJ_TraduireTexte( TYPE_LIGNE, LGN_GARDE_I0    , txt_faisceau_incident);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_GARDER_I0    , txt_garder);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_TRANS    , txt_transmission);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_GARDER_ITRANS, txt_garder);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_TRANS_DET, txt_detecteurs);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_TRANS_COL, txt_colonnes);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_FLUO     , txt_fluorescence);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_FLUO_COL , txt_colonnes);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_FLUO_DET , txt_detecteurs);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_GARDER_IFLUO , txt_garder);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_ELECTRONS, txt_det_electrons);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_EL_COLS  , txt_colonnes);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_EL_DET   , txt_detecteurs);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_EL_GARDER    , txt_garder);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_MULTI    , txt_plusieurs_detecteurs);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_MD_MOYENNER  , txt_moyenner);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_MD_SEPARER   , txt_separer);

  OBJ_TraduireTexte( TYPE_LIGNE, TL_TXT_APRES    , txt_apres_chargement);
  OBJ_TraduireTexte( TYPE_LIGNE, EXTRAIT_TRIER   , txt_trier);
  OBJ_TraduireTexte( TYPE_LIGNE, EXTRAIT_DOUBLONS, txt_moy_doublons);
  OBJ_TraduireTexte( TYPE_LIGNE, TL_NORMALISER   , txt_normaliser);
  OBJ_TraduireTexte( TYPE_LIGNE, EXTRAIT_CORRIGE , txt_corriger);

  OBJ_TraduireTexte( TYPE_LIGNE, EXTRAIT_OK     , txt_lancer);
  OBJ_TraduireTexte( TYPE_LIGNE, EXTRAIT_ANNULER, txt_annuler);
#endif

#ifndef _DEVELOPPE_ /* Fonctions encore inactives */
  OBJ_active( TYPE_LIGNE, TL_LIGNE_AUTO  , FORCE_NON );
  OBJ_active( TYPE_LIGNE, EXTRAIT_CORRIGE, FORCE_NON );
  OBJ_active( TYPE_LIGNE, TL_ELECTRONS   , FORCE_NON );
#endif

  arbres[ TYPE_LIGNE ].drapeaux.prepare = VRAI;
}

/* ~-~-~-~-~-~-~-~-~ Informations sur les lignes ~-~-~-~-~-~-~-~-~ */

static boolean I_ES_ImportXAS_LigneEstASCII(unsigned int ligne)
{
  switch ( ligne )
  {
  case LIGNE_BESSY2_:		/* Format HDF : binaire */
    return FAUX;
  }

  /* En général, les fichiers sont en ASCII */
  return VRAI;
}

/* ~-~-~-~-~-~-~-~-~ Paramétrages complémentaires ~-~-~-~-~-~-~-~-~ */

/* -------- Correction du temps mort en fluorescence -------- */

static void I_ES_ImportXAS_Parametrer_CorrectionTempsMort(struct infos_corrige_tps_mort *infos)
{
  /* Valeurs par défaut : fichier Hasylab/DORIS avec SCA analogique */
  struct infos_demande infos_question[ 4 ] = {
    { BI_CHAMP_DEMANDE_ENTIER_U, "Tps mort col.", NULL, BI_EDITABLE_MIN | BI_EDITABLE_ENTIER, 0.0, 0.0, 0.0, 0, NULL, {3}, {3}, {0} }, /* Colonne du temps mort */
    { BI_CHAMP_DEMANDE_REEL    , "Constante"    , NULL, BI_EDITABLE_MIN                     , 0.0, 0.0, 0.0, 0, NULL, {0}, {0}, {0} }, /* Constante de conversion */
    { BI_CHAMP_DEMANDE_ENTIER_U, "Tps mes. col.", NULL, BI_EDITABLE_MIN | BI_EDITABLE_ENTIER, 0.0, 0.0, 0.0, 0, NULL, {12}, {12}, {0} }, /* Colonne du temps de mesure */
    { BI_CHAMP_DEMANDE_REEL    , "Temps mes."   , NULL, BI_EDITABLE_MIN                     , 0.0, 0.0, 0.0, 0, NULL, {0}, {0}, {0} }, /* Temps de comptage fixe */
  };
  int retour __attribute__((__unused__));

  infos_question[ 1 ].valeur_initiale.reel = 2.17752e-6;
  infos_question[ 1 ].valeur_defaut.reel   = 2.17752e-6;

  infos_question[ 3 ].valeur_initiale.reel = 10.0;
  infos_question[ 3 ].valeur_defaut.reel   = 10.0;

  /* Demande à l'utilisateur */
  retour = BI_DemandeMultiple( NULL, &Q_ParametresTempsMort, 4, infos_question );

  /* Lecture des valeurs */
  infos->constante    = infos_question[ 1 ].valeur_donnee.reel;
  infos->col_tpsmort  = (unsigned int) ( infos_question[ 0 ].valeur_donnee.entier_U );
  infos->col_temps    = (unsigned int) ( infos_question[ 2 ].valeur_donnee.entier_U );
  infos->temps_mesure = infos_question[ 3 ].valeur_donnee.reel;
}

/* ~-~-~-~-~-~-~-~-~ Chargement proprement dit ~-~-~-~-~-~-~-~-~ */

/* -------- Lignes avec fichier au format ASCII : une colonne par variable -------- */

static int I_ES_ImportXAS_FichierASCII(unsigned int type_ligne, Infos_Extraction *infos)
{
  char **noms = NULL;      /* Les adresses des noms des fichiers à charger */
  int retour;
  unsigned long i, nombre_fichiers = 0;

/*   EXT_ListeGraphes resultats = { NULL, NULL, 0, NULL, NULL };  /\* Les graphes créés par le chargement *\/ */
  Serie_Graphes *serie = NULL;

  /* On prépare les extensions utilisables comme masque */
  I_ES_ImportXAS_PreparerSelecteur( type_ligne, &infos_selecteur );

  /* On demande les fichiers à charger */
  retour = BI_SelecteurChargement( &infos_selecteur, &nombre_fichiers, &noms, principale );
  if ( ( BI_SELECTEUR_ANNULER == retour ) || ( 0 == nombre_fichiers ) || ( NULL == noms ) ) {
    /* Chargement annulé */
#ifdef _DEVELOPPE_
    if ( ( ( NULL == noms ) && ( nombre_fichiers > 0 ) ) ||
	 ( ( NULL != noms ) && ( 0 == nombre_fichiers ) ) ) {
      (void) fprintf(stderr, "%s, %s l.%u : Incohérence dans le retour du sélecteur : %lu fichiers mais fichiers = %p\n",
		     __FILE__, __FUNCTION__, __LINE__, nombre_fichiers, (void *) noms);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }
#endif
    return retour;
  }

  /* On crée la série qui contient les spectres créés */
  if ( OBJ_choisit( TYPE_LIGNE, TL_NLE_SERIE, QUESTION ) == VRAI ) {
    char *titre;

    if ( ( 1 == nombre_fichiers ) || ( VRAI == OBJ_choisit(TYPE_LIGNE, TL_SERIE_FICHIER, QUESTION) ) ) {
      titre = TXT_genere_chaine( txt_titre_serie_un_fichier, noms[ 0 ] );
    } else {
      titre = TXT_genere_chaine( txt_titre_serie_n_fichiers, nombre_fichiers );
    }
    serie = SERIE_GRF_CreerSerie(serie_courante, titre);
    free( titre ); titre = NULL;
  } else if ( ( NULL == serie_courante ) || ( VRAI == serie_courante->drapeaux.protegee ) ) {
    char *titre;

    if ( 1 == nombre_fichiers ) {
      titre = TXT_genere_chaine( txt_titre_serie_un_fichier, noms[ 0 ]);
    } else {
      titre = TXT_genere_chaine( txt_titre_serie_n_fichiers, nombre_fichiers );
    }
    serie = SERIE_GRF_CreerSerie(serie_courante, titre);
    free( titre ); titre = NULL;
  } else {
    serie = serie_courante;
  }

  if ( NULL == serie ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer une série\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) BI_ERR_alerte( NULL, &ERR_ImpossibleCreerSerie, FAUX ); /* Pas besoin qu'elle soit bloquante ! */

    for( i = 0 ; i < nombre_fichiers ; i++ )
      free( noms[ i ] );
    free( noms );
    return -3;
  }

  /* Pour chaque fichier indiqué, on réalise l'extraction */
  for ( i = 0 ; i < nombre_fichiers ; i++ ) {
    retour = EXTRAIT_ChargerFichier( noms[ i ], infos, serie, type_ligne );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans EXTRAIT_ChargerFichier [%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }

    /* Toutes les informations sur les spectres chargés sont dans infos :
       on pourrait donc ici s'en servir pour faire des travaux "croisés" entre spectres... */

    /* Si besoin, on crée une nouvelle série avant de passer au fichier suivant */
    if ( ( VRAI == OBJ_choisit(TYPE_LIGNE, TL_NLE_SERIE, QUESTION) ) &&
	 ( VRAI == OBJ_choisit(TYPE_LIGNE, TL_SERIE_FICHIER, QUESTION) ) && ( i < nombre_fichiers - 1 ) ) {
      char *titre;
      
      titre = TXT_genere_chaine( txt_titre_serie_un_fichier, noms[ i + 1 ] );
      serie = SERIE_GRF_CreerSerie( serie, titre );
      free( titre );
    }

    free( noms[ i ]); noms[ i ] = NULL; /* Plus besoin de ce nom */
  }
  free( noms ); noms = NULL;

  if ( NULL == serie->premier_graphe ) {
    /* On n'a rien chargé -> ne pas garder une série vide ! */
    SERIE_GRF_EffacerSerie( serie, VRAI );
  }

  /* Terminé sans problème */
  return 0;
}

/* -------- Lignes avec fichier au format binaire -------- */

static int I_ES_ImportXAS_FichierBinaire(unsigned int type_ligne, Infos_Extraction *infos)
{
  int retour;

/*   EXT_ListeGraphes resultats = { NULL, NULL, 0, NULL, NULL };  /\* Les graphes créés par le chargement *\/ */

  switch ( type_ligne )
  {
  case LIGNE_BESSY2_:		/* Dérivé du format HDF */
    retour = I_ES_ImportXAS_Bessy2( infos );
    break;

  default:
    (void) fprintf( stderr, "%s, %s l.%u : Code ligne inconnu ou non-binaire [%u]\n",
		    __FILE__, __FUNCTION__, __LINE__, type_ligne );
    return -20;
  }

  return retour;
}

/* -------- Initialiser le sélecteur de fichier ---------- */

static void I_ES_ImportXAS_PreparerSelecteur(unsigned int type_ligne, BI_InfosSelecteur *infos_selecteur)
{
  const char *nom_ligne = NULL/* , *extensions[] = NULL */;
  unsigned long nbr_extensions = 0, decalage_ligne = 0;

  free( infos_selecteur->titre ); infos_selecteur->titre = NULL;

  /* On récupère le nom de la ligne */
  if ( type_ligne < NOMBRE_LIGNES_EXPERIMENTALES ) {
    nom_ligne = textes_type_ligne[ type_ligne ];
  } else {
    nom_ligne = "<<Inconnue>>";
  }

  /* On en déduit le titre du sélecteur */
  infos_selecteur->titre = TXT_genere_chaine( txt_titre_selecteur_extrait, nom_ligne );

  /* On s'occupe des filtres classiques */
  switch ( type_ligne )
  {
  case LIGNE_LURE_AQUVISU:
    nbr_extensions = 7;
    decalage_ligne = 0;
    break;

  case LIGNE_LURE_D21:
    nbr_extensions = 7;
    decalage_ligne = 0;
    break;

  case LIGNE_HASYLAB_DORIS:
    break;

  case LIGNE_NSLS_MXANES:
    break;

  case LIGNE_ESRF_SNBL:
    break;

  case LIGNE_ESRF_BM05:
    break;

  case LIGNE_ESRF_ID21:
    break;

  case LIGNE_ESRF_ID22:
    break;

  default:
    nbr_extensions = 0;
  }

  /* On libère les anciennes extensions [table seulement : extensions en const] */
  free( infos_selecteur->filtres ); infos_selecteur->filtres = NULL;
  infos_selecteur->nbr_filtres = 0;

  /* On remplit le tableau */
  if ( nbr_extensions > 0 ) {
    infos_selecteur->filtres = (const char **) malloc( sizeof(char *) * nbr_extensions );
    if ( NULL == infos_selecteur->filtres ) {
      ERR_Memoire( "Table des extensions de fichier", sizeof(char *) * nbr_extensions );
    } else {
      unsigned int i;

      for ( i = 0 ; i < nbr_extensions ; i++ ) {
	infos_selecteur->filtres[ i ] = filtres[ decalage_ligne + i ];
      }
    }
  }

  /* Terminé */
}
