/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 2006, Emmanuel CURIS
   Création : 2 juin 2006

   Interface graphique

   Soustraction d'un fond continu à un graphe
   Dans le cas du XAS : ôter la contribution de la matrice (µm)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "automatise.h"
#include "interface.h"
#include "operations.h"		/* Pour la dérivée */
#include "serie.h"
#include "tracer.h"
#include "xanes.h"

/* -------- Variables locales -------- */

struct infos_fond_continu {
  /* Les paramètres nécessaires à l'extraction */
  MAT_infos parametres;

  /* Infos complémentaires pour l'affichage */
  Graphe *resultat;
  double residu;
  boolean affiche_derivees;
  Graphe *derivee_premiere, *derivee_seconde;

  /* Infos pour la post-analyse */
  unsigned int type_chaine;
};

/* -------- Les textes [isolés pour traduction] -------- */

#include "int_xan_matrice.txt"

/* -------- Fonctions locales -------- */

static struct infos_fond_continu *I_XAN_MUM_CreerInfos     (Graphe *brut);
static boolean                    I_XAN_MUM_GrapheConvient (Graphe *brut);
static Fenetre                   *I_XAN_MUM_RappelerFenetre(Graphe *brut);

static void I_XAN_MUM_ChangerMethode  (unsigned int numero_arbre, unsigned int numero_objet, unsigned long option_choisie);
static int  I_XAN_MUM_DemanderLesE    (Fenetre *fenetre, Graphe *brut, struct infos_fond_continu *infos);
static int  I_XAN_MUM_GererArbre      (unsigned int numero_objet, int x, int y);
static void I_XAN_MUM_InstallerMethode(struct infos_fond_continu *infos, unsigned int nv_methode);
static void I_XAN_MUM_LireParametres  (struct infos_fond_continu *infos);
static int  I_XAN_MUM_PreparerArbre   (unsigned int numero_arbre);

static void I_XAN_MUM_CalculDerivees(Graphe *brut, struct infos_fond_continu *infos);
static int  I_XAN_MUM_LibererFenetre(Fenetre *fenetre);
static void I_XAN_MUM_MettreAJour   (Fenetre *fenetre, Graphe *brut, struct infos_fond_continu *infos);

static void I_XAN_MUM_DessinDerivee(Fenetre *fenetre, int x_HG, int y_HG, unsigned int larg, unsigned int haut,
				    Graphe *graphe, struct infos_fond_continu *infos, boolean derivee_seconde);
static void I_XAN_MUM_DessinTravail(Fenetre *fenetre, int x_HG, int y_HG, unsigned int larg, unsigned int haut,
				    Graphe *graphe, struct infos_fond_continu *infos);

static int  CLC_matrice(Fenetre *fenetre, int bouton, int n_clic, int x, int y, int etat);
static int  CLV_matrice(Fenetre *fenetre, unsigned int touche, int etat);
static int  MSG_matrice(Fenetre *fenetre, Message *message);
static void RCT_matrice(Fenetre *fenetre, int x, int y, int vx_x, int vx_y);
static void RDS_matrice(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Routine d'appel principale : ouverture de la fenêtre et initialisation -------- */

Graphe *I_XAN_OterFondContinu(Graphe *brut, unsigned int type_chaine)
{
  Fenetre *fenetre;
  struct infos_fond_continu *infos;

  /* Contrôle sur le graphe */
  if ( FAUX == I_XAN_MUM_GrapheConvient( brut ) ) {
    AVERTISSEMENT( "Le graphe %s ne convient pas !", ( brut != NULL ) ? ( brut->nom ) : "NULL" );
    return NULL;
  }

  /* Est-ce que l'arbre est déjà ouvert ? */
  if ( VRAI == ARB_EstOuvert( ELIM_MUMATRICE ) ) {
    fenetre = I_XAN_MUM_RappelerFenetre( brut );
    if ( NULL == fenetre ) {
      /* On ne souhaite pas mettre à jour la fenêtre avec le graphe indiqué -> terminé ! */
      return NULL;
    }
  }

  /* On crée les informations */
  infos = I_XAN_MUM_CreerInfos( brut );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer les informations nécessaires : calcul annulé\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On associe ces informations à l'arbre */
  ARB_MettreComplement( ELIM_MUMATRICE, infos );

  /* Si l'on est ici, c'est que la fenêtre est fermée : on l'ouvre donc */
  fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, titre_fenetre_mum, NULL,
				ELIM_MUMATRICE, BI_FEN_ARBRE_HAUT, I_XAN_MUM_PreparerArbre,
				RDS_matrice, CLV_matrice, CLC_matrice, MSG_matrice, RCT_matrice,
				( int (*)(Fenetre *) ) NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la fenêtre pour l'interface !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( infos->parametres.fond_continu ); infos->parametres.fond_continu = NULL;
    GRF_efface( infos->resultat, VRAI, VRAI ); infos->resultat = NULL;
    free( infos ); infos = NULL;
    return NULL; /* Rien à faire... */
  }
  fenetre->drapeaux.redessin    = VRAI;
  fenetre->msg_masque.fermeture = VRAI; /* Messages traités par MSG_matrice */

  /* Quel suite donner à l'extraction du fond continu ? */
  infos->type_chaine = type_chaine;

  /* On associe le graphe brut & les informations à la fenêtre */
  FEN_GRF_LierGraphe( fenetre, brut );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, infos );

  /* On calcule le modèle de fond continu et le résultat
     Après toute l'initialisation pour avoir accès à la méthode,...
  */
  I_XAN_MUM_MettreAJour( fenetre, brut, infos );
  OBJ_MettreNombre( ELIM_MUMATRICE, OFC_E1, infos->parametres.e_1, 0.0 );

  /* On ouvre la fenêtre */
  FEN_Ouvrir( fenetre );

  /* On indique quel est le graphe résultat */
  return infos->resultat;
}

/* -------- Contrôle lors de l'effacement d'un graphe -------- */

int I_XAN_ControleEffacement(Fenetre *fenetre, struct graphe *graphe_efface)
{
  struct infos_fond_continu *infos;
  Graphe *brut;

  if ( NULL == graphe_efface ) return 0;

  /* On prépare l'eventuel message de fermeture */

  /* Le graphe traité */
  brut = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  if ( brut == graphe_efface ) {
    /* Cela à revient à fermer la fenêtre en forçant l'annulation */
    FEN_MessageFermeture( fenetre, BI_FEN_FERMER_FORCER_ANNULE );
    return VRAI;
  }

  /* Les informations complémentaires */
  infos = (struct infos_fond_continu *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return FAUX;

  if ( graphe_efface == infos->resultat ) { /* Le spectre contenant le résultat */
    infos->resultat = NULL;
    /* Cela à revient à fermer la fenêtre en forçant l'annulation */
    FEN_MessageFermeture( fenetre, BI_FEN_FERMER_FORCER_ANNULE );
    return VRAI;
  }
  if ( graphe_efface == infos->parametres.mu_calc ) { /* µm théorique utilisé */
    /* Pas besoin de fermer, on remplace juste par une victorine 3 */
    I_XAN_MUM_ChangerMethode( ELIM_MUMATRICE, OFC_CHX_METHODE, MUM_VICTORINE_1 );
    I_XAN_MUM_MettreAJour   ( fenetre, brut, infos );

    infos->parametres.mu_calc = NULL;
  }
  if ( graphe_efface == infos->parametres.reference ) { /* Référence sur laquelle caler le tout */
    /* Pas besoin de fermer, on se contente de faire sauter la référence ! */
    infos->parametres.reference = NULL;
  }
  if ( graphe_efface == infos->parametres.fond_continu ) { /* Modèle calculé du fond continu */
    (void) fprintf(stderr, "%s,%s l.%u : Modèle calculé : graphe isolé mais demande d'effacement extérieure à la fermeture de la fenêtre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }

  /* Le graphe n'était pas impliqué */
  return FAUX;
}

/* -------- Créer et initialiser les informations associées à la fenêtre -------- */

static struct infos_fond_continu *I_XAN_MUM_CreerInfos(Graphe *brut)
{
  struct infos_fond_continu *infos;
  char *titre;

  /* On réserve la structure */
  infos = (struct infos_fond_continu *) malloc( (size_t) sizeof( struct infos_fond_continu ) );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : \n",
		   __FILE__, __FUNCTION__, __LINE__);
    ERR_Memoire( "Informations d'extraction --- Calcul annul"e_AIGU, sizeof( struct infos_fond_continu ) );
    return NULL;
  }

  /* On initialise */
  /* a. Les paramètres de l'ancien appel, s'il y en a [au début car certaines valeurs pourront être changées ensuite !] */
  I_XAN_MUM_LireParametres( infos );

  /* b. Autres paramètres */
  /* Méthode : on réutilise celle d'avant. Mais si l'arbre est ouvert pour la première fois, on récupère 0 car l'objet n'est pas encore converti en menu ! */
  infos->parametres.methode = OBJ_OptionMenu( ELIM_MUMATRICE, OFC_CHX_METHODE, 0, QUESTION );
  infos->parametres.coefficients = NULL;

  infos->parametres.fond_continu = NULL;
  infos->parametres.mu_calc      = NULL;
  infos->parametres.reference    = NULL;

  /* c. Valeurs uniquement utilisées pour l'affichage */
  infos->resultat = NULL;
  infos->derivee_premiere = NULL;
  infos->derivee_seconde = NULL;

  /* d. Le cas de E1 */
  if ( ( infos->parametres.e_1 >= brut->x_max ) || ( infos->parametres.e_1 <= brut->x_min ) ) {
    /* On ne peut pas garder la valeur précédemment indiquée ==> on en propose une */
    infos->parametres.e_1 = XAN_MUM_ProposerE1( brut );
  }

  /* On prépare le graphe [isolé] qui va contenir le modèle.
     Ce graphe a forcément les mêmes X que le brut : inutile de les stocker en double !
     Il est isolé ==> pas de légendes ni d'unités [qui seraient les mêmes que celles du brut] 
   */
  infos->parametres.fond_continu = GRF_nv_TableX( NULL, GR_MUMATRICE, GRF_SERIE, "Fond continu calculé", VRAI,
						  brut->x, brut->nombre_point, FAUX );
  if ( NULL == infos->parametres.fond_continu ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de préparer le fond à soustraire\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( infos ); infos = NULL;
    return NULL;
  }

  /* On prépare le graphe [en série] qui va contenir le résultat.
     Ce graphe a forcément les mêmes X que le brut, mais on les stocke isolément
               car on peut les changer sans nécessairement changer les originaux
     Il a les mêmes légendes et unités que le brut, en X et en Y
   */
  titre = TXT_genere_chaine( "%s - "MU"m (en cours)", brut->nom );
  infos->resultat = GRF_nv_TableX( brut, GR_MOINSMAT, GRF_FCT_REELLE, titre, FAUX,
				   brut->x, brut->nombre_point, VRAI );
  free( titre ); titre = NULL;
  if ( NULL == infos->resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de préparer le résultat\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( infos->parametres.fond_continu ); infos->parametres.fond_continu = NULL;
    free( infos ); infos = NULL;
    return NULL;
  }
  /* Les unités sont celles du graphe originel ! */
  GRF_ChangerUnite( infos->resultat, AXE_X, GRF_QuelleUnite( brut, AXE_X ) );
  GRF_ChangerUnite( infos->resultat, AXE_Y, GRF_QuelleUnite( brut, AXE_Y ) );

  /* Les légendes sont aussi celles du graphe originel, du moins en X ! */
/*   GRF_ChangerLegende( infos->resultat, AXE_X, GRF_QuelleLegende( brut, AXE_X ) ); */
/*   GRF_Changer( infos->extrait, AXE_Y, GRF_QuelleUnite( brut, AXE_Y ) ); */

  /* On installe la méthode de calcul du fond continu : initialise correctement les deux graphes ci-dessus ! */
  I_XAN_MUM_ChangerMethode( ELIM_MUMATRICE, OFC_CHX_METHODE, infos->parametres.methode );

  /* Terminé, tout s'est bien passé */
  return infos;
}

/* -------- Que faire si la fenêtre est déjà ouverte ? -------- */

static Fenetre *I_XAN_MUM_RappelerFenetre(Graphe *brut)
{
  Fenetre *fenetre;
  struct infos_fond_continu *infos __attribute__((__unused__));
  Graphe *graphe;
  int reponse;

  /* Quelle est la fenêtre associée à l'arbre ? */
  fenetre = ARB_QuelleFenetre( &(arbres[ELIM_MUMATRICE]), FAUX );

  /* Quelles sont les informations associées à cette fenêtre ? */
  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (struct infos_fond_continu *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  /* Est-ce que le graphe est le même ? */
  if ( graphe == brut ) {
    /* On travaille sur le même graphe : on se contente de faire repasser au premier plan */
    FEN_premierplan( fenetre );
    return NULL;
  }

  /* Est-ce que le nouveau graphe est utilisable ? */

  /* On pose la question : faut-il rappeler,... ? */
  reponse = BI_ERR_alerte( fenetre, &Q_ReactiverFondContinu, VRAI, graphe->nom, brut->nom );
  switch ( reponse )
  {
  case I_XAN_MUM_RAPPELER:
    FEN_premierplan( fenetre );
    return NULL;

  case I_XAN_MUM_RELANCER:
    /* On termine les calculs pour l'ancien graphe : ferme la fenêtre et libère tout */
    OPTION_A_FAIRE( "Libérer les anciennes infos, le graphe..." )

    /* On rappelera donc une nouvelle fenêtre propre... */
    return NULL;

  case I_XAN_MUM_ANNULER:
    /* On ne fait rien de spécial */
    return NULL;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en posant la question, ou réponse inconnue [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, reponse);
    return fenetre;
  }
}

/* -------- Renvoie VRAI si le graphe convient, FAUX s'il n'est pas légitime de lui ôter un fond continu -------- */

static boolean I_XAN_MUM_GrapheConvient(Graphe *brut)
{
  if ( NULL == brut ) return FAUX; /* Rien à faire ! */

  /* Le graphe doit être une fonction réelle : y=f(x) */
  if ( FAUX == GRF_EstFonctionReelle( brut ) ) {
    (void) BI_ERR_alerte( principale, &ERR_FonctionReelle, VRAI, brut->nom );
    return FAUX;
  }

  /* Pour l'instant, pas d'autre contrôle, mais certains types devraient être interdits */
  return VRAI;
}

/* -------- Préparer la fermeture de la fenêtre proprement -------- */

static int I_XAN_MUM_LibererFenetre(Fenetre *fenetre)
{
  Graphe *brut;
  struct infos_fond_continu *infos;

  if ( NULL == fenetre ) return -1;  /* Ne devrait jamais arriver, mais bon... */

  /* On récupère les infos associées à la fenêtre */
  brut  = FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  /* On dissocie les infos de la fenêtre */
  FEN_GRF_LierGraphe( fenetre, NULL );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );
  ARB_MettreComplement( ELIM_MUMATRICE, NULL );

  if ( brut != NULL ) {
    /* Le graphe n'est plus utilisé par cette fenêtre : on le déprotège */
    brut->drapeaux.protege = FAUX;
  }

  if ( NULL != infos ) {
    GRF_efface( infos->resultat, VRAI, VRAI ); infos->resultat = NULL;
    GRF_eff_isole( infos->parametres.fond_continu ); infos->parametres.fond_continu = NULL;
    free( infos->parametres.coefficients ); infos->parametres.coefficients = NULL;
    free( infos ); infos = NULL;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Gestion de la fenêtre ~-~-~-~-~-~-~-~-~ */

static void I_XAN_MUM_CalculDerivees(Graphe *brut __attribute__((__unused__)), struct infos_fond_continu *infos)
{
  /* On efface les anciennes dérivées */
  GRF_eff_isole( infos->derivee_premiere ); infos->derivee_premiere = NULL;
  GRF_eff_isole( infos->derivee_seconde ); infos->derivee_seconde = NULL;

  /* Calcul des dérivées : méthode la plus simple possible ! */
  infos->derivee_premiere = GRF_DeriveeSimple( infos->resultat );
  infos->derivee_seconde  = GRF_DeriveeSimple( infos->derivee_premiere );
}

/* -------- Mise à jour de la fenêtre après une modification -------- */

static void I_XAN_MUM_MettreAJour(Fenetre *fenetre, Graphe *brut, struct infos_fond_continu *infos)
{
  int retour;
  unsigned long ddl;

  if ( NULL == infos ) return;	/* Peut arriver quand on appelle à l'ouverture de la fenêtre */

  /* On calcule le modèle de fond continu */
  retour = XAN_MUM_CalculerFond( brut, &(infos->parametres) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans XAN_MUM_CalculerFond() : code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* On le soustrait du brut */
  retour = XAN_MUM_SoustraireFond( brut, infos->resultat, &(infos->parametres) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans XAN_MUM_SoustraireFond() : code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* On calcule le résidu */
  infos->residu = XAN_MUM_CalculerResidu( brut, infos->resultat, &(infos->parametres), &ddl );
  if ( ddl > 0 ) {
    /* On affiche l'incertitude estimée ["bruit" expérimental], comme écart-type */
    OBJ_MettreNombre( ELIM_MUMATRICE, OFC_RESIDU, sqrt( infos->residu / (double) ddl ), 0.0 );
  } else {
    /* On affiche une valeur absurde pour indiquer le problème */
    OBJ_MettreNombre( ELIM_MUMATRICE, OFC_RESIDU, -1.0, 0.0 );
  }

  /* On signale que le résultat a été modifié */
  GRF_modifier( infos->resultat, infos->resultat, FAUX );

  /* Si demandé, on recalcule les dérivées */
  if ( VRAI == infos->affiche_derivees ) {
    I_XAN_MUM_CalculDerivees( brut, infos );
  }

  /* Si la fenêtre est ouverte, on redessine */
  if ( fenetre != NULL ) {
    FEN_redessin( fenetre, BI_RDS_UTIL );
  }
}

/* -------------- Clic sur la fenêtre d'élimination de mu_matrice ------------------------ */

static int CLC_matrice(Fenetre *fenetre, int bouton, int n_clic, int etat __attribute__((__unused__)), int x_clic, int y_clic)
{
  Graphe *brut;
  double echelle_X, x_vrai;
  int retour;
  struct infos_fond_continu *infos;

#ifdef _TOS_ /* Version GEM : un clic passe en premier plan, car pas d'auto-focus... */
  int handle;

  wind_get(0, WF_TOP, &handle, &n_clic, &etat, &bouton);
  if (handle != fenetre->handle) {
    wind_set(fenetre->handle, WF_TOP);
    return VRAI;
  }
#endif
#ifdef _DEVELOPPE_
  if ( n_clic > 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Clic multiple, sans fonction\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return BI_EVENEMENT_NON_TRAITE;
  }
#endif

  if ( ( x_clic < 0 ) || ( x_clic > (int) (fenetre->w_larg) ) ||
       ( y_clic < 0 ) || ( y_clic > (int) (fenetre->w_haut) ) ) {
    return BI_EVENEMENT_TRAITE; /* Clic en-dehors de la zone de travail */
  }

  infos = FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  brut  = FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  if ( ( NULL == infos ) || ( NULL == brut ) ) {
    return BI_EVENEMENT_TRAITE; /* Clic sans rien dans la fenêtre */
  }

  /* Calcul de l'échelle en X [la seule utile !] */
  if ( VRAI == infos->affiche_derivees ) {
    int x_separateur;
    unsigned int lg_principale;

    /* On repère la région où les clics on un sens  */
    x_separateur  = fenetre->w_larg - (fenetre->w_haut / 2 );
    lg_principale = x_separateur;

    if ( x_clic >= x_separateur ) return BI_EVENEMENT_TRAITE; /* Le clic sur les dérivées n'a aucun intérêt */

    retour = TRC_Echelles( lg_principale, fenetre->w_haut, brut->x_min, brut->x_max, 0.0, 1.0, &echelle_X, NULL );
  } else {
    retour = TRC_Echelles( fenetre->w_larg, fenetre->w_haut, brut->x_min, brut->x_max, 0.0, 1.0, &echelle_X, NULL );
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans le calcul de l'échelle en X\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return BI_EVENEMENT_TRAITE; /* Clic sans rien dans la fenêtre */
  }

  /* On en déduit le X cliqué */
  x_vrai = TRC_ConvertitX_EvP( x_clic, brut->x_min, 0, echelle_X );

  if ( ( UN_POINT == infos->parametres.mode ) || ( 1 == bouton ) ) { /* Clic gauche, 2 points ou quelconque, 1 point */
    infos->parametres.e_1 = x_vrai;
    OBJ_MettreNombre( ELIM_MUMATRICE, OFC_E1, infos->parametres.e_1, 0.0 );
  }
  if ( ( 2 == bouton ) && ( DEUX_POINTS == infos->parametres.mode ) ) {
    infos->parametres.e_1p = x_vrai;
    OBJ_MettreNombre( ELIM_MUMATRICE, OFC_E1P, infos->parametres.e_1p, 0.0 );
  }

  I_XAN_MUM_MettreAJour( fenetre, brut, infos );

  return BI_EVENEMENT_TRAITE;
}

/* -------- Gestion du clavier pour cette fenêtre ------------\
 |  Renvoie VRAI si la touche a été reconnue comme raccourci  |
 |                                                            |
 | espace  = quitter le module                                |
 | 1       = demander la valeur de E1                         |
 | shift-1 = demander la valeur de E1', si elle existe        |
 | -, +    = modifier le degré du polynôme                    |
 | 2 ou d  = mode 1 point / 2 points                          |
 \----------------------------------------------------------- */

static int CLV_matrice(Fenetre *fenetre, unsigned int touche, int etat)
{
  int retour;
  struct infos_fond_continu *infos;
  unsigned long nv_methode;
  Graphe *brut;
  Message message;

  if ( etat & K_CTRL ) {
    /* Ctrl-touche : aucune combinaison -> renvoyée aux menus */
    return BI_EVENEMENT_NON_TRAITE;
  }

  brut  = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = (struct infos_fond_continu *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return BI_EVENEMENT_TRAITE;

  switch (touche)
  {
  case '1':
    (void) I_XAN_MUM_DemanderLesE( fenetre, brut, infos );
    return BI_EVENEMENT_TRAITE;

  case '-':
    if ( MUM_POLYNOME == infos->parametres.methode ) {
      retour = ARB_SimulerClic( ELIM_MUMATRICE, MUM_DEGRE_AVANT, 1 );
      if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    }
    return BI_EVENEMENT_TRAITE;

  case '+':
    if ( MUM_POLYNOME == infos->parametres.methode ) {
      retour = ARB_SimulerClic( ELIM_MUMATRICE, MUM_DEG_APRES, 1 );
      if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    }
    return BI_EVENEMENT_TRAITE;

  case '2':
  case 'd':
    retour = ARB_SimulerClic( ELIM_MUMATRICE, OFC_AVEC_2EPOINT, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;

  case ' ':
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas où != 0, fermeture forcée... */
    retour = (fenetre->message)(fenetre, &message);
    if ( retour != BI_EVENEMENT_NON_TRAITE ) return retour;
    return BI_EVENEMENT_TRAITE;

  case Fleche_Haut:		/* On remonte dans le menu : méthode PRÉCÉDENTE */
    nv_methode = infos->parametres.methode;
    if ( 0 == nv_methode ) {
      nv_methode = NBR_MUM_METHODES - 1;
    } else {
      nv_methode--;
    }
    OBJ_OptionMenu( ELIM_MUMATRICE, OFC_CHX_METHODE, nv_methode, METTRE );
    return BI_EVENEMENT_TRAITE;

  case Fleche_Bas:		/* On descend dans le menu : méthode SUIVANTE */
    /* Préparation de l'appel */
    nv_methode = infos->parametres.methode;

    /* Appel « clavier » du défileur */
    if ( ( NBR_MUM_METHODES - 1 ) <= nv_methode ) {
      nv_methode = 0;
    } else {
      nv_methode++;
    }

    /* Mémorisation des résultats de l'appel */
    OBJ_OptionMenu( ELIM_MUMATRICE, OFC_CHX_METHODE, nv_methode, METTRE );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE; /* Touche non traitée -> raccourci général ? */
}

/* -------- Gestion des messages, en particulier de fermeture -------- */

static int MSG_matrice(Fenetre *fenetre, Message *message)
{
  int retour, choix;
  Graphe *brut;
  struct infos_fond_continu *infos;

  if ( message->type != WM_CLOSED ) {
    AVERTISSEMENT( "Message non traité [%d]", message->type );
    return BI_EVENEMENT_NON_TRAITE;
  }

  /* Si on est là, on veut fermer la fenêtre : WM_CLOSED */
  if ( BI_FEN_FERMER_FORCER_ANNULE == message->fenetre.haut ) { /* On force la fermeture */
    choix = I_XAN_MUM_FIN_MODULE_FERMER;
  } else {
    choix = BI_ERR_alerte( fenetre, &Q_QuitterModule, VRAI );
  }

  if ( I_XAN_MUM_FIN_MODULE_ANNULER == choix ) {
    /* On annule la fermeture */
    return BI_EVENEMENT_TRAITE;
  } else if ( I_XAN_MUM_FIN_MODULE_FERMER == choix ) {
    /* On ferme en perdant les calculs */
    I_XAN_MUM_LibererFenetre( fenetre );
    FEN_Detruire( fenetre );
    return BI_EVENEMENT_FEN_FERMEE;
  }

  /* Là, on ferme en gardant les calculs */
  /* On récupère les informations */
  infos = FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  brut  = FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );

  /* On dissocie les informations, pour pouvoir effacer tranquillement */
  FEN_GRF_LierGraphe( fenetre, NULL );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );
  ARB_MettreComplement( ELIM_MUMATRICE, NULL );

  /* On ferme la fenêtre de toute façon */
  FEN_Detruire( fenetre );

  /* On déprotège les graphes impliqués */
  if ( infos->resultat != NULL) /* N'a pas été effacé... */
    infos->resultat->drapeaux.protege = FAUX;
  if ( brut != NULL) /* Le graphe existe encore : ce n'est pas un effacement qui cause la fermeture */
    brut->drapeaux.protege = FAUX;

  OPTION_A_FAIRE( "Mémorisation pour la reproduction" );
#ifdef ancien
  num_methode = infos->type; /* On mémorise pour l'appel suivant */
  OBJ_MettreNombre( EXAFS_GENERAL, EXF_VALEUR_E1 , infos->e_1, 0.0 );
  OBJ_choisit(EXAFS_GENERAL, EXF_MUM_DROITE, infos->type == MUM_DROITE);
  OBJ_choisit(EXAFS_GENERAL, EXF_MUM_VICT1 , infos->type == MUM_VICTORINE_1);
  OBJ_choisit(EXAFS_GENERAL, EXF_MUM_VICT2 , infos->type == MUM_VICTORINE_2);
#endif

  /* Travail sur le graphe sélectionné (déjà presque fait) */
  retour = XAN_MUM_PostSoustraction( brut, infos->resultat, &(infos->parametres) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans XAN_MUM_PostSoustraction() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return BI_EVENEMENT_FEN_FERMEE;
  }

  if ( I_XAN_MUM_FIN_MODULE_SERIE == choix ) { /* On traite la série complète de la même façon */
    Serie_Graphes *serie;
    Graphe **graphes = NULL;
    unsigned long i, nombre_graphes = 0;
    struct automatiser *lot;

    /* On récupère la série du graphe déjà traité */
    serie = GRF_QuelleSerie( brut ); /* Pas forcément la courante ! */
    if ( NULL == serie ) {
      AVERTISSEMENT( "Graphe %s hors série !", brut->nom );
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }

    /* On en extrait la liste des graphes qui conviennent : même type que le brut, même unité en X */
    graphes = GRF_ChoisirMasque( serie, brut, GRF_CHOIX_GENRE | GRF_CHOIX_TYPE, 0, &nombre_graphes );
    lot = AUTO_GRF_PreparerStructure( nombre_graphes, graphes );
    if ( NULL == lot ) {
      OPTION_A_FAIRE( "Fin propre" );
    }
    free( graphes ); graphes = NULL;

    /* Préparation de l'automatisation : fonction à utiliser,... */
    lot->classe_fonction = FONCTION_EST_PAR_TYPE;
    lot->genre = GRF_FCT_REELLE;
    lot->type = GR_BRUT;
    lot->numero_fonction = FCT_GRF_BRUT_MATRICE;

    /* Infos nécessaires à l'automatisation */
    lot->infos_necessaires = infos;
    lot->reference.graphe  = infos->resultat;

    /* On perd le résultat dans les infos, pour éviter des problèmes éventuels... */
    infos->resultat = NULL;

    /* On appelle le traitement par lots */
    retour = I_AUTOMATISE_SurListeConnue( lot, 0 );
    if ( retour != 0 ) {
      if ( retour < 0 ) {
	OPTION_A_FAIRE( "Message : erreur survenue !" );
      }

      OPTION_A_FAIRE( "Quitter proprement !" );
    }

    /* Doit-on insérer dans une nouvelle série ? */
    if ( lot->nom_serie != NULL ) {
      /* Oui : on la crée après la série en cours */
      serie = SERIE_GRF_CreerSerie( serie, lot->nom_serie );
      free( lot->nom_serie ); lot->nom_serie = NULL;
    } else {
      /* On reste dans la série en cours */
    }

    /* On insère tous les résultats dans la série indiquée */
    serie = I_GRF_InsererDansSerie( serie, brut, lot->resultats[ 0 ].graphe );
    for ( i = 1 ; i < lot->nombre_objets ; i++ ) {
      serie = I_GRF_InsererDansSerie( serie, lot->resultats[ i - 1 ].graphe, lot->resultats[ i ].graphe );
    }
  }

  /* On n'a plus besoin des graphes internes */
  GRF_eff_isole( infos->parametres.fond_continu ); infos->parametres.fond_continu = NULL;
  free( infos->parametres.coefficients ); infos->parametres.coefficients = NULL;

  /* On met le graphe résultat comme graphe courant */
  GRF_aller( infos->resultat );

  /* Veut-on chaîner avec d'autres opérations ? */
  if ( I_XAS_CHAINE_XANES == infos->type_chaine ) {
    /* On veut faire le traitement du XANES dans son ensemble */
    Graphe *derivee; /* Graphe renvoyé */

    /* Remarque : l'interpolation était demandée dans EDA.
       Elle n'a aucun intérêt dans LASE ==> cette étape est sautée [elle peut toujours être faite à la main si nécessaire !] */

    /* On demande la méthode pour calculer la dérivée première */
    derivee = I_GRF_Derivee( infos->resultat );
    if ( derivee != NULL ) {
      /* On demande la dérivée seconde */
      derivee = I_GRF_Derivee( infos->resultat );
    }
  }

  /* On n'a plus besoin des infos de traitement */
  free( infos ); infos = NULL;

  /* On a bien fermé la fenêtre : on le signale ! */
  return BI_EVENEMENT_FEN_FERMEE;
}

/* -------- Pointeur/réticule ------------- */

static void RCT_matrice(Fenetre *fenetre, int x, int y, int vx_x, int vx_y)
{
  int x_separateur;
  struct infos_fond_continu *infos;
  unsigned int lg_principale, lg_derivee __attribute__((__unused__)), ht_derivee;
  Graphe *graphe;

  /* On prépare pour le réticule */
  TRC_InitPointeur;

  /* On vérifie qu'il y a quelque chose à dessiner */
  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (struct infos_fond_continu *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( ( NULL == graphe ) || ( NULL == infos ) ) return;

  /* On définit les zones ; si pas d'affichage, dérivées hors écran : devrait être négligé ! */
  if ( VRAI == infos->affiche_derivees ) {
    x_separateur  = fenetre->w_larg - (fenetre->w_haut / 2 );
  } else {
    x_separateur  = fenetre->w_larg;
  }
  lg_principale = x_separateur;
  lg_derivee    = fenetre->w_larg - x_separateur + 1;
  ht_derivee    = fenetre->w_haut / 2;

  /* Doit-on effacer l'ancien réticule ? */
  if ( ( vx_x >= 0 ) && ( vx_x < (int) fenetre->w_larg ) && ( vx_y >= 0 ) && ( vx_y < (int) fenetre->w_haut ) ) {
    if ( vx_x < x_separateur ) {
      TRC_LigneV( fenetre, vx_x, 0, fenetre->w_haut );
    } else {
      if ( vx_y < (int) ht_derivee ) {
	TRC_LigneV( fenetre, vx_x, 0, ht_derivee );
      } else {
	TRC_LigneV( fenetre, vx_x, ht_derivee, fenetre->w_haut - ht_derivee );
      }
    }
  }

  /* Doit-on placer le nouveau réticule */
  if ( ( x >= 0 ) && ( x < (int) fenetre->w_larg ) && ( y >= 0 ) && ( y < (int) fenetre->w_haut ) ) {
    double X_reel, echelle_X;
    int retour, x_debut;
    char *infos;
    const char *texte_infos = NULL;

    if ( x < x_separateur ) {
      TRC_LigneV( fenetre, x, 0, fenetre->w_haut );

      /* On calcule l'échelle en X */
      x_debut = 0;
      texte_infos = infos_zone_principale;
      retour = TRC_Echelles( lg_principale, fenetre->w_haut, graphe->x_min, graphe->x_max, 0.0, 1.0, &echelle_X, NULL );
    } else {
      if ( y < (int) ht_derivee ) {
	TRC_LigneV( fenetre, x, 0, ht_derivee );
	texte_infos = infos_zone_derivee1;
      } else {
	TRC_LigneV( fenetre, x, ht_derivee, fenetre->w_haut - ht_derivee );
	texte_infos = infos_zone_derivee2;
      }

      x_debut = x_separateur;
      retour  = TRC_Echelles( lg_principale, ht_derivee, graphe->x_min, graphe->x_max, 0.0, 1.0, &echelle_X, NULL );
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur dans le calcul de l'échelle [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return;
    }

    /* On calcule la valeur de X réelle */
    X_reel = TRC_ConvertitX_EvP( x, graphe->x_min, x_debut, echelle_X );

    /* On met les informations */
    infos = TXT_genere_chaine( texte_infos, X_reel, GRF_QuelleUnite( graphe, AXE_X ) );
    FEN_infos( fenetre, infos );
    free( infos ); infos = NULL;
  }
}

/* -------- Redessin de la fenêtre -------- */

static void RDS_matrice(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Graphe *graphe;
  struct infos_fond_continu *infos;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut);

  /* On vérifie qu'il y a quelque chose à dessiner */
  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (struct infos_fond_continu *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == graphe ) {
    TRC_Message( fenetre, ERR_aucun_graphe );
    return;
  }

  if ( VRAI == infos->affiche_derivees ) {
    int x_separateur;
    unsigned int lg_principale, lg_derivee;

    /* On s'arrange pour que les dérivées occupent des carrés à droite de la fenêtre */
    x_separateur  = fenetre->w_larg - (fenetre->w_haut / 2 );
    lg_principale = x_separateur;
    lg_derivee    = fenetre->w_larg - x_separateur + 1;

    /* On place donc les séparateurs */
    TRC_Separateur( fenetre, VERTICAL, x_separateur, 0, fenetre->w_haut );
    TRC_Separateur( fenetre, HORIZONTAL, x_separateur, fenetre->w_haut / 2, lg_derivee );

    /* On dessine la partie principale */
    I_XAN_MUM_DessinTravail( fenetre, 0, 0, lg_principale, fenetre->w_haut, graphe, infos );

    /* On dessine la dérivée première [en haut] */
    I_XAN_MUM_DessinDerivee( fenetre, x_separateur, 0, lg_derivee, fenetre->w_haut / 2, graphe, infos, FAUX );

    /* On dessine la dérivée seconde [en bas] */
    I_XAN_MUM_DessinDerivee( fenetre, x_separateur, fenetre->w_haut / 2, lg_derivee, fenetre->w_haut - (fenetre->w_haut / 2), graphe, infos, VRAI );
  } else {
    /* On occupe tout l'espace disponible */
    I_XAN_MUM_DessinTravail( fenetre, 0, 0, fenetre->w_larg, fenetre->w_haut, graphe, infos );
  }
}

/* -------- Dessin de la dérivée, première ou seconde -------- */

static void I_XAN_MUM_DessinDerivee(Fenetre *fenetre, int x_HG, int y_HG, unsigned int largeur, unsigned int hauteur,
				    Graphe *graphe, struct infos_fond_continu *infos, boolean derivee_seconde)
{
  Graphe *derivee;
  InfosRegion2D region;
  double echelle_X, echelle_Y;
  int retour, x_barre;

  /* Quel est le graphe pertinent ? */
  if ( VRAI == derivee_seconde ) {
    derivee = infos->derivee_seconde;
  } else {
    derivee = infos->derivee_premiere;
  }

  /* On initialise la région de redessin */
  retour = TRC_Z2D_Initialiser( &region, fenetre, derivee );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant initialiser la zone de dessin\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* On adapte ses coordonnées pour la région choisie */
  region.x_BG = x_HG;
  region.y_BG = y_HG + hauteur - 1;
  region.largeur = largeur;
  region.hauteur = hauteur;

  /* Et on s'assure d'avoir des X corrects à tous les coups */
  region.axe_X.min = graphe->x_min;
  region.axe_X.max = graphe->x_max;

  if ( NULL == derivee ) {
    TRC_Couleur( ROUGE );
    TRC_Z_Message( fenetre, x_HG, y_HG, largeur, hauteur, ERR_derivee_nulle );

    /* On aura besoin de l'échelle en X quand même */
  } else {
    int y_axe;

    /* On trace le graphe dans la région 2D et on récupère les échelles au passage */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, derivee, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &echelle_X, &echelle_Y );

    /* On place l'axe des Y */
    TRC_Couleur( VIOLET );
    TRC_Style( BI_CONTINU );
    y_axe = TRC_ConvertitY_PvE( 0.0, region.axe_Y.min, region.y_BG, echelle_Y );
    TRC_LigneH( fenetre, x_HG, y_axe, largeur );
  }

  /* On met la barre rouge associée à E1 */
  TRC_Couleur( ROUGE );
  TRC_Style( BI_CONTINU );
  x_barre = TRC_ConvertitX_PvE( infos->parametres.e_1, graphe->x_min, x_HG, echelle_X );
  TRC_LigneV( fenetre, x_barre, y_HG, hauteur );
}

/* -------- Dessin de la partie utile : graphe source, résultat ; clics ; ... -------- */

static void I_XAN_MUM_DessinTravail(Fenetre *fenetre, int x_HG, int y_HG, unsigned int largeur, unsigned int hauteur,
				    Graphe *graphe, struct infos_fond_continu *infos)
{
  InfosRegion2D region;
  double x_echelle, y_echelle;
  int retour, x_barre;

  /* On initialise la région de redessin */
  retour = TRC_Z2D_Initialiser( &region, fenetre, graphe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant initialiser la zone de dessin\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* On adapte ses coordonnées pour la région choisie */
  region.x_BG = x_HG;
  region.y_BG = y_HG + hauteur - 1;
  region.largeur = largeur;
  region.hauteur = hauteur;

  /* On adapte la partie verticale en fonction du modèle */
  if ( infos != NULL ) {
    if ( infos->parametres.fond_continu != NULL ) {
      if ( infos->parametres.fond_continu->y_min < region.axe_Y.min ) region.axe_Y.min = infos->parametres.fond_continu->y_min;
      if ( infos->parametres.fond_continu->y_max > region.axe_Y.max ) region.axe_Y.max = infos->parametres.fond_continu->y_max;
    }
  }

  /* On trace le graphe dans la région 2D et on récupère les échelles au passage */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &x_echelle, &y_echelle );

  /* Reste-t-il des choses à tracer */
  if ( NULL == infos ) {
    TRC_Message( fenetre, ERR_pas_d_informations );
    return;
  }

  /* On trace la courbe à soustraire - en bleu */
  if ( infos->parametres.fond_continu != NULL ) {
    /* On recalcule les échelles pour rien, mais pas grave ! */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->parametres.fond_continu, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &x_echelle, &y_echelle );
  } else {
    TRC_Message( fenetre, ERR_pas_de_modele );
  }

  /* On met la barre rouge associée à E1 */
  TRC_Couleur( ROUGE );
  x_barre = TRC_ConvertitX_PvE( infos->parametres.e_1, graphe->x_min, 0, x_echelle );
  TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );

  /* Si nécessaire, on met la barre verte associée à E1' */
  if ( infos->parametres.mode == DEUX_POINTS ) {
    TRC_Couleur( VERT );
    x_barre = TRC_ConvertitX_PvE( infos->parametres.e_1p, graphe->x_min, 0, x_echelle );
    TRC_LigneV( fenetre, x_barre, 0, hauteur );
  }

  /* On trace le résultat obtenu - en gris foncé */
  if ( ( NULL == infos->resultat ) || ( infos->resultat->y_max == infos->resultat->y_min ) ) {
    TRC_Message( fenetre, ERR_pas_extrait );
  } else {
    int y_barre;

    /* L'échelle en X n'a pas changé, mais celle en Y est différente ! */
    region.axe_Y.min = infos->resultat->y_min;
    region.axe_Y.max = infos->resultat->y_max;

    TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->resultat, GRIS_FONCE, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &x_echelle, &y_echelle );

    /* L'axe Y = 0, pour vérifier que le fond a bien été modélisé */
    TRC_Couleur( VIOLET );
    TRC_Style( BI_CONTINU );
    y_barre = TRC_ConvertitY_PvE( 0.0, region.axe_Y.min, region.y_BG, y_echelle );
    TRC_LigneH( fenetre, 0, y_barre, largeur );
  }
}

/* ~-~-~-~-~-~-~-~-~ Gestion de l'arbre ~-~-~-~-~-~-~-~-~ */

/* -------- Changer la méthode de modélisation du fond continu -------- */

static void I_XAN_MUM_ChangerMethode(unsigned int numero_arbre, unsigned int numero_objet __attribute__((__unused__)), unsigned long option_choisie)
{
  Fenetre *fenetre;
  Graphe *brut;
  struct infos_fond_continu *infos;

  /* On récupère les informations de l'arbre & de la fenêtre associée */
  fenetre = ARB_QuelleFenetre( arbres + numero_arbre, FAUX );
  brut  = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos = (struct infos_fond_continu *) ARB_LireComplement( numero_arbre );

  /* On installe la nouvelle méthode */
  I_XAN_MUM_InstallerMethode( infos, option_choisie );

  /* On force le recalcul & le redessin */
  I_XAN_MUM_MettreAJour( fenetre, brut, infos );

  /* Terminé. Inutile de mémoriser dans l'arbre : la routine appelante s'en charge ! */
}

/* -------- Demander les valeurs de E1 et E1' (ou E1 seul) -------- */

static int I_XAN_MUM_DemanderLesE(Fenetre *fenetre, Graphe *brut, struct infos_fond_continu *infos)
{
  struct infos_demande nv_valeurs[ 2 ] = {
    { BI_CHAMP_DEMANDE_REEL, "E1   =", NULL, BI_EDITABLE_MIN | BI_EDITABLE_MAX, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, "E1'  =", NULL, BI_EDITABLE_MIN | BI_EDITABLE_MAX, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } }
  };
  int reponse;
  unsigned int i;
  const char *unite_X;

  /* Cela a-t-il un sens de demander les valeurs ? */
  if ( NULL == brut ) return -1;

  /* Quelle serait l'unité en X ? */
  unite_X = GRF_QuelleUnite( brut, AXE_X );

  /* Initialisation des valeurs */
  for ( i = 0 ; i < 2 ; i++ ) {
    /* Les bornes */
    nv_valeurs[ i ].min = brut->x_min;
    nv_valeurs[ i ].max = brut->x_max;

    /* L'unité nécessaire */
    if ( unite_X != NULL ) {
      nv_valeurs[ i ].proprietes |= BI_EDITABLE_UNITE;
      nv_valeurs[ i ].unite = unite_X;
    }
  }
  nv_valeurs[ 0 ].valeur_initiale.reel = infos->parametres.e_1;
  nv_valeurs[ 1 ].valeur_initiale.reel = infos->parametres.e_1p;

  OPTION_A_FAIRE( "Proposer une valeur de E1 plausible par défaut" );
  nv_valeurs[ 1 ].valeur_defaut.reel = brut->x_min;

  /* On demande les nouvelles valeurs à l'utilisateur */
  if ( DEUX_POINTS == infos->parametres.mode ) {
    reponse = BI_DemandeMultiple( fenetre, &Q_ValeursE1, 2, nv_valeurs );
    if ( reponse != (int) Q_ValeursE1.bouton_ok ) {
      return +1;		/* On annule */
    }

    infos->parametres.e_1p = nv_valeurs[ 1 ].valeur_donnee.reel;
    OBJ_MettreNombre( ELIM_MUMATRICE, OFC_E1P, infos->parametres.e_1p, 0.0 );
  } else {
    reponse = BI_DemandeValeur( fenetre, &Q_ValeurE1, nv_valeurs );
    if ( reponse != (int) Q_ValeurE1.bouton_ok ) {
      return +1;		/* On annule */
    }
  }

  /* On mémorise les valeurs */
  infos->parametres.e_1 = nv_valeurs[ 0 ].valeur_donnee.reel;
  OBJ_MettreNombre( ELIM_MUMATRICE, OFC_E1, infos->parametres.e_1, 0.0 );

  /* On met à jour les dessins */
  I_XAN_MUM_MettreAJour( fenetre, brut, infos );

  /* Terminé sans souci */
  return 0;
}

/* -------- Gérer l'arbre -------- */

static int I_XAN_MUM_GererArbre(unsigned int numero_objet, int x_souris, int y_souris)
{
  Graphe *brut;
  Fenetre *fenetre;
  int nv_etat;
  struct infos_fond_continu *infos;

  /* On récupère les informations */
  fenetre = ARB_QuelleFenetre( arbres + ELIM_MUMATRICE, FAUX ); /* Fenêtre actuelle */
  infos = (struct infos_fond_continu *) ARB_LireComplement( ELIM_MUMATRICE );
  brut  = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );

  if ( ( NULL == brut ) || ( NULL == infos ) ) return BI_EVENEMENT_TRAITE;

  switch ( numero_objet )
  {
  case OFC_TXT_E1:
  case OFC_E1:
  case OFC_TXT_EP1:
  case OFC_E1P:
    (void) I_XAN_MUM_DemanderLesE( fenetre, brut, infos );
    return BI_EVENEMENT_TRAITE;

  case OFC_TXT_2EPOINT:
  case OFC_AVEC_2EPOINT: /* Travail avec 1 ou 2 points de contrôle ? */
    nv_etat = OBJ_choisit( ELIM_MUMATRICE, OFC_AVEC_2EPOINT, ECHANGE );
    OBJ_active( ELIM_MUMATRICE, OFC_E1P, nv_etat );
    if ( nv_etat ) { /* On travaille avec deux points */
      infos->parametres.mode = DEUX_POINTS;
    } else {
      infos->parametres.mode = UN_POINT;
    }

    I_XAN_MUM_MettreAJour( fenetre, brut, infos );
    return BI_EVENEMENT_TRAITE;		

  case MUM_DEGRE_AVANT: /* Choix du degré du polynôme */
  case MUM_DEGRE_FOND:
  case MUM_DEGRE:
  case MUM_DEG_APRES:
    infos->parametres.degre = gere_ascenseur( ELIM_MUMATRICE, MUM_DEGRE_AVANT, MUM_DEGRE_FOND, MUM_DEGRE, MUM_DEG_APRES, ASC_HORIZONTAL,
					      1, MUM_DEGRE_MAXIMAL, 1, infos->parametres.degre, numero_objet, x_souris, y_souris );

    I_XAN_MUM_MettreAJour( fenetre, brut, infos );
    return BI_EVENEMENT_TRAITE;

  case OFC_AVEC_DERIVEE:
    nv_etat = OBJ_choisit( ELIM_MUMATRICE, OFC_AVEC_DERIVEE, ECHANGE );
    if ( VRAI == nv_etat ) {
      infos->affiche_derivees = VRAI;

      /* On calcule les dérivées par la méthode la plus simple possible */
      I_XAN_MUM_CalculDerivees( brut, infos );
    } else {
      /* On annule l'affichage des dérivées : on peut les effacer */
      infos->affiche_derivees = FAUX;
      GRF_eff_isole( infos->derivee_premiere ); infos->derivee_premiere = NULL;
      GRF_eff_isole( infos->derivee_seconde ); infos->derivee_seconde = NULL;
    }

    /* On redessine [et cela suffit !] */
    FEN_redessin( fenetre, BI_RDS_UTIL );

    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Lire les paramètres [du moins ceux stockés directement dans l'arbre !] -------- */

static void I_XAN_MUM_LireParametres(struct infos_fond_continu *infos)
{
  /* Combien de bornes ? */
  if ( OBJ_choisit( ELIM_MUMATRICE, OFC_AVEC_2EPOINT, QUESTION ) ) {
    infos->parametres.mode = DEUX_POINTS;
  } else {
    infos->parametres.mode = UN_POINT;
  }

  /* Les deux bornes */
  infos->parametres.e_1  = OBJ_LireNombre( ELIM_MUMATRICE, OFC_E1, NULL );
  infos->parametres.e_1p = OBJ_LireNombre( ELIM_MUMATRICE, OFC_E1P, NULL );

  /* Le degré du polynôme */
  infos->parametres.degre = OBJ_LireEntier_U( ELIM_MUMATRICE, MUM_DEGRE );

  /* Options d'affichage */
  if ( OBJ_choisit( ELIM_MUMATRICE, OFC_AVEC_DERIVEE, QUESTION ) ) {
    infos->affiche_derivees = VRAI;
  } else {
    infos->affiche_derivees = FAUX;
  }
}

/* -------- Changer la méthode d'estimation du fond continu -------- */

static void I_XAN_MUM_InstallerMethode(struct infos_fond_continu *infos, unsigned int nv_methode)
{
  /* On initialise l'arbre avec les valeurs de la méthode */
  /* a. Le sélecteur de méthodes */
  /* On ne le fait pas ici, sinon appel par I_XAN_MUM_ChangerMethode() -> boucle sans fin */

  /* b. Les options qui sont actives en fonction de cette méthode */
  OBJ_active( ELIM_MUMATRICE, MUM_DEGRE_AVANT, nv_methode == MUM_POLYNOME );
  OBJ_active( ELIM_MUMATRICE, MUM_DEGRE_FOND , nv_methode == MUM_POLYNOME );
  OBJ_active( ELIM_MUMATRICE, MUM_DEGRE      , nv_methode == MUM_POLYNOME );
  OBJ_active( ELIM_MUMATRICE, MUM_DEG_APRES  , nv_methode == MUM_POLYNOME );

  /* On mémorise la nouvelle méthode */
  if ( infos != NULL ) {
    /* Peut être NULL quand on appelle à l'ouverture de la fenêtre */
    infos->parametres.methode = nv_methode;
  }
}

/* -------- Préparation de l'arbre dans la fenêtre -------- */

static int I_XAN_MUM_PreparerArbre(unsigned int numero_arbre)
{
  int retour;

  /* On transforme en sous-menu l'appel de la courbe modèle */
  OBJ_ChangerType( numero_arbre, OFC_CHX_METHODE, G_SSMENU );
  OBJ_MettreMenu ( numero_arbre, OFC_CHX_METHODE, NBR_MUM_METHODES, Meth_mu_matrice, NULL, I_XAN_MUM_ChangerMethode );

  /* On installe la méthode par défaut */
  OBJ_OptionMenu( numero_arbre, OFC_CHX_METHODE, MUM_VICTORINE_1, METTRE );

  retour = ARB_Initialiser( numero_arbre, "menu_absorption.html#matrice", I_XAN_MUM_GererArbre, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en initialisant l'arbre [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -1;
  }

  /* Améliorations esthétiques */
  OBJ_agrandit( numero_arbre, ROOT, 0, +2 );

  ARB_EnjoliverAscenseur( numero_arbre, MUM_DEGRE_AVANT, MUM_DEGRE_FOND, MUM_DEGRE, MUM_DEG_APRES, ASC_HORIZONTAL, 0 );

  OBJ_deplace( numero_arbre, OFC_E1P        , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, OFC_AVEC_ERREUR, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, MUM_MU_0       , 0, +2, DEPLACE_PIXELS );

#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( numero_arbre, EMM_COURBE, "Method:" );
#endif

#ifndef _DEVELOPPE_
/*   OBJ_active( numero_arbre, MUM_AVEC_ERREUR, FORCE_NON ); */
#endif

  arbres[ numero_arbre ].drapeaux.prepare = VRAI;
  return 0;
}
