/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Recherche du modle adquat pour ajuster un spectre
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "atomes.h"
#include "feff.h" /* Pour les modles feff */
#include "graphes.h"
#include "modeles.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define MODELE_INADAPTE -1
#define MODELE_EXACT     0
#define MODELE_VIDE      1 /* et adapt au type de graphe */
#define MODELE_ADAPTE    2 /* mais un autre graphe est dj not comme modle */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int modele_convient(struct modele_ajuste *modele, Graphe *graphe);

static struct modele_ajuste *modele_dans_absorbeur(Absorbeur *absorbeur, Graphe *graphe, struct modele_ajuste **premier, struct modele_ajuste **vide);
static struct modele_ajuste *modele_dans_molecule (Molecule *molecule, Graphe *graphe, struct modele_ajuste **premier, struct modele_ajuste **vide);

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

struct modele_ajuste *MODELE_ChercherModele(struct graphe *graphe)
{
  /* exact   : modle avec le graphe indiqu comme graphe exprimental
     vide    : premier modle sans graphe exprimental adapt au graphe pass
     premier : premier modle adapt au graphe pass */
  struct modele_ajuste *exact = NULL, *vide = NULL, *premier = NULL;
  Molecule *molecule;

  if (NULL == graphe) return NULL;

  /* On cherche dans les molcules [les absorbeurs, en fait] */
  if (NULL != mol_courante) {
    /* On commence par la molcule courante */
    exact = modele_dans_molecule(mol_courante, graphe, &premier, &vide);
    if (exact != NULL)
      return exact;

    /* Si pas trouv, on essaye dans les autres molcules */
    for (molecule = premiere_molecule ; molecule != NULL ; molecule = molecule->suivante ) {
      if ( mol_courante == molecule ) /* Dj contrle */
	continue;

      exact = modele_dans_molecule(molecule, graphe, &premier, &vide);
      if (exact != NULL)
	return exact;
    }
  }

  /* Si on est l, c'est qu'un modle exact n'a pas t trouv */
  if ( vide != NULL ) { /* On donne prsance aux modles vides */
    vide->experience = graphe; /* Comme cela, il ne sera plus vide ! */
    return vide;
  }

  /* Rien trouv dans les molcules : on cherche donc dans les modles gnriques */

  return premier;
}

/* -------- Recherche dans les modles de la molcule indique -------- */

static struct modele_ajuste *modele_dans_molecule(Molecule *molecule, Graphe *graphe, struct modele_ajuste **premier, struct modele_ajuste **vide)
{
  struct modele_ajuste *exact;
  Absorbeur *absorbeur;
  unsigned int i;

  if (0 == molecule->nbr_absorbeurs) /* Aucun absorbeur = aucun modle ! */
    return NULL;

  /* On regarde d'abord l'absorbeur actif */
  absorbeur = molecule->absorbeurs + molecule->absorbeur_courant;
  exact = modele_dans_absorbeur(absorbeur, graphe, premier, vide);
  if (NULL != exact) /* On a trouv pile celui qu'il fallait */
    return exact;

  /* On contrle ensuite dans les autres absorbeurs */
  for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
    if (i == molecule->absorbeur_courant) /* on l'a test juste avant */
      continue;

    exact = modele_dans_absorbeur(molecule->absorbeurs + i, graphe, premier, vide);
    if (NULL != exact) /* On a trouv pile celui qu'il fallait */
      return exact;
  }

  return NULL; /* Non trouv en exact dans la molcule... */
}

/* -------- Chercher dasn les modles d'un atome absorbeur -------- */

static struct modele_ajuste *modele_dans_absorbeur(Absorbeur *absorbeur, Graphe *graphe, struct modele_ajuste **premier, struct modele_ajuste **vide)
{
  unsigned int i;
  int retour;
  struct modele_ajuste *modele;

  if (0 == absorbeur->nbr_modeles) return NULL;

  /* On commence par le modle actif */
  retour = modele_convient(absorbeur->modele_actif, graphe);
  if (MODELE_EXACT == retour)
    return absorbeur->modele_actif;
  else if (MODELE_VIDE == retour) {
    if ( (vide != NULL) && (NULL == *vide) ) /* Si non-NULL, dj un modle vide trouv,
						plus proche de celui attendu par l'utilisateur */
      *vide = absorbeur->modele_actif;
  } else if (MODELE_ADAPTE == retour) {
    if ( (premier != NULL) && (NULL == *premier) ) /* Si non-NULL, dj un modle adapt trouv,
						      plus proche de celui attendu par l'utilisateur */
      *premier = absorbeur->modele_actif;
  }

  /* On regarde les autres modles - si par hasard il y a un exact... */
  for (i = 0, modele = absorbeur->modeles ; i < absorbeur->nbr_modeles ; i++) {
    if (modele == absorbeur->modele_actif) /* Dj contrl */
      continue;

    retour = modele_convient(modele, graphe);
    if (MODELE_EXACT == retour)
      return modele;
    else if (MODELE_VIDE == retour) {
      if ( (vide != NULL) && (NULL == *vide) ) /* Mme remarque qu'avant */
	*vide = modele;
    } else if (MODELE_ADAPTE == retour) {
      if ( (premier != NULL) && (NULL == *premier) ) /* Mme remarque qu'avant */
	*premier = modele;
    }
  }

  /* Si on est l, c'est que l'on n'a pas trouv de modle exact */
  return NULL;
}

/* -------- Est-ce que le modle indiqu convient pour le graphe indiqu ? -------- */

static int modele_convient(struct modele_ajuste *modele, Graphe *graphe)
{
  boolean convient = FAUX;

  if ( modele->experience == graphe ) /* Il est marqu comme tel ! */
    return MODELE_EXACT;

  if ( graphe->type != GR_SSTYPE ) {
    switch (modele->type)
    {
    case MODELE_EXAFS_CHEMINS:
      if ( (GR_TF == graphe->type) || (GR_EXAFS == graphe->type) || /* EXAFS doit tre normalis ! */
	   ( (GR_AJUSTE == graphe->type) && (VRAI == graphe->drapeaux.exafs) ) /* Modle d'un ajustement prcdent */
	   )
	convient = VRAI; /* Ajustable par ce type de modle */
      else if (VRAI == graphe->drapeaux.exafs) {
	printf("modele_convient : attention, graphe EXAFS non ajustable (non normalis ?) : %s\n", graphe->nom);
      }
      break;

    case MODELE_GENERIQUE:
      convient = VRAI;		/* un modle gnrique peut ajuster n'importe-quoi par dfinition ! */
      break;

    default:
      fprintf(stderr, "modele_convient : type de modle pas encore gr = %d\n", modele->type);
      break;
    }
  } else { /* Les graphes sans type sont ajustables par n'importe quel modle ! */
    convient = VRAI;
  }

  if (FAUX == convient)
    return MODELE_INADAPTE;
  else {
    if (NULL == modele->experience)
      return MODELE_VIDE;
    else
      return MODELE_ADAPTE;
  }
}

