/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Calcul d'une matrice de covariance, intra-modle
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "couches.h"
#include "feff.h"
#include "graphes.h"
#include "matrices.h"
#include "modeles.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int calc_mat_simple(double *matrice, unsigned long nbr_elements, struct modele_ajuste *modele);

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Cration de la matrice de covariance associe au modle -------- */

Matrice *MODELE_CreerMatriceCovariance(struct modele_ajuste *modele)
{
  if (NULL == modele) return NULL;
  if ( NULL != modele->covariance ) {
    /* On efface l'ancienne */
    effacer_matrice( modele->covariance );
    modele->covariance = NULL;
  }

  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) fprintf(stderr, "MODELE_CreerMatriceCovariance : ajustement en cours sur le modle !\n");
    return NULL;
  }

  if ( 2 < modele->nbr_variables ) {
    (void) fprintf(stderr, "MODELE_CreerMatriceCovariance : trop peu de variables ajustes !\n");
    return NULL;
  }

  modele->covariance = creer_matrice(modele->nbr_variables, modele->nbr_variables, MATRICE_SYMETRIQUE);
  if (NULL == modele->covariance) {
    (void) fprintf(stderr, "MODELE_CreerMatriceCovariance : impossible de crer la matrice de covariance !\n");
    return NULL;
  }

  /* Tout s'est bien pass */
  return modele->covariance;
}

/* -------- Interface gnrale, vue de l'extrieur -------- */

Matrice *MODELE_MatriceCovariance(struct modele_ajuste *modele)
{
  Matrice *covariance, *hessienne;
  int retour;
  unsigned int nbr_variables;

  if (NULL == modele) return NULL;
  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : ajustement en cours sur ce modle !\n");
    return NULL;
  }

  if ( NULL == modele->reference ) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : pas de rfrence exprimentale !\n");
    return NULL;
  }
  if ( NULL == modele->reference->erreur_y ) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : pas de barres d'erreur sur la rfrence disponibles !\n");
    return NULL;
  }

  if ( NULL == modele->modele ) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : pas de spectre ajust calcul !\n");
    return NULL;
  }

  nbr_variables = modele->nbr_variables;
  if (nbr_variables < 2) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : trop peu de variables pour calculer une matrice de covariance !\n");
    return NULL;
  }

  /* On initialise le modle */
  retour = MODELE_Initialiser( modele );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : impossible d'initialiser le modle (code %d).\n", retour);
    return NULL;
  }

  /* On initialise les variables du modle */
  retour = MODELE_InitVariables( modele );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : ERREUR dans MODELE_InitVariables (code %d)\n", retour);
    return NULL;
  }

  hessienne = creer_matrice(nbr_variables, nbr_variables, MATRICE_SYMETRIQUE);
  if (hessienne == NULL) {
    MODELE_TerminerAjustement(modele, FAUX);
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "N_couches/covariance.c", "NC_covariance", "matrice");
    return NULL;
  }

/* Calcul de la matrice des drives secondes. Mis  part dans le cas d'une option ventuelle
   pour le choix entre un calcul exact ou simplifi. */
  if ( calc_mat_simple(hessienne->valeurs, nbr_variables, modele) == FAUX ) {
    MODELE_TerminerAjustement(modele, FAUX);
    return NULL;
  }

  covariance = inverse_matrice(hessienne, VRAI);
  effacer_matrice(hessienne); /* Plus besoin de la hessienne */
  if (NULL == covariance) {
    (void) fprintf(stderr, "MODELE_MatriceCovariance : impossible d'inverser la matrice des drives secondes\n");
    MODELE_TerminerAjustement(modele, FAUX);
    return NULL;
  }

  return covariance;
}


/* -------- Calcul de la matrice de base : formule approche (produit des drives) --------

                  N    1  [ dF_modele(x )  dF_modele(x ) ]
     alpha    = Somme --- [            k              k  ]
          i,j    k=1  s  [ -------------  ------------- ]
                       k  [      da             da       ]
                          [        i              j      ]

     Remarque : ceci est une approximation, en ralit (alpha) est la moiti de la hessienne,
  c'est--dire de la matrice des drives secondes...
     Cela correspond  l'approximation linaire, qui est de toute faon  la base du lien entre
  la hessienne et la matrice de covariance, donc a devrait tre bon (tant que l'on travaille
  avec le maximum de vraisemblance)

     Renvoie 0 si tout s'est bien pass, != 0 en cas d'erreur

 matrice    : pointeur sur les lments  calculer (triangulaire infrieure)
 N_elements : nombre d'lments dans une ligne de la matrice de covariance
 modele     : les renseignements sur le modle utilis pour l'ajustement
*/

static int calc_mat_simple(double *matrice, unsigned long N_elements, struct modele_ajuste *modele)
{
  Graphe *ajuste_i, *ajuste_j, *grf_modele;
  unsigned long i, j, k;

  for (i = 0 ; i < N_elements ; i++) {
    PG_variable var_i;

    var_i = modele->infos_variables.pseudogradient[i];
    if (var_i.modele == NULL) { /* Pas de modle associ : problme ! */
      (void) fprintf(stderr, "calc_mat_simple : pas de modle pour la variable %ld, calcul annul\n", i + 1);
      return -1;
    }

    ajuste_i = var_i.modele->modele;
    grf_modele = modele->reference; /* Spectre _total_ ajust */

    for (j = 0 ; j <= i ; j++) {
      PG_variable var_j;

      var_j = modele->infos_variables.pseudogradient[j];
      if (var_j.modele == NULL) { /* Pas de modle associ : problme ! */
	(void) fprintf(stderr, "calc_mat_simple : pas de modle pour la variable %ld, calcul annul\n", j + 1);
	return -1;
      }
      ajuste_j = var_j.modele->modele;

      if (ajuste_i != ajuste_j) { /* Variables de spectres croiss _donc_ diffrentes ->
				     Pas de corrlation (car, actuellement, pas de relations
				     possibles entre les variables */
#ifdef _DEVELOPPE_
	printf("Variables de spectres croiss : non-corrls\n");
#endif
	matrice[MAT_TINF_element(i, j)] = 0.0;
      } else {
	double somme = 0.0;
	double der_i, der_j;
	register double *x, *y_erreur, *y_modele;

	y_erreur = modele->reference->erreur_y->y;

	for (k = 0, x = ajuste_i->x, y_modele = grf_modele->y ;
	     k < ajuste_i->nombre_point ;
	     k++, x++, y_modele++, y_erreur++) {
	  double sigma = 1.0;
	  double DE_i, DE_j;
	  double khi_i, khi_j;

	  /* Calcul de la pondration par sigma */
	  sigma = *y_erreur * *y_erreur;

	  /* Inits pour la variable i */
	  if ( VRAI == modele->infos_modele.feff->drapeaux.DE0_general ) {
	    DE_i = VAR_LireValeur( &( modele->infos_modele.feff->Delta_E ), VARIABLE_LIRE_VALEUR_AJUSTEE );
	  } else {
	    DE_i = VAR_LireValeur( &( var_i.origine.couche->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE );
	  }
	  if (var_i.origine.couche != NULL) {
	    khi_i = calc_exafs_1point(*x, modele->infos_modele.feff, var_i.origine.couche, DE_i);
#ifdef _DEVELOPPE_
	    /* Les exafs calculs ici sont corrects */
/*	    printf("Variable avec couche associe - valeur de l'exafs : %g\n", khi_i); */
#endif
	  }
	  else
	    khi_i = *y_modele;
	  der_i = calc_derivee(*x, var_i.origine.couche, modele->infos_modele.feff, var_i.type,
			       VAR_LireValeur( var_i.var, VARIABLE_LIRE_VALEUR_AJUSTEE ), khi_i, DE_i);

	  /* Inits pour la variable j */
	  if ( VRAI == modele->infos_modele.feff->drapeaux.DE0_general ) {
	    DE_j = VAR_LireValeur( &( modele->infos_modele.feff->Delta_E ), VARIABLE_LIRE_VALEUR_AJUSTEE );
	  } else {
	    DE_j = VAR_LireValeur( &( var_j.origine.couche->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE );
	  }
	  if (var_j.origine.couche != NULL) {
	    khi_j = calc_exafs_1point(*x, modele->infos_modele.feff, var_j.origine.couche, DE_j);
#ifdef _DEVELOPPE_
	    /* Les exafs calculs ici sont corrects */
/*	    printf("Variable avec couche associe - valeur de l'exafs : %g\n", khi_i); */
#endif
	  }
	  else
	    khi_j = *y_modele;
	  der_j = calc_derivee(*x, var_j.origine.couche, modele->infos_modele.feff, var_j.type,
			       VAR_LireValeur( var_j.var, VARIABLE_LIRE_VALEUR_AJUSTEE ), khi_j, DE_j);
/* *	  printf("Drives, pour i : %g - pour j : %g\n", der_i, der_j); */

	  somme += der_i * der_j / sigma;
	}

/*	printf("lment (%ld, %ld) calcul : %g\n", i, j, somme); */
	matrice[MAT_TINF_element(i, j)] = 2.0 * somme;
      }
    }
  }

  return VRAI;
}
