/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   "Libration" des modles & analyse des rsultats aprs l'ajustement (partie gnrique)
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "exafs.h" /* Pour arbre des modles */

#include "couches.h"
#include "graphes.h"
#include "modeles.h"

#include "modele_fin.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int Modele_LibereInformations(ModeleAjustement *modele);
static int Modele_LibereVariables   (ModeleAjustement *modele);

static int Modele_AfficherVariables (ModeleAjustement *modele, boolean historique);

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Travail en interne : libration des graphes,... -------- */

int MODELE_TerminerAjustement(ModeleAjustement *modele, boolean aff_resultats)
{
  if ( NULL == modele ) return -1;

  /* La prparation est, forcment, finie !
     Normalement, dj FAUX mais peut tre VRAI si problme en cours de prparation...
   */
  modele->drapeaux.preparation = FAUX;

  /* L'ajustement n'est plus en cours */
  modele->drapeaux.en_cours = FAUX;

  /* On dprotge donc les graphes */
  GRF_DeprotegerGraphe( modele->experience );
  GRF_DeprotegerGraphe( modele->reference  );
  GRF_DeprotegerGraphe( modele->modele     );
  GRF_DeprotegerGraphe( modele->residu     );

  /* On libre les informations spcialises du modle */
  Modele_LibereInformations( modele );
  Modele_LibereVariables   ( modele );

  /* Tout s'est bien pass */
  if ( VRAI == aff_resultats ) {
    MODELE_AfficherResultats(modele);
  }

  return 0;
}

/* -------- Libration des informations complmentaires spcifiques du modle -------- */

static int Modele_LibereInformations(ModeleAjustement *modele)
{

  if (NULL == modele->infos_modele.cl_modeles) return -1;

  switch (modele->type)
  {
  case MODELE_MULTI_MODELES:
    {
      struct multi_modeles *infos;
      unsigned int i;

      infos = modele->infos_modele.modeles;
      for (i = 0 ; i < infos->nbr_modeles ; i++) {
	Modele_LibereInformations(infos->modeles[i]);
      }
    }
    break;

  case MODELE_CL_MODELES:
    {
      struct cl_modeles *infos;
      unsigned int i;

      infos = modele->infos_modele.cl_modeles;
      for (i = 0 ; i < infos->nbr_termes ; i++) {
	Modele_LibereInformations(infos->modeles[i]);
      }
    }
    break;

  case MODELE_EXAFS_CHEMINS:
    NC_LibererModele(modele->infos_modele.feff, VRAI);
    break;

  case MODELE_EXAFS_GNXAS:
  case MODELE_SEUIL_BASIQUE:
  case MODELE_FLUOX:
  case MODELE_DECOMPOSITION:
    fprintf(stderr, "Modele_LibereInformations : type de modle pas encore gr [%d]\n", modele->type);
    break;

  case MODELE_GENERIQUE:
    OPTION_A_FAIRE( "Libration des arbres d'expressions, modle gnrique" );
    break;

  default:
    (void) fprintf(stderr, "Modele_LibereInformations : type de modle inconnu [%d]\n", modele->type);
  }

  return 0;
}

/* -------- Libration des informations complmentaires pour l'algorithme considr -------- */

static int Modele_LibereVariables(ModeleAjustement *modele)
{
  switch ( modele->algorithme )
  {
  case ALGO_PSEUDOGRADIENT:
    free( modele->infos_variables.pseudogradient );
    modele->infos_variables.pseudogradient = NULL;
    break;

  default:
    (void) fprintf(stderr, "Modele_LibereVariables : algorithme inconnu ou pas encore support !\n");
    break;
  }

  return 0;
}

/* -~-~-~-~-~-~-~-~ Affichage des rsultats -~-~-~-~-~-~-~-~ */

int MODELE_AfficherResultats(ModeleAjustement *modele)
{
  int reponse;
  double erreur, somme_y2;

  if ( NULL == modele ) {
    return -1;
  }
  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) fprintf( stderr, "%s,%s l.%u : modle en cours d'ajustement !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return -2;
  }
  if ( ( NULL == modele->experience ) || ( NULL == modele->reference ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : modle sans graphe exprimental !\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* Affichage  l'cran */
  reponse = BI_ERR_alerte( NULL, &INF_ResultatsAjustement, VRAI,
			   modele->nom_modele, modele->experience->nom,
			   modele->erreur, modele->erreur / (double) modele->reference->nombre_point );
  if ( 2 == reponse ) { /* On ne veut pas d'historique [ni d'infos avances] */
    Modele_AfficherVariables( modele, FAUX );
    return 0;
  }

  /* Gnralits */
  HIST_Titre(1, "R"e_AIGU"sultats d'un ajustement", "Ajustement");
  HIST_Texte(2, 0, "Mod"e_GRAVE"le utilis"e_AIGU" : %s\n", modele->nom_modele);
  HIST_Texte(2, 0, "Graphe ajust"e_AIGU"    : %s\n", modele->experience->nom);
  HIST_Texte(3, 0, "Intervalle d'ajustement : [%g, %g]\n", modele->x_min, modele->x_max);
  HIST_Texte(2, 0, "Graphe du mod"e_GRAVE"le    : %s\n", modele->modele->nom);
  HIST_Texte(3, 0, "Erreur finale (minims"e_AIGU"e) : %g\n", modele->erreur);
  HIST_Texte(3, 0, "%u variables ajust"e_AIGU"es sur %lu points exp"e_AIGU"rimentaux", modele->nbr_variables,
	     modele->reference->nombre_point );
  HIST_Texte(5, 0, "It"e_AIGU"rations : %lu (sur %lu max.)\n", modele->iteration, modele->max_iteration);

  if ( NULL == modele->residu ) {
    (void) fprintf( stderr, "%s,%s l.%u : modle sans rsidu !\n",
		    __FILE__, __FUNCTION__, __LINE__);
    HIST_Erreur( 1, "Analyse annul"e_AIGU"e", "Pas de r"e_AIGU"sidu dans ce mod"e_GRAVE"le !\n" );
    return -3;
  }

  HIST_Texte( 4, 0, "\n\t\t\t Absolue \t Absolue pond"e_AIGU"r"e_AIGU" \t Relative \t Pond"e_AIGU"r"e_AIGU" relative\n" );

  /* Les diffrents types de rsidu : on calcule x, y et on sauve  x [abs], x/N [abs pond], x/y [rel], (x/y)/N [rel pond]
     1.0 Moindres carrs simples : x = S((th-exp)) & y = S(exp)*/
  somme_y2 = MODELE_CalculerErreur( modele, MODELE_ERREUR_NORME2, FAUX, FAUX );
  erreur   = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, FAUX, FAUX );
  HIST_Texte(4, 0, "Moindres carr"e_AIGU"s simples  : %g \t %g \t %g \t %g\n",
	     erreur, erreur / (double) (modele->reference->nombre_point),
	     erreur / somme_y2, (erreur / somme_y2) / (double) (modele->reference->nombre_point));

  /* 2.0 Moindres carrs pondrs : x = S((th-exp)/s), y = S(exp/s)
         Remarque : pour s = Cte, on doit retrouver les mmes valeurs pour les relatives, x/y et (x/y)/N */
  somme_y2 = MODELE_CalculerErreur( modele, MODELE_ERREUR_NORME2, VRAI, FAUX );
  if (somme_y2 > 0.0) { /* On peut pondrer par les barres d'erreur */
    erreur = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, VRAI, FAUX );
    HIST_Texte(4, 0, "Moindres carr"e_AIGU"s pond"e_AIGU"r"e_AIGU"s : %g \t %g \t %g \t %g\n",
	       erreur, erreur / (double) (modele->reference->nombre_point),
	       erreur / somme_y2, (erreur / somme_y2) / (double) (modele->reference->nombre_point));
  }

  /* 3.0 Moindres carrs relatifs : x = S((th-exp)/exp) & y = 1 */
  erreur = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, FAUX, VRAI );
  HIST_Texte(4, 0, "Moindres carr"e_AIGU"s relatifs : %g \t %g \t / \t /\n",
	     erreur, erreur / (double) (modele->reference->nombre_point) );

  /* 4.0 Moindres distances simples : x = S(|th-exp|), y = S(|exp|) */
  somme_y2 = MODELE_CalculerErreur( modele, MODELE_ERREUR_NORME1, FAUX, FAUX );
  erreur   = MODELE_CalculerErreur( modele, MODELE_ERREUR_ABSOLUE, FAUX, FAUX );
  HIST_Texte(4, 0, "Moindres distances simples   : %g \t %g \t %g \t %g\n",
	     erreur, erreur / (double) (modele->reference->nombre_point),
	     erreur / somme_y2, (erreur / somme_y2) / (double) (modele->reference->nombre_point));

  /* 5.0 Moindres distances pondres : x = S(|th-exp|/s), y = S(|exp|/s)
         Remarque : pour s = Cte, on doit retrouver les mmes valeurs pour les relatives, x/y et (x/y)/N */
  somme_y2 = MODELE_CalculerErreur(modele, MODELE_ERREUR_NORME1, VRAI, FAUX);
  if (somme_y2 > 0.0) { /* On peut pondrer par les barres d'erreur */
    erreur = MODELE_CalculerErreur(modele, MODELE_ERREUR_ABSOLUE, VRAI, FAUX);
    HIST_Texte(4, 0, "Moindres distances pond"e_AIGU"r"e_AIGU"es : %g \t %g \t %g \t %g\n",
	       erreur, erreur / (double) (modele->reference->nombre_point),
	       erreur / somme_y2, (erreur / somme_y2) / (double) (modele->reference->nombre_point));
  }

  /* 6.0 Moindres distances relatives : x = S(|th-exp|/|exp|) & y = 1 */
  erreur = MODELE_CalculerErreur(modele, MODELE_ERREUR_ABSOLUE, FAUX, VRAI);
  HIST_Texte(4, 0, "Moindres distances relatives  : %g \t %g \t / \t /\n",
	     erreur, erreur / (double) (modele->reference->nombre_point) );

  /* Faire, sur demande, les tests statistiques en fin d'ajustement */
  if ( ARRET_AJUSTEMENT_TESTS & modele->condition_arret ) {
    /* TEST DU KHI-DEUX, BASIQUE & VERSION ANOVA */
    (void) MODELE_TestKhiDeux(modele);

    (void) MODELE_TestFisherSnedecor(modele);
  }

  /* 7 Rsultats variable par variable : dpend du type de modle tudi */
  HIST_Titre(1, "R"e_AIGU"sultats d'un ajustement", "MOD");
  HIST_Texte(2, 0, "Mod"e_GRAVE"le : %s\n", modele->nom_modele);
  (void) Modele_AfficherVariables(modele, VRAI);
  return 0;
}

/* -------- Affichage pour un seul modle des rsultats pour chaque variable -------- */

static int Modele_AfficherVariables(ModeleAjustement *modele, boolean historique)
{
  if (NULL == modele) return -1;

  if (VRAI == historique) {
    printf("Modele_AfficherVariables :  FAIRE, indiquer le type de modle\n");
    if ( modele->type != MODELE_MULTI_MODELES ) {
      HIST_Texte(2, 0, "Graphe ajust"e_AIGU" : %s\n", modele->experience->nom);
      HIST_Texte(2, 0, "Intervalle d'ajustement : [%g; %g]\n", modele->x_min, modele->x_max);
      HIST_Texte(5, 2, "soit %lu points utiliss dans l'ajustement...\n", modele->reference->nombre_point);
    }
  }

  printf( "Erreur absolue : %.10lg\n", modele->erreur );

  switch (modele->type)
  {
  case MODELE_MULTI_MODELES:
    {
      struct multi_modeles *infos;
      unsigned int i;

      infos = modele->infos_modele.modeles;
      for (i = 0 ; i < infos->nbr_modeles ; i++) {
	Modele_AfficherVariables(infos->modeles[i], historique);
      }
    }
    break;

  case MODELE_CL_MODELES:
    {
      struct cl_modeles *infos;
      unsigned int i;

      infos = modele->infos_modele.cl_modeles;
      printf("Modele_AfficherVariables :  FAIRE, indiquer les rsultats pour la constante de la CL\n");
      for (i = 0 ; i < infos->nbr_termes ; i++) {
	Modele_AfficherVariables(infos->modeles[i], historique);
      }
    }
    break;

  case MODELE_EXAFS_CHEMINS:
    NC_AfficherResultats(modele->infos_modele.feff, historique);
    break;

  case MODELE_EXAFS_GNXAS:
  case MODELE_SEUIL_BASIQUE:
  case MODELE_FLUOX:
  case MODELE_DECOMPOSITION:
    fprintf(stderr, "Modele_AfficherVariables : type de modle pas encore gr [%d]\n", modele->type);
    break;

  case MODELE_GENERIQUE:
    MODELE_Generique_AfficherResultats( modele->infos_modele.generique, historique );
    break;

  default:
    (void) fprintf(stderr, "Modele_AfficherVariables : type de modle inconnu [%d]\n", modele->type);
  }

  /* Tout s'est bien pass */
  return 0;
}
