/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis
   
   Cration et gestion de modles en vue d'un ajustement
   
   Tests de validit en fin d'ajustement
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "graphes.h"
#include "modeles.h"
#include "science.h"      /* Pour gamma_q, si n'existe pas */
#include "statistiques.h" /* Pour STAT_VarianceCommune */

#include "modele_test.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Test de Fisher-Snedecor (type ANVA) ---------\
 |   Renvoie la valeur du critre calcul (<0 si erreur) |
 |                                                       |
 | modele : le modle  tester (en fin d'ajustement)     |
 \------------------------------------------------------ */

double MODELE_TestFisherSnedecor(struct modele_ajuste *modele)
{
  unsigned long ddl_numerateur, ddl_denominateur;
  double residu, variance, critere;
  double moyenne, validite;

  if (NULL == modele) return -1.0;
  if (VRAI == modele->drapeaux.en_cours) {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : modle en cours d'ajustement\n");
    return -2.0;
  }
  if ( NULL == modele->reference ) {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : modle sans rfrence exprimentale !\n");
    return -3.0;
  }
  if ( (NULL == modele->modele) || (NULL == modele->residu) ) {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : modle non ajust ou ajustement perdu !\n");
    return -4.0;
  }

  /* Conditions prliminaires */
  if ( NULL == modele->reference->erreur_y ) { /* Doit avoir des barres d'erreur associes ! */
    (void) BI_ERR_alerte(NULL, &ERR_FisherSnedecor_PasDErreur, VRAI);
    return -100.0;
  }
  if ( VRAI == modele->drapeaux.contraintes ) { /* Ne doit pas tre fait sous contrainte */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &ERR_FisherSnedecor_Contraintes, VRAI);
    if (2 == reponse) return -10.0;
  }
  if ( modele->experience->covariance != NULL ) { /* Les points doivent tre indpendants */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &ERR_KhiDeux_Covariance, VRAI);
    if (2 == reponse) return -10.0;    
  }
  /* Doit aussi tre linaire, avec des erreurs gaussiennes parfaitement connues, mais non contrl ici...
     Suppose aussi que la variance est constante : on peut utiliser la variance commune
  */

  /* Rsultat fondamental : Soit un vecteur alatoire gaussien Y de dimension N, Xk q vecteurs de dimension N.
                            Supposons que Y = a1 X1 + a2 X2 + ... + aq Xq

			    Si (a1,...,aq) est obtenu par la mthode des moindres carrs, alors
                1 t(Y - a1X1 - ...) (Y - a1X1 - ...) / (N - q)
           X = - --------------------------------------------
                N      (variance commune)/(ddl_variance)
                            suit une loi de Fisher-Snedecor  (N-q, ddl_variance) ddl
  */
  if (modele->nbr_variables > modele->reference->nombre_point) {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : plus de variables que de points exprimentaux !\n");
    return -5.0;
  }

  /* Le numrateur */
  residu = MODELE_CalculerErreur(modele, MODELE_ERREUR_CARRES, FAUX, FAUX);
  ddl_numerateur = modele->reference->nombre_point - modele->nbr_variables;

  /* Le dnominateur */
  variance = STAT_VarianceCommune(modele->reference->erreur_y->y, modele->reference->erreur_y->z.ddl,
				  modele->reference->nombre_point, &ddl_denominateur);
  if ( variance <= 0.0 ) {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : erreur dans le calcul de la variance commune !\n");
    return -6.0;
  }

  /* Critre de test rel */
  critere  = (residu / (double) ddl_numerateur) / (variance/ (double) ddl_denominateur);
  critere /= (double) (modele->reference->nombre_point);

  /* Affichage du rsultat */
  if (ddl_denominateur > 2) {
    moyenne = ((double) ddl_denominateur) / (double) (ddl_denominateur - 2);
  } else {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : la loi F(%lu, %lu) n'a pas d'esprance !\n",
		   ddl_numerateur, ddl_denominateur);
    moyenne = -1.0;
  }
  if (ddl_denominateur > 4) {
    variance  = 2.0 * moyenne * moyenne;
    variance *= (double) (ddl_numerateur + ddl_denominateur - 2);
    variance /= (double) (ddl_numerateur * (ddl_denominateur - 4));
  } else {
    (void) fprintf(stderr, "MODELE_TestFisherSnedecor : la loi F(%lu, %lu) n'a pas de variance !\n",
		   ddl_numerateur, ddl_denominateur);
    variance = -1.0;
  }

  validite = -1.0; /* Il faudrait calculer la vraie valeur */
  (void) BI_ERR_alerte(NULL, &MSG_NC_ResultatFisherSnedecor, VRAI,
		       moyenne, sqrt(variance), critere, validite);

  return validite;
}

/* -------- Test du khi-deux ----------------------------\
 |   Renvoie la valeur du critre calcul (<0 si erreur) |
 |                                                       |
 | modele : le modle  tester (en fin d'ajustement)     |
 \------------------------------------------------------ */

double MODELE_TestKhiDeux(struct modele_ajuste *modele)
{
  int *ddl;
  unsigned long i, nbr_deg_liberte;
  double *y_residu, *y_erreur;
  double residu = 0.0, validite = -1.0;

  if (NULL == modele) return -1.0;
  if (VRAI == modele->drapeaux.en_cours) {
    (void) fprintf(stderr, "MODELE_TestKhiDeux : modle en cours d'ajustement\n");
    return -2.0;
  }
  if ( NULL == modele->reference ) {
    (void) fprintf(stderr, "MODELE_TestKhiDeux : modle sans rfrence exprimentale !\n");
    return -3.0;
  }
  if ( (NULL == modele->modele) || (NULL == modele->residu) ) {
    (void) fprintf(stderr, "MODELE_TestKhiDeux : modle non ajust ou ajustement perdu !\n");
    return -4.0;
  }

  /* Conditions prliminaires */
  if ( VRAI == modele->drapeaux.contraintes ) { /* Ne doit pas tre fait sous contrainte */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &ERR_KhiDeux_Contraintes, VRAI);
    if (2 == reponse) return -10.0;
  }
  if ( FAUX == modele->drapeaux.pondere ) { /* Doit tre pondr par les barres d'erreur */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &ERR_KhiDeux_Ponderation, VRAI);
    if (2 == reponse) return -10.0;
  }
  if ( modele->experience->covariance != NULL ) { /* Les points doivent tre indpendants */
    int reponse;

    reponse = BI_ERR_alerte(NULL, &ERR_KhiDeux_Covariance, VRAI);
    if (2 == reponse) return -10.0;    
  }
  /* Doit aussi tre linaire, avec des erreurs gaussiennes parfaitement connues, mais non contrl ici...*/

  /* Rsultat fondamental : Soit un vecteur alatoire gaussien Y de dimension N, Xk q vecteurs de dimension N.
                            Supposons que Y = a1 X1 + a2 X2 + ... + aq Xq

			    Si (a1,...,aq) est obtenu par la mthode des moindres carrs, alors
           Khi = t(Y - a1X1 - ...) Sigma_Y^-1 (Y - a1X1 - ...)
                            suit une loi du khi-deux  N-q ddl
  */
  if (modele->nbr_variables > modele->reference->nombre_point) {
    (void) fprintf(stderr, "MODELE_TestKhiDeux : plus de variables que de points exprimentaux !\n");
    return -5.0;
  }

  nbr_deg_liberte = modele->reference->nombre_point - modele->nbr_variables;

  /* Calcul du rsidu correct : somme des carrs pondrs par les erreurs sur la moyenne */
  if ( NULL == modele->reference->erreur_y ) {
    (void) fprintf(stderr, "MODELE_TestKhiDeux : aucune erreur disponible !\n");
    return -6.0;
  }
  for (i = 0, y_residu = modele->residu->y, y_erreur = modele->reference->erreur_y->y, ddl = modele->reference->erreur_y->z.ddl ;
       i < modele->residu->nombre_point ; i++, y_residu++, y_erreur++, ddl++) {
    double erreur;

    erreur = *y_erreur * *y_erreur;
    if (0.0 == erreur) {
      (void) fprintf(stdout, "MODELE_TestKhiDeux : point %lu ignor, car erreur associe nulle !\n", i);
      if (nbr_deg_liberte == 0) {
	(void) fprintf(stderr, "MODELE_TestKhiDeux : plus de variables que de points exprimentaux !\n");
	return -5.0;
      }
      nbr_deg_liberte--;
      continue;
    }
    if ( *ddl > 0 ) { /* On convertit en erreur sur la moyenne */
      erreur *= (double) ( *ddl + 1 ); /* nombre de mesures = ddl + 1 */
    }

    residu += *y_residu * *y_residu / erreur;
  }
  /* IL FAUDRAIT S'ASSURER QUE L'ON UTILISE BIEN LES ERREURS SUR LA MOYENNE POUR L'AJUSTEMENT
     En pratique, on a souvent le mme nombre de mesures => facteur 1/sqrt(n) partout :
      - ne change pas l'ajustement ;
      - divise le rsidu par n : ATTENTION
 */

  /* Test (approch) du khi-deux */
  validite = gamma_q(0.5 * (double) nbr_deg_liberte, residu * 0.5);

#ifdef _DEVELOPPE_
  printf("khi suit une loi du khi-deux  %ld degrs de libert\n", nbr_deg_liberte);
  printf("Probabilit de trouver un khi-deux suprieur  %f : %g\n", residu, validite);
  printf("Elle doit tre la plus proche possible de 1\n");
#endif
  /* REMARQUE : CONTROLER S'IL FAUT UTILISER LES SIGMA DIRECTEMENT
     OU LES SIGMA SUR LA MOYENNE AVEC LES DDL MEMORISES...
  */

  (void) BI_ERR_alerte(NULL, &MSG_NC_ResultatKhiDeux, VRAI,
		       (double) nbr_deg_liberte, sqrt(2.0 * (double) nbr_deg_liberte), residu, validite);
  return validite;
}
