/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Cration : 9 fvrier 1999

   Calculs statistiques divers

   Calcul d'une analyse de variance  un facteur
                                    de rgression linaire simple (droite)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "stats_utils.h"
#include "systeme.h"

#include "interface.h"
#include "lissage.h" /* Pour moindre2 */
#include "serie.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static Graphe *STAT_ANVA_Preparer_1F(Graphe **graphes, unsigned long nbr_modalites, unsigned long *ddl_total);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Initialiser un facteur : prparer le nombre de modalits ! -------- */

int ANVA_InitialiserFacteur(struct anva_facteur *facteur, unsigned int nombre_modalites)
{
  unsigned int i;

  if ( NULL == facteur ) return -1;

  /* On libre d'ventuelles anciennes modalits */
  if ( facteur->modalites != NULL ) {
    while ( facteur->nbr_modalites > 0 ) {
      ANVA_FACTEUR_OterModalite( facteur, facteur->nbr_modalites - 1 );
    }

    /* ainsi que la table des modalits... */
    free( facteur->modalites );
    facteur->modalites = NULL;
  }

  /* On mmorise le nouveau nombre de modalits */
  facteur->nbr_modalites = nombre_modalites;

  /* On contrle que cela  un intrt... */
  if ( nombre_modalites < 2 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Facteur  0 ou 1 modalit : inutile dans l'analyse !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }

  /* On rserve les modalits */
  facteur->modalites = (struct modalite *) malloc( sizeof( struct modalite ) * nombre_modalites );
  if ( NULL == facteur->modalites ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mmoire insuffisante pour %u modalits\n",
		   __FILE__, __FUNCTION__, __LINE__, nombre_modalites );
    facteur->nbr_modalites = 0;
  }

  /* On les initialise */
  for ( i = 0 ; i < facteur->nbr_modalites ; i++ ) {
    facteur->modalites[ i ].nom      = NULL;
    facteur->modalites[ i ].valeur   = 0.0;
    facteur->modalites[ i ].effet    = 0.0;
    facteur->modalites[ i ].moyenne  = 0.0;
    facteur->modalites[ i ].SCR      = 0.0;
    facteur->modalites[ i ].valeurs  = NULL;
  }

  /* Termin sans souci */
  return 0;
}

/* -------- Changer le nom du facteur indiqu -------- */

int ANVA_FACTEUR_Renommer(struct anva_facteur *facteur, const char *nouveau_nom)
{
  if ( NULL == facteur ) return -1;

  /* On libre l'ancien nom */
  free( facteur->nom ); facteur->nom = NULL;

  /* On fait une copie, si possible, du nouveau nom */
  if ( nouveau_nom != NULL ) {
    facteur->nom = strdup( nouveau_nom );
  }

  /* Termin sans problme */
  return 0;
}

/* -------- Ajouter une modalit au facteur indiqu ------------- */

int ANVA_FACTEUR_AjouterModalite(struct anva_facteur *facteur, unsigned int numero_modalite_avant)
{
  struct modalite *nv_modalites;
  unsigned int i, numero_modalite;

  if ( NULL == facteur ) return -1;
  if ( ( facteur->nbr_modalites > 0 ) && ( numero_modalite_avant >= facteur->nbr_modalites ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Numro de modalit absurde : %u (valide : 0  %u)\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_modalite_avant, facteur->nbr_modalites - 1);
  }

  /* On rserve la place pour un modalit de plus */
  nv_modalites = (struct modalite *) realloc( facteur->modalites, sizeof( struct modalite ) * (facteur->nbr_modalites + 1) );
  if ( NULL == nv_modalites ) {
    /* Mmoire insuffisante */
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de crer une modalit supplmentaire\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On repre la nouvelle modalit */
  if ( 0 == facteur->nbr_modalites ) {
    numero_modalite = 0;
  } else {
    numero_modalite = numero_modalite_avant + 1;
  }

  /* On dplace les modalits suivantes */
  for ( i = facteur->nbr_modalites ; i > numero_modalite_avant ; i-- ) {
    nv_modalites[ i ] = nv_modalites[ i - 1 ];
  }

  /* On initialise cette modalit */
  nv_modalites[ numero_modalite ].nom      = NULL;
  nv_modalites[ numero_modalite ].valeur   = 0.0;
  nv_modalites[ numero_modalite ].effet    = 0.0;
  nv_modalites[ numero_modalite ].moyenne  = 0.0;
  nv_modalites[ numero_modalite ].SCR      = 0.0;
  nv_modalites[ numero_modalite ].valeurs  = NULL;
  
  /* On mmorise cette nouvelle table de modalits */
  facteur->modalites = nv_modalites;
  facteur->nbr_modalites++;

  /* Termin sans incident */
  return 0;
}

/* -------- ter la modalit indique du facteur indiqu -------- */

int ANVA_FACTEUR_OterModalite(struct anva_facteur *facteur, unsigned int numero_modalite)
{
  unsigned int j;

  /* Contrles prliminaires */
  if ( NULL == facteur ) return -1;
  if ( ( 0 == facteur->nbr_modalites ) || ( NULL == facteur->modalites ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Le facteur %s n'a plus de modalits !\n",
		   __FILE__, __FUNCTION__, __LINE__, facteur->nom);
    return -2;
  }
  if ( numero_modalite >= facteur->nbr_modalites ) {
    numero_modalite = facteur->nbr_modalites - 1;
  }

  /* On libre les infos de la modalit considre */
  free( facteur->modalites[ numero_modalite ].nom ); facteur->modalites[ numero_modalite ].nom = NULL;
  facteur->modalites[ numero_modalite ].valeurs = NULL;

  /* On dplace les modalits suivantes */
  for ( j = numero_modalite ; j < facteur->nbr_modalites - 1 ; j++ ) {
    facteur->modalites[ j ] = facteur->modalites[ j + 1 ];
  }
  /* La dernire est donc vide [mais on ne libre pas : elle a juste t dplace dans la nouvelle avant-dernire modalit !] */
  facteur->modalites[ facteur->nbr_modalites - 1 ].nom = NULL;
  facteur->modalites[ facteur->nbr_modalites - 1 ].valeurs = NULL;

  /* Il y a donc une modalit de moins */
  facteur->nbr_modalites--;

  /* Pas de souci */
  return 0;
}

/* -------- Prparer un graphe (isol) qui contient en X la modalit, en Y les moyennes, avec en erreur les carts-types -------- */

static Graphe *STAT_ANVA_Preparer_1F(Graphe **graphes, unsigned long nbr_modalites, unsigned long *ddl_total)
{
  Graphe *moyenne, *ecart_type;
  unsigned long i;

  /* Contrles */
  *ddl_total = 0;
  if ( ( NULL == graphes ) || ( 0 == nbr_modalites ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune modalit indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On cre un graphe isol adapt */
  moyenne = GRF_nv_isole( nbr_modalites, GR_SSTYPE, GRF_SERIE );
  if ( NULL == moyenne ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe moyen\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On lui associe des erreurs en Y */
  ecart_type = GRF_creer_erreurs( moyenne, AXE_Y );
  if ( NULL == ecart_type ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe des carts-types\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( moyenne ); moyenne = NULL;
    return NULL;
  }

  /* On calcule les moyennes et variances de chaque modalit */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    moyenne->y[ i ]           = esperance( graphes[ i ] );
    moyenne->erreur_y->y[ i ] = variance ( graphes[ i ], moyenne->y[ i ] );

    /* ddl associ [ >0 car on a mmoris la variance sur les valeurs */
    moyenne->erreur_y->z.ddl[ i ] = graphes[ i ]->nombre_point - 1;

    /* Dcompte des ddl totaux */
    *ddl_total += graphes[ i ]->nombre_point;
  }

  /* Calcul des bornes sur les moyennes pour du dessin ventuel */
  GRF_bornes( moyenne, NE_PAS_FAIRE, DESORDONNE );

  /* Termin sans problme */
  *ddl_total = *ddl_total - 1;
  return moyenne;
}

/* -------- Interface d'appel, ANVA  1 facteur --------
   Renvoie 0 si tout va bien
           != 0 en cas de problme

   graphes       : la liste des graphes. Chaque graphe correspond aux valeurs pour 1 modalit du facteur
   nbr_modalites : nombre de graphes dans la liste = nombre de modalits du facteur
   resultat      : la description des rsultats en tant que test statistique
 */

int STAT_ANVA_1facteur(Graphe **graphes, unsigned long nbr_modalites, struct test_stats *resultat)
{
  Graphe *moyenne;
  double critere;
  double moyenne_globale = 0.0, carre_total, carre_factoriel = 0.0, carre_residuel = 0.0;
  unsigned long ddl_factoriel = 0, ddl_total = 0, ddl_residuel = 0, i;

  /* Contrles prliminaires */
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de mmoriser le rsultat du test\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }
  if ( 1 == nbr_modalites ) {
    (void) fprintf(stderr, "%s,%s l.%u : Un seul facteur, impossible de voir des diffrences !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Cration du graphe des moyennes : point de dpart du travail */
  moyenne = STAT_ANVA_Preparer_1F( graphes, nbr_modalites, &ddl_total );
  if ( NULL == moyenne ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en voulant prparer les donnes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* Affichage initial */
  printf( "*** Analyse de variance : 1 facteur  %lu modalits ***\n", nbr_modalites );

  /* Calcul des sommes de carrs */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    double sigma2;

    /* Variance pour cette modalit : dans les erreurs du graphe moyenne */
    sigma2 = moyenne->erreur_y->y[ i ];

    /* On en dduit la somme des carrs rsiduels */
    carre_residuel += sigma2 * (double) ( moyenne->erreur_y->z.ddl[ i ] );

    /* Affichage */
    printf( "Modalit %lu : %lu valeurs, moyenne = %g, variance = %g, cart-type = %g\n",
	    i + 1, graphes[ i ]->nombre_point, moyenne->y[ i ], sigma2, sqrt( sigma2 ) );
  }

  /* Calcul des autres ddl */
  ddl_factoriel = nbr_modalites - 1;
  ddl_residuel  = ddl_total - ddl_factoriel;

  /* Calcul de la moyenne globale */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    moyenne_globale += moyenne->y[ i ] * (double) ( graphes[ i ]->nombre_point );
  }
  moyenne_globale /= (double) (ddl_total + 1);

  /* Calcul de la somme des carrs factorielle */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    double ecart;

    ecart = moyenne_globale - moyenne->y[ i ];
    carre_factoriel += ecart * ecart * (double) ( graphes[ i ]->nombre_point );
  }

  /* Somme des carrs totale, juste pour info */
  carre_total = carre_factoriel + carre_residuel;

  /* Calcul du critre de test */
  critere  = carre_factoriel / (double) ddl_factoriel;
  critere /= carre_residuel  / (double) ddl_residuel;

  /* Affichage des rsultats */
  printf( "Moyenne globale : %g\n", moyenne_globale );
  printf( "--------------------\n" );
  printf( "Factoriel : %lu ddl, QF = %.15e => SF = %.15e\n", ddl_factoriel, carre_factoriel, carre_factoriel / (double) ddl_factoriel );
  printf( "Rsiduel  : %lu ddl, QR = %.15e => SR = %.15e\n", ddl_residuel , carre_residuel , carre_residuel  / (double) ddl_residuel  );
  printf( "Total     : %lu ddl, QT = %.15e => ST = %.15e\n", ddl_total, carre_total, carre_total / (double) ddl_total );
  printf( "--------------------\n" );
  printf( "Critre de test : SF/SR = %.15e suit un F(%lu, %lu)\n", critere, ddl_factoriel, ddl_residuel );

  /* On initialise les rsultats */
  resultat->proprietes = STAT_TEST_EXACT;
  resultat->nom = "Analyse de variance "a_GRAVE" un facteur";
  resultat->H0 = "Toutes les esp"e_AIGU"rances sont "e_AIGU"gales";
  resultat->H1 = "Au moins une esp"e_AIGU"rance est diff"e_AIGU"rente des autres";

  resultat->v_ideale  = 0;
  if ( ddl_residuel > 2 ) {
    resultat->v_esperee = ( (double) ddl_residuel ) / (double) ( ddl_residuel - 2 );
  } else {
    /* L'esprance n'est pas dfinie si F(n,p) et p < 2 ! */
    resultat->v_esperee = -1.0;
  }

  if ( ddl_residuel > 4 ) {
    resultat->variance  = resultat->v_esperee * resultat->v_esperee;
    resultat->variance *= (double) 2 * ( ddl_factoriel + ddl_residuel - 2);
    resultat->variance /= (double) ddl_factoriel * ( ddl_residuel - 4 );
  } else {
    /* La variance n'est pas dfinie si F(n,p) et p <= 4 ! */
    resultat->variance = -1.0;
  }

  resultat->nombre_valeurs = ddl_total + 1;
  resultat->resultat = critere;

  /* Calcul termin */
  GRF_eff_isole( moyenne ); moyenne = NULL;
  return 0;
}

/* -------- Interface d'appel, ANVA de rgression linaire sur une droite --------
   Renvoie 0 si tout va bien
           != 0 en cas de problme

   graphes       : la liste des graphes. Chaque graphe correspond aux valeurs pour 1 modalit du facteur
   nbr_modalites : nombre de graphes dans la liste = nombre de modalits du facteur
   x             : le graphe contenant les X
   resultat      : la description des rsultats en tant que test statistique
 */

int STAT_ANVA_droite(Graphe **graphes, unsigned long nbr_modalites, Graphe *x, struct test_stats *resultat)
{
  Graphe *moyennes, *modele;
  Serie_Graphes *serie_resultats;
  int retour;
  double critere, coefs[ 2 ];
  double Sy, Sxy, Sx, Sx2, Dp;	/* Sommes pour les moindres carrs */
  double moyenne_globale = 0.0, carre_total, carre_factoriel = 0.0, carre_residuel = 0.0, carre_ecart = 0.0, carre_lineaire = 0.0;
  unsigned long ddl_factoriel = 0, ddl_total = 0, ddl_residuel = 0, i;

  /* Contrles prliminaires */
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de mmoriser le rsultat du test\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }
  if ( ( NULL == x ) || ( x->nombre_point != nbr_modalites ) ) {
    AVERTISSEMENT( "table des X absente ou avec un mauvais nombre de points [%lu pour %lu modalits]",
		   (x == NULL) ? (0) : (x->nombre_point), nbr_modalites );
    return -2;
  }
  if ( nbr_modalites < 3 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Moins de deux modalits, le modle linaire est forcment correct !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* Cration du graphe des moyennes : point de dpart du travail */
  moyennes = STAT_ANVA_Preparer_1F( graphes, nbr_modalites, &ddl_total );
  if ( NULL == moyennes ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en voulant prparer les donnes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* On assigne les X au graphe des moyennes */
  retour = GRF_ChangerGenre( moyennes, GRF_FCT_REELLE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer les X du graphe [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    GRF_eff_isole( moyennes ); moyennes = NULL;
    return -4;
  }
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    moyennes->x[ i ] = x->y[ i ];
  }
  moyennes->x_min = x->y_min;	/* Les bornes aussi sont conserves ! */
  moyennes->x_max = x->y_max;

  /* Affichage initial */
  printf( "*** Analyse de variance : 1 facteur  %lu modalits, rgression linaire ***\n", nbr_modalites );

  /* Somme des carrs rsiduels */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    double sigma2;

    /* Variance pour cette modalit */
    sigma2 = moyennes->erreur_y->y[ i ];

    /* On en dduit la somme des carrs rsiduels */
    carre_residuel += sigma2 * (double) ( moyennes->erreur_y->z.ddl[ i ] );

    /* Affichage */
    printf( "Modalit %lu [X=%g] : %lu valeurs, moyenne = %g, variance = %g, cart-type = %g\n",
	    i + 1, moyennes->x[ i ], graphes[ i ]->nombre_point, moyennes->y[ i ], sigma2, sqrt( sigma2 ) );
  }

  /* Calcul des autres ddl */
  ddl_factoriel = nbr_modalites - 1;
  ddl_residuel  = ddl_total - ddl_factoriel;

  /* Calcul de la moyenne globale */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    moyenne_globale += moyennes->y[ i ] * (double) ( graphes[ i ]->nombre_point );
  }
  moyenne_globale /= (double) (ddl_total + 1);

  /* Calcul de la somme des carrs factorielle */
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    double ecart;

    ecart = moyenne_globale - moyennes->y[ i ];
    carre_factoriel += ecart * ecart * (double) ( graphes[ i ]->nombre_point );
  }

  /* Estimation de la pente et de l'ordonne  l'origine : moindres carrs simples sur le graphe des moyennes
     ATTENTION, en cas de mesures de rptition diffrente, on ne peut appliquer les formules classiques... */
  /* a. Sommes */
  Sx = 0.0; Sxy = 0.0; Sx2 = 0.0; Sy = 0.0;
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    Sx  += moyennes->x[ i ] * (double) ( graphes[ i ]->nombre_point );
    Sx2 += moyennes->x[ i ] * moyennes->x[ i ] * (double) ( graphes[ i ]->nombre_point );
    Sy  += moyennes->y[ i ] * (double) ( graphes[ i ]->nombre_point );
    Sxy += moyennes->x[ i ] * moyennes->y[ i ] * (double) ( graphes[ i ]->nombre_point );
  }

  /* b. Dterminant principal */
  Dp =  Sx2 * (double) (ddl_total + 1) - Sx * Sx;
  if ( 0.0 == Dp ) {
    (void) fprintf(stderr, "%s,%s l.%u : Dterminant nul !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -20;
  }

  /* c. Coefficients */
  coefs[ 0 ]  = Sy * Sx2 - Sxy * Sx;
  coefs[ 0 ] /= Dp; /* Ordonne  l'origine */

  coefs[ 1 ]  = Sxy * (double) (ddl_total + 1) - Sx * Sy;
  coefs[ 1 ] /= Dp; /* Pente */

  /* Somme des carrs pour l'cart au modle = somme des rsidus pondre */
  /* Somme des carrs linaires */  
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    double m, ecart;

    m = coefs[ 0 ] + moyennes->x[ i ] * coefs[ 1 ]; /* Valeur prdite */

    ecart = moyenne_globale - m; /* Variance prdite par le modle */
    carre_lineaire += ecart * ecart * (double) ( graphes[ i ]->nombre_point );

    ecart = moyennes->y[ i ] - m; /* Variance due aux carts  la linarit */
    carre_ecart += ecart * ecart *  (double) ( graphes[ i ]->nombre_point );
  }

  /* Somme des carrs totale, juste pour info */
  carre_total = carre_factoriel + carre_residuel;

  /* Calcul du critre de test */
  critere  = carre_factoriel / (double) ddl_factoriel;
  critere /= carre_residuel  / (double) ddl_residuel;

  /* Affichage des rsultats */
  printf( "Moyenne globale : %g\n", moyenne_globale );
  printf( "--------------------\n" );
  printf( "Factoriel : %lu ddl, QF = %.15e => SF = %.15e\n", ddl_factoriel, carre_factoriel, carre_factoriel / (double) ddl_factoriel );
  printf( "  Lin.    : %u ddl, QL = %.15e => SF = %.15e [%g + X * %g]\n", 1, carre_lineaire, carre_lineaire, coefs[ 0 ], coefs[ 1 ] );
  printf( "  cart   : %lu ddl, QE = %.15e => SF = %.15e\n", ddl_factoriel - 1, carre_ecart, carre_ecart / (double) (ddl_factoriel - 1) );
  printf( "Rsiduel  : %lu ddl, QR = %.15e => SR = %.15e\n", ddl_residuel , carre_residuel , carre_residuel  / (double) ddl_residuel  );
  printf( "Total     : %lu ddl, QT = %.15e => ST = %.15e\n", ddl_total, carre_total, carre_total / (double) ddl_total );
  printf( "--------------------\n" );
  printf( "Critres de test : SF/SR = %.15e suit un F(%lu, %lu)\n", critere, ddl_factoriel, ddl_residuel );
  printf( "                   SE/SR = %.15e suit un F(%lu, %lu)\n",
	  ( carre_ecart / (double) (ddl_factoriel - 1) ) / ( carre_residuel  / (double) ddl_residuel ), ddl_factoriel - 1, ddl_residuel );

  /* On initialise les rsultats */
  resultat->proprietes = STAT_TEST_EXACT;
  resultat->nom = "R"e_AIGU"gression lin"e_AIGU"aire simple";
  resultat->H0 = "Un mod"e_GRAVE"le affine relie les esp"e_AIGU"rances de Y "a_GRAVE" X";
  resultat->H1 = "Un autre type de mod"e_GRAVE"le relie les esp"e_AIGU"rances de Y "a_GRAVE" X";

  resultat->v_ideale  = 0;
  if ( ddl_residuel > 2 ) {
    resultat->v_esperee = ( (double) ddl_residuel ) / (double) ( ddl_residuel - 2 );
  } else {
    /* L'esprance n'est pas dfinie si F(n,p) et p < 2 ! */
    resultat->v_esperee = -1.0;
  }

  if ( ddl_residuel > 4 ) {
    resultat->variance  = resultat->v_esperee * resultat->v_esperee;
    resultat->variance *= (double) 2 * ( ddl_factoriel + ddl_residuel - 2);
    resultat->variance /= (double) ddl_factoriel * ( ddl_residuel - 4 );
  } else {
    /* La variance n'est pas dfinie si F(n,p) et p <= 4 ! */
    resultat->variance = -1.0;
  }

  resultat->nombre_valeurs = ddl_total + 1;
  resultat->resultat = critere;

  /* On cre une srie contenant les rsultats */
  serie_resultats = SERIE_GRF_CreerSerie( serie_courante, "R"e_AIGU"sultats de la r"e_AIGU"gression lin"e_AIGU"aire" );
  if ( NULL == serie_resultats ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de mmoriser les rsultats !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( moyennes ); moyennes = NULL;
    return +1;
  }

  /* Le graphe des moyennes exprimentales, avec leur incertitudes */
  GRF_InsererDansSerie( moyennes, serie_resultats, NULL );
  GRF_titre( moyennes, "Donn"e_AIGU"es exp"e_AIGU"rimentales (moyennes)" );
  GRF_ChangerUnite( moyennes, AXE_X, GRF_QuelleUnite( x, AXE_Y ) );

  /* Le graphe du modle linaire */
  modele = GRF_nv_TableX( moyennes, GR_SSTYPE, GRF_FCT_REELLE, "Mod"e_GRAVE"le lin"e_AIGU"aire", FAUX, moyennes->x, nbr_modalites, VRAI );
  if ( NULL == modele ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe contenant le modle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +2;
  }
  for ( i = 0 ; i < nbr_modalites ; i++ ) {
    modele->y[ i ] = coefs[ 0 ] + coefs[ 1 ] * modele->x[ i ];
  }
  GRF_bornes( modele, NE_PAS_FAIRE, DESORDONNE ); /* Rien ne prouve que les x soient tris par ordre croissant ! */

  /* On va dans la srie des rsultats */
  I_SERIE_GRF_Aller( serie_resultats );

  /* Calcul termin */
  return 0;
}

/* ~-~-~-~-~-~-~-~ Travail sur une modalit ~-~-~-~-~-~ */

/* -------- Calculs qui ne concernent qu'une modalit -------- */

int ANVA_MODALITE_CalculsInitiaux(struct modalite *modalite)
{
  double ecart;
  unsigned long i;

  if ( NULL == modalite ) return -1;
  if ( NULL == modalite->valeurs ) {
    /* Aucune valeur disponible pour cette modalit */
    modalite->moyenne = 0.0;
    modalite->SCR = 0.0;

    modalite->effet = 0.0;

    /* Rien  faire donc ! Mais pas normal ==> erreur mineure*/
    return +1;
  }

  /* Calcul de la moyenne */
  STAT_Moyenne_D( modalite->valeurs->y, modalite->valeurs->nombre_point, &(modalite->moyenne) );

  /* Calcul de la somme des carrs des carts */
  for ( i = 0, modalite->SCR = 0.0 ; i < modalite->valeurs->nombre_point ; i++ ) {
    ecart = modalite->valeurs->y[ i ] - modalite->moyenne;
    modalite->SCR += ecart * ecart;
  }

  /* Termin sans souci */
  return 0;
}

/* -------- Renommer une modalit --------
     Renvoie  0 si tout s'est bien pass
             <0 en cas de problme

     Le texte pass est stock directement.
     Le texte doit tre librable avec free(), il le sera automatiquement quand la modalit sera renomme ou efface
 */

int ANVA_MODALITE_Renommer(struct modalite *modalite, char *nouveau_nom)
{
  if ( ( NULL == modalite ) || ( NULL == nouveau_nom ) ) return -1;

  free( modalite->nom ); modalite->nom = NULL;

  modalite->nom = strdup( nouveau_nom );
  return 0;
}
