/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999, Emmanuel CURIS
   Cration de ce fichier : 26 avril 1999

   Module d'analyse statistique
   Routines permettant de tirer des nombres alatoires suivant une loi donne
*/

/* #undef AVEC_GSL_ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#ifdef AVEC_GSL_
# include <gsl/gsl_rng.h>
# include <gsl/gsl_randist.h>
#endif

#include "systeme.h"

#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define CLASSIQUE   0 /* Utiliser un gnrateur classique, moyennement performant mais efficace */
#define PERFORMANT  1 /* Utiliser un gnrateur trs performant, mais plus lent */
#define RAPIDE      2 /* Utiliser un gnrateur plus rapide, mais moins tudi */
#define GSL_MT19937 3 /* Gnrateur MT19937 [Matsumoto & Nishimura] de la GSL */

#if !defined(TYPE_UNIFORME) /* Permet une dfinition ventuelle dans le Makefile */
# ifdef AVEC_GSL_ 
#  define TYPE_UNIFORME GSL_MT19937
# else
#  define TYPE_UNIFORME CLASSIQUE /* Par dfaut, on utilise le gnrateur classique */
# endif
#endif

#if TYPE_UNIFORME == CLASSIQUE
# define IA1       16807L /* Diffrentes valeurs utiliss par le gnrateur uniforme */
# define IM1  2147483647L
# define IQ1      127773L
# define IR1        2836L
# define NTAB 32 /* lments stocks dans la table pour le gnrateur uniforme */
#elif TYPE_UNIFORME == PERFORMANT
# define IM1 2147483563L
# define IM2 2147483399L
# define IA1      40014L
# define IA2      40692L
# define IQ1      53668L
# define IQ2      52774L
# define IR1      12211L
# define IR2       3791L
# define NTAB 32 /* lments stocks dans la table pour le gnrateur uniforme */
#elif TYPE_UNIFORME == RAPIDE
# define NTAB 56
#endif

static long racine = 0L; /* Racine pour le gnrateur alatoire - doit tre > 0 */
#ifdef NTAB
static long table_alea[NTAB], i_table; /* Index dans la table + table de nombres alatoires */
#endif

#if TYPE_UNIFORME == PERFORMANT
static long racine2 = 123456789L;
#endif

#ifdef AVEC_GSL_
static gsl_rng *generateur_aleatoire = NULL;
#else
static boolean g_existe = FAUX; /* VRAI si une valeur de gaussienne est dj en mmoire */
static double v_gauss = 0.0;  /* Cette valeur */
#endif

const char *txt_distributions[] = 
{
  "Gauss-Laplace N("MU", "SIGMA")",
  "Uniforme U([a, b])",
  "Cauchy C(a)",
  "Exponentielle E(l)",
  "Double exponentielle E(l, "MU")",
  "Vecteur gaussien"
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Initialisation des gnrateurs --------\
 |  Renvoie <0 en cas d'erreur                     |
 |                                                 |
 | si racine > 0, utilise l'horloge interne        |
 | si racine <= 0, prend cette racine comme racine |
 \------------------------------------------------ */

int init_aleatoire(long racine_donnee)
{
#ifndef AVEC_GSL_
  int j;
  long k;
#endif

  /* Recherche de la racine */
  if (racine_donnee <= 0)
    racine = -racine_donnee;
  else {
    racine= (long) time(NULL);
#ifdef _gettimeofday_
    struct timeval temps;
# ifdef _HP_UNIX_ /* Sous HP Unix, le 2e argument doit tre NULL ! */
    gettimeofday(&temps, NULL);
# else
    struct timezone a;
    gettimeofday(&temps, &a); /* On rcupre l'heure systme */
# endif

    racine = (long) (temps.tv_usec);
#endif
  }

  if (racine < 0) racine = -racine;
  if (racine == 0) racine = 1L;

#if (TYPE_UNIFORME == PERFORMANT) || (TYPE_UNIFORME == CLASSIQUE)
# if TYPE_UNIFORME == PERFORMANT
  racine2 = racine;
# endif

  /* Gnration de la table : 8 tours pour rien avant
     Fonctionne pour le classique et le "performant" */

  for ( j = NTAB + 7 ; j >= 0 ; j-- ) {
    k = racine / IQ1;
    racine = IA1 * (racine - k * IQ1) - k * IR1;
    if (racine < 0) racine += IM1;
    if (j < NTAB) table_alea[j] = racine;
  }
  i_table = table_alea[0];
#else
# ifdef AVEC_GSL_
#  if GSL_MT19937 == TYPE_UNIFORME
  generateur_aleatoire = gsl_rng_alloc( gsl_rng_mt19937 );
#  endif
  if ( NULL == generateur_aleatoire ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'initialiser le gnrateur de nombres alatoires [GSL, type %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, TYPE_UNIFORME );
    return -5;
  }

  /* On mmorise la racine */
  gsl_rng_set( generateur_aleatoire, racine );
# else
  OPTION_A_FAIRE( "Initialisation, gnrateur de type GSL sans la GSL" );
# endif
#endif

  return 0; /* Pas de problme survenu */
}

/* -------- Librer les structures de donnes ventuellement gnres  l'initialisation -------- */

void STAT_TerminerAleatoire(void)
{
#ifdef AVEC_GSL_
  gsl_rng_free( generateur_aleatoire );
  generateur_aleatoire = NULL;
#else
  /* Normalement, rien  librer si l'on n'utilise pas la GSL */
#endif
}

/* -------- Interface gnrique de gnration de nombres alatoires -------- */

double STAT_NombreAleatoire(struct loi_probabilite *loi)
{
  if ( NULL == loi ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune loi indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }

  switch ( loi->famille )
  {
  case DP_GAUSSIENNE:
#ifdef _DEVELOPPE_
    if ( loi->parametres.Gauss.sigma <= 0.0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : cart-type absurde [%g], gaussienne\n",
		     __FILE__, __FUNCTION__, __LINE__, loi->parametres.Gauss.sigma );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      return loi->parametres.Gauss.mu;
    }
#endif
    return gaussienne() * loi->parametres.Gauss.sigma + loi->parametres.Gauss.mu;

  case DP_UNIFORME:
    if ( loi->proprietes & DP_PROPRIETE_DISCRETE ) {
      OPTION_A_FAIRE( "Loi uniforme discrte" );
    } else {
      return STAT_NombreAleatoire_UniformeC( loi->parametres.Uniforme.min, loi->parametres.Uniforme.max );
    }
  }

  (void) fprintf(stderr, "%s, %s l.%u : Type de loi inconnu [%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, loi->famille);
  return 0.0;
}

/* ~-~-~-~-~-~-~-~-~ Generation de nombres aleatoires : loi uniforme ~-~-~-~-~-~-~-~-~ */

/* -------- Loi uniforme continue sur [a, b] -------- */

double STAT_NombreAleatoire_UniformeC(double min, double max)
{
  return uniforme() * ( max - min ) + min;
}

#if TYPE_UNIFORME == CLASSIQUE

/* -------- Loi uniforme sur [0, 1] --------

   Gnrateur emprunt aux Numerical Recipes 2e dition
      initialement : ran1 = congruence linaire (Ij+1 = a*Ij modulo m),
                            avec sortie alatoire des valeurs. Priode : environ 10^8

   Pour avoir une loi uniforme sur [a, b] : X' = a + (b - a) X
*/

#define NDIV (1 + (IM1 - 1L) / NTAB )

double uniforme(void)
{
  int j;
  long k;

  if ( (i_table == 0) || (racine <= 0) ) /* Il faut rinitialiser */
  {
#ifdef _DEVELOPPE_
    printf("Demande de rinitialisation du gnrateur de nombres alatoires\n");
#endif
    init_aleatoire(0);
  }

  k = racine / IQ1;
  racine = IA1 * (racine - k * IQ1) - IR1 * k;
  if (racine < 0) racine += IM1;

  j = i_table / NDIV;
  i_table = table_alea[j];
  table_alea[j] = racine;

  return ((double) i_table) / (double) IM1;
}

#elif TYPE_UNIFORME == PERFORMANT

/* -------- Loi uniforme sur [0, 1] --------

   Gnrateur emprunt aux Numerical Recipes 2e dition
      initialement : ran2 = gnrateur de L'cuyer avec mlange de Bays-Durham
                            priode estime > 2^18 (!)

   Pour avoir une loi uniforme sur [a, b] : X' = a + (b - a) X
*/

#define NDIV (1 + (IM1 - 1L) / NTAB )

double uniforme(void)
{
  int j;
  long k;

  if ( (i_table == 0) || (racine <= 0) ) /* Il faut rinitialiser */
  {
#ifdef _DEVELOPPE_
    printf("Demande de rinitialisation du gnrateur de nombres alatoires\n");
#endif
    init_aleatoire(0);
  }

  k = racine / IQ1;
  racine = IA1 * (racine - k * IQ1) - IR1 * k;
  if (racine < 0) racine += IM1;

  k = racine2 / IQ2;
  racine2 = IA2 * (racine2 - k * IQ2) - IR2 * k;
  if (racine2 < 0) racine2 += IM2; 

  j = i_table / NDIV;
  i_table = table_alea[j] - racine2;
  table_alea[j] = racine;

  if (i_table < 1) i_table += (IM1 - 1);

  return ( (double) i_table) / (double) IM1;
}

#elif TYPE_UNIFORME == RAPIDE

/* -------- Loi uniforme sur [0, 1] --------

   Gnrateur emprunt aux Numerical Recipes 2e dition
      initialement : ran3 = mthode soustractive de Knuth

   Pour avoir une loi uniforme sur [a, b] : X' = a + (b - a) X
*/

#else
# ifdef AVEC_GSL_
double uniforme( void )
{
  return gsl_rng_uniform( generateur_aleatoire );
}
# else
/*  FAIRE : Gnrateur uniforme type GSL, sans la GSL */
# endif

#endif

/* -------- Loi de Laplace-Gauss centre rduite : N(0,1) --------

   Mthode de Box-Muller (cf. les Numerical Recipes 2e dition)

   Pour avoir une gaussienne N(m, s) : X' = m + s X
*/

double gaussienne(void)
{
#ifdef AVEC_GSL_
  return gsl_ran_ugaussian( generateur_aleatoire );
#else
  double v1, v2, dist;
  float tmp;

  if (g_existe == VRAI) { /* Il y a une valeur toute prte ! */
    g_existe = FAUX;
    return v_gauss;
  }

  do {
    v1 = 2.0 * uniforme() - 1.0; /* On tire un point dans le carr unit */
    v2 = 2.0 * uniforme() - 1.0;
    dist = v1 * v1 + v2 * v2; /* Leur distance  l'origine */
  }
  while ( (dist >= 1.0) || (dist == 0.0) ); /* En-dehors du cercle ou  l'origine */

  tmp = sqrt( -2.0 * log(dist) / dist );
  v_gauss = v1 * tmp;
  g_existe = VRAI;
  return (v2 * tmp);
#endif
}
