/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2006, Emmanuel CURIS
   Cration : 3 aott 2006

   Analyses statistiques

   Comparaisons multiples [de moyennes]
    (pour les comparaisons fonctionnelles de graphes : cf. comparer_graphes.c)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "graphes.h"
#include "operations.h"
#include "proba.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static Graphe *STAT_GRF_Comparer2Graphes                   (Graphe *graphe_1, Graphe *graphe_2, Graphe **difference);
static Graphe *STAT_GRF_Comparer2Graphes_StudentIndependant(Graphe *graphe_1, Graphe *graphe_2);

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Fonction d'appel gnrique, visible de l'extrieur ~-~-~-~-~-~-~-~-~ *
 * Renvoie : les degrs de signification                                                  *
 *           le graphe avec la diffrence dans r_difference (si non-NULL)                 *
 * Normalement,  charge de l'appelant de vrifier que les graphes sont compatibles,...   */

Graphe *STAT_GRF_ComparerGraphes(unsigned long nombre_graphes, Graphe **graphes, Graphe *reference __attribute__((__unused__)), Graphe **difference)
{
  Graphe *resultat = NULL;

  if ( NULL == graphes ) return NULL;
  if ( nombre_graphes < 2 ) return NULL;

  /* En fonction du nombre de graphes, on appelle les bonnes sous-fonctions */
  if ( 2 == nombre_graphes ) {
    resultat = STAT_GRF_Comparer2Graphes( graphes[ 0 ], graphes[ 1 ], difference );
  } else {
    OPTION_A_FAIRE( "Comparer plus de deux graphes" );
  }

  return resultat;
}

/* ~-~-~-~-~-~-~-~-~ Comparaisons de deux graphes : version Student ~-~-~-~-~-~-~-~-~ */

static Graphe *STAT_GRF_Comparer2Graphes(Graphe *graphe_1, Graphe *graphe_2, Graphe **r_difference)
{
  Graphe *resultat = NULL;
  Graphe *difference = NULL;

  if ( ( NULL == graphe_1 ) || ( NULL == graphe_2 ) ) {
    return NULL;
  }

  /* Commun  tous : le calcul de la diffrence des deux spectres ; isole... */
  difference = GRF_Copier( graphe_1, NULL, VRAI );
  GRF_Rclin( difference, 1.0, graphe_2, -1.0 );

  /* En fonction des barres d'erreur [en Y] */
  if ( ( graphe_1->erreur_y != NULL ) && ( graphe_2->erreur_y != NULL ) ) {
    /* Barres d'erreurs sur les 2 ==> Student de 2 chantillons indpendants */
    resultat = STAT_GRF_Comparer2Graphes_StudentIndependant( graphe_1, graphe_2 );
  } else if ( graphe_1->erreur_y != NULL ) {
    /* Barres d'erreurs sur l'un seulement ==> Student de comparaison  une valeur thorique */
    OPTION_A_FAIRE( "Un seul graphe avec des incertitudes" )
  } else if ( graphe_2->erreur_y != NULL ) {
    /* Barres d'erreurs sur l'un seulement ==> Student de comparaison  une valeur thorique */
    OPTION_A_FAIRE( "Un seul graphe avec des incertitudes" )
  } else {
    /* Aucune barres d'erreur : seule la diffrence a un sens... */
    OPTION_A_FAIRE( "Calcul de la diffrence..." )
  }

  /* Calculs termins */
  if ( r_difference != NULL ) {
    *r_difference = difference;
  } else {
    GRF_eff_isole( difference );
    difference = NULL;
  }
  return resultat;
}

static Graphe *STAT_GRF_Comparer2Graphes_StudentIndependant(Graphe *graphe_1, Graphe *graphe_2)
{
  Graphe *resultat = NULL;

  /* Les X sont-ils compatibles ? */
  if ( VRAI == GRF_ComparerX( graphe_1, graphe_2 ) ) {
    unsigned long i;
    register double *y_1, *y_2, *err_y1, *err_y2, *y_ds;
    register int *n_1, *n_2;

    /* Mme X exactement : aucune difficult pour la comparaison... */
    
    /* --- Le graphe rsultat a donc les mmes X, exactement */
    resultat = GRF_nv_TableX( NULL, GR_DEGRE_SIGNIFICATION, GRF_FCT_REELLE, "p-values/Student", VRAI,
			      graphe_1->x, graphe_1->nombre_point, VRAI );
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : Problme en crant le graphe des d.s.\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Point par point, on calcule le d. s. */
    for ( i = 0, y_1 = graphe_1->y, y_2 = graphe_2->y, y_ds = resultat->y,
	    err_y1 = graphe_1->erreur_y->y, err_y2 = graphe_2->erreur_y->y,
	    n_1 = graphe_1->erreur_y->z.ddl, n_2 = graphe_2->erreur_y->z.ddl ;
	  i < graphe_1->nombre_point ;
	  i++, y_1++, y_2++, y_ds++, err_y1++, err_y2++, n_1++, n_2++ ) {
      double d, sd_d;
      unsigned int ddl;

      /* La diffrence */
      d = *y_2 - *y_1;

      /* L'cart-type de la diffrence [moyenne]
	 (si n>0 : cart-type de la population ; si n<0 : cart-type de la moyenne -- en DDL !)
       */
      if ( *n_1 > 0 ) {
	if ( *n_2 > 0 ) {
	  sd_d = *err_y1 * *err_y1 / (*n_1 + 1) + *err_y2 * *err_y2 / (*n_2 + 1);
	} else {
	  sd_d = *err_y1 * *err_y1 / *n_1 + *err_y2 * *err_y2;
	}
      } else {
	if ( *n_2 > 0 ) {
	  sd_d = *err_y1 * *err_y1 + *err_y2 * *err_y2 / *n_2;
	} else {
	  sd_d = *err_y1 * *err_y1 + *err_y2 * *err_y2;
	}
      }
      sd_d = sqrt( sd_d );

      /* Les ddl */
      ddl = abs( *n_1 ) + abs( *n_2 );

      /* Le t de Student */
      d = d/sqrt( sd_d );

      /* Le p qui va avec, en utilisant la symtrie [test bilatral] */
      *y_ds = 2 * (1 - PROBA_FdR_Student( fabs( d ), ddl ) );
    }

    /* On met les bornes en Y : pas par le calcul, mais sachant que ce sont des probabilits */
    resultat->y_min = 0.0;     resultat->y_max = 1.0;
  } else {
    OPTION_A_FAIRE( "Cas des X diffrents : interpolations ? et pour les s ?" )
  }

  /* Termin */
  return resultat;
}
