/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999, Emmanuel CURIS
   Création de ce fichier : 26 avril 1999

   Module d'analyse statistique
   Routines de calcul des densités de probabilités des lois usuelles
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "systeme.h"

#include "exafs.h"

#include "science.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_TXT_DISTRIBUTION 34

static int modele = DP_GAUSSIENNE;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

Graphe *calc_densite(int modele, double x_min, double x_max, unsigned long nbr_pts);

void prepare_arbre_densite(void); /* Non static car utilisé aussi pour les tirages et les FR */
int GER_densites(unsigned int objet, int x, int y);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Sélection du modèle de densité à utiliser -------- */

Graphe *calc_densite(int modele, double x_min, double x_max, unsigned long nbr_pts)
{
  double p1, p2, p3, p4 __attribute__((__unused__));
  Graphe *densite = NULL;

  p1 = OBJ_LireNombre(CREER_DIST_CONT, CDC_PARAM_1, NULL );
  p2 = OBJ_LireNombre(CREER_DIST_CONT, CDC_PARAM2, NULL );
  p3 = OBJ_LireNombre(CREER_DIST_CONT, CDC_PARAM3, NULL );
  p4 = OBJ_LireNombre(CREER_DIST_CONT, CDC_PARAM4, NULL );

  switch (modele)
  {
  case DP_GAUSSIENNE:
    if (p1 < 0.0) p1 = -p1;
    if (p1 == 0.0) {
      EXF_ERREUR(DIV_ZERO, "Variance nulle", txt_err_calc_annule);
      return NULL;
    }
    densite = DP_gaussienne(p1, p2, x_min, x_max, nbr_pts);
    break;

  case DP_UNIFORME:
    if (p1 > p2) {double p = p1; p1 = p2; p2 = p;}
    if (p1 == p2) {
      EXF_ERREUR(DIV_ZERO, "Intervalle vide", txt_err_calc_annule);
      return NULL;
    }
    densite = DP_uniforme(p1, p2, x_min, x_max, nbr_pts);
    break;

  case DP_CAUCHY:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_cauchy(p1, p2, x_min, x_max, nbr_pts);
    break;

  case DP_EXPONENTIEL:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_exponentielle(p1, p2, x_min, x_max, nbr_pts);
    break;

  case DP_EXP_SYMETRIQUE:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_double_exp(p1, p2, x_min, x_max, nbr_pts);
    break;

  case DP_BETA_I:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    if (p3 < 0.0) p3 = -p3;
    if (p3 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_beta_I(p1, p2, p3, x_min, x_max, nbr_pts);
    break;

  case DP_BETA_II:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    if (p3 < 0.0) p3 = -p3;
    if (p3 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_beta_II(p1, p2, p3, x_min, x_max, nbr_pts);
    break;

  case DP_FISHER:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    if (p3 < 0.0) p3 = -p3;
    if (p3 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_fisher(p1, p2, p3, x_min, x_max, nbr_pts);
    break;

  case DP_CHIDEUX: /* Loi du khi-deux à n degrés de liberté = loi gamma(n/2, 1/2) */
    p2 /= 2.0;
    p3 = 0.5;
  case DP_GAMMA:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      return NULL;
    }
    densite = DP_gamma(p1, p2, p3, x_min, x_max, nbr_pts);
    break;

  case DP_GUMBEL:
    densite = DP_gumbel(p1, x_min, x_max, nbr_pts);
    break;

  case DP_STUDENT:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) { /* Densité indéfinie */
      EXF_ERREUR(DIV_ZERO, "Aucun degr"e_AIGU" de libert"e_AIGU, txt_err_calc_annule);
      return NULL;
    }
    densite = DP_student(p1, p2, x_min, x_max, nbr_pts);
    break;

  case DP_PEARSON:
    break;

  case DP_LOGNORMALE:
    if (p2 < 0.0) p2 = -p2;
    if (p2 == 0.0) {
      EXF_ERREUR(DIV_ZERO, "\"Variance\" nulle", txt_err_calc_annule);
      return NULL;
    }
    densite = DP_lognormale(p1, p2, p3, x_min, x_max, nbr_pts);
    break;

  case DP_WEIBULL:
    densite = DP_weibull(p1, p2, p3, x_min, x_max, nbr_pts);
    break;

  default:
    fprintf(stderr, "modele_probas : loi de probabilité inconnue, %d\n", modele);
  }

  return densite;
}

/* -------- Calcul pour une loi bêta de type I, BI(n, p), éventuellement décalée -------- */

Graphe *DP_beta_I(double delta, double n, double p, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  if ( (x_max - delta < 0.0) || (x_min - delta > 1.0) ) {
    fprintf(stderr, "DP_beta_I : intervalle demandé hors de l'intervalle de définition\n");
    fprintf(stderr, "            la densité est nulle partout ==> graphe non créé\n");
    return NULL;
  }

  pas = (x_max - x_min) / (double) (nbr_pts - 1);
  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = 1.0 / beta(n, p);
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta;  /* Centrage */
    if ( (x <= 0.0) || (x >= 1.0) )
      (densite->y)[i] = 0.0;
    else {
      y = pow(x, n - 1.0) * pow(1.0 - x, p - 1.0);
      y *= coef;
      (densite->y)[i] = y;
    }
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une loi bêta de type II, BI(n, p), éventuellement décalée -------- */

Graphe *DP_beta_II(double delta, double n, double p, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  if ( x_max - delta < 0.0 ) {
    fprintf(stderr, "DP_beta_II : intervalle demandé hors de l'intervalle de définition\n");
    fprintf(stderr, "              la densité est nulle partout ==> graphe non créé\n");
    return NULL;
  }

  pas = (x_max - x_min) / (double) (nbr_pts - 1);
  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = 1.0 / beta(n, p);
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta;  /* Centrage */
    if ( x <= 0.0 )
      (densite->y)[i] = 0.0;
    else {
      y = pow(x, n - 1.0) / pow(1.0 + x, n + p);
      y *= coef;
      (densite->y)[i] = y;
    }
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une loi de Cauchy C(a) centrée sur le mode -------- */

Graphe *DP_cauchy(double mode, double a, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);
  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = a / M_PI;
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= mode;  /* Centrage */
    y = 1/(x * x + a * a);
    y *= coef;
    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une exponentielle double E(lambda), décalée de delta -------- */

Graphe *DP_double_exp(double delta, double lambda, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta; /* Centrage */
    y = lambda * exp(- fabs(x) * lambda) / 2.0;
    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une exponentielle simple E(lambda), décalée de delta -------- */

Graphe *DP_exponentielle(double delta, double lambda, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas;
  unsigned long i;

  if (x_max < -delta) {
    fprintf(stderr, "DP_exponentielle : intervalle demandé hors de l'intervalle de définition\n");
    fprintf(stderr, "              la densité est nulle partout ==> graphe non créé\n");
    return NULL;
  }

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta; /* Centrage */
    y = lambda * exp(- x * lambda);
    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

Graphe *DP_fisher(double mode, double n, double p, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  unsigned long i;
  double coef, *y;

  densite = DP_beta_I(mode, n/2.0, p/2.0, x_min, x_max, nbr_pts);
  if (densite == NULL) return NULL;

  coef = n/p;
  for (i = 0, y = densite->y ; i < densite->nombre_point ; i++, y++)
    *y *= coef;
  densite->y_min *= coef;
  densite->y_max *= coef;

  return densite;
}

/* -------- Calcul pour une loi gamma g(alpha, beta) décalée de delta -------- */

Graphe *DP_gamma(double delta, double alpha, double beta, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  if ( x_max - delta < 0.0 ) {
    fprintf(stderr, "DP_gamma : intervalle demandé hors de l'intervalle de définition\n");
    fprintf(stderr, "           la densité est nulle partout ==> graphe non créé\n");
    return NULL;
  }

  pas = (x_max - x_min) / (double) (nbr_pts - 1);
  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = alpha * log(beta) - lgamma(alpha);
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta;  /* Centrage */
    if ( x <= 0.0 )
      (densite->y)[i] = 0.0;
    else {
      y = coef + (alpha - 1.0) * log(x) - beta * x;
      (densite->y)[i] = exp(y);
    }
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une gaussienne N(µ, sigma) -------- */

Graphe *DP_gaussienne(double mu, double sigma, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = sqrt(2 * M_PI) * sigma;
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= mu;    /* Centrage */
    x /= sigma; /* Réduction */
    y = exp(- x * x / 2.0);
    y /= coef;
    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une loi de Gumbel décalée de delta -------- */

Graphe *DP_gumbel(double delta, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta; /* Décalage */
    y = exp(-x);
    y = exp(-x - y);
    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une loi log-normale(µ, sigma) -------- */

Graphe *DP_lognormale(double mu, double sigma, double gamma, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  if (x_max <= gamma) {
    fprintf(stderr, "DP_lognormale : la densité est nulle partout ==> graphe non créé\n");
    return NULL;
  }

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = sqrt(2 * M_PI) * sigma;
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    if ( x > gamma ) {
      x -= gamma; /* Centrage */
      y = (log(x) - mu) / sigma;
      y = exp(- y * y / 2.0);
      y /= coef * sqrt(x);
    } else
      y = 0.0;
    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une loi de Student à n degrés de liberté, Tn, décalée de delta -------- */

Graphe *DP_student(double delta, double n, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = sqrt(n) * beta(0.5, n/2.0);
  for (i = 0 ; i < densite->nombre_point ; i++) {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    x -= delta; /* "centrage" */
    y  = pow(1.0 + x*x/n, -(n+1.0)/2.0);
    y /= coef;

    (densite->y)[i] = y;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;
}

/* -------- Calcul pour une loi uniforme U([min, max]) -------- */

Graphe *DP_uniforme(double min, double  max, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  if ( (max <= x_min) || (min >= x_max) ) {
    fprintf(stderr, "DP_uniforme : intervalle demandé disjoint de l'intervalle de définition\n");
    fprintf(stderr, "              la densité est nulle partout ==> graphe non créé\n");
    return NULL;
  }

  if ( (max >= x_max) && (min <= x_min) ) {
    fprintf(stderr, "DP_uniforme : intervalle demandé inclus dans l'intervalle de définition\n");
    fprintf(stderr, "              la densité est constante partout, égale à %g ==> graphe non créé\n", 1/(max - min));
    return NULL;
  }

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = 1 / (max - min);
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    if ( (x > min) && (x < max) )
      (densite->y)[i] = coef;
    else
      (densite->y)[i] = 0.0;
  }
  densite->x_min = x_min;
  densite->x_max = x_max;
  densite->y_min = 0.0;
  densite->y_max = coef;

  return densite;
}

/* -------- Calcul pour une loi de Weibull générale W(alpha, beta, gamma) -------- */

Graphe *DP_weibull(double alpha, double beta, double gamma, double x_min, double x_max, unsigned long nbr_pts)
{
  Graphe *densite;
  double pas, coef;
  unsigned long i;

  pas = (x_max - x_min) / (double) (nbr_pts - 1);

  densite = GRF_nouveau( NULL, nbr_pts, GR_DENSITE, GR_NORMAL, "Densit"e_AIGU" de probabilit"e_AIGU );
  if (densite == NULL)
    return NULL;

  coef = beta / fabs(alpha);
  for (i = 0 ; i < densite->nombre_point ; i++)
  {
    double x, y;

    x = x_min + pas * (double) i;
    (densite->x)[i] = x;

    if ( x <= 0.0 )
      (densite->y)[i] = 0.0;
    else {
      y = pow(x, beta) - gamma;
      y /= alpha;
      if (y < 0.0)
	y = 0.0;
      else {
	y  = exp(-y);
	y *= coef * pow(x, beta - 1.0);
      }
      (densite->y)[i] = y;
    }
  }
  densite->x_min = x_min;
  densite->x_max = x_max;

  GRF_bornes(densite, NE_PAS_FAIRE, DESORDONNE);
  return densite;  
}

/* -------- Interface utilisateur -------- */

Graphe *modele_probas(void)
{
  Graphe *densite = NULL;
  double x_min, x_max;
  unsigned long nbr_pts;

  if (arbres[CREER_DIST_CONT].drapeaux.prepare == FAUX)
    prepare_arbre_densite();

  OBJ_active(CREER_DIST_CONT, CDC_XMIN, FORCE_OUI);
  OBJ_active(CREER_DIST_CONT, CDC_X_MAX, FORCE_OUI);
  OBJ_active(CREER_DIST_CONT, CDC_N_POINTS, FORCE_OUI);

  if ( ARB_Bloquant( CREER_DIST_CONT, "Mod"e_GRAVE"le de distribution", principale, 0 ) == CDC_ANNULER)
    return NULL;

  x_min = OBJ_LireNombre(CREER_DIST_CONT, CDC_XMIN, NULL );
  x_max = OBJ_LireNombre(CREER_DIST_CONT, CDC_X_MAX, NULL );
  if (x_max < x_min) {double x = x_min; x_min = x_max; x_max = x;}
  if (x_max == x_min) {
    EXF_ERREUR(PAS_FAUX);
    return NULL;
  }

  nbr_pts = OBJ_LireEntier_U( CREER_DIST_CONT, CDC_N_POINTS );
  if (nbr_pts < 3) {
    EXF_ERREUR(TROP_PEU_POINTS);
    return NULL;
  }

  if ( OBJ_choisit(CREER_DIST_CONT, CDC_DENSITE, QUESTION) == VRAI ) { /* On calcule la densité */
    return calc_densite(modele, x_min, x_max, nbr_pts);
  }


  return densite;
}

/* -------- Gestion de l'interface -------- */

int GER_densites(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int nv_modele;
  double p1, p2;

  switch (objet)
  {
  case CDC_DISTRIB:
    OBJ_choisit(CREER_DIST_CONT, objet, FORCE_OUI);
    nv_modele = SS_menu_texte(txt_loi_probas_continues, DP_NOMBRE_LOIS, "Loi de probabilit"e_AIGU, CREER_DIST_CONT, CDC_DISTRIB);
    if ( (nv_modele >= 0) && (nv_modele != modele) ) {
      met_chaine(CREER_DIST_CONT, CDC_DISTRIB, LGMAX_TXT_DISTRIBUTION, txt_loi_probas_continues[nv_modele], TXT_COPIER);
      modele = nv_modele;

      OBJ_active(CREER_DIST_CONT, CDC_PARAM3,
		 (modele == DP_BETA_I) || (modele == DP_BETA_II) || (modele == DP_GAMMA) || (modele == DP_FISHER) ||
		 (modele == DP_WEIBULL) || (modele == DP_LOGNORMALE) );
    }
    OBJ_choisit(CREER_DIST_CONT, objet, FORCE_NON);
    return BI_EVENEMENT_TRAITE;

  case CDC_PROPOSER:
    OBJ_choisit(CREER_DIST_CONT, objet, FORCE_OUI);
    p1 = OBJ_LireNombre( CREER_DIST_CONT, CDC_PARAM_1, NULL );
    p2 = OBJ_LireNombre( CREER_DIST_CONT, CDC_PARAM2, NULL );
    switch (modele)
    {
    case DP_GAUSSIENNE:
      OBJ_MettreNombre( CREER_DIST_CONT, CDC_XMIN, p1 - 5.0 * fabs(p2), 0.0 );
      OBJ_MettreNombre( CREER_DIST_CONT, CDC_X_MAX, p1 + 5.0 * fabs(p2), 0.0 );
      break;
    }
    OBJ_choisit(CREER_DIST_CONT, objet, FORCE_NON);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Préparation de l'interface -------- */

void prepare_arbre_densite(void)
{
  OBJ_agrandit(CREER_DIST_CONT, CDC_FND_GENERER, +3, +1);
  OBJ_deplace(CREER_DIST_CONT, CDC_F_REPART, +1, 0, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_POINTS, +2, 0, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_XMIN, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_X_MAX, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_N_POINTS, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PROPOSER, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_TXT_PARAM2, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PARAM2, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_TXT_PARAM3, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PARAM3, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_TXT_PARAM4, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PARAM4, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_TXT_PARAM6, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PARAM6, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_TXT_PARAM7, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PARAM7, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_TXT_PARAM8, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace(CREER_DIST_CONT, CDC_PARAM8, 0, +3, DEPLACE_PIXELS);

  met_chaine(CREER_DIST_CONT, CDC_DISTRIB, LGMAX_TXT_DISTRIBUTION, txt_loi_probas_continues[modele], TXT_COPIER);
  OBJ_MettreNombre( CREER_DIST_CONT, CDC_PARAM_1, 0.0, 0.0 ); /* Centrée réduite par défaut */
  OBJ_MettreNombre( CREER_DIST_CONT, CDC_PARAM2, 1.0, 0.0 );
  GER_densites(CDC_PROPOSER, 0, 0);

  arbres[CREER_DIST_CONT].fonction.arbre = GER_densites;
  arbres[CREER_DIST_CONT].drapeaux.prepare = VRAI;
}
