/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2006, Emmanuel CURIS
   Création : 3 aoûtt 2006

   Analyses statistiques

   Comparaisons multiples [de moyennes]
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "graphes.h"
#include "proba.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Réalisation pratique ~~~~~~~~ */

/* -------- Fonction générique, visible depuis l'extérieur -------- */

int STAT_ComparerAUnTemoin_Dunnett(unsigned int nombre_comparaisons, struct comparaison *comparaisons, int sens, double alpha_global, unsigned int approximation,
				   double variance_commune, unsigned long ddl_SR, boolean est_equilibree)
{
  double d_Dunnett = 0.0;
  double alpha_individuel;
  unsigned int i;
  unsigned long n_temoin, n_modalite;

  if ( ( 0 == nombre_comparaisons ) || ( NULL == comparaisons ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune comparaison à faire !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }

  /* 1re étape : déterminer le "fractile" */
  switch ( approximation )
  {
  case STAT_IC_FRACTILE_EXACT:
    d_Dunnett = STAT_Fractile_Dunnett( alpha_global, nombre_comparaisons, comparaisons, ddl_SR, sens );
    break;

  case STAT_IC_FRACTILE_BONFERONI:
    alpha_individuel = PROBA_INEG_Bonferroni( alpha_global, nombre_comparaisons );
    if ( 0 == sens ) {
      d_Dunnett = PROBA_Fractile_Student( 1.0 - alpha_individuel / 2.0, ddl_SR );
      /* Test bilatéral */
    } else {
      d_Dunnett = PROBA_Fractile_Student( 1.0 - alpha_individuel, ddl_SR );
    }
    break;

  case STAT_IC_FRACTILE_SIDAK:
  case STAT_IC_FRACTILE_SLEPIAN:
    /* Conduisent ici à la même formule ! */
    alpha_individuel = PROBA_INEG_Sidak( alpha_global, nombre_comparaisons );
    if ( 0 == sens ) {
      d_Dunnett = PROBA_Fractile_Student( 1.0 - alpha_individuel / 2.0, ddl_SR );
      /* Test bilatéral */
    } else {
      d_Dunnett = PROBA_Fractile_Student( 1.0 - alpha_individuel, ddl_SR );
    }
    break;

  case STAT_IC_FRACTILE_HW:
    OPTION_A_FAIRE( "Approximation avec inégalité de Hunter-Worsley" );
    break;

  case STAT_IC_FRACTILE_MILLERW:
    if ( VRAI == est_equilibree ) {
      OPTION_A_FAIRE( "Méthode de Miller-Winner, cas équilibré = cas exact !" );
    } else {
      OPTION_A_FAIRE( "Méthode de Miller-Winner, cas non équilibré !" );
    }
    break;

  default:
    AVERTISSEMENT( "Méthode à implémenter : %u", approximation );
    break;
  }

  /* Combien de valeurs dans la modalité contrôle ? [toujours la 2e des comparaison] */
  if ( NULL == comparaisons[ 0 ].modalites[ 1 ]->valeurs ) {
    AVERTISSEMENT( "Pas de valeurs dans la modalité témoin [%p]", NULL );
    return -1;
  } else {
    n_temoin = comparaisons[ 0 ].modalites[ 1 ]->valeurs->nombre_point;
  }

  /* Une fois que l'on a le d de Dunnett, on peut construire les IC de chaque différence */
  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    /* Combien de valeurs dans la modalité considérée ? */
    if ( NULL == comparaisons[ i ].modalites[ 0 ]->valeurs ) {
      AVERTISSEMENT( "Pas de valeurs dans la modalité %u --- calcul impossible", i );
      continue;
    }
    n_modalite = comparaisons[ i ].modalites[ 0 ]->valeurs->nombre_point;

    /* On calcule la variance de la différence [formule valable autant pour équilibré que non-équilibré */
    comparaisons[ i ].variance_difference = sqrt( variance_commune * ( 1.0 / n_modalite + 1.0 / n_temoin ) );

    /* On mémorise le "fractile" utilisé */
    comparaisons[ i ].fractile = d_Dunnett;

    /* Ainsi que la méthode de calcul & le sens */
    comparaisons[ i ].methode = approximation;
    comparaisons[ i ].sens = sens;

    /* On en déduit l'intervalle de confiance de cette différence */
    comparaisons[ i ].drapeaux |= STAT_CMP_AVEC_IC;
    STAT_CMP_CalculerIC( comparaisons + i );
  }

  /* Terminé sans souci */
  return 0;
}

/* -------- Calcul du d exact de Dunnett : valeur d pour laquelle p(T_multi<d)=1-alpha_global -------- */

double STAT_Fractile_Dunnett(double alpha_global, unsigned int nombre_comparaisons, struct comparaison *comparaisons, unsigned long ddl, int sens)
{
  double d0, d1, d2, d = 0.0;
  double alpha_Benferroni, alpha_Sidak;
  double p0, p1, p2, risque;
  double *facteurs, *x;
  unsigned int i;
  unsigned long n_controle, n_modalite;

  if ( NULL == comparaisons ) return 0.0;

  /* Vérifications préliminaires sur le contrôle */
  if ( NULL == comparaisons[ 0 ].modalites[ 1 ]->valeurs ) {
    AVERTISSEMENT( "Pas de valeurs dans la modalité témoin [%p]", NULL );
    return -1;
  } else {
    n_controle = comparaisons[ 0 ].modalites[ 1 ]->valeurs->nombre_point;
  }

  /* 1re estimation [supérieure] : inégalité de Bonferroni */
  /* 2e estimation  [supérieure] : inégalité de Sidak */
  /* 3e estimation  [inférieure] : gaussienne / risque global */
  alpha_Benferroni = PROBA_INEG_Bonferroni( alpha_global, nombre_comparaisons );
  alpha_Sidak = PROBA_INEG_Sidak( alpha_global, nombre_comparaisons );

  if ( 0 == sens ) {
    /* Test bilatéral */
    d0 = PROBA_Fractile_Student( 1.0 - alpha_Benferroni / 2.0, ddl );
    d1 = PROBA_Fractile_Student( 1.0 - alpha_Sidak / 2.0, ddl );

    alpha_global /= 2.0;
  } else {
    d0 = PROBA_Fractile_Student( 1.0 - alpha_Benferroni, ddl );
    d1 = PROBA_Fractile_Student( 1.0 - alpha_Sidak, ddl );
  }
  d2 = PROBA_Fractile_Gauss( 1.0 - alpha_global );

  /* Initialisation des tableaux */
  x = (double *) malloc( sizeof( double ) * nombre_comparaisons );
  if ( NULL == x ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mémoire insuffisante, dimension = %u\n",
		   __FILE__, __FUNCTION__, __LINE__, nombre_comparaisons );
    return 0.0;
  }

  facteurs = (double *) malloc( sizeof( double ) * nombre_comparaisons );
  if ( NULL == facteurs ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mémoire insuffisante, dimension = %u\n",
		   __FILE__, __FUNCTION__, __LINE__, nombre_comparaisons );
    free( x ); x = NULL;
    return 0.0;
  }

  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    if ( NULL == comparaisons[ i ].modalites[ 0 ]->valeurs ) {
      AVERTISSEMENT( "Pas de valeurs dans la modalité %u --- calcul impossible", i );
      free( facteurs ); facteurs = NULL;
      free( x ); x = NULL;
      return 0.0;
    }
    n_modalite = comparaisons[ i ].modalites[ 0 ]->valeurs->nombre_point;

    facteurs[ i ] = 1.0 / sqrt( 1.0 + ( (double) n_controle ) / ( (double) n_modalite ) );
  }

  /* Calcul des probabilités initiales */
  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    x[ i ] = d2;
  }
  p2 = PROBA_FdR_MultiStudent_Multiplicatif( nombre_comparaisons, x, facteurs, ddl );

  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    x[ i ] = d1;
  }
  p1 = PROBA_FdR_MultiStudent_Multiplicatif( nombre_comparaisons, x, facteurs, ddl );

  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    x[ i ] = d0;
  }
  p0 = PROBA_FdR_MultiStudent_Multiplicatif( nombre_comparaisons, x, facteurs, ddl );

  printf( "Encadrement initial : %g < d < %g < %g [%g < %g < %g < %g]\n",
	  d2, d1, d0, 1.0 - p0, 1.0 - p1, alpha_global, 1.0 - p2 );

  /* On travaille pour l'instant par dichotomie */
  OPTION_A_FAIRE( "Remplacer dichotomie par une méthode plus rapide !" );
  do {
    d = (d2 + d1) / 2.0;

    /* Calcul de la valeur pour ce d */
    for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
      x[ i ] = d;
    }
    risque = 1.0 - PROBA_FdR_MultiStudent_Multiplicatif( nombre_comparaisons, x, facteurs, ddl );
    if ( risque < alpha_global ) {
      /* a_C < a <=> 1 - a_C > 1 - a <=> dc > d : on est entre a = d, donc d1 [borne supérieure] devient d*/
      d1 = d;
    } else {
      d2 = d;
    }
/*     printf( "[%g,%g] d=%g ==> risque = %g\n", d2, d1, d, risque ); */
  } while ( ( fabs( risque - alpha_global ) > 0.001 ) || ( ( d1 - d2 ) > 1e-4 ) );  /* Une précision de 1/100 suffit _sur le risque_ */

  /* On en déduit d : milieu de l'intervalle */
  d = (d1 + d2) / 2.0;

  /* Calcul de la valeur pour ce d */
  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    x[ i ] = d;
  }
  risque = 1.0 - PROBA_FdR_MultiStudent_Multiplicatif( nombre_comparaisons, x, facteurs, ddl );

  printf( "Encadrement trouvé : %g < d < %g [risque : %g, voulu = %g] => d ~ %g\n",
	  d2, d1, risque, alpha_global, d );

  /* Plus besoin des tableaux */
  free( facteurs ); facteurs = NULL;
  free( x ); x = NULL;

  return d;
}
