/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'un spectre XAS par transmission
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "extrait.h" /* Pour MU_Fluorescence */
#include "experience.h"
#include "graphes.h"
#include "operations.h" /* Pour GRF_Integre */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int XAS_Fluo_I0Ct_PrjCte(Experience *experience, Graphe *grf_i0, Graphe *grf_i1);
static int XAS_Fluo_PrjCte     (Experience *experience, Graphe *grf_i0, Graphe *grf_i1);
static int XAS_Fluo_Quelconque(Experience *experience, Graphe *grf_i0, Graphe *grf_i1);

static double FLUO_CalculerSignal(Experience *experience, Faisceau *faiceau, double energie, Faisceau *f_transmis);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Calcul d'un spectre de fluorescence complet [appelable extrieurement] -------- */

Graphe *EXP_SpectreXAS_Fluorescence( Experience *experience )
{
  int retour = -1;
  boolean i0_constant = VRAI; /* Idalement, il n'y a pas besoin de recalculer I0  chaque pas */
  boolean proj_constante = VRAI; /* Mme chose, mais pour la projection du faisceau sur l'chantillon */
  Graphe *spectre, *grf_i0, *grf_i1;

  ENTREE_FONCTION("EXP_SpectreXAS_Fluorescence", "experience=%p", experience);

  /* Vrifications */
  if ( NULL == experience ) {
    SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p [ERREUR : experience==NULL]", NULL);
    return NULL;
  }
  if ( experience->E_fin <= experience->E_debut ) {
    (void) fprintf( stderr, "EXP_SpectreXAS_Fluorescence : intervalle de mesure en nergie incorrect [%g -> %g eV]\n",
		    experience->E_debut, experience->E_fin );
    SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p [ERREUR : E_fin <= E_debut]", NULL);
    return NULL;
  }
  if ( NULL == experience->i0 ) {
    SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p [ERREUR : pas de dtecteur I0]", NULL);
    return NULL;
  }
  if ( NULL == experience->i1 ) {
    SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p [ERREUR : pas de dtecteur I1]", NULL);
    return NULL;
  }

  /* Premire tape : construire les spectres */
  grf_i0 = GRF_nv_GrilleXP(NULL, GR_I0, GR_NORMAL, "Signal I0", FAUX,
			   experience->E_debut, experience->E_fin, experience->pas_E );
  if ( NULL == grf_i0 ) {
    (void) fprintf(stderr, "EXP_SpectreXAS_Fluorescence : ERREUR dans GRF_nv_GrilleXP() / graphe I0\n");
    SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p [ERREUR : grf_i0]", NULL);
    return NULL;
  }
  grf_i1 = GRF_nv_GrilleXP(grf_i0, GR_I1, GR_NORMAL, "Signal I fluoresc", FAUX,
			   experience->E_debut, experience->E_fin, experience->pas_E );
  if ( NULL == grf_i1 ) {
    (void) fprintf(stderr, "EXP_SpectreXAS_Fluorescence : ERREUR dans GRF_nv_GrilleXP() / graphe I1\n");
    GRF_efface( grf_i0, VRAI, VRAI );
    SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p [ERREUR : grf_i1]", NULL);
    return NULL;
  }

  /* Faut-il recalculer le i0 en chaque point ? */
  if ( experience->faisceau->proprietes & FAISCEAU_DECROIT ) {
    i0_constant = FAUX; /* Il faut, car l'intensit du faisceau change avec l'nergie */
    proj_constante = FAUX; /* La projection du faisceau sur l'chantillon est donc modifie */
  }
  if ( experience->i0->proprietes & DETECTEUR_ABSORBANT ) {
    i0_constant = FAUX; /* Il faut, car le dtecteur ne rpond alors pas pareil pour chaque nergie */
    proj_constante = FAUX; /* La projection du faisceau sur l'chantillon est donc modifie */
  }
  if ( experience->i0->proprietes & DETECTEUR_NRJ_VARIABLE ) {
    i0_constant = FAUX; /* Il faut, car le dtecteur ne rpond pas pareil pour chaque nergie */
  }

  if ( ( VRAI == i0_constant ) && ( VRAI == proj_constante ) ) {
    retour = XAS_Fluo_I0Ct_PrjCte( experience, grf_i0, grf_i1 );
    /* On calcule le signal sur I0, et son effet sur le faisceau */
  } else if ( VRAI == proj_constante ) {
    retour = XAS_Fluo_PrjCte( experience, grf_i0, grf_i1 );
  } else {
    retour = XAS_Fluo_Quelconque( experience, grf_i0, grf_i1 );
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le calcul du signal\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Maintenant que l'on connat I0 et I1, on peut en dduire le mu observ */
  spectre = MU_Fluorescence( grf_i0, grf_i1 ); /* S'occupe de tout, y compris les erreurs si ncessaire ! */
  if ( NULL == spectre ) {
    (void) fprintf(stderr, "EXP_SpectreXAS_Fluorescence : ERREUR dans MU_Fluorescence \n");
    spectre = grf_i1; /* On renvoie alors I1 tout simplement... */
  }

  /* On a fini, tout s'est bien pass */
  SORTIE_FONCTION("EXP_SpectreXAS_Fluorescence", "%p", spectre);
  return spectre;
}

/* -------- Cas le plus simple : calcul en fluorescence
 * Le courant I0 ne varie pas avec l'nergie
 * La projection du faisceau sur l'chantillon ne varie pas avec l'nergie
 * ------------------- */

static int XAS_Fluo_I0Ct_PrjCte(Experience *experience, Graphe *grf_i0, Graphe *grf_i1)
{
  Faisceau *f_echantillon, *f_apres_i0, *f_detecteur;
  double i0, *energie, *i1;
  unsigned long i;

  ENTREE_FONCTION("XAS_Fluo_I0Ct_PrjCte", "experience=%p grf_i0=%p grf_i1=%p", experience, grf_i0, grf_i1);

  /* On calcule I0 : copie de l'original => il faudra l'effacer, quoi qu'il arrive ! */
  f_apres_i0 = FAISCEAU_Copier( experience->faisceau );
  if ( NULL == f_apres_i0 ) {
    (void) fprintf(stderr, "XAS_Fluo_I0Ct_PrjCte : impossible de crer le faisceau aprs I0\n");
    SORTIE_FONCTION("XAS_Fluo_I0Ct_PrjCte", "%d [ERREUR faisceau aprs I0]", -1);
    return -1;
  }
  i0 = DETECTEUR_FaisceauTraverse_I(experience->i0, experience->faisceau, grf_i0->x[ 0 ], f_apres_i0 );

  /* Valable  toutes les nergies */
  for ( i = 0 ; i < grf_i0->nombre_point ; i++ ) {
    grf_i0->y[ i ] = i0;
  }
  grf_i0->y_min = i0; grf_i0->y_max = i0;

  /* Et on projette cela sur l'chantillon */
  f_echantillon = FAISCEAU_ProjeterSurEchantillon( f_apres_i0, experience->echantillon );
  if ( NULL == f_echantillon ) {
    (void) fprintf(stderr, "XAS_Fluo_I0Cte_PrjCte : ERREUR dans FAISCEAU_ProjeterSurEchantillon()\n");
    FAISCEAU_Effacer( f_apres_i0 );
    SORTIE_FONCTION("XAS_Fluo_I0Ct_PrjCte", "%d [ERREUR faisceau avant chantillon]", -2);
    return -2;
  }
  if ( f_echantillon != f_apres_i0 ) { /* On n'a plus besoin de lui */
    FAISCEAU_Effacer( f_apres_i0 );
  }
  f_apres_i0 = NULL; /* Soit on l'a effac, soit il est gal  f_echantillon => inutile ! */

  /* On construit, ensuite, le "squelette" du faisceau peru par le dtecteur de fluo. */
  f_detecteur = FAISCEAU_GeometrieFluorescence( experience->i1, experience->table_X_fluo, experience->nbr_X_fluo );
  if ( NULL == f_detecteur ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_GeometrieFluorescence\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( f_echantillon );
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR faisceau fluorescence]", -3);
    return -3;
  }

  /* On calcule le signal mesur par I1 pour chaque nergie */
  for ( i = 0, energie = grf_i1->x, i1 = grf_i1->y ; i < grf_i1->nombre_point ; i++, energie++, i1++ ) {
#ifdef _DEVELOPPE_
    if ( *energie <= 0.0 ) {
      (void) fprintf(stderr, "XAS_Fluo_I0Ct_PrjCte : ERREUR - nergie ngative [i=%lu]\n", i);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }
#endif
    *i1 = FLUO_CalculerSignal( experience, f_echantillon, *energie, f_detecteur );
    printf( "%s : E=%g eV\r", __FUNCTION__, *energie ); fflush( stdout );
/*     break; */
  }

  /* Bornes... */
  GRF_bornes( grf_i1, NE_PAS_FAIRE, DESORDONNE );

  /* Plus besoin du faisceau / dtecteur, ni du faisceau chantillon */
  FAISCEAU_Effacer( f_detecteur );
  FAISCEAU_Effacer( f_echantillon );

  /* On a fini, tout s'est bien pass */
  SORTIE_FONCTION("XAS_Fluo_I0Ct_PrjCte", "%d", 0);
  return 0;
}


/* -------- Cas intermdiaire : I0 change, mais pas le faisceau sur l'chantillon -------- */

static int XAS_Fluo_PrjCte(Experience *experience, Graphe *grf_i0, Graphe *grf_i1)
{
  Faisceau *f_echantillon, *f_apres_i0, *f_detecteur;
  double *energie, *i1, *i0;
  unsigned long i;

  ENTREE_FONCTION( __FUNCTION__, "experience=%p grf_i0=%p grf_i1=%p", experience, grf_i0, grf_i1 );

  /* Prparation des faisceaux intermdiaires */
  /* 1. En sortie de I0 : copie de l'original => il faudra l'effacer, quoi qu'il arrive ! */
  f_apres_i0 = FAISCEAU_Copier( experience->faisceau );
  if ( NULL == f_apres_i0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le faisceau aprs I0\n",
		   __FILE__, __FUNCTION__, __LINE__);
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR copie faisceau]", -1);
    return -1;
  }

  /* 2. Sur l'chantillon */
  f_echantillon = FAISCEAU_ProjeterSurEchantillon( f_apres_i0, experience->echantillon );
  if ( NULL == f_echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_ProjeterSurEchantillon()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( f_apres_i0 ); /* Plus besoin de lui */
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR faisceau chantillon]", -2);
    return -2;
  }
  if ( f_echantillon != f_apres_i0 ) {
    FAISCEAU_Effacer( f_apres_i0 ); /* Plus besoin de lui */
  }
  f_apres_i0 = NULL; /* Soit on l'a effac, soit il est gal  f_echantillon => inutile ! */

  /* 3. "Squelette" du faisceau peru par le dtecteur de fluo. */
  f_detecteur = FAISCEAU_GeometrieFluorescence( experience->i1, experience->table_X_fluo, experience->nbr_X_fluo );
  if ( NULL == f_detecteur ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_GeometrieFluorescence\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( f_echantillon ); /* f_echantillon <= f_apres_i0, forcment diffrent de exp->faisceau car copie */
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR faisceau fluorescence]", -3);
    return -3;
  }

  /* En chaque point du spectre... */
  for ( i = 0, energie = grf_i1->x, i0 = grf_i0->y, i1 = grf_i1->y ; i < grf_i1->nombre_point ; i++, energie++, i0++, i1++ ) {
    /* Signal avant l'chantillon */
    *i0 = DETECTEUR_FaisceauTraverse_I(experience->i0, experience->faisceau, *energie, NULL );

    /* Signal aprs l'chantillon */
    *i1 = FLUO_CalculerSignal( experience, f_echantillon, *energie, f_detecteur );

#ifdef _DEVELOPPE_
    printf("%s : E=%g I0=%g I1=%g\n", __FUNCTION__, *energie, *i0, *i1);
#endif
  }

  /* Plus besoin d'eux */
  FAISCEAU_Effacer( f_detecteur );
  FAISCEAU_Effacer( f_echantillon );

  /* Tout s'est bien pass */
  GRF_bornes( grf_i0, NE_PAS_FAIRE, DESORDONNE );
  GRF_bornes( grf_i1, NE_PAS_FAIRE, DESORDONNE );

  SORTIE_FONCTION( __FUNCTION__, "%d", 0);
  return 0;
}

/* -------- Cas le plus gnral : il faut tout refaire  chaque fois -------- */

static int XAS_Fluo_Quelconque(Experience *experience, Graphe *grf_i0, Graphe *grf_i1)
{
  Faisceau *f_echantillon, *f_apres_i0, *f_detecteur = NULL;
  double *energie, *i1, *i0;
  unsigned long i;

  ENTREE_FONCTION( __FUNCTION__, "experience=%p grf_i0=%p grf_i1=%p", experience, grf_i0, grf_i1 );

  /* Prparation des faisceaux intermdiaires */
  /* 1. En sortie de I0 : copie de l'original => il faudra l'effacer, quoi qu'il arrive ! */
  f_apres_i0 = FAISCEAU_Copier( experience->faisceau );
  if ( NULL == f_apres_i0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le faisceau aprs I0\n",
		   __FILE__, __FUNCTION__, __LINE__);
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR copie faisceau]", -1);
    return -1;
  }

  /* Squelette du faisceau sur le dtecteur */
  f_detecteur = FAISCEAU_GeometrieFluorescence( experience->i1, experience->table_X_fluo, experience->nbr_X_fluo );
  if ( NULL == f_detecteur ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_GeometrieFluorescence\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( f_apres_i0 );
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR faisceau fluorescence]", -3);
    return -3;
  }

  /* En chaque point du spectre... */
  for ( i = 0, energie = grf_i1->x, i0 = grf_i0->y, i1 = grf_i1->y ; i < grf_i1->nombre_point ; i++, energie++, i0++, i1++ ) {
    /* Signal avant l'chantillon */
    *i0 = DETECTEUR_FaisceauTraverse_I(experience->i0, experience->faisceau, *energie, f_apres_i0 );

    /* Et on projette cela sur l'chantillon */
    f_echantillon = FAISCEAU_ProjeterSurEchantillon( f_apres_i0, experience->echantillon );
    if ( NULL == f_echantillon ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_ProjeterSurEchantillon()\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FAISCEAU_Effacer( f_detecteur );
      FAISCEAU_Effacer( f_apres_i0 );
      SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR faisceau projet]", -2);
      return -2;
    }

    /* Signal aprs l'chantillon */
    *i1 = FLUO_CalculerSignal( experience, f_echantillon, *energie, f_detecteur );

    /* On efface pour le recrer correctement  l'nergie suivante... */
    if ( f_echantillon != f_apres_i0)
      FAISCEAU_Effacer( f_echantillon );

#ifdef _DEVELOPPE_
    printf("%s : E=%g I0=%g I1=%g\n", __FUNCTION__, *energie, *i0, *i1);
#endif
  }

  /* Plus besoin d'eux */
  FAISCEAU_Effacer( f_detecteur );
  FAISCEAU_Effacer( f_apres_i0 );

  /* Tout s'est bien pass */
  GRF_bornes( grf_i0, NE_PAS_FAIRE, DESORDONNE );
  GRF_bornes( grf_i1, NE_PAS_FAIRE, DESORDONNE );

  SORTIE_FONCTION( __FUNCTION__, "%d", 0);
  return 0;
}

/* -------- Calculer le signal de fluorescence -------- */

static double FLUO_CalculerSignal(Experience *experience, Faisceau *faisceau, double energie, Faisceau *f_detecteur)
{
  int retour;
  double i_fluo;
  Graphe *spectre_fluo;

  ENTREE_FONCTION("FLUO_CalculerSignal", "experience=%p faisceau=%p energie=%g f_detecteur=%p", experience, faisceau, energie, f_detecteur);

  /* 1. On remplit l'chantillon avec le nombre _total_ de photons absorbs par l'chantillon */
  retour = FLUO_InitialiserEchantillon( experience->echantillon, faisceau, energie );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FLUO_CalculerSignal : ERREUR dans FLUO_InitialiserEchantillon()\n");
    SORTIE_FONCTION("FLUO_CalculerSignal", "%g [ERREUR %d dans FLUO_InitialiserEchantillon]", -1.0, retour);
    return -1.0;
  }

  /* 2. On calcule le spectre de fluo _total_ vu par le dtecteur, suppos rsolu */
  spectre_fluo = FLUO_CalculerSpectre( experience, f_detecteur, energie );
  if ( NULL == spectre_fluo ) {
    (void) fprintf(stderr, "FLUO_CalculerSignal : ERREUR dans FLUO_CalculerSpectre\n");
    SORTIE_FONCTION("FLUO_CalculerSignal", "%g [ERREUR dans FLUO_CalculerSpectre]", -2.0);
    return -2.0;
  }

  /* 3. On intgre ce spectre pour la ROI considre */
  i_fluo = GRF_Integrer_R( spectre_fluo, experience->ROI_debut, experience->ROI_fin, INTG_TRAPEZE, NULL );

  SORTIE_FONCTION("FLUO_CalculerSignal", "%g", i_fluo);
  return i_fluo;
}
