/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'une description de configuration exprimentale
   Construction du spectre rsultant de cette description
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "absorption.h"
#include "experience.h"
#include "graphes.h" /* Pour GRF_eff_isole */
#include "images.h"
#include "theorie.h"
#include "utils_tables.h"

#include "th_experience.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int EXP_PreparerTables(Experience *experience);
static int EXP_InitialiserExperience(Experience *experience);
static int EXP_ControlerOptions(Experience *experience);
static int EXP_CalculerMu(Experience *experience, struct matiere *matiere);

/* void exporter(Graphe *, const char *); */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Manipulation de la structure ~-~-~-~-~-~-~-~-~ */

Experience *EXP_CreerExperience(int methode)
{
  Experience *experience;

  /* On rserve la mmoire ncessaire */
  experience = (Experience *) calloc( 1, sizeof( Experience ) );
  if ( NULL == experience ) {
    (void) fprintf(stderr, "EXP_CreerExperience : Mmoire insuffisante\n");
    return NULL;
  }

  /* On initialise [calloc met  zro => seulement le reste]
     Tous les pointeurs sont  NULL, l'exprience n'absorbe pas,... */
  experience->methode = methode;

  /* Pb : rien ne prouve que 0.0 soit cod avec 0 seulement, mme si c'est le plus frquent */
  experience->E_debut = 0.0;
  experience->E_fin = 0.0;
  experience->pas_E = 0.0;

  experience->Ef_debut = 0.0;
  experience->Ef_fin = 0.0;
  experience->pas_Ef = 0.0;

  experience->ROI_debut = 0.0;
  experience->ROI_fin = 0.0;

  experience->x_min = -5.0;  experience->y_min = -5.0;
  experience->x_max = +5.0;  experience->y_max = +5.0;
  experience->pas_x =  1.0;  experience->pas_y =  1.0;

  /* On a fini, tout s'est bien pass */
  return experience;
}

/* -------- Renvoie VRAI si la mthode demande un balayage en nergie -------- */

boolean EXP_EstSpectroscopie(int methode)
{
  if ( EXPERIENCE_XAS_TRANSMISSION == methode ) return VRAI;
  if ( EXPERIENCE_XAS_FLUORESCENCE == methode ) return VRAI;

  /* Autres mthodes : une seule nergie d'excitation */
  return FAUX;
}

/* -------- Renvoie VRAI si la mthode demande le suivi de la fluorescence --- */

boolean EXP_UtiliseFluorescence(Experience *experience)
{
  if ( NULL == experience ) return FAUX;

  if ( EXPERIENCE_XAS_TRANSMISSION == experience->methode ) {
    if ( experience->ordre > 0 ) return VRAI; /* On tient compte du rayonnement fluoresc qui arrive sur les dtecteurs */

    return FAUX;
  }

  if ( EXPERIENCE_XAS_FLUORESCENCE == experience->methode ) return VRAI;
  if ( EXPERIENCE_XRF == experience->methode ) return VRAI;
  if ( EXPERIENCE_XRF_TOMOGRAPHIE == experience->methode ) return VRAI;
  if ( EXPERIENCE_PIXE == experience->methode ) return VRAI;

  (void) fprintf(stderr, "EXP_UtiliseFluorescence : Mthode inconnue [%d]\n", experience->methode);
  return FAUX;
}

/* ~-~-~-~-~-~-~-~- Construction du spectre obtenu par l'exprience crite ~-~-~-~-~-~-~-~- */

/* -------- Construction pas  pas -------- */

Graphe *EXP_ConstruireSpectre( Experience *experience )
{
  int retour;
  struct graphe *spectre = NULL;

  ENTREE_FONCTION("EXP_ConstruireSpectre", "experience=%p", experience);

  retour = EXP_ControlerOptions( experience );
  if ( retour != 0 ) { 
    (void) fprintf(stderr, "EXP_ConstruireSpectre : ERREUR dans EXP_ControlerOptions() [%d]\n", retour);
    SORTIE_FONCTION("EXP_ConstruireSpectre", "%p [ERREUR : options incohrentes, code %d]", NULL, retour);
    return NULL;
  }

  retour = EXP_InitialiserExperience( experience );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_ConstruireSpectre : ERREUR dans EXP_InitialiserExperience\n");
    SORTIE_FONCTION("EXP_ConstruireSpectre", "%p [ERREUR]", NULL);
    return NULL;
  }

  switch ( experience->methode )
  {
  case EXPERIENCE_XAS_TRANSMISSION:
    spectre = EXP_SpectreXAS_Transmission( experience );
    break;

  case EXPERIENCE_XAS_FLUORESCENCE:
    spectre = EXP_SpectreXAS_Fluorescence( experience );
    break;

  case EXPERIENCE_XRF:
    spectre = EXP_SpectreXRF( experience, NULL );
    break;

  default:
    (void) fprintf(stderr, "EXP_ConstruireSpectre : Technique inconnue [%d]\n", experience->methode);
    break;
  }
/*   exporter( spectre, "apres_modelisation.asc" ); */

  /* On a fini */
  SORTIE_FONCTION("EXP_ConstruireSpectre", "%p", spectre);
  return spectre;
}

/* -------- Construire une cartographie de l'chantillon -------- */

struct image *EXP_ConstruireCartographie(Experience *experience)
{
  Faisceau *faisceau;
  Graphe *spectre = NULL;
  Image *cartographie;
  Orientation3D orientation_originale;
  int retour;
  unsigned int lg_carto, ht_carto, i_lignes, i_colonnes;
  unsigned long idx_image;
  double x_reel, y_reel;

  if ( FAUX == ( experience->proprietes & EXPERIENCE_CARTOGRAPHIE ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Appel alors que l'on ne veut pas une cartographie !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On valide les options choisies */
  retour = EXP_ControlerOptions( experience );
  if ( retour != 0 ) { 
    (void) fprintf(stderr, "EXP_ConstruireSpectre : ERREUR dans EXP_ControlerOptions() [%d]\n", retour);
    SORTIE_FONCTION("EXP_ConstruireSpectre", "%p [ERREUR : options incohrentes, code %d]", NULL, retour);
    return NULL;
  }

  /* On prpare tout pour l'exprience */
  retour = EXP_InitialiserExperience( experience );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_ConstruireSpectre : ERREUR dans EXP_InitialiserExperience\n");
    SORTIE_FONCTION("EXP_ConstruireSpectre", "%p [ERREUR]", NULL);
    return NULL;
  }

  /* On prpare l'image qui va contenir la cartographie [un spectre par point !] */
  lg_carto = 1;
  if ( experience->pas_x > 0 ) {
    lg_carto += (unsigned int) rint( experience->x_max - experience->x_min ) / ( experience->pas_x );
  }

  ht_carto = 1;
  if ( experience->pas_y > 0 ) {
    ht_carto += (unsigned int) rint( experience->y_max - experience->y_min ) / ( experience->pas_y );
  }

  cartographie = IMG_CreerImage( IMG_serie_courante, NULL, IMG_GENRE_GRAPHE, IMG_TYPE_CARTOGRAPHIE, lg_carto, ht_carto );
  if ( NULL == cartographie ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de crer l'image contenant le rsultat de la cartographie !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  cartographie->l_reel = experience->pas_x; cartographie->x_min = experience->x_min;
  cartographie->h_reel = experience->pas_y; cartographie->y_min = experience->y_min;

  /* On initialise le faisceau */
  faisceau = experience->faisceau;
  orientation_originale = faisceau->orientation;

  faisceau->orientation.position.y = 0.0;  /* On se repre  la surface de l'chantillon ! */

  /* On dplace le faisceau sur l'chantillon */
  for ( i_lignes = 0 ; i_lignes < ht_carto ; i_lignes++ ) {
    y_reel = experience->y_min + i_lignes * experience->pas_y;
    faisceau->orientation.position.z = y_reel;

    for ( i_colonnes = 0 ; i_colonnes < lg_carto ; i_colonnes++ ) {
      x_reel = experience->x_min + i_colonnes * experience->pas_x;
      faisceau->orientation.position.x = x_reel;

      switch ( experience->methode )
      {
      case EXPERIENCE_XAS_TRANSMISSION:
	spectre = EXP_SpectreXAS_Transmission( experience );
	break;

      case EXPERIENCE_XAS_FLUORESCENCE:
	spectre = EXP_SpectreXAS_Fluorescence( experience );
	break;

      case EXPERIENCE_XRF:
	spectre = EXP_SpectreXRF( experience, NULL );
	break;

      default:
	(void) fprintf(stderr, "EXP_ConstruireSpectre : Technique inconnue [%d]\n", experience->methode);
	break;
      }

      if ( NULL == spectre ) {
	(void) fprintf(stderr, "%s,%s l.%u : spectre = NULL comme rsultat de simulation !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	printf( "  --- se produit pour I=%u, J=%u soit en (%g, %g)\n", i_colonnes, i_lignes, x_reel, y_reel );
	/* Se produit parfois sur les bords de l'chantillon : il faudrait l'amliorer, mais problmes d'arrondi un peu partout... */
/* 	FIN_ANTICIPEE( FIN_ERREUR_MINEURE ); */
      }

      /* On mmorise le spectre  sa position dans l'image */
      (void) IMG_IJ_v_N( cartographie, i_colonnes, i_lignes, &idx_image );
      cartographie->valeurs.G[ idx_image ] = spectre;
    }
  }

  /* On remet le faisceau  sa position initiale */
  faisceau->orientation = orientation_originale;

  /* Termin */
  return cartographie;
}

/* -------- Contrle des options utilises -------- */

static int EXP_ControlerOptions(Experience *experience)
{
  int retour;

  if ( NULL == experience ) return -1;

  /* -- Contrle des options de l'exprience elle-mme -- */
  /* 1. composition [si absorbante, sinon info. inutile] */
  if ( experience->proprietes & EXPERIENCE_ABSORBANTE ) {
    retour = Matiere_ControlerComposition( experience->composition );
    if ( retour != 0 ) {
      (void) BI_ERR_alerte( NULL, &ERR_SansComposition, VRAI );
      (void) fprintf(stderr, "EXP_ControlerOptions : ERREUR dans Matiere_ControlerComposition() [code %d]\n", retour);
      return -2;
    }
  }

  /* 2. Intervalle d'excitation [si spectro. uniquement] */
  if ( VRAI == EXP_EstSpectroscopie( experience->methode ) ) {
    if ( experience->E_debut >= experience->E_fin ) {
      (void) fprintf(stderr, "EXP_ControlerOptions : Bornes absurdes [min=%g, max=%g]\n", experience->E_debut, experience->E_fin);
      return -3;
    }
    if ( experience->pas_E >= ( experience->E_fin - experience->E_debut ) ) {
      (void) fprintf(stderr, "EXP_ControlerOptions : Pas absurde [pas=%g, intervalle=%g]\n", experience->pas_E, experience->E_fin - experience->E_debut);
      return -3;
    }
  }

  /* 3. Intervalle de mesure / fluorescence */
  if ( VRAI == EXP_UtiliseFluorescence( experience ) ) {
    if ( experience->Ef_debut >= experience->Ef_fin ) {
      (void) fprintf(stderr, "EXP_ControlerOptions : Bornes absurdes, fluorescence [min=%g, max=%g]\n", experience->Ef_debut, experience->Ef_fin);
      return -4;
    }
    if ( experience->pas_Ef >= ( experience->Ef_fin - experience->Ef_debut ) ) {
      (void) fprintf(stderr, "EXP_ControlerOptions : Pas absurde, fluorescence [pas=%g, intervalle=%g]\n",
		     experience->pas_Ef, experience->Ef_fin - experience->Ef_debut);
      return -4;
    }

    if ( experience->ROI_debut >= experience->ROI_fin ) {
      (void) fprintf(stderr, "EXP_ControlerOptions : Bornes absurdes, ROI [min=%g, max=%g]\n", experience->ROI_debut, experience->ROI_fin);
      return -5;
    }
    if ( ( experience->ROI_fin <= experience->Ef_debut ) || ( experience->ROI_debut >= experience->Ef_fin ) ) {
      (void) fprintf(stderr, "EXP_ControlerOptions : ROI hors de la rgion calcule\n");
      return -5;
    }
  }

  /* 4. Zone cartographie, si cartographie */
  if ( experience->proprietes & EXPERIENCE_CARTOGRAPHIE ) {
    if ( ( experience->x_min > experience->x_max ) || ( experience->y_min > experience->y_max ) ) {
      (void) fprintf(stderr, "%s, %s l.%u : Intervalle cartographi absurde [X : %g  %g --- Y : %g  %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, experience->x_min, experience->x_max, experience->y_min, experience->y_max );
      return -6;
    }

    if ( ( ( experience->pas_x < 0 ) || ( experience->pas_x > ( experience->x_max - experience->x_min ) ) ) ||
	 ( ( experience->pas_y < 0 ) || ( experience->pas_y > ( experience->y_max - experience->y_min ) ) ) ) {
      (void) fprintf(stderr, "%s, %s l.%u : Pas de cartographie absurde [X : %g, Y : %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, experience->pas_x, experience->pas_y );
      return -6;
    }
  }

  /* -- Contrle des options du faisceau incident -- */
  if ( NULL == experience->faisceau ) {
    (void) BI_ERR_alerte( NULL, &ERR_SansFaisceau, VRAI );
    (void) fprintf(stderr, "EXP_ControlerOptions : Pas de faisceau\n");
    return -2;
  }
  retour = FAISCEAU_ControlerOptions( experience->faisceau, experience->E_debut );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_ControlerOptions : ERREUR dans FAISCEAU_ControlerOptions() [code %d]\n", retour);
    return -10;
  }

  /* -- Contrle des options de l'chantillon -- */
  if ( NULL == experience->echantillon ) {
    (void) BI_ERR_alerte( NULL, &ERR_SansEchantillon, VRAI );
    (void) fprintf(stderr, "EXP_ControlerOptions : Pas d'chantillon\n");
    return -4;
  }
  retour = ECH_ControlerOptions( experience->echantillon );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_ControlerOptions : ERREUR dans ECH_ControlerOptions() [code %d]\n", retour);
    return -10;
  }

  /* -- Contrle des options du dtecteur I0 -- */
  if ( VRAI == EXP_EstSpectroscopie( experience->methode ) ) {
    retour = DET_ControlerOptions( experience->i0 );
  } else if ( NULL != experience->i0 ) { /* Pas obligatoire, mais s'il est l, on le contrle */
    retour = DET_ControlerOptions( experience->i0 );
  } /* Sinon, retour vaut 0 sinon on aurait quitt avant -> tout va bien */
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_ControlerOptions : ERREUR dans DET_ControlerOptions(), I0 [code %d]\n", retour);
    return -11;
  }

  /* -- Contrle des options du dtecteur I1/If, obligatoire -- */
  retour = DET_ControlerOptions( experience->i1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_ControlerOptions : ERREUR dans DET_ControlerOptions(), I1 [code %d]\n", retour);
    return -12;
  }

  /* Tous les contrles ont t passs correctement */
  return 0;
}

/* -------- Prparations ncessaires  la simulation -------- */

static int EXP_InitialiserExperience(Experience *experience)
{
  Echantillon *echantillon;
  int retour;
  struct matiere *composition;

  /* Prparation des tables de X qui seront utilises dans les calculs ultrieurs */
  retour = EXP_PreparerTables( experience );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_InitialiserExperience : ERREUR dans EXP_PreparerTables() [%d]\n", retour);
    return -1;
  }

  echantillon = experience->echantillon;

  /* Calcul des coefficients d'absorption avec l'nergie */
  /* 1. Pour l'chantillon */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    /* On construit le spectre d'absorption */
    retour = EXP_CalculerMu( experience, echantillon->valeurs.homogene.composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "EXP_InitialiserExperience : ERREUR dans EXP_CalculerMu() [code %d]\n", retour);
      return -2;
    }
  } else {
    if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      /* On construit le spectre d'absorption */
      retour = EXP_CalculerMu( experience, echantillon->valeurs.homogene.composition );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "EXP_InitialiserExperience : ERREUR dans EXP_CalculerMu() [code %d]\n", retour);
	return -2;
      }
    } else {
      /* On calcule celui de chaque composition, pour viter d'en oublier */
      for ( composition = premiere_composition ; composition != NULL ; composition = composition->suivante ) {
	retour = EXP_CalculerMu( experience, composition );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans EXP_CalculerMu(,%s) [code %d]\n",
			 __FILE__, __FUNCTION__, __LINE__, composition->nom, retour);
	  return -2;
	}
      }
    }
  }

  /* 2. Exprience absorbante : on calcule aussi les sections efficaces correspondantes */
  if ( experience->proprietes & EXPERIENCE_ABSORBANTE ) {
    retour = EXP_CalculerMu( experience, experience->composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "EXP_InitialiserExperience : ERREUR dans EXP_CalculerMu() / exp. absorbante [code %d]\n", retour);
      return -2;
    }
  }

  /* 3. Dtecteurs absorbants ? */
  if ( experience->i0->proprietes & DETECTEUR_ABSORBANT ) {
    retour = EXP_CalculerMu( experience, experience->i0->composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "EXP_InitialiserExperience : ERREUR dans EXP_CalculerMu() / I0 absorbante [code %d]\n", retour);
      return -2;
    }
  }
  if ( experience->i1->proprietes & DETECTEUR_ABSORBANT ) {
    retour = EXP_CalculerMu( experience, experience->i1->composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "EXP_InitialiserExperience : ERREUR dans EXP_CalculerMu() / I1 absorbante [code %d]\n", retour);
      return -2;
    }
  }

  /* Calcul des spectres de fluorescence, si ncessaire */
  if ( VRAI == EXP_UtiliseFluorescence( experience ) ) {
    /* 0. On efface tous les graphes de fluo. pouvant exister dans les compo. existantes */
    for ( composition = premiere_composition ; composition != NULL ; composition = composition->suivante ) {
      GRF_eff_isole( composition->fluo );
      composition->fluo = NULL;
    }

    /* 1. On cre les spectres de fluo. pour chaque composition */
    if ( ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) ||
	 ( ECHANTILLON_GRADIENT == echantillon->type ) ) {
      composition = echantillon->valeurs.homogene.composition;
      composition->fluo = GRF_nv_TableX( NULL, GR_FLUOX, GRF_SERIE, NULL, VRAI,
					 experience->table_X_fluo, experience->nbr_X_fluo, FAUX );
      if ( NULL == composition->fluo ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR en crant la fluorescence pour la composition %s\n",
		       __FILE__, __FUNCTION__, __LINE__, composition->nom);
	return -12;
      }
    } else {
      /* L'idal serait de le faire juste pour les compositions utilises... */
      for ( composition = premiere_composition ; composition != NULL ; composition = composition->suivante ) {
	composition->fluo = GRF_nv_TableX( NULL, GR_FLUOX, GRF_SERIE, NULL, VRAI,
					   experience->table_X_fluo, experience->nbr_X_fluo, FAUX );
	if ( NULL == composition->fluo ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR en crant la fluorescence pour la composition %s\n",
			 __FILE__, __FUNCTION__, __LINE__, composition->nom);
	  return -13;
	}
      }
    }
  }

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Calculs prliminaires  la simulation ~-~-~-~-~-~-~-~-~ */

/* -------- Prparation des tables des X : seront communes  plein de graphes ensuite ! -------- */

static int EXP_PreparerTables(Experience *experience)
{
  double **tables = NULL, *x_excitation;
  unsigned int i;
  unsigned long *nbr = NULL, nbr_tables = 1, n_excitation;

  /* On libre d'ventuelles anciennes tables */
  free( experience->table_X_mu   ); experience->table_X_mu = NULL; experience->nbr_X_mu = 0;
  free( experience->table_X_fluo ); experience->table_X_fluo = NULL; experience->nbr_X_fluo = 0;

  /* La table de fluorescence, si ncessaire */
  if ( VRAI == EXP_UtiliseFluorescence( experience ) ) {
    experience->table_X_fluo = TABLE_CreerTable_mMp( experience->Ef_debut, experience->Ef_fin, experience->pas_Ef, &(experience->nbr_X_fluo) );
    if ( NULL == experience->table_X_fluo ) {
      (void) fprintf(stderr, "EXP_PreparerTables : ERREUR dans TABLE_CreerTable_mMp() / X fluo\n");
      return -1;
    }
    nbr_tables++; /* Une table de plus !*/
  }

  /* Combien de tables au total ? */
  if ( experience->faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    nbr_tables += experience->faisceau->nbr_harmoniques; /* Une table par harmonique */
  }

  /* On rserve la table des tables */
  tables = ( double ** ) calloc( nbr_tables, sizeof( double * ) );
  if ( NULL == tables ) {
    (void) fprintf(stderr, "EXP_PreparerTables : Mmoire insuffisante --- table des tables\n");
    return -2;
  }
  nbr = ( unsigned long * ) calloc( nbr_tables, sizeof( unsigned long ) );
  if ( NULL == nbr ) {
    (void) fprintf(stderr, "EXP_PreparerTables : Mmoire insuffisante --- table des n\n");
    free( tables );
    return -2;
  }

  /* Les tables d'excitation */
  /* 1. Le fondamental */
  if ( VRAI == EXP_EstSpectroscopie( experience->methode ) ) {
    double E_debut = 0.0, E_fin = 0.0, pas_E = -1.0;

    E_debut = experience->E_debut;
    E_fin = experience->E_fin;
    pas_E = experience->pas_E;

    /* Faisceau polychromatique : on dborde un peu... */
    if ( experience->faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
      /* Dans le spectre, les X sont en relatif par rapport  l'nergie nominale */
      while ( E_debut >= experience->E_debut * ( 1.0 + experience->faisceau->g_E->x_min ) ) {
	E_debut -= pas_E; /* On reste cohrent avec le pas de l'exprience */
      } 

      while ( E_fin <= experience->E_fin * ( 1.0 + experience->faisceau->g_E->x_max ) ) {
	E_fin += pas_E; /* On reste cohrent avec le pas de l'exprience */
      }
    }

    /* Construction de la table des X */
    x_excitation = TABLE_CreerTable_mMp( E_debut, E_fin, pas_E, &n_excitation );
    if ( NULL == x_excitation ) {
      (void) fprintf(stderr, "EXP_PreparerTables : ERREUR dans TABLE_CreerTable_mMp() / X en excitation\n");
      free( tables );
      free( nbr );
      return -3;
    }
  } else {
    double E_excitation;

    E_excitation = experience->E_debut;

    /* Faisceau polychromatique */
    if ( experience->faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
      unsigned long i;

      n_excitation = experience->faisceau->g_E->nombre_point;
      x_excitation = TABLE_CreerTable( n_excitation );
      if ( NULL == x_excitation ) {
	(void) fprintf(stderr, "EXP_PreparerTables : ERREUR dans TABLE_CreerTable(%lu) / X en excitation\n", n_excitation);
	free( tables );
	free( nbr );
	return -3;
      }

      /* Remplissage de la table */
      for ( i = 0 ; i < n_excitation ; i++ ) {
	x_excitation[ i ] = E_excitation * ( 1.0 + experience->faisceau->g_E->x[ i ] );
      }
    } else {
      x_excitation = TABLE_CreerTable( 1 );
      if ( NULL == x_excitation ) {
	(void) fprintf(stderr, "EXP_PreparerTables : ERREUR dans TABLE_CreerTable(1) / X en excitation\n");
	free( tables );
	free( nbr );
	return -3;
      }
      n_excitation = 1;

      *x_excitation = E_excitation;
    }
  }
  tables[ 0 ] = x_excitation;
  nbr[ 0 ] = n_excitation;

  /* 2. Les harmoniques */
  if ( experience->faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    for ( i = 0 ; i < experience->faisceau->nbr_harmoniques ; i++ ) {
      if ( experience->faisceau->i_harmoniques[ i ] > 0.0 ) { /* harmonique non teinte */
	unsigned long idx_x;
	
	tables[ i + 1 ] = TABLE_CopierTable( x_excitation, n_excitation );
	if ( NULL == tables[ i + 1 ] ) {
	  unsigned int j;

	  (void) fprintf(stderr, "EXP_PreparerTables : ERREUR dans TABLE_CopierTable() [harmonique %u]\n", i + 1);
	  for ( j = 0 ; j < i ; j++ )
	    free( tables[ j + 1 ] );
	  free( x_excitation );
	  free( tables );
	  free( nbr );
	  return -4;
	}

	nbr[ i + 2 ] = n_excitation;
	for ( idx_x = 0 ; idx_x < n_excitation ; idx_x++ ) {
	  *(tables[ i + 1 ] + idx_x) *= (double) (i + 2);
	}
      } else {
	tables[ i + 1 ] = NULL; /* Normalement, inutile car fait par calloc, mais bon */
	nbr[ i + 1 ] = 0;
      }
    }
  }

  /* 3. La fluorescence, si ncessaire */
  if ( VRAI == EXP_UtiliseFluorescence( experience ) ) {
    tables[ nbr_tables - 1 ] = experience->table_X_fluo;
    nbr[ nbr_tables - 1 ] = experience->nbr_X_fluo;
  }

  /* 4. On fusionne tout cela en une seule table */
  experience->table_X_mu = TABLE_FondreTables( tables, nbr, nbr_tables, VRAI, &(experience->nbr_X_mu) );

  /* Libration des intermdiaires */
  for ( i = 0 ; i < nbr_tables ; i++ ) {
    if ( tables[ i ] != experience->table_X_fluo ) { /* Pour ne pas l'effacer accidentellement ! */
      free( tables[ i ] );
    }
  }
  free( tables );
  free( nbr );

  /* Contrle */
  if ( NULL == experience->table_X_mu ) {
    (void) fprintf(stderr, "EXP_PreparerTables : ERREUR dans TABLE_FondreTables()\n");
    return -10;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Calculer, pour une composition donne, le coefficient d'absorption pour toutes les valeurs ncessaires -------- */

static int EXP_CalculerMu(Experience *experience, struct matiere *matiere)
{
  int retour;

  /* On efface les anciens spectres d'absorption pour tre tranquille */
  GRF_eff_isole( matiere->mu_total ); matiere->mu_total = NULL;
  GRF_eff_isole( matiere->te_total ); matiere->te_total = NULL;
  GRF_eff_isole( matiere->ti_total ); matiere->ti_total = NULL;

  /* Calcul du coefficient d'absorption pour la table des X utilise ensuite */
  matiere->mu_total = Matiere_CalculerMu( matiere, experience->table_X_mu, experience->nbr_X_mu, 0.0, 0.0, -1.0 );
  if ( NULL == matiere->mu_total ) {
    (void) fprintf(stderr, "EXP_CalculerMu : ERREUR dans Matiere_CalculerMu()\n");
    return -1;
  }

  /* Calcul du coefficient de diffusion lastique */
  matiere->te_total = GRF_nv_TableX( NULL, GR_SEC_EFFICACE, GRF_SERIE, NULL, VRAI, experience->table_X_mu, experience->nbr_X_mu, FAUX );
  if ( NULL == matiere->te_total ) {
    (void) fprintf(stderr, "EXP_CalculerMu : ERREUR dans GRF_nv_TableX() [tau lastique]\n");
    return -2;
  }
  retour = Matiere_CalculerTau( matiere, matiere->te_total, VRAI, DIFFUSION_COEF_McMaster, FAUX );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_CalculerMu : ERREUR dans Matiere_CalculerTau() [tau lastique]\n");
    return -2;
  }

  /* Calcul du coefficient de diffusion inlastique */
  matiere->ti_total = GRF_nv_TableX( NULL, GR_SEC_EFFICACE, GRF_SERIE, NULL, VRAI, experience->table_X_mu, experience->nbr_X_mu, FAUX );
  if ( NULL == matiere->ti_total ) {
    (void) fprintf(stderr, "EXP_CalculerMu : ERREUR dans GRF_nv_TableX() [tau inlastique]\n");
    return -3;
  }
  retour = Matiere_CalculerTau( matiere, matiere->ti_total, FAUX, DIFFUSION_COEF_McMaster, FAUX );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "EXP_CalculerMu : ERREUR dans Matiere_CalculerTau() [tau inlastique]\n");
    return -3;
  }

  if ( experience->table_X_fluo != NULL ) { /* On a besoin du spectre de fluo. idal */
    matiere->fluo = GRF_nv_TableX( NULL, GR_FLUOX, GRF_SERIE, NULL, /*FAUX*/ VRAI, experience->table_X_fluo, experience->nbr_X_fluo, FAUX );
    if ( NULL == matiere->fluo ) {
      (void) fprintf(stderr, "EXP_CalculerMu : ERREUR dans GRF_nv_TableX() [fluo]\n");
      return -4;
    }
  }

  /* Tout s'est bien pass */
  return 0;
}
