/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'une description de configuration expérimentale
   Interface graphique pour le paramétrage de cette expérience
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "experience.h"
#include "graphes.h"
#include "images.h"
#include "interface.h"
#include "tracer.h"

#include "th_interface.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define SSFENETRE_ECHANTILLON 0 /* Index de la sous-fenêtre affichant l'échantillon */
#define SSFENETRE_GEOMETRIE   1 /* Index de la sous-fenêtre affichant la géométrie de l'expérience */
#define SSFENETRE_FAISCEAU    2 /* Index de la sous-fenêtre affichant les propriétés du faisceau */
#define SSFENETRE_DETECTEUR   3 /* Index de la sous-fenêtre affichant les propriétés des détecteurs */

struct infos_echantillon {
  Echantillon *echantillon;
  Image *coupe;         /* Tranche affichée */
  unsigned int x, y;    /* Coordonnées du point [case] dont on affiche le spectre & la composition */
  unsigned int tranche; /* Numéro de la tranche affichée à l'écran */
} infos = { NULL, NULL, 0, 0, 0 };

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void ASC_Detecteur( Fenetre *, int, int, int );
static int  CLC_Detecteur( Fenetre *, int bouton, int n_clic, int etat, int x, int y);
static void RDS_Detecteur( Fenetre *, int, int, unsigned int, unsigned int );

static void ASC_Echantillon( Fenetre *, int, int, int );
static int  CLC_Echantillon( Fenetre *, int bouton, int n_clic, int etat, int x, int y);
static void RDS_Echantillon( Fenetre *, int, int, unsigned int, unsigned int );

static int  CLV_Experience( Fenetre *, unsigned int, int );
static void RDS_Experience3D(Fenetre *, int, int, unsigned int, unsigned int);

static int  CLC_Faisceau( Fenetre *, int bouton, int n_clic, int etat, int x, int y);
static void RDS_Faisceau( Fenetre *, int, int, unsigned int, unsigned int );

static int RDS_BoiteEnglobante(Experience *experience, Vecteur3D *coin_BGAv, Vecteur3D *coin_HDAr);

static int ECH_TrouverPoint(int x_HG, int y_HG, unsigned int larg, unsigned int haut, int x_clic, int y_clic,
			    struct infos_echantillon *infos, unsigned int *i, unsigned int *j);

static int ECH_DemanderComposition(struct infos_echantillon *, unsigned int i, unsigned int j, unsigned int k);
static int ECH_DemanderDensite(struct infos_echantillon *);
static int ECH_PreparerCoupe( struct infos_echantillon * );

static void RDS_SpectresComposition(Fenetre *fenetre, struct matiere *composition, Experience *experience, int x, int y, unsigned int larg, unsigned int haut);
static void RDS_ValeursComposition (Fenetre *fenetre, struct matiere *composition, double correction, int x, int y, unsigned int larg, unsigned int haut);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

int EXP_OuvrirFenetre( void )
{
  int attributs[ 4 ] = { INFO | ASC_VERTICAL, INFO, INFO, INFO | ASC_HORIZONTAL },
      types[ 4 ] = { BI_FENETRE_SIMPLE, BI_FENETRE_GL, BI_FENETRE_SIMPLE, BI_FENETRE_SIMPLE };
  Fenetre *fenetre, **sous_fenetres;

  if ( NULL == experience_courante ) {
    experience_courante = EXP_CreerExperience( EXPERIENCE_XAS_TRANSMISSION );
    if ( NULL == experience_courante ) {
      (void) fprintf(stderr, "EXP_OuvrirFenetre : Impossible de créer une expérience\n");
      return -2;
    }

    experience_courante->E_debut =  9500.0;
    experience_courante->E_fin   = 10600.0;
    experience_courante->pas_E = 2.0;

    experience_courante->Z = ZINC;
    experience_courante->seuil = 0; /* SEUIL_K; */

    /* On crée un détecteur idéal pour I0, par défaut ; de type chambre d'ionisation */
    experience_courante->i0 = DET_CreerDetecteur( DETECTEUR_CHAMBRE, 0 );
    experience_courante->i0->orientation.position.y = -10000.0; /* 10 cm avant échantillon : ne devrait pas poser de souci */
    DET_ChainerDetecteur( NULL, experience_courante->i0 );
    DET_NommerDetecteur( experience_courante->i0, "Chambre I0" );

    /* On crée un détecteur idéal pour I1/If, par défaut ; de type chambre d'ionisation ; derrière l'échantillon, sur l'axe */
    experience_courante->i1 = DET_CreerDetecteur( DETECTEUR_CHAMBRE, 0 );
    experience_courante->i1->orientation.position.y = 10000.0; /* 10 cm après échantillon : ne devrait pas poser de souci */
    DET_ChainerDetecteur( experience_courante->i0, experience_courante->i1 );
    DET_NommerDetecteur( experience_courante->i1, "Chambre I1" );

    /* On crée un faisceau, s'il n'en existe aucun */
    if ( NULL == premier_faisceau ) {
      experience_courante->faisceau = FAISCEAU_Creer( 0.0, 0.0, FAISCEAU_PONCTUEL, 0, 1e8 );
      if ( NULL == experience_courante->faisceau ) {
	(void) fprintf( stderr, "%s,%s l.%u : ERREUR dans la création du faisceau\n",
			__FILE__, __FUNCTION__, __LINE__ );
	OPTION_A_FAIRE( "Annuler la création de l'expérience" );
      }
      FAISCEAU_ChainerFaisceau( NULL, experience_courante->faisceau );
    }

    /* Paramétrage de l'expérience */
    EXP_Parametrer( experience_courante );
  }

  fenetre = FEN_CreerFenetreMultiple( titre_fenetre_experience, 4, attributs, types );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "EXP_OuvrirFenetre : Impossible d'ouvrir la fenêtre...\n");
    return -1;
  }
  FEN_AssocierEvenements( fenetre, NULL, NULL, CLV_Experience, NULL, NULL, NULL, NULL );
  FEN_Ouvrir( fenetre );

  /* On répartit les informations dans les sous-fenêtres */
  sous_fenetres = fenetre->ss_fenetres;
  FEN_AssocierEvenements( sous_fenetres[ SSFENETRE_ECHANTILLON ], RDS_Echantillon, ASC_Echantillon, NULL, CLC_Echantillon, NULL, NULL, NULL );
  FEN_AssocierEvenements( sous_fenetres[ SSFENETRE_GEOMETRIE ], RDS_Experience3D, NULL, NULL, NULL, NULL, NULL, NULL );
  FEN_AssocierEvenements( sous_fenetres[ SSFENETRE_DETECTEUR ], RDS_Detecteur, ASC_Detecteur, NULL, CLC_Detecteur, NULL, NULL, NULL );
  FEN_AssocierEvenements( sous_fenetres[ SSFENETRE_FAISCEAU ], RDS_Faisceau, NULL, NULL, CLC_Faisceau, NULL, NULL, NULL );

  /* L'échantillon */
  OPTION_A_FAIRE( "Réserver au vol les infos pour l'échantillon !" );
  infos.echantillon = experience_courante->echantillon;
  infos.x = 0; infos.y = 0; /* On affiche le contenu du coin en bas, à gauche */
  infos.tranche = 0;        /* On affiche la tranche de surface côté faisceau  */
  ECH_PreparerCoupe( &infos );
  FEN_StockerComplement( sous_fenetres[ SSFENETRE_ECHANTILLON ], FEN_COMPLEMENT_1, &infos );
  if ( NULL == infos.echantillon ) {
    FEN_ascenseur( sous_fenetres[ SSFENETRE_ECHANTILLON ], ASC_VERTICAL, 0, 1000 );
    FEN_infos( sous_fenetres[ SSFENETRE_ECHANTILLON ], infos_echantillon_aucun );
  } else if ( infos.echantillon->nbr_y > 1 ) {
    int taille;
    char *txt_infos;

    taille = (int) ( 1000.0 / ( float ) infos.echantillon->nbr_y );
    FEN_ascenseur( sous_fenetres[ SSFENETRE_ECHANTILLON ], ASC_VERTICAL, 1000, taille );
    txt_infos = TXT_genere_chaine( E_AIGU"chantillon, tranche %u/%u [%s]", infos.tranche + 1, infos.echantillon->nbr_y,
				   infos.echantillon->nom );
    FEN_infos( sous_fenetres[ SSFENETRE_ECHANTILLON ], txt_infos );
    free( txt_infos ); txt_infos = NULL;
  } else {
    FEN_ascenseur( sous_fenetres[ SSFENETRE_ECHANTILLON ], ASC_VERTICAL, 0, 1000 ); 
    FEN_infos( sous_fenetres[ SSFENETRE_ECHANTILLON ], "Composition de l'échantillon [« infiniment » fin]" );
  }

  /* Le faisceau... */
  FEN_StockerComplement( sous_fenetres[ SSFENETRE_FAISCEAU ], FEN_COMPLEMENT_1, experience_courante->faisceau );
  if ( experience_courante->faisceau == NULL ) {
    FEN_infos( sous_fenetres[ SSFENETRE_FAISCEAU ], infos_faisceau_aucun );
  } else {
    char *texte;

    texte = TXT_genere_chaine( infos_faisceau, experience_courante->faisceau->nom );
    FEN_infos( sous_fenetres[ SSFENETRE_FAISCEAU ], texte );
    free( texte );
  }

  /* Deux détecteurs possibles : I0 et I fluo/transmis ->I0 par défaut */
  FEN_ascenseur( sous_fenetres[ SSFENETRE_DETECTEUR ] , ASC_HORIZONTAL, 0, 500 );
  FEN_StockerComplement( sous_fenetres[ SSFENETRE_DETECTEUR ], FEN_COMPLEMENT_1, experience_courante->i0 );
  FEN_infos( sous_fenetres[ SSFENETRE_DETECTEUR ], infos_detecteur );

  /* On redessine tout pour être sûr d'être à jour */
  FEN_redessin( fenetre, BI_RDS_TOUT );

  /* Ouverture correcte */
  return 0;
}

/* Gestion du clavier : commune à toutes les sous-fenetres */

static int CLV_Experience(Fenetre *fenetre, unsigned int touche, int etat)
{
  Fenetre *racine, **sous_fenetres;
  struct infos_echantillon *infos;

  racine = FEN_FenetreMere( fenetre );
  if ( NULL == racine ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FEN_FenetreMere(), fenêtre bizarre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }
  sous_fenetres = racine->ss_fenetres;
  if ( NULL == sous_fenetres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Table des sous-fenêtres nulle\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }

  if ( etat & K_CTRL ) return BI_EVENEMENT_NON_TRAITE; /* Ctrl : raccourci du menu */

  switch ( touche )
  {
  case ' ': /* Lancer la simulation de l'expérience */
    if ( experience_courante->proprietes & EXPERIENCE_CARTOGRAPHIE ) {
      struct image *image;

      image = EXP_ConstruireCartographie( experience_courante );
      if ( NULL == image ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans EXP_ConstruireCartographie()\n",
		       __FILE__, __FUNCTION__, __LINE__);
      }
    } else {
      Graphe *graphe;

      graphe = EXP_ConstruireSpectre( experience_courante );
      if ( NULL == graphe ) {
	(void) fprintf(stderr, "%s, %s l.%u : ERREUR dans EXP_ConstruireSpectre()\n",
		       __FILE__, __FUNCTION__, __LINE__);
      }
    }
    return BI_EVENEMENT_TRAITE;

  case 'e': /* Paramétrer l'échantillon */
    OPTION_A_FAIRE( "e : Paramétrer l'échantillon !" );
    return BI_EVENEMENT_TRAITE;

  case 'f': /* Paramétrer le faisceau */
    FAISCEAU_Parametrer( (Faisceau *) FEN_LireComplement( sous_fenetres[ SSFENETRE_FAISCEAU ], FEN_COMPLEMENT_1 ) );
    FEN_redessin( sous_fenetres[ SSFENETRE_FAISCEAU ], VRAI );
    if ( experience_courante->faisceau == NULL ) {
      FEN_infos( sous_fenetres[ SSFENETRE_FAISCEAU ], infos_faisceau_aucun );
    } else {
      char *texte;

      texte = TXT_genere_chaine( infos_faisceau, experience_courante->faisceau->nom );
      FEN_infos( sous_fenetres[ SSFENETRE_FAISCEAU ], texte );
      free( texte ); texte = NULL;
    }
    return BI_EVENEMENT_TRAITE;

  case 's': /* Sauver l'expérience actuelle */
    I_EXP_Sauver( experience_courante );
    return BI_EVENEMENT_TRAITE;

  case 't': /* Paramétrer l'expérience [_t_ype] */
  case 'x':	/* Idem [e_x_périence, par analogie avec l'appel !] */
    (void) EXP_Parametrer( experience_courante );

    /* On refait les associations, car a pu changer ! */
    infos = (struct infos_echantillon *) FEN_LireComplement( sous_fenetres[ SSFENETRE_ECHANTILLON ], FEN_COMPLEMENT_1 );

    /* On a pu changer d'échantillon ou en changer la taille... */
    IMG_EffacerImage( infos->coupe, VRAI ); /* La coupe n'est plus valable */
    infos->coupe = NULL;

    if ( experience_courante->echantillon != infos->echantillon ) {
      infos->x = 0; infos->y = 0; /* On s'assure que l'on est bien dans l'échantillon */
      infos->tranche = 0;
    } else if ( infos->echantillon != NULL ) { /* Peut être NULL dans les deux si aucun échantillon créé ! */
      if ( infos->x >= infos->echantillon->nbr_x ) infos->x = infos->echantillon->nbr_x;
      if ( infos->y >= infos->echantillon->nbr_z ) infos->y = infos->echantillon->nbr_z;
      if ( infos->tranche >= infos->echantillon->nbr_y ) infos->tranche = infos->echantillon->nbr_y;
    }

    infos->echantillon = experience_courante->echantillon; /* Nouvel échantillon affiché */

    ECH_PreparerCoupe( infos ); /* On prépare la coupe */

    FEN_StockerComplement( sous_fenetres[ SSFENETRE_FAISCEAU ], FEN_COMPLEMENT_1, experience_courante->faisceau );
    if ( experience_courante->faisceau == NULL ) {
      FEN_infos( sous_fenetres[ SSFENETRE_FAISCEAU ], infos_faisceau_aucun );
    } else {
      char *texte;

      texte = TXT_genere_chaine( infos_faisceau, experience_courante->faisceau->nom );
      FEN_infos( sous_fenetres[ SSFENETRE_FAISCEAU ], texte );
      free( texte ); texte = NULL;
    }

    /* On redessine */
    FEN_redessin( racine, BI_RDS_TOUT );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* ~-~-~-~-~-~-~-~- Gestion de la sous-fenêtre associée aux détecteurs ~-~-~-~-~-~-~-~- */

static void ASC_Detecteur(Fenetre *fenetre, int type, int ss_type, int position)
{
  boolean changement = FAUX;
  Detecteur *detecteur;

  detecteur = (Detecteur *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );

  if ( WM_ARROWED == type ) {
    if ( ( WA_LFPAGE == ss_type ) || ( WA_LFLINE == ss_type ) ) { /* Passer sur I0 */
      detecteur = experience_courante->i0;
      changement = VRAI;
    } else if ( ( WA_RTPAGE == ss_type ) || ( WA_RTLINE == ss_type ) ) { /* Passer sur I1 */
      detecteur = experience_courante->i1;
      changement = VRAI;
    }
  } else if ( WM_VSLID == type ) {
    changement = VRAI;
    if ( position < 500 ) { /* On travaille en 1/1000 */
      detecteur = experience_courante->i0;
    } else {
      detecteur = experience_courante->i1;
    }
  }

  if ( FAUX == changement ) return; /* On n'a rien changé */

  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_1, detecteur );
  FEN_ascenseur(fenetre, ASC_HORIZONTAL, ( detecteur == experience_courante->i0 ) ? ( 0 ) : ( 1000 ), -1);

  if ( NULL == detecteur ) {
    if ( detecteur == experience_courante->i0 )
      FEN_infos( fenetre, infos_detecteur_I0_aucun );
    else
      FEN_infos( fenetre, infos_detecteur_I1_aucun );
  } else {
    char *infos;

    infos = TXT_genere_chaine( infos_detecteur, ( detecteur == experience_courante->i0 ) ? ( "I0" ) : ( "I1" ), detecteur->nom );
    FEN_infos( fenetre, infos );
    free( infos ); infos = NULL;
  }

  FEN_redessin( fenetre, BI_RDS_UTIL );
}

static int CLC_Detecteur( Fenetre *fenetre, int bouton, int n_clic, int etat, int x, int y)
{
  int retour;
  Detecteur *detecteur;

  detecteur = (Detecteur *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );

  if ( 1 == bouton ) {
    retour = DETECTEUR_Parametrer( detecteur );
    if ( 0 == retour ) {
      FEN_redessin( fenetre, BI_RDS_UTIL );
    }
  } else {
    (void) fprintf(stderr, "CLC_Detecteur : bouton=%d, n_clic=%d, etat=%d, x=%d, y=%d\n", bouton, n_clic, etat, x, y);
  }

  return BI_EVENEMENT_TRAITE;
}

static void RDS_Detecteur(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Detecteur *detecteur;
  unsigned int lg_case, ht_case __attribute__((__unused__));

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  /* On cherche à afficher le détecteur */
  detecteur = (Detecteur *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == detecteur ) {
    TRC_Message( fenetre, texte_aucun_detecteur );
    return;
  }

  /* On partage la fenêtre en régions : trois colonnes, trois rangées */
  lg_case = fenetre->w_larg / 3;
  ht_case = fenetre->w_haut / 3;

  /* Séparateurs entre cases : lignes noires */
  TRC_Separateur( fenetre, VERTICAL, 2 * lg_case, 0, fenetre->w_haut );
  TRC_Separateur( fenetre, HORIZONTAL, 2 * lg_case, fenetre->w_haut / 2, lg_case );

  /* Case de gauche [la plus grande] : réponse avec la surface */
  if ( NULL == detecteur->r_surface ) {
    TRC_Z_Message( fenetre, 0, 0, 2 * lg_case, fenetre->w_haut, texte_detecteur_uniforme );
  } else {
    OPTION_A_FAIRE( "Détecteur non ponctuel" );
  }

  /* Case en haut, à droite : r(E) */
  if ( NULL == detecteur->composition ) {
    TRC_Z_Message( fenetre, 2 * lg_case, 0, fenetre->w_larg - (2 * lg_case), fenetre->w_haut / 2,
		texte_detecteur_ideal );
  } else {
    if ( NULL == detecteur->composition->mu_total ) {
      detecteur->composition->mu_total =
	Matiere_CalculerMu( detecteur->composition, NULL, 0, experience_courante->E_debut, experience_courante->E_fin, experience_courante->pas_E );
    }
    if ( detecteur->proprietes & DETECTEUR_ABSORBANT ) {
      TRC_GrapheDansRectangle( fenetre, detecteur->composition->mu_total, VIOLET, 1, 2 * lg_case + 2, 2,
			       fenetre->w_larg - (2 * lg_case) - 4, fenetre->w_haut / 2 - 4 );
    } else {
      /* Il y a une composition pour l'absorbance, mais on ne l'utilise pas ! */
      TRC_GrapheDansRectangle( fenetre, detecteur->composition->mu_total, GRIS, 1, 2 * lg_case + 2, 2,
			       fenetre->w_larg - (2 * lg_case) - 4, fenetre->w_haut / 2 - 4 );
      TRC_Z_Message( fenetre, 2 * lg_case, 0, fenetre->w_larg - (2 * lg_case), fenetre->w_haut / 2,
		     texte_detecteur_nabsorbant );
    }
  }

  /* Case en bas, à droite : I(E) */
  if ( NULL == detecteur->r_signal ) {
    TRC_Z_Message( fenetre, 2 * lg_case, fenetre->w_haut / 2,
		   fenetre->w_larg - (2 * lg_case), fenetre->w_haut - ( fenetre->w_haut / 2 ),
		   texte_detecteur_lineaire );
  } else {
    OPTION_A_FAIRE( "Détecteur non-linéaire" );
  }
}

/* ~-~-~-~-~-~-~-~- Gestion de la sous-fenêtre associée à l'échantillon ~-~-~-~-~-~-~-~- */

static void ASC_Echantillon(Fenetre *fenetre, int type, int ss_type, int position)
{
  struct infos_echantillon *infos;

  infos = (struct infos_echantillon *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune information pour l'échantillon\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }
  if ( NULL == infos->echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun échantillon\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  if (type == WM_ARROWED) { /* On a utilisé une des flèches de l'ascenseur */
    switch (ss_type) /* Action qui peut être envisagée */
    {
      /* ------ Ascenseur vertical : passage d'une tranche à l'autre */
    case WA_UPPAGE: /* On remonte de 10 tranches */
      if ( infos->tranche + 10 < infos->echantillon->nbr_y ) {
	infos->tranche += 10;
      } else {
	infos->tranche = infos->echantillon->nbr_y - 1;
      }
      break;

    case WA_DNPAGE: /* On redescend de 10 tranches */
      if ( infos->tranche > 10 ) {
	infos->tranche -= 10;
      } else {
	infos->tranche = 0;
      }
      break;

    case WA_UPLINE: /* Tranche précédente [plus loin de la face d'entrée] */
      if ( infos->tranche + 1 < infos->echantillon->nbr_y ) {
	infos->tranche++;
      }
      break;

    case WA_DNLINE: /* Tranche suivante [plus proche de la face d'entrée] */
      if ( infos->tranche > 0 ) {
	infos->tranche--;
      }
      break;
    }
  } else if (type == WM_VSLID) { /* Ascenseur vertical déplacé */
    int n_tranche;

    n_tranche = (int) rint( (double) ( 1000.0 - position ) * (double) ( infos->echantillon->nbr_y - 1.0 ) / 1000.0 );
    if ( n_tranche < 0 ) n_tranche = 0;
    else if ( (unsigned int) n_tranche >= infos->echantillon->nbr_y ) n_tranche = infos->echantillon->nbr_y - 1;
    infos->tranche = n_tranche;
  }

  /* On met à jour la tranche */
  ECH_TrancheVersImage_Y( infos->echantillon, infos->tranche, infos->coupe );

  /* On place l'ascenseur */
  if ( infos->echantillon->nbr_y > 1 ) {
    int taille;
    char *txt_infos;

    taille   = (int) ( 1000.0 / ( float ) infos->echantillon->nbr_y );
    position = 1000 - (int) ( infos->tranche * 1000.0 / ( float ) ( infos->echantillon->nbr_y - 1 ) );
    FEN_ascenseur( fenetre, ASC_VERTICAL, position, taille );

    txt_infos = TXT_genere_chaine( infos_echantillon, infos->tranche + 1, infos->echantillon->nbr_y, infos->echantillon->nom );
    FEN_infos( fenetre, txt_infos );
    free( txt_infos );
  } else {
    FEN_ascenseur( fenetre, ASC_VERTICAL, 0, 1000 ); 
  }
  /* Et on redessine la tranche */
  FEN_redessin( fenetre, BI_RDS_UTIL );
}

static int CLC_Echantillon(Fenetre *fenetre, int bouton, int n_clic __attribute__((__unused__)), int etat __attribute__((__unused__)), int x, int y)
{
  struct infos_echantillon *infos;
  Echantillon *echantillon;
  int retour __attribute__((__unused__));
  int x_barre_V;

  /* On récupère l'échantillon */
  infos = (struct infos_echantillon *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune information\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return BI_EVENEMENT_TRAITE;
  }
  echantillon = infos->echantillon;
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun échantillon défini\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return BI_EVENEMENT_TRAITE;
  }

  /* Comment se répartissent les zones intéressantes ? */
  /* Séparateur vertical */
  x_barre_V = 2 * ( fenetre->w_larg / 3 ) + 2; 

  /* De quel côté de la barre a-t-on cliqué ? */
  if ( x > x_barre_V ) {
    int y_barre_H;

    /* Séparateur horizontal */
    y_barre_H = 2;
    if ( y > y_barre_H ) { /* Clic sur l'affichage de la composition -> modifier la composition */
      if ( 1 == bouton ) { /* Modifier la composition courante */
	struct matiere *composition = NULL;

	/* On trouve la composition */
	if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
	  composition = echantillon->valeurs.homogene.composition;
	} else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	  composition = echantillon->valeurs.homogene.composition;
	} else {
	  unsigned long n_case;

	  n_case = ECH_IJK_v_Case( echantillon, infos->x, infos->tranche, infos->y );
	  composition = echantillon->valeurs.inhomogene[ n_case ];
	  if ( NULL == composition ) { /* Trou */
	    return BI_EVENEMENT_TRAITE; /* Clic traité : rien à faire pour un trou ! */
	  }
	}

	Matiere_ModifierComposition( composition );
      } else if ( 2 == bouton ) { /* Choisir la composition courante */
	struct matiere *composition = NULL;

	/* On demande la nouvelle composition */
	composition = Matiere_ChoisirComposition( 0, ROOT, 0, NULL );
	if ( NULL == composition ) {
	  return BI_EVENEMENT_TRAITE; /* Traité, mais finalement annulé */
	}

	if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
	  echantillon->valeurs.homogene.composition = composition;
	} else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	  echantillon->valeurs.homogene.composition = composition;
	} else {
	  unsigned long n_case;
	  Image *nv_coupe;

	  /* On travaille sur le point courant -> ses index de case sont dans les infos ! */
	  n_case = ECH_IJK_v_Case( echantillon, infos->x, infos->tranche, infos->y );
	  echantillon->valeurs.inhomogene[ n_case ] = composition;

	  /* On doit aussi refaire l'image, mais ses dimensions n'ont pas changé */
	  nv_coupe = ECH_TrancheVersImage_Y( echantillon, infos->tranche, infos->coupe );
	  if ( ( NULL == nv_coupe ) || ( infos->coupe != nv_coupe ) ) {
	    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le calcul de la coupe\n",
			   __FILE__, __FUNCTION__, __LINE__);
	  }
	}
      } else { /* Choisir le paramètre associé */
	if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	  ECH_DemanderDensite( infos );
	}
      }
    } else { /* Clic sur l'affichage des graphes */
      OPTION_A_FAIRE( "Clic sur les graphes [Demande pour calcul ?]\n" );
    }
  } else { /* Clic sur la coupe de l'échantillon */
    if ( 1 == bouton ) { /* Choisir le point courant */
      if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) { /* Pas de sens -> modifier la composition */
	Matiere_ModifierComposition( echantillon->valeurs.homogene.composition );
      } else {
	int retour;
	unsigned int i, j;

	retour = ECH_TrouverPoint( 0, 0, x_barre_V, fenetre->w_haut, x, y, infos, &i, &j );
	if ( retour != 0 ) {
	  if ( retour < 0 ) {
	    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans ECH_TrouverPoint [code %d]\n",
			   __FILE__, __FUNCTION__, __LINE__, retour);
#ifdef _DEVELOPPE_
	  } else {
	    /* Si on est là : code > 0 ==> hors de l'image */
	    printf("%s : Clic hors de l'image... [code %d]\n", __FUNCTION__, retour );
#endif
	  }
	  return BI_EVENEMENT_TRAITE;
	}

	/* La tranche n'est pas modifiée, juste la position dedans */
	infos->x = j; infos->y = i; /* Attention : i = ligne => y, j = colonne => x ! */
      }
    } else if ( 2 == bouton ) { /* Menu contextuel */
      retour = ECH_SousMenuComposition( echantillon, infos->coupe, infos->tranche );
    } else { /* Choisir la composition / densité en ce point */
      if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) { 
	(void) fprintf(stderr, "%s,%s l.%u : Echantillon homogene, demander la composition\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	ECH_DemanderDensite( infos );
      } else {
	int retour;
	unsigned int i, j;

	/* On repère la position du clic */
	retour = ECH_TrouverPoint( 0, 0, x_barre_V, fenetre->w_haut, x, y, infos, &i, &j );
	if ( retour != 0 ) {
	  if ( retour < 0 ) {
	    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans ECH_TrouverPoint [code %d]\n",
			   __FILE__, __FUNCTION__, __LINE__, retour);
#ifdef _DEVELOPPE_
	  } else {
	    /* Si on est là : code > 0 ==> hors de l'image */
	    printf("%s : Clic hors de l'image... [code %d]\n", __FUNCTION__, retour );
#endif
	  }
	  return BI_EVENEMENT_TRAITE;
	}

	/* On demande la composition en ce point [j = colonne = X] */
	retour = ECH_DemanderComposition( infos, j, infos->tranche, i );
	if ( retour != 0 ) {
#ifdef _DEVELOPPE_
	  (void) fprintf(stderr, "%s, %s l.%u : ERREUR survenue en demandant la composition\n",
			 __FILE__, __FUNCTION__, __LINE__);
#endif
	  return BI_EVENEMENT_TRAITE;
	}
      }
    }
  }

  FEN_redessin( fenetre, BI_RDS_TOUT );

  /* Le clic a été géré */
  return BI_EVENEMENT_TRAITE;
}

/* -------- Trouver les coordonnées (i;j) du point cliqué sur l'image --------
   Renvoie 0 si tout s'est bien passé
           i, j --- i = ligne donc coordonnée Y, j = colonne donc coordonée X
 */

static int ECH_TrouverPoint(int x_HG, int y_HG, unsigned int larg, unsigned int haut, int x_clic, int y_clic,
			    struct infos_echantillon *infos, unsigned int *i, unsigned int *j)
{
  double echelle;
  double x_image, y_image;
  int retour;
  unsigned int lg_image, ht_image;

  /* On récupère l'échelle et les dimensions de l'image */
  echelle = IMG_dim_trace( infos->coupe, larg - 4, haut - 4, &lg_image, &ht_image );
  if ( 0.0 == echelle ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de trouver l'échelle\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On s'affranchit des bords de la fenêtre */
  x_clic -= x_HG;
  y_clic -= y_HG;

  /* On passe dans le repère local */
  x_image = ( (double) x_clic - larg / 2.0 ) /  echelle;
  y_image = ( (double) y_clic - haut / 2.0 ) / -echelle;

  /* On en déduit I (ligne) & J (colonne) */
  retour = IMG_XY_v_IJ( infos->coupe, x_image, y_image, i, j );
  if ( retour != 0 ) {
#ifdef _DEVELOPPE_
    printf("%s : Clic hors de l'image... [code %d]\n", __FUNCTION__, retour );
#endif
    return +1; /* Clic en dehors */
  }

  return 0;
}

/* -------- Redessiner la portion de fenêtre dédiée à l'échantillon -------- */

static void RDS_Echantillon(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  struct infos_echantillon *infos;
  Echantillon *echantillon;
  Image *coupe;
  int i, points[ 10 ], x_centre, y_centre;
  unsigned int lg_case, lg_coupe __attribute__((__unused__)), ht_coupe __attribute__((__unused__)), lg_image, ht_image;
  double echelle;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  /* On cherche à afficher l'échantillon */
  infos = (struct infos_echantillon *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune information\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  echantillon = infos->echantillon;
  if ( NULL == echantillon ) {
    TRC_Message( fenetre, texte_aucun_echantillon );
    return;
  }

  /* On partage la fenêtre en régions : trois colonnes, deux rangées */
  lg_case = fenetre->w_larg / 3;
  if ( ( 0 == lg_case ) || ( 0 == fenetre->w_haut ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Zone de dessin trop petite...\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return;
  }

  /* Séparateurs entre cases : lignes noires */
  TRC_Separateur( fenetre, VERTICAL  , 2 * lg_case, 0, fenetre->w_haut );
  TRC_Separateur( fenetre, HORIZONTAL, 2 * lg_case, fenetre->w_haut / 2, lg_case );

  /* Dessin de l'échantillon
     1. Calcul de l'échelle */
  x_centre = lg_case;
  y_centre = fenetre->w_haut / 2;
  lg_case *= 2;

  coupe = infos->coupe;
  if ( NULL == coupe ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de coupe...\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }
  echelle = IMG_dim_trace( coupe, lg_case - 4, fenetre->w_haut - 4, &lg_image, &ht_image );
  if ( 0.0 == echelle ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_dim_trace\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return;
  }
  /* lg_image, ht_image : dimensions de l'image. L'échantillon est un peu moins grand, car l'image déborde
     => dimensions corrigées dans lg_coupe, ht_coupe */
  lg_coupe = (unsigned int) rint( echantillon->larg * echelle );
  ht_coupe = (unsigned int) rint( echantillon->haut * echelle );

  /* 2. Tracé du contenu */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    TRC_Texte( fenetre, lg_case >> 1, fenetre->w_haut >> 1, BI_TEXTE_CENTRER_H, BI_TEXTE_CENTRER_V, 0, texte_echantillon_homogene );

    RDS_SpectresComposition( fenetre, echantillon->valeurs.homogene.composition, experience_courante,
			     lg_case + 2, 2, fenetre->w_larg - lg_case - 4, fenetre->w_haut / 2 - 4 );
    RDS_ValeursComposition ( fenetre, echantillon->valeurs.homogene.composition, -1.0,
			     lg_case + 2, (fenetre->w_haut / 2) + 2, fenetre->w_larg - lg_case - 4, fenetre->w_haut / 2 - 4 );
  } else {
    Vecteur3D point_affiche;
    int x_affiche, y_affiche;

    /* Quel point est le point courant ? */
    ECH_IJK_v_XYZ( echantillon, infos->x, infos->tranche, infos->y, &(point_affiche.x), &(point_affiche.y), &(point_affiche.z) );

    TRC_ImagePositionCentre( fenetre, coupe, VERT, x_centre, y_centre, echelle );

    /* On affiche le point courant, en bleu (avec tour vert) */
    x_affiche = lg_case / 2.0 + echelle * point_affiche.x;
    y_affiche = ( fenetre->w_haut - fenetre->w_haut / 2.0 ) - echelle * point_affiche.z ;
    TRC_Couleur( VERT );
    TRC_Z_Cercle( fenetre, 0, 0, lg_case, fenetre->w_haut, x_affiche, y_affiche, 4 );
    TRC_Couleur( BLEU );
    TRC_Z_Disque( fenetre, 0, 0, lg_case, fenetre->w_haut, x_affiche, y_affiche, 4 );

    if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      double correction;

      correction = ECH_LireDensite( echantillon, &point_affiche );
      RDS_SpectresComposition( fenetre, echantillon->valeurs.homogene.composition, experience_courante,
			       lg_case + 2, 2, fenetre->w_larg - lg_case - 4, fenetre->w_haut / 2 - 4 );
      RDS_ValeursComposition ( fenetre, echantillon->valeurs.homogene.composition, correction,
			       lg_case + 2, (fenetre->w_haut / 2) + 2, fenetre->w_larg - lg_case - 4, fenetre->w_haut / 2 - 4 );
    } else {
      unsigned long n_case;

      /* Composition du point actif */
      n_case = ECH_IJK_v_Case( echantillon, infos->x, infos->tranche, infos->y );

      RDS_SpectresComposition( fenetre, echantillon->valeurs.inhomogene[ n_case ], experience_courante,
			       lg_case + 2, 2, fenetre->w_larg - lg_case - 4, fenetre->w_haut / 2 - 4 );
      RDS_ValeursComposition ( fenetre, echantillon->valeurs.inhomogene[ n_case ], 1.0,
			       lg_case + 2, (fenetre->w_haut / 2) + 2, fenetre->w_larg - lg_case - 4, fenetre->w_haut / 2 - 4 );
    }
  }

  /* 3. Tracé du repère : bleu, au centre */
  TRC_Couleur( BLEU );
  TRC_LigneH( fenetre, 2, y_centre, lg_case - 4 );
  TRC_LigneV( fenetre, x_centre, 2, fenetre->w_haut - 4 );

  /* 4. Tracé du tour, en bleu aussi */
  TRC_Rectangle( fenetre, (lg_case - echantillon->larg * echelle) / 2.0, ( fenetre->w_haut - echantillon->haut * echelle ) / 2.0,
		 echantillon->larg * echelle, echantillon->haut * echelle );

  /* --- Tracés complémentaires : du faisceau ---
     1. Point d'impact (centre du faisceau) = disque rouge */
  if ( experience_courante->faisceau != NULL ) {
    Faisceau *faisceau;
    double x_vise = 0.0, y_vise = 0.0, e_impact;
    int retour;

    faisceau = experience_courante->faisceau;

    /* À quelle profondeur de l'échantillon est-on ? */
    e_impact = infos->tranche * infos->echantillon->pas_y;

    /* Coordonnées du centre du faisceau */
    retour = FAISCEAU_ProjeterCentre( faisceau, e_impact, &x_vise, &y_vise );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "RDS_Echantillon : Faisceau parallèle à l'échantillon\n");
      return;
    }
    /* printf("Point visé : X= %g, Y=%g (microns)", x_vise, y_vise); */

    /* On convertit dans le repère de la fenêtre */
    x_vise = x_vise * echelle + lg_case / 2.0;
    y_vise = ( fenetre->w_haut - fenetre->w_haut / 2.0 ) - y_vise * echelle ;
    /*       printf(" ==> X= %g, Y=%g (pixels)\n", x_vise, y_vise); */
    if ( ( x_vise < 0 ) || ( y_vise < 0 ) || ( x_vise > fenetre->w_larg ) || ( y_vise > fenetre->w_haut ) ) {
      (void) fprintf(stderr, "RDS_Echantillon : ATTENTION, le centre du faisceau ne passe pas par l'échantillon\n");
    } else {
      /* Le disque --- rouge, bord vert */
      TRC_Couleur( VERT );
      TRC_Z_Cercle( fenetre, 0, 0, lg_case, fenetre->w_haut, (int) rint( x_vise ), (int) rint( y_vise ), 2 );
      TRC_Couleur( ROUGE );
      TRC_Z_Disque( fenetre, 0, 0, lg_case, fenetre->w_haut, (int) rint( x_vise ), (int) rint( y_vise ), 2 );

/*       if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) { */
/* 	TRC_ImagePositionCentre( _PRG_station, faisceau->g_espace, ROUGE, x_vise, y_vise, echelle ); */
/*       } */
    }

    /* Faisceau non ponctuel : on trace donc le tour aussi */
    if ( ( faisceau->proprietes & FAISCEAU_PONCTUEL ) == FAUX ) {
      double x_coin[ 5 ], y_coin[ 5 ];
      Point_Ecran sommets[ 5 ];

      FAISCEAU_ProjeterLimites( faisceau, e_impact, x_coin, y_coin);
/*       printf("Faisceau : (%g,%g)-(%g,%g)-(%g,%g)-(%g,%g)\n", x_coin[0],y_coin[0], x_coin[1],y_coin[1], x_coin[2],y_coin[2], x_coin[3],y_coin[3]); */
      for ( i = 0 ; i < 4 ; i++ ) {
	sommets[ i ].x = x_coin[ i ] * echelle + lg_case / 2.0;
	sommets[ i ].y = ( fenetre->w_haut - fenetre->w_haut / 2.0 ) - y_coin[ i ] * echelle ;
      }
      sommets[ 4 ] = sommets[ 0 ]; /* On ferme le rectangle _incliné_ */

      TRC_Couleur( ROUGE );
      TRC_LigneBrisee( fenetre, 5, sommets );

#ifdef _DEVELOPPE_
      points[ 0 ] = x_coin[ 4 ] * echelle + lg_case / 2.0;
      points[ 1 ] = ( fenetre->w_haut - fenetre->w_haut / 2.0 ) - y_coin[ 4 ] * echelle ;
	
      /* Le disque --- vert --- du centre obtenu par la 2e méthode */
      TRC_Couleur( VERT );
      TRC_Z_Disque( fenetre, 0, 0, lg_case, fenetre->w_haut, points[ 0 ], points[ 1 ], 2 );
#endif
    }
  }

  /* ---- Tracé complémentaire : du point visé par le détecteur en fluorescence ---- */
  if ( ( EXP_UtiliseFluorescence( experience_courante ) ) && ( experience_courante->i1 != NULL ) ) {
    Detecteur *det_fluo;
    Vecteur3D v_directeur, point_vise;
    double x_vise = 0.0, y_vise = 0.0, e_impact;
    int retour;

    det_fluo = experience_courante->i1;

    /* À quelle profondeur de l'échantillon est-on ? */
    e_impact = infos->tranche * infos->echantillon->pas_y;

    /* Coordonnées du centre du détecteur */
    /* a. Direction du faisceau */
    REPERE_VecteurNormal( &(det_fluo->orientation), &v_directeur );

    /* Intersection de cette droite avec le plan y=profondeur */
    retour = U3D_Inter_Droite_Plan_YCt( &(det_fluo->orientation.position), &v_directeur, e_impact, &point_vise);
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans U3D_Inter_Droite_Plan_YCt() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour );
    }

    /* On convertit dans le repère de la fenêtre */
    x_vise = point_vise.x * echelle + lg_case / 2.0;
    y_vise = ( fenetre->w_haut - fenetre->w_haut / 2.0 ) - point_vise.z * echelle ;
    if ( ( x_vise < 0 ) || ( y_vise < 0 ) || ( x_vise > fenetre->w_larg ) || ( y_vise > fenetre->w_haut ) ) {
      (void) fprintf(stderr, "RDS_Echantillon : ATTENTION, le détecteur ne vise pas l'échantillon\n");
    } else {
      /* Le disque --- violet, bord vert */
      TRC_Couleur( VERT );
      TRC_Z_Cercle( fenetre, 0, 0, lg_case, fenetre->w_haut, (int) rint( x_vise ), (int) rint( y_vise ), 2 );
      TRC_Couleur( VIOLET );
      TRC_Z_Disque( fenetre, 0, 0, lg_case, fenetre->w_haut, (int) rint( x_vise ), (int) rint( y_vise ), 2 );
    }
  }

  /* ---- Tracé complémentaire : de la carte de fluorescence, si expérience de cartographie ---- */
  if ( experience_courante->proprietes & EXPERIENCE_CARTOGRAPHIE ) {
    int x_BG, y_BG;
    unsigned int largeur, hauteur;
    unsigned long i;

    /* On convertit en coordonnées de la fenêtre */
    x_BG = (int) rint( experience_courante->x_min * echelle + lg_case / 2.0 );
    y_BG = (int) rint( ( fenetre->w_haut - fenetre->w_haut / 2.0 ) - experience_courante->y_min * echelle );

    largeur = (unsigned int) rint( fabs( experience_courante->x_max - experience_courante->x_min ) * echelle );
    hauteur = (unsigned int) rint( fabs( experience_courante->y_max - experience_courante->y_min ) * echelle );

    /* On trace le tout de la carto --- en jaune */
    TRC_Couleur( JAUNE );
    TRC_Rectangle( fenetre, x_BG, y_BG - hauteur, largeur, hauteur );

    /* On trace la grille en X */
    if ( experience_courante->pas_x > 0.0 ) {
      for ( i = 0 ;
	    i < 1 + (unsigned int) rint( fabs( experience_courante->y_max - experience_courante->y_min ) / experience_courante->pas_x ) ;
	    i++ ) {
	TRC_LigneV( fenetre, x_BG + (int) rint( i * experience_courante->pas_x * echelle ), y_BG - hauteur, hauteur );
      }
    }

    /* On trace la grille en Y */
    if ( experience_courante->pas_y > 0.0 ) {
      for ( i = 0 ;
	    i < 1 + (unsigned int) rint( fabs( experience_courante->y_max - experience_courante->y_min ) / experience_courante->pas_y ) ;
	    i++ ) {
	TRC_LigneH( fenetre, x_BG, y_BG - hauteur + (int) rint( i * experience_courante->pas_y * echelle ), largeur );
      }
    }
  }
}

static void RDS_SpectresComposition(Fenetre *fenetre, struct matiere *composition, Experience *experience, int x, int y, unsigned int larg, unsigned int haut)
{
  if ( NULL == composition ) {
    TRC_Z_Message( fenetre, x, y, larg, haut, texte_sans_composition );
    return;
  }

  if ( EXP_EstSpectroscopie( experience->methode ) == VRAI ) {
    /* Tracé du spectre d'absorption de l'échantillon */
    if ( NULL == composition->mu_total ) {
      composition->mu_total = Matiere_CalculerMu( composition, NULL, 0, experience->E_debut, experience->E_fin, experience->pas_E );
    }
    TRC_GrapheDansRectangle( fenetre, composition->mu_total, BLEU, 1, x, y, larg, haut );
  }

  if ( EXP_UtiliseFluorescence( experience ) == VRAI ) {
    if ( NULL == composition->fluo ) {
      double E_excitation;

      if ( EXP_EstSpectroscopie( experience->methode ) == VRAI ) {
	E_excitation = ( experience->E_debut + experience->E_fin ) / 2.0;
      } else {
	E_excitation = experience->E_debut;
      }
      composition->fluo = GRF_nv_GrilleXP( NULL, GR_FLUOX, GRF_FCT_REELLE, NULL, VRAI, experience->Ef_debut, experience->Ef_fin, experience->pas_Ef );
      if ( NULL == composition->fluo ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la préparation du graphe de fluorescence\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return;
      }
      Matiere_CalculerFluo( composition, E_excitation, VRAI );
    }
    TRC_GrapheDansRectangle( fenetre, composition->fluo, VERT, 1, x, y, larg, haut );
  }
}

static void RDS_ValeursComposition(Fenetre *fenetre, struct matiere *composition, double correction, int x, int y, unsigned int larg, unsigned int haut)
{
  int y_texte __attribute__((__unused__));
  int attributs_texte[ 10 ];
  unsigned int i_cmp;
  unsigned int larg_texte, haut_texte;
  char *texte;

  /* Dimensions d'un caractère */
  vqt_attributes( _PRG_station->handle, attributs_texte);
  larg_texte = attributs_texte[6];
  haut_texte = attributs_texte[7] + 2; /* Interligne de 2 pixels entre les textes */

  /* Première ligne : le nom de la composition */
  y_texte = y;

  if ( NULL == composition ) {
    TRC_Couleur( ROUGE );
    TRC_Texte( fenetre, x + larg / 2, y, BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_HAUT, 0, texte_trou );
    return;
  } else {
    unsigned int larg_titre;

    TRC_Couleur( BLEU );
    larg_titre = larg_texte * strlen( composition->nom );
    if ( larg_titre > larg ) {
      TRC_Texte( fenetre, x, y, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, composition->nom );
    } else {
      TRC_Texte( fenetre, x + larg/2, y, BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_HAUT, 0, composition->nom );
    }
    y += haut_texte;
  }

  /* Textes suivants en noir par défaut */
  TRC_Couleur( NOIR );

  /* Deuxième ligne : la densité */
  texte = TXT_genere_chaine( texte_densite, composition->densite );
  TRC_Texte( fenetre, x, y, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  y += haut_texte;
  free( texte ); texte = NULL;

  if ( correction >= 0.0 ) {
    texte = TXT_genere_chaine( "Correction : %f", correction );
    TRC_Texte( fenetre, x + haut_texte, y, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
    y += haut_texte;
    free( texte ); texte = NULL;
  }

  /* Ensuite : les constituants & leurs fractions massiques */
  for ( i_cmp = 0 ; i_cmp < composition->nbr_composants ; i_cmp++ ) {
    unsigned int i_atome;
    struct composant *composant;

    composant = composition->composants[ i_cmp ];
    if ( NULL == composant ) continue;

    texte = TXT_genere_chaine( "%P %s", composition->fraction_massique[ i_cmp ], composant->nom );
    TRC_Texte( fenetre, x, y, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
    y += haut_texte;
    free( texte ); texte = NULL;

    for ( i_atome = 0 ; i_atome < ZMAX_TABLE_CONSTITUANTS ; i_atome++ ) {
      if ( composant->atomes[ i_atome ] <= 0.0 ) continue;

      texte = TXT_genere_chaine( "%P %s [%s]", composant->atomes[ i_atome ], ATM_Z_vers_symbole( i_atome + 1 ), ATM_Z_vers_nom( i_atome + 1 ) );
      if ( composant->mu[ i_atome ] != NULL ) {
	TRC_Couleur( ROUGE );
      }
      TRC_Texte( fenetre, x + larg_texte, y, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
      y += haut_texte;
      free( texte ); texte = NULL;

      TRC_Couleur( NOIR );
      if ( (unsigned int ) y >= y + haut ) break;
    }

    if ( (unsigned int ) y >= y + haut ) break;
  }
}

/* -------- Demander la composition pour un point de l'échantillon -------- */

static int ECH_DemanderComposition(struct infos_echantillon *infos, unsigned int i, unsigned int j, unsigned int k)
{
  struct matiere *composition;
  unsigned long n_case;
  Image *nv_coupe;

  /* Quelle case est modifiée ? */
  n_case = ECH_IJK_v_Case( infos->echantillon, i, j, k );

  /* On demande la nouvelle composition en cette case */
  composition = Matiere_ChoisirComposition( 0, ROOT, 0, NULL );

  /* On la place dans la case, si elle est différente de l'ancienne */
  if ( composition == infos->echantillon->valeurs.inhomogene[ n_case ] ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : Composition inchangée\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return +1;
  }
  infos->echantillon->valeurs.inhomogene[ n_case ] = composition;

  /* On reconstruit la tranche : on doit refaire l'image, mais ses dimensions n'ont pas changé */
  nv_coupe = ECH_TrancheVersImage_Y( infos->echantillon, infos->tranche, infos->coupe );
  if ( ( NULL == nv_coupe ) || ( infos->coupe != nv_coupe ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le calcul de la coupe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +5;
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Demander la densité pour un point de l'échantillon -------- */

static int ECH_DemanderDensite(struct infos_echantillon *infos)
{
  double densite, nv_densite;
  double x, y, z;
  int reponse;
  struct matiere *composition;
  struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, BI_EDITABLE_MIN, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };
  unsigned long n_case;

  composition = infos->echantillon->valeurs.homogene.composition;
  n_case  = ECH_IJK_v_Case( infos->echantillon, infos->x, infos->tranche, infos->y ); 
  densite = infos->echantillon->valeurs.homogene.concentration[ n_case ];

  ECH_IJK_v_XYZ( infos->echantillon, infos->x, infos->tranche, infos->y, &x, &y, &z );

  nv_valeur.texte = "d =";
  nv_valeur.valeur_initiale.reel = densite;

  reponse = BI_DemandeValeur( NULL, &Q_ValeurDensite, &nv_valeur, x, y, z, composition->densite );
  if ( reponse == (int) Q_ValeurDensite.bouton_annuler ) return +1; /* On annule */
  nv_densite = nv_valeur.valeur_donnee.reel;

  infos->echantillon->valeurs.homogene.concentration[ n_case ] = nv_densite;

  /* On refait la coupe */
  ECH_TrancheVersImage_Y( infos->echantillon, infos->tranche, infos->coupe );

  /* Pas d'erreur */
  return 0;
}

/* -------- Préparer l'image qui contiendra la coupe de l'échantillon affichée -------- */

static int ECH_PreparerCoupe( struct infos_echantillon *infos )
{
  int type_image = IMG_TYPE_DENSITE, genre_image = IMG_GENRE_REELLE;

  if ( NULL == infos ) return -1;
  if ( NULL == infos->echantillon ) {
    (void) fprintf(stderr, "ECH_PreparerCoupe : Pas d'échantillon\n");
    return -2;
  }

  ENTREE_FONCTION( __FUNCTION__, "infos=%p", infos);

  /* On efface éventuellement l'ancienne image */
  if ( infos->coupe != NULL ) {
    IMG_EffacerImage( infos->coupe, VRAI );
    infos->coupe = NULL;
  }

  /* De quel type sera l'image ? */
  if ( ECHANTILLON_GRADIENT == infos->echantillon->type )
    type_image = IMG_TYPE_DENSITE;
  else if ( FAUX == ( infos->echantillon->proprietes & ECHANTILLON_HOMOGENE ) ) {
    type_image = IMG_TYPE_COMPOSITION;
    genre_image = IMG_GENRE_COMPO;
  }

  /* On crée l'image qui contiendra la coupe
     Remarque : il faut la créer même pour un homogène, car utilisée pour les échelles de tracé et tout ça...
  */
#ifdef _DEVELOPPE_
  infos->coupe = IMG_CreerImage( NULL, NULL, genre_image, type_image, infos->echantillon->nbr_x, infos->echantillon->nbr_z );
  IMG_Titre( infos->coupe, "Coupe de l'échantillon" );
#else
  infos->coupe = IMG_CreerImageIsolee( genre_image, type_image, infos->echantillon->nbr_x, infos->echantillon->nbr_z );
#endif
  if ( NULL == infos->coupe ) {
    (void) fprintf(stderr, "ECH_PreparerCoupe : Impossible de préparer la coupe\n");
    SORTIE_FONCTION( __FUNCTION__, "%d [ERREUR dans la création de l'image]", -3);
    return -3;
  }

  /* On met à jour la tranche */
  ECH_TrancheVersImage_Y( infos->echantillon, infos->tranche, infos->coupe );
#ifdef _DEVELOPPE_
  printf("%s, image : %g de large, %g de haut --- Echantillon : %g de large, %g de haut\n", __FUNCTION__,
	 infos->coupe->larg * infos->coupe->l_reel,  infos->coupe->haut * infos->coupe->h_reel, infos->echantillon->larg, infos->echantillon->haut);
#endif

  /* Tout s'est bien passé */
  SORTIE_FONCTION( __FUNCTION__, "%d", 0);
  return 0;
}

/* ~-~-~-~-~-~-~-~- Gestion de la sous-fenêtre associée au faisceau ~-~-~-~-~-~-~-~- */

static int CLC_Faisceau(Fenetre *fenetre, int bouton __attribute__((__unused__)), int n_clic __attribute__((__unused__)), int etat __attribute__((__unused__)), int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  Faisceau *faisceau;
  Fenetre *racine;

  /* On cherche à modifier le faisceau */
  faisceau = (Faisceau *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == faisceau ) {
    int code_retour;
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Pas de faisceau\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    
    code_retour = FAISCEAU_Nouveau_I( 0 );
    if ( code_retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_Nouveau_I() [code %d]\n",
		      __FILE__, __FUNCTION__, __LINE__, code_retour );
      return BI_EVENEMENT_ERREUR;
    }

    /* La fonction de création offre de le paramétrer... */

    return BI_EVENEMENT_TRAITE;
  }

  /* On modifie les paramètres */
  FAISCEAU_Parametrer( faisceau );

  /* Mise à jour du titre */
  if ( NULL == faisceau ) {
    FEN_infos( fenetre, infos_faisceau_aucun );
  } else {
    char *texte;

    texte = TXT_genere_chaine( infos_faisceau, faisceau->nom );
    FEN_infos( fenetre, texte );
    free( texte ); texte = NULL;
  }

  /* Et on redessine pour mettre à jour */
  FEN_redessin( fenetre, BI_RDS_UTIL );

  /* Il faut aussi redessiner la partie échantillon : trace du faisceau... */
  racine = FEN_FenetreMere( fenetre );
  if ( NULL != racine ) {
    Fenetre **sous_fenetres;

    sous_fenetres = racine->ss_fenetres;
    if ( NULL == sous_fenetres ) {
      (void) fprintf( stderr, "%s,%s l.%u : Table des sous-fenêtres nulle\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      return FAUX;
    }
    FEN_redessin( sous_fenetres[ SSFENETRE_ECHANTILLON ], BI_RDS_UTIL );
  }

  /* On a réussi, tout va bien */
  return BI_EVENEMENT_TRAITE;
}

static void RDS_Faisceau(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Faisceau *faisceau;
  int x_centre, y_centre;
  unsigned int lg_case, ht_case, lg_image, ht_image;
  double echelle;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  /* On cherche à afficher le faisceau */
  faisceau = (Faisceau *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == faisceau ) {
    TRC_Message( fenetre, texte_aucun_faisceau );
    return;
  }

  /* On partage la fenêtre en régions : trois colonnes, trois rangées */
  lg_case = fenetre->w_larg / 3;
  ht_case = fenetre->w_haut / 3;

  /* Séparateurs entre cases : lignes noires */
  TRC_Separateur( fenetre, VERTICAL  , 2 * lg_case, 0, fenetre->w_haut );
  TRC_Separateur( fenetre, HORIZONTAL, 0, 2 * ht_case, 2 * lg_case );
  TRC_Separateur( fenetre, HORIZONTAL, 2 * lg_case, fenetre->w_haut / 2, fenetre->w_larg );

  /* Case en haut, à droite : g(E) */
  if ( NULL == faisceau->g_E ) {
    TRC_Z_Message( fenetre, 2 * lg_case, 0, lg_case, fenetre->w_haut / 2,
		   texte_faisceau_monochrome );
  } else {
    TRC_GrapheDansRectangle( fenetre, faisceau->g_E, ROUGE, 1, fenetre->w_x + 2 * lg_case + 2, fenetre->w_y + 2,
			     fenetre->w_larg - (2 * lg_case) - 4, fenetre->w_haut / 2 - 4 );
  }

  /* Case en bas, à droite : I(E) */
  if ( NULL == faisceau->g_energie ) {
    TRC_Z_Message( fenetre, 2 * lg_case, fenetre->w_haut / 2, lg_case, fenetre->w_haut - fenetre->w_haut / 2,
		   texte_faisceau_constant );
  } else {
    TRC_GrapheDansRectangle( fenetre, faisceau->g_energie, ROUGE, 1,
			     2 * lg_case + 2, fenetre->w_haut / 2 + 2,
			     fenetre->w_larg - (2 * lg_case) - 4, fenetre->w_haut - ( fenetre->w_haut / 2 ) - 4 );
  }

  /* Case en bas, à gauche : harmoniques */
  if ( faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    OPTION_A_FAIRE( "Affichage des harmoniques" );
  }

  /* Case en haut, à gauche [la plus grande] : coupe du faisceau */
  lg_case *= 2;
  ht_case *= 2;

  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    TRC_Z_Message( fenetre, 0, 0, lg_case, ht_case, texte_faisceau_ponctuel );
    return;
  }

  if ( NULL == faisceau->g_espace ) {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau rectangulaire sans coupe !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* Calcul de l'échelle [2 pixels de marge tout autour] */
  echelle = IMG_dim_trace( faisceau->g_espace, lg_case - 4, ht_case - 4, &lg_image, &ht_image );
  if ( 0.0 == echelle ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_dim_trace()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* Position du centre du faisceau à l'écran */
  x_centre = lg_case / 2;
  y_centre = ht_case / 2;

  /* Tracé du contenu */
  if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) {
    TRC_ImagePositionCentre( fenetre, faisceau->g_espace, ROUGE, x_centre, y_centre, echelle );
  } else {
    TRC_Z_Message( fenetre, 0, 0, lg_case, ht_case, texte_faisceau_homogene );
  }

  /* 3. Tracé du repère : bleu, au centre */
  TRC_Couleur( BLEU );
  TRC_LigneH( fenetre, 2, y_centre, lg_case - 4 );
  TRC_LigneV( fenetre, x_centre, 2, ht_case - 4 );

  /* 4. Tracé du cadre : bleu, limites du faisceau */
  TRC_Rectangle( fenetre, x_centre - ( faisceau->largeur / 2.0) * echelle, y_centre - ( faisceau->hauteur / 2.0) * echelle,
		 faisceau->largeur * echelle, faisceau->hauteur * echelle );
}

/* ~-~-~-~-~-~-~-~-~ Vue en 3D de l'expérience ~-~-~-~-~-~-~-~-~ */

/* -------- Détermination de la boîte qui contient tout -------- */

static int RDS_BoiteEnglobante(Experience *experience, Vecteur3D *coin_BGAv, Vecteur3D *coin_HDAr)
{
  double /*min,*/ dimension;

  if ( NULL == experience ) return -1;
#ifdef _DEVELOPPE_
  if ( ( NULL == coin_BGAv ) || ( NULL == coin_HDAr ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : pas de vecteur passé en paramètre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -157;
  }
#endif

  /* 0. On initialise pour avoir 2 microns de boîte dans le pire des cas */
  coin_BGAv->x = -1.0; coin_BGAv->y = -1.0;  coin_BGAv->z = -1.0;
  coin_HDAr->x =  1.0; coin_HDAr->y =  1.0;  coin_HDAr->z =  1.0;

  /* 1. On contient au moins l'échantillon, s'il y en a un */
  if ( experience->echantillon != NULL ) {
    Echantillon *echantillon;

    echantillon = experience->echantillon;
    coin_BGAv->x = -echantillon->larg / 2.0;
    coin_BGAv->y = 0.0;
    coin_BGAv->z = -echantillon->haut / 2.0;

    coin_HDAr->x = echantillon->larg / 2.0;
    coin_HDAr->y = echantillon->epaisseur;
    coin_HDAr->z = echantillon->haut / 2.0;
  }

  /* 2. On veut que les détecteurs figurent aussi, si pas à l'infini ! */
  if ( experience->i0 != NULL ) {
    Vecteur3D *position;

    position = &(experience->i0->orientation.position);
    if ( position->x < coin_BGAv->x ) coin_BGAv->x = position->x;
    if ( position->y < coin_BGAv->y ) coin_BGAv->y = position->y;
    if ( position->z < coin_BGAv->z ) coin_BGAv->z = position->z;

    if ( position->x > coin_HDAr->x ) coin_HDAr->x = position->x;
    if ( position->y > coin_HDAr->y ) coin_HDAr->y = position->y;
    if ( position->z > coin_HDAr->z ) coin_HDAr->z = position->z;
  }

  if ( experience->i1 != NULL ) {
    Vecteur3D *position;

    position = &(experience->i1->orientation.position);
    if ( position->x < coin_BGAv->x ) coin_BGAv->x = position->x;
    if ( position->y < coin_BGAv->y ) coin_BGAv->y = position->y;
    if ( position->z < coin_BGAv->z ) coin_BGAv->z = position->z;

    if ( position->x > coin_HDAr->x ) coin_HDAr->x = position->x;
    if ( position->y > coin_HDAr->y ) coin_HDAr->y = position->y;
    if ( position->z > coin_HDAr->z ) coin_HDAr->z = position->z;
  }

  /* Essai pour avoir une boîte cubique */
#ifdef essai
  min = coin_BGAv->x;
  if ( coin_BGAv->y < min ) min = coin_BGAv->y;
  if ( coin_BGAv->z < min ) min = coin_BGAv->z;
  coin_BGAv->x = min; coin_BGAv->y = min; coin_BGAv->z = min;

  min = coin_HDAr->x;
  if ( coin_HDAr->y > min ) min = coin_HDAr->y;
  if ( coin_HDAr->z > min ) min = coin_HDAr->z;
  coin_HDAr->x = min; coin_HDAr->y = min; coin_HDAr->z = min;
#endif

  /* On ajoute 10% de cadre autour */
  dimension = ( coin_HDAr->x - coin_BGAv->x );
  coin_BGAv->x -= dimension * 0.05;
  coin_HDAr->x += dimension * 0.05;

  dimension = ( coin_HDAr->y - coin_BGAv->y );
  coin_BGAv->y -= dimension * 0.05;
  coin_HDAr->y += dimension * 0.05;

  dimension = ( coin_HDAr->z - coin_BGAv->z );
  coin_BGAv->z -= dimension * 0.05;
  coin_HDAr->z += dimension * 0.05;

#ifdef _DEVELOPPE_
  printf( "Boîte englobante : [%g, %g] × [%g, %g] × [%g, %g]\n",
	  coin_BGAv->x, coin_HDAr->x, coin_BGAv->y, coin_HDAr->y, coin_BGAv->z, coin_HDAr->z );
#endif

  /* Terminé */
  return 0;
}

static void RDS_Experience3D(Fenetre *fenetre, int x __attribute__((__unused__)), int y __attribute__((__unused__)),
			     unsigned int larg __attribute__((__unused__)), unsigned int haut __attribute__((__unused__)))
{
  int retour;
  Vecteur3D coin_BGAv, coin_HDAr;

  if ( FAUX == BI_OpenGL_actif ) {
    TRC_Message( fenetre, "Open-GL non disponible" );
    return;
  }

  /* On repère la boîte qui entoure l'expérience */
  retour = RDS_BoiteEnglobante( experience_courante, &coin_BGAv, &coin_HDAr );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en construisant la boîte englobante [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

#ifdef AVEC_OPEN_GL_
  /* Matrice de projection */
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glOrtho( coin_BGAv.x * 4.0, coin_HDAr.x * 4.0, coin_BGAv.y * 2.0, coin_HDAr.y * 2.0, coin_BGAv.z * 2.0, coin_HDAr.z * 4.0 );

  /* Placement de la vue */
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  gluLookAt( coin_HDAr.x, coin_BGAv.y, coin_HDAr.z,
	     (coin_HDAr.x + coin_BGAv.x) / 2.0, (coin_HDAr.y + coin_BGAv.y) / 2.0, (coin_HDAr.z + coin_BGAv.z) / 2.0,
	     0.0, 0.0, 1.0 );

#ifdef _DEVELOPPE_
  TRC_Boite3D( &coin_BGAv, &coin_HDAr, -1, BLANC, 1.0 );
#endif

  /* On trace les axes - en vert */
  glDisable(GL_LIGHTING);
  glShadeModel(GL_FLAT);
  glLineWidth(1.0); /* Épaisseur décente... */
  glColor3ub( 0, 255, 0 ); /* On passe en vert */
  glBegin(GL_LINES);
    glVertex3f( 0.0  , 0.0  , 0.0   ); /* Axe des X */
    glVertex3f( 1.0e8, 0.0  , 0.0   );
    glVertex3f( 0.0  , 0.0  , 0.0   ); /* Axe des Y */
    glVertex3f( 0.0  , 1.0e8, 0.0   );
    glVertex3f( 0.0  , 0.0  , 0.0   ); /* Axe des Z */
    glVertex3f( 0.0  , 0.0  , 1.0e8 );
  glEnd();

  /* On dessine l'échantillon : boîte [bleue] en fil de fer */
  if ( experience_courante->echantillon != NULL ) {
    Echantillon *echantillon;

    echantillon = experience_courante->echantillon;

    coin_BGAv.x = -echantillon->larg / 2.0;
    coin_BGAv.y = 0.0;
    coin_BGAv.z = -echantillon->haut / 2.0;

    coin_HDAr.x = echantillon->larg / 2.0;
    coin_HDAr.y = echantillon->epaisseur;
    coin_HDAr.z = echantillon->haut / 2.0;

    TRC_Boite3D( &coin_BGAv, &coin_HDAr, -1, BLEU, 2.0 );
  }

  /* On dessine le faisceau : en rouge */
  if ( experience_courante->faisceau != NULL ) {
    Faisceau *faisceau;
    double x, y, z;
    Vecteur3D direction, constante;

    faisceau = experience_courante->faisceau;
    constante = faisceau->orientation.position;
/*     if ( 0 != FAISCEAU_ProjeterCentre( faisceau, 0.0, &(constante.x), &(constante.z) ) ) { */
/*       constante = faisceau->orientation.position; */
/*     } else { */
/*       constante.y = 0.0; */
/*     } */
    REPERE_VecteurNormal( &(faisceau->orientation), &direction );

    glLineWidth( 2.0 );
    glColor3ub( 255, 0, 0 ); /* Rouge */

    /* 1. Tracé du centre du faisceau */
    glBegin( GL_LINES ); {
      /* Point de départ : pour t = -1.0e5, très négatif */
      x = constante.x - 1.0e8 * direction.x;
      y = constante.y - 1.0e8 * direction.y;
      z = constante.z - 1.0e8 * direction.z;
      glVertex3f( x, y, z );
      printf( "Départ : (%g, %g, %g) -", x, y, z );

      /* Point d'arrivée : pour t = +1.0e5, très positif */
      x = constante.x + 1.0e8 * direction.x;
      y = constante.y + 1.0e8 * direction.y;
      z = constante.z + 1.0e8 * direction.z;
      glVertex3f( x, y, z );
      printf( " arrivée : (%g, %g, %g)\n", x, y, z );
    } glEnd();

    if ( FAUX == ( faisceau->proprietes & FAISCEAU_PONCTUEL ) ) {
      OPTION_A_FAIRE( "Tracer le cadre du faisceau" );
    }
  }
#endif
}
