/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 2003, Emmanuel CURIS

   Lire des tables de donnes pour les lments : routines gnriques
   Format gnrique suppos de ces tables :
   - Une ligne par lment
   - Symbole de l'lment en dbut de ligne
   - Une grandeur par colonne, toutes numriques
   - Commentaires : # en dbut de ligne, ignors
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include "systeme.h"

#include "ascii_utils.h"
#include "chimie_utils.h"
#include "utils_tables.h"
#include "theorie.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Chargement des valeurs  partir des fichiers ~-~-~-~-~-~-~-~-~ */

/* -------- Ouverture du fichier, dcompte des colonnes,... -------- */

FILE *TH_OuvrirFichier( const char *nom, unsigned long *nbr_lignes, unsigned int *nbr_colonnes, double **valeurs )
{
  FILE *fichier;
  int c;

  fichier = fopen( nom , "r" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "TH_OuvrirFichier : Impossible d'ouvrir le fichier %s\n", nom);
    *valeurs = NULL;
    *nbr_colonnes = 0;
    return NULL;
  }

  /* On saute les commentaires initiaux */
  c = fgetc( fichier );
  if ( c != '#' ) {
    (void) fprintf(stderr, "TH_OuvrirFichier : Format de fichier incorrect [%s] / commence par par des commentaires\n", nom);
    (void) fclose( fichier );
    return NULL;
  }
  while ( '#' == c ) {
    c = ignore_ligne( fichier );
    if ( EOF == c ) {
      (void) fprintf(stderr, "TH_OuvrirFichier : Format de fichier incorrect [%s] / contient que des commentaires\n", nom);
      (void) fclose( fichier );
      return NULL;
    }
  }

  /* On compte le nombre de colonnes -> rservation de la table de lecture
     On sait que la 1re colonne contient les symboles -> inutile de rserver pour elle
  */
  *nbr_colonnes = compte_colonnes( fichier, 0 );
  *valeurs = TABLE_CreerTable( *nbr_colonnes - 1 );
  if ( NULL == *valeurs ) {
    (void) fprintf(stderr, "TH_OuvrirFichier : ERREUR dans TABLE_CreerTable() [%s]\n", nom);
    (void) fclose( fichier );
    return NULL;
  }

  /* On compte le nombre de lignes utiles */
  *nbr_lignes = compte_lignes( fichier, '#' );

#ifdef _DEVELOPPE_
  printf("Table de donnes : %s --- %lu lignes, %u colonnes\n", nom, *nbr_lignes, *nbr_colonnes );
#endif

  /* Tout d'est bien pass */
  return fichier;
}

/* -------- Lecture d'une ligne -------- */

int TH_LireLigne( FILE *fichier, unsigned int nbr_colonnes, double *valeurs, double *erreurs )
{
  char symbole[ 4 ] = {0, 0, 0, 0}; /* Si codage numrique de l'lement, 3 chiffres */
  int c, Z;

  if ( NULL == fichier ) return -10; /* Pas de fichier : arrive si manque un fichier d'nergies... */

  ENTREE_FONCTION( __FUNCTION__, "fichier=%p nbr_colonnes=%u valeurs=%p erreurs=%p", fichier, nbr_colonnes, valeurs, erreurs );

  /* On saute les commentaires qui pourraient prcder la ligne intressante */
  c = fgetc( fichier );
  while ( '#' == c ) {
    c = ignore_ligne( fichier );
    if ( EOF == c ) {
      (void) fprintf(stderr, "TH_LireLigne : Fin de fichier atteinte prmaturment\n");
      return -2;
    }
    c = fgetc( fichier ); /* Attention, ignore_ligne ne lit pas le 1er caractre de la ligne, mme s'il le renvoie */
  }

  /* Lecture du symbole. Premire colonne, peut tre rduite  un caractre ->
     alors symbole[1] contient l'espace de fin de colonne, qui peut tre unique : ne pas lire au-del ! */
  symbole[ 0 ] = c;
  symbole[ 1 ] = fgetc( fichier );
  c = symbole[ 1 ];
/*   c = fgetc( fichier ); */
  while ( FAUX == isspace( c ) ) {
    symbole[ 2 ] = c;
    c = fgetc( fichier );
    if ( EOF == c ) {
      (void) fprintf(stderr, "TH_LireLigne : Fin de fichier atteinte prmaturment [%s]\n", symbole);
      SORTIE_FONCTION( __FUNCTION__, "%d [Fin de fichier]", -2);
      return -2;
    }
  }

  /* De quel atome s'occupe-t-on ? */
  Z = ATM_num_atomique( symbole );
  if ( SYMBOLE_CHIMIQUE_INCONNU == Z ) {
    Z = atoi( symbole );
    if ( ( Z < HYDROGENE ) || ( Z > 120 ) ) {
      (void) fprintf(stderr, "TH_LireLigne : lment inconnu [%s]\n", symbole);
      (void) ignore_ligne( fichier );
      SORTIE_FONCTION( __FUNCTION__, "%d [lment Z inconnu / %s]", 0, symbole);
      return 0;
    }
  }

  /* On lit les diffrents nombres */
  c = lit_valeurs( fichier, nbr_colonnes - 1, 0, valeurs, erreurs );
  if ( c!= 0 ) {
    (void) fprintf(stderr, "TH_LireLigne : ERREUR dans lit_valeurs() [%s]\n", symbole);
    SORTIE_FONCTION( __FUNCTION__, "%d [Erreur dans lit_valeurs / %s]", -3, symbole);
    return -3;
  }

/*   printf("%s :", __FUNCTION__); */
/*   for ( c = 0 ; c < (int) nbr_colonnes - 1 ; c++ ) { */
/*     printf(" %g", valeurs[ c ]); */
/*   } */
/*   printf("\n"); */

  /* Chargement termin sans problme */
  SORTIE_FONCTION( __FUNCTION__, "%d [%s / %s]", Z, symbole, ATM_Z_vers_symbole( Z ));
  return Z;
}

/* ~-~-~-~-~-~-~-~-~ Manipulation des tables de valeurs ~-~-~-~-~-~-~-~-~ */

/* -------- Crer une table pour les lments entre Z_min et Z_max (inclus) -------- */

float *TH_Creer_Table( unsigned int Z_min, unsigned int Z_max )
{
  float *table;
  unsigned int i;

  if ( Z_min > Z_max ) {
    (void) fprintf(stderr, "TH_Creer_Table : Z_max < Z_min, table non-cre\n");
    return NULL;
  }

  /* Rservation */
  table = (float *) malloc( sizeof( float ) * ( Z_max - Z_min + 1 ) );
  if ( NULL == table ) {
    (void) fprintf(stderr, "TH_Creer_Table : Mmoire insuffisante [table de %u  %u]\n", Z_min, Z_max);
    return NULL;
  }

  /* Initialisation : -2.0 = valeur inconnue */
  for ( i = 0 ; i <= ( Z_max - Z_min ) ; i++ ) {
    table[ i ] = -2.0;
  }

  /* Termin */
  return table;
}

/* -------- Lire une valeur dans une table pour les lments entre Z_min et Z_max (inclus) -------- */

float TH_Table_Lire( float *table, int Z, unsigned int Z_min, unsigned int Z_max )
{
  ENTREE_FONCTION("TH_Table_Lire", "table=%p Z=%d Z_min=%u Z_max=%u", table, Z, Z_min, Z_max);

  if ( NULL == table ) {
    SORTIE_FONCTION("TH_Table_Lire", "%f [table = NULL]", 0.0);
    return 0.0;
  }

  if ( Z < (signed) Z_min ) {
    SORTIE_FONCTION("TH_Table_Lire", "%f [Z trop petit]", -1.0);
    return -1.0; /* Pas de sens ! */
  }
  if ( Z > (signed) Z_max ) {
#ifdef _DEVELOPPE_
    printf("%s : Z trop grand / Z_min = %u, Z_max = %u, Z = %u\n", __FUNCTION__, Z_min, Z_max, Z );
    SORTIE_FONCTION("TH_Table_Lire", "%f [Z trop grand]", -2.0);
#endif
    return -2.0; /* Valeur inconnue */
  }

  SORTIE_FONCTION("TH_Table_Lire", "%f", table[ Z - Z_min ]);
  return table[ Z - Z_min ];
}

/* -------- crire une valeur dans une table pour les lments entre Z_min et Z_max (inclus) -------- */

int TH_Table_Remplir( float *table, int Z, unsigned int Z_min, unsigned int Z_max, double valeur )
{
  ENTREE_FONCTION("TH_Table_Remplir", "table=%p Z=%d Z_min=%u Z_max=%u valeur=%f", table, Z, Z_min, Z_max, valeur);

  if ( NULL == table ) {
    SORTIE_FONCTION("TH_Table_Remplir", "%d [table inexistante]", -1);
    return -1;
  }

  if ( Z < (signed) Z_min ) {
    SORTIE_FONCTION("TH_Table_Remplir", "%d [Z trop petit]", +1);
    return +1;
  }
  if ( Z > (signed) Z_max ) {
    SORTIE_FONCTION("TH_Table_Remplir", "%d [Z trop grand]", +2);
    return +2;
  }

  table[ Z - Z_min ] = ( float ) valeur;

  SORTIE_FONCTION("TH_Table_Remplir", "%d / criture en %p", 0, (void *) (table + Z - Z_min) );
  return 0;
}
