/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1997 - 2002, Emmanuel CURIS

   Affichage graphique (rsultat,...)
   Modle structural : molcule (ie l'ensemble des atomes)
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "atomes.h"
#include "chemins.h"
#include "feff.h"
#include "modeles.h"
#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#ifdef AVEC_OPEN_GL_
static int fc_x = 10, fc_z = 10; /* Nombre de facettes par sphre */

static GLUquadricObj *sphere = NULL;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void TRC_AtomesModeleFEFF(Modele_FEFF *modele, int mode);
static void TRC_Atome           (Atome *atome, int mode, boolean cadre);
static void TRC_Liaison         (Liaison *liaison, int mode);
static void TRC_SegmentsChemin  (Atome *absorbeur, Chemin *chemin);

static void gl_Couleur_Atome(unsigned int num_atomique);
static void gl_Type_Liaison (unsigned int type);

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* -------- Dessin d'un seul atome -------- */

static void gl_Couleur_Atome(unsigned int num_atomique)
{
  switch (num_atomique)
  {
  case AZOTE: /* Azote bleu */
    glColor3ub(0, 0, 255);
    break;

  case CALCIUM: /* Calcium gris trs clair */
    glColor3f(0.75, 0.75, 0.75);
    break;
    
  case CARBONE: /* Carbone gris clair */
    glColor3f(0.5, 0.5, 0.5);
    break;

  case CHLORE: /* Chlore en vert vif */
    glColor3ub(0, 255, 0);
    break;

  case HYDROGENE: /* En "blanc" = gris trs clair */
    glColor3f(0.9, 0.9, 0.9);
    break;

  case OXYGENE: /* Oxygne rouge */
    glColor3ub(255, 0, 0);
    break;

  case PHOSPHORE: /* Phosphore bleu fonc */
    glColor3f(0.0, 0.0, 0.5);
    break;

  case SOUFRE: /* Soufre, jaune vif */
    glColor3ub(255, 255, 0);
    break;

  default: /* On met du violet... */
    glColor3ub(255, 0, 255);  
  }
}

/* ---- Trace un atome, avec la couleur active ---- */

static void TRC_Atome(Atome *atome, int mode, boolean cadre)
{
  double rayon;

  if (NULL == atome) return;

  if (atome->num_atomique < HYDROGENE) return; /* lment inconnu */
  if (atome->num_atomique > PLUTONIUM) return; /* > Pu : Inconnu ou trs improbable */

  if ( mode & V3D_ATOME_REDUIT ) {
    rayon = atome->rayon / 3.0;
  } else {
    rayon = atome->rayon;
  }
  rayon /= 100.0; /* Les rayons sont en pm, mais l'unit spatiale est l'angstrom */
  if (rayon <= 0.0) {
    (void) fprintf(stderr, "TRC_Atome : atome de rayon ngatif, ignor\n");
    return; /* Rayon absurde */
  }

  glPushMatrix(); /* Normalement, contient l'identit donc... */
  glTranslatef(atome->position.x, atome->position.y, atome->position.z); /* On se place l o est l'atome */
  gluSphere(sphere, rayon, fc_x, fc_z);

  if ( VRAI == cadre ) { /* On veut un cube en fil de fer autour de l'atome */
    glBegin(GL_LINE_LOOP); /* Face basse */
      glVertex3f(-rayon, -rayon, -rayon);
      glVertex3f(-rayon,  rayon, -rayon);
      glVertex3f( rayon,  rayon, -rayon);
      glVertex3f( rayon, -rayon, -rayon);
    glEnd();
    glBegin(GL_LINE_LOOP); /* Face haute */
      glVertex3f(-rayon, -rayon, rayon);
      glVertex3f(-rayon,  rayon, rayon);
      glVertex3f( rayon,  rayon, rayon);
      glVertex3f( rayon, -rayon, rayon);
    glEnd();
    glBegin(GL_LINES); /* Montants */
      glVertex3f(-rayon, -rayon, -rayon);
      glVertex3f(-rayon, -rayon,  rayon);
      glVertex3f(-rayon,  rayon, -rayon);
      glVertex3f(-rayon,  rayon,  rayon);
      glVertex3f( rayon,  rayon, -rayon);
      glVertex3f( rayon,  rayon,  rayon);
      glVertex3f( rayon, -rayon, -rayon);
      glVertex3f( rayon, -rayon,  rayon);
    glEnd();
  }
  glPopMatrix(); /* On remet celle d'avant */
}

/* -------- Ensemble des atomes, modle FEFF -------- */

static void TRC_AtomesModeleFEFF(Modele_FEFF *modele, int mode)
{
  FEFF_atome *atm_feff;

  if (NULL == modele) return;

  /* L'absorbeur a dj t dessin, avant l'appel */

  atm_feff = modele->parametres.prem_atome;
  while ( atm_feff != NULL ) {
    if ( ((mode & V3D_VOIR_SELECTION) == FAUX) || ( VRAI == atm_feff->atome->marques.visible) ) {
      /* Couleur : en fonction du potentiel choisi */
      gl_Couleur_Atome( (unsigned int) (modele->potentiels[ atm_feff->potentiel ]) );

      TRC_Atome(atm_feff->atome, mode, ATM_courant == atm_feff->atome);
    }

    atm_feff = atm_feff->suivant;
  }
}

/* -------- Liste OpenGL pour une molcule -------- */

void TRC_GL_Molecule(Molecule *molecule, int mode)
{
  Atome *atome;

  if (NULL == molecule) return;
  if (NULL == molecule->premier) {
    (void) fprintf(stderr, "dessine_molecule : molcule sans atome\n");
    return;
  }

  if ( ( (mode & V3D_VOIR_ATOMES) == FAUX ) && /* On ne cherche pas  afficher les atomes */
       ( (mode & V3D_VOIR_MODELE) == FAUX ) &&
       ( (mode & V3D_VOIR_SELECTION) == FAUX ) )
    return;

  /* Init. Open-GL */
  glCullFace(GL_BACK); /* back = intrieur des sphres, donc sans intrt */
  glEnable(GL_LIGHTING); /* Pour avoir la 3D */
  glShadeModel(GL_SMOOTH);

  sphere = gluNewQuadric();
  if (sphere == NULL) {
    (void) fprintf(stderr, "dessine_molecule : impossible de rserver une sphre OpenGL !\n");
    return;
  }
  gluQuadricDrawStyle(sphere, GLU_FILL);
  gluQuadricNormals(sphere, GLU_SMOOTH); /* Pour avoir des couleurs avec la lumire */

  if ( mode & V3D_VOIR_MODELE ) { /* On n'affiche que le modle actif */
    Absorbeur *absorbeur;

    if (NULL == molecule->absorbeurs) {
      (void) fprintf(stderr, "dessine_molecule : molcule sans absorbeur, pas de modle\n");
      gluDeleteQuadric(sphere);
      return;
    }
    absorbeur = molecule->absorbeurs + molecule->absorbeur_courant;

    /* On dessine l'absorbeur, qui appartient toujours au modle de toute faon */
    if ( ((mode & V3D_VOIR_SELECTION) == FAUX) || ( VRAI == absorbeur->absorbeur->marques.visible) ) {
      glColor3f(1.0, 0.5, 0.0); /* Absorbeur actif : en orange */

      TRC_Atome(absorbeur->absorbeur, mode, ATM_courant == absorbeur->absorbeur);
    }

    if (NULL == absorbeur->modeles) {
      (void) fprintf(stderr, "dessine_molecule : absorbeur sans modle\n");
      gluDeleteQuadric(sphere);
      return;
    }

    /* Reste  dessiner le modle... */
    switch (absorbeur->modele_actif->type)
    {
    case MODELE_EXAFS_CHEMINS:
      TRC_AtomesModeleFEFF(absorbeur->modele_actif->infos_modele.feff, mode);
      break;
    }

    gluDeleteQuadric(sphere);
    return; /* Fini pour les modles */
  }

  /* On dessine tous les atomes, grce  la liste chane de la molcule */
  for ( atome = molecule->premier ; atome != NULL ; atome = atome->suivant ) {
    Atome *absorbeur = NULL;

    if ( molecule->absorbeurs != NULL ) {
      absorbeur = molecule->absorbeurs[ molecule->absorbeur_courant ].absorbeur;
    }

    if ( (V3D_VOIR_ATOMES & mode) || ( VRAI == atome->marques.visible) ) {
      /* Choix de la couleur */
      if ( atome == absorbeur ) { /* Accolade indispensable avec le hack affreux pour Mesa / atari */
	glColor3f(1.0, 0.5, 0.0); /* Absorbeur actif : orange */
      } else if ( VRAI == atome->marques.absorbeur ) { /* Absorbeur potentiel */
	/* Il faudrait, idalement, faire quelque chose pour le signaler... */
	gl_Couleur_Atome(atome->num_atomique); /* Couleur Open-GL par dfaut */
      } else {
	gl_Couleur_Atome(atome->num_atomique); /* On met la couleur Open-GL par dfaut */
      }

      TRC_Atome(atome, mode, ATM_courant == atome);
    }
  }

  gluDeleteQuadric(sphere);
}

/* -------- Type de pointill  utiliser pour la liaison -------- */

static void gl_Type_Liaison(unsigned int type)
{
  switch (type)
  {
  case LIAISON_COVALENTE:
    glLineStipple(1, 0xFFFF); /* Liaisons covalentes : lignes continues */
    break;

  case LIAISON_HYDROGENE:
    glLineStipple(2, 0xAAAA); /* Liaison H : pointill long */
    break;

  default:
    glLineStipple(1, 0xAAAA); /* pointill court */
  }
}

/* -------- Dessin d'une liaison -------- */

static void TRC_Liaison(Liaison *liaison, int mode)
{
  if ( ( liaison->type != LIAISON_COVALENTE ) &&
       ( ( mode & V3D_LIAISON_SPECIALE ) == FAUX ) ) {
    /* On n'affiche pas les liaisons spciales */
    return;
  }

  gl_Type_Liaison(liaison->type);

  if ( ( 1 == liaison->multiplicite ) ||
       ( ( mode & V3D_LIAISON_MULTIPLE ) == FAUX ) ) {
    glBegin(GL_LINE_STRIP);
      gl_Couleur_Atome(liaison->atome_1->num_atomique); /* Couleur de l'atome 1 */
      glVertex3f(liaison->atome_1->position.x, liaison->atome_1->position.y, liaison->atome_1->position.z); /* Position de l'atome 1 */
      glVertex3f( (liaison->atome_1->position.x + liaison->atome_2->position.x) / 2.0,
		  (liaison->atome_1->position.y + liaison->atome_2->position.y) / 2.0,
		  (liaison->atome_1->position.z + liaison->atome_2->position.z) / 2.0); /* Position du milieu */
      gl_Couleur_Atome(liaison->atome_2->num_atomique); /* Couleur de l'atome 2 */
      glVertex3f(liaison->atome_2->position.x, liaison->atome_2->position.y, liaison->atome_2->position.z); /* Position de l'atome 2 */
    glEnd();
  } else {
    printf("TRC_Liaison : liaison multiple  afficher comme telle  FAIRE\n");
  }
}

/* -------- Liste Open-GL pour le schma de liaisons -------- */

void TRC_GL_Liaisons(Molecule *molecule, int mode)
{
  Liaison *liaison;
  Fragment *fragment;
  unsigned long i;

  if ( ( mode & V3D_VOIR_LIAISONS ) == FAUX ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "TRC_GL_Liaisons : on ne veut pas voir les liaisons [mode = %d]\n", mode);
#endif
    return;
  }

  glDisable(GL_LIGHTING);
  glEnable(GL_LINE_STIPPLE);
  glShadeModel(GL_FLAT);
  glLineWidth(3.0);

  /* Premire tape : liaisons dans les fragments */
  for (fragment = molecule->fragments, i = 0 ; i < molecule->nbr_fragments ;
       i++, fragment = fragment->suivant) {
    if (fragment->liaisons == NULL) /* Aucune liaison dans le fragment */
      continue;

    for (liaison = fragment->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
      TRC_Liaison(liaison, mode);
    }
  }

  /* Deuxime tape : liaisons entre fragments */
  for (liaison = molecule->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
    TRC_Liaison(liaison, mode);
  }

  glDisable(GL_LINE_STIPPLE);
}

/* -~-~-~-~-~-~-~-~ Trac des chemins -~-~-~-~-~-~-~-~ */

/* -------- Dessin du pass en paramtre --------

   Principe du dessin : trait pais allant d'un atome  l'autre, dans l'ordre
 */

void TRC_GL_Chemin(Atome *absorbeur, Chemin *chemin, unsigned int num_equivalent)
{
  if ( (NULL == absorbeur) || (NULL == chemin) ) return;

  /* On commence par dessiner le squelette, sans 3d pour tre bien visible */
  glDisable(GL_LIGHTING);
  glLineWidth(3.0);
  glColor3ub(0U, 255U, 0U); /* En vert, pais */

  TRC_SegmentsChemin(absorbeur, chemin);

  if ( num_equivalent > 0 ) { /* On veut tracer un quivalent */
    glColor3ub(255U, 0U, 0U); /* En cyan, pais */
    num_equivalent--;
    printf("TRC_GL_Chemin : On trace les quivalents\n");

    if ( NULL == chemin->equivalents ) { /* Pas d'quivalents ! */
      return;
    }

    if (num_equivalent >= chemin->nbr_equivalents) { /* On les trace tous */
      unsigned int i;

      printf("  [tous : %u]\n", chemin->nbr_equivalents);
      for (i = 0 ; i < chemin->nbr_equivalents ; i++) {
	TRC_SegmentsChemin(absorbeur, chemin->equivalents[ i ]);
      }
    } else { /* On ne trace que celui prcis */
      printf("  [le n %u]\n", num_equivalent + 1);
      TRC_SegmentsChemin(absorbeur, chemin->equivalents[ num_equivalent ]);
    }
  }

  glEnable(GL_LIGHTING);
}

static void TRC_SegmentsChemin(Atome *absorbeur, Chemin *chemin)
{
  Atome *atome = NULL;
  GLfloat x_dep, y_dep, z_dep;
  unsigned int i = 0;

  x_dep = (GLfloat) (absorbeur->position.x); /* Premier atome du chemin : forcment l'absorbeur */
  y_dep = (GLfloat) (absorbeur->position.y);
  z_dep = (GLfloat) (absorbeur->position.z);

  glBegin(GL_LINE_LOOP);
    glVertex3f(x_dep, y_dep, z_dep);

  for (i = 0 ; i < chemin->nbr_atomes ; i++) {
    atome = chemin->atomes[i];
    glVertex3f( (GLfloat) (atome->position.x), (GLfloat) (atome->position.y), (GLfloat) (atome->position.z));
  }
  glEnd();

  /* Maintenant, il faut positionner les triangles pour faire les flches... */
}

#endif /* AVEC_OPEN_GL_ */
