/* Bibliothèque d'interface
   (C) 1998-2002, Emmanuel CURIS

   Fonctions générales de traitement des chaînes de caractère
   Découpage d'un texte en plusieurs lignes, de longueur maximale imposée
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h> /* Pour is... */

#include "deboguage.h"
#include "txt_utils.h"

/* #undef _DEVELOPPE_ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned long TXT_CompterSeparateurs(const char *texte, int separateur);

static int TXT_DecomposerChaine_Espaces(const char *texte, char **table_mots);
static int TXT_DecomposerChaine_Impose (const char *texte, char **table_mots, int separateur);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Construit un tableau avec chaque ligne créée --------
   Le tableau est rempli avec la chaîne originale, qui est MODIFIÉE avec des 0 bien placés
   ==> NE PAS LIBÉRER LES LIGNES DU TABLEAU
 */

char **TXT_GenereLignes(char *texte, unsigned int lg_max, int separateur, unsigned long *nbr_lignes)
{
  unsigned int nombre_lignes = 0, i = 0;
  char **lignes, *deb_chaine;

  if (NULL == texte) return NULL;
  if (NULL == nbr_lignes) return NULL;
  if ( lg_max < 2 ) {
    (void) fprintf(stderr, "TXT_GenereLignes : lignes de 2 caractères au moins !\n");
    return NULL;
  }

  /* On compte le nombre de lignes nécessaires */
  deb_chaine = texte;
  do {
    char *pos_separateur = NULL;

    nombre_lignes++;
    if (separateur != 0) {
      pos_separateur = strchr(deb_chaine, separateur);
    }
    if ( (pos_separateur != NULL) && ( (pos_separateur - deb_chaine) < (long) lg_max ) ) {
      /* Retour à la ligne forcé */
      deb_chaine = pos_separateur + 1;
    } else if ( strlen(deb_chaine) <= lg_max ) { /* Dernière ligne, plus courte */
      deb_chaine = NULL;
    } else { /* Texte plus long que la ligne : on coupe */
      char *fin_ligne;
      int c;
      
      fin_ligne = deb_chaine + lg_max;
      c = *fin_ligne;
      while ( (c != ' ') && (c != '-') && (c != '/') ) { /* Coupure préférentielle sur espace, - ou / */
	fin_ligne--;
	if (fin_ligne <= deb_chaine) { /* Pas de coupure automatique simple -> après 50 caractères, arbitraire */
	  fin_ligne = deb_chaine + lg_max;
	  break;
	}
	c = *fin_ligne;
      }
      deb_chaine = fin_ligne + 1;
    }
  } while ( deb_chaine != NULL );

  /* On réserve pour les lignes */
  lignes = (char **) calloc( (size_t) nombre_lignes, sizeof(char *));
  if ( NULL == lignes ) {
    (void) fprintf(stderr, "TXT_GenereLignes : mémoire insuffisante pour générer la table des lignes !\n");
    return NULL;
  }
  *nbr_lignes = nombre_lignes;

  /* On initialise le tableau */
  for (i = 0, deb_chaine = texte ; i < nombre_lignes ; i++) {
    if (NULL == deb_chaine) { /* Ligne vide */
/* printf("Ligne %d (objet %d sur %d) vide...\n", i, bouton + i, nbr_objets - 1); */
      lignes[i] = NULL;
    } else if ( strlen(deb_chaine) <= lg_max ) { /* La chaîne tient */
      char *pos_separateur = NULL;
/*      printf("Texte ligne %d (objet %d sur %d): %s\n", i, bouton + i, nbr_objets - 1, deb_chaine); */
      lignes[i] = deb_chaine;

      if ( separateur != 0 ) {
	pos_separateur = strchr(deb_chaine, separateur);
      }
      if (NULL != pos_separateur) {
	*pos_separateur = 0;
	deb_chaine = pos_separateur + 1;
      } else
	deb_chaine = NULL;
    } else {/* Ligne trop longue : il faut découper */
      char *fin_chaine = deb_chaine + lg_max, *pos_separateur = NULL;

/*      printf("Texte ligne %d avant coupure : %s\n", i, deb_chaine); */
      if ( separateur != 0 ) {
	pos_separateur = strchr(deb_chaine, separateur);
      }
      if ( (NULL != pos_separateur) && (pos_separateur <= fin_chaine) ) { /* retour à la ligne forcé... */
/*	printf("  coupure imposée : barre |\n"); */
	fin_chaine = pos_separateur;
      } else {
/*	printf("  coupure automatique au 50e caractère : %c (%d)\n", *fin_chaine, *fin_chaine); */
	while ( (*fin_chaine != ' ') && (*fin_chaine != '-') && (*fin_chaine != '/') ) {
	  fin_chaine--;
	  if (fin_chaine == deb_chaine) {
	    (void) fprintf(stderr, "BI_ERR_alerte : chaîne insécable > %u caractères : %s\n", lg_max, deb_chaine);
	    fin_chaine = deb_chaine + lg_max;
	  }
	}
/*	printf("  coupure automatique au %u caractère : %c (%d)\n", fin_chaine - deb_chaine, *fin_chaine, *fin_chaine); */
      }
      lignes[i] = deb_chaine;
      *fin_chaine = 0; /* Pour avoir une belle fin de ligne */
/*      printf("Texte ligne %d (objet %d sur %d): %s\n", i, bouton + i, nbr_objets - 1, deb_chaine); */
      deb_chaine = fin_chaine + 1;
    }
  }

#ifdef _DEVELOPPE_
/*   for (i = 0 ; i < nombre_lignes ; i++) { */
/*     printf("Ligne %u : %s\n", i, lignes[i]); */
/*   } */
#endif

  return lignes;
}

/* -------- Décomposer une ligne en table de "mots" --------
    Renvoie une table de chaînes ; table réservée avec malloc
    Le texte est copié dans table[ 0 ] et les séparateurs sont remplacés par des 0
     ==> il n'y a QUE table et table[ 0 ] à libérer !
 */

char **TXT_DecomposerChaine(const char *texte, int separateur, unsigned long *nbr_elements)
{
  char **table_mots, *copie_texte;
  int retour;
  unsigned long nbr_separateurs;

  if ( NULL == texte ) return NULL;
  if ( NULL == nbr_elements ) {
    (void) fprintf(stderr, "%s,%s l.%u : nbr_elements = NULL\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  if ( 0 == separateur ) { 
    /* On fait sauter les espaces initiaux si le séparateur est n'importe quoi comme espace */
    while ( isspace( *texte ) )
      texte++;
    if ( 0 == *texte ) {
      (void) fprintf(stderr, "%s,%s l.%u : Texte vide (ou que des espaces)\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
  }

  /* Combien de séparateurs dans le texte ? */
  nbr_separateurs = TXT_CompterSeparateurs( texte, separateur );
#ifdef _DEVELOPPE_
  (void) printf("%s : On a trouvé %lu séparateurs => %lu champs\n", __FUNCTION__, nbr_separateurs, nbr_separateurs + 1);
#endif

  /* On réserve la table */
  table_mots = ( char ** ) calloc( nbr_separateurs + 1, sizeof( char * ) );
  if ( NULL == table_mots ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver la table des mots\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On copie le texte */
  copie_texte = strdup( texte );
  if ( NULL == copie_texte ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de faire la copie du texte\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( table_mots );
    return NULL;
  }

  table_mots[ 0 ] = copie_texte;
  *nbr_elements = nbr_separateurs + 1; /* n séparateurs ==> n+1 valeurs, puisque avant et après ! */

  if ( 0 == nbr_separateurs ) { /* la phrase = un seul élément */
    *nbr_elements = 1;
    return table_mots;
  }

  if ( 0 == separateur ) {
    retour = TXT_DecomposerChaine_Espaces( texte, table_mots );
  } else {
    retour = TXT_DecomposerChaine_Impose( texte, table_mots, separateur );
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la décomposition [texte = %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, texte);
    free( table_mots[ 0 ] );
    free( table_mots );
    return NULL;
  }

  /* La décomposition s'est bien passée */
  return table_mots;
}

/* -------- Compter le nombre de séparateurs -------- */

static unsigned long TXT_CompterSeparateurs(const char *texte, int separateur)
{
  unsigned long nbr_espaces = 0;

  if ( separateur != 0 ) {
    /* Séparateur = 1 caractère, on compte le nombre de fois que ce caractère apparaît */
    return TXT_compte_caractere( texte, separateur );
  }

  /* Ici, un lot d'espaces sert de séparateur */
  /* a. On fait sauter les espaces initiaux */
  while ( isspace( *texte ) )
    texte++;

  /* b. On analyse jusqu'à la fin de la chaîne */
  while ( *texte != 0 ) {
    while ( FAUX == isspace( *texte ) ) { /* On est sur autre chose qu'un espace */
      texte++;
      if ( 0 == *texte ) { /* La chaîne se termine sur autre chose qu'un espace */
	return nbr_espaces;
      }
    }

    /* Ici, on a trouvé un espace : un séparateur de plus */
    nbr_espaces++;
    while( isspace( *texte ) ) {
      texte++;
    }
  }

  /* Si on sort, c'est que la chaîne se finit par des espaces, on néglige les finaux */
  return (nbr_espaces - 1);
}

/* -------- Décomposition en table de mots, séparateur = espaces (simples ou multiples) -------- */

static int TXT_DecomposerChaine_Espaces(const char *texte, char **table_mots)
{
  char *position, *debut_item;
  unsigned long item = 0;

  debut_item = (char *) texte; /* table_mots[ 0 ]; / * Début du champ */
  position = debut_item;
  do {
    /* On saute le champ */
    while ( FAUX == isspace( *position ) ) {
      position++;
      if ( 0 == *position ) {
	table_mots[ item ] = debut_item; /* On mémorise le champ */
	return 0; /* Terminé */
      }
    }

    /* Si on est là : séparateur atteint */
    table_mots[ item ] = debut_item; /* On mémorise le champ */
    *position = 0; /* On le coupe, sans garder les espaces */
#ifdef _DEVELOPPE_
    (void) printf("%s : champ %lu = %s [%s => reste %s]\n", __FUNCTION__, item, table_mots[ item ], texte, position + 1);
#endif
    item++; /* On s'occupe du suivant */
    position++;

    while ( isspace( *position ) ) {
      position++; /* On fait sauter tous les espaces */
    }

    debut_item = position; /* On a trouvé le début du champ suivant */
  } while ( *position != 0 ); /* Tant que l'on n'a pas atteint la fin */

  /* Tout s'est bien passé */
  return 0;
}

static int TXT_DecomposerChaine_Impose(const char *texte, char **table_mots, int separateur)
{
  const char *position;
  char *copie_texte, *debut_item;
  unsigned long item = 0;

  copie_texte = table_mots[ 0 ];

  for ( position = texte, debut_item = copie_texte, item = 0 ; *position != 0 ; position++ ) {
    if ( *position == separateur ) { /* On tombe sur un séparateur */
      item++; /* On passe à l'item suivant */

      debut_item = copie_texte + ( position - texte ) + 1; /* L'item suivant commence _après_ le séparateur */
      copie_texte[ position - texte ] = 0; /* 0 pour marquer la fin du morceau */

      table_mots[ item ] = debut_item;
#ifdef _DEVELOPPE_
      (void) printf("%s : champ %lu = %s [%s => reste %s]\n", __FUNCTION__, item, table_mots[ item - 1 ], texte, position);
#endif
    }
  }
#ifdef _DEVELOPPE_
  (void) printf("%s : champ %lu = %s [%s => reste %s]\n", __FUNCTION__, item + 1, table_mots[ item ], texte, position);
#endif

  /* Tout s'est bien passé */
  return 0;
}
