/* Bibliothèque d'interface
   (C) 1998-2015, Emmanuel CURIS

   Fonctions permettant de faciliter la gestion de l'UTF-8 dans les chaînes de caractères
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h> /* Pour is... */
#include <math.h>  /* Pour log10 */

#include "txt_utils.h"

char TXT_UTF8_vers_nombre(const char *texte, unsigned int *n_octets)
{
  char c;

  /* Par défaut, inconnu => on ne convertit pas */
  n_octets = 0;

  /* Quel est le 1er caractère ? */
  c = *texte;

  /* Les nombres ASCII classiques */
  if ( ( c >= '0' ) && ( c <= '9' ) ) {
    *n_octets = 1;
    return c;
  }

  /* Les caractères en exposant */
  if ( '\xC2' == c ) {		/* Codes en C2 xx : ¹, ² et ³ (car existent en latin-1) */
    c = *(texte + 1);

    *n_octets = 2;
    if ( '\xB9' == c ) return '1';	/* C2 B9 = ¹ */
    if ( '\xB2' == c ) return '2';	/* C2 B2 = ² */
    if ( '\xB3' == c ) return '3';	/* C2 B3 = ³ */
    *n_octets = 0;
  }

  if ( '\xE2' == c ) {
    c = *(texte + 1);

    if ( '\x81' == c ) {	/* Codes en E2 81 xx : ⁰ ⁴ ⁵ ⁶ ⁷ ⁸ ⁹ ⁺ ⁻ ⁽ ⁾ ⁼ */
      c = *(texte + 2);

      *n_octets = 3;
      if ( '\xB0' == c ) return '0'; /* E2 81 B0 = ⁰ */
      if ( '\xB4' == c ) return '4'; /* E2 81 B4 = ⁴ */
      if ( '\xB5' == c ) return '5'; /* E2 81 B5 = ⁵ */
      if ( '\xB6' == c ) return '6'; /* E2 81 B6 = ⁶ */
      if ( '\xB7' == c ) return '7'; /* E2 81 B7 = ⁷ */
      if ( '\xB8' == c ) return '8'; /* E2 81 B8 = ⁸ */
      if ( '\xB9' == c ) return '9'; /* E2 81 B9 = ⁹ */
      if ( '\xBA' == c ) return '+'; /* E2 81 BA = ⁺ */
      if ( '\xBB' == c ) return '-'; /* E2 81 BB = ⁻ */
      if ( '\xBD' == c ) return '('; /* E2 81 BD = ⁽ */
      if ( '\xBE' == c ) return ')'; /* E2 81 BE = ⁾ */
      *n_octets = 0;		   /* Code non-convertible */
    }

    /* Les chiffres en indices... */
    if ( '\x82' == c ) {	/* Codes en E2 82 xx : ₀ ₁ ₂ ₃ ₄ ₅ ₆ ₇ ₈ ₉ ₊ ₋ ₌ ₍ ₎ */
      c = *(texte + 2);

      *n_octets = 3;
      if ( '\xB0' == c ) return '0'; /* E2 81 B0 = ₀ */
      if ( '\xB1' == c ) return '1'; /* E2 81 B1 = ₁ */
      if ( '\xB2' == c ) return '2'; /* E2 81 B2 = ₂ */
      if ( '\xB3' == c ) return '3'; /* E2 81 B3 = ₃ */
      if ( '\xB4' == c ) return '4'; /* E2 81 B4 = ₄ */
      if ( '\xB5' == c ) return '5'; /* E2 81 B5 = ₅ */
      if ( '\xB6' == c ) return '6'; /* E2 81 B6 = ₆ */
      if ( '\xB7' == c ) return '7'; /* E2 81 B7 = ₇ */
      if ( '\xB8' == c ) return '8'; /* E2 81 B8 = ₈ */
      if ( '\xB9' == c ) return '9'; /* E2 81 B9 = ₉ */
      if ( '\xBA' == c ) return '+'; /* E2 81 BA = ₊ */
      if ( '\xBB' == c ) return '-'; /* E2 81 BB = ₋ */
      if ( '\xBD' == c ) return '('; /* E2 81 BD = ₍ */
      if ( '\xBE' == c ) return ')'; /* E2 81 BE = ₎ */
      *n_octets = 0;		   /* Code non-convertible */
    }

    /* Le signe moins, − */
    if ( '\x88' == c ) {
      c = *(texte + 2);

      *n_octets = 3;
      if ( '\x92' == c ) return '-'; /* E2 88 92 = − */
      *n_octets = 0;		   /* Code non-convertible */
    }
  }

  /* Si l'on n'a rien trouvé : on renvoie 0 */
  return 0;
}

char TXT_UTF8_vers_operateur(const char *texte, unsigned int *n_octets)
{
  char c;

  if ( NULL == n_octets ) {
    (void) fprintf( stderr, "%s,%s l.%u : n_octets = NULL, interdit\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    /* FIN_ANTICIPEE( FIN_ERREUR_GRAVE ); */
  }

  /* Par défaut, inconnu =>  on ne convertit pas */
  *n_octets = 0;

  /* Quel est le 1er caractère ? */
  c = *texte;

  /* Les opérateurs classiques : + - / * () [] {} = */
  if ( ( '+' == c ) || ( '-' == c ) || ( '/' == c ) || ( '*' == c ) ||
       ( '(' == c ) || ( ')' == c ) || ( '[' == c ) || ( ']' == c ) || ( '{' == c ) || ( '}' == c ) ||
       ( '=' == c ) ) {
    *n_octets = 1;
    return c;
  }

  /* Les variantes tyographiques des opérateurs usuels */
  if ( '\xC2' == c ) {
    c = *(texte + 1);

    *n_octets = 2;
    if ( '\xB7' == c ) return '*';	/* C2 B7 = · */
    *n_octets = 0;
  }
  if ( '\xC3' == c ) {
    c = *(texte + 1);

    *n_octets = 2;
    if ( '\x97' == c ) return '*';	/* C3 97 = × */
    if ( '\xB7' == c ) return '/';	/* C3 B7 = ÷ */
    *n_octets = 0;
  }

  /* Les opérateurs en indice ou exposant */
  if ( '\xE2' == c ) {
    c = *(texte + 1);

    if ( '\x81' == c ) {	/* Codes en E2 81 xx : ⁺ ⁻ ⁽ ⁾ ⁼ */
      c = *(texte + 2);

      *n_octets = 3;
      if ( '\xBA' == c ) return '+'; /* E2 81 BA = ⁺ */
      if ( '\xBB' == c ) return '-'; /* E2 81 BB = ⁻ */
      if ( '\xBC' == c ) return '='; /* E2 81 BB = = */
      if ( '\xBD' == c ) return '('; /* E2 81 BD = ⁽ */
      if ( '\xBE' == c ) return ')'; /* E2 81 BE = ⁾ */
      *n_octets = 0;		   /* Code non-convertible */
    }

    /* Les chiffres en indices... */
    if ( '\x82' == c ) {	/* Codes en E2 82 xx : ₀ ₁ ₂ ₃ ₄ ₅ ₆ ₇ ₈ ₉ ₊ ₋ ₌ ₍ ₎ */
      c = *(texte + 2);

      *n_octets = 3;
      if ( '\xBA' == c ) return '+'; /* E2 81 BA = ₊ */
      if ( '\xBB' == c ) return '-'; /* E2 81 BB = ₋ */
      if ( '\xBC' == c ) return '='; /* E2 81 BB = = */
      if ( '\xBD' == c ) return '('; /* E2 81 BD = ₍ */
      if ( '\xBE' == c ) return ')'; /* E2 81 BE = ₎ */
      *n_octets = 0;		   /* Code non-convertible */
    }
  }

  /* Si l'on n'a rien trouvé : on renvoie 0 */
  return 0;
}

/* -------- Compter le nombre de _caractères_ d'une chaîne codée en UTF-8 -------- */

unsigned long TXT_UTF8_CompterCaracteres(const char *texte)
{
  int lg_caractere;
  unsigned long i, nombre_caracteres, nombre_octets;
  const char *c;

  if ( NULL == texte ) return 0; /* Aucune chaîne passée... */

  /* Combien d'_octets_ dans la chaîne ? */
  nombre_octets = strlen( texte );

  /* On se place au début de la chaîne */
  i = 0; nombre_caracteres = 0 ; c = texte;
  while ( i < nombre_octets ) {
    if ( 0 == *c ) break;	/* Fin de la chaîne atteinte */

    /* Longueur du caractère UTF-8 qui démarre la chaîne */
    lg_caractere = TXT_UTF8_LongueurCaractere( c );
    if ( 0 == lg_caractere ) {
      (void) fprintf( stderr, "%s,%s l.%u : Situation inattendue, pointeur devient NULL en milieu de chaîne\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      /* FIN_ANTICIPEE( FIN_ERREUR_GRAVE ); */
      return 0;
    }
    if ( lg_caractere == -1 ) lg_caractere = 1; /* Code UTF-8 incorrect : compté comme un seul caractère [latin-1 ?] */

    /* Un caractère de plus */
    nombre_caracteres++;

    /* On avance du nombre de caractères indiqués dans la chaîne */
    i += lg_caractere;
    c += lg_caractere;
  }

  /* On renvoie le nombre de caractères trouvé */
  return nombre_caracteres;
}

/* -------- Compter le nombre d'octets du premier caractère d'une chaîne codée en UTF-8 -------- */

int TXT_UTF8_LongueurCaractere(const char *caractere)
{
  unsigned short int bit_7, bit_6, bit_5, bit_4, bit_3;

  /* Contrôles triviaux préliminaires */
  if ( NULL == caractere  ) return 0; /* Aucune chaîne passée... */
  if ( 0    == *caractere ) return 1; /* 0 final = un seul octet, même si pas vraiment un caractère... */

  /* Le 7e bit est-il mis ? [0x80 = 1000 0000] */
  bit_7 = ( *caractere & 0x80U );
  if ( 0 == bit_7 ) return 1; /*  Codes de 00 à 7F : code ASCII, occupe un seul octet */

  /* Le 6e bit est-il mis ? [0x40 = 0100 0000] */
  bit_6 = ( *caractere & 0x40U);
  if ( 0 == bit_6 ) {
    /* 10xxxxxx, 80 à C1 : invalide comme 1er octet d'une séquence UTF-8
       mais pourrait être un octet de continuation si erreur auparavant
       plus probablement : chaîne en latin-1 transmise, non-convertie...

       Caractères latin-1 couverts : 80 -> 9F : contrôle
				     A0       : espace insécable
				     A1 -> AC : ¡¢£¤¥¦§¨©ª«¬
				     AD       : césure optionnelle invisible
				     AE -> BF : ®¯°±²³´µ¶·¸¹º»¼½¾¿
    */
    (void) fprintf( stderr, "%s,%s l.%u : Code UTF-8 incorrect en 1re position [%2hhX], considéré comme du latin-1 résiduel\n",
		    __FILE__, __FUNCTION__, __LINE__, *caractere );
    return -1;
  }
  if ( ( '\xC0' == *caractere ) || ( '\xC1' == *caractere ) ) {
    /* 11000000 et 11000001 = interdits pour raisons de sécurite (formes « longues » de l'ASCII)
       Caractères latin-1 couverts : C0 : À --- C1 : Á
     */
    (void) fprintf( stderr, "%s,%s l.%u : Code UTF-8 interdit [%2hhX], considéré comme du latin-1 résiduel\n",
		    __FILE__, __FUNCTION__, __LINE__, *caractere );
    return -1;
  }

  /* Si on est là, séquence multi-caractères, les bits 6 et 7 sont mis :
     séquence 11xxxxxx 10xxxxxx ...
     donc, doit avoir au-moins 2 octets, est-ce le cas ? */
  if ( 0 == *(caractere + 1) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 incomplète [%2hhX terminal] - compté comme un seul octet\n",
		    __FILE__, __FUNCTION__, __LINE__, *caractere );
    return -1;
  }

  /* Le 2e octet doit commencer par 10, soit être compris entre 80 et BF */
  if ( 0 == ( *(caractere + 1) & 0x80 ) ) {
    /* 11xxxxxx 0xxxxxxx : impossible, le 2e devrait commencer par 10... */
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX]\n",
		    __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1) );
    return -1;
  }
  if ( *(caractere + 1) & 0x40 ) {
    /* 11xxxxxx 11xxxxxx : impossible, le 2e devrait commencer par 10... */
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX]\n",
		    __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1) );
    return -1;
  }

  /* Le 5e bit est-il mis ? [0x20 = 0010 0000]*/
  bit_5 = ( *caractere & 0x20U );
  if ( 0 == bit_5 ) {
    /* 110xxxxx, C2 à DF  : 1er octet d'une séquence de 2 octets, normalement...
       On a déjà vérifié que l'octet suivant n'est pas nul et commençait bien par 10xxxxxx.
     */

    /* Pas de contrainte supplémentaires... */
    /* 110xxxxx 10xxxxxx : séquence UTF-8 correcte pour deux octets */
    return 2;
  }

  /* Si on est là : séquence multi-caractères sur au moins 3 octets
     séquence 111xxxxx 10xxxxxx 10xxxxxx ...
     donc, doit avoir au-moins 3 octets, est-ce le cas ? */
  if ( 0 == *(caractere + 2) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 incomplète [%2hhX %2hhX terminal]\n",
		    __FILE__, __FUNCTION__, __LINE__, *caractere, *(caractere +1) );
    return -1;
  }

  /* Le 3e octet doit commencer par 10, soit être compris entre 80 et BF */
  if ( 0 == ( *(caractere + 2) & 0x80 ) ) {
    /* 111xxxxx 10xxxxxx 0xxxxxxx : impossible, le 3e devrait commencer par 10... */
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX]\n",
		    __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2) );
    return -1;
  }
  if ( *(caractere + 2) & 0x40 ) {
    /* 111xxxxx 10xxxxxx 11xxxxxx : impossible, le 3e devrait commencer par 10... */
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX]\n",
		    __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2) );
    return -1;
  }

  /* Le 4e bit est-il mis ? [0x10 = 0001 0000]*/
  bit_4 = ( *caractere & 0x10U );
  if ( 0 == bit_4 ) {
    /* 1110xxxx, E0 à EF : 1er octet d'une séquence de 3 octets, normalement...
       Les deux octets suivants existent bien (vérifié juste avant)
       On a bien une séquence de la forme 1110xxxx 10xxxxxx 10xxxxxx
     */

    /* Contraintes supplémentaires :
       - si 1er octet = E0, le 2e octet doit avoir un bit 5 à 1
     */
    if ( ( '\xE0' == *caractere ) && ( 0 == ( *(caractere + 1) & 0x40U ) ) ) {
      (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX]\n",
		      __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2) );
      return -1;
    }

    /* - si 1er octet = ED, le 2e octet doit avoir un bit 5 à 0
     */
    if ( ( '\xED' == *caractere ) && ( *(caractere + 1) & 0x40U ) ) {
      (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX]\n",
		      __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2) );
      return -1;
    }

      /* 110xxxxx 10xxxxxx 10xxxxxx : séquence UTF-8 correcte pour trois octets
       */
    return 3;
  }

  /* Si on est là : séquence multi-caractères sur au moins 4 octets
     séquence 1111xxxx 10xxxxxx 10xxxxxx 10xxxxxx ...
     donc, doit avoir au-moins 4 octets, est-ce le cas ? */
  if ( 0 == *(caractere + 3) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 incomplète [%2hhX %2hhX %2hhX terminal]\n",
		    __FILE__, __FUNCTION__, __LINE__, *caractere, *(caractere + 1), *(caractere + 2) );
    return -1;
  }

  /* Le 4e octet doit commencer par 10, soit être compris entre 80 et BF */
  if ( 0 == ( *(caractere + 3) & 0x80 ) ) {
    /* 1111xxxx 10xxxxxx 10xxxxxx 0xxxxxxx : impossible, le 4e devrait commencer par 10... */
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX]\n",
		    __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2) );
    return -1;
  }
  if ( *(caractere + 3) & 0x40 ) {
    /* 1111xxxx 10xxxxxx 10xxxxxx 11xxxxxx : impossible, le 4e devrait commencer par 10... */
    (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX]\n",
		    __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2) );
    return -1;
  }

  /* Le bit 3 est-il mis ? [0x08 = 0000 1000] */
  bit_3 = ( *caractere & 0x08 );
  if ( 0 == bit_3 ) {
    /* 11110xxx, F0 à F7 : 1er octet d'une séquence de 4 octets, normalement...
       Les trois octets suivants existent bien (vérifié juste avant)
       On a bien une séquence de la forme 1110xxxx 10xxxxxx 10xxxxxx 10xxxxxx
     */

    /* Contraintes supplémentaires :
       - le 1er octet ne peut pas valoir F5, F6, F7, F8, F9, FA, FB, FC, FF
     */
    if ( ( *caractere != '\xF0' ) && ( *caractere != '\xF1' ) && ( *caractere != '\xF2' ) &&
	 ( *caractere != '\xF3' ) && ( *caractere != '\xF4' ) ) {
      (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX %2hhX]\n",
		      __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2), *(caractere + 3) );
      return -1;
    }

    /*  - si 1er octet = F0, le 2e octet doit être compris entre 0x90 et 0xBF
	  soit 2e octet : bit 5 à 1 ou ( bit 5 à 0 et bit 4 à 1) ==> bits 4 et 5 à 0 interdits
     */
    if ( ( '\xF0' == *caractere ) && ( 0 == ( *(caractere + 1) & 0x30U ) ) ) {
      (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX %2hhX]\n",
		      __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2), *(caractere + 3) );
      return -1;
    }

    /*  - si 1er octet = F4, le 2e octet doit être compris entre 0x80 et 0x8F
	  soit 2e octet : bits 4 et 5 à 0 tous les deux
     */
    if ( ( '\xF4' == *caractere ) && ( *(caractere + 1) & 0x30U ) ) {
      (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX %2hhX]\n",
		      __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2), *(caractere + 3) );
      return -1;
    }

    /* Séquence valide 11110xx 10xxxxxx 10xxxxxx 10xxxxxx sur 4 octets */
    return 4;
  }

  /* Si on est là : séquence sur + de 4 octets, interdite actuellement
     11111xx 10xxxxxx 10xxxxxx 10xxxxxx ... 

     Couvre les caractères latin-1 F8 -> FF : øùúûüýþÿ
  */
  (void) fprintf( stderr, "%s,%s l.%u : Séquence UTF-8 invalide [%2hhX %2hhX %2hhX %2hhX]\n",
		  __FILE__, __FUNCTION__, __LINE__,  *caractere, *(caractere + 1), *(caractere + 2), *(caractere + 3) );
  return -1;
}
