/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999, Emmanuel CURIS
   Cration : le 10 juin 1999

   Corrections des spectres exprimentaux :
    - correction d'nergie (recalibration)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"

#include "cristaux.h"
#include "extrait.h" /* Pour cherche_e0 */
#include "science.h"

#include "corr_nrj.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_CRISTAL 11

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GER_bragg(unsigned int objet, int x, int y);
static void prepare_arbre_bragg(void);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Recalibration en nergie ~-~-~-~-~-~-~-~- */

void corr_energie(Graphe *spectre)
{
  double e0_th, e0_exp;
  register double *x;
  unsigned long i;

  if (spectre == NULL) return;

  if ( arbres[CORR_ENERGIE].drapeaux.prepare == FAUX ) {
    prepare_arbre_bragg();
    e0_exp = cherche_e0(spectre, FAUX); /* On cherche le E0 exprimental */
    OBJ_MettreNombre( CORR_ENERGIE, SEUIL_MESURE , e0_exp, 0.0 );
    OBJ_MettreNombre( CORR_ENERGIE, SEUIL_ATTENDU, e0_exp, 0.0 ); /* On suppose le mme par dfaut */
  }
  
  if ( ARB_Bloquant( CORR_ENERGIE, titre_arbre_corrige_nrj, principale, 0 ) == COR_NRJ_ANNULE)
    return; /* On annule le calcul */

  e0_exp = OBJ_LireNombre(CORR_ENERGIE, SEUIL_MESURE , NULL ); /* E0 exprimental */
  e0_th  = OBJ_LireNombre(CORR_ENERGIE, SEUIL_ATTENDU, NULL ); /* E0 thorique    */
/*   if ( e0_exp == e0_th ) { */
/* #ifdef _DEVELOPPE_ */
/*     (void) fprintf(stderr, "%s : pas de correction demandee !\n", __FUNCTION__); */
/* #endif */
/*     return; */
/*   } */

  HIST_Titre(1, "Correction des "e_AIGU"nergies", "CORR");
  HIST_Texte(1, 0, "Spectre trait"e_AIGU" : %s\n", spectre->nom);
  HIST_Texte(3, 0, E_AIGU"nergie mesur"e_AIGU"e : %f eV\n", e0_exp);
  HIST_Texte(3, 0, E_AIGU"nergie attendue : %f eV\n", e0_th);

  if ( OBJ_choisit(CORR_ENERGIE, COR_NRJ_DECALE, QUESTION) == VRAI ) { /* Simple dcalage */
    double delta_x = e0_th - e0_exp;

    GRF_FacteurAdditif( spectre, delta_x, 0.0, AXE_X );

    HIST_Texte(2, 0, "M"e_AIGU"thode : translation des X : E' = E + %f\n", delta_x);
  } else { /* On utilise la loi de Bragg */
    double e_v_angle, dist_plans;
    double d_theta;
    double theta_exp;

    if ( (e0_th == 0.0) || (e0_exp == 0.0) ) {
      EXF_ERREUR(DIV_ZERO, txt_erreur_nrj_nulle, txt_err_calc_annule);
      return;
    }

    dist_plans = OBJ_LireNombre( CORR_ENERGIE, COR_NRJ_DIST, NULL ); /* Distance interrticulaire, en A */
    if (dist_plans == 0.0) {
      EXF_ERREUR(DIV_ZERO, txt_erreur_d_nulle, txt_err_calc_annule);
      return;
    }

    e_v_angle  = PLANCK * (VITESSE_LUMIERE * M_V_ANGSTROM); /* Conversion theta <-> nergie */
    e_v_angle /= 2.0 * dist_plans * EV_V_JOULE; /* Conversion eV <-> J */

    /* Contrles de validit */
    if ( fabs( e_v_angle / e0_th ) > 1.0 ) {
      (void) BI_ERR_alerte( NULL, &ERR_EnergieIncorrecte, VRAI, e0_th, dist_plans, e_v_angle );
    }
    if ( fabs( e_v_angle / e0_exp ) > 1.0 ) {
      (void) BI_ERR_alerte( NULL, &ERR_EnergieIncorrecte, VRAI, e0_exp, dist_plans, e_v_angle );
    }

    /* Quelle nergie minimale est acceptable ? */
    if ( spectre->x_min < e_v_angle ) {
      (void) BI_ERR_alerte( NULL, &ERR_EnergieIncorrecte, VRAI, spectre->x_min, dist_plans, e_v_angle );
      return;
    }
    
    d_theta = asin( e_v_angle / e0_th) - asin(e_v_angle / e0_exp); /* Correction */
	  
    x = spectre->x;
    for (i = 0 ; i < spectre->nombre_point ; i++, x++) {
      if (*x == 0.0) {
	EXF_ERREUR(DIV_ZERO, txt_erreur_abs_nulle, txt_err_val_inchangee);
      } else {
	double truc;

	theta_exp = asin(e_v_angle / *x);
	truc = sin(d_theta + theta_exp);
	if (truc == 0.0)
	  EXF_ERREUR(DIV_ZERO, txt_erreur_angle_nul, txt_err_val_inchangee);
	else
	  *x = e_v_angle / truc;
      }
    }

    HIST_Texte(2, 0, "M"e_AIGU"thode : utilisation de la loi de Bragg");
    HIST_Texte(3, 0, " (d = %f "ANGSTROM")", dist_plans);
    HIST_Texte(2, 0, "\n");

    GRF_bornes(spectre, DESORDONNE, DESORDONNE); /* On recalcule les bornes */
  }
  spectre->drapeaux.sauve = FAUX; /* Puisque l'on vient de changer les X */
  GRF_modifier(spectre, spectre, FAUX);
}

/* -------- Gestion de l'interface -------- */

static int GER_bragg(unsigned int objet, int x, int y)
{
  int cristal;

  switch (objet)
  {
  case CE_CRISTAL:
    cristal = SS_menu_texte(noms_cristaux, NOMBRE_CRISTAUX, titre_ssmenu_cristal, CORR_ENERGIE, CE_CRISTAL);
    if (cristal >= 0) {
      OBJ_MettreNombre( CORR_ENERGIE, COR_NRJ_DIST, cristaux[cristal].d_hkl, 0.0 );
      met_chaine(CORR_ENERGIE, CE_CRISTAL, LGMAX_NOM_CRISTAL, noms_cristaux[cristal], TXT_COPIER);
    }
    return BI_EVENEMENT_TRAITE;

  case COR_NRJ_DECALE:
    OBJ_choisit(CORR_ENERGIE, COR_NRJ_DECALE, FORCE_OUI);
    OBJ_active(CORR_ENERGIE, COR_NRJ_DIST, FORCE_NON);
    OBJ_active(CORR_ENERGIE, CE_CRISTAL, FORCE_NON);
    return BI_EVENEMENT_TRAITE;

  case COR_NRJ_BRAGG:
    OBJ_choisit(CORR_ENERGIE, COR_NRJ_BRAGG, FORCE_OUI);
    OBJ_active(CORR_ENERGIE, COR_NRJ_DIST, FORCE_OUI);
    OBJ_active(CORR_ENERGIE, CE_CRISTAL, FORCE_OUI);
    return BI_EVENEMENT_TRAITE;
  }

#ifdef _DEVELOPPE_
  printf("GER_bragg : objet = %u, x = %d, y = %d\n", objet, x, y);
#endif
  return BI_EVENEMENT_NON_TRAITE;
}

static void prepare_arbre_bragg(void)
{
  OBJ_deplace(CORR_ENERGIE, SEUIL_ATTENDU, 0, +1, DEPLACE_PIXELS);
  OBJ_MettreNombre( CORR_ENERGIE, COR_NRJ_DIST, cristaux[SILICIUM_111].d_hkl, 0.0 );
  met_chaine(CORR_ENERGIE, CE_CRISTAL, LGMAX_NOM_CRISTAL, noms_cristaux[SILICIUM_111], TXT_COPIER);

  arbres[CORR_ENERGIE].fonction.arbre = GER_bragg;
  arbres[CORR_ENERGIE].drapeaux.prepare = VRAI;
}
